#! /bin/sh
#
# script to setup pipeline installation and sample reduction blocks
# after unpacking of the distribution.
#
# Usage:
#        setup.sh


HERE=`pwd`
HERE_=`echo $HERE | sed 's/\//\\\\\//g'`
PIPE_VERSION=fors1-1.8.2
TEMPFILE=tempfile
TEMPDIR=tempdir

PIPE_DIR=${HERE}/pipeline
PIPE_FILE=${PIPE_DIR}/${PIPE_VERSION}.tar.gz
INST_PREFIX_DIR=${PIPE_DIR}/${PIPE_VERSION}

REC_DIR=${HERE}/pipeline/rec/
RUL_DIR=${HERE}/pipeline/rul/

REDBLOCK_DIR_1=${HERE}/examples/fors1/redblock
REDBLOCK_DIR_2=${HERE}/examples/fors2/redblock

MID_STARTUP_1=${REDBLOCK_DIR_1}/pipe.prg
MID_STARTUP_2=${REDBLOCK_DIR_2}/pipe.prg

echo
echo 'This script installs the FORS1+2 pipeline and'
echo 'configures the sample reduction blocks for immediate use.'
echo
echo


echo '1. Verifying correct MIDAS installation'
echo '---------------------------------------'
echo
echo

if [ -z "$MIDASHOME" ]; then
   echo "Environment variable \$MIDASHOME is not defined. Abort."
   exit 1
fi

if [ -z "$MIDVERS" ]; then
   echo "Environment variable \$MIDVERS is not defined. Abort."
   exit 1
fi

MIDASV1=`echo $MIDVERS | cut -c1-5`
MIDASV2=`echo $MIDVERS | cut -c1-2`
if [ "$MIDASV1" != "01SEP" -a "$MIDASV2" != "02" ]; then
  echo "MIDAS version is older than 01SEP."
  echo "Please install MIDAS release 01SEP or later. EXIT."
  exit 1
fi

if [ ! -d "$MIDASHOME" ]; then
  echo "${MIDASHOME}: no such directory."
  echo "Call your MIDAS operator. EXIT."
  exit 1
fi

if [ ! -d "${MIDASHOME}/${MIDVERS}" -a ! -h "${MIDASHOME}/${MIDVERS}" ]; then
  echo "${MIDASHOME}/${MIDVERS}: no such directory."
  echo "Call your MIDAS operator. EXIT."
  exit 1
fi


echo '2. Installation of the FORS1+FORS2 pipeline'
echo '-------------------------------------------'
echo
echo

if [ ! -d $TEMPDIR ]; then
  mkdir $TEMPDIR
fi

( cd $TEMPDIR
  zcat $PIPE_FILE | tar xvf - > /dev/null
  cd $PIPE_VERSION
  ./setup $INST_PREFIX_DIR
)

if [ $? != 0 ]
then
  echo 'Pipeline installation failed. Exiting'
  exit 2
fi

( cd $PIPE_DIR
  ln -s $PIPE_VERSION fors1
  ln -s $PIPE_VERSION fors2
)

rm -rf $TEMPDIR

echo
echo done.
echo
echo
echo '3. Configuring the reduction blocks'
echo '-----------------------------------'
echo
echo

( cd $REDBLOCK_DIR_1
  for file in *.rb
  do
    sed "s/@@PIPE_DIR@@/$HERE_/" $file > $TEMPFILE
    mv $TEMPFILE $file
  done
)

( cd $REDBLOCK_DIR_2
  for file in *.rb
  do
    sed "s/@@PIPE_DIR@@/$HERE_/" $file > $TEMPFILE
    mv $TEMPFILE $file
  done
)


echo done.
echo
echo
echo '4. Creating a MIDAS startup file'
echo '--------------------------------'
echo
echo

cat > $MID_STARTUP_1 <<EOF
!
! MIDAS startup file to define settings for the FORS pipeline.
! The RBS context is then enabled and the pipeline with displays
! started.
!
! 
! Usage:
!  - type '@@ pipe.prg' in a running MIDAS session
!    or
!  - type 'inmidas -j "@@ pipe.prg"' in a UNIX shell
!
!
! Available commands:
!  - start/pipeline Display/NoDisplay 
!                       to start the pipeline with/without Display
!  - execute/RB rbfile  to execute given RB
!  - RBS/help           to get help about this context
!
!

EOF

echo 'write/key CALIBDB_RUL/c/1/100  "'$RUL_DIR'"' >> $MID_STARTUP_1
echo 'write/key CALIBDB_REC/c/1/100  "'$REC_DIR'"' >> $MID_STARTUP_1

cat >> $MID_STARTUP_1 <<EOF

set/context RBS >Null
start/pipeline Display >Null
pipeline(11) = 1

WRITE/OUT " "
WRITE/OUT "Enable/disable displays using the command"
WRITE/OUT "  START/PIPELINE Display/Nodisplay (currently enabled)"
WRITE/OUT " "
WRITE/OUT "Execute a reduction block using the command"
WRITE/OUT "  EXECUTE/RB <rbfile>"
WRITE/OUT " "

EOF

cp $MID_STARTUP_1 $MID_STARTUP_2

echo done.
echo 
echo 
echo 'Configuration sucessfully finished.'
echo
echo 'To use the FORS pipeline now you have to do the following:'
echo
echo ' - Define the following system variable:'
echo '   % setenv PIPE_HOME '$PIPE_DIR
echo '   (for the csh, accordingly for other shells,'
echo '    you can also put this into your .cshrc, .profile, etc.)'
echo 
echo ' - To execute one of the FORS1 sample reduction blocks:'
echo '   % cd '$REDBLOCK_DIR_1
echo '   % inmidas -j "@@ pipe.prg"'
echo '   Midas 001> EXECUTE/RB <rbfile>'
echo '   (accordingly for FORS2).'
echo

exit 0
