/* $Id: rebin.c,v 1.2 1999/03/15 09:09:21 rpalsa Exp $   
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 *.IDENT       $RCSfile: rebin.c,v $
 *
 *.KEYWORDS    Driver for specific rebinning functions
 *
 *.USAGE       int rebin(BinData *sp, BinData *rp, Method rbinmtd)
 *
 *.INPUT       sp      Binned input data set
 *             rp      Output data set. The new data grid is taken from
 *                     this structure
 *             Method  Reference to the actual rebinning function.
 *
 *.OUTPUT      None
 *
 *.RETURNS     The function returns the exit status. The return value
 *             is -1 if the input and output data grids do not overlap,
 *             -2 if one of the data grids is not sorted in ascending order
 *             and -3 if one of the data grids is not continuous. In all
 *             other cases the function returns the exit status of the
 *             called rebinning function.
 *
 *.PURPOSE     Check input and output data grid if they meet all basic
 *             requirements (see above). Call selected rebinning method.
 *
 *.ENVIRON     None
 *
 *.LANGUAGE    ANSI C/KR C
 *
 *.AUTHOR      R. Palsa, Univ.-Sternwarte Muenchen
 *
 *.VERSION     0.1 1999/02/28: Creation (RP)
 *
 *----------------------------------------------------------------------------
 */


#include <math.h>
#include <rebin.h>


#define TINY  1.e-10



#ifdef __STDC__
int rebin(BinData *sp, BinData *rp, Method rbinmtd)
#else
int rebin(sp, rp, rbinmtd)
     BinData *sp, *rp;
     Method rbinmtd;
#endif
{

  int i, state;

  double smin, smax, rmin, rmax;
  double low, high;


  
  /*
   * Check if wavelength regions overlap
   */

  smin = sp->bval[0] - 0.5 * sp->bwidth[0];
  smax = sp->bval[sp->nbin - 1] + 0.5 * sp->bwidth[sp->nbin - 1];

  rmin = rp->bval[0] - 0.5 * rp->bwidth[0];
  rmax = rp->bval[rp->nbin - 1] + 0.5 * rp->bwidth[rp->nbin - 1];

  if (smax < rmin || smin > rmax)
    return -1;


  /*
   * Check if data grids are continuous and sorted in ascending order.
   */

  for (i = 0; i < sp->nbin - 1; i++) {
    if (sp->bval[i] >= sp->bval[i + 1])
      return -2;

    low = sp->bval[i + 1] - 0.5 * sp->bwidth[i + 1];
    high = sp->bval[i] + 0.5 * sp->bwidth[i];

    if (fabs(high - low) > TINY)
      return -3;
  }

  for (i = 0; i < rp->nbin - 1; i++) {
    if (rp->bval[i] >= rp->bval[i + 1])
      return -2;

    low = rp->bval[i + 1] - 0.5 * rp->bwidth[i + 1];
    high = rp->bval[i] + 0.5 * rp->bwidth[i];

    if (fabs(high - low) > TINY)
      return -3;
  }


  /*
   * Rebinn spectrum using selected method
   */

  state = rbinmtd(sp->bval, sp->bwidth, sp->data, sp->nbin, rp->bval,
		  rp->bwidth, rp->data, rp->nbin);

  return state;

}
