/* $Id: reblin.c,v 1.1 1999/03/04 12:17:57 rpalsa Exp $   
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *.IDENT       $RCSfile: reblin.c,v $
 *
 *.KEYWORDS    Rebinning of 1d data using linear interpolation
 *
 *.USAGE       int reblin(double *x1, double *w1, double *y1, int n1,
 *                        double *x2, double *w2, double *y2, int n2)
 *
 *.INPUT       x1      Central values of input bins
 *             w1      Width of input bins
 *             y1      Integral value of bin normalized to the bin width.
 *             n1      Number of bins in input grid
 *             x2      Central values of output bins
 *             w2      Width of output bins
 *             n2      Number of bins in output grid
 *
 *.OUTPUT      y2      Integral value of the output bin normalized to the
 *                     output bin width.
 *
 *.RETURNS     The function returns 0 if the rebinning was successful,
 *             otherwise the return value is 1.
 *
 *.PURPOSE     Rebin the binned input data set to the new grid structure
 *             using linear interpolation for input bins that do not fall
 *             completely in a single bin of the new grid.
 *             The total flux may not be conserved!
 *
 *             Note:  This method has been implemented since it is used by
 *                    the INTEGRATE/LONG and INTEGRATE/MOS commands defined
 *                    in the MIDAS contexts LONG and MOS.
 *
 *.ENVIRON     None
 *
 *.LANGUAGE    ANSI C/KR C
 *
 *.AUTHOR      R. Palsa, (ESO-DMO)
 *
 *.VERSION     0.1 1999/02/28: Derived from rebpix (RP)
 *
 *----------------------------------------------------------------------------
 */


#include <stdlib.h>


#ifdef __STDC__
unsigned long locate(double, double *, unsigned long);
#else
unsigned long locate();
#endif


#ifdef __STDC__
double linpol(double, double, double, double, double);
#else
double linpol();
#endif



#ifdef __STDC__
int reblin(double *x1, double *w1, double *y1, int n1, double *x2,
	    double *w2, double *y2, int n2)
#else
int reblin(x1, w1, y1, n1, x2, w2, y2, n2)
     double *x1, *w1, *y1, *x2, *w2, *y2;
     int n1, n2;
#endif
{

  int i, k, l, kmin, kmax, lmin, lmax;

  double *b1, *b2, *y;
  double ylow, yhigh;


  /*
   * Initialize rebinned spectrum
   */

  for (i = 0; i < n2; i++)
    y2[i] = 0.;


  /*
   * Create arrays of bin boundary values for both data grids
   * for the first one we do a quasi refinement by using the
   * bin centers as additional boundaries.
   */

  if ((b1 = (double *)osmmget(( 2 * n1 + 1) * sizeof(double))) == NULL)
    return 1;
  else {
    for (i = 0; i < n1; i++) {
      b1[2 * i] = x1[i] - 0.5 * w1[i];
      b1[2 * i + 1] = x1[i];
    }
    b1[2 * n1] = x1[n1 - 1] + 0.5 * w1[n1 - 1];
  }


  if ((b2 = (double *)osmmget((n2 + 1) * sizeof(double))) == NULL) {
    osmmfree(b1);
    return 1;
  }
  else {
    for (i = 0; i < n2; i++)
      b2[i] = x2[i] - 0.5 * w2[i];
    b2[n2] = x2[n2 - 1] + 0.5 * w2[n2 - 1];
  }


  /*
   * Create array for remapping the input data to the input
   * boundary values by linear interpolation.
   */

  if ((y = (double *)osmmget((2 * n1 + 1) * sizeof(double))) == NULL) {
    osmmfree(b1);
    osmmfree(b2);
    return 1;
  }
  else {
    y[1] = y1[0];

    for (i = 1; i < n1; i++) {
      y[2 * i] = linpol(x1[i - 1], y1[i - 1], x1[i], y1[i], b1[2 * i]);
      y[2 * i + 1] = y1[i];
    }


    /*
     * Construct function values at first and last bin boundary
     * so that the integral over the first and the last bin is
     * conserved.
     */

    y[0] = 2. * y[1] - y[2];
    y[2 * n1] = 2. * y[2 * n1 - 1] - y[2 * n1 - 2];

    n1 *= 2;
  }


  /*
   * Determine index range of the new grid where the new and the old
   * grid overlap. The indices kmin and kmax are the first and the last
   * bin values of the new grid which must be located in the old one.
   * In the case that the new grid exceeds the lower or upper limit of
   * the old grid the first and last bin of the new grid are here
   * computed separately.
   */

  if (b2[0] < b1[0]) {
    if ((kmin = locate(b1[0], b2, n2 + 1)) < 0) {
      osmmfree(b1);
      osmmfree(b2);
      osmmfree(y);
      return 1;
    }
    else {
      if ((lmax = locate(b2[kmin + 1], b1, n1 + 1)) == -2)
	lmax = n1;

      for (l = 0; l < lmax / 2; l++)
	y2[kmin] += y1[l] * w1[l];
      y2[kmin] /= w2[kmin];

      if (lmax == n1) {
	osmmfree(b1);
	osmmfree(b2);
	osmmfree(y);
	return 0;
      }
      else {
	if (lmax % 2 == 1)
	  y2[kmin] += 0.5 * (y[lmax - 1] + y[lmax]) * 
                     (b1[lmax] - b1[lmax - 1]) / w2[kmin];

	yhigh = linpol(b1[lmax], y[lmax], b1[lmax + 1], y[lmax + 1],
		       b2[kmin + 1]);
	y2[kmin] += 0.5 * (y[lmax] + yhigh) * (b2[kmin + 1] - b1[lmax]) /
	            w2[kmin];
	kmin++;
      }
    }
  }
  else 
    kmin = 0;

  if (b2[n2] > b1[n1]) {
    if ((kmax = locate(b1[n1], b2, n2 + 1)) < 0) {
      osmmfree(b1);
      osmmfree(b2);
      osmmfree(y);
      return 1;
    }
    else {
      lmin = locate(b2[kmax], b1, n1 + 1);
      ylow = linpol(b1[lmin], y[lmin], b1[lmin + 1], y[lmin + 1], b2[kmax]);
      y2[kmax] = 0.5 * (ylow + y[lmin + 1]) * (b1[lmin + 1] - b2[kmax]);

      if (lmin % 2 == 0) {
	y2[kmax] += 0.5 * (y[lmin + 1] + y[lmin + 2]) * 
                    (b1[lmin + 2] - b1[lmin + 1]);
	lmin++;
      }

      for (l = (lmin + 1) / 2; l < n1 / 2; l++)
	y2[kmax] += y1[l] * w1[l];

      y2[kmax] /= w2[kmax];
    }
  }
  else 
    kmax = n2;


  /*
   * Integrate tabulated function and normalize the result to the
   * new bin width.
   */

  for (k = kmin; k < kmax; k++) {
    lmin = locate(b2[k], b1, n1 + 1);
    lmax = locate(b2[k + 1], b1, n1 + 1);
	
    ylow = linpol(b1[lmin], y[lmin], b1[lmin + 1], y[lmin + 1], b2[k]);
    yhigh = linpol(b1[lmax], y[lmax], b1[lmax + 1], y[lmax + 1], b2[k + 1]);

    if (lmin == lmax)
      y2[k] = 0.5 * (ylow + yhigh);
    else {
      y2[k] = 0.5 * (ylow + y[lmin + 1]) * (b1[lmin + 1] - b2[k]);
      y2[k] += 0.5 * (y[lmax] + yhigh) * (b2[k + 1] - b1[lmax]);


      /*
       * Account for bins of the original grid which are not
       * completely within the integration boundaries.
       */

      if (lmin % 2 == 0) {
	y2[k] += 0.5 * (y[lmin + 1] + y[lmin + 2]) * 
                 (b1[lmin + 2] - b1[lmin + 1]);
	lmin++;
      }

      if (lmax % 2 == 1) {
	y2[k] += 0.5 * (y[lmax - 1] + y[lmax]) * (b1[lmax] - b1[lmax - 1]);
        lmax--;
      }

      for (l = (lmin + 1) / 2; l < lmax / 2; l++)
	y2[k] += y1[l] * w1[l];

      y2[k] /= w2[k];
    }
  }

  osmmfree(b1);
  osmmfree(b2);
  osmmfree(y);

  return 0;

}
