/* $Id: rebpix.c,v 1.1 1999/03/04 12:07:15 rpalsa Exp $   
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *.IDENT       $RCSfile: rebpix.c,v $
 *
 *.KEYWORDS    Rebinning of 1d data using bin fractions
 *
 *.USAGE       int rebpix(double *x1, double *w1, double *y1, int n1,
 *                        double *x2, double *w2, double *y2, int n2)
 *
 *.INPUT       x1      Central values of input bins
 *             w1      Width of input bins
 *             y1      Integral value of bin normalized to the bin width.
 *             n1      Number of bins in input grid
 *             x2      Central values of output bins
 *             w2      Width of output bins
 *             n2      Number of bins in output grid
 *
 *.OUTPUT      y2      Integral value of the output bin normalized to the
 *                     output bin width.
 *
 *.RETURNS     The function returns 0 if the rebinning was successful,
 *             otherwise the return value is 1.
 *
 *.PURPOSE     Rebin the binned input data set to the new grid structure
 *             using fractions of bin. The total flux is conserved.
 *
 *.ENVIRON     None
 *
 *.LANGUAGE    ANSI C/KR C
 *
 *.AUTHOR      R. Palsa, (ESO-DMO)
 *
 *.VERSION     0.1 1999/02/28: Derived from the original version of
 *                             R. Palsa (Univ.-Sternwarte Muenchen).
 *
 *----------------------------------------------------------------------------
 */


#include <stdlib.h>


#ifdef __STDC__
unsigned long locate(double, double *, unsigned long);
#else
unsigned long locate();
#endif



#ifdef __STDC__
int rebpix(double *x1, double *w1, double *y1, int n1, double *x2,
	    double *w2, double *y2, int n2)
#else
int rebpix(x1, w1, y1, n1, x2, w2, y2, n2)
     double *x1, *w1, *y1, *x2, *w2, *y2;
     int n1, n2;
#endif
{

  int i, k, l, kmin, kmax, lmin, lmax;

  double *b1, *b2;



  /*
   * Initialize rebinned spectrum
   */

  for (i = 0; i < n2; i++)
    y2[i] = 0.;


  /*
   * Create arrays of bin boundary values for both data grids
   */

  if ((b1 = (double *)osmmget((n1 + 1) * sizeof(double))) == NULL)
    return 1;
  else {
    for (i = 0; i < n1; i++)
      b1[i] = x1[i] - 0.5 * w1[i];
    b1[n1] = x1[n1 - 1] + 0.5 * w1[n1 - 1];
  }


  if ((b2 = (double *)osmmget((n2 + 1) * sizeof(double))) == NULL) {
    osmmfree(b1);
    return 1;
  }
  else {
    for (i = 0; i < n2; i++)
      b2[i] = x2[i] - 0.5 * w2[i];
    b2[n2] = x2[n2 - 1] + 0.5 * w2[n2 - 1];
  }


  /*
   * Account for parts of the output grid that exceeds the input grid
   * boundaries. Also the special case that the whole input grid falls
   * into a single bin of the new grid is handled.
   * The indices kmin and kmax are the first and the last bin values
   * of the new grid which must be located in the old one.
   */

  if (b2[0] < b1[0]) {
    if ((kmin = locate(b1[0], b2, n2 + 1)) < 0) {
      osmmfree(b1);
      osmmfree(b2);
      return 1;
    }
    else {
      if ((lmax = locate(b2[kmin + 1], b1, n1 + 1)) == -2)
	lmax = n1;

      for (l = 0; l < lmax; l++)
	y2[kmin] += y1[l] * w1[l];
      y2[kmin] /= w2[kmin];

      if (lmax == n1) {
	osmmfree(b1);
	osmmfree(b2);
	return 0;
      }
      else {
	y2[kmin] += y1[lmax] * (b2[kmin + 1] - b1[lmax]) / w2[kmin];
	kmin++;
      }
    }
  }
  else 
    kmin = 0;

  if (b2[n2] > b1[n1]) {
    if ((kmax = locate(b1[n1], b2, n2 + 1)) < 0) {
      osmmfree(b1);
      osmmfree(b2);
      return 1;
    }
    else {
      lmin = locate(b2[kmax], b1, n1 + 1);
      y2[kmax] = y1[lmin] * (b1[lmin + 1] - b2[kmax]);

      for (l = lmin + 1; l < n1; l++)
	y2[kmax] += y1[l] * w1[l];

      y2[kmax] /= w2[kmax];
    }
  }
  else 
    kmax = n2;


  /*
   * Integrate tabulated function and normalize the result to the
   * new bin width.
   */

  for (k = kmin; k < kmax; k++) {
    lmin = locate(b2[k], b1, n1 + 1);
    lmax = locate(b2[k + 1], b1, n1 + 1);
	
    if (lmin == lmax)
      y2[k] = y1[lmin];
    else {
      y2[k] = y1[lmin] * (b1[lmin + 1] - b2[k]);

      for (l = lmin + 1; l < lmax; l++)
	y2[k] += y1[l] * w1[l];

      y2[k] += y1[lmax] * (b2[k + 1] - b1[lmax]);
      y2[k] /= w2[k];
    }
  }

  osmmfree(b1);
  osmmfree(b2);

  return 0;

}
