! $Id: fors_calcoeff.prg,v 1.63 2001/04/06 16:19:25 sbogun Exp $
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
! .COPYRIGHT  (C) 2000  European Southern Observatory (ESO)
!
! .IDENT      fors_calcoeff.prg
!
! .KEYWORDS   Calibration coefficients
!
! .PURPOSE    Computation of magnitude zero point, extinction coefficient
!             and colour coefficient for each filter.
!             The different regression steps can be plotted - the plots
!             are saved in "{OUT_B}".
! __________
!
! .USAGE      P1(ncoef)
!             =0:  @p (IDENT) [ncoef]     alicat       [labels(18)] -
!                             [thresh(5)] ?            ?            [disp] [re]
!
!             =1:  @p (IDENT) [ncoef]     alicat       [labels(18)] -
!                             [thresh(8)] coltabs(6|5) ?            [disp] [re]
!
!             =2:  @p (IDENT) [ncoef]     alicat       [labels(17)] -
!                             ?           coltabs(6|5) exttabs(6|5) [disp] [re]
!
! .INPUT      ncoef   (P1): number of different types of input tables with
!                           calibration coefficients:
!                             |      => per filter:
!                             V      zero_point  extinction.coef.  colour.coef.
!                                 + - - - - - - - - - - - - - - - - - - - - - -
!                            <0>  |  calculated  calculated        calculated
!                            <1>  |  calculated  calculated        input_tables
!                            <2>  |  calculated  input_tables      input_tables
!                             default: <0>
!
!             alicat  (P2): image catalogue with aligned photometry tables of
!                           standard stars, at least for one colour from
!                           "UBVRI"(Bessel);
!                             default: none
!
!             labels  (P3): If P1(ncoef)=[0|1|2]:
!                           --------------------
!                             names of columns in the ali.phot.tables with
!                                1.     star_names
!                                2.- 6. measured_mags.in_U/B/V/R/I
!                                7.     airmasses
!                                8.     modified julian dates
!                                9.     object quality flags
!                               10.-14. standard_mags.in_U/B/V/R/I
!                             and colours to be calculated from the std.mags.
!                               15.-19. for the filters  U/B/V/R/I
!                           and if P1(ncoef)=[0|1] (else ignored):
!                           -------------------------------------
!                             name of column in the ali.phot.tables with
!                               20.     number of measurements which had been
!                                       used to derive the std.mags.,
!                           all separated by commas;
!                           from columns with measured mags. only used one must
!                           be present in each table - else the table will be
!                           skipped;
!                             default:OBJECT,
!                                     MAG_U,MAG_B,MAG_V,MAG_R,MAG_I,
!                                     AIRMASS,
!                                     MJD,
!                                     FLAGS,
!                                     U,B,V,R,I,
!                                     U-B,B-V,B-V,V-R,V-R,
!                                     N
!
!             thresh  (P4): If P1(ncoef)=[0|1] (else ignored):
!                           ---------------------------------
!                             1. minimum number of measurements which should
!                                have been used to derive the std.mags.
!                             2. binsize for determination of number of diffe-
!                                rent airmasses, >= 10^(-AIRACC) (see below)
!                             3. minimum number of different airmasses, >= 2
!                             4. minimum range of airmasses
!                             5. minimum number of stars within the thresholds
!                           and if P1(ncoef)=[0] (else ignored):
!                           -----------------------------------
!                             6. binsize for determination of number of diffe-
!                                rent colours, >= 10^(-COLACC) (see below)
!                             7. minimum number of different colours, >= 5
!                             8. minimum range of colours
!                           all separated by commas;
!                             default:5,0.001,2,0.5,3,0.001,5,1.0
!
!             coltabs (P5): If P1(ncoef)=[1|2] (else ignored):
!                           ---------------------------------
!                             tables with colour coefficients for the filters
!                             U,B,V,R,I -for filters which should not be used
!                             either a "?" or a "+" can be entered-
!                             AND number of column with the coef. (must be the
!                             same for each table),
!                           separated by commas;
!                             default:?,?,?,?,?,1
!
!             exttabs (P6): If P1(ncoef)=[2] (else ignored):
!                           -------------------------------
!                             tables with extinction coefficients for the
!                             filters U,B,V,R,I -for filters which should not
!                             be used either a "?" or a "+" can be entered-
!                             AND number of column with the coef. (must be the
!                             same for each table),
!                           separated by commas;
!                             default:?,?,?,?,?,1
!
!             disp    (P7): display selection: <p>:pipeline mode:read global
!                           key NICE_CREA [y|n], <n>:none, <l>:local, or
!                           host:display[.screen] - with the latter option
!                           you can specify the name or IP-address of a
!                           display;
!                             default: <l>
!
!             re      (P8): flag for re-calculation of calibration coefficients
!                           with different thresholds P4(thresh): <n>o, <y>es -
!                           can be used if this procedure has already created
!                           at least one merged table (see text output) to skip
!                           all creations in a further run (saves time) -
!                           P1(ncoef) must not change between the runs;
!                             default: <n>
!
! .OUTPUT     OUT_A: String with names of output catalogues which contain all
!                    output tables; the number of output catalogues depends
!                    on P1(ncoef) according to the following scheme:
!
!                          |   P2(alicat)_//
!                          |   //zero.cat  //kext.cat  //kcol.cat
!             - - - - - - - - - - - - - - - - - - - - - - - - - -
!             P1(ncoef)=0  |        x           x           x
!                       1  |        x           x
!                       2  |        x
!
!             If ( P7(disp)="p" and nice_crea!="n" )
!             or ( P7(disp)!="p" and P7(disp)!="n" ):
!     
!             OUT_B: Name of saved plot file
!
! .RETURN     Q1:  0: successful return
!                 -1: a warning message occured
!                  1: an error occured
! __________
!
! .ALGORITHM  **************************************************
!             * THE CODE CAN BE BROWSED BY SEARCHING FOR "%" ! *
!             **************************************************
!
!             1   a  Start of code
!                 b  Tests of parameters
!                 c  Definition of further variables
!                 d  Managing displays
!
!             2   LOOP WITH ALIGNED PHOTOMETRY TABLES:
!                 a  Determination of used filter
!                 b  Merging tables per filter
!
!             3   LOOP WITH MERGED TABLES:
!                 a  Calculation of colours
!                 b  Calculation of standard minus instrument magnitudes
!                 c  Selection of high quality objects (column FLAGS .ne. "Y")
!                 d  Extinction coefficient:
!                    P1(ncoef)=[0|1]:
!                    ---------------
!                    - Checking extinction thresholds / indexing of good obj.
!                    - Calculation of extinction coef.s / zero points without
!                      colour correction per good object
!                    - Calc. of mean ext.coef.
!                    P1(ncoef)=[2]:
!                    -------------
!                    - Getting ext.coeff. from input
!                    P1(ncoef)=[0|1]:
!                    ---------------
!                    - Re-calc. of zero points without colour correction with
!                      mean ext.coef.
!                    P1(ncoef)=[2]:
!                    -------------
!                    - Calc. of zero points without colour correction with
!                      input ext.coef.
!                 e  Colour coefficient / zero point:
!                    P1(ncoef)=[0]:
!                    -------------
!                    - Selection of good objects / checking colour thresholds
!                    - Calc. of colour coef. / zero point with colour correc.
!                      with all good objects
!                    P1(ncoef)=[1]:
!                    -------------
!                    - Calc. of zero point with colour correction with input
!                      colour coef. with all good objects
!                    P1(ncoef)=[2]:
!                    -------------
!                    - Calc. of zero point with colour correction with input
!                      colour coef. with all objects
!                    P1(ncoef)=[0|1|2]:
!                    -----------------
!                    - Calc. of all single zero points with extinction
!                      and colour correction
!
!             4a  Displaying of results
!             4b  Creation of output tables / catalogues
!
!             USED FORMULA:    /U\                       /col=P3(labels)[13]\
!                             | B |                     | col=P3(labels)[14] |
!             mag_lit -mag_mea| V | = z +k_ext*X +k_col*| col=P3(labels)[15] |
!                             | R |   \__________/      | col=P3(labels)[16] |
!                              \I/              = z_col  \col=P3(labels)[17]/
!
! .COMMENT    With the entered column labels P3(labels)[2-6] (measured magni-
!             tudes) it is checked whether exactly one of these columns is
!             present per aligned photometry table. Then the last character of
!             the name of the found column determines the filter of this table
!             (e.g. "MAG_V" is recognized as filter "V") !
!
!             Warning:
!             If ( P7(disp)="p" and nice_crea!="n" )
!             or ( P7(disp)!="p" and P7(disp)!="n" ),
!             a fixed-sized graphic window {FIXGRAPHNO} (see below) is created !
!             Technical reason: No normalized graphic coordinates are present
!             up to Midas version 99NOVpl1.1 - they exist from pl2.0 on.
!
!             Used UNIX commands:
!             $ /bin/echo (###A as long as mess/out is still too slow)
!
! .ENVIRON    -
!
! .LANGUAGE   MIDAS 99NOV
!
! .AUTHOR     N.Rainer  DPG/DMD/ESO
!
! .VERSION    1.5  00/01/19  Creation
!                  00/       Release
!
! .TODO       ###A /bin/echo -> mess/out, when mess/out in monitor !
!             ###B missing errors for reg/pol
!
!------------------------------------------------------------------------------
!
!
!     %1a  START OF CODE :
!          *************
!
crossref ncoef alicat labels thresh coltabs exttabs disp re
!
!
! "p_dspmode" set in "fors_nice.prg" - for fixed-sized graphic window:
if m$existk("p_dspmode") .eq. 0 define/local p_dspmode/i/1/1 {ididev(7)}
! for create/graphic: G.C.[3]=(1592-G.C.[1])/2 ; G.C.[4]=(1171-G.C.[2])/2+27
define/local GRAPHCOORD/c/1/19 "744,480,424,372"
define/local FIXGRAPHNO/i/1/1 8 ? +low
define/local PLOT_FILE/c/1/40 " " all
!
define/local TOTFILT/i/1/1 5 ? +low
define/local ALLFILT/c/1/{TOTFILT} "UBVRI" ? +low  ! fixed to these filters
if m$existk("AIRACC") .eq. 0 then
  define/local AIRACC/i/1/1 3                  ! #deci.f.airmass column
endif
if m$existk("COLACC") .eq. 0 then
  define/local COLACC/i/1/1 3 ? +low           ! #deci.f.col.column / col.coef.
endif
define/local EXTACC/i/1/1 3                    ! #deci.f.ext.coef.
!
!
!  For `set/format':
define/local FMTDEF/c/1/17 "I1 E14.7,E23.15"
!
!  For `mess/out':
define/local PROID/c/1/40 "calcoeff"
if m$existk("mess_level") .eq. 0 write/keyword mess_level/i/1/1 3
define/local line/c/1/51 "- " all
!
!  For `longlog/out':
write/keyword out_a "given command/qualif not defined ..."
show/command longlog/out | write/keyword in_a
-delete Mid{mid$sess(11:12)}1Pipe
-delete midlis{mid$sess(11:12)}.prg
if in_a .eq. out_a create/command longlog/out "write/out"
!
!  For return value:
define/local warn/i/1/1 0
write/keyword q1 0
!
!
set/format {FMTDEF}
!
!
if p1 .eq. "??" then
  w/o "      ******************************************************************"
  w/o "      ncoef   (P1): number of different types of input tables with"
  w/o "                    calibration coefficients:"
  w/o "                      |      => per filter:"
  w/o "                      V      zero_point  extinction.coef.  colour.coef."
  w/o "                          + - - - - - - - - - - - - - - - - - - - - - -"
  w/o "                     <0>  |  calculated  calculated        calculated"
  w/o "                     <1>  |  calculated  calculated        input_tables"
  w/o "                     <2>  |  calculated  input_tables      input_tables"
  w/o "                      default: <0>"
  w/o " "
  w/o "      alicat  (P2): image catalogue with aligned photometry tables of"
  w/o "                    standard stars, at least for one colour from"
  w/o "                    "UBVRI"(Bessel);"
  w/o "                      default: none"
  w/o " "
  w/o "      labels  (P3): If P1(ncoef)=[0|1|2]:"
  w/o "                    --------------------"
  w/o "                      names of columns in the ali.phot.tables with"
  w/o "                         1.     star_names"
  w/o "                         2.- 6. measured_mags.in_U/B/V/R/I"
  w/o "                         7.     airmasses"
  w/o "                         8.     modified julian dates"
  w/o "                         9.     object quality flags"
  w/o "                        10.-14. standard_mags.in_U/B/V/R/I"
  w/o "                      and colours to be calculated from the std.mags."
  w/o "                        15.-19. for the filters  U/B/V/R/I"
  w/o "                    and if P1(ncoef)=[0|1] (else ignored):"
  w/o "                    -------------------------------------"
  w/o "                      name of column in the ali.phot.tables with"
  w/o "                        20.     number of measurements which had been"
  w/o "                                used to derive the std.mags.,"
  w/o "                    all separated by commas;"
  w/o "                    from columns with measured mags. only used one must"
  w/o "                    be present in each table - else the table will be"
  w/o "                    skipped;"
  w/o "                      default:OBJECT,"
  w/o "                              MAG_U,MAG_B,MAG_V,MAG_R,MAG_I,"
  w/o "                              AIRMASS,"
  w/o "                              MJD,"
  w/o "                              FLAGS,"
  w/o "                              U,B,V,R,I,"
  w/o "                              U-B,B-V,B-V,V-R,V-R,"
  w/o "                              N"
  w/o " "
  w/o "      thresh  (P4): If P1(ncoef)=[0|1] (else ignored):"
  w/o "                    ---------------------------------"
  w/o "                      1. minimum number of measurements which should"
  w/o "                         have been used to derive the std.mags."
  w/o "                      2. binsize for determination of number of diffe-"
  w/o "                         rent airmasses, >= 10^(-{AIRACC})"
  w/o "                      3. minimum number of different airmasses, >= 2"
  w/o "                      4. minimum range of airmasses"
  w/o "                      5. minimum number of stars within the thresholds"
  w/o "                    and if P1(ncoef)=[0] (else ignored):"
  w/o "                    -----------------------------------"
  w/o "                      6. binsize for determination of number of diffe-"
  w/o "                         rent colours, >= 10^(-{COLACC})"
  w/o "                      7. minimum number of different colours, >= 5"
  w/o "                      8. minimum range of colours"
  w/o "                    all separated by commas;"
  w/o "                      default:5,0.001,2,0.5,3,0.001,5,1.0"
  w/o " "
  w/o "      coltabs (P5): If P1(ncoef)=[1|2] (else ignored):"
  w/o "                    ---------------------------------"
  w/o "                      tables with colour coefficients for the filters"
  w/o "                      U,B,V,R,I -for filters which should not be used"
  w/o "                      either a "?" or a "+" can be entered-"
  w/o "                      AND number of column with the coef. (must be the"
  w/o "                      same for each table),"
  w/o "                    separated by commas;"
  w/o "                      default:?,?,?,?,?,1"
  w/o " "
  w/o "      exttabs (P6): If P1(ncoef)=[2] (else ignored):"
  w/o "                    -------------------------------"
  w/o "                      tables with extinction coefficients for the"
  w/o "                      filters U,B,V,R,I -for filters which should not"
  w/o "                      be used either a "?" or a "+" can be entered-"
  w/o "                      AND number of column with the coef. (must be the"
  w/o "                      same for each table),"
  w/o "                    separated by commas;"
  w/o "                      default:?,?,?,?,?,1"
  w/o " "
  w/o "      disp    (P7): display selection: <p>:pipeline mode:read global"
  w/o "                    key NICE_CREA [y|n], <n>:none, <l>:local, or"
  w/o "                    host:display[.screen] - with the latter option"
  w/o "                    you can specify the name or IP-address of a"
  w/o "                    display;"
  w/o "                      default: <l>"
  w/o " "
  w/o "      re      (P8): flag for re-calculation of calibration coefficients"
  w/o "                    with different thresholds P4(thresh): <n>o, <y>es -"
  w/o "                    can be used if this procedure has already created"
  w/o "                    at least one merged table (see text output) to skip"
  w/o "                    all creations in a further run (saves time) -"
  w/o "                    P1(ncoef) must not change between the runs;"
  w/o "                      default: <n>"
  w/o "      ******************************************************************"
  return/exit
endif
!
!
!     %1b  INPUT AND TESTS OF PARAMETERS :
!          *****************************
!
!P1(ncoef):
!---------
define/parameter p1 0 n -
  "Mode = no. of input coeffs. [def.:<0>|col=<1>|col+ext=<2>]: "
define/local ncoef/i/1/1 {p1} ? +low
if ncoef .lt. 0 .or. ncoef .gt. 2 then
        write/out " "
        mess/out 1 {PROID} "Entered no. of input coeffs. out of range !"
        mess/out 1 {PROID} "({p1})"
  goto error
endif
!
!
!P2(alicat):
!----------
if p2 .eq. "?" mess/out 2 {PROID} " "
define/parameter p2 ? t -
  "Icat. w. ali.phot.tbls.: "
define/local alicat/c/1/64 "{p2}" ? +low
if m$index(alicat,".cat") .eq. 0 alicat = alicat//".cat"
!
if m$exist(alicat) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered catalogue is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({alicat})"
  goto error
endif
!
show/icat {alicat} n
define/local catentries/i/1/1 {outputi}  ! needed below, too
if catentries .lt. 1 then
        write/out " "
        mess/out 1 {PROID} "Entered cat. with aligned phot. tables is empty !"
        mess/out 1 {PROID} "({alicat})"
  goto error
endif
!
define/local catal/i/1/1 0
inloop:
  store/frame inputc {alicat} 1 end_inloop
  if m$exist(inputc) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "At least one catalogue entry is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({alicat})"
    goto error
  endif
  goto inloop
end_inloop:
!
!
!P3(labels):
!----------
define/parameter p3 + ? -
  "[20|19] tbl.cols. (see procedure header): "
outputi = m$parse(p3,"ali")
if ncoef .le. 1 then
  inputi(1) = 20
  inputi(2) = 20
elseif ncoef .eq. 2 then
  inputi(1) = 19
  inputi(2) = 20  ! allowed, but unused
endif
if outputi .ne. inputi(1) .and. outputi .ne. inputi(2) then
  if outputi .ne. 1 then  ! 1 allowed for defaults
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P3 !"
    goto error
  endif
endif
if outputi .eq. inputi(1) .or. outputi .eq. inputi(2) then
  define/local badcol/i/1/1 0
  define/local cstar/c/1/16 "{ali01}"; if cstar(1:1) .eq. ":" cstar = cstar(2:)
  define/local cmeaU/c/1/16 "{ali02}"; if cmeaU(1:1) .eq. ":" cmeaU = cmeaU(2:)
  define/local cmeaB/c/1/16 "{ali03}"; if cmeaB(1:1) .eq. ":" cmeaB = cmeaB(2:)
  define/local cmeaV/c/1/16 "{ali04}"; if cmeaV(1:1) .eq. ":" cmeaV = cmeaV(2:)
  define/local cmeaR/c/1/16 "{ali05}"; if cmeaR(1:1) .eq. ":" cmeaR = cmeaR(2:)
  define/local cmeaI/c/1/16 "{ali06}"; if cmeaI(1:1) .eq. ":" cmeaI = cmeaI(2:)
  define/local cairm/c/1/16 "{ali07}"; if cairm(1:1) .eq. ":" cairm = cairm(2:)
  define/local cmjd/c/1/16 "{ali08}"; if cmjd(1:1) .eq. ":" cmjd = cmjd(2:)
  define/local cflag/c/1/16 "{ali09}"; if cflag(1:1) .eq. ":" cflag = cflag(2:)
  define/local cstdU/c/1/16 "{ali10}"; if cstdU(1:1) .eq. ":" cstdU = cstdU(2:)
  define/local cstdB/c/1/16 "{ali11}"; if cstdB(1:1) .eq. ":" cstdB = cstdB(2:)
  define/local cstdV/c/1/16 "{ali12}"; if cstdV(1:1) .eq. ":" cstdV = cstdV(2:)
  define/local cstdR/c/1/16 "{ali13}"; if cstdR(1:1) .eq. ":" cstdR = cstdR(2:)
  define/local cstdI/c/1/16 "{ali14}"; if cstdI(1:1) .eq. ":" cstdI = cstdI(2:)
  define/local colU/c/1/16 "{ali15}"; if colU(1:1) .eq. ":" colU = colU(2:)
  define/local colB/c/1/16 "{ali16}"; if colB(1:1) .eq. ":" colB = colB(2:)
  define/local colV/c/1/16 "{ali17}"; if colV(1:1) .eq. ":" colV = colV(2:)
  define/local colR/c/1/16 "{ali18}"; if colR(1:1) .eq. ":" colR = colR(2:)
  define/local colI/c/1/16 "{ali19}"; if colI(1:1) .eq. ":" colI = colI(2:)
  !
  if m$len(colU) .lt. 3 badcol = 1
  if m$len(colB) .lt. 3 badcol = 1
  if m$len(colV) .lt. 3 badcol = 1
  if m$len(colR) .lt. 3 badcol = 1
  if m$len(colI) .lt. 3 badcol = 1
  if badcol .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "At least one colour string (P3) is too short !"
        mess/out 1 {PROID} "(Characters.: 1.<filterX> 2.<-> 3.<filterY>)"
    goto error
  endif
  if m$index(ALLFILT,colU(1:1)) .eq. 0 .or. -
     m$index(ALLFILT,colU(3:3)) .eq. 0 badcol = 1
  if m$index(ALLFILT,colB(1:1)) .eq. 0 .or. -
     m$index(ALLFILT,colB(3:3)) .eq. 0 badcol = 1
  if m$index(ALLFILT,colV(1:1)) .eq. 0 .or. -
     m$index(ALLFILT,colV(3:3)) .eq. 0 badcol = 1
  if m$index(ALLFILT,colR(1:1)) .eq. 0 .or. -
     m$index(ALLFILT,colR(3:3)) .eq. 0 badcol = 1
  if m$index(ALLFILT,colI(1:1)) .eq. 0 .or. -
     m$index(ALLFILT,colI(3:3)) .eq. 0 badcol = 1
  if badcol .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "At least one colour string (P3) has a bad filter !"
        mess/out 1 {PROID} "(Only <{ALLFILT}> are possible)"
    goto error
  endif
  !
  if ncoef .le. 1 then
    define/local cnstm/c/1/16 "{ali20}"
    if cnstm(1:1) .eq. ":" cnstm = cnstm(2:)
  endif
elseif outputi .eq. 1 then  ! defaults
  define/local cstar/c/1/16 "OBJECT"
  define/local cmeaU/c/1/16 "MAG_U"
  define/local cmeaB/c/1/16 "MAG_B"
  define/local cmeaV/c/1/16 "MAG_V"
  define/local cmeaR/c/1/16 "MAG_R"
  define/local cmeaI/c/1/16 "MAG_I"
  define/local cairm/c/1/16 "AIRMASS"
  define/local cmjd/c/1/16 "MJD"
  define/local cflag/c/1/16 "FLAGS"
  define/local cstdU/c/1/16 "U"
  define/local cstdB/c/1/16 "B"
  define/local cstdV/c/1/16 "V"
  define/local cstdR/c/1/16 "R"
  define/local cstdI/c/1/16 "I"
  define/local colU/c/1/16 "U-B"
  define/local colB/c/1/16 "B-V"
  define/local colV/c/1/16 "B-V"
  define/local colR/c/1/16 "V-R"
  define/local colI/c/1/16 "V-R"
  if ncoef .le. 1 define/local cnstm/c/1/16 "N"
endif
!
!
!P4(thresh):
!----------
if ncoef .le. 1 then
  !
  define/parameter p4 + ? -
    "[8|5] thresholds (see procedure header): "
  outputi = m$parse(p4,"limit")
  if ncoef .eq. 0 then
    inputi(1) = 8
    inputi(2) = 8
  elseif ncoef .eq. 1 then
    inputi(1) = 5
    inputi(2) = 8  ! allowed, but #6-8 unused
  endif
  if outputi .ne. inputi(1) .and. outputi .ne. inputi(2) then
    if outputi .ne. 1 then  ! 1 allowed for defaults
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P4 !"
      goto error
    endif
  endif
  if outputi .eq. inputi(1) .or. outputi .eq. inputi(2) then
    define/local badthr/i/1/1 0
    define/local min_stdmea/i/1/1 {limit01} ? +low
    define/local airbinsize/r/1/1 {limit02} ? +low
    inputr = 10.**(-AIRACC)
    if airbinsize .lt. inputr badthr = 1
    define/local min_noairs/i/1/1 {limit03} ? +low
    if min_noairs .lt. 2 badthr = 1
    define/local min_airrng/r/1/1 {limit04} ? +low
    define/local min_nstars/i/1/1 {limit05} ? +low
    if ncoef .eq. 0 then
      define/local colbinsize/r/1/1 {limit06} ? +low
      inputr = 10.**(-COLACC)
      if colbinsize .lt. inputr badthr = 1
      define/local min_nocols/i/1/1 {limit07} ? +low
      if min_nocols .lt. 5 badthr = 1
      define/local min_colrng/r/1/1 {limit08} ? +low
    endif
    !
    if badthr .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "At least one threshold (P4) is out of range !"
      goto error
    endif
    !
  elseif outputi .eq. 1 then  ! defaults
    define/local min_stdmea/i/1/1 5     ? +low
    define/local airbinsize/r/1/1 0.001 ? +low    ! >= 10^(-AIRACC) !!
    define/local min_noairs/i/1/1 2     ? +low    ! >= 2 !!
    define/local min_airrng/r/1/1 0.5   ? +low
    define/local min_nstars/i/1/1 3     ? +low
    if ncoef .eq. 0 then
      define/local colbinsize/r/1/1 0.001 ? +low  ! >= 10^(-COLACC) !!
      define/local min_nocols/i/1/1 5     ? +low  ! >= 5 !!
      define/local min_colrng/r/1/1 1.0   ? +low
    endif
  endif
  !
elseif ncoef .eq. 2 then
  !
  define/local min_nstars/i/1/1 1 ? +low
  !
endif
!
!
!P5(coltabs):
!-----------
define/local allfiltloop/i/1/1 0           ! needed for below, too
define/local filter/c/1/{TOTFILT} " " all  ! needed for below, too (w.TOTFILT!)
if ncoef .ge. 1 then
  !
  if p5 .eq. "?" mess/out 2 {PROID} " "
  define/parameter p5 ? ? -
    "Tabs.w.colour_coeff._f.U,B,V,R,I,column_number [def.:?,?,?,?,?,1]: "
  outputi = m$parse(p5,"col")
  if outputi .ne. 6 .and. outputi .ne. 5 then  ! 5: def.f.number
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P5 !"
    goto error
  endif
  !
  if outputi .eq. 6 then
    if m$tstno(col06(1:1)) .eq. 0 col06 = col06(2:)  ! "#" possible
    if m$tstno(col06) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "P5[6] must be a number !"
    goto error
    endif
    define/local ckcol/i/1/1 {col06}
  elseif outputi .eq. 5 then  ! default
    define/local ckcol/i/1/1 1
  endif
  !
  define/local allcolfilt/c/1/{TOTFILT} " " all
  do allfiltloop = 1 TOTFILT
    filter = ALLFILT({allfiltloop}:{allfiltloop})
    set/format I2
    define/local coltbl{filter}/c/1/64 "{col{allfiltloop}}"
    set/format {FMTDEF}
    !
    ! "?|+" mean that this single filter should not be used:
    if coltbl{filter} .ne. "?" .and. coltbl{filter} .ne. "+" then
      allcolfilt = allcolfilt // filter
      if m$index(coltbl{filter},".tbl") .eq. 0 -
        coltbl{filter} = coltbl{filter}//".tbl"
      if m$exist(coltbl{filter}) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered table is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({coltbl{filter}})"
        goto error
      elseif {{coltbl{filter}},TBLCONTR(3)} .lt. ckcol then
        write/out " "
        mess/out 1 {PROID} "Entered column number is not present !"
        mess/out 1 {PROID} "({coltbl{filter}},#{ckcol})"
        goto error
      elseif {{coltbl{filter}},TBLCONTR(4)} .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered table is empty !"
        mess/out 1 {PROID} "({coltbl{filter}})"
        goto error
      endif
    endif
    !
  enddo
  !
  if m$len(allcolfilt) .ge. 1 then
    !
        mess/out 3 {PROID} "--> found colour coef. for filters <{allcolfilt}>"
    !
  else
        write/out " "
        mess/out 1 {PROID} "No colour coefficients present !"
    goto error
  endif
  !
  define/local allcalfilt/c/1/{TOTFILT} " " all  ! necessary for P6 (ncoef=2)
  if ncoef .eq. 1 allcalfilt = allcolfilt
  !
endif
!
!
!P6(exttabs):
!-----------
if ncoef .eq. 2 then
  !
  if p6 .eq. "?" mess/out 2 {PROID} " "
  define/parameter p6 ? ? -
    "Tabs.w.ext.coeff._f.U,B,V,R,I,column_number [def.:?,?,?,?,?,1]: "
  outputi = m$parse(p6,"ext")
  if outputi .ne. 6 .and. outputi .ne. 5 then  ! 5: def.f.number
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P6 !"
    goto error
  endif
  !
  if outputi .eq. 6 then
    if m$tstno(ext06(1:1)) .eq. 0 ext06 = ext06(2:)  ! "#" possible
    if m$tstno(ext06) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "P6[6] must be a number !"
    goto error
    endif
    define/local ckext/i/1/1 {ext06}
  elseif outputi .eq. 5 then  ! default
    define/local ckext/i/1/1 1
  endif
  !
  define/local allextfilt/c/1/{TOTFILT} " " all
  do allfiltloop = 1 TOTFILT
    filter = ALLFILT({allfiltloop}:{allfiltloop})
    set/format I2
    define/local exttbl{filter}/c/1/64 "{ext{allfiltloop}}"
    set/format {FMTDEF}
    !
    ! "?|+" mean that this single filter should not be used:
    if exttbl{filter} .ne. "?" .and. exttbl{filter} .ne. "+" then
      allextfilt = allextfilt // filter
      if m$index(exttbl{filter},".tbl") .eq. 0 -
        exttbl{filter} = exttbl{filter}//".tbl"
      if m$exist(exttbl{filter}) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered table is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({exttbl{filter}})"
        goto error
      elseif {{exttbl{filter}},TBLCONTR(3)} .lt. ckext then
        write/out " "
        mess/out 1 {PROID} "Entered column number is not present !"
        mess/out 1 {PROID} "({exttbl{filter}},#{ckext})"
        goto error
      elseif {{exttbl{filter}},TBLCONTR(4)} .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered table is empty !"
        mess/out 1 {PROID} "({exttbl{filter}})"
        goto error
      endif
    endif
    !
  enddo
  !
  if m$len(allextfilt) .ge. 1 then
    !
        mess/out 3 {PROID} "--> found extin. coef. for filters <{allextfilt}>"
    !
  else
        write/out " "
        mess/out 1 {PROID} "No extin. coefficients present !"
    goto error
  endif
  !
  allcalfilt = " "
  if allextfilt .eq. allcolfilt then
    !
    allcalfilt = allextfilt
    !
  else  ! extract filters common to ext./col.tbl.
    !
    do allfiltloop = 1 TOTFILT
      filter = ALLFILT({allfiltloop}:{allfiltloop})
      !
      if m$index(allcolfilt,filter) .gt. 0 -
        .and. m$index(allextfilt,filter) .gt. 0 then
        allcalfilt = allcalfilt // filter
      endif
      !
    enddo
    !
  endif
  !
  if m$len(allcalfilt) .ge. 1 then
    !
        mess/out 3 {PROID} "==> found common coef. for filters <{allcalfilt}>"
    !
  else
        write/out " "
        mess/out 1 {PROID} "No common coefficients present !"
    goto error
  endif
  !
endif
!
!
!P7(disp):
!--------
define/parameter p7 l ? -
  "        Enter display [def.:<l>ocal|<p>ipeline|<n>one|host:disp[.screen]]: "
define/local dispflag/c/1/30 "{p7}"
!
!P8(re):
!------
define/parameter p8 n ? -
  "        Enter flag for re-calculation [def.:<n>o|<y>es]: "
define/local recalc/c/1/1 "{p8}"
!
!
!
!     %1c  DEFINITION OF FURTHER VARIABLES :
!          *******************************
!
!     "Managing displays" :
  define/local rgbflag/i/1/1 0
!
!     "L o o p (aligned)" :
  define/local topictbl/c/1/64 " " all
  define/local loopcount/i/1/1 0
  !
  !   "Getting used filter" :
  define/local tblfilt/c/1/{TOTFILT} " " all +low
  !
  !   "Preparing merged tables" :
  define/local alltblfilt/c/1/{TOTFILT} " " all +low
  define/local char/i/1/1 0
  !
  define/local tblfiltloop/i/1/1 0
  define/local tottblfilt/i/1/1 0
  !
  !   "Saving filters" :
  define/local refile/c/1/64 " " all
  define/local refileid/i/1/2 0 all
!     "E n d (aligned)"
!
!     "L o o p (merged)" :
  !   "Reading filters" :
  define/local realltblfilt/c/1/{TOTFILT} " " all
  define/local retottblfilt/i/1/1 0
  !
  define/local currtbl/c/1/64 " " all
  define/local outidx/i/1/1 0 ? +low
  !
  !   "Removing underscores" :
  define/local allrows/i/1/1 0
  define/local alloop/i/1/1 0
  define/local nameloop/i/1/1 0
  !
  !   "Calculating colours" :
  define/local colour/c*16/1/{TOTFILT} " " all +low
  !
  !   "Selecting high quality obj." :
  define/local hiqualobjno/i/1/1 0
  define/local mess_type_i/i/1/1 0
  define/local hiqualtbl/c/1/64 " " all +low
  define/local loqualobjno/i/1/1 0 ? +low
  define/local loqualtbl/c/1/64 " " all +low
  !
  !   "Extinction coefficient" :
  define/local skip/i/1/1 0
  define/local gdobjno/i/1/1 0 ? +low
  do inputi = 1 TOTFILT
    filter = ALLFILT({inputi}:{inputi})
    define/local kext{filter}/r/1/512 0. all +low
    define/local zcol{filter}/r/1/512 0. all +low
  enddo
  define/local kextmea/d/1/{TOTFILT} 0. all +low
  define/local kextsig/r/1/{TOTFILT} 0. all +low
  !
  !   "Colour coefficient / zero point" :
  define/local graphflag/c/1/3 " " all
  define/local kcolmea/d/1/{TOTFILT} 0. all +low
  define/local kcolsig/r/1/{TOTFILT} 0. all +low
  define/local zmea/d/1/{TOTFILT} 0. all +low
  define/local zsig/r/1/{TOTFILT} 0. all +low
  define/local ux/r/1/4 0. all +low
  define/local uy/r/1/4 0. all +low
  define/local mjdmin/r/1/1 500000. ? +low  ! "500000." !!
  define/local mjdmax/r/1/1 0. ? +low
  !
  !   "Which filter .. not skipped ?" :
  define/local goodtblfilt/c/1/{TOTFILT} " " all +low
  !
!     "E n d (merged)"
!
!     "Results" :
  define/local outi/i/1/1 0
!
!     "Output" :
  define/local zerocat/c/1/64 " " all
  define/local kextcat/c/1/64 " " all
  define/local kcolcat/c/1/64 " " all
  define/local zerotbl/c/1/64 " " all
  define/local kexttbl/c/1/64 " " all
  define/local kcoltbl/c/1/64 " " all
!
!
!
!     %1d  MANAGING DISPLAYS :
!          *****************
!
if dispflag .ne. "p" then
  @p fors_nicecrea {dispflag}
  !~~~~~~~~~~~~~~~~~~~~~~~~~~
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Managing of display options failed !"
    goto error
  endif
endif
!
if nice_crea .eq. "y" then  ! written by fors_nicecrea.prg
  if winopen(12) .eq. 1 then
    assign/graphic g,1 >Null
    clear/graphic
  endif
  if winopen(13) .eq. 1 then
    assign/graphic g,2 >Null
    clear/graphic
  endif
  if winopen(2) .eq. 1 then
    assign/display d,1 >Null
    clear/display
    display/lut off
  endif
  if winopen(3) .eq. 1 then
    assign/display d,2 >Null
    clear/display
    display/lut off
  endif
  if winopen(4) .eq. 1 then
    assign/display d,3 >Null
    clear/display
    display/lut off
  endif
endif
!
!       CREATION OF FIXED-SIZED GRAPHIC WINDOW {FIXGRAPHNO}:
if nice_crea .eq. "y" then
  if winopen(19) .gt. 0 then
        mess/out 4 {PROID} "Deleting old graphic window {FIXGRAPHNO}"
    delete/graphic {FIXGRAPHNO} >Null
  endif
  ididev(7) = p_dspmode  ! for TrueColour displays
  !
  rgbflag = ididev(7)
  if nice_disp .eq. "l" then
        mess/out 4 {PROID} -
          "Creating graph.wnd.{FIXGRAPHNO} (fixed size) on local screen"
    create/graphic {FIXGRAPHNO} {GRAPHCOORD} 900000 ? >Null
  else
        mess/out 4 {PROID} -
          "Creating graph.wnd.{FIXGRAPHNO} (fixed size) on <{nice_disp}>"
    create/graphic {FIXGRAPHNO} {GRAPHCOORD} 900000 {nice_disp} >Null
  endif
  ididev(7) = rgbflag
endif
!
!
!
!     %2   L O O P   W I T H   A L I G N E D
!          P H O T O M E T R Y   T A B L E S :
!          * * * * * * * * * * * * * * * * *
!
!
! this loop is only necessary for a 1st run of an input catalogue:
if recalc .eq. "y" goto skipped_merging
!
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "merging photometry tables"
!
catal = 0
catloop:
  store/frame topictbl {alicat} 1 end_catloop
  loopcount = loopcount+1
        mess/out 3 {PROID} -
          "  extracting from table {loopcount} / {catentries} : {topictbl}"
  !
  !
  !   %2a  GETTING USED FILTER :
  !        *******************
  !
        mess/out 4 {PROID} "    getting used filter"
  !
  tblfilt = " "
  do allfiltloop = 1 TOTFILT
    filter = ALLFILT({allfiltloop}:{allfiltloop})
    if m$existc(topictbl,":{cmea{filter}}") .gt. 0 then
      ! filter is last char. in column label f.measured mags.:
      inputi = m$len(cmea{filter})
      tblfilt = tblfilt // cmea{filter}({inputi}:{inputi})
    endif
  enddo
  !
  if m$len(tblfilt) .eq. 0 then
        mess/out 2 {PROID} "    No column with measured mags. found -"
        mess/out 2 {PROID} "    skipping topical ali.phot.tbl. !"
    warn = 1
    goto catloop
  elseif m$len(tblfilt) .gt. 1 then
        mess/out 2 {PROID} "    Measured mags. for more than one filter found -"
        mess/out 2 {PROID} "    no automatic colour assignment possible -"
        mess/out 2 {PROID} "    skipping topical ali.phot.tbl. !"
    warn = 1
    goto catloop
  endif
  !
  !
  !        CHECKING USED FILTER :
  !        ********************
  !
        mess/out 4 {PROID} "    checking used filter"
  !
  if m$index(ALLFILT,tblfilt) .eq. 0 then
        mess/out 2 {PROID} -
          "    Unknown filter <{tblfilt}> - only <{ALLFILT}> are possible"
        mess/out 2 {PROID} -
          "    as 1st char. in colu. label for measured mags. -"
        mess/out 2 {PROID} "    skipping topical ali.phot.tbl. !"
    warn = 1
    goto catloop
  endif
  if ncoef .ge. 1 then
    !
    if m$index(allcalfilt,tblfilt) .eq. 0 then
        mess/out 2 {PROID} "    No calib.coef. present for filter <{tblfilt}> -"
        mess/out 2 {PROID} "    skipping topical ali.phot.tbl. !"
      warn = 1
      goto catloop
    endif
    !
  endif
  !
        mess/out 4 {PROID} "    found filter: <{tblfilt}>"
  !
  !
  !        CHECKING EXISTENCE OF FURTHER COLUMNS
  !        (FITTING TO CHECKED FILTER AND COLOUR USED FOR FILTER) :
  !        ******************************************************
  !
        mess/out 4 {PROID} "    checking columns"
  !
  if m$existc(topictbl,":{cstar}") .eq. -1 then
        mess/out 2 {PROID} "    Column is not present (:{cstar}) -"
        mess/out 2 {PROID} "    skipping table !"
    goto catloop
  elseif m$existc(topictbl,":{cairm}") .eq. -1 then
        mess/out 2 {PROID} "    Column is not present (:{cairm}) -"
        mess/out 2 {PROID} "    skipping table !"
    goto catloop
  elseif m$existc(topictbl,":{cmjd}") .eq. -1 then
        mess/out 2 {PROID} "    Column is not present (:{cmjd}) -"
        mess/out 2 {PROID} "    skipping table !"
    goto catloop
  elseif m$existc(topictbl,":{cflag}") .eq. -1 then
        mess/out 2 {PROID} "    Column is not present (:{cflag}) -"
        mess/out 2 {PROID} "    skipping table !"
    goto catloop
  elseif m$existc(topictbl,":{cstd{tblfilt}}") .eq. -1 then
        mess/out 2 {PROID} "    Column is not present (:{cstd{tblfilt}}) -"
        mess/out 2 {PROID} "    skipping table !"
    goto catloop
  elseif m$existc(topictbl,":{cstd{col{tblfilt}(1:1)}}") .eq. -1 then
        mess/out 2 {PROID} -
          "    Column is not present (:{cstd{col{tblfilt}(1:1)}}) -"
        mess/out 2 {PROID} "   skipping table !"
    goto catloop
  elseif m$existc(topictbl,":{cstd{col{tblfilt}(3:3)}}") .eq. -1 then
        mess/out 2 {PROID} -
          "    Column is not present (:{cstd{col{tblfilt}(3:3)}}) -"
        mess/out 2 {PROID} "    skipping table !"
    goto catloop
  endif
  if ncoef .le. 1 then
    if m$existc(topictbl,":{cnstm}") .eq. -1 then
        mess/out 2 {PROID} "    Column is not present (:{cnstm}) -"
        mess/out 2 {PROID} "    skipping table !"
      goto catloop
    endif
  endif
  if {{topictbl},TBLCONTR(4)} .eq. 0 then
        mess/out 2 {PROID} "    Table is empty - skipping !"
    goto catloop
  endif
  !
  !
  !        PREPARING MERGED ALI.PHOT.TABLES (ONE PER FILTER) :
  !        *************************************************
  !
  if m$index(alltblfilt,tblfilt) .eq. 0 then  ! colour not present multi.
    alltblfilt = alltblfilt // tblfilt
  else
    goto alltbl_exists
  endif
  !
        mess/out 4 {PROID} "    preparing merged table for filter <{tblfilt}>"
  !
    char = m$index(alicat,".cat")-1
    define/local alltbl{tblfilt}/c/1/64 " " all
    alltbl{tblfilt} = alicat(1:{char})//"_{tblfilt}.tbl"
  create/table {alltbl{tblfilt}} 16 1 null
  !
  create/column {alltbl{tblfilt}} :IDX     " "   I4    I*4
  create/column {alltbl{tblfilt}} :OBJECT  " "   A11   C*11
  set/format I1
  create/column {alltbl{tblfilt}} :AIRMASS "  "  F7.{AIRACC} R*4
  create/column {alltbl{tblfilt}} :INS_{tblfilt} -
                                           "mag" F8.3  R*4
  create/column {alltbl{tblfilt}} :MJD     "day" F13.6 R*8
  create/column {alltbl{tblfilt}} :FLAGS   " "   A9    C*9
  create/column {alltbl{tblfilt}} :NSTD    " "   I4    I*4
  create/column {alltbl{tblfilt}} :STD_U   "mag" F7.3  R*4
  create/column {alltbl{tblfilt}} :STD_B   "mag" F7.3  R*4
  create/column {alltbl{tblfilt}} :STD_V   "mag" F7.3  R*4
  create/column {alltbl{tblfilt}} :STD_R   "mag" F7.3  R*4
  create/column {alltbl{tblfilt}} :STD_I   "mag" F7.3  R*4
  create/column {alltbl{tblfilt}} :{col{tblfilt}(1:1)}_{col{tblfilt}(3:3)} -
                                           "mag" F6.{COLACC} R*4
  create/column {alltbl{tblfilt}} :STD_INS "mag" F7.3  R*4
  create/column {alltbl{tblfilt}} :Z_COL   "mag" F7.3  R*4
  create/column {alltbl{tblfilt}} :Z       "mag" F7.3  R*4
  !
  ! for copy/tt below:
    define/local emptytbl{tblfilt}/c/1/64 " " all
    emptytbl{tblfilt} = alicat(1:{char})//"_{tblfilt}tpl.tbl"
  -copy {alltbl{tblfilt}} {emptytbl{tblfilt}}
  ! for copy/tt,merge/table below:
    define/local tmptbl{tblfilt}/c/1/64 " " all
    tmptbl{tblfilt} = alicat(1:{char})//"_{tblfilt}tmp.tbl"
  !
  alltbl_exists:
  !
  -copy {emptytbl{tblfilt}} {tmptbl{tblfilt}}  ! => tmptbls are empty (copy/tt
  !                                            !    doesn't delete old rows)
  !
  !
  !   %2b  ADDING TABLE VALUES TO MERGED TABLES :
  !        ************************************
  !
        mess/out 4 {PROID} "    merging table"
  !
  copy/tt {topictbl} :{cstar}         {tmptbl{tblfilt}} :OBJECT
  copy/tt {topictbl} :{cairm}         {tmptbl{tblfilt}} :AIRMASS
  copy/tt {topictbl} :{cmea{tblfilt}} {tmptbl{tblfilt}} :INS_{tblfilt}
  copy/tt {topictbl} :{cmjd}          {tmptbl{tblfilt}} :MJD
  copy/tt {topictbl} :{cflag}         {tmptbl{tblfilt}} :FLAGS
  if ncoef .le. 1 then
    copy/tt {topictbl} :{cnstm}         {tmptbl{tblfilt}} :NSTD
  endif
  copy/tt {topictbl} :{cstd{tblfilt}} {tmptbl{tblfilt}} :STD_{tblfilt}
  if tblfilt .ne. col{tblfilt}(1:1) -
    copy/tt {topictbl} :{cstd{col{tblfilt}(1:1)}} -
      {tmptbl{tblfilt}} :STD_{col{tblfilt}(1:1)}
  if tblfilt .ne. col{tblfilt}(3:3) -
    copy/tt {topictbl} :{cstd{col{tblfilt}(3:3)}} -
      {tmptbl{tblfilt}} :STD_{col{tblfilt}(3:3)}
  !
  merge/table {alltbl{tblfilt}} {tmptbl{tblfilt}} tmp_a.tbl
  -rename tmp_a.tbl {alltbl{tblfilt}}
  !
  !
  goto catloop
end_catloop:
!
!
tottblfilt = m$len(alltblfilt)
if tottblfilt .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No ali.phot.table left after technical checks !"
  goto error
endif
!
do tblfiltloop = 1 tottblfilt
  tblfilt = alltblfilt({tblfiltloop}:{tblfiltloop})
  -delete {tmptbl{tblfilt}}
  -delete {emptytbl{tblfilt}}
enddo
!
!
!          SAVING FILTERS OF MERGED TABLES
!          FOR POSSIBLE FURTHER RUN WITH-
!          OUT REPEATED MERGING :
!          *******************************
!
        mess/out 4 {PROID} "saving filters for possible further run"
!
  char = m$index(alicat,".cat")-1
  refile = alicat(1:{char})//"_filt.dat"
open/file {refile} write refileid
if refileid(1) .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Unable to open new file {refile} !"
  goto error
endif
write/file {refileid(1)} {ncoef}
write/file {refileid(1)} {alltblfilt}
close/file {refileid(1)}
!
!
skipped_merging:
!
!
!          E N D   O F   L O O P   W I T H
!            A L I G N E D   T A B L E S
!          * * * * * * * * * * * * * * * *
!
!
!
!     %3   L O O P   W I T H   M E R G E D
!          P H O T O M E T R Y   T A B L E S :
!          * * * * * * * * * * * * * * * * *
!
!
        mess/out 3 {PROID} " "
!
!
!          READING FILTERS OF MERGED TABLES
!          IN FURTHER RUN WITHOUT REPEATED
!          MERGING :
!          *******************************
!
if recalc .eq. "y" then
  !
        mess/out 4 {PROID} "reading filters from first run"
  !
  char = m$index(alicat,".cat")-1
  refile = alicat(1:{char})//"_filt.dat"
  if m$exist(refile) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Chosen flag P8(re)=<y> needs file from 1st run -"
        mess/out 1 {PROID} "it is not present or can not be opened !"
        mess/out 1 {PROID} "({refile})"
    goto error
  endif
  open/file {refile} read refileid
  !
  read/file {refileid(1)} outputc
  if m$tstno(outputc) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Chosen flag P8(re)=<y> needs file from 1st run -"
        mess/out 1 {PROID} "it is corrupt !"
        mess/out 1 {PROID} "({refile})"
    close/file {refileid(1)}
    goto error
  endif
  if {outputc} .ne. ncoef then
        write/out " "
        mess/out 1 {PROID} "For chosen flag P8(re)=<y> P1(ncoef) must not"
        mess/out 1 {PROID} "differ from 1st run - P1_1.run=<{outputc}> !"
    close/file {refileid(1)}
    goto error
  endif
  !
  read/file {refileid(1)} outputc  ! read/file does not work with keywords
  alltblfilt = outputc             ! longer than 8 characters !
  tottblfilt = m$len(alltblfilt)
  if tottblfilt .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Chosen flag P8(re)=<y> needs file from 1st run -"
        mess/out 1 {PROID} "it is corrupt !"
        mess/out 1 {PROID} "({refile})"
    close/file {refileid(1)}
    goto error
  endif
  close/file {refileid(1)}
  !
  char = m$index(alicat,".cat")-1
  do tblfiltloop = 1 tottblfilt
    tblfilt = alltblfilt({tblfiltloop}:{tblfiltloop})
    define/local alltbl{tblfilt}/c/1/64 " " all
    alltbl{tblfilt} = alicat(1:{char})//"_{tblfilt}.tbl"
    if m$exist(alltbl{tblfilt}) .eq. 0 then
        mess/out 2 {PROID} -
          "Skipping filter <{tblfilt}> - merged table from 1st run"
        mess/out 2 {PROID} "is not present or can not be opened !"
        mess/out 2 {PROID} "({alltbl{tblfilt}})"
    else
      realltblfilt = realltblfilt // tblfilt
    endif
  enddo
  retottblfilt = m$len(realltblfilt)
  if retottblfilt .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No merged ali.phot.table found !"
    goto error
  endif
  alltblfilt = realltblfilt
  tottblfilt = retottblfilt
  !
        mess/out 3 {PROID} "merged tables from first run:"
else
        mess/out 3 {PROID} "merged tables:"
  !
endif  ! "if recalc .eq. "y""
!
do tblfiltloop = 1 tottblfilt
  tblfilt = alltblfilt({tblfiltloop}:{tblfiltloop})
  currtbl = alltbl{tblfilt}
        mess/out 3 {PROID} "  {tblfilt}: {currtbl}"
enddo
!
!
do tblfiltloop = 1 tottblfilt
  tblfilt = alltblfilt({tblfiltloop}:{tblfiltloop})
  currtbl = alltbl{tblfilt}
  outidx = m$index(ALLFILT,tblfilt)
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "{line}"
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} -
          "processing {tblfilt} filter (table {tblfiltloop} / {tottblfilt}) :"
        mess/out 3 {PROID} "{currtbl}"
  !
  !
  if recalc .ne. "y" then  ! only necessary for 1st run
    !
    !
    !      REMOVING BLANKS AND UNDERSCORES
    !      IN ALL OBJECT NAMES :
    !      *******************************
    !
        mess/out 4 {PROID} " "
        mess/out 4 {PROID} "  unifying object names :"
    !
    ! Object names are sometimes with underscores, sometimes without;
    ! blanks may be in front / at the end of the names, too:
    !
    allrows = {{currtbl},TBLCONTR(4)}
    do alloop = 1 allrows
        ! ###A mess/out still too slow !
        ! mess/out -4 {PROID} "    name {alloop} / {allrows}\r"
        if mess_level .gt. 3 -
          $ /bin/echo -
            "INFO    \0133{PROID}\0135:\t\tname {alloop} / {allrows}\r\c"

      !  removes leading blanks, cuts at first non-leading blank:
      write/keyword inputc {{currtbl},:OBJECT,@{alloop}}
      inputi = m$len(inputc)
      outputc = " "
      do nameloop = 1 inputi
        in_a = inputc({nameloop}:{nameloop})
        if in_a .ne. "_" then  ! current char. of object name is no underscore
          outputc = outputc // in_a
        endif
      enddo
      write/table {currtbl} :OBJECT @{alloop} {outputc}  ! rem.longer old names
    enddo
        ! ###A mess/out 4 {PROID}
        if mess_level .gt. 3 $ /bin/echo "INFO"
    !
    !
    ! %3a  CALCULATION OF COLOURS :
    !      **********************
    !
    colour({outidx}) = "{col{tblfilt}(1:1)}_{col{tblfilt}(3:3)}"
    !
        mess/out 3 {PROID} " "
        in_a = colour({outidx})
        in_a = in_a // " = {col{tblfilt}(1:1)}_lit-{col{tblfilt}(3:3)}_lit"
        mess/out 3 {PROID} "  calculating colours {in_a}"
    !
    compute/table {currtbl} :{colour({outidx})} = -
      :STD_{col{tblfilt}(1:1)} - :STD_{col{tblfilt}(3:3)} >Null
    !
    !
    ! %3b  CALCULATION OF STANDARD MINUS INSTRUMENT MAGNITUDES :
    !      ***************************************************
    !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "  calculating {tblfilt}_lit-{tblfilt}_inst"
    !
    compute/table {currtbl} :STD_INS = :STD_{tblfilt} - :INS_{tblfilt}
    !
    !
  else   ! also necessary for further runs
    !
    colour({outidx}) = "{col{tblfilt}(1:1)}_{col{tblfilt}(3:3)}"
    !
  endif  ! "if recalc .ne. "y""
  !
  !
  !   %3c  SELECTION OF OBJECTS WHICH ARE NOT
  !        MANUALLY FLAGGED AS BAD ONES :
  !        **********************************
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "  removing previously flagged objects"
  !
  select/table {currtbl} :FLAGS.ne."Y*" >Null  ! "Y" for "Yes, remove it";
  hiqualobjno = outputi                        ! "*" for "tbl,:col,row ="
  if hiqualobjno .ge. min_nstars then          ! (no cleaning before writing)
    mess_type_i = 3                            ! instead of "wri/tab" !
  else
    mess_type_i = 2
  endif
        set/format J3
        mess/out {mess_type_i} {PROID} -
          "  - unflagged objects: {hiqualobjno} (min.{min_nstars}\b)"
        set/format {FMTDEF}
  if hiqualobjno .lt. min_nstars then
    select/table {currtbl} all
        mess/out 2 {PROID} -
          "  Too few unflagged ones - skipping filter <{tblfilt}> !"
    warn = 1
    skip = skip+1
    goto skipped_main
  endif
  char = m$index(currtbl,".tbl")-1
  hiqualtbl = currtbl(1:{char})//"_higq.tbl"
  copy/table {currtbl} {hiqualtbl}
  !
  select/table {currtbl} :FLAGS.eq."Y*" >Null
  loqualobjno = outputi
        set/format J3
        mess/out {mess_type_i} {PROID} -
          "  - flagged objects  : {loqualobjno}"
        set/format {FMTDEF}
  char = m$index(currtbl,".tbl")-1
  loqualtbl = currtbl(1:{char})//"_lowq.tbl"
  copy/table {currtbl} {loqualtbl}
  !
  !
  !        *************************************************
  !        *                       COMPUTATION OF          *
  !        *              ZERO_POINT  EXT.COEF.  COL.COEF. *
  !        *************************************************
  !        * MODE  0  =>      +           +          +     *
  !        *       1  =>      +           +       (given)  *
  !        *       2  =>      +        (given)    (given)  *
  !        *************************************************
  !
        mess/out 4 {PROID} " "
        mess/out 4 {PROID} "  branching to mode: {ncoef}"
  !
  !
  !   %3d  EXTINCTION COEFFICIENT :
  !        **********************
  !
  !-------------------------------------
  if ncoef .eq. 0 .or. ncoef .eq. 1 then
  !-------------------------------------
    !
    !   CHECKING OF EXTIN. THRESHOLDS / INDEXING OF GOOD OBJECTS:
    @p fors_calcoeff_check e
    !~~~~~~~~~~~~~~~~~~~~~~~
    if {q1} .eq. -1 then
      warn = 1
    elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Checking of extin. thresholds failed !"
      goto error
    endif
    if {q2} .gt. 0 then
      skip = skip+{q2}
      goto skipped_main
    endif
    gdobjno = {q3}
    !
    !   CALCULATION OF EXTINCTION COEF.S / ZERO POINTS WITHOUT
    !   COLOUR CORRECTION PER GOOD OBJECT,
    !   CALCULATION OF MEAN EXT.COEFF.:
    @p fors_calcoeff_reg
    !~~~~~~~~~~~~~~~~~~~
    if {q1} .eq. -1 then
      warn = 1
    elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Calculation of mean ext.coef. failed !"
      goto error
    endif
    if {q2} .gt. 0 then
      skip = skip+{q2}
      goto skipped_main
    endif
    !
  !-----------------------
  elseif ncoef .eq. 2 then
  !-----------------------
    !
    !   GETTING EXT.COEFF. FROM INPUT:
    kextmea({outidx}) = {{exttbl{tblfilt}},#{ckext},@1}
    kextsig({outidx}) = 0.
        mess/out 3 {PROID} " "
        set/format I1
        inputi = EXTACC+3
        set/format ,T{inputi}.{EXTACC}
        mess/out 3 {PROID} "  using k_ext: {kextmea({outidx})}"
        set/format {FMTDEF}
    !
  !-------------------------------------------
  endif  ! "if ncoef .eq. 0 .or. ncoef .eq. 1"
  !-------------------------------------------
  !
  if ncoef .eq. 0 .or. ncoef .eq. 1 then
    !   RE-CALC. OF ZERO POINTS WITHOUT COLOUR CORRECTION WITH
    !   MEAN EXT.COEF.:
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} -
          "  calculating {tblfilt}_lit-{tblfilt}_inst-<k_ext>*X"
  elseif ncoef .eq. 2 then
    !   CALC. OF ZERO POINTS WITHOUT COLOUR CORRECTION WITH
    !   INPUT EXT.COEF.:
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} -
          "  calculating {tblfilt}_lit-{tblfilt}_inst-k_ext*X"
  endif
  !  Selects all objects automatically:
  compute/table {hiqualtbl} :Z_COL = :STD_INS - {kextmea({outidx})}*:AIRMASS
  !~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  if ncoef .eq. 0 .or. ncoef .eq. 1 then
    select/table {hiqualtbl} :IDX.gt.0 >Null
  endif
  !
  if loqualobjno .gt. 0 then
    compute/table {loqualtbl} :Z_COL = :STD_INS - {kextmea({outidx})}*:AIRMASS
    !~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  endif
  !
  if nice_crea .eq. "y" then
    !   PLOT 1 - GOOD AND BAD OBJ.:
    @p fors_calcoeff_plt 1
    !~~~~~~~~~~~~~~~~~~~~~
    if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "<{tblfilt}>: first plot failed !"
      goto error
    endif
  endif
  !
  !
  !   %3e  COLOUR COEFFICIENT / ZERO POINT :
  !        *******************************
  !
  !-------------------
  if ncoef .eq. 0 then
  !-------------------
    !
    !   SELECTION OF GOOD OBJECTS / CHECKING OF COLOUR THRESHOLDS:
    @p fors_calcoeff_check c
    !~~~~~~~~~~~~~~~~~~~~~~~
    if {q1} .eq. -1 then
      warn = 1
    elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Checking of extin. thresholds failed !"
      goto error
    endif
    if {q2} .gt. 0 then
      skip = skip+{q2}
      goto skipped_main
    endif
    !
    !   CALC. OF COLOUR COEF. / ZERO POINT WITH COLOUR CORREC.
    !   WITH ALL GOOD OBJECTS:
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "  calculating colour coefficient and zeropoint :"
    if nice_crea .eq. "y" then
      ! PLOT 2 - GOOD OBJ.:
      @p fors_calcoeff_plt 21
      !~~~~~~~~~~~~~~~~~~~~~~
      if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "<{tblfilt}>: second plot (good obj.) failed !"
        goto error
      endif
      graphflag = "{FIXGRAPHNO},w"
    else
      graphflag = "n"
    endif
    regress/general {hiqualtbl} {colour({outidx})},Z_COL -
    lin rob {graphflag} n n n,n,n
    !~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Regression failed - :Z_COL(:{colour({outidx})}) !"
      goto error
    endif
    kcolmea({outidx}) = outputd(2)
    kcolsig({outidx}) = outputd(12)  ! ###B still no std.dev. from reg/pol
    zmea({outidx}) = outputd(1)
    zsig({outidx}) = outputd(11)  ! ###B still no std.dev. from reg/pol
    !
        mess/out 3 {PROID} " "
        set/format T7.4
        mess/out 3 {PROID} -
          "    k_col({colour({outidx})}) =  {kcolmea({outidx})} +-{kcolsig({outidx})}"
        mess/out 3 {PROID} -
          "    z({colour({outidx})})     =  {zmea({outidx})} +-{zsig({outidx})}"
        set/format {FMTDEF}
    !
  !-----------------------------------------
  elseif ncoef .eq. 1 .or. ncoef .eq. 2 then
  !-----------------------------------------
    !
    kcolmea({outidx}) = {{coltbl{tblfilt}},#{ckcol},@1}
    kcolsig({outidx}) = 0.
    !
    !   CALC. OF ZERO POINT WITH COLOUR CORRECTION WITH INPUT COLOUR COEF. ..
    !   NCOEF=1: .. WITH ALL GOOD OBJECTS (GOOD OBJECTS ARE STILL SELECTED):
    !   NCOEF=2: .. WITH ALL OBJECTS:
        mess/out 3 {PROID} " "
        set/format I1
        inputi = COLACC+3
        set/format ,T{inputi}.{COLACC}
        mess/out 3 {PROID} "  using k_col: {kcolmea({outidx})}"
        mess/out 3 {PROID} "  calculating zero point :"
        set/format {FMTDEF}
    if nice_crea .eq. "y" then
      ! PLOT 2 - GOOD OBJ.:
      @p fors_calcoeff_plt 21
      !~~~~~~~~~~~~~~~~~~~~~~
      if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "<{tblfilt}>: second plot (good obj.) failed !"
        goto error
      endif
      graphflag = "{FIXGRAPHNO},w"
    else
      graphflag = "n"
    endif
    regress/general {hiqualtbl} {colour({outidx})},Z_COL -
    slope,{kcolmea({outidx})} avg {graphflag} n n n,n,n
    !~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} -
          "Regression w.const.slope failed - :Z_COL(:{colour({outidx})}) !"
      goto error
    endif
    zmea({outidx}) = outputd(1)
    zsig({outidx}) = outputd(11)  ! ###B still no std.dev. from reg/pol
    !
        mess/out 3 {PROID} " "
        set/format T7.4
        mess/out 3 {PROID} -
          "    z({colour({outidx})})     = {zmea({outidx})} +-{zsig({outidx})}"
        set/format {FMTDEF}
    !
  !-------------------------
  endif  ! "if ncoef .eq. 0"
  !-------------------------
  !
  if nice_crea .eq. "y" then
    !   PLOT 2 - BAD OBJ.:
    @p fors_calcoeff_plt 22
    !~~~~~~~~~~~~~~~~~~~~~~
    if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "<{tblfilt}>: second plot (bad obj.) failed !"
      goto error
    endif
  endif
  !
  !     CALC. OF ALL SINGLE ZERO POINTS
  !     WITH EXTINCTION AND COLOUR CORRECTION:
        mess/out 3 {PROID} " "
  if ncoef .eq. 0 .or. ncoef .eq. 1 then
        mess/out 3 {PROID} -
          "  calculating {tblfilt}_lit-{tblfilt}_inst-<k_ext>*X-k_col*({colour({outidx})})"
  elseif ncoef .eq. 2 then
        mess/out 3 {PROID} -
          "  calculating {tblfilt}_lit-{tblfilt}_inst-k_ext*X-k_col*({colour({outidx})})"
  endif
  !  Selects all objects automatically:
  compute/table {hiqualtbl} :Z = :STD_INS - {kextmea({outidx})}*:AIRMASS -
    - {kcolmea({outidx})}*:{colour({outidx})}
  !~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  if ncoef .eq. 0 .or. ncoef .eq. 1 then
    select/table {hiqualtbl} :IDX.gt.0 >Null
  endif
  !
  if loqualobjno .gt. 0 then
    compute/table {loqualtbl} :Z = :STD_INS - {kextmea({outidx})}*:AIRMASS -
      - {kcolmea({outidx})}*:{colour({outidx})}
    !~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  endif
  !
  if nice_crea .eq. "y" then
    !   PLOT 3 - GOOD AND BAD OBJ.:
    @p fors_calcoeff_plt 3
    !~~~~~~~~~~~~~~~~~~~~~
    if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "<{tblfilt}>: third plot failed !"
      goto error
    endif
  endif
  !
  !
  !        WHICH FILTER HAS NOT BEEN SKIPPED ?
  !        *********************************
  !
  goodtblfilt = goodtblfilt // tblfilt
  !
  !
skipped_main:
enddo  ! "tblfiltloop = 1 tottblfilt"
!
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "{line}"
!
if nice_crea .eq. "y" then
  !   MARKING MISSING OR SKIPPED FILTERS:
  @p fors_calcoeff_plt 4
  !~~~~~~~~~~~~~~~~~~~~~
  if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Labels for missing or skipped filters failed !"
    goto error
  endif
  !   SAVING PLOT FILE:
  @p fors_calcoeff_plt 5
  !~~~~~~~~~~~~~~~~~~~~~
  if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Saving of plot file failed !"
    goto error
  endif
  PLOT_FILE = q2
endif
!
if skip .eq. tottblfilt then
        write/out " "
        mess/out 1 {PROID} "No filter left after threshold checks !"
  goto error
endif
!
!
!          E N D   O F   L O O P   W I T H
!             M E R G E D   T A B L E S
!          * * * * * * * * * * * * * * * *
!
!
!
!     %4a  DISPLAYING OF RESULTS :
!          *********************
!
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "Used formula:"
        mess/out 3 {PROID} -
          " /U\       /U\                           /{colU}\\b"  ! "\b" !!
        mess/out 3 {PROID} -
          "| B |     | B |                         | {colB} |"
        mess/out 3 {PROID} -
          "| V |lit -| V |inst = z +k_ext*X +k_col*| {colV} |lit"
        mess/out 3 {PROID} -
          "| R |     | R |       \__________/      | {colR} |"
        mess/out 3 {PROID} -
          " \I/       \I/          = z_col          \ \b{colI}/"  ! " \b" !!
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} -
          "Calibration coefficients:"
        set/format T5.2,T6.2
    if ncoef .eq. 0 then
        mess/out 3 {PROID} "  |    z (OUT)    |  k_ext (OUT)  |  k_col (OUT)"
    elseif ncoef .eq. 1 then
        mess/out 3 {PROID} "  |    z (OUT)    |  k_ext (OUT)  |  k_col (IN)"
    elseif ncoef .eq. 2 then
        mess/out 3 {PROID} "  |    z (OUT)    |  k_ext (IN)   |  k_col (IN)"
    endif
        mess/out 3 {PROID} "--+---------------+---------------+--------------"
    do outi = 1 TOTFILT
      filter = ALLFILT({outi}:{outi})
      if m$index(goodtblfilt,filter) .gt. 0 then
        append/out -
          "{filter} |{zmea({outi})} +-{zsig({outi})} |"
        append/out -
          "{kextmea({outi})} +-{kextsig({outi})} |{kcolmea({outi})} +-{kcolsig({outi})}"
        mess/out 3 {PROID}
      elseif m$index(alltblfilt,filter) .gt. 0 then
        mess/out 3 {PROID} -
          "{filter} |    failed     |    failed     |    failed"
      else
        mess/out 3 {PROID} -
          "{filter} |    no data    |    no data    |    no data"
      endif
    enddo
        mess/out 3 {PROID} "--+---------------+---------------+--------------"
        set/format {FMTDEF}
!
        longlog/out 1 " " >Null
        longlog/out 1 "     Taken from input tables:" >Null
    if ncoef .eq. 0 then
        longlog/out 1 "     no coefficients" >Null
    elseif ncoef .eq. 1 then
        longlog/out 1 "     colour coefficients  (k_col)" >Null
    elseif ncoef .eq. 2 then
        longlog/out 1 "     colour coefficients      (k_col)" >Null
        longlog/out 1 "     extinction coefficients  (k_ext)" >Null
    endif
!
        longlog/out 1 " " >Null
        longlog/out 1 -
  "     Used formula:" >Null
        longlog/out 1 -
     "      /U\       /U\                           /{colU}\ " >Null  ! " " !!
        longlog/out 1 -
     "     | B |     | B |                         | {colB} |" >Null
        longlog/out 1 -
     "     | V |lit -| V |inst = z +k_ext*X +k_col*| {colV} |lit" >Null
        longlog/out 1 -
     "     | R |     | R |                         | {colR} |" >Null
        longlog/out 1 -
     "      \I/       \I/                           \'colI'/" >Null  ! "''" !!
        longlog/out 1 " " >Null
        longlog/out 1 -
  "     Calibration coefficients:" >Null
        set/format T5.2,T6.2
    if ncoef .eq. 0 then
        longlog/out 1 -
     "       |    z (OUT)    |  k_ext (OUT)  |  k_col (OUT)" >Null
    elseif ncoef .eq. 1 then
        longlog/out 1 -
     "       |    z (OUT)    |  k_ext (OUT)  |  k_col (IN)" >Null
    elseif ncoef .eq. 2 then
        longlog/out 1 -
     "       |    z (OUT)    |  k_ext (IN)   |  k_col (IN)" >Null
    endif
        longlog/out 1 -
     "     --+---------------+---------------+--------------" >Null
    do outi = 1 TOTFILT
      filter = ALLFILT({outi}:{outi})
      if m$index(goodtblfilt,filter) .gt. 0 then
        append/out -
          "{filter} |{zmea({outi})} +-{zsig({outi})} |"
        append/out -
          "{kextmea({outi})} +-{kextsig({outi})} |{kcolmea({outi})} +-{kcolsig({outi})}"
        longlog/out 1 "     {cbuf}" >Null
        write/keyword cbuf " " all
      elseif m$index(alltblfilt,filter) .gt. 0 then
        longlog/out 1 -
       "     {filter} |    failed     |    failed     |    failed" >Null
      else
        longlog/out 1 -
       "     {filter} |    no data    |    no data    |    no data" >Null
      endif
    enddo
        longlog/out 1 -
       "     --+---------------+---------------+--------------" >Null
        set/format {FMTDEF}
!
!
!     %4b  CREATION OF OUTPUT TABLES / CATALOGUES :
!          **************************************
!
  char = m$index(alicat,".cat")-1
!
zerocat = alicat(1:{char})//"_zero.cat"
kextcat = alicat(1:{char})//"_kext.cat"
kcolcat = alicat(1:{char})//"_kcol.cat"
!
create/icat {zerocat} null >Null
if ncoef .eq. 0 .or. ncoef .eq. 1 then
  create/icat {kextcat} null >Null
endif
if ncoef .eq. 0 then
  create/icat {kcolcat} null >Null
endif
!
do outi = 1 TOTFILT
  filter = ALLFILT({outi}:{outi})
  if m$index(goodtblfilt,filter) .gt. 0 then
    !
    zerotbl = alicat(1:{char})//"_{filter}_zero.tbl"
    kexttbl = alicat(1:{char})//"_{filter}_kext.tbl"
    kcoltbl = alicat(1:{char})//"_{filter}_kcol.tbl"
    !
    create/table {zerotbl} 3 1 null
    create/column {zerotbl} :z_{filter} "mag" F9.3 R*4
    create/column {zerotbl} :zsig_{filter} "mag" F9.3 R*4
    create/column {zerotbl} :mode "#ofUsedCalTbls" I4 I*4
    {zerotbl},:z_{filter},@1 = {zmea({outi})}
    {zerotbl},:zsig_{filter},@1 = {zsig({outi})}
    {zerotbl},:mode,@1 = {ncoef}
    add/icat {zerocat} {zerotbl} >Null
    if ncoef .eq. 0 .or. ncoef .eq. 1 then
      create/table {kexttbl} 2 1 null
      create/column {kexttbl} :kext_{filter} "mag" F9.{EXTACC} R*4
      create/column {kexttbl} :kextsig_{filter} "mag" F9.{EXTACC} R*4
      {kexttbl},:kext_{filter},@1 = {kextmea({outi})}
      {kexttbl},:kextsig_{filter},@1 = {kextsig({outi})}
      add/icat {kextcat} {kexttbl} >Null
    endif
    if ncoef .eq. 0 then
      create/table {kcoltbl} 2 1 null
      create/column {kcoltbl} :kcol_{filter} " " F9.{COLACC} R*4
      create/column {kcoltbl} :kcolsig_{filter} " " F9.{COLACC} R*4
      {kcoltbl},:kcol_{filter},@1 = {kcolmea({outi})}
      {kcoltbl},:kcolsig_{filter},@1 = {kcolsig({outi})}
      add/icat {kcolcat} {kcoltbl} >Null
    endif
    !
  endif
enddo
!
if mess_level .eq. 4 then
        mess/out 4 {PROID} " "
        inputc = "Zero points  - <{zerocat}>:"
        mess/out 4 {PROID} "{inputc}"
        inputi = m$strlen(inputc)-1
        inputc = " "
        write/key inputc/c/1/{inputi} "-" all
        mess/out 4 {PROID} "{inputc}"
        catal = 0
        zero_loop:
          store/frame outputc {zerocat} 1 end_zero_loop
          mess/out 4 {PROID} "  {outputc}"
          goto zero_loop
        end_zero_loop:
    if ncoef .eq. 0 .or. ncoef .eq. 1 then
        inputc = "Extin. coef. - <{kextcat}>:"
        mess/out 4 {PROID} "{inputc}"
        inputi = m$strlen(inputc)-1
        inputc = " "
        write/key inputc/c/1/{inputi} "-" all
        mess/out 4 {PROID} "{inputc}"
        catal = 0
        kext_loop:
          store/frame outputc {kextcat} 1 end_kext_loop
          mess/out 4 {PROID} "  {outputc}"
          goto kext_loop
        end_kext_loop:
    endif
    if ncoef .eq. 0 then
        inputc = "Colour coef. - <{kcolcat}>:"
        mess/out 4 {PROID} "{inputc}"
        inputi = m$strlen(inputc)-1
        inputc = " "
        write/key inputc/c/1/{inputi} "-" all
        mess/out 4 {PROID} "{inputc}"
        catal = 0
        kcol_loop:
          store/frame outputc {kcolcat} 1 end_kcol_loop
          mess/out 4 {PROID} "  {outputc}"
          goto kcol_loop
        end_kcol_loop:
    endif
    if nice_crea .eq. "y" then
        inputc = "Plot file:"
        mess/out 4 {PROID} "{inputc}"
        inputi = m$strlen(inputc)-1
        inputc = " "
        write/key inputc/c/1/{inputi} "-" all
        mess/out 4 {PROID} "{inputc}"
        mess/out 4 {PROID} "  {PLOT_FILE}"
    endif
endif
!
        longlog/out 1 " " >Null
        longlog/out 1 -
     "     Output tables:" >Null
do outi = 1 TOTFILT
  filter = ALLFILT({outi}:{outi})
  if m$index(goodtblfilt,filter) .gt. 0 then
    !
    zerotbl = alicat(1:{char})//"_{filter}_zero.tbl"
    kexttbl = alicat(1:{char})//"_{filter}_kext.tbl"
    kcoltbl = alicat(1:{char})//"_{filter}_kcol.tbl"
    !
        longlog/out 1 -
          "     {filter}:  zero points :            {zerotbl}" >Null
    if ncoef .eq. 0 .or. ncoef .eq. 1 then
        longlog/out 1 -
          "         extinction coefficients: {kexttbl}" >Null
    endif
    if ncoef .eq. 0 then
        longlog/out 1 -
          "         colour coefficients:     {kcoltbl}" >Null
    endif
    !
  endif
enddo
!
!
set/format
if nice_crea .eq. "y" set/graphic default
!
!
if ncoef .eq. 0 out_a = zerocat//","//kextcat//","//kcolcat
if ncoef .eq. 1 out_a = zerocat//","//kextcat
if ncoef .eq. 2 out_a = zerocat

if nice_crea .eq. "y" then 
  out_b = PLOT_FILE
else
  out_b = " "
endif

if warn .eq. 0 then
  return 0
else
  return -1
endif
!
error:
  set/format
  if m$existk("nice_crea") .eq. 1 then
    if nice_crea .eq. "y" set/graphic default
  endif
  return 1
!
!
!-------------------------------------------------------------------------------
