! $Id: fors_calcoeff_reg.prg,v 1.13 2000/12/01 16:25:02 sbogun Exp $
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
! .COPYRIGHT  (C) 2000  European Southern Observatory (ESO)
!
! .IDENT      fors_calcoeff_reg.prg
!
! .KEYWORDS   Special regression per object
!
! .PURPOSE    Calculation of extinction coefficient and zero point (no colour
!             correction) for each good object in a merged (per filter) aligned
!             photometry table. "Good" object means that "fors_calcoeff_check
!             .prg" has written a positive number into the column ":IDX".
! __________
!
! .USAGE      @p fors_calcoeff_reg
!
! .INPUT      No parameters
!             -------------
!
!             "+lower_level" variables from "fors_calcoeff.prg":
!             -------------------------------------------------
!             hiqualtbl
!             tblfilt
!             outidx
!
!             gdobjno
!             min_nstars
!
!             kext{tblfilt}
!             zcol{tblfilt}
!             kextmea
!             kextsig
!
! .OUTPUT     kext{tblfilt}(..): filled with the extinction coefficients
!                                of all stars
!
!             zcol{tblfilt}(..): filled with the zero points without colour
!                                correction of all stars
!
!             kextmea({outidx}): filled with the mean extinction coefficient
!                                of all stars with negative ext. coef.
!                                (outidx = m$index(ALLFILT,tblfilt))
!
!             kextsig({outidx}): filled with the std.dev. of the mean ext.coef.
!                                of all stars with negative ext.coef.
!
!             {hiqualtbl}: stars with a non-negative ext.coef. are indexed
!                          with "0"
!
!             Q2: 0: table is O.K.
!                 1: table should be skipped because too few stars are left
!                    (<{min_nstars})
!
! .RETURN     Q1:  0: successful return
!                 -1: a warning message occured
!                  1: an error occured
! __________
!
! .ALGORITHM  USED FORMULA:
!                 mag_lit -mag_mea(filter) = z +k_ext*X +k_col*col(P2)
!                                            \__________/
!                                                      = z_col
!                 \______________________/   |        |        |
!                            |               |        |        |
!             Tbl. columns:  :STD_INS        :Z       :AIRMASS e.g.:U_B
!                                                        |
!                                                        :Z_COL
!
! .COMMENT    Special procedure for calib. coeffs. - not to be used generally !
!
!             Used UNIX command:
!             $ /bin/echo (### as long as mess/out is still too slow)
!
! .ENVIRON    -
!
! .LANGUAGE   MIDAS 99NOV
!
! .AUTHOR     N.Rainer  DPG/DMD/ESO
!
! .VERSION    1.5  00/02/18  Creation
!                  00/       Release
!
!------------------------------------------------------------------------------
!
!
!  For `set/format':
define/local FMTDEF/c/1/17 "I1 E14.7,E23.15"
!
!  For `mess/out':
define/local PROID/c/1/40 "calcoeff_reg"
if m$existk("mess_level") .eq. 0 write/keyword mess_level/i/1/1 3
!
!  For return value:
define/local warn/i/1/1 0
write/keyword q1 0
write/keyword q2 0
!
!
if p1 .eq. "??" then
  w/o "      *****************************************************************"
  w/o "      No parameters"
  w/o "      -------------"
  w/o " "
  w/o "      "+lower_level" variables from "fors_calcoeff.prg":"
  w/o "      -------------------------------------------------"
  w/o "      hiqualtbl"
  w/o "      tblfilt"
  w/o "      outidx"
  w/o " "
  w/o "      gdobjno"
  w/o "      min_nstars"
  w/o " "
  w/o "      kext{tblfilt}"
  w/o "      zcol{tblfilt}"
  w/o "      kextmea"
  w/o "      kextsig"
  w/o "      *****************************************************************"
  return/exit
endif
!
!
set/format {FMTDEF}
!
!
!          INPUT (PARAMETERS) :
!          ******************
!
! None.
!
!          INPUT (USED "+LOWER_LEVEL" VARIABLES) :
!          *************************************
!
! See header of this procedure.
!
!
!
!          DEFINITION OF FURTHER VARIABLES :
!          *******************************
!
  define/local lenobjno/i/1/1 0
  define/local negkextobjno/i/1/1 0
  define/local objloop/i/1/1 0
  define/local mid$selidx/i/1/1 0 ? +lower_levels
  define/local firstsel/i/1/1 0
  define/local selrows/i/1/1 0
  define/local lastsel/i/1/1 0
  define/local mess_level_sav/i/1/1 3
  define/local mess_type_s/c/1/7 " " all
  define/local blanks/c/1/52 " " all
  blanks(51:52) = "\r"
  define/local clean/c/1/79 " " all
  clean(79:79) = "\\\"  ! IS EXACTLY O.K. !
  define/local mess_type_i/i/1/1 0
  define/local skip/i/1/1 0
!
!
!
!          REGRESSION :STD_INS(:AIRMASS) PER OBJECT :
!          ****************************************
!
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} -
          "  calculating extinction coefficient k_ext"
        mess/out 3 {PROID} -
          "  for each object :"
!
negkextobjno = 0
write/keyword kext{tblfilt} 0. all
write/keyword zcol{tblfilt} 0. all
kextmea({outidx}) = 0.
!
        set/format I1
        lenobjno = m$len("{gdobjno}")
!
do objloop = 1 gdobjno
  !                                                 ! check for #vals.p.obj.
  select/table {hiqualtbl} :IDX.eq.{objloop} >Null  ! done in f.c._check.prg
  firstsel = mid$selidx
  selrows = outputi
  lastsel = firstsel+selrows-1
  !
  mess_level_sav = mess_level
  mess_level = 3
  regress/general {hiqualtbl} AIRMASS,STD_INS poly,1 avg n n
  !~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  mess_level = mess_level_sav
  if {q1} .eq. 1 then
        mess/out 2 {PROID} -
          "Obj. {objloop}: Regression failed - :STD_INS(:AIRMASS) !"
    warn = 1
    goto skipped_object
  endif
  zcol{tblfilt}({objloop}) = outputd(1)
  kext{tblfilt}({objloop}) = outputd(2)
  !
  if kext{tblfilt}({objloop}) .lt. 0. then
    !
    negkextobjno = negkextobjno+1
    kextmea({outidx}) = kextmea({outidx})+kext{tblfilt}({objloop})
        inputr(1) = kext{tblfilt}({objloop})
        inputr(2) = zcol{tblfilt}({objloop})
        set/format J{lenobjno} T7.4
        ! ### mess/out still too slow !
        ! if mess_type_s .eq. "WARNING" -
        ! mess/out -3 {PROID} "{blanks}"
        ! mess/out -3 {PROID} -
        !   "    obj. {objloop}/ {gdobjno}:  k_ext= {inputr(1)}  z_col= {inputr(2)}\r"
        if mess_type_s .eq. "WARNING" -
        $ /bin/echo "{clean}\r\c"
        ! following command split because of token length (>99):
        $ /bin/echo -
          "INFO    \0133{PROID}\0135:\t\tobj. {objloop}/ {gdobjno}:\c"
        $ /bin/echo -
          "  k_ext= {inputr(1)}  z_col= {inputr(2)}\r\c"
        set/format {FMTDEF}
        mess_type_s = "INFO"
    !
  else
    !  THIS OBJECT WILL NOT BE USED:
    write/table {hiqualtbl} :IDX @{firstsel}..{lastsel} 0
        set/format J{lenobjno}
        ! ### mess/out still too slow !
        ! mess/out -2 {PROID} "{blanks}"
        ! mess/out -2 {PROID} -
        !   "    obj. {objloop}/ {gdobjno}:  k_ext>=0. => skipping obj. !\r"
        $ /bin/echo "{clean}\r\c"
        $ /bin/echo -
          "WARNING \0133{PROID}\0135:\t\tobj. {objloop}/ {gdobjno}:  k_ext>=0. => skipping obj. !\r\c"
        set/format {FMTDEF}
        mess_type_s = "WARNING"
    warn = 1
    !
  endif
  !
  skipped_object:
enddo  ! "objloop = 1 gdobjno"
      !if mess_type_s .eq. "INFO" then
        ! ### mess/out 3 {PROID}
      !elseif mess_type_s .eq. "WARNING" then
        ! ### mess/out 2 {PROID}
      !endif
      !  $ /bin/echo "{mess_type_s}"
!
!
!          STILL ENOUGH OBJECTS WITH NEGATIVE EXT.COEF. LEFT ?
!          *************************************************
!
if negkextobjno .ge. min_nstars then
  mess_type_i = 3
else
  mess_type_i = 2
endif
        mess/out {mess_type_i} {PROID} -
          "  => objects with k_ext < 0 : {negkextobjno} (min.{min_nstars})"
if negkextobjno .ge. min_nstars then
  select/table {hiqualtbl} :IDX.gt.0 >Null
else
  select/table {hiqualtbl} all
        mess/out 2 {PROID} -
          "  Too few good ones - skipping filter <{tblfilt}> !"
  warn = 1
  skip = 1
  goto skipped_entry
endif
!
!
!          CALCULATION OF MEAN EXT.COEF. :
!          *****************************
!
kextmea({outidx}) = kextmea({outidx})/negkextobjno
kextsig({outidx}) = 0.
do objloop = 1 gdobjno
  if kext{tblfilt}({objloop}) .lt. 0. then
    kextsig({outidx}) = kextsig({outidx})+ -
      (kext{tblfilt}({objloop})-kextmea({outidx}))**2
  endif
enddo
kextsig({outidx}) = m$sqrt(kextsig({outidx})/(negkextobjno-1))
        mess/out 3 {PROID} " "
        set/format T7.4
        mess/out 3 {PROID} -
          "    mean k_ext =  {kextmea({outidx})} +-{kextsig({outidx})}"
        set/format {FMTDEF}
!
!
skipped_entry:
  if warn .eq. 0 then
    return 0 {skip}
  else
    return -1 {skip}
  endif
!
error:
  return 1 -99
!
!
!------------------------------------------------------------------------------
