! $Id: fors_checkdisp.prg,v 1.11 2000/09/26 11:21:58 nrainer Exp $
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1999  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_checkdisp.prg,v $
!
!.KEYWORDS    Simple check of dispersion solution.
!
!.USAGE       check/dispersion frame disptbl slits [tolerance]
!
!.INPUT       frame       (P1)    Wavelength calibration frame.
!             disptbl     (P2)    Table with dispersion solution.
!             slits       (P3)    Slit positions table.
!             tolerance   (P4)    Tolerance of central wavelength and average
!                                 dispersion in percent.
!             slitno      (P5)    If entered and >0: In addition to errors
!                                 also information about this slit is shown
!                                 (central wavelength, linear dispersion)
!
!.OUTPUT      In OUTPUTI(1-19) -with the real slit number as index- "0" means
!             that the dispersion solution for this slit is O.K., "1" means
!             that it could not pass the quality checks and "2" that the slit
!             is closed.
!
!.RETURNS     The recipe returns 0 on success, otherwise 1. Keyword Q1 is
!             updated with the exit status. The keyword Q2 is 0 if the
!             dispersion solution is ok (in at least one MOS slit), otherwise
!             it is set to 1.
!
!.PURPOSE     The procedure compares the central wavelength and the 
!             linear dispersion of the current grism from the frame
!             header with the values computed from the dispersion relation.
!             The dispersion solution is accepted if the computed residuals
!             are smaller than the provided tolerance values.
!
!.ENVIRON     FORS context, XMOS context
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1999/03/10:  Creation
!             1.5  2000/08/01:  Calculation of central wavelength and linear
!                               dispersion improved;
!                               all slits must fail before procedure returns
!                               fail status;
!                               additional output with fail flag for each slit;
!                               text and longlog output improved (NR).
!
!------------------------------------------------------------------------------

crossref  frm disptbl slits tolerance

define/param p1 ? c "Enter calibration frame:"
define/param p2 ? c "Enter dispersion relation table:"
define/param p3 ? c "Enter slit positions table:"
define/param p4 + c "Enter tolerance values (wlen,disp):"
define/param p5 0 n "Enter slit (MOS: chosen as example): "
define/maxpar 5


! Procedure identification
define/local PROID/c/1/20      "CheckDisp"


! Local constants
define/local nm2A/r/1/1       10.
define/local mu2mm/r/1/1      1.e-3
define/local dscbase/c/1/8    "COEFSLIT"
define/local separator1/c/1/42 "." all
define/local separator2/c/1/51 "." all


! Local variables
define/local i/i/1/1         0
define/local j/i/1/1         0
define/local failed/i/1/1    0
define/local failcnt/i/1/1   0
define/local large/i/1/1     0
define/local pos1/i/1/1      0
define/local pos2/i/1/1      0
define/local nslt/i/1/1      0
define/local ncoef/i/1/1     0
define/local act_slit/i/1/1  0
define/local twlen/r/1/1     0.
define/local tdisp/r/1/1     0.
define/local fwlen/r/1/1     0.
define/local fdisp/r/1/1     0.
define/local cwlen/r/1/1     0.
define/local cdisp/r/1/1     0.
define/local reswlen/r/1/1   0.
define/local resdisp/r/1/1   0.
define/local xoff/r/1/1      0.
define/local coeff/r/1/1     0.
define/local centxy/r/1/1    0.
define/local dscname/c/1/15  " "   all
define/local frm/c/1/100     {p1}
define/local disp/c/1/100    {p2}
define/local spos/c/1/100    {p3}
define/local snum/i/1/1      {p5}
define/local xdeg/i/1/1      0
define/local degloop/i/1/1   0
define/local deriv/i/1/1     0
define/local coeffidx/i/1/1  0
define/local slittext/i/1/1  0
define/local texttype/i/1/1  0
define/local message/i/1/1   0
define/local slitchk/i/1/{FORS_MAXSLITLET} 2 all  ! "2" = "slits closed"


! Process parameters
if p4(1:1) .eq. "+" then
  twlen = tolcoef(1)
  tdisp = tolcoef(2)
else
  pos1 = m$indexb(p4,",") - 1
  if pos1 .lt. 0 then
    if m$tstno(p4) .ne. 1 then
      mess/out 1 {PROID} "Invalid tolerance values!"
      return 1
    else
      twlen = {p4}
      tdisp = tolcoef(2)
    endif
  elseif pos1 .eq. 0 then
    if m$tstno(p4(2:2)) .ne. 1 then
      mess/out 1 {PROID} "Invalid tolerance values!"
      return 1
    else
      twlen = tolcoef(1)
      tdisp = {p4(2:)}
    endif
  else
    pos2 = pos1 + 2
    if m$tstno(p4(1:{pos1})) .ne. 1 .or. m$tstno(p4({pos2}:)) .ne. 1 then
      mess/out 1 {PROID} "Invalid tolerance values!"
      return 1
    else
      twlen = {p4(1:{pos1})}
      tdisp = {p4({pos2}:)}
    endif
  endif
endif


! Get central wavelength and average dispersion from calibration
! frame and convert units
fwlen = m$value({frm},{h_wcenter}) * nm2A
fdisp = m$value({frm},{h_avdisp}) * nm2A * m$value({frm},{h_xpixsize}) * mu2mm


! Get CCD parameters from dispersion solution table
if m$existd(disp,"CENTERXY") .eq. 0 then
  mess/out 1 {PROID} "Invalid dispersion relation table encountered!"
  return 1
else
  centxy(1) = m$value({disp},CENTERXY(3))
endif
xdeg = {{disp},NORMCOORD(2)}-1


! Get number of coefficients of polynomial along dispersion axis
! (without mixed terms "x*y")
ncoef = m$value({disp},NORMCOORD(2))


! Get number of active slitlets from slit position table
nslt = m$value({spos},tblcontr(4))
if nslt .ne. nslit then
  mess/out 1 {PROID} "Invalid slit position table encountered!"
  return 1
endif



! CHECK ALL ACTIVE SLITS:
! ======================

longlog/out 1 " "
longlog/out 1 "     Checked computed dispersion relation"

do i = 1 nslt
  !
  !
  failed = 0
  slittext = 0
  xoff = m$value({spos},:{c_sloffset},{i})
  set/format I1
  act_slit = {{spos},:{c_slsno},@{i}}   ! slitlet index may differ from loop
  dscname = dscbase // "{act_slit}"     ! counter
  !
  !
  ! CENTRAL WAVELENGTH:
  ! ------------------
  inputr(1) = {{disp},CENTERXY(1)}+xoff
  inputr(2) = ({{spos},:YSTART,{i}}+{{spos},:YEND,{i}})/2.
  ! ### accuracy without search for used row near to middle row high enough ###
  apply/calib {inputr(1)},{inputr(2)} {spos} {disp} n
  cwlen = outputr(1)
  !
  ! check if computed central wavelength matches header value within tol. limit
  reswlen = (cwlen / fwlen - 1.) * 100.
  message = 0
  if m$abs(reswlen) .gt. twlen then
    texttype = 1
    message = 1
    failed = 1
  elseif act_slit .eq. snum then
    texttype = 3
    message = 1
  endif
  if message .eq. 1 then
    set/format I2 T7.2
    mess/out {texttype} {PROID} "Slit #{act_slit}:"
    slittext = 1
    mess/out {texttype} {PROID} "  Central wavelength:"
    mess/out {texttype} {PROID} -
      "    header:   {fwlen} A      allowed dev.:{twlen} %"
    mess/out {texttype} {PROID} -
      "    computed: {cwlen} A      actual dev.: {reswlen} %"
    longlog/out 1 "     Slit #{act_slit}:"
    longlog/out 1 "       Central wavelength:"
    longlog/out 1 "         header:   {fwlen} A      allowed dev.:{twlen} %"
    longlog/out 1 "         computed: {cwlen} A      actual dev.: {reswlen} %"
    set/format
  endif
  !
  !
  ! LINEAR DISPERSION:
  ! -----------------
  cdisp = m$value({disp},{dscname}(2))/centxy
  if xdeg .ge. 2 then
    do degloop = 2 xdeg
      deriv = degloop-1
      coeffidx = degloop+1
      cdisp = cdisp -
        +{degloop}. *{{disp},{dscname}({coeffidx})}/centxy *(xoff/centxy)**deriv
    enddo
  endif
  ! ### accuracy without derivations of mixed terms high enough ###
  !
  if cdisp .lt. 0 then
    if slittext .eq. 0 then
      set/format I2
      mess/out 1 {PROID} "Slit #{act_slit}:"
      slittext = 1
      texttype = 1
      longlog/out 1 "     Slit #{act_slit}:"
    endif
    mess/out 1 {PROID} "  Linear term:        <0"
    longlog/out 1 "       Linear term:        <0"
    failed = 1
    goto skipped
  endif
  !
  ! check if computed linear dispersion matches header value within tol. limit
  resdisp = (cdisp / fdisp - 1.) * 100.
  message = 0
  if m$abs(resdisp) .gt. tdisp then
    texttype = 1
    message = 1
    failed = 1
  elseif act_slit .eq. snum then
    texttype = 3
    message = 1
  endif
  if message .eq. 1 then
    if slittext .eq. 0 then
      set/format I2
      mess/out 1 {PROID} "Slit #{act_slit}:"
      slittext = 1
      longlog/out 1 "     Slit #{act_slit}:"
    endif
    set/format T7.2
    mess/out {texttype} {PROID} "  Linear term:"
    mess/out {texttype} {PROID} -
      "    header:   {fdisp} A/px.  allowed dev.:{tdisp} %"
    mess/out {texttype} {PROID} -
      "    computed: {cdisp} A/px.  actual dev.: {resdisp} %"
    longlog/out 1 "       Linear term:"
    longlog/out 1 "         header:   {fdisp} A/px.  allowed dev.:{tdisp} %"
    longlog/out 1 "         computed: {cdisp} A/px.  actual dev.: {resdisp} %"
    set/format
  endif
  !
  !
  ! HIGHER ORDERS: check if they are small compared to the linear term
  ! -------------
  if ncoef .gt. 2 then
    large = 0
    do j = 3 ncoef
      coeff = m$value({disp},{dscname}({j}))
      inputr = cdisp*centxy
      if m$abs(coeff) .ge. inputr then
        large = 1
      endif
    enddo
    if large .eq. 1 then
      if slittext .eq. 0 then
        set/format I2
        mess/out 1 {PROID} "Slit #{act_slit}:"
        slittext = 1
        texttype = 1
        longlog/out 1 "     Slit #{act_slit}:"
      endif
      set/format I1 T5.2
      inputi = centxy
      mess/out 1 {PROID} "  High order term(s):"
      mess/out 1 {PROID} "             >= {cdisp} A/pixel*{inputi}"
      longlog/out 1 "       High order term(s):"
      longlog/out 1 "                  >= {cdisp} A/pixel*{inputi}"
      set/format
      failed = 1
    endif
  endif
  !
  !
  if failed .eq. 1 then
    skipped:
    failcnt = failcnt+1
    slitchk({act_slit}) = 1
  else
    slitchk({act_slit}) = 0
  endif
  !
  !
enddo
!
!
if texttype .ne. 0 then
      mess/out {texttype} {PROID}
endif
!
write/keyword outputi -99 all
if failcnt .lt. nslt then
  copy/keyword slitchk outputi
  return 0 0
else
  return 0 1
endif
