! $Id: fors_chkcal.prg,v 1.226 2001/04/06 16:19:23 sbogun Exp $   
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
! .COPYRIGHT  (C) 1998  European Southern Observatory (ESO)
!
! .IDENT       $RCSfile: fors_chkcal.prg,v $
!
! .KEYWORDS   Statistics
!
! .PURPOSE    Determine differences between (reduced) testframes and a master
!             frame by the use of statistical tests. Outliers ("cosmics") are
!             removed within the tests.
!             At the end bad testframes are removed from the input catalogue.
!             The images and statistical plots can be displayed.
! __________
!
! .USAGE      @p fors_chkcal imkebi master type -
!                            [disp] [calc] [scan] [ps] thresh
!
! .INPUT      imkebi (P1): _im_age catalogue with input frames AND name of real
!                          _ke_yword with mean vals. of unreduced testframes AND
!                          _bi_nning factor, separated by a comma;
!                            default: none,none,none
!
!             master (P2): reference frame;
!                            default: none
!
!             type   (P3): number for type of input frames: <1>:bias,<2>:dark,
!                          <3>:flatfield (<4>-<6> are treated as <3>);
!                            default: none
!
!             disp   (P4): display selection: <p>:pipeline mode:read global key
!                          NICE_CREA [y|n], <n>:none, <l>:local, or
!                          host:display[.screen] - with the latter option
!                          you can specify the name or IP-address of a display;
!                            default: <l>
!
!             calc   (P5): calculation flag: <>:calculate statistics of frames,
!                          or <u>se descriptor STATISTIC - a missing descriptor
!                          is calculated;
!                          if an input frame contains the special descriptor
!                          {MEDDESC} (see below), it is always used as median;
!                            default: <>
!
!             scan   (P6): number of pre- AND overscan columns, separated by
!                          a comma;
!                            default: 16,16
!
!             ps     (P7): flag for creation of histogramme postscript
!                          files: <>:do not create, <y>es,create;
!                            default: <>
!                            hint: ignored, if P4=n
!
!             thresh (P8): 7 float numbers with following thresholds, separated
!                          by commas:
!                           1. median+FACTOR*mediansigma -> outlier limit,
!                           2. max. allowed PERCENTAGE of outliers,
!                           3. |MEAN| limit for binned difference frame,
!                           4. SLOPE(Poisson-Fit): x=sqrt(cnt.level) y=std.dev.,
!                           5. ZERO POINT(Poisson-Fit),
!                           6. STD.DEV. limit for binned difference frame,
!                           7. STD.DEV. limit for quadrants of diff.frame;
!                          If one of the numbers in 2./3./6./7. is "-1.",
!                          the concerning limit test is not applied.
!                            default:
!                              none
!                            examples:
!                              bias/dark      : 7.,0.5,10.,0.034,-0.13,0.75,0.4
!                              sky  flat.(IMG): 7.,1.0,70.,0.034,-0.13,10.,40.
!                              scr. flat.(IMG): 7.,0.5,10.,0.034,-0.13,2.,2.
!                              norm.flat.(LSS): 7.,0.5,10.,0.034,-0.13,4.,-1.
!                              nightflat.(IMG): 7.,1.0,10.,0.034,-0.13,35.,10.
!
! .OUTPUT     Q2        :  0: all frames are O.K.
!                          1: frames have been removed from catalogue and at
!                             least {MINFRAMES} (see below) are left
!                         -1: frames have been removed from catalogue and less
!                             then {MINFRAMES} (see below) are left
!
!             OUTPUTI(1): total number of left frames
!
!             if lower-level key USEIDX exists: contains all loop counter values
!                             of good frames
!
!             Q3        : two integers, separated by a comma:
!                         1st element:
!                          0: all frames have same sizes
!                         >0: total number of removed frames because sizes
!                             differ from size of master frame
!                         2nd element:
!                          0: all frames have positive mean
!                         >0: total number of removed frames because mean
!                             values are negative or zero
!
!             "<P1>"        : input catalogue without bad frames
!
!             "<P1>_chk.tbl": statistics -  see ALGORITHM
!
! .RETURN     Q1:  0: successful return
!                 -1: a warning message occured
!                  1: an error occured
! __________
!
! .ALGORITHM  Comparisons between
!             each binned/outlier-cleaned difference frame "(master-scalefactor*
!               testframe)", with the following scalefactors: for bias none, for
!               darks exp.time(master)/exp.time(testframe), for flats median(ma-
!               ster)/median(testframe),
!             and pre-defined, empirically received thresholds:
!               - mean (64x64 bins)
!               - standard deviation (64x64 bins)
!               - standard deviation (2x2 bins = 4 quadrants)
!               - percentage of outliers
!               The thresholds are:
!                1. a FACTOR for calculating the outlier limit: (median value of
!                   the frame + this FACTOR * the median sigma of the frame)
!                   yields the limit for detecting outliers;
!                2. maximal allowed PERCENTAGE of outliers;
!                3. limit for the ABSOLUTE value of the MEAN of the binned
!                   difference frame - difference as explained above;
!                4. the SLOPE calculated by a `Poisson fit': the abscissa is the
!                   square root of the count level of every possible raw frame
!                   type, the ordinate is the standard deviation of the counts;
!                5. the ZERO POINT (standard deviation when count level is zero)
!                   of the `Poisson fit' from 4. - can be negative;
!                6. limit for the STANDARD DEVIATION of the binned (64x64)
!                   difference frame;
!                7. limit for the STANDARD DEVIATION of the MEANS of the
!                   quadrants of the difference frame;
!             If one of the numbers in 2./3./6./7. is "-1.",
!             the concerning limit test is not applied.
!             If at least one of the tests notices that the input image is bad,
!             it is rejected.
!
! .ENVIRON    -
!
! .LANGUAGE   MIDAS 98NOV
!
! .AUTHOR     N.Rainer  DPG/DMD/ESO
!
! .VERSION    0.9  98/03/05  Creation
!                  98/07/15  Release
!             1.0  98/12/15  Release
!             1.1  99/01/14  Release
!             1.2  99/03/15  Release
!             1.3  99/09/23  Release
!
!-------------------------------------------------------------------------------
!
!
crossref imkebi master type disp calc scan ps thresh
!
!
if m$existk("MEDDESC") .eq. 0 -
  define/local MEDDESC/c/1/30 "ESO.PRO.DATAMEDI"
!
if m$existk("MINFRAMES") .eq. 0 define/local MINFRAMES/i/1/1 4  ! f. output (q2)
!
!
!  For `set/format':
define/local FMTDEF/c/1/17 "I1 E14.7,E23.15"
!
!  For `mess/out':
define/local PROID/c/1/40 "chkcal"
if m$existk("mess_level") .eq. 0 write/keyword mess_level/i/1/1 3
!
!  For `shortlog/out':
write/keyword out_a "given command/qualif not defined ..."
show/command shortlog/out | write/keyword in_a
-delete Mid{mid$sess(11:12)}1Pipe
-delete midlis{mid$sess(11:12)}.prg
if in_a .eq. out_a then
  create/command shortlog/out "write/out"
endif
!
!  For `longlog/out':
write/keyword out_a "given command/qualif not defined ..."
show/command longlog/out | write/keyword in_a
-delete Mid{mid$sess(11:12)}1Pipe
-delete midlis{mid$sess(11:12)}.prg
define/local longyes/i/1/1 1
if in_a .eq. out_a then
  create/command longlog/out "write/out"
  longyes = 0
endif
!
!  For return value:
define/local warn/i/1/1 0
write/keyword q1 0
write/keyword q2 0
write/keyword q3 0
!
!
if p1 .eq. "??" then
  w/o "      ******************************************************************"
  w/o "      imkebi (P1): _im_age catalogue with input frames AND name of real"
  w/o "                   _ke_yword with mean vals. of unreduced testframes AND"
  w/o "                   _bi_nning factor, separated by a comma;"
  w/o "                     default: none,none,none"
  w/o " "
  w/o "      master (P2): reference frame;"
  w/o "                     default: none"
  w/o " "
  w/o "      type   (P3): number for type of input frames: <1>:bias,<2>:dark,"
  w/o "                   <3>:flatfield (<4>-<6> are treated as <3>);"
  w/o "                     default: none"
  w/o " "
  w/o "      disp   (P4): display selection: <p>:pipeline mode:read global key"
  w/o "                   NICE_CREA [y|n], <n>:none, <l>:local, or"
  w/o "                   host:display[.screen] - with the latter option"
  w/o "                   you can specify the name or IP-address of a display;"
  w/o "                     default: <l>"
  w/o " "
  w/o "      calc   (P5): calculation flag: <>:calculate statistics of frames,"
  w/o "                   or <u>se descriptor STATISTIC - a missing descriptor"
  w/o "                   is calculated;"
  w/o "                   if an input frame contains the special descriptor"
  w/o "                   {MEDDESC}, it is always used as median;"
  w/o "                     default: <>"
  w/o " "
  w/o "      scan   (P6): number of pre- and overscan columns, separated"
  w/o "                   by a comma;"
  w/o "                     default: 16,16"
  w/o " "
  w/o "      ps     (P7): flag for creation of histogramme postscript"
  w/o "                   files: <>:do not create, <y>es,create;"
  w/o "                     default: <>"
  w/o "                     hint: ignored, if P4=n"
  w/o " "
  w/o "      thresh (P8): 7 float numbers with following thresholds, separated"
  w/o "                   by commas:"
  w/o "                    1. median+FACTOR*mediansigma -> outlier limit,"
  w/o "                    2. max. allowed PERCENTAGE of outliers,"
  w/o "                    3. |MEAN| limit for binned difference frame,"
  w/o "                    4. SLOPE(Poisson-Fit): x=sqrt(cnt.level) y=std.dev.,"
  w/o "                    5. ZERO POINT(Poisson-Fit),"
  w/o "                    6. STD.DEV. limit for binned difference frame,"
  w/o "                    7. STD.DEV. limit for quadrants of diff.frame;"
  w/o "                   If one of the numbers in 2./3./6./7. is "-1.","
  w/o "                   the concerning limit test is not applied."
  w/o "                     default:"
  w/o "                       none"
  w/o "                     examples:"
  w/o "                       bias/dark      : 7.,0.5,10.,0.034,-0.13,0.75,0.4"
  w/o "                       sky  flat.(IMG): 7.,1.0,70.,0.034,-0.13,10.,40."
  w/o "                       scr. flat.(IMG): 7.,0.5,10.,0.034,-0.13,2.,2."
  w/o "                       norm.flat.(LSS): 7.,0.5,10.,0.034,-0.13,4.,-1."
  w/o "                       nightflat.(IMG): 7.,1.0,10.,0.034,-0.13,35.,10."
  w/o "      ******************************************************************"
  return/exit
endif
!
!
set/format {FMTDEF}
!
!
!          INPUT :
!          *****
!
!P1(imkebi):
!----------
if p1 .eq. "?" mess/out 2 {PROID} " "
define/parameter p1 ? ? -
  "        Cat. with ima. AND key. with orig.mean vals. AND bin fact.: "
outputi = m$parse(p1,"catval")
if outputi(1) .ne. 3 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P1 !"
  goto error
endif
define/local icat/c/1/64 "{catval01}"
if m$index(icat,".cat") .eq. 0 icat = icat//".cat"
!
if m$exist(icat) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered catalogue is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({icat})"
  goto error
endif
!
show/icat {icat} n
define/local totentries/i/1/1 {outputi(1)}
if totentries .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered catalogue is empty !"
        mess/out 1 {PROID} "({icat})"
  goto error
endif
define/local meanorig/r/1/{outputi(1)} 0. all
!
define/local catal/i/1/1 0
inloop:
  store/frame inputc {icat} 1 end_inloop
  if m$exist(inputc) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "At least one catalogue entry is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({icat})"
    goto error
  endif
  goto inloop
end_inloop:
!
copy/keyword {catval02} meanorig
!
! Following "+lower" for context independent
! test version of subprocedure "fors_bindif":
define/local binfact/i/1/1 {catval03} ? +lower_levels
!
if binfact .ne. 1 .and. binfact .ne. 2 then
  if binfact .ne. 4 then
        write/out " "
        mess/out 1 {PROID} "Binning factor not supported !"
        mess/out 1 {PROID} "({binfact})"
    goto error
  endif
endif
!
!
!P2(master):
!----------
if p2 .eq. "?" mess/out 2 {PROID} " "
define/parameter p2 ? i -
  "        Enter name of master frame for comparison: "
define/local master/c/1/64 "{p2}"
if m$index(master,".bdf") .eq. 0 master = master//".bdf"
!
if m$exist(master) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered master frame is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({master})"
  goto error
endif
!
!
!P3(type):
!--------
define/parameter p3 ? ? -
  "        Enter image type [1:bias|2:dark|3:flat]:"
define/local type/i/1/1 {p3}
!
if type .ge. 4 .and. type .le. 6 then  ! for compatibility with other procedures
  type = 3
elseif type .lt. 1 .or. type .gt. 6 then
        write/out " "
        mess/out 1 {PROID} "Impossible frame type !"
  goto error
endif
!
!
!P4(disp):
!--------
define/parameter p4 l ? -
  "        Enter display [def.:<l>ocal|<p>ipeline|<n>one|host:disp[.screen]]: "
define/local dispflag/c/1/30 "{p4}"
!
!
!P5(calc):
!--------
define/parameter p5 + ? -
  "        Enter calculation flag [def.:calc_stat|<u>se_desc]:"
define/local calc/c/1/1 "{p5}"
!
!
!P6(scan):
!--------
define/parameter p6 16,16 ? -
  "        Enter pre-/overscan values [def.:16,16]: "
outputi = m$parse(p6,"scan")
if outputi(1) .ne. 2 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P6 !"
  goto error
endif
! Following "+lower" for context independent
! test version of subprocedure "fors_bindif":
define/local prescan/i/1/1 {scan01} ? +lower_levels
define/local ovescan/i/1/1 {scan02} ? +lower_levels
!
if prescan .lt. 0 .or. ovescan .lt. 0 then
        write/out " "
        mess/out 1 {PROID} "Impossible pre- or overscan value !"
  goto error
endif
!
!
!P7(ps):
!------
define/parameter p7 + ? -
  "        Enter flag for PS files of histo.s [def.:n|<y>es,create]: "
define/local ps/c/1/1 "{p7}"
!
!
!P8(thresh):
!----------
define/parameter p8 ? ? -
  "        Enter 7 thresholds (see procedure header) : "
outputi = m$parse(p8,"limit")
if outputi(1) .ne. 7 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P8 !"
  goto error
endif                                   ! ALLOWED :
define/local tf_cosm64/r/1/1 {limit01}  ! >0.
define/local tf_outp64/r/1/1 {limit02}  ! 0.-100. or -1.
define/local tf_mea64/r/1/1 {limit03}   ! >0.     or -1.
define/local tf_slope/r/1/1 {limit04}   ! >0.
define/local tf_zero/r/1/1 {limit05}    ! all
define/local tf_sig64/r/1/1 {limit06}   ! >0.     or -1.
define/local tf_sig2/r/1/1 {limit07}    ! >0.     or -1.
!
define/local badthr/i/1/1 0
if tf_cosm64 .le. 0. badthr = 1
if tf_outp64 .lt. 0. .or. tf_outp64 .gt. 100. then
   if tf_outp64 .ne. -1. badthr = 1
endif
if tf_mea64 .le. 0. .and. tf_mea64 .ne. -1. badthr = 1
if tf_slope .le. 0. badthr = 1
if tf_sig64 .le. 0. .and. tf_sig64 .ne. -1. badthr = 1
if tf_sig2 .le. 0. .and. tf_sig2 .ne. -1. badthr = 1
if badthr .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "At least one threshold is out of range !"
  goto error
endif
!
!
!
!          DEFINITION OF VARIABLES :
!          ***********************
!
!     "Preparing" :
  define/local char/i/1/2 0 all
  define/local stattable/c/1/64 " " all +lower_levels
!
!     "thresholds" :
  define/local thr_cosm64/r/1/1 0.
  define/local thr_outp64/r/1/1 0.
  define/local thr_mea64/r/1/1 0.
  define/local keyloop/i/1/1 0
  define/local thr_meaoriavg/r/1/1 0.
  define/local thr_sig64fit/r/1/1 0.
  define/local thr_sig64/r/1/1 0.
  define/local thr_sig2/r/1/1 0.
!
!     "size of master" :
  define/local mpix/i/1/2 0 all
!
!     "displaying of master" :
  define/local mlowcut/r/1/1 0.
  define/local mhighcut/r/1/1 0.
!
!     "median of master" :
  define/local mmedian/r/1/1 0.
  define/local msubxstart/i/1/1 0
  define/local msubxend/i/1/1 0
!
!     "L o o p   w i t h   t e s t" :
  define/local topicima/c/1/64 " " all
  define/local lopcount/i/1/1 0
  !
  !   "size of test" :
  define/local tpix/i/1/2 0 all
  define/local xsize/i/1/1 0
  define/local skipsize/i/1/1 0
  !
  !   "level of test" :
  define/local tsubxstart/i/1/1 0
  define/local tsubxend/i/1/1 0
  define/local tmean/r/1/1 0.
  define/local skiplevel/i/1/1 0
  !
  !   "median of test" :
  define/local tmedian/r/1/1 0.
  !
  !   "creation ... diff" :
  define/local testscale/r/1/1 0.
  define/local outpix/i/1/1 0
  define/local bdmean2q/r/1/4 0.
  define/local bdmean64frm/c/1/64 " " all
  !
  !   "statistics ... diff" :
  define/local percent/r/1/1 0.
  define/local bdmean2/r/1/1 0.
  define/local bdstddev2/r/1/1 0.
  define/local bdmean64/r/1/1 0.
  define/local bdstddev64/r/1/1 0.
  define/local abs_bdmean64/r/1/1 0.
  define/local badflag/i/1/1 0
  define/local goodcount/i/1/1 0
  !
  !   "Displaying histogrammes" :
  define/local pscr/c/1/64 " " all
!
!     "Test frames :  Removing" :
  define/local bad/i/1/1 0
  define/local rflag/i/1/1 0
  define/local all/i/1/1 0
  define/local good/i/1/1 0
!
!
!
!          MANAGING DISPLAYS / ITT / LUT :
!          *****************************
!
if dispflag .ne. "p" then
  !
  !**************************
  @p fors_nicecrea {dispflag}
  !**************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Managing of display options failed !"
    goto error
  endif
  !
endif
!
if nice_crea .eq. "y" then  ! written by fors_nicecrea.prg
  !
  if nice_dids(1) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 1 is missing !"
    goto error
  endif
  if nice_dids(2) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 2 is missing !"
    goto error
  endif
  if nice_dids(3) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 3 is missing !"
    goto error
  endif
  if nice_gids(1) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Graphic window 1 is missing !"
    goto error
  endif
  if nice_gids(2) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Graphic window 2 is missing !"
    goto error
  endif
  !
  assign/display d,1 >Null
  clear/display
  display/lut off
  label/display "MASTER" {nice_dlab}
  assign/display d,2 >Null
  clear/display
  display/lut off
  label/display "TEST" {nice_dlab}
  assign/display d,3 >Null
  clear/display
  display/lut off
  label/display "DIFFERENCE" {nice_dlab}
  load/lut heat
  load/itt ramp
  assign/graphic g,1 >Null
  clear/graphic
  assign/graphic g,2 >Null
  clear/graphic
  set/graphic default
  set/graphic pmode=0 bcolour=1
  !
endif
!
!
!          PREPARING TESTS IN LOOP :
!          ***********************
!
        mess/out 4 {PROID} "preparing statistical tests"
!
create/icat bad.cat null >Null
!
  char = m$index(icat,".cat")-1
  stattable = icat(1:{char})//"_chk.tbl"
outputi(1) = totentries+2  ! empty line + "limit" line
create/table {stattable} 11 {outputi(1)} null
!
create/column {stattable} :ACCEPT " " A6 C*6
create/column {stattable} :FRAME " " A30 C*64
create/column {stattable} :MEAN_ORIG "ADU" F10.2 R*4
create/column {stattable} :MEAN_DIFF "ADU" F9.2 R*4
create/column {stattable} :SIGM_DIFF "1_sigma_ADU" F9.2 R*4
create/column {stattable} :MEAN_A "ADU" F6.2 R*4
create/column {stattable} :MEAN_B "ADU" F6.2 R*4
create/column {stattable} :MEAN_C "ADU" F6.2 R*4
create/column {stattable} :MEAN_D "ADU" F6.2 R*4
create/column {stattable} :SIGM_ABCD "1_sigma_ADU" F9.2 R*4
create/column {stattable} :OUTL_PCENT "%" F10.3 R*4
!
!
!          CALCULATING THRESHOLDS :
!          **********************
!
        mess/out 4 {PROID} "calculating statistical limits"
!
thr_cosm64 = tf_cosm64
thr_outp64 = tf_outp64
thr_mea64 = tf_mea64
!
info/keyword meanorig
do keyloop = 1 mid$info(2)
  thr_meaoriavg = thr_meaoriavg+meanorig({keyloop})
enddo
thr_meaoriavg = thr_meaoriavg/mid$info(2)
thr_sig64fit = tf_slope * m$sqrt(thr_meaoriavg) + tf_zero  ! Poisson statistics
!
if tf_sig64 .ne. -1. then
  thr_sig64 = tf_sig64 * binfact * thr_sig64fit
else
  thr_sig64 = -1.
endif
!
thr_sig2 = tf_sig2
!
!
!          DETERMINING SIZE OF MASTER FRAME :
!          ********************************
!
! FOR "DETERMINING MEDIAN" AND "LOOP WITH TEST FRAMES" :
mpix(1) = {{master},NPIX(1)}
mpix(2) = {{master},NPIX(2)}
!
!
!          DISPLAYING OF MASTER FRAME :
!          **************************
!
if nice_crea .eq. "y" then
  !
        mess/out 4 {PROID} " "
        mess/out 4 {PROID} "displaying master frame"
        mess/out 4 {PROID} "({master})"
  !
  !******************************************************************
  @p fors_nicedisp {master} ? 7. {prescan},{ovescan} {calc} ? 3.,3. n
  !******************************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Displaying of master frame failed !"
        mess/out 1 {PROID} "({master})"
    goto error
  endif
  !
        mess/out 4 {PROID} "  loading image into display 1"
  assign/display d,1 >Null
  load/image {master} ? {outputi(1)},avg c,c
  label/display "MASTER" {nice_dlab}  ! removed by special load/image
  !
  !        FOR "LOOP WITH TEST FRAMES" :
  mlowcut = {{master},LHCUTS(1)}
  mhighcut = {{master},LHCUTS(2)}
  !
endif
!
!
!          FOR DIFFERENCE FLAT FRAME :
!          DETERMINING MEDIAN OF MASTER FRAME :
!          **********************************
!
if type .eq. 3 then  ! only needed for flats
  !
        mess/out 4 {PROID} "determining median of master frame"
  !
  if m$existd(master,MEDDESC) .eq. 1 then
    !
        mess/out 4 {PROID} "  from descriptor {MEDDESC}"
    !
    mmedian = {{master},{MEDDESC}}
  else
    !
        mess/out 4 {PROID} "  by new calculation"
    !
    msubxstart = prescan+1
    msubxend = mpix(1)-ovescan
    statistics/image {master} [@{msubxstart},<:@{msubxend},>] ? ? gn >Null
    mmedian = outputr(8)
  endif
  !
endif
!
!
!
!          L O O P   W I T H   T E S T   F R A M E S :
!          * * * * * * * * * * * * * * * * * * * * *
!
!
        if mess_level .eq. 3 mess/out 3 {PROID} " "
!
catal = 0
catloop:
  store/frame topicima {icat} 1 end_catloop
  lopcount = lopcount+1
        mess/out 4 {PROID} " "
        mess/out 3 {PROID} -
          "test image {lopcount} / {totentries} : {topicima}"
  !
  {stattable},:FRAME,{lopcount} = "{topicima}"
  {stattable},:MEAN_ORIG,{lopcount} = {meanorig({lopcount})}
  !
  !
  !        CHECKING SIZE OF TEST FRAME :
  !        ***************************
  !
  tpix(1) = {{topicima},NPIX(1)}
  tpix(2) = {{topicima},NPIX(2)}
  xsize = tpix(1)-prescan-ovescan
  if xsize .ne. mpix(1) .or. tpix(2) .ne. mpix(2) then
    !
        mess/out 2 {PROID} "  Test and master frame have different sizes -"
        mess/out 2 {PROID} "  (pre-/overscan is considered)"
        mess/out 2 {PROID} "  test frame will be removed from input cat. !"
        mess/out 2 {PROID} "  ({topicima})"
    warn = 1
    !
    add/icat bad.cat {topicima} >Null
    skipsize = skipsize+1
    {stattable},:ACCEPT,{lopcount} = "-Sizes"
    goto catloop
  endif
  !
  !
  !        CHECKING LEVEL OF TEST FRAME :
  !        ****************************
  !
        mess/out 3 {PROID} "  checking count level"
  !
  if calc .eq. "u" .and. m$existd(topicima,"STATISTIC") .eq. 0 then
        mess/out 2 {PROID} "  Test frame: Descriptor "STATISTIC" is missing -"
        mess/out 2 {PROID} "  statistics will be calculated !"
        mess/out 2 {PROID} "  ({topicima})"
    warn = 1
  endif
  if calc .ne. "u" .or. m$existd(topicima,"STATISTIC") .eq. 0 then
    tsubxstart = prescan+1
    tsubxend = tpix(1)-ovescan
    statistics/image {topicima} [@{tsubxstart},<:@{tsubxend},>] ? ? sn >Null
  endif
  tmean = {{topicima},STATISTIC(3)}  ! NOT `outputr' because of above `if'
  !
  if tmean .le. 0. then  ! bias is already subtracted (if no bias frame) !
    !
        mess/out 2 {PROID} "  Test frame has negative or zero mean value -"
        mess/out 2 {PROID} "  it will be removed from input cat. !"
        mess/out 2 {PROID} "  ({topicima})"
    warn = 1
    !
    add/icat bad.cat {topicima} >Null
    skiplevel = skiplevel+1
    {stattable},:ACCEPT,{lopcount} = "-Level"
    goto catloop
  endif
  !
  !
  !        DISPLAYING OF TEST FRAME :
  !        ************************
  !
  if nice_crea .eq. "y" then
    !
        mess/out 4 {PROID} "  displaying test frame"
    !
    !********************************************************************
    @p fors_nicedisp {topicima} ? 7. {prescan},{ovescan} {calc} ? 3.,3. n
    !********************************************************************
    if {q1} .eq. -1 then
      warn = 1
    elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Displaying of test frame failed !"
        mess/out 1 {PROID} "({topicima})"
      goto error
    endif
    !
        mess/out 4 {PROID} "    loading image into display 2"
    assign/display d,2 >Null
    load/image {topicima} ? {outputi(1)},avg c,c
    label/display "TEST" {nice_dlab}  ! removed by special load/image
    !
  endif
  !
  !
  !        FOR DIFFERENCE FLAT FRAME :
  !        DETERMINING MEDIAN OF TEST FRAME :
  !        ********************************
  !
  if type .eq. 3 then  ! only needed for flats
    !
        mess/out 4 {PROID} "  determining median of test frame"
    !
    if m$existd(topicima,MEDDESC) .eq. 1 then
      !
        mess/out 4 {PROID} "    from descriptor {MEDDESC}"
      !
      tmedian = {{topicima},{MEDDESC}}
    else
      !
        mess/out 4 {PROID} "    by new calculation"
      !
      tsubxstart = prescan+1
      tsubxend = mpix(1)-ovescan
      statistics/image {topicima} [@{tsubxstart},<:@{tsubxend},>] ? ? gn >Null
      tmedian = outputr(8)
    endif
    !
  endif
  !
  !
  !        CREATION OF BINNED DIFFERENCE FRAME :
  !        ***********************************
  !
        mess/out 3 {PROID} "  creating binned difference frame (64x64 bins)"
  !
  branch type 1,2,3 bias,dark,flat
    bias:
      testscale = 1.
      goto end_scaletype
    dark:
      testscale = {{topicima},O_TIME(7)} / {{master},O_TIME(7)}
      goto end_scaletype
    flat:
      testscale = tmedian / mmedian
  end_scaletype:
  !
  ! So scaling is done according to the formula :
  ! diff. = master - test/testscale
  !
  !*********************************************************************
  @p fors_bindif {master} {topicima} 64 64 {thr_cosm64} 1 1. {testscale}
  !*********************************************************************
  if {q1} .eq. -1 then  ! negative error is exception
        write/out " "
        mess/out 1 {PROID} "Creation of binned frames failed !"
    goto error
  endif
  !
  outpix = {q1}
  bdmean2q(1) = outputd(1)  ! means of quadrants:
  bdmean2q(2) = outputd(2)  ! 3(C) | 4(D)
  bdmean2q(3) = outputd(3)  ! -----+-----
  bdmean2q(4) = outputd(4)  ! 1(A) | 2(B)
  bdmean64frm = out_a
  !
  !
  !        DISPLAYING OF BINNED DIFFERENCE FRAME :
  !        *************************************
  !
  if nice_crea .eq. "y" then
    !
        mess/out 3 {PROID} "  displaying difference frame"
    !
    !*************************************************
    @p fors_nicedisp {bdmean64frm} ? n 0,0 ? ? 2.,2. n  ! no pre-/overscan
    !*************************************************
    if {q1} .eq. -1 then
      warn = 1
    elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Displaying of difference frame failed !"
        mess/out 1 {PROID} "({bdmean64frm})"
      goto error
    endif
    !
        mess/out 4 {PROID} "    loading image into display 3"
    assign/display d,3 >Null
    if outputi(1) .le. -2 then
      load/image {bdmean64frm} ? {outputi(1)},avg c,c
    else
      load/image {bdmean64frm} ? {outputi(1)} c,c
    endif
    label/display "DIFFERENCE" {nice_dlab}  ! removed by special load/image
    !
  endif
  !
  !
  !        STATISTICS OF BINNED DIFFERENCE FRAME
  !           FOR DETERMINATION OF GOODNESS :
  !        *************************************
  !
        mess/out 3 {PROID} "  calculating statistics of difference frame"
        mess/out 4 {PROID} "    checking number of outliers"
  !
  outputr(1) = outpix
  percent = outputr(1)/((tpix(1)-prescan-ovescan)*tpix(2))*100.
  {stattable},:OUTL_PCENT,{lopcount} = {percent}
  !
        mess/out 4 {PROID} "    checking quadrants"
  !
  {stattable},:MEAN_A,{lopcount} = {bdmean2q(1)}
  {stattable},:MEAN_B,{lopcount} = {bdmean2q(2)}
  {stattable},:MEAN_C,{lopcount} = {bdmean2q(3)}
  {stattable},:MEAN_D,{lopcount} = {bdmean2q(4)}
  bdmean2 = (bdmean2q(1)+bdmean2q(2)+bdmean2q(3)+bdmean2q(4))/4.
  bdmean2q(1) = (bdmean2q(1)-bdmean2)**2
  bdmean2q(2) = (bdmean2q(2)-bdmean2)**2
  bdmean2q(3) = (bdmean2q(3)-bdmean2)**2
  bdmean2q(4) = (bdmean2q(4)-bdmean2)**2
  bdstddev2 = m$sqrt( (bdmean2q(1)+bdmean2q(2)+bdmean2q(3)+bdmean2q(4))/3. )
  {stattable},:SIGM_ABCD,{lopcount} = {bdstddev2}
  !
        mess/out 4 {PROID} "    checking all 64x64 bins"
  !
  statistics/image {bdmean64frm} + ? ? sn >Null  ! no pre-/overscan present
  bdmean64 = outputr(3)
  bdstddev64 = outputr(4)
  abs_bdmean64 = m$abs(bdmean64)
  {stattable},:MEAN_DIFF,{lopcount} = {bdmean64}
  {stattable},:SIGM_DIFF,{lopcount} = {bdstddev64}
  !
  !
  badflag = 0
  !
  if percent .gt. m$value(thr_outp64) -
    .and. m$value(thr_outp64) .ne. -1. badflag = 1
  if bdstddev2 .gt. m$value(thr_sig2) -
    .and. m$value(thr_sig2) .ne. -1. badflag = 1
  if abs_bdmean64 .gt. m$value(thr_mea64) -
    .and. m$value(thr_mea64) .ne. -1. badflag = 1
  if bdstddev64 .gt. m$value(thr_sig64) -
    .and. m$value(thr_sig64) .ne. -1. badflag = 1
  !
  if badflag .eq. 1 then
    !
        mess/out 3 {PROID} "    bad {topicima}"
        mess/out 3 {PROID} "    will be REMOVED from input catalogue"
    !
    add/icat bad.cat {topicima} >Null
  else
    !
        mess/out 3 {PROID} "    {topicima} is O.K."
    !
    if m$existk("useidx") .eq. 1 then  ! lower-level key from calling procedure
      goodcount = goodcount+1
      useidx({goodcount}) = lopcount
    endif
    {stattable},:ACCEPT,{lopcount} = "   #  "
  endif
  !
  !
  !        DISPLAYING HISTOGRAMMES :
  !        ***********************
  !
  if nice_crea .eq. "y" then
    !
    create/icat histo.cat null >Null
    add/icat histo.cat {master} >Null
    add/icat histo.cat {topicima} >Null
    !
    !***********************************************************
    @p fors_nicehist histo.cat 7. {prescan},{ovescan} {calc} ? -
      5.,5. s,{testscale} n
    !***********************************************************
    if {q1} .eq. -1 then
      warn = 1
    elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Creation of histogrammes failed !"
        mess/out 1 {PROID} "({master}/{topicima})"
      goto error
    endif
    !
    assign/graphic g,1 >Null
        mess/out 4 {PROID} "  plotting master histogramme (red)"
    set/graphic colour=2
    plot/histogram {master}
        mess/out 4 {PROID} "  plotting test histogramme (yellow)"
    set/graphic colour=5
    overplot/histogram {topicima}
    !
    !***************
    @p fors_nicegrap
    !***************
    if {q1} .eq. -1 then
      warn = 1
    elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Label of histogrammes failed !"
        mess/out 1 {PROID} "({master}/{topicima})"
      goto error
    endif
    set/graphic colour=3
    label/graphic "{master} (red)  {topicima} (yellow)" {nice_glabw}
    !
    if ps .eq. "y" then
        mess/out 4 {PROID} "  creating postscript file of master/test hist."
      assign/graphic g,1 >Null
      copy/graphic postscript
        char = m$index(master,".bdf")-1
        char(2) = m$index(topicima,".bdf")-1
        pscr = master(1:{char}) // "__" // topicima(1:{char(2)}) // ".ps"
      -rename postscript.ps {pscr}
    endif
    !
  endif
  !
  if nice_crea .eq. "y" then
    !
    create/icat histo.cat null >Null
    add/icat histo.cat {bdmean64frm} >Null
    !
    !**************************************************
    @p fors_nicehist histo.cat n 0,0 {calc} ? 5.,5. ? n  ! no pre-/overscan
    !**************************************************
    if {q1} .eq. -1 then
      warn = 1
      if "{{out_a},:total_flat,1}" .eq. "x" goto hist_end
    elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Creation of histogramme failed !"
        mess/out 1 {PROID} "({bdmean64frm})"
      goto error
    endif
    !
    assign/graphic g,2 >Null
        mess/out 4 {PROID} "  plotting difference histogramme"
    set/graphic colour=7
    plot/histogram {bdmean64frm}
    !
    !***************
    @p fors_nicegrap
    !***************
    if {q1} .eq. -1 then
      warn = 1
    elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Label of histogramme failed !"
        mess/out 1 {PROID} "({bdmean64frm})"
      goto error
    endif
    set/graphic colour=3
    label/graphic "{bdmean64frm}  {lopcount} / {totentries}" {nice_glabw}
    !
    if ps .eq. "y" then
        mess/out 4 {PROID} "  creating postscript file of diff. hist."
      assign/graphic g,2 >Null
      copy/graphic postscript
        pscr = master(1:{char}) // "__" // topicima(1:{char(2)}) // "__diff.ps"
      -rename postscript.ps {pscr}
    endif
    !
  endif
  hist_end:
  !
  !
  !        RESETTING ORIGINAL MEDIAN VALUE IN
  !        STATISTIC DESCRIPTORS (TEST FRAMES) :
  !        ***********************************
  !
  if m$existd(topicima,MEDDESC) .eq. 1 then
    {topicima},STATISTIC(8) = {tmedian}
  endif
  !
  !
  goto catloop
end_catloop:
!
!
-delete histo.cat
-delete histo_his.tbl
!
!
!          RESETTING ORIGINAL MEDIAN VALUE IN
!          STATISTIC DESCRIPTOR (MASTER FRAME) :
!          ***********************************
!
if m$existd(master,MEDDESC) .eq. 1 -
  .and. m$existd(master,"STATISTIC") .eq. 1 then
  {master},STATISTIC(8) = {mmedian}
endif
!
!
!          E N D   O F   L O O P   W I T H   T E S T   F R A M E S
!          * * * * * * * * * * * * * * * * * * * * * * * * * * * *
!
!
!
!          TEST FRAMES :  REMOVING BAD ONES FROM INPUT CATALOGUE :
!          *****************************************************
!
lopcount = lopcount+2
{stattable},:FRAME,{lopcount} = "              LIMITS (absolut)"  ! 14 blanks
if m$value(thr_mea64) .ne. -1. {stattable},:MEAN_DIFF,{lopcount} = {thr_mea64}
if m$value(thr_sig64) .ne. -1. {stattable},:SIGM_DIFF,{lopcount} = {thr_sig64}
if m$value(thr_sig2) .ne. -1. {stattable},:SIGM_ABCD,{lopcount} = {thr_sig2}
if m$value(thr_outp64) .ne. -1. -
  {stattable},:OUTL_PCENT,{lopcount} = {thr_outp64}
!
if mess_level .ge. 3 then
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "Statistical values:"
        write/out " "
  read/table {stattable}
        write/out " "
endif
!
if longyes .eq. 1 then
  @p fors_chkcal,longlog
endif
!
show/icat bad.cat n
bad = outputi(1)
!
    longlog/out 1 " " >Null
!
if bad .eq. 0 then
        mess/out 3 {PROID} "Result: All frames are O.K."
        mess/out 3 {PROID} "------"
        longlog/out 1 "     Result: All frames are O.K." >Null
        longlog/out 1 "     ------" >Null
  !
  rflag = 0
else
        mess/out 4 {PROID} "removing bad test frames from input catalogue"
  !
  subtract/icat {icat} bad.cat >Null
  -delete dirfile.ascii
  show/icat {icat} n
  good = outputi(1)
  !
        mess/out 3 {PROID} -
               "Result: Frames: {totentries} / bad: {bad} / good: {good}"
        mess/out 3 {PROID} "------"
        longlog/out 1 -
          "     Result: Frames: {totentries} / bad: {bad} / good: {good}" >Null
        longlog/out 1 "     ------" >Null
  !
  if good .ge. MINFRAMES then
    rflag = 1
  else
    rflag = -1
  endif
endif
-delete bad.cat
!
shortlog/out 1 "Determined diff. between test and master frames" >Null
!
if skipsize .gt. 0 then
        mess/out 2 {PROID} -
          "{skipsize} frames were removed from input catalogue"
        mess/out 2 {PROID} "because sizes are not equal to master frame !"
  warn = 1
endif
if skiplevel .gt. 0 then
        mess/out 2 {PROID} -
          "{skiplevel} frames were removed from input catalogue"
        mess/out 2 {PROID} "because mean values are negative !"
  warn = 1
endif
!
!
!          OUTPUT :
!          ******
!
        mess/out 4 {PROID} " "
        mess/out 4 {PROID} ">>> Statistics are in <{stattable}>. <<<"
!
if nice_crea .eq. "y" .and. ps .eq. "y" then
        mess/out 4 {PROID} ">>> PS files of the plots are <*__*.ps>. <<<"
endif
!
!
set/format
if nice_crea .eq. "y" set/graphic default
!
!
if warn .eq. 0 then
  return 0 {rflag} {skipsize},{skiplevel}
else
  return -1 {rflag} {skipsize},{skiplevel}
endif
!
error:
  set/format
  if m$existk("nice_crea") .eq. 1 then
    if nice_crea .eq. "y" set/graphic default
  endif
  return 1 -99 -99,-99
!
!
!-------------------------------------------------------------------------------
!
entry longlog
!
  define/local tmpcode/i/1/2 0 all
  define/local limitrow/i/1/1 0
  define/local txtline/i/1/1 0
  define/local oneline/c/1/128 " " all
  define/local head/c/1/79 " " all
  define/local minus/c/1/79 " " all
  define/local allimits/c/1/79 " " all
  define/local lastrow/i/1/1 0
  !
  assign/print file tmp.txt
  print/table {stattable} 128
  assign/print terminal
  !
  open/file tmp.txt read tmpcode
  if tmpcode(1) .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Unable to open tmp.txt !"
    goto error
  endif
  show/table {stattable} >Null
  limitrow = outputi(2)+3
  do txtline = 1 limitrow
    read/file {tmpcode(1)} oneline
    if txtline .eq. 2 head = oneline(49:)
    if txtline .eq. 3 minus = oneline(49:)
    if txtline .eq. limitrow allimits = oneline(49:)
  enddo
  close/file {tmpcode(1)}
  longlog/out 1 "     {head}" >Null
  longlog/out 1 "     {minus}" >Null
  longlog/out 1 "     LIMITS (absolut):" >Null
  longlog/out 1 "     {allimits}" >Null
  longlog/out 1 "     {minus}" >Null
  open/file tmp.txt read tmpcode
  lastrow = limitrow-2
  do txtline = 1 lastrow
    read/file {tmpcode(1)} oneline
    if txtline .ge. 4 then
      !
      out_a = oneline(14:14)
      out_b = oneline(18:47)
      if out_a .eq. "#" then
        longlog/out 1 "     ACCEPTED: {out_b}:" >Null
      else
        longlog/out 1 "     REJECTED: {out_b}:" >Null
      endif
      longlog/out 1 "     {oneline(49:)}" >Null
      !
    endif
  enddo
  close/file {tmpcode(1)}
  longlog/out 1 "     {minus}" >Null
  !
  -delete tmp.txt
  !
  longlog/out 1 "     Table with statistics: {stattable}" >Null
!
!
!-------------------------------------------------------------------------------
