! $Id: fors_chkobs.prg,v 1.40 2001/04/06 16:19:23 sbogun Exp $   
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
! .COPYRIGHT  (C) 1999  European Southern Observatory (ESO)
!
! .IDENT       $RCSfile: fors_chkobs.prg,v $
!
! .KEYWORDS   Spectroscopical quality
!
! .PURPOSE    Compute signal quality values of each object spectrum, and that
!             seeing (FWHM) per wavelength bin and signal-to-noise per
!             wavelength.
! __________
!
! .USAGE      @p fors_chkobs red sky dim3 [binsz] [descs] [disp] [scan]
!
! .INPUT      red   (P1): input standard (or science) frame (reduced,
!                         distortion corrected);
!                           default: none
!
!             sky   (P2): input sky background frame (distortion corrected);
!                           default: none
!
!             dim3  (P3): three-dimensional input spectroscopy table AND names
!                         of columns with slit_no./r.a./dec./wavelengths/fluxes/
!                         obj_y-low_coords./obj_y-high_coords., separated by
!                         commas;
!                           default: none,Obj_Slit,RA_DEG,DEC_DEG,LAMBDA,FLUX,
!                                    Obj_Strt,Obj_End
!
!             binsz (P4): number of x-pixels per wavelength bin for computation
!                         of FWHM, {MINFITXPIX} (see below) <= binsz << total
!                         no. of x-pixels;
!                           default: <c>:sqrt(total no. of x-pixels), at least
!                                    {MINFITXPIX}
!
!             descs (P5): names of descriptors with readout noise (e-/pixel) AND
!                         electrons to ADU conversion factor, separated by a
!                         comma;
!                           default: RON,CONAD
!
!             disp  (P6): display selection: <p>:pipeline mode:read global key
!                         NICE_CREA [y|n], <n>:none, <l>:local, or
!                         host:display[.screen] - with the latter option
!                         you can specify the name or IP-address of a display;
!                           default: <l>
!
!             scan  (P7): number of pre- AND overscan columns, separated by
!                         a comma;
!                           default: 16,16
!                           hint: ignored, if P6=n
!
!             P2 is copied to <P1(no ext.)>_skymap.bdf
!             P3 is copied to <P1(no ext.)>_in3d.tbl
!
! .OUTPUT     "<P1>_skysub.bdf":  sky-subtracted frame
!
!             "<P1>_qc3d.tbl"  :  This is a three-dimensional table !
!                                 It is built as follows:
!
!                                     _  array (values of columns;
!                                     /| depth = tot. no. of <P1>'s x-axis pix.)
!                                    /
!                                   +-------> columns (see below)
!                                   |
!                                   V rows (objects)
!
!                                 The object planes can be converted to normal
!                                 two-dimensional tables with the help of
!
!                                   extract/t3d 3Dtbl 3Dcol 3Drow 2Dtbl 2Dcol +
!
!                                 for each column; `2Dtbl' will be created.
!
!               The columns are:  1 SLIT    : Slit number
!                                 2 RA,
!                                 3 DEC     : Positions
!                                 4 LAMBDA  : WAVELENGTHS
!                                 5 YSTART,
!                                 6 YEND    : Y-RANGES OF OBJECTS
!                                -----------
!                                 7 FLUX,     (sky-subtracted)
!                                 8 SKY     : BRIGHTNESSES
!                                -----------
!                                 9 LAM_CEN : Central wavelenghts of bins
!                                10 STATUS  : BIN QUALITY (mnd.f.FWHM,opt.f.S2N)
!                                11 FWHM    : SEEING (per bin)
!                                12 S2N     : SIGNAL-TO-NOISE (per wavelength)
!
!               Legend for column 10 (bin qual.flags):
!
!                 <0 : estimation failed (during Gauss fit)
!                  0 : O.K.
!                  1 : ambiguous sources
!                  2 : no source
!                  3 : iteration failed
!                 >3 : routine error
!
!               Formula for column 12 (signal-to-noise):
!
!                 All fluxes and readout noise in e- , w. original expo. time ,
!                 the object and sky fluxes are integrated (per wavelength,
!                 perpendicular to dispersion direction):
!
!                                     flux_obj.
!                 ----------------------------------------------------
!                 sqrt( flux_obj. + flux_sky + #ypix./wave.*(RON**2) )
!
!             Q2: total number of skipped objects
!
! .RETURN     Q1:  0: successful return
!                 -1: a warning message occured
!                  1: an error occured
! __________
!
! .ALGORITHM  -
!
! .ENVIRON    -
!
! .LANGUAGE   MIDAS 98NOV
!
! .AUTHOR     N.Rainer  DPG/DMD/ESO
!
! .VERSION    1.2  99/02/09  Creation
!                  99/03/15  Release
!             1.3  99/09/23  Release
!
!-------------------------------------------------------------------------------
!
!
crossref red sky dim3 binsz descs disp scan
!
!
if m$existk("MINFITXPIX") .eq. 0 -
  define/local MINFITXPIX/i/1/1 27 ? +lower_levels
!                             ! cen/gau min.(x,y): 3 pixel ; better: 27
define/local MAXFITNPIX/i/1/1 40000 ? +lower_levels
!                             ! cen/gau max.(x*y): 40000 pixel
!
!  For `set/format':
define/local FMTDEF/c/1/17 "I1 E14.7,E23.15"
!
!  For `mess/out':
define/local PROID/c/1/40 "chkobs"
if m$existk("mess_level") .eq. 0 write/keyword mess_level/i/1/1 3
!
!  For return value:
define/local warn/i/1/1 0
write/keyword q1 0
write/keyword q2 0
!
!
if p1 .eq. "??" then
  set/format I1
  w/o "      ******************************************************************"
  w/o "      red   (P1): input standard (or science) frame (reduced,"
  w/o "                  distortion corrected);"
  w/o "                    default: none"
  w/o " "
  w/o "      sky   (P2): input sky background frame (distortion corrected);"
  w/o "                    default: none"
  w/o " "
  w/o "      dim3  (P3): three-dimensional input spectroscopy table AND names"
  w/o "                  of columns with slit_no./r.a./dec./wavelengths/fluxes/"
  w/o "                  obj_y-low_coords./obj_y-high_coords., separated by"
  w/o "                  commas;"
  w/o "                    default: none,Obj_Slit,RA_DEG,DEC_DEG,LAMBDA,FLUX,"
  w/o "                             Obj_Strt,Obj_End"
  w/o " "
  w/o "      binsz (P4): number of x-pixels per wavelength bin for computation"
  w/o "                  of FWHM, {MINFITXPIX} <= binsz << total no. of"
  w/o "                  x-pixels;"
  w/o "                    default: <c>:sqrt(total no. of x-pixels), at least"
  w/o "                             {MINFITXPIX}"
  w/o " "
  w/o "      descs (P5): names of descriptors with readout noise (e-/pixel) AND"
  w/o "                  electrons to ADU conversion factor, separated by a"
  w/o "                  comma;"
  w/o "                    default: RON,CONAD"
  w/o " "
  w/o "      disp  (P6): display selection: <p>:pipeline mode:read global key"
  w/o "                  NICE_CREA [y|n], <n>:none, <l>:local, or"
  w/o "                  host:display[.screen] - with the latter option"
  w/o "                  you can specify the name or IP-address of a display;"
  w/o "                    default: <l>"
  w/o " "
  w/o "      scan  (P7): number of pre- AND overscan columns, separated by"
  w/o "                  a comma;"
  w/o "                    default: 16,16"
  w/o "                    hint: ignored, if P6=n"
  w/o " "
  w/o "      P2 is copied to <P1(no ext.)>_skymap.bdf"
  w/o "      P3 is copied to <P1(no ext.)>_in3d.tbl"
  w/o "      ******************************************************************"
  set/format
  return/exit
endif
!
!
set/format {FMTDEF}
!
!
!          INPUT :
!          *****
!
!P1(red):
!-------
if p1 .eq. "?" mess/out 2 {PROID} " "
define/parameter p1 ? i -
  "        Enter name of reduced frame: "
define/local redima/c/1/64 "{p1}"
if m$index(redima,".bdf") .eq. 0 redima = redima//".bdf"
!
if m$exist(redima) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered reduced frame is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({redima})"
  goto error
endif
!
!
!P2(sky):
!-------
if p2 .eq. "?" mess/out 2 {PROID} " "
define/parameter p2 ? i -
  "        Enter name of sky background frame: "
if m$index(p2,".bdf") .eq. 0 p2 = p2//".bdf"
!
if m$exist(p2) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered sky background frame is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({p2})"
  goto error
endif
!
! COPY IMAGE:
define/local backima/c/1/64 " " all
  define/local char/i/1/1 0
  char = m$indexb(redima,".")-1
  backima = redima(1:{char})//"_skymap.bdf"
if p2 .ne. backima -copy {p2} {backima}
!
!
!P3(dim3):
!--------
if p3 .eq. "?" mess/out 2 {PROID} " "
define/parameter p3 ? t -
  "  3dTbl,cols [d:none,Obj_Slit,RA_DEG,DEC_DEG,LAMBDA,FLUX,Obj_Strt,Obj_End]: "
outputi = m$parse(p3,"posflx")
if outputi(1) .ne. 8 .and. outputi(1) .ne. 1 then  ! 1 allowed for col.defaults
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P3 !"
  goto error
endif
if m$index(posflx01,".tbl") .eq. 0 posflx01 = posflx01//".tbl"
!
if m$exist(posflx01) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered table is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({posflx01})"
  goto error
endif
!
! COPY TABLE:
define/local dim3tbl/c/1/64 " " all
  char = m$indexb(redima,".")-1
  dim3tbl = redima(1:{char})//"_ex3d.tbl"
if posflx01 .ne. dim3tbl -copy {posflx01} {dim3tbl}
!
if outputi(1) .eq. 8 then
  define/local cslit/c/1/16 "{posflx02}" ? +lower_levels
  define/local cra/c/1/16 "{posflx03}" ? +lower_levels
  define/local cdec/c/1/16 "{posflx04}" ? +lower_levels
  define/local clambda/c/1/16 "{posflx05}" ? +lower_levels
  define/local cflux/c/1/16 "{posflx06}" ? +lower_levels
  define/local cystart/c/1/16 "{posflx07}" ? +lower_levels
  define/local cyend/c/1/16 "{posflx08}" ? +lower_levels
  if cslit(1:1) .eq. ":" cslit = cslit(2:)
  if cra(1:1) .eq. ":" cra = cra(2:)
  if cdec(1:1) .eq. ":" cdec = cdec(2:)
  if clambda(1:1) .eq. ":" clambda = clambda(2:)
  if cflux(1:1) .eq. ":" cflux = cflux(2:)
  if cystart(1:1) .eq. ":" cystart = cystart(2:)
  if cyend(1:1) .eq. ":" cyend = cyend(2:)
elseif outputi(1) .eq. 1 then  ! defaults
  define/local cslit/c/1/16 "Obj_Slit" ? +lower_levels
  define/local cra/c/1/16 "RA_DEG" ? +lower_levels
  define/local cdec/c/1/16 "DEC_DEG" ? +lower_levels
  define/local clambda/c/1/16 "LAMBDA" ? +lower_levels
  define/local cflux/c/1/16 "FLUX" ? +lower_levels
  define/local cystart/c/1/16 "Obj_Strt" ? +lower_levels
  define/local cyend/c/1/16 "Obj_End" ? +lower_levels
endif
!
if m$existc(dim3tbl,":{cslit}") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Entered column is not present !"
        mess/out 1 {PROID} "({dim3tbl},:{cslit})"
  goto error
elseif m$existc(dim3tbl,":{cra}") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Entered column is not present !"
        mess/out 1 {PROID} "({dim3tbl},:{cra})"
  goto error
elseif m$existc(dim3tbl,":{cdec}") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Entered column is not present !"
        mess/out 1 {PROID} "({dim3tbl},:{cdec})"
  goto error
elseif m$existc(dim3tbl,":{clambda}") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Entered column is not present !"
        mess/out 1 {PROID} "({dim3tbl},:{clambda})"
  goto error
elseif m$existc(dim3tbl,":{cflux}") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Entered column is not present !"
        mess/out 1 {PROID} "({dim3tbl},:{cflux})"
  goto error
elseif m$existc(dim3tbl,":{cystart}") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Entered column is not present !"
        mess/out 1 {PROID} "({dim3tbl},:{cystart})"
  goto error
elseif m$existc(dim3tbl,":{cyend}") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Entered column is not present !"
        mess/out 1 {PROID} "({dim3tbl},:{cyend})"
  goto error
endif
!
select/table {dim3tbl} :{cslit}.ne.NULL >Null  ! removal of empty rows
copy/table {dim3tbl} tmp.tbl
-rename tmp.tbl {dim3tbl}
!
show/table {dim3tbl} >Null
define/local totobjno/i/1/1 {outputi(2)}  ! needed below, too
if totobjno .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered table is empty !"
        mess/out 1 {PROID} "({dim3tbl})"
  goto error
endif
!
!
!P4(binsz):
!---------
outputi(1) = {{redima},NPIX(1)}
outputi(2) = m$sqrt(outputi(1))
if outputi(2) .lt. MINFITXPIX outputi(2) = MINFITXPIX
define/parameter p4 {outputi(2)} n -
  "        Enter binsize [def.:{outputi(2)};=sqrt(xpix),>={MINFITXPIX}]: "
define/local binsize/i/1/1 {p4} ? +lower_levels  ! "+low." f.f._c._ex3d/_fwhm
!
if binsize .lt. MINFITXPIX .or. binsize .gt. outputi(1) then
        write/out " "
        mess/out 1 {PROID} "Binsize out of range !"
        mess/out 1 {PROID} -
                          "({binsize} ; {MINFITXPIX}<=binsize<<{outputi(1)})"
  goto error
endif
!
!
!P5(descs):
!---------
define/parameter p5 RON,CONAD ? -
  "        Enter descs. f.readout_noise/conv.fact. [def.:RON,CONAD]: "
outputi = m$parse(p5,"sig")
if outputi(1) .ne. 2 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P5 !"
  goto error
endif
define/local ron/c/1/72 "{sig01}"
define/local conad/c/1/72 "{sig02}"
!
if m$existd(redima,ron) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered desc. for readout noise does not exist -"
        mess/out 1 {PROID} "({redima},{ron})"
        mess/out 1 {PROID} "no calculation of signal-to-noise possible !"
  goto error
endif
if {{redima},{ron}} .le. 0 then
        set/format F15.14
        write/out " "
        mess/out 1 {PROID} "Impossible value for descriptor"
        mess/out 1 {PROID} "`{ron}': {{redima},{ron}}"
        set/format {FMTDEF}
  goto error
endif
!
if m$existd(redima,conad) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered desc. for convers. factor does not exist -"
        mess/out 1 {PROID} "({redima},{conad})"
        mess/out 1 {PROID} "no calculation of signal-to-noise possible !"
  goto error
endif
if {{redima},{conad}} .le. 0 then
        set/format F15.14
        write/out " "
        mess/out 1 {PROID} "Impossible value for descriptor"
        mess/out 1 {PROID} "`{conad}': {{redima},{conad}}"
        set/format {FMTDEF}
  goto error
endif
!
!
!P6(disp):
!--------
define/parameter p6 l ? -
  "        Enter display [def.:<l>ocal|<p>ipeline|<n>one|host:disp[.screen]]: "
define/local dispflag/c/1/30 "{p6}"
!
!
!P7(scan):
!--------
define/parameter p7 16,16 ? -
  "        Enter pre-/overscan values [def.:16,16]: "
outputi = m$parse(p7,"scan")
if outputi(1) .ne. 2 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P7 !"
  goto error
endif
define/local prescan/i/1/1 {scan01}
define/local ovescan/i/1/1 {scan02}
!
if prescan .lt. 0 then
        write/out " "
        mess/out 1 {PROID} "Impossible prescan value ({prescan}) !"
  goto error
endif
if ovescan .lt. 0 then
        write/out " "
        mess/out 1 {PROID} "Impossible overscan value ({ovescan}) !"
  goto error
endif
!
!
!
!          DEFINITION OF VARIABLES :
!          ***********************
!
!     "Reduced frame" :
  define/local rlowcut/r/1/1 0.
  define/local rhighcut/r/1/1 0.
!
!     "Sky-subtracted frame" :
  define/local subima/c/1/64 " " all
!
!     "L o o p  ...  o b j e c t s" :
  !
  define/local rowloop/i/1/1 0
  !
  !   "Extraction" :
  define/local dim2tbl/c/1/64 " " all
  define/local yrngpx/r/1/2 0. all
  define/local yrngwd/r/1/2 0. all
  define/local ystartpix/r/1/1 0.
  define/local yendpix/r/1/1 0.
  define/local ylow/r/1/1 0.
  define/local yhigh/r/1/1 0.
  !
  !   "Table with coords." :
  define/local bintbl/c/1/64 " " all
  !
  !   "FWHM":
  define/local skip/i/1/1 0 ? +lower_levels  ! "+low." f. f._c._fwhm
  !
  !   "Signal-to-noise":
  define/local eron/r/1/1 0.
  define/local cfadu/r/1/1 0.
  !
  !   "Output table":
  define/local qc3dtbl/c/1/64 " " all
!
!
!
!          MANAGING DISPLAYS / ITT / LUT :
!          *****************************
!
if dispflag .ne. "p" then
  !
  !**************************
  @p fors_nicecrea {dispflag}
  !**************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Managing of display options failed !"
    goto error
  endif
  !
endif
!
if nice_crea .eq. "y" then
  !
  if nice_dids(1) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 1 is missing !"
    goto error
  endif
  if nice_dids(2) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 2 is missing !"
    goto error
  endif
  if nice_dids(3) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 3 is missing !"
    goto error
  endif
  if nice_gids(1) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Graphic window 1 is missing !"
    goto error
  endif
  if nice_gids(1) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Graphic window 2 is missing !"
    goto error
  endif
  !
  assign/display d,1 >Null
  clear/display
  display/lut off
  label/display "REDUCED" {nice_dlab}
  assign/display d,2 >Null
  clear/display
  display/lut off
  label/display "BACKGROUND" {nice_dlab}
  assign/display d,3 >Null
  clear/display
  display/lut off
  label/display "SUBTRACTED" {nice_dlab}
  load/lut heat
  load/itt ramp
  assign/graphic g,1 >Null
  clear/graphic
  assign/graphic g,2 >Null
  clear/graphic
  set/graphic default
  set/graphic pmode=0 bcolour=1
  !
endif
!
!
!          DISPLAYING OF REDUCED FRAME :
!          ***************************
!
if nice_crea .eq. "y" then
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "displaying reduced frame"
        mess/out 4 {PROID} "  ({redima})"
  !
  !****************************************************************
  @p fors_nicedisp {redima} ? 7. {prescan},{ovescan} ? ? -1.5,14. n
  !****************************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Displaying of reduced frame failed !"
        mess/out 1 {PROID} "({redima})"
    goto error
  endif
  !
        mess/out 4 {PROID} "  loading reduced image into display 1"
  assign/display d,1 >Null
  load/image {redima} ? {outputi(1)},avg c,c
  label/display "REDUCED" {nice_dlab}  ! removed by special load/image
  !
  ! for background and sky-subtracted frames :
  rlowcut = {{redima},LHCUTS(1)}
  rhighcut = {{redima},LHCUTS(2)}
  !
endif
!
!
!          DISPLAYING OF BACKGROUND FRAME :
!          ******************************
!
if nice_crea .eq. "y" then
  !
        mess/out 3 {PROID} "displaying background frame"
        mess/out 4 {PROID} "  ({backima})"
  !
  !*************************************************************
  !@p fors_nicedisp {backima} ? n {prescan},{ovescan} ? ? 1.,2. n
  @p fors_nicedisp {backima} n ? ? ? ? ? n
  !*************************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Displaying of background frame failed !"
        mess/out 1 {PROID} "({backima})"
    goto error
  endif
  !
        mess/out 4 {PROID} "  loading background image into display 2"
  assign/display d,2 >Null
  load/image {backima} ? {outputi(1)},avg c,c {rlowcut},{rhighcut}
  label/display "BACKGROUND" {nice_dlab}  ! removed by special load/image
  !
endif
!
!
!          CALCULATION OF SKY-SUBTRACTED FRAME :
!          ***********************************
!
    char = m$indexb(redima,".")-1
    subima = redima(1:{char})//"_skysub.bdf"
!
        mess/out 3 {PROID} "calculating sky-subtracted frame"
        mess/out 4 {PROID} "  ({subima})"
!
compute/pixel {subima} = {redima} - {backima}
!
!
!          DISPLAYING OF SKY-SUBTRACTED FRAME :
!          **********************************
!
if nice_crea .eq. "y" then
  !
        mess/out 3 {PROID} "displaying sky-subtracted frame"
  !
  !*************************************************************
  !@p fors_nicedisp {subima} ? 7. {prescan},{ovescan} ? ? 1.,3. n
  @p fors_nicedisp {subima} n ? ? ? ? ? n
  !*************************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Displaying of sky-subtracted frame failed !"
        mess/out 1 {PROID} "({subima})"
    goto error
  endif
  !
        mess/out 4 {PROID} "  loading sky-subtracted image into display 3"
  assign/display d,3 >Null
  load/image {subima} ? {outputi(1)},avg c,c {rlowcut},{rhighcut}
  label/display "SUBTRACTED" {nice_dlab}  ! removed by special load/image
  !
endif
!
!
!
!          L O O P   W I T H   O B J E C T S :
!          * * * * * * * * * * * * * * * * *
!
!
do rowloop = 1 totobjno
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "object {rowloop} / {totobjno} :"
  !
  !                                              _  array (lam./pos./flx.-vals.;
  !                                              /| depth = {redima} x-ax. pix.)
  !        EXTRACTION OF ARRAYS FROM TOPICAL    /
  !        OBJECT PLANE IN 3D INPUT TABLE :    +-------> cols.(lambda,pos.,flux)
  !        *********************************   |
  !                                            V rows (objects)
    char = m$indexb(redima,".")-1
    dim2tbl = redima(1:{char})//"_qc2d.tbl"
  !
        mess/out 3 {PROID} "  extracting object plane from 3D input table"
        mess/out 4 {PROID} "    ({dim3tbl} -> {dim2tbl})"
  !
  !*********************************************************
  @p fors_chkobs_ex3d {dim3tbl} {rowloop} {subima} {dim2tbl}
  !*********************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Extraction from 3D table failed !"
    goto error
  endif
  !
  !outputi = m$parse(q2,"yrngpx")  ! for FWHM and S/N below
  !if outputi .ne. 2 then
  !      write/out " "
  !      mess/out 1 {PROID} "Wrong number of elements in Q2 !"
  !  goto error
  !endif
  !ystartpix = {yrngpx01}
  !yendpix = {yrngpx02}
  !
  !if m$existk(yrngwd01) .eq. 1 delete/key yrngwd01
  !if m$existk(yrngwd02) .eq. 1 delete/key yrngwd02
  !outputi = m$parse(q3,"yrngwd")  ! f. crea. of tbl. w. input coords. below
  !if outputi .ne. 2 then
  !      write/out " "
  !      mess/out 1 {PROID} "Wrong number of elements in Q3 !"
  !  goto error
  !endif
  !ylow = {yrngwd01}
  !yhigh = {yrngwd02}
  !
  write/keyword yrngpx {q2}
  ystartpix = yrngpx(1)
  yendpix = yrngpx(2)
  if ystartpix .ge. yendpix then              ! possible error in input table !
        mess/out 2 {PROID} "    Obj. y-start >= y-end in 3D input table"
        mess/out 2 {PROID} "    - skipping object !"
    skip = skip+1
    warn = 1
    goto skipped
  endif
  !
  write/keyword yrngwd {q3}
  ylow = yrngwd(1)
  yhigh = yrngwd(2)
  !
  !
  !        CREATION OF TABLE WITH COORDS. OF INPUT FRAME (Y-RANGE
  !        OF TOPICAL OBJECT AND X-COORDS. OF WAVELENGTH BINS) :
  !        ******************************************************
  !
    char = m$indexb(redima,".")-1
    bintbl = redima(1:{char})//"_bins.tbl"
  !
        mess/out 3 {PROID} "  binning wavelengths"
        mess/out 4 {PROID} "    (-> {bintbl})"
  !
  !***************************************************************
  @p fors_chkobs_crea {subima} {ylow},{yhigh} {binsize} b {bintbl}
  !***************************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Creation of table with bin positions failed !"
    goto error
  endif
  !
  !
  !        COMPUTING FWHM VALUES
  !        BY FITTING EACH BIN OF THE OBJECT
  !        (PERPENDICULAR TO DISPERSION DIRECTION) :
  !        ***************************************
  !
        mess/out 3 {PROID} "  seeing (per good bin) ..."
        mess/out 4 {PROID} "    ({bintbl} -> {dim2tbl})"
  !
  !************************************************
  @p fors_chkobs_fwhm {subima} {bintbl} {dim2tbl} -
     {rowloop},{ystartpix},{totobjno}
  !************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Computation of FWHM values failed !"
    goto error
  endif
  if {q2} .eq. 1 goto skipped
  !
  !
  !        COMPUTING SIGNAL-TO-NOISE VALUES
  !        BY CALCULATING FLUX, SKY AND RON VALUES
  !        FOR EACH WAVELENGTH PIXEL OF THE OBJECT :
  !        ***************************************
  !
        mess/out 3 {PROID} "  signal-to-noise (per wavelength) ..."
        mess/out 4 {PROID} "    (-> {dim2tbl})"
  !
  if rowloop .eq. 1 then
    eron = {{redima},{ron}}
    cfadu = {{redima},{conad}}
  endif
  !
  !****************************************************************************
  @p fors_chkobs_s2n {backima} {dim2tbl} {ystartpix},{yendpix} {eron},{cfadu} -
     {rowloop},{totobjno}
  !****************************************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Computation of S/N values failed !"
    goto error
  endif
  if {q2} .eq. 1 goto skipped
  !
  !
  !        ADDING SINGLE OBJECT TABLE
  !        TO COMMON 3D OUTPUT TABLE :
  !        **************************
  !
  !
    char = m$indexb(redima,".")-1
    qc3dtbl = redima(1:{char})//"_qc3d.tbl"
  !
        mess/out 3 {PROID} "  adding results to 3D output table"
        mess/out 4 {PROID} "    ({dim2tbl} -> {qc3dtbl})"
  !
  !***********************************************************
  @p fors_chkobs_in3d {dim2tbl} {totobjno} {qc3dtbl} {rowloop}
  !***********************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Insertion into 3D table failed !"
    goto error
  endif
  !
  !
skipped:
enddo
!
copy/dd {redima} *,5 {qc3dtbl}
copy/dd {redima} IDENT {qc3dtbl}
!
!
!          E N D   O F   L O O P   W I T H   O B J E C T S
!          * * * * * * * * * * * * * * * * * * * * * * * *
!
!
!
!          OUTPUT :
!          ******
!
        mess/out 4 {PROID} " "
        mess/out 4 {PROID} ">>> Sky-subtr. frame: {subima} <<<"
        mess/out 4 {PROID} ">>> FWHMs  and S/Ns : {qc3dtbl}   <<<"
        mess/out 3 {PROID} " "
!
!
set/format
if nice_crea .eq. "y" set/graphic default
!
!
if warn .eq. 0 then
  return 0 {skip}
else
  return -1 {skip}
endif
!
error:
  if m$existk("dim2tbl") .eq. 1 then
    if m$exist(dim2tbl) .eq. 1 select/table {dim2tbl} all >Null
  endif
  set/format
  if m$existk("nice_crea") .eq. 1 then
    if nice_crea .eq. "y" set/graphic default
  endif
  return 1 -99
!
!
!-------------------------------------------------------------------------------
