! $Id: fors_defpositions.prg,v 1.31 2001/03/29 13:29:05 sbogun Exp $   
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1998  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_defpositions.prg,v $
!
!.KEYWORDS    Define slit boundaries and offsets
!
!.USAGE       define/positions  frame slits obsmode [flength]
!
!.INPUT       frame       (P1)    Input frame
!             slits       (P2)    Input slit position table
!             obsmode     (P3)    Observing mode (LSS,MOS)
!             flength     (P4)    Focal length of collimator, camera
!
!.OUTPUT      None
!
!.RETURNS     The recipe returns 0 on success or 1 on error. Keyword Q1 is
!             updated with the exit status. 
!
!.PURPOSE     The procedure selects an entry in a slit position table. For the
!             LSS mode of the FORS instrument the selection is made using the
!             collimator name and the name of the LSS slit. The FORS LSS slits
!             cover the full field of view. Therefore the first row of the
!             slits is the first row of the image frame and the last frame row
!             indicates the end of the slit. Also the start and end
!             coordinates for the MOS slitlets are fixed values. Therefore the
!             start and end coordinates are provided by a generic slit
!             position table as in LSS mode. For both instrument modes the
!             slit offset from the detector center is calculated from the 
!             LSS and MOS setup header information of the input frame and the 
!             focal length of the collimator in use and of the camera. A set
!             readout window is taken into account.
!
!.ENVIRON     FORS context
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1998/04/27:  Creation
!                  1999/10/12   determination of representative slitlet 
!------------------------------------------------------------------------------

crossref  frame slits obsmode
define/param p1 ?    c    "Enter input frame:"
define/param p2 +    c    "Enter input slit table:"
define/param p3 LSS  c    "Enter observing mode:"
define/param p4 +    c    "Enter focal length of collimator,camera:"
define/maxpar 4


! Procedure identification
define/local procID/c/1/80      "DefinePositions"


! Local constants
define/local mu2mm/r/1/1   1.0e-3
define/local unused/d/1/1  -9999.0
define/local tmp/c/1/12    "middumma.tbl"


! Local variables
define/local i/i/1/1                        0
define/local j/i/1/1                        0
define.local tbl_max_rows/i/1/1             0
define.local cynpix/i/1/1                   0
define.local wynpix/i/1/1                   0
define/local wxstart/i/1/1                  0
define/local wxend/i/1/1                    0
define/local wystart/i/1/1                  0
define/local wyend/i/1/1                    0
define/local active/i/1/1                   0
define/local pxlsize/r/1/1                  0.
define/local flen/d/1/2                     0.   all
define/local stab/c/1/80                    " "  all
define/local sname/c/1/20                   " "  all
define/local cname/c/1/10                   " "  all
define/local obsmode/c/1/3                  " "  all
define/local xoffset/d/1/{FORS_MAXSLITLET}  0.  all
define/local slitno/i/1/{FORS_MAXSLITLET}   0   all
define/local frm/c/1/100                    {p1}



! Process parameters
if p2 .ne. "+" then
  stab = p2
else
  stab = mos
endif

obsmode = m$upper(p3(1:3))

! Get collimator name and select appropriate entries in the slit
! position table
cname = m$value({frm},{h_coll})
if p4(1:1) .ne. "+" then
  write/key flen {p4}
else
  if cname .eq. "COLL_HR" then
    flen(1) = fors_fcoll(2)
  else
    flen(1) = fors_fcoll(1)
  endif
  flen(2) = fors_fcam
endif

pxlsize = m$value({frm},{h_xpixsize}) * mu2mm

@p fors_getwin {frm}
wxstart =  - m$nint(m$value({frm},refpix(1))) * m$value({frm},{h_xwinbin})
wxend = wxstart + m$value({frm},{h_xwinn}) * m$value({frm},{h_xwinbin}) - 1
wystart = (outputi(2) - 1) * m$value({frm},{h_ywinbin}) + 1
wyend = wystart + m$value({frm},{h_ywinn}) * m$value({frm},{h_ywinbin}) - 1

compute/table {stab} :{c_slcoll} = collapse(:{c_slcoll})
compute/table {stab} :{c_slsname} = collapse(:{c_slsname})

select/tab {stab} :{c_slcoll} .eq. "{cname}*" >Null
if outputi(1) .lt. 1 then
  mess/out 1 {procID} "Collimator {cname} not defined in {stab}!"
  return 1
endif

if obsmode .eq. "LSS" then
  sname = m$value({frm},{h_slit})
  select/tab {stab} sel .and. :{c_slsname} .eq. "{sname}*" >Null
  if outputi(1) .ne. 1 then
    mess/out 1 {procID} "LSS slit >{sname}< not found in {stab}!"
    return 1
  endif

  nslit = outputi(1)

  if m$value({stab},tblcontr(4)) .ne. 1 then
    copy/table {stab} {tmp}
    -rename {tmp} {stab}
  else
    select/table {stab} all >Null
  endif

  if m$existd(frm,h_slpos) .ne. 1 then
    mess/out 1 {procID} -
       "Undetermined slit position, descriptor {h_slpos} is missing in {frm}!"
    return 1
  else
    xoffset(1) = m$value({frm},{h_slpos}) * flen(2) / (flen(1) * pxlsize)
    {stab},:{c_sloffset},@1 = xoffset(1)
    set/format F7.2
    mess/out 4 {procID} -
        "Slit >{sname}< detected: xoffset = {xoffset(1)} px"
    set/format
  endif


! Set slit boundaries to edge pixels in y-direction
  if wystart .gt. m$value({stab},:{c_slystart},1) then
    {stab},:{c_slystart},1 = {wystart}
    mess/out 3 {procID} "Lower slit boundary set to CCD pixel {wystart}."
  endif

  if wyend .lt. m$value({stab},:{c_slyend},1) then
    {stab},:{c_slyend},1 = {wyend}
    mess/out 3 {procID} "Upper slit boundary set to CCD pixel {wyend}."
  endif
else
  copy/table {stab} {tmp}
  -rename {tmp} {stab}

  nslit = 0
  tbl_max_rows = {{stab},TBLCONTR(4)}

  do i = 1 tbl_max_rows
    slitno({i}) = m$nint({{stab},:{c_slsno},@{i}})
  enddo

  slitidx = 0

  do i = 1 FORS_MAXSLITLET 
    set/format i1 t8.2
    active = m$existd(frm,h_mospos({i})) * m$existd(frm,h_moswid({i}))

    do j = 1 tbl_max_rows 
      if slitno({j}) .eq. i goto next
    enddo
   
    goto endloop

    next:
    set/format i2
    if active .eq. 1 then
      if m$value({frm},{h_moswid({i})}) .le. 0. then
        active = 0
        xoffset({i}) = unused
        mess/out 3 {procID} "Slit #{i} is closed and will be ignored"
      else
        xoffset({i}) = m$value({frm},{h_mospos({i})}) * flen(2) / (flen(1) -
                       * pxlsize)
        if  xoffset({i}) .ge. {wxstart} .and.  xoffset({i}) .le. {wxend} then
          if m$value({stab},:{c_slystart},@{j}) .ge. {wystart} -
            .and. m$value({stab},:{c_slyend},@{j}) .le. {wyend} then
            IF slitidx .le. 0 slitidx = i  ! index of representative slitlet 
                                           ! assuming that the slitwidths of all 
                                           ! activ slitlets are equal (assumption
                                           ! currently made by moscalifors.c)
            nslit = nslit + 1
            mess/out 3 {procID} "Slit #{i}: xoffset = {xoffset({i})} pixel"
          else
            mess/out 2 {procID} "Slit #{i} not on CCD and will be ignored!"
            xoffset({i}) = unused
          endif
        else
          mess/out 2 {procID} "Slit #{i} not on CCD and will be ignored!"
          xoffset({i}) = unused
        endif
      endif
    else
      xoffset({i}) = unused
      mess/out 3 {procID} "Slit #{i}: not active"
    endif

    {stab},:{c_sloffset},@{j} = xoffset({i})

  endloop:
  enddo

  if nslit .lt. 1 then
    mess/out 1 {procID} "No active slitlets found in {frm}!"
    return 1
  else
    set/format i1
    mess/out 3 {procID} "{nslit} active slits found."
    set/format

    select/tab {stab} :{c_sloffset} .gt. {unused} >Null
    copy/table {stab} {tmp}
    -rename {tmp} {stab}
  endif
endif

return 0
