! $Id: fors_echwave.prg,v 1.13 2001/04/06 16:19:23 sbogun Exp $
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
! .COPYRIGHT  (C) 1999  European Southern Observatory (ESO)
!
! .IDENT       $RCSfile: fors_echwave.prg,v $
!
! .KEYWORDS   Line identification, dispersion coefficients
!
! .PURPOSE    Determine dispersion coefficients on a lamp frame. For it
!              - extract orders on the frame,
!              - search emission lines on extracted frame,
!              either:
!              - identify lines manually and calculate 1st guess disp. coeffs.
!              or:
!              - use existing 1st guess disp.coeffs. for comparison of line
!                positions and laboratory wavelengths for line identification
!                and calculate final disp. coeffs.
!             The extracted orders, found lines, identified wavelengths,
!             residuals (identified minus calculated wavelengths) and
!             pixel sizes can be displayed.
! __________
!
! .USAGE      @p fors_echwave lampima ordtbl [thresh] [idemod] p5 [p6] -
!                             [test] [disp]
!
! .INPUT      lampima (P1): lamp frame;
!                             default: none
!
!             ordtbl  (P2): output table from "fors_findorder.prg";
!                             default: <ORDDEF> (see below)
!
!             thresh  (P3): approx. line width in dispersion direction [pixels,
!                           >0.] AND detecting threshold above(!) local back-
!                           ground [counts,>0.], separated by a comma;
!                             default: 12.,10.
!
!             idemod  (P4): mode flag: either <manu0>:manual line identification
!                           or <manu1>:creation of 1st guess dispersion coeffi-
!                           cients after manual line identification or <auto>:
!                           use 1st guess dispersion coefficients from P5;
!                             default: auto
!
!             if P4=manu1:
!             +----------
!             |
!             | absord  (P5): highest(1st) and lowest(2nd) absol. order number,
!             |               separated by a comma;
!             |                 default: none
!             !
!             ! degrees (P6): degrees for fitting x-pixels and absolute orders,
!             !               separated by a comma;
!             !                 default: 3,3
!
!             if P4=auto:
!             +---------
!             !
!             ! reftabs (P5): 1st guess reference table AND table w. laboratory
!             !               wavelengths, separated by a comma;
!             !                 default: none,none
!             !
!             ! degree  (P6): degree for fitting both x-pix. and absol. orders;
!             !                 default: 3
!
!             test    (P7): flag for testing calculated disperson coefficients
!                           via rebinning and merging: <y>es,<n>o;
!                             default: <y>
!
!             disp    (P8): display selection:<p>:pipeline mode:read global key
!                           NICE_CREA [y|n], <n>:none, <l>:local, or
!                           host:display[.screen] - with the latter option
!                           you can specify the name or IP-address of a display;
!                             default: <l>
!
! .OUTPUT     Q2: if P4=manu1: table with 1st guess dispersion coefficients
!                              (to be stored for further use in the autom. mode)
!                 if P4=auto:  table with final dispersion coefficients
!
!             Q3: minimal pixel size [angstroem]
!
! .RETURN     Q1:  0: successful return
!                 -1: a warning message occured
!                  1: an error occured
! __________
!
! .ALGORITHM  - order extraction:   see "fors_extorder.prg"
!             - line search:        see "fors_echwave_sea.prg"
!             - 1.guess disp.coef.: see "fors_echwave_gue.prg"
!             - final disp.coef.:   see "fors_echwave_dis.prg"
!
! .COMMENT    THE MANUAL MODES MUST NOT BE USED WITHIN THE PIPELINE -
!             THEIR PURPOSE IS TO FILL THE CALIBRATION DATABASE !!!
!
!             Technical information for identification modes:
!
!               invoked procedures   | manu0 | manu1 | auto
!               ---------------------+-------+-------+-----
!               fors_extorder        |   x   |       |  x
!               fors_echwave_sea     |   x   |       |  x
!                                    |       |       |
!               fors_echwave_ide (0) |   x   |       |
!               fors_echwave_ide (1) |       |   x   |
!               fors_echwave_gue     |       |   x   |
!               fors_echwave_dis     |       |       |  x
!                                    |       |       |
!               (fors_reborder)      |       |  (x)  | (x)
!               (fors_merorder)      |       |  (x)  | (x)
!
! .ENVIRON    -
!
! .LANGUAGE   MIDAS 99NOV
!
! .AUTHOR     N.Rainer  DPG/DMD/ESO
!
! .VERSION    1.4  99/10/18  Creation
!                  99/       Release
!
!------------------------------------------------------------------------------
!
!
crossref lampima ordtbl thresh idemod p5 p6 test disp
!
!
define/local ORDDEF/c/1/20 "order.tbl"
define/local MAXDEGREE/i/1/1 6  ! technically fixed to 6 - DON'T CHANGE !!!
!
!
!  For `set/format':
define/local FMTDEF/c/1/17 "I1 E14.7,E23.15"
!
!  For `mess/out':
define/local PROID/c/1/40 "echwave"
if m$existk("mess_level") .eq. 0 write/keyword mess_level/i/1/1 3
!
!  For return value:
define/local warn/i/1/1 0
write/keyword q1 0
write/keyword q2 0
write/keyword q3 0
!
!
if p1 .eq. "??" then
  w/o "      ******************************************************************"
  w/o "      lampima (P1): lamp frame;"
  w/o "                      default: none"
  w/o " "
  w/o "      ordtbl  (P2): output table from "fors_findorder.prg";"
  w/o "                      default: {ORDDEF}"
  w/o " "
  w/o "      thresh  (P3): approx. line width in dispersion direction [pixels,"
  w/o "                    >0.] AND detecting threshold above(!) local back-"
  w/o "                    ground [counts,>0.], separated by a comma;"
  w/o "                      default: 12.,10."
  w/o " "
  w/o "      idemod  (P4): mode flag: either <manu0>:manual line identification"
  w/o "                    or <manu1>:creation of 1st guess dispersion coeffi-"
  w/o "                    cients after manual line identification or <auto>:"
  w/o "                    use 1st guess dispersion coefficients from P5;"
  w/o "                      default: auto"
  w/o " "
  w/o "      if P4=manu1:"
  w/o "      +----------"
  w/o "      |"
  w/o "      | absord  (P5): highest(1st) and lowest(2nd) absol. order number,"
  w/o "      |               separated by a comma;"
  w/o "      |                 default: none"
  w/o "      !"
  w/o "      ! degrees (P6): degrees for fitting x-pixels and absolute orders,"
  w/o "      !               separated by a comma;"
  w/o "      !                 default: 3,3"
  w/o " "
  w/o "      if P4=auto:"
  w/o "      +----------"
  w/o "      !"
  w/o "      ! reftabs (P5): 1st guess reference table AND table w. laboratory"
  w/o "      !               wavelengths, separated by a comma;"
  w/o "      !                 default: none,none"
  w/o "      !"
  w/o "      ! degree  (P6): degree for fitting both x-pix. and absol. orders;"
  w/o "      !                 default: 3"
  w/o " "
  w/o "      test    (P7): flag for testing calculated disperson coefficients"
  w/o "                    via rebinning and merging: <y>es,<n>o;"
  w/o "                      default: <y>"
  w/o " "
  w/o "      disp    (P8): display selection:<p>:pipeline mode:read global key"
  w/o "                    NICE_CREA [y|n], <n>:none, <l>:local, or"
  w/o "                    host:display[.screen] - with the latter option"
  w/o "                    you can specify the name or IP-address of a display;"
  w/o "                      default: <l>"
  w/o "      ******************************************************************"
  return/exit
endif
!
!
set/format {FMTDEF}
!
!
!          INPUT :
!          *****
!
!P1(lampima):
!-----------
if p1 .eq. "?" mess/out 2 {PROID} " "
define/parameter p1 ? i -
  "        Enter name of lamp frame: "
define/local lampima/c/1/64 "{p1}"
if m$index(lampima,".bdf") .eq. 0 lampima = lampima//".bdf"
!
if m$exist(lampima) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered frame is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({lampima})"
  goto error
endif
!
!
!P2(ordtbl):
!----------
define/parameter p2 {ORDDEF} ? -
  "        Enter order positions table [def.:{ORDDEF}]: "
define/local ordtbl/c/1/60 "{p2}"
if m$index(ordtbl,".tbl") .eq. 0 ordtbl = ordtbl //".tbl"
!
if m$exist(ordtbl) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered table is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({ordtbl})"
  goto error
endif
! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
outputi = 0
if m$existd(ordtbl,"REGRESSC") .eq. 0 outputi = 1
if m$existd(ordtbl,"REGRESSI") .eq. 0 outputi = 1
if m$existd(ordtbl,"REGRESSR") .eq. 0 outputi = 1
if m$existd(ordtbl,"REGRESSD") .eq. 0 outputi = 1
if outputi(1) .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Descriptor is not present !"
        mess/out 1 {PROID} "({ordtbl},REGRESS[ C/I/R/D ])"
  goto error
endif
if m$existd(ordtbl,"ORDWIDTH") .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Descriptor is not present !"
        mess/out 1 {PROID} "({ordtbl},ORDWIDTH)"
  goto error
endif
! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
if {{ordtbl},TBLCONTR(4)} .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered table is empty !"
        mess/out 1 {PROID} "({ordtbl})"
  goto error
endif
!
!
!P3(thresh):
!----------
define/parameter p3 12.,10. ? -
  "        Enter line width,detect.thresh. [def.:12.,10.]: "
outputi = m$parse(p3,"thresh")
if outputi(1) .ne. 2 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P3 !"
  goto error
endif
define/local linwid/r/1/1 {thresh01}
define/local level/r/1/1 {thresh02}
!
! Tests of line width and detect. thresh. after creation
! of extracted frame in "fors_echwave_sea.prg".
!
!
!P4(idemod):
!----------
define/parameter p4 auto ? -
  "        Enter flag for calc. of disp. coef. [def.:auto|manu0|manu1]: "
define/local idemod/c/1/5 "{p4}"
!
if idemod .ne. "auto" then
  if idemod .ne. "manu0" .and. idemod .ne. "manu1" then
        write/out " "
        mess/out 1 {PROID} "Undefined calculation mode - "{idemod}" -"
        mess/out 1 {PROID} "enter "manu0" or "manu1" or "auto" !"
    goto error
  endif
endif
!
!
if idemod .eq. "manu1" then  ! not necessary for "manu0"
  !
  !
  !P5(absord):
  !----------
  if p5 .eq. "?" mess/out 2 {PROID} " "
  define/parameter p5 ? n -
    "        Enter highest,lowest absolute order number: "
  outputi = m$parse(p5,"order")
  if outputi(1) .ne. 2 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P5 !"
    goto error
  endif
  define/local absord/i/1/2 {order01},{order02}
  !
  if absord(1) .le. absord(2) then
        write/out " "
        mess/out 1 {PROID} "Highest abs.ord.no. ({absord(1)}) <="
        mess/out 1 {PROID} " lowest abs.ord.no. ({absord(2)}) !"
    goto error
  endif
  !
  !
  !P6(degrees):
  !-----------
  define/parameter p6 3,3 n -
    "        Enter fit degree for x-pixels,abs.orders: "
  outputi = m$parse(p6,"degree")
  if outputi(1) .ne. 2 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P6 !"
    goto error
  endif
  define/local xdeg/i/1/1 {degree01}
  define/local orderdeg/i/1/1 {degree02}
  !
  if xdeg .gt. MAXDEGREE then
        write/out " "
        mess/out 1 {PROID} "Fit degree for x-pixels too high !"
        mess/out 1 {PROID} "({xdeg} ; x-degree<={MAXDEGREE})"
    goto error
  endif
  if orderdeg .gt. MAXDEGREE then
        write/out " "
        mess/out 1 {PROID} "Fit degree for abs.orders too high !"
        mess/out 1 {PROID} "({orderdeg} ; order-degree<={MAXDEGREE})"
    goto error
  endif
  !
  ! Comparison of degrees with number of found lines after creation
  ! of line table in "fors_echwave_gue.prg".
  !
  !
elseif idemod .eq. "auto" then
  !
  !
  !P5(reftabs):
  !-----------
  if p5 .eq. "?" mess/out 2 {PROID} " "
  define/parameter p5 ? t -
    "        Enter guess_table,line_catalogue: "
  outputi = m$parse(p5,"reftabs")
  if outputi(1) .ne. 2 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P5 !"
    goto error
  endif
  define/local guetbl/c/1/60 "{reftabs01}"
  if m$index(guetbl,".tbl") .eq. 0 guetbl = guetbl //".tbl"
  define/local wavcat/c/1/60 "{reftabs02}"
  if m$index(wavcat,".tbl") .eq. 0 wavcat = wavcat //".tbl"
  !
  if m$existc(guetbl,":X") .eq. -2 then
        write/out " "
        mess/out 1 {PROID} "Entered table is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({guetbl})"
    goto error
  elseif m$existc(guetbl,":X") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Column is not present !"
        mess/out 1 {PROID} "({guetbl},:X)"
    goto error
  elseif m$existc(guetbl,":YNEW") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Column is not present !"
        mess/out 1 {PROID} "({guetbl},:YNEW)"
    goto error
  elseif m$existc(guetbl,":ORDER") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Column is not present !"
        mess/out 1 {PROID} "({guetbl},:ORDER)"
    goto error
  endif
  ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if m$existd(guetbl,"PIXEL") .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Descriptor is not present !"
        mess/out 1 {PROID} "({guetbl},PIXEL)"
    goto error
  endif
  outputi = 0
  if m$existd(guetbl,"RORDC") .eq. 0 outputi = 1
  if m$existd(guetbl,"RORDI") .eq. 0 outputi = 1
  if m$existd(guetbl,"RORDR") .eq. 0 outputi = 1
  if m$existd(guetbl,"RORDD") .eq. 0 outputi = 1
  if outputi(1) .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Descriptor is not present !"
        mess/out 1 {PROID} "({guetbl},RORD[ C/I/R/D ])"
    goto error
  endif
  if m$existd(guetbl,"ORDER") .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Descriptor is not present !"
        mess/out 1 {PROID} "({guetbl},ORDER)"
    goto error
  endif
  show/descriptor {guetbl} ORDER h
  if outputi(3) .lt. 2 then
        write/out " "
        mess/out 1 {PROID} "Descriptor has too few elements"
        mess/out 1 {PROID} "({guetbl},ORDER) - 2 needed !"
    goto error
  endif
  ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if {{guetbl},TBLCONTR(4)} .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered guess table is empty !"
        mess/out 1 {PROID} "({guetbl})"
    goto error
  endif
  statistics/table {guetbl} :ORDER >Null
  define/local absord/i/1/2 {outputr(2)},{outputr(1)}
  if absord(1) .eq. absord(2) then
        write/out " "
        mess/out 1 {PROID} "Highest abs.ord.no. ({absord(1)}) =="
        mess/out 1 {PROID} " lowest abs.ord.no. ({absord(2)}) !"
        mess/out 1 {PROID} " ({guetbl},:ORDER)"
    goto error
  endif
  ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  if m$existc(wavcat,":WAVE") .eq. -2 then
        write/out " "
        mess/out 1 {PROID} "Entered table is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({wavcat})"
    goto error
  elseif m$existc(wavcat,":WAVE") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Column is not present !"
        mess/out 1 {PROID} "({wavcat},:WAVE)"
    goto error
  endif
  inputi = absord(1)+1-absord(2)
  define/local totordno/i/1/1 {inputi(1)}
  inputi(1) = totordno*2
  inputi(2) = {{wavcat},TBLCONTR(4)}
  if inputi(2) .lt. inputi(1) then
        write/out " "
        mess/out 1 {PROID} "Too few lines present in line catalogue !"
        mess/out 1 {PROID} "({inputi(2)} ; >= {inputi(1)} (2 per order) needed)"
    goto error
  endif
  !
  !
  !P6(degree):
  !----------
  define/parameter p6 3 n -
    "        Enter fit degree for x-pixels and abs.orders: "
  define/local x_ord_deg/i/1/1 {p6}
  !
  if x_ord_deg .gt. MAXDEGREE then
        write/out " "
        mess/out 1 {PROID} "Fit degree for x-pixels and abs.orders too high !"
        mess/out 1 {PROID} "({x_ord_deg} ; degree<={MAXDEGREE})"
    goto error
  endif
  !
  ! Comparison of degree with number of found lines after creation
  ! of line table in "fors_echwave_dis.prg".
  !
  !
endif
!
!
!P7(test):
!--------
define/parameter p7 y ? -
  "        Enter flag for rebinning / merging [def.:y|n]: "
define/local test/c/1/1 "{p7}"
!
!
!P8(disp):
!--------
define/parameter p8 l ? -
  "        Enter display [def.:<l>ocal|<p>ipeline|<n>one|host:disp[.screen]]: "
define/local dispflag/c/1/30 "{p8}"
!
!
!          DEFINITION OF FURTHER VARIABLES :
!          *******************************
!
define/local extimg/c/1/64 " " all
define/local lintbl/c/1/64 " " all
if idemod .eq. "manu1" then
  define/local char/i/1/1 0
  define/local idetbl/c/1/64 " " all
endif
define/local disptbl/c/1/64 " " all
define/local pszmin/d/1/1 0.
define/local rebima/c/1/64 " " all
define/local merima/c/1/64 " " all
!
!
!
!          MANAGING DISPLAYS / ITT / LUT :
!          *****************************
!
if dispflag .ne. "p" then
  !
  !**************************
  @p fors_nicecrea {dispflag}
  !**************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Managing of display options failed !"
    goto error
  endif
  !
endif
!
if nice_crea .eq. "y" then
  !
  if nice_dids(1) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 1 is missing !"
    goto error
  endif
  if nice_dids(2) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 2 is missing !"
    goto error
  endif
  if nice_dids(3) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 3 is missing !"
    goto error
  endif
  if nice_gids(1) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Graphic window 1 is missing !"
    goto error
  endif
  if nice_gids(2) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Graphic window 2 is missing !"
    goto error
  endif
  !
  assign/display d,1 >Null
  clear/display
  display/lut off
  assign/display d,2 >Null
  clear/display
  display/lut off
  assign/display d,3 >Null
  clear/display
  display/lut off
  load/lut heat
  load/itt ramp
  assign/graphic g,1 >Null
  clear/graphic
  assign/graphic g,2 >Null
  clear/graphic
  assign/graphic g,3 >Null
  clear/graphic
  modify/graphic icon >Null
  set/graphic default
  set/graphic pmode=0
  !
endif
!
!
if idemod .eq. "manu0" .or. idemod .eq. "auto" then
  !
  !
  !        AVERAGING PIXELS WITHIN ORDERS OF LAMP FRAME
  !        PERPENDICULAR TO DISPERSION DIRECTION:
  !        ********************************************
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "extracting orders of comparison spectrum"
  !
  !************************************
  @p fors_extorder {lampima} {ordtbl} p
  !************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Extract. of orders of comparison spectrum failed !"
    goto error
  endif
  extimg = q2
  !
  !
  !        DETECTING AND CENTERING LINES ON ORDER FRAME :
  !        ********************************************
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "searching for emission lines on extracted orders"
  !
  !*****************************************************************
  @p fors_echwave_sea {lampima} {extimg} {ordtbl} {linwid},{level} p
  !*****************************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Search for emiss. lines on extr. orders failed !"
    goto error
  endif
  lintbl = q2
  !
  !
endif
!
!
if idemod .eq. "manu0" then
  !
  !
  !        MANUAL IDENTIFICATION OF WAVELENGTHS :
  !        ************************************
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "identifying wavelengths manually"
  !
  !*****************************
  @p fors_echwave_ide {lintbl} 0  ! includes "return/exit"
  !*****************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Manual identification of wavelengths failed !"
    goto error
  endif
  !
  !
elseif idemod .eq. "manu1" then
  !
  !
  !        WRITING MANUALLY IDENTIFIED WAVELENGTHS
  !        INTO COPIED LINE TABLE:
  !        ***************************************
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "using manually identified wavelengths"
  !
  char = m$indexb(lampima,".")-1
  lintbl = lampima(1:{char})//"_wav.tbl"
  !*****************************
  @p fors_echwave_ide {lintbl} 1
  !*****************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Storage of manually identified wavelengths failed !"
    goto error
  endif
  idetbl = q2
  !
  !        * FITTING OF MANUALLY IDENTIFIED LINES FOR
  !            CALCULATION OF 1ST GUESS DISPERSION COEFFS.
  !            (FOR LAMP FRAME / ORDER FRAME / PER ORDER),
  !        * CALCULATION OF PIXEL SIZES:
  !          *********************************************
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "calculating 1st guess dispersion coefficients"
  !
  char = m$indexb(lampima,".")-1
  extimg = lampima(1:{char})//"_ext.bdf"
  !**************************************************************
  @p fors_echwave_gue {extimg} {idetbl} -
                      {absord(1)},{absord(2)} {xdeg},{orderdeg} p
  !**************************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Calc. of 1st guess dispersion coefficients failed !"
    goto error
  endif
  disptbl = q2  ! ="guetbl", only "disptbl" because of further procedures below
  pszmin = {q3}
  !
  !
elseif idemod .eq. "auto" then
  !
  !
  !        * CALCULATION OF A SOLUTION OF 1ST GUESS DISP.
  !            COEFFS. FOR COMPARISON OF LINE POSITIONS,
  !        * IDENTIFICATION OF LINES VIA LINE CATALOGUE,
  !        * FITTING OF IDENTIFIED LINES FOR
  !            CALCULATION OF FINAL DISPERSION COEFFS.
  !            (FOR LAMP FRAME / ORDER FRAME / PER ORDER),
  !        * CALCULATION OF PIXEL SIZES:          
  !          *********************************************
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "calculating dispersion coefficients"
  !
  !**************************************************
  @p fors_echwave_dis {lampima} {extimg} {lintbl} -
                      {guetbl} {wavcat} {x_ord_deg} p
  !**************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Calculation of final dispers. coefficients failed !"
    goto error
  endif
  disptbl = q2
  pszmin = {q3}
  !
  !
endif
!
!
if test .ne. "n" then
  !
  !
  !        TESTING DISPERSION COEFFICIENTS VIA REBINNING :
  !        *********************************************
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "testing disp.coef.: rebinning per order"
  !
  !*********************************************
  @p fors_reborder {extimg} {disptbl} {pszmin} p
  !*********************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Rebinning per order failed !"
    goto error
  endif
  rebima = q2
  !
  !
  !        TESTING DISPERSION COEFFICIENTS VIA MERGING :
  !        *******************************************
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "testing disp.coef.: merging all orders"
  !
  !************************************
  @p fors_merorder {rebima} {disptbl} p
  !************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Merging of orders failed !"
    goto error
  endif
  merima = q2
  !
  !
endif
!
!
!          OUTPUT :
!          ******
!
if idemod .eq. "manu1" .or. mess_level .eq. 4 then
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} ">>> These are the output files : <<<"
        mess/out 3 {PROID} ">>> Extracted orders:          {extimg} <<<"
        mess/out 3 {PROID} ">>> Line table:                {lintbl} <<<"
      if idemod .eq. "manu1" then
        mess/out 3 {PROID} ">>> Manu.ident.lines:          {idetbl} <<<"
        mess/out 3 {PROID} ">>> 1.GUESS DISP.COEF. DESCS.: {disptbl} <<<"
      elseif idemod .eq. "auto" then
        mess/out 3 {PROID} ">>> DISPERSION COEFS. DESCS.:  {disptbl} <<<"
      endif
      if test .ne. "n" then
        mess/out 3 {PROID} ">>> Rebinned orders:           {rebima} <<<"
        mess/out 3 {PROID} ">>> Merged orders:             {merima} <<<"
      endif
  !
endif
!
        mess/out 3 {PROID} " "
!
!
set/format
if nice_crea .eq. "y" set/graphic default
!
!
if warn .eq. 0 then
  return 0 {disptbl} {pszmin}
else
  return -1 {disptbl} {pszmin}
endif
!
error:
  set/format
  if m$existk("nice_crea") .eq. 1 then
    if nice_crea .eq. "y" set/graphic default
  endif
  return 1 -99 -99
!
!
!------------------------------------------------------------------------------
