! $Id: fors_findorder.prg,v 1.4 2001/04/06 16:19:22 sbogun Exp $
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
! .COPYRIGHT  (C) 1999  European Southern Observatory (ESO)
!
! .IDENT       $RCSfile: fors_findorder.prg,v $
!
! .KEYWORDS   Order detection
!
! .PURPOSE    Determine positions of orders on a reference frame.
!             The found orders can be displayed.
! __________
!
! .USAGE      @p fors_findorder ima [geom] [hough] [extra] [disp]
!
! .INPUT      ima   (P1): (screen flatfield) frame for order detection;
!                           default: none
!
!             geom  (P2): distance of chosen columns AND
!                         no. of neighbour pixels for median filtering AND
!                         minimal value for first derivative AND
!                         y-width of one order AND
!                         tolerance for this width, separated by commas;
!                           default: 50,43,5,220,10
!
!             hough (P3): slope_start AND slope_end AND slope_step AND
!                         zero_point_step for the Hough transformation,
!                         separated by commas;
!                           default: 0.35,0.75,0.008,1.
!
!             extra (P4): flag for extrapolation of traced orders: <>:do not
!                         extrapolate, <y>es,extrapolate;
!                           default: <>
!
!             disp  (P5): display selection:<p>:pipeline mode:read global key
!                         NICE_CREA [y|n], <n>:none, <l>:local, or
!                         host:display[.screen] - with the latter option
!                         you can specify the name or IP-address of a display;
!                           default: <l>
!
! .OUTPUT     Q2: order detection table "order.tbl" with relative order numbers
!                 and fitted positions  AND  table "back.tbl" with background
!                 positions, separated by a comma; the informations are stored
!                 both in columns and descriptors
!
!             Q3: total number of identified orders
!
! .RETURN     Q1:  0: successful return
!                 -1: a warning message occured
!                  1: an error occured
! __________
!
! .ALGORITHM  1. Detection of order positions in chosen columns by calculating
!                the column derivatives.
!             2. Computing straight coefficients for all orders by performing
!                a Hough transformation and cluster detection and thereby
!                finding the number of orders.
!             3. Assigning relative order numbers to order positions with the
!                help of the straight coefficients.
!             4. Improving the order positions with a 2nd order regression.
!             5. Computing positions of background areas with the help of the
!                regression coefficients.
!
! .COMMENT    The orientation of the input frame MUST be as follows, as
!             "fors_findorder_lin" (Hough trafo.) only works with POSITIVE
!             slopes !!!
!
!                           -------------
!                    low   |  / /    / / | higher
!                   (red)  |/ /    / /   |
!               Abs.       |/    / /    /|     Rel.
!               ord.       |   / /    / /|     ord.
!                    high  | / /    / /  | 1
!                   (blue) |      / /    |
!                          |             |
!                           -------------
!                           low      high
!                              Lambda
!
! .ENVIRON    -
!
! .LANGUAGE   MIDAS 99NOV
!
! .AUTHOR     N.Rainer  DPG/DMD/ESO
!
! .VERSION    1.4  99/07/15  Creation
!                  99/10/21  Release
!
!------------------------------------------------------------------------------
!
!
crossref ima geom hough extra disp
!
!
!  For `set/format':
define/local FMTDEF/c/1/17 "I1 E14.7,E23.15"
!
!  For `mess/out':
define/local PROID/c/1/40 "findorder"
if m$existk("mess_level") .eq. 0 write/keyword mess_level/i/1/1 3
!
!  For return value:
define/local warn/i/1/1 0
write/keyword q1 0
write/keyword q2 0
write/keyword q3 0
!
!
if p1 .eq. "??" then
  w/o "      ******************************************************************"
  w/o "      ima   (P1): (screen flatfield) frame for order detection;"
  w/o "                   default: none"
  w/o " "
  w/o "      geom  (P2): distance of chosen columns AND"
  w/o "                  no. of neighbour pixels for median filtering AND"
  w/o "                  minimal value for first derivative AND"
  w/o "                  y-width of one order AND"
  w/o "                  tolerance for this width, separated by commas;"
  w/o "                    default: 50,43,5,220,10"
  w/o " "
  w/o "      hough (P3): slope_start AND slope_end AND slope_step AND"
  w/o "                  zero_point_step for the Hough transformation,"
  w/o "                  separated by commas;"
  w/o "                    default: 0.35,0.75,0.008,1."
  w/o " "
  w/o "      extra (P4): flag for extrapolation of traced orders: <>:do not"
  w/o "                  extrapolate, <y>es,extrapolate;"
  w/o "                    default: <>"
  w/o " "
  w/o "      disp  (P5): display selection: <p>:pipeline mode:read global key"
  w/o "                  NICE_CREA [y|n], <n>:none, <l>:local, or"
  w/o "                  host:display[.screen] - with the latter option"
  w/o "                  you can specify the name or IP-address of a display;"
  w/o "                    default: <l>"
  w/o "      ******************************************************************"
  return/exit
endif
!
!
set/format {FMTDEF}
!
!
!          INPUT :
!          *****
!
!P1(ima):
!-------
if p1 .eq. "?" mess/out 2 {PROID} " "
define/parameter p1 ? i -
  "        Enter name of flatfield frame: "
define/local orderflat/c/1/64 "{p1}"
if m$index(orderflat,".bdf") .eq. 0 orderflat = orderflat//".bdf"
!
if m$exist(orderflat) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered frame is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({orderflat})"
  goto error
endif
!
define/local npix/i/1/2 {{orderflat},NPIX(1)},{{orderflat},NPIX(2)}
!
!
!P2(geom):
!--------
define/parameter p2 50,43,5,220,10 ? -
  "        Enter geom.params. (see proc.header) [def:50,43,5,220,10]: "
outputi = m$parse(p2,"geom")
if outputi(1) .ne. 5 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P2 !"
  goto error
endif
define/local COLDIST/i/1/1 {geom01}
define/local YMED/i/1/1    {geom02}
define/local YDERTHR/i/1/1 {geom03}
define/local ORDWID/i/1/1  {geom04}
define/local WIDTOL/i/1/1  {geom05}
!
define/local geombad/i/1/1 0
inputi(1) = m$nint(npix(1)/2.-1.-0.5)
if COLDIST .lt. 1 .or. COLDIST .gt. inputi(1) then
        write/out " "
        mess/out 1 {PROID} "P2[1]: Column y-distance out of range !"
        mess/out 1 {PROID} "({COLDIST} ; 1<=col.dist.<={inputi(1)})"
  geombad = 1
endif
if YDERTHR .lt. 1 then
        write/out " "
        mess/out 1 {PROID} "P2[3]: Minimal first y-derivative out of range !"
        mess/out 1 {PROID} "({YDERTHR} ; 1<=first_der.)"
  geombad = 1
endif
inputi(1) = npix(2)-3
if ORDWID .lt. 1 .or. ORDWID .gt. inputi(1) then
        write/out " "
        mess/out 1 {PROID} "P2[4]: Y-width of one order out of range !"
        mess/out 1 {PROID} "({ORDWID} ; 1<=width<={inputi(1)})"
  geombad = 1
endif
inputi(1) = ORDWID/2
if WIDTOL .lt. 1 .or. WIDTOL .gt. inputi(1) then
        write/out " "
        mess/out 1 {PROID} "P2[5]: Toler.f.y-width of one order out of range !"
        mess/out 1 {PROID} "({WIDTOL} ; 1<=width_tol.<={inputi(1)})"
  geombad = 1
endif
inputi(1) = ORDWID/2
if YMED .lt. 1 .or. YMED .gt. inputi(1) then
        write/out " "
        mess/out 1 {PROID} "P2[2]: No.of neighb.pix. f.med.filt. out of range !"
        mess/out 1 {PROID} "({YMED} ; 1<=neighb.pix.<={inputi(1)})"
  geombad = 1
endif
if geombad .eq. 1 goto error
!
!
!P3(hough):
!---------
define/parameter p3 0.35,0.75,0.008,1. ? -
  "        Enter Hough params. (see proc.header) [def:0.35,0.75,0.008,1.]: "
outputi = m$parse(p3,"hough")
if outputi(1) .ne. 4 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P3 !"
  goto error
endif
define/local SLOPMIN/r/1/1 {hough01}
define/local SLOPMAX/r/1/1 {hough02}
define/local SLOPSTEP/r/1/1 {hough03}
define/local ZEPOSTEP/r/1/1 {hough04}
!
define/local houghbad/i/1/1 0
set/format F5.3,
if SLOPMIN .ge. SLOPMAX then
        write/out " "
        mess/out 1 {PROID} "P3[1:2]: slope minimum >= slope maximum !"
        mess/out 1 {PROID} "({SLOPMIN} >= {SLOPMAX})"
  houghbad = 1
endif
inputr(1) = (SLOPMAX-SLOPMIN)/2.
if SLOPSTEP .ge. inputr(1) .and. houghbad .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "P3[3]: slope step too high !"
        mess/out 1 {PROID} "({SLOPSTEP} ; slope_step<{inputr(1)})"
  houghbad = 1
endif
! ZEPOSTEP can first be checked in "fors_findorder_lin.prg" !
set/format {FMTDEF}
if houghbad .eq. 1 goto error
!
!
!P4(extra):
!---------
define/parameter p4 + ? -
  "        Enter flag for extrapolation of orders [def.:n|<y>es]: "
define/local extra/c/1/1 "{p4}"
!
!
!P5(disp):
!--------
define/parameter p5 l ? -
  "        Enter display [def.:<l>ocal|<p>ipeline|<n>one|host:disp[.screen]]: "
define/local dispflag/c/1/30 "{p5}"
!
!
!          DEFINITION OF FURTHER VARIABLES :
!          *******************************
!
define/local postbl/c/1/60 " " all
define/local houima/c/1/60 " " all
define/local cluima/c/1/60 " " all
define/local lintbl/c/1/60 " " all
define/local tratbl/c/1/60 " " all
define/local foundords/i/1/1 0
define/local ordtbl/c/1/60 " " all
define/local bactbl/c/1/60 " " all
define/local ordloop/i/1/1 0
!
!
!
!          MANAGING DISPLAYS / ITT / LUT :
!          *****************************
!
if dispflag .ne. "p" then
  !
  !**************************
  @p fors_nicecrea {dispflag}
  !**************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Managing of display options failed !"
    goto error
  endif
  !
endif
!
if nice_crea .eq. "y" then
  !
  if nice_dids(1) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 1 is missing !"
    goto error
  endif
  !
  assign/display d,1 >Null
  clear/display
  display/lut off
  if nice_dids(2) .eq. 1 then
    assign/display d,2 >Null
    clear/display
    display/lut off
  endif
  if nice_dids(3) .eq. 1 then
    assign/display d,3 >Null
    clear/display
    display/lut off
  endif
  load/lut heat
  load/itt ramp
  if nice_gids(1) .eq. 1 then
    assign/graphic g,1 >Null
    clear/graphic
  endif
  if nice_gids(2) .eq. 1 then
    assign/graphic g,2 >Null
    clear/graphic
  endif
  !
endif
!
!
if nice_crea .eq. "y" then
  !
  !
  !        DISPLAYING OF REFERENCE FRAME :
  !        *****************************
  !
        mess/out 4 {PROID} "displaying image"
  !
  !************************************************
  @p fors_nicedisp {orderflat} ? n 0,0 ? ? 0.4,3. n
  !************************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Displaying of reference frame failed !"
        mess/out 1 {PROID} "({orderflat})"
    goto error
  endif
  !
        mess/out 4 {PROID} "  loading reference image into display 1"
  assign/display d,1 >Null
  load/image {orderflat} ? {outputi(1)},avg c,c
  !
endif
!
!
!          CALCULATING COLUMN DERIVATIVES / ORDER Y-CENTRES :
!          ************************************************
!
        mess/out 3 {PROID} "detecting order positions"
!
!*****************************************************************
@p fors_findorder_pos {orderflat} -
                      {COLDIST},{YMED},{YDERTHR},{ORDWID},{WIDTOL}
!*****************************************************************
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Detection of order positions failed !"
  goto error
endif
postbl = q2
!
!
!          PERFORMING HOUGH TRANSFORMATION / CLUSTER DETECTION :
!          ***************************************************
!
        mess/out 3 {PROID} "computing straight coefficients / order numbers"
!
!****************************************************************
@p fors_findorder_lin {postbl} -
                      {SLOPMIN},{SLOPMAX},{SLOPSTEP},{ZEPOSTEP} -
                      {ORDWID},{npix(1)}
!****************************************************************
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Comp. of straight coeff. / ord. no. failed !"
  goto error
endif
outputi = m$parse(q2,"lin")
lintbl = lin01
houima = lin02
cluima = lin03
!
!
!          TRACING POSITIONS PER ORDER :
!          ***************************
!
        mess/out 3 {PROID} "assigning order numbers to positions"
!
!***************************************************
@p fors_findorder_tra {postbl} {lintbl},ORIG,SLOPE -
                      {ORDWID}
!***************************************************
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Assignm. of order numbers to positions failed !"
  goto error
endif
tratbl = q2
foundords = {q3}
!
!
!          FITTING POSITIONS PER ORDER :
!          ***************************
!
        mess/out 3 {PROID} "improving order positions"
!
!*******************************************************************
@p fors_findorder_fit {tratbl} {extra} {npix(1)},{npix(2)},{COLDIST}
!*******************************************************************
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Improving of order positions failed !"
  goto error
endif
ordtbl = q2
!
write/descriptor {ordtbl} ORDWIDTH/i/1/1 {ORDWID}  ! for later order extraction
!
!
if nice_crea .eq. "y" then
  !
  !
  !        DISPLAYING FOUND ORDERS :
  !        ***********************
  !
  label/display "FOUND RELATIVE ORDERS" {nice_dpos} ? cyan 1
  load/table {ordtbl} :X :YFIT ? 5 3 5
  do ordloop = 1 foundords
    select/table {ordtbl} :ORDER.eq.{ordloop} >Null
    load/table {ordtbl} :X :YFIT :ORDER 5 0 8
  enddo
  select/table {ordtbl} all
  !
endif
!
!
!          RE-FITTING BACKGROUND AREAS :
!          ***************************
!
        mess/out 3 {PROID} "computing background areas"
!
!***********************************************************
@p fors_findorder_bac {ordtbl} {npix(1)},{npix(2)} {COLDIST}
!***********************************************************
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Computation of background areas failed !"
  goto error
endif
bactbl = q2
!
!
if nice_crea .eq. "y" then
  !
  !
  !        DISPLAYING CHOSEN BACKGROUND AREAS :
  !        **********************************
  !
  load/table {bactbl} :X :YBKG :ORDER,nodraw 5 0 6
  name/column {bactbl} :XSTA :XSTARTPIX
  name/column {bactbl} :XEND :XENDPIX
  name/column {bactbl} :YSTA :YSTARTPIX
  name/column {bactbl} :YEND :YENDPIX
  draw/rectangle {bactbl} p n 4,0 >Null
  name/column {bactbl} :XSTARTPIX :XSTA
  name/column {bactbl} :XENDPIX :XEND
  name/column {bactbl} :YSTARTPIX :YSTA
  name/column {bactbl} :YENDPIX :YEND
  !
endif
!
!
!          OUTPUT :
!          ******
!
        mess/out 4 {PROID} " "
        mess/out 4 {PROID} ">>> These are the output files :               <<<"
        mess/out 4 {PROID} ">>> 1st guess order positions:  {postbl}   <<<"
        mess/out 4 {PROID} ">>> Hough image:                {houima} <<<"
        mess/out 4 {PROID} ">>> Cluster image:              {cluima} <<<"
        mess/out 4 {PROID} ">>> Straight coeff. per order:  {lintbl}   <<<"
        mess/out 4 {PROID} ">>> Rel.ord.no. assi. to pos.:  {tratbl}   <<<"
        mess/out 4 {PROID} ">>> REL.ORD.NO. / FITTED POS.:  {ordtbl}      <<<"
        mess/out 4 {PROID} ">>> POS. OF BACKGROUND AREAS:   {bactbl}       <<<"
!
!
set/format
!
!
if warn .eq. 0 then
  return 0 {ordtbl},{bactbl} {foundords}
else
  return -1 {ordtbl},{bactbl} {foundords}
endif
!
error:
  set/format
  return 1 -99,-99 -99
!
!
!------------------------------------------------------------------------------
