! $Id: fors_oscmodel.prg,v 1.4 2000/07/04 19:21:04 nrainer Exp $
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1999  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_oscmodel.prg,v $
!
!.KEYWORDS    Basic CCD reduction, bias computation from overscan area
!
!.USAGE       model/overscan  frame region readout mtd
!
!.INPUT       frame       (P1)   Input frame for which the bias is determined
!                                from the overscan area(s).
!             region      (P2)   Image area to be used for overscan
!                                computation. Expected format for the region
!                                string is: x1,y1,x2,y2 (with xi,yi CCD
!                                pixels although world coordinates are used)
!             readout     (P3)   CCD readout direction
!             mtd         (P4)   Method used for bias computation: MEDIAN or
!                                FIT. For method fit the fit accuracy, i.e.
!                                the difference of successive chi^2 values,
!                                has to be given in the form FIT,<value>.
!
!.OUTPUT      none.
!
!.RETURNS     The recipe returns 0 on success, otherwise 1. Keyword Q1 is 
!             updated with the exit status. The computed bias value and the
!             associated noise are written to the keywords OUTPUTR(1) and
!             OUTPUTR(2).
!
!.PURPOSE     Get the bias value from the specified image area.
!
!.ENVIRON     FORS context
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1999/08/12:  Creation
!
!------------------------------------------------------------------------------

crossref  frame region readout mtd

define/param p1 ?   c "Enter name of input frame:"
define/param p2 ?   c "Enter image area:"
define/param p3 X   c "Enter CCD readout direction:"
define/param p4 MED c "Enter name of computational method:"
define/maxpar 4


! Procedure identification
define/local procID/c/1/80      "ModelOscan"


! Local constants
define/local ov_imode/c/1/10  "no"             ! fit overscan interactively
define/local ov_funct/c/1/20  "poly"           ! fitting function
define/local ov_order/i/1/1   3                ! number of polynomial terms
define/local ov_aver/i/1/1    1                ! number of points to combine
define/local ov_iter/i/1/1    1                ! number of iterations
define/local ov_rejec/r/1/1   3.               ! sigma rejection factor
define/local tmp0/c/1/20      "tmp_ovarea.bdf"
define/local tmp1/c/1/20      "tmp_ovscan.bdf"
define/local tmp2/c/1/20      "tmp_ovdata.tbl"
define/local tmp3/c/1/20      "tmp_ovparam.tbl"


! Local variables
define/local pos/i/1/1       0
define/local bias/r/1/1      0.
define/local sigma/r/1/1     0.
define/local accuracy/r/1/1  0.
define/local mtd/c/1/3       " "  all
define/local rdir/c/1/1      " "  all
define/local ov_sec/c/1/80   " "  all
define/local ifrm/c/1/80     {p1}
define/local oreg/r/1/4      {p2}



! Process parameters
rdir = m$upper(p3(1:1))
if rdir .ne. "X" .and. rdir .ne. "Y" then
  mess/out 1 {procID} "Invalid CCD readout direction!"
  return 1
endif

mtd = m$upper(p4(1:3))
if mtd .ne. "MED" .and. mtd .ne. "FIT" then
  mess/out 1 {procID} "Invalid overscan method {mtd} encountered!"
  return 1
else
  if mtd .eq. "FIT" then
    pos = m$index(p4,",")

    if pos .le. 0 then
      mess/out 1 {procID} "Missing accuracy value for method >{mtd}<!"
      return 1
    else
      pos = pos + 1
      if m$tstno(p4({pos}:)) .eq. 1 then
        accuracy = {p4({pos}:)}
      else
        mess/out 1 {procID} "Invalid fit accuracy >{p4({pos}:)}<!"
        return 1
      endif
    endif

    if accuracy .le. 0 then
      mess/out 1 {procID} "Invalid fit accuracy {p4({pos}:)} for method fit!"
      return 1
    endif
  endif
endif


! Extract overscan region
ov_sec = "[{oreg(1)},{oreg(2)}:{oreg(3)},{oreg(4)}]"
extract/image {tmp0} = {ifrm} {ov_sec}


! Average overscan area perpendicular to the readout direction
if rdir .eq. "X" then
  average/row {tmp1} = {tmp0}
else
  average/column {tmp1} = {tmp0}
endif


! Use given method to compute overscan bias
set/format F7.2
mess/out 4 {procID} "Computing bias value using frame >{ifrm}<"
mess/out 4 {procID} -
  "Overscan area [ {oreg(1)} {oreg(2)} : {oreg(3)} {oreg(4)} ]"
mess/out 4 {procID} "Computational method:\t{mtd}"
set/format

if mtd .eq. "FIT" then
  fit/exponential {tmp1} {tmp2} {tmp3} {accuracy} >Null
  if {q1} .ne. 0 then
    -delete {tmp0} {tmp1} {tmp2} {tmp3}
    return 1
  else
    mess/out 4 {procID} "Chi-square of fit:\t{outputr(1)}"
    bias = m$value({tmp3},:result,{ov_order})
    -delete {tmp2} {tmp3}
  endif
else
  statistics/image {tmp1} option=GN >Null
  bias = outputr(8)
endif


! Determine noise
statistics/image {tmp0} option=SN >Null
sigma = outputr(4)


! Fill output keyword
outputr(1) = bias
outputr(2) = sigma

set/format f1.3
mess/out 4 {procID} "Computed bias value:\t{bias} adu"
mess/out 4 {procID} "Computed noise value:\t{sigma} adu"
set/format


! Cleanup
-delete {tmp0}
-delete {tmp1}

return 0
