! $Id: fors_photometry.prg,v 1.253 2001/07/13 09:43:40 sbogun Exp $   
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
! .COPYRIGHT  (C) 1998  European Southern Observatory (ESO)
!
! .IDENT       $RCSfile: fors_photometry.prg,v $
!
! .KEYWORDS   Photometry
!
! .PURPOSE    *) Before the invocation of fors_photometry.prg the input frames
!                should be normalized by their exposure times. If an overexpo-
!                sure mask is present for an input frame, it should be applied.
!                Thus pixels which were overexposed (before the basic reduc-
!                tions) are reset to their original values and objects with
!                such pixels can be flagged.
!
!             B) Now, by means of the SExtractor object extraction software,
!                object positions (pixel and WCS), sky background frame, sky-
!                subtracted frame and brightnesses of each object are computed.
!                SExtractor works as explained in .ALGORITHM under B).
!
!             C) Afterwards generally objects with properties which are ex-
!                plained in .ALGORITHM under C) are REMOVED.
!
!             D) Furthermore signal quality values, and that signal-to-noise,
!                limiting magnitude, seeing (FWHM) and trailing (ellipticity),
!                are calculated.
!                The values are SELECTED and calculated as explained in
!                .ALGORITHM under D).
!
!             E) Additionally the weighted airmass is computed.
!
!             *) After the invocation of fors_photometry.prg the object bright-
!                nesses should be converted from ADU to electrons.
! __________
!
! .USAGE      @p fors_photometry cattyp [sext] [ron] [pce] [geo] -
!                                [disp] [over] [col]
!
! .INPUT      cattyp (P1): image _cat_alogue with input frames AND _typ_e of
!                          input frames: s<t>andard stars (only good detected
!                          objects are used), s<c>ience (all objects are listed
!                          in the output table (and displayed)), separated by a
!                          comma;
!                            default: none,<c>
!
!             sext   (P2): 4 ascii files for SExtractor (input parameters, con-
!                          volution mask, output parameters, neuronal network
!                          weights), separated by commas;
!                            default: sex__inpar.dat,sex_convol.dat,sex_out-
!                                     par.dat,sex_neupar.dat
!
!             ron    (P3): name of descriptors with readout noise (e-/pixel);
!                            default: RON
!
!             pce    (P4): names of descriptors with _p_ixel scale AND electrons
!                          to ADU _c_onversion factor AND original _e_xposure
!                          time, separated by commas; if the name of the desc.
!                          with the original exposure time is NOT entered, NOT
!                          normalized (by exp.time) images are assumed - affects
!                          calculation of signal-to-noise;
!                            default: PIXSCALE,CONAD,none
!
!             geo    (P5): names of descriptors with east longitude AND latitude
!                          of observatory (both descriptor values in degrees and
!                          only as single float numbers), separated by a comma;
!                            default: GEOLON,GEOLAT
!
!             disp   (P6): display selection: <p>:pipeline mode:read global key
!                          NICE_CREA [y|n], <n>:none, <l>:local, or
!                          host:display[.screen] - with the latter option
!                          you can specify the name or IP-address of a display;
!                            default: <l>
!
!             over   (P7): numbers of pre- AND overscan columns AND float thre-
!                          shold (ADU) for overexposed pixels, separated by
!                          commas (first two elem. only for calc. of cut vals.);
!                            default: 16,16,65534.
!                            hint: first two elements ignored, if P6=n
!
!             col    (P8): column labels #3,4,5,6,20,21 of output table (are
!                          used in further procedures), separated by commas;
!                            default: RA,DEC,FLUX,MAG,S2N,AIRMASS
!
! .OUTPUT     A catalogue "photo_out.cat" will be created, where all output
!             files are stored; this name is written into Q3, too.
!
!             "<P1-entries>_skymap.bdf":  sky background frames
!
!             "<P1-entries>_skysub.bdf":  sky-subtracted frames
!
!             If objects were found:
!
!               "<P1-entries>_phot.tbl":  1 XPIX,
!                (`*':default names)      2 YPIX,
!                (`N':do NOT change)   *  3 RA,
!                                      *  4 DEC     : POSITIONS
!                                        ----------
!                                      *  5 FLUX,     (sky-subtracted)
!                                      *  6 MAG,      (sky-subtracted)
!                                         7 SKY     : BRIGHTNESSES
!                                        ----------
!                                         8 FWHM    : SEEING
!                                         9 ELLIP   : TRAILING
!                                        10 NPIX    : # pixels per object
!                                        11 ELL_A   : maj.ell.half-axis (1sigma)
!                                        12 ELL_B   : min.ell.half-axis (1sigma)
!                                      N 13 ROTANG  : angle(maj.axis,x-axis)
!                                        14 FLAGS   : OBJ.QUAL. (not f. std.)
!                                        ==========
!                                      N 15 COLOR   : colours f. ellipses
!                                      N 16 XSTARTPIX,
!                                      N 17 YSTARTPIX,
!                                      N 18 XENDPIX,
!                                      N 19 YENDPIX : rectangle surround. ellip.
!                                      * 20 S2N     : SIGNAL-TO-NOISE
!                                      * 21 AIRMASS : WEIGHTED AIRMASS
!                                        22 MJD     : mod.Jul.date
!                                        23 STELLAR : stellarity index
!
!               Hint for columns 13, 15 and 16-19 (object marks):
!                 With `draw/ellipse <P1-entries>_phot.tbl p' the
!                 found objects can be marked onto the displayed current
!                 <P1-entry> frame.
!
!               Legend for column 14 (obj.qual.flags):
!
!                 <C>lose    : area of neighbour overlaps more than 10%
!                 <M>ultiple : multiple object detection after deblending
!                 <S>aturated: at least one pixel of object is saturated
!                 ( <T>runcated: object is too close to image boundary,
!                                already removed )
!                 ( <A>perture data corrupt : incomplete data
!                                already removed )
!                 memory overflow during <D>eblending: memory not enough
!                 memory <O>verflow during extraction: memory not enough
!
!               Formula for column 20 (signal-to-noise):
!
!                 All fluxes and readout noise in e- , w. original expo. time ,
!                 the object flux is integrated :
!
!                                     flux_obj.
!                 -------------------------------------------------------------
!                 sqrt( flux_obj. + #pix./obj.*flux_sky + #pix./obj.*(RON**2) )
!
!               The airmass (X) in column 21 is weighted as follows:
!                 <X> = (X_expostart + 4*X_expomiddle + X_expoend) / 6.
!
!               In the last but one row the medians of SEEING and TRAILING
!               are given (negative number: calculation failed),
!               in the last row the median deviations from the median ("more
!               robust sigma").
!
!               In the last but one row of the MAG column the relative (i.e.,
!               not flux calibrated) limiting magnitude is given (for S/N=5 ,
!               defined here in fors_photometry.prg; negative number: calcu-
!               lation failed).
!
!             Q2: tot. number of skipped frames, because no objects were found
!
! .RETURN     Q1:  0: successful return
!                 -1: a warning message occured
!                  1: an error occured
! __________
!
! .ALGORITHM  See .PURPOSE under B):
!               SExtractor works as follows:
!               1. Estimation of the sky background (kappa-sigma and mode)
!               2. Detection of objects by thresholding instead of peak finding
!                  (improved by a convolution mask entered as ascii file P2[2])
!               3. Deblending (splitting of close detections, reallocation of
!                  low-flux pixels)
!               4. Filtering of the detections (eliminating noise peaks in
!                  wings)
!               5. Photometry (fixed aperture in the pipeline; this and the
!                  other output values are determined by the ascii file P2[3])
!               Important input parameters (passed by the ascii file P2[1]):
!               for 1. mesh size of grid, size of median filter;
!                   2. convolution mask, threshold, min. number of connected
!                      pixels, hardware dependent buffer sizes;
!                   3. number of additional thresholds, fraction of total
!                      intensity above which a component should be considered
!                      to be separate;
!                   4. factor for strength of cleaning, hardware dependent
!                      buffer size;
!                   5. aperture diameter in pixels, saturation level, no zero-
!                      point correction;
!               For more detailed descriptions see corresponding documents in
!               $PIPE_HOME/fors/sext/doc/, especially sex1_doc.ps.gz and
!               sex2_doc.ps.gz (the latter one for the valid current version
!               2.0.15 has not been completed by the author E.Bertin yet).
!
!             See .PURPOSE under C):
!               Generally objects with the following properties are REMOVED
!               (the constants `("...")' are defined below):
!               - negative flux (because of sky subtraction),
!               - negative FWHM (bug in SExtractor),
!               - too close to the image boundary,
!               - incomplete data,
!               - ellipticity > 0.9 ("BADELLIP"),
!               - only for standard stars: #pixels/object < 40 ("MINSTDPIX").
!
!             See .PURPOSE under D):
!               The values are SELECTED and calculated in the following way:
!               * S/N:
!                   (no selection and no median value)
!               * LIMITING MAGNITUDE:
!                   - S/N <= 10. ("MAXS2N"),
!                   - without error flags;
!                   at least 9 ("MINOBJ") objects must be left, then S/N is
!                   fitted against the flux (with mode of all quotients (S/N /
!                   flux) as slope and 0.,0. as zero point). The limiting mag-
!                   nitude is taken at S/N=5 ("LIMS2N") .
!               * FWHM:
!                   - ellipticity < 0.5 ("MAXELLIP")
!                   - stellarity index > 0.5 ("MINSTELL")
!                   - FWHM > 0.3 arcsec ("MINFWHM")
!                   - without error flags (created by SExtractor);
!                   at least 9 ("MINOBJ") objects must be left, then the median
!                   of the FWHMs of the remaining objects is calculated.
!               * ELLIPTICITY:
!                   - #pixels/object >= 25 ("MINPIX"),
!                   - without error flags;
!                   at least 3 objects must be left, then the median
!                   of the ellipticities of the remaining objects is calculated.
!
! .COMMENT    The ascii files with input parameters and output parameters (P2)
!             must NOT be changed; otherwise this programme must be altered
!             partially (see comments in section "Photometry" and numbers in
!             section "Creation of output table") !!!
!
!             The last three rows in the photometry output table must NOT be
!             removed and additional rows must NOT be created; otherwise fol-
!             lowing procedures will fail (those use all but the last three
!             rows) !!!
!
! .ENVIRON    -
!
! .LANGUAGE   MIDAS 98NOV
!
! .AUTHOR     N.Rainer  DPG/DMD/ESO
!
! .VERSION    0.9  98/07/07  Creation
!                  98/07/15  Release
!             1.0  98/12/15  Release
!             1.1  99/01/14  Release
!             1.2  99/03/15  Release
!             1.3  99/09/23  Release
!
!-------------------------------------------------------------------------------
!
!
crossref cattyp sext ron pce geo disp over col
!
!
define/local BADELLIP/r/1/1    0.5   ! for classification of worst objects
define/local MINSTDPIX/i/1/1  40     ! for detection as standard stars
define/local ELLSIZE/r/1/1     5.    ! # of sigmas of half-axes for drawing ell.
define/local MAXS2N/r/1/1     10.    ! for limiting magnitude
define/local MINOBJ/i/1/1      9 ? +lower_levels  ! f. lim.mag. and fwhm - >=3 !
define/local LIMS2N/r/1/1      5.    ! for limiting magnitude
define/local MINS2N/r/1/1     60.    ! for fwhm
define/local MAXELLIP/r/1/1    0.5   ! for fwhm and graphics
define/local MINSTELL/r/1/1    0.5   ! for fwhm
define/local MINFWHM/r/1/1     0.3   ! for fwhm
define/local FWHMBINSZ/r/1/1   0.1   ! binsize for fwhm statistics
define/local MINPIX/i/1/1     25     ! for fwhm, med.ellip. and graphics
define/local OUTCAT/c/1/20   "photo_out.cat"
define/local FAILNO/r/1/1   -99999.9 ? +lower_levels  ! for failed calculations
define/local dictinst/c/1/10  " " ALL
define/local dpr_type/c/1/64  " " ALL
define/local tpl_id/c/1/64    " " ALL
define/local arcfile/c/1/64   " " ALL
define/local filter/c/1/64    " " ALL
!
!  For `set/format':
define/local FMTDEF/c/1/17 "I1 E14.7,E23.15"
!
!  For `mess/out':
define/local PROID/c/1/40 "photometry"
if m$existk("mess_level") .eq. 0 write/keyword mess_level/i/1/1 3
!
!  For "Creation of output table":
if m$existk("normalized") .eq. 0 define/local normalized/i/1/1 0
!
!  For `shortlog/out':
write/keyword out_a "given command/qualif not defined ..."
show/command shortlog/out | write/keyword in_a
-delete Mid{mid$sess(11:12)}1Pipe
-delete midlis{mid$sess(11:12)}.prg
if in_a .eq. out_a create/command shortlog/out "write/out"
!
!  For `longlog/out':
write/keyword out_a "given command/qualif not defined ..."
show/command longlog/out | write/keyword in_a
-delete Mid{mid$sess(11:12)}1Pipe
-delete midlis{mid$sess(11:12)}.prg
if in_a .eq. out_a create/command longlog/out "write/out"
!
!  For return value:
define/local warn/i/1/1 0
write/keyword q1 0
write/keyword q2 0
write/keyword q3 0
!
!
if p1 .eq. "??" then
  w/o "      ******************************************************************"
  w/o "      cattyp (P1): image _cat_alogue with input frames AND _typ_e of"
  w/o "                   input frames: s<t>andard stars (only good detected"
  w/o "                   objects are used), s<c>ience (all objects are listed"
  w/o "                   in the output table (and displayed)), separated by a"
  w/o "                   comma;"
  w/o "                     default: none,<c>"
  w/o " "
  w/o "      sext   (P2): 4 ascii files for SExtractor (input parameters, con-"
  w/o "                   volution mask, output parameters), separated by a"
  w/o "                   comma;"
  w/o "                     default: sex__inpar.dat,sex_convol.dat,sex_out-"
  w/o "                              par.dat,sex_neupar.dat"
  w/o " "
  w/o "      ron    (P3): name of descriptors with readout noise (e-/pixel);"
  w/o "                     default: RON"
  w/o " "
  w/o "      pce    (P4): names of descriptors with _p_ixel scale AND electrons"
  w/o "                   to ADU _c_onversion factor AND original _e_xposure"
  w/o "                   time, separated by commas; if the name of the desc."
  w/o "                   with the original exposure time is NOT entered, NOT"
  w/o "                   normalized (by exp.time) images are assumed - affects"
  w/o "                   calculation of signal-to-noise;"
  w/o "                     default: PIXSCALE,CONAD,none"
  w/o " "
  w/o "      geo    (P5): names of descriptors with east longitude AND latitude"
  w/o "                   of observatory (both descriptor values in degrees and"
  w/o "                   only as single float numbers), separated by a comma;"
  w/o "                     default: GEOLON,GEOLAT"
  w/o " "
  w/o "      disp   (P6): display selection: <p>:pipeline mode:read global key"
  w/o "                   NICE_CREA [y|n], <n>:none, <l>:local, or"
  w/o "                   host:display[.screen] - with the latter option"
  w/o "                   you can specify the name or IP-address of a display;"
  w/o "                     default: <l>"
  w/o " "
  w/o "      over   (P7): numbers of pre- AND overscan columns AND float thre-"
  w/o "                   shold (ADU) for overexposed pixels, separated by"
  w/o "                   commas (first two elem. only for calc. of cut vals.);"
  w/o "                     default: 16,16,65534."
  w/o "                     hint: first two elements ignored, if P6=n"
  w/o " "
  w/o "      col    (P8): column labels #3,4,5,6,20,21 of output table (are"
  w/o "                   used in further procedures), separated by commas;"
  w/o "                     default: RA,DEC,FLUX,MAG,S2N,AIRMASS"
  w/o "      ******************************************************************"
  return/exit
endif
!
!
set/format {FMTDEF}
!
!
!          INPUT :
!          *****
!
!P1(cattyp):
!----------
if p1 .eq. "?" mess/out 2 {PROID} " "
define/parameter p1 ? ? -
  "        Enter cat.(BDF) / type of inp.frm. [def.:s<c>i|s<t>d]: "
outputi = m$parse(p1,"ima")
if outputi(1) .gt. 2 then  ! 1 allowed for type default
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P1 !"
  goto error
endif
define/local icat/c/1/64 "{ima01}"
if m$index(icat,".cat") .eq. 0 icat = icat//".cat"
if outputi(1) .eq. 2 then
  define/local type/c/1/1 "{ima02}"
elseif outputi(1) .eq. 1 then  ! default
  define/local type/c/1/1 "c"
endif
!
if m$exist(icat) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered catalogue is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({icat})"
  goto error
endif
!
show/icat {icat} n
if outputi(1) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered catalogue is empty !"
        mess/out 1 {PROID} "({icat})"
  goto error
endif
!
define/local catal/i/1/1 0
inloop:
  store/frame inputc {icat} 1 end_inloop
  if m$exist(inputc) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "At least one catalogue entry is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({icat})"
    goto error
  endif
  goto inloop
end_inloop:
!
if type .ne. "c" .and. type .ne. "t" then
        write/out " "
        mess/out 1 {PROID} "Wrong frame type selection !"
  goto error
endif
!
!
!P2(sext):
!--------
define/parameter p2 sex__inpar.dat,sex_convol.dat,sex_outpar.dat,sex_neupar.dat -
  ? "        4 ASCII files for SExtractor [if different from def.]: "
outputi = m$parse(p2,"sex")
if outputi(1) .ne. 4 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P2 !"
  goto error
endif
define/local inparfile/c/1/64 "{sex01}"
if m$index(inparfile,".") .eq. 0 inparfile = inparfile//".dat"
define/local convolfile/c/1/64 "{sex02}"
if m$index(convolfile,".") .eq. 0 convolfile = convolfile//".dat"
define/local outparfile/c/1/64 "{sex03}"
if m$index(outparfile,".") .eq. 0 outparfile = outparfile//".dat"
define/local neuparfile/c/1/64 "{sex04}"
if m$index(neuparfile,".") .eq. 0 neuparfile = neuparfile//".dat"
!
if m$exist(inparfile) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered file with sextractor input parameters"
        mess/out 1 {PROID} "is not present or can not be opened !"
        mess/out 1 {PROID} "({inparfile})"
  goto error
endif
!
if m$filtyp(inparfile," ") .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered file with sextractor input parameters"
        mess/out 1 {PROID} "is no ascii file ({inparfile}) !"
  goto error
endif
!
if m$exist(convolfile) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered file with sextractor convolution mask"
        mess/out 1 {PROID} "is not present or can not be opened !"
        mess/out 1 {PROID} "({convolfile})"
  goto error
endif
!
if m$filtyp(convolfile," ") .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered file with sextractor convolution mask"
        mess/out 1 {PROID} "is no ascii file ({convolfile}) !"
  goto error
endif
!
if m$exist(outparfile) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered file with sextractor output parameters"
        mess/out 1 {PROID} "is not present or can not be opened !"
        mess/out 1 {PROID} "({outparfile})"
  goto error
endif
!
if m$filtyp(outparfile," ") .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered file with sextractor output parameters"
        mess/out 1 {PROID} "is no ascii file ({outparfile}) !"
  goto error
endif
!
if m$exist(neuparfile) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered file with neuronal network weights"
        mess/out 1 {PROID} "is not present or can not be opened !"
        mess/out 1 {PROID} "({neuparfile})"
  goto error
endif
!
if m$filtyp(neuparfile," ") .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered file with neuronal network weights"
        mess/out 1 {PROID} "is no ascii file ({neuparfile}) !"
  goto error
endif
!
!
!P3(ron):
!-------
define/parameter p3 RON ? -
  "        Enter desc. with readout noise (e-/pixel) [def.:RON]: "
define/local ron/c/1/72 "{p3}"
!
!
!P4(pce):
!-------
define/parameter p4 PIXSCALE,CONAD,DIT ? -
  "        Enter descs. f.pix.scl./conv./ori.expo. [def.:PIXSCALE,CONAD,DIT]: "
outputi = m$parse(p4,"sig")
if outputi(1) .ne. 2 .and. outputi(1) .ne. 3 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P4 !"
  goto error
endif
define/local pixscale/c/1/72 "{sig01}"
define/local conad/c/1/72 "{sig02}"
if outputi(1) .eq. 3 then
  define/local exporig/c/1/72 "{sig03}"
  define/local normalized/i/1/1 1
else
  define/local normalized/i/1/1 0
endif
!
!
!P5(geo):
!-------
define/parameter p5 GEOLON,GEOLAT ? -
  "        Enter desc.names f.long./lat. [def.:GEOLON,GEOLAT]: "
outputi = m$parse(p5,"geo")
if outputi(1) .ne. 2 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P5 !"
  goto error
endif
define/local geolon/c/1/72 "{geo01}"
define/local geolat/c/1/72 "{geo02}"
!
!
!P6(disp):
!--------
define/parameter p6 l ? -
  "        Enter display [def.:<l>ocal|<p>ipeline|<n>one|host:disp[.screen]]: "
define/local dispflag/c/1/30 "{p6}"
!
!
!P7(over):
!--------
define/parameter p7 16,16,65534. ? -
  "        Enter pre-/overscan vals., overexp.thresh [def.:16,16,65534.]: "
outputi = m$parse(p7,"over")
if outputi(1) .ne. 3 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P7 !"
  goto error
endif
define/local prescan/i/1/1 {over01}
define/local ovescan/i/1/1 {over02}
define/local thresh/r/1/1 {over03}
!
if prescan .lt. 0 .or. ovescan .lt. 0 then
        write/out " "
        mess/out 1 {PROID} "Impossible pre- or overscan value !"
  goto error
endif
!
!
!P8(col):
!-------
define/parameter p8 RA,DEC,FLUX,MAG,S2N,AIRMASS ? -
  "        Tbl.col.names [def.:RA,DEC,FLUX,MAG,S2N,AIRMASS]: "
outputi = m$parse(p8,"col")
if outputi(1) .ne. 6 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P8 !"
  goto error
endif
define/local ra/c/1/20 "{col01}"
define/local dec/c/1/20 "{col02}"
define/local flux/c/1/20 "{col03}"
define/local mag/c/1/20 "{col04}"
define/local s2n/c/1/20 "{col05}"
define/local airmass/c/1/20 "{col06}"
!
!
!
!          DEFINITION OF VARIABLES :
!          ***********************
!
!     "L o o p" :
  define/local topicima/c/1/64 " " all
  define/local char0/i/1/1 0
  define/local topiorig/c/1/64 " " all
  define/local lopcount/i/1/1 0
  !
  !   "Photometry" :
  define/local char/i/1/1 0
  define/local topicfits/c/1/64 " " all
  define/local photascii/c/1/64 " " all
  define/local backima/c/1/64 " " all
  define/local foundima/c/1/64 " " all
  !
  !   "Checking ... objects" :
  define/local listcode/i/1/2 0 all
  define/local skip/i/1/1 0
  !
  !   "Creation ... table" :
  define/local formcode/i/1/2 0 all
  define/local ctype1/c/1/16 " " all
  define/local ctype2/c/1/16 " " all
  define/local delworld/i/1/1 0
  !
  !   "Removal ... from table" :
  define/local phottable/c/1/64 " " all
  define/local selrows/i/1/1 0
  define/local nnegobj/i/1/1 0
  define/local iloop/i/1/1 0
  define/local nbadobj/i/1/1 0
  define/local goodcount/i/1/1 0
  define/local totrowno/i/1/1 0
  !
  !   "Decision" :
  define/local nogood/i/1/1 0
  !
  !   "signal-to-noise" :
  define/local eron/r/1/1 0.
  define/local cfadu/r/1/1 0.
  define/local exorg/r/1/1 0.
  define/local conexp/r/1/1 0.
  define/local ronsqr/r/1/1 0.
  !
  !   "limiting magnitude" :
  define/local silent/i/1/1 0
  define/local limflux/r/1/1 0.
  define/local limmag/r/1/1 0.
  !
  !   "seeing" :
  define/local fmode/r/1/1 0.
  define/local fsigma/r/1/1 0.
  !
  !   "trailing" :
  define/local emedian/r/1/1 0.
  define/local esigma/r/1/1 0.
!
!
!
!          MANAGING DISPLAYS / ITT / LUT :
!          *****************************
!
if dispflag .ne. "p" then
  !
  !**************************
  @p fors_nicecrea {dispflag}
  !**************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Managing of display options failed !"
    goto error
  endif
  !
endif
!
if nice_crea .eq. "y" then
  !
  if nice_dids(1) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 1 is missing !"
    goto error
  endif
  if nice_dids(2) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 2 is missing !"
    goto error
  endif
  if nice_dids(3) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Display window 3 is missing !"
    goto error
  endif
  if nice_gids(1) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Graphic window 1 is missing !"
    goto error
  endif
  if nice_gids(1) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Graphic window 2 is missing !"
    goto error
  endif
  !
  assign/display d,1 >Null
  clear/display
  display/lut off
  label/display "ORIGINAL" {nice_dlab}
  assign/display d,2 >Null
  clear/display
  display/lut off
  label/display "BACKGROUND" {nice_dlab}
  assign/display d,3 >Null
  clear/display
  display/lut off
  label/display "FOUND OBJECTS" {nice_dlab}
  load/lut heat
  load/itt ramp
  assign/graphic g,1 >Null
  clear/graphic
  assign/graphic g,2 >Null
  clear/graphic
  set/graphic default
  set/graphic pmode=0 bcolour=1
  !
endif
!
!
!          PREPARING OUTPUT :
!          ****************
!
create/icat {OUTCAT} null >Null
!
!
!
!          L O O P   W I T H   F R A M E S :
!          * * * * * * * * * * * * * * * *
!
!
catal = 0
catloop:
  store/frame topicima {icat} 1 end_catloop
  lopcount = lopcount+1
  show/icat {icat} n
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "image {lopcount} / {outputi(1)} : {topicima}"
        char0 = m$indexb(topicima,"_norm")-1
        topiorig = topicima(1:{char0})//".bdf"
        longlog/out 3 " " >Null
        longlog/out 3 "Photometry performed on {topiorig} (normalized):" >Null
  !
  if nice_crea .eq. "y" then
    !
    !
    !      DISPLAYING OF ORIGINAL FRAME :
    !      ****************************
    !
        mess/out 4 {PROID} " "
        mess/out 4 {PROID} "  Displaying image :"
    !
    !***************************************************************
    @p fors_nicedisp {topicima} ? 7. {prescan},{ovescan} ? ? 1.,3. n
    !***************************************************************
    if {q1} .eq. -1 then
      warn = 1
    elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Displaying of original frame failed !"
        mess/out 1 {PROID} "({topicima})"
      goto error
    endif
    !
        mess/out 4 {PROID} "    loading original image into display 1"
    assign/display d,1 >Null
    load/image {topicima} ? {outputi(1)},avg c,c
    label/display "ORIGINAL" {nice_dlab}  ! removed by special load/image
    !
  endif
  !
  !
  !        PHOTOMETRY :
  !        **********
  !
        mess/out 3 {PROID} " "
        mess/out 3 {PROID} "  Positions, Sky and Brightnesses :"
  !
        mess/out 3 {PROID} "    converting input frame: BDF -> FITS"
        mess/out 3 {PROID} " "
  !
  char = m$indexb(topicima,".")-1
  topicfits = topicima(1:{char})//".fits"
  outdisk/fits {topicima} {topicfits} >Null


  ! Call Sextractor
  $ {p_bindir}/sex {topicfits} -c               {inparfile}  -
                               -FILTER_NAME     {convolfile} -
                               -PARAMETERS_NAME {outparfile} -
                               -STARNNW_NAME    {neuparfile} -
                               -SATUR_LEVEL     {thresh}


  if m$exist("sex_output.dat") .eq. 0 then
        mess/out 1 {PROID} "Object detection failed !"
    goto error
  endif
  !
    photascii = topicima(1:{char})//"_phot.dat"
  -rename sex_output.dat {photascii}
  -delete {topicfits}
  !
  inputc = "{{topicima},CUNIT}"  ! "CUNIT":
  ctype1 = inputc(17:32)         ! Stored as multiple 16 char.
  ctype2 = inputc(33:48)         ! words in the frame header
  if ctype1 .ne. "RA---TAN" .or. ctype2 .ne. "DEC--TAN" then
    !
        mess/out 2 {PROID} "    No world coordinates in header -"
        mess/out 2 {PROID} "    using pixel coordinates instead !"
        mess/out 2 {PROID} "    ({topicima})"
    warn = 1
    !
    delworld = 1
  endif
  !
        mess/out 3 {PROID} "    converting output frames: FITS -> BDF"
  !
    char = m$indexb(topicima,".")-1
    backima = topicima(1:{char})//"_skymap.bdf"
  indisk/fits sex_skymap.fits {backima} >Null
  -delete sex_skymap.fits
    foundima = topicima(1:{char})//"_skysub.bdf"
  indisk/fits sex_skysub.fits {foundima} >Null
  -delete sex_skysub.fits
  !
  delete/descriptor {backima} ESO* n  ! indisk/fits converts descs. to double
  delete/descriptor {foundima} ESO* n
  copy/dd {topicima} *,5 {backima}
  copy/dd {topicima} IDENT {backima}
  copy/dd {topicima} *,5 {foundima}
  copy/dd {topicima} IDENT {foundima}
  !
  add/icat {OUTCAT} {backima} >Null
  add/icat {OUTCAT} {foundima} >Null
  !
        longlog/out 1 "     sky background frame: {backima}" >Null
        longlog/out 1 "     sky subtracted frame: {foundima}" >Null
        shortlog/out 1 "Sky background subtracted." >Null
  !
  !
  !        DISPLAYING OF BACKGROUND :
  !        ************************
  !
  if nice_crea .eq. "y" then
    !
    !*************************************************************
    @p fors_nicedisp {backima} ? n {prescan},{ovescan} ? ? 1.,2. n
    !*************************************************************
    if {q1} .eq. -1 then
      warn = 1
    elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Displaying of background frame failed !"
        mess/out 1 {PROID} "({backima})"
      goto error
    endif
    !
        mess/out 4 {PROID} "    loading background image into display 2"
    assign/display d,2 >Null
    load/image {backima} ? {outputi(1)},avg c,c
    label/display "BACKGROUND" {nice_dlab}  ! removed by special load/image
    !
  endif
  !
  !
        mess/out 4 {PROID} " "
        mess/out 3 {PROID} "  Object quality determination :"
  !
  !
  !        CHECKING OF NUMBER OF FOUND OBJECTS :
  !        ***********************************
  !
        mess/out 4 {PROID} "    checking number of found objects"
  !
  open/file {photascii} read listcode
  read/file {listcode(1)} outputc
  if listcode(2) .eq. -1 then
        mess/out 2 {PROID} "    Not a single object found -"
        mess/out 2 {PROID} "    skipping {topicima} !"
    close/file {listcode(1)}
    skip = skip+1
    warn = 1
    !
    goto catloop
  endif
  close/file {listcode(1)}
  !
  !
  !        CREATION OF OUTPUT TABLE :
  !        ************************
  !
        mess/out 4 {PROID} "    creating object table"
    !
    open/file sex.fmt write formcode
    if {formcode(1)} .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Unable to open new file sex.fmt !"
      goto error
    endif
    write/file {formcode(1)} define/field  56  66 R*4 F10.3 :XPIX   "pixel"  ! 1
    write/file {formcode(1)} define/field  67  77 R*4 F10.3 :YPIX   "pixel"  ! 2
    write/file {formcode(1)} define/field  78  89 R*8 F12.7 :{ra}   "deg"    ! 3
    write/file {formcode(1)} define/field  90 101 R*8 F12.7 :{dec}  "deg"    ! 4
  if normalized .eq. 0 then
    write/file {formcode(1)} define/field  11  23 R*4 G13.6 :{flux} "ADU"    ! 5
    write/file {formcode(1)} define/field  24  32 R*4 F9.4  :{mag}  "mag"    ! 6
    write/file {formcode(1)} define/field  33  45 R*4 G12.6 :SKY    "ADU"    ! 7
  else
    write/file {formcode(1)} define/field  11 23 R*4 G13.6 :{flux} "ADU/sec" ! 5
    write/file {formcode(1)} define/field  24 32 R*4 F9.4  :{mag}  "mag/sec" ! 6
    write/file {formcode(1)} define/field  33 45 R*4 G12.6 :SKY    "ADU/sec" ! 7
  endif
    write/file {formcode(1)} define/field 135 143 R*4 F8.2  :FWHM   "arcsec" ! 8
    write/file {formcode(1)} define/field 126 134 R*4 F6.3  :ELLIP  "0-1"    ! 9
    write/file {formcode(1)} define/field  46  55 I*4 I8    :NPIX   "#"      !10
    write/file {formcode(1)} define/field 102 110 R*4 F8.2  :ELL_A "pix1sig" !11
    write/file {formcode(1)} define/field 111 119 R*4 F8.2  :ELL_B "pix1sig" !12
    write/file {formcode(1)} define/field 120 125 R*4 F6.1  :ROTANG "deg"    !13
    write/file {formcode(1)} define/field 144 147 C*9 A9    :FLAGS "CMSTADO" !14
    write/file {formcode(1)} define/field 148 153 R*4 F6.2  :STELLAR " "     !15
    close/file {formcode(1)}
    !
  create/table tmp.tbl 23 1 {photascii} sex.fmt  ! 9 cols. added later (1 tmp).,
  -delete {photascii}                            ! rows are expanded autom.
  -delete sex.fmt
  !
  copy/dd {topicima} *,5 tmp.tbl
  copy/dd {topicima} IDENT tmp.tbl
  !
  !
  !        CONVERSION OR DELETION OF OBJ. QUAL. FLAGS
  !             (DECIMAL -> BINARY -> LETTERS) :
  !        ******************************************
  !
        mess/out 3 {PROID} "    creating object quality flags"
  !
  show/table tmp.tbl >Null
  !*********************************************
  @p fors_photo_flags tmp.tbl,FLAGS {outputi(2)}
  !*********************************************
  if {q1} .eq. -1 then
    warn = 1
  elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Conversion of object quality flags failed !"
    goto error
  endif
  !
  !
  !        REMOVAL OF VERY BAD OBJECTS FROM OUTPUT TABLE
  !        (AND FOR STANDARD STAR FRAMES ALSO "NORMAL" BAD ONES) :
  !        *****************************************************
  !
    char = m$indexb(topicima,".")-1
    phottable = topicima(1:{char})//"_phot.tbl"
  !
        mess/out 3 {PROID} "    cleaning object table"
  !
  select/table tmp.tbl :{flux}.gt.0. -
                  .and.:FWHM.gt.0. -
                  .and.:FLAGS.ne."*T*" -
                  .and.:FLAGS.ne."*A*" -
                  .and.:ELLIP.lt.{BADELLIP} >Null
  selrows = outputi(1)                                   ! FLUX because of prior
  if selrows .eq. 0 then                                 ! sky subtraction and
        mess/out 2 {PROID} "    No objects with positive flux found -"
        mess/out 2 {PROID} "    skipping {topicima} !"   ! FWHM because of bug
    skip = skip+1                                        ! in SExtractor !!!
    warn = 1
        longlog/out 1 "     No objects left after cleaning !" >Null
    !
    goto catloop
  endif
  show/table tmp.tbl >Null
  nnegobj = outputi(2)-selrows
  copy/table tmp.tbl gt0.tbl
  -rename gt0.tbl tmp.tbl
  !
  if type .eq. "t" then  ! only for standard star frames
    !
        mess/out 3 {PROID} "    extracting high quality detections"
    !
    select/table tmp.tbl :FLAGS.eq."  -------".and.:NPIX.ge.{MINSTDPIX} >Null
                                 ! "  -------": 2 blanks, 7 dashes !
    !  flux.(mult.obj. after debl.) O.K.:
    select/table tmp.tbl sel.or.:FLAGS.eq."  -M-----" >Null
    !
    show/table tmp.tbl >Null
    outputi(2) = outputi(2)+nnegobj  ! => tot.no. equal to SExt. output and obj.
    nbadobj = outputi(2)-outputi(8)  !    w.neg.flux are counted as "BAD" below
    copy/table tmp.tbl {phottable}
    -delete tmp.tbl
    !
    add/icat {OUTCAT} {phottable} >Null
    !
        mess/out 3 {PROID} -
       "    Obj.: {outputi(2)} / good: {outputi(8)} / spurious: {nbadobj}"
        longlog/out 1 -
      "     obj.: {outputi(2)} / good: {outputi(8)} / spurious: {nbadobj}" >Null
    !
    if outputi(8) .gt. 0 then
      goodcount = 1
    else
        mess/out 2 {PROID} "    No proper objects found !"
        mess/out 2 {PROID} "    ({topicima})"
      warn = 1
        longlog/out 1 -
          "     No objects left after high quality extraction" >Null
      !
      goto catloop
    endif
    !
  else
    !
    -rename tmp.tbl {phottable}
    goodcount = 1  ! if no objects were found, frame was skipped above
    !
        mess/out 3 {PROID} "      {selrows} objects left"
        longlog/out 1 "     {selrows} objects left after cleaning"
    !
    add/icat {OUTCAT} {phottable} >Null
    !
  endif
  -delete dirfile.ascii
  !
  show/table {phottable} >Null
  totrowno = outputi(2)
  !
        longlog/out 1 "     photometry table:     {phottable}"
        shortlog/out 1 "Object positions and brightnesses calculated." >Null
  !
  !
  !        DISPLAYING OF FOUND OBJECTS :
  !        ***************************
  !
  ! For drawing ellipses around found objects:
  !
  create/column {phottable} :COLOR "1-8" F5.0 R*4  ! col.15
  do iloop = 1 totrowno
    if "{{phottable},:FLAGS,{iloop}}" .eq. "  -------" then  ! good objects
      {phottable},:COLOR,{iloop} = 4.    ! "  -------": 2 blanks, 7 dashes !
    else
      {phottable},:COLOR,{iloop} = 5.
    endif
  enddo
  !
  create/column {phottable} :XSTARTPIX "pixel" F12.7 R*8  ! col.16
  create/column {phottable} :YSTARTPIX "pixel" F12.7 R*8  ! col.17
  create/column {phottable} :XENDPIX   "pixel" F12.7 R*8  ! col.18
  create/column {phottable} :YENDPIX   "pixel" F12.7 R*8  ! col.19
  !
  compute/table {phottable} :XSTARTPIX = :XPIX-{ELLSIZE}*:ELL_A
  compute/table {phottable} :YSTARTPIX = :YPIX-{ELLSIZE}*:ELL_B
  compute/table {phottable} :XENDPIX   = :XPIX+{ELLSIZE}*:ELL_A
  compute/table {phottable} :YENDPIX   = :YPIX+{ELLSIZE}*:ELL_B
  !
  if nice_crea .eq. "y" then
    !
    !***************************************************************
    @p fors_nicedisp {foundima} ? 7. {prescan},{ovescan} ? ? 1.,3. n
    !***************************************************************
    if {q1} .eq. -1 then
      warn = 1
    elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Displaying of found objects failed !"
        mess/out 1 {PROID} "({foundima})"
      goto error
    endif
    !
        mess/out 4 {PROID} "    loading image with found objects into disp. 3"
    assign/display d,3 >Null
    clear/channel overlay
    load/image {foundima} ? {outputi(1)},avg c,c
    label/display "FOUND OBJECTS" {nice_dlab}  ! after load/image - otherwise
    !                                          ! removed by special load/image
    if type .eq. "c" then  ! For standard star frames bad objects are removed !
      label/display "Good" 10,10 ? green 1
      label/display "Spurious" 10,80 ? blue 1
    endif
    draw/ellipse {phottable} p n >Null
    !
  endif
  !
  !
  !        DECISION FOR SIGNAL CHECKS :
  !        **************************
  !
  nogood = 0
  if type .eq. "c" then  ! For science frames NO bad objects are removed !
    !
    select/table {phottable} :FLAGS.eq."  -------" >Null
                                     ! "  -------": 2 blanks, 7 dashes !
    !!  flux.(mult.obj. after debl.) for QC N_O_T O.K. (next line NOT used !):
    !!select/table {phottable} sel.or.:FLAGS.eq."  -M-----" >Null
    !
    if outputi(1) .eq. 0 then
      !
        mess/out 2 {PROID} "    No good objects found - skipping calculation"
        mess/out 2 {PROID} "    of medians (FWHM/Ellipticity) !"
        mess/out 2 {PROID} "    ({foundima})"
      warn = 1
        longlog/out 1 "       No good objects found !" >Null
      !
      select/table {phottable} all >Null
      nogood = 1
    endif
    !
  endif
  !
  !
  !        COMPUTATION OF SIGNAL-TO-NOISE VALUES :
  !        *************************************
  !
  ! also done if we DON'T have good objects for cal-
  ! culation of MEDIAN s/n or limiting magnitude !
  !
        mess/out 4 {PROID} " "
        mess/out 3 {PROID} "  signal-to-noise values ..."
  !
  create/column {phottable} :{s2n} " " F10.3 R*4  ! col.20
  !
  !
  if m$existd(topicima,ron) .eq. 0 then
        write/out " "
        mess/out 2 {PROID} "Entered desc. for readout noise does not exist -"
        mess/out 2 {PROID} "({topicima},{ron})"
        mess/out 2 {PROID} "no calculation of signal-to-noise possible !"
        write/out " "
    warn = 1
    goto no_s2n
  endif
  if {{topicima},{ron}} .le. 0 then
        set/format F15.14
        write/out " "
        mess/out 2 {PROID} "Impossible value for descriptor"
        mess/out 2 {PROID} "`{ron}': {{topicima},{ron}} -"
        mess/out 2 {PROID} "no calculation of signal-to-noise possible !"
        write/out " "
        set/format {FMTDEF}
    warn = 1
    goto no_s2n
  endif
  eron = {{topicima},{ron}}
  !
  if m$existd(topicima,conad) .eq. 0 then
        write/out " "
        mess/out 2 {PROID} "Entered desc. for convers. factor does not exist -"
        mess/out 2 {PROID} "({topicima},{conad})"
        mess/out 2 {PROID} "no calculation of signal-to-noise possible !"
        write/out " "
    warn = 1
    goto no_s2n
  endif
  if {{topicima},{conad}} .le. 0 then
        set/format F15.14
        write/out " "
        mess/out 2 {PROID} "Impossible value for descriptor"
        mess/out 2 {PROID} "`{conad}': {{topicima},{conad}} -"
        mess/out 2 {PROID} "no calculation of signal-to-noise possible !"
        write/out " "
        set/format {FMTDEF}
    warn = 1
    goto no_s2n
  endif
  cfadu = {{topicima},{conad}}
    ! S/N is defined in e-
  !
  if normalized .eq. 1 then  ! input frames were already devided by exp.time
    if m$existd(topicima,exporig) .eq. 0 then
        write/out " "
        mess/out 2 {PROID} "Entered desc. for orig. exp.time does not exist -"
        mess/out 2 {PROID} "({topicima},{exporig})"
        mess/out 2 {PROID} "no calculation of signal-to-noise possible !"
        write/out " "
      warn = 1
      goto no_s2n
    endif
    if {{topicima},{exporig}} .le. 0 then
        set/format F15.14
        write/out " "
        mess/out 2 {PROID} "Impossible value for descriptor"
        mess/out 2 {PROID} "`{exporig}': {{topicima},{exporig}} -"
        mess/out 2 {PROID} "no calculation of signal-to-noise possible !"
        write/out " "
        set/format {FMTDEF}
      warn = 1
      goto no_s2n
    endif
    exorg = {{topicima},{exporig}}
      ! S/N is defined for exp.time before normalization by exp.time
  else
    exorg = 1.
  endif
  !
  conexp = cfadu*exorg
  ronsqr = (eron*cfadu)**2
  ! following commands split because string exceeded max. number of characters:
  compute/table {phottable} :{s2n} = :{flux}*{conexp} +:NPIX*:SKY*{conexp} -
    +:NPIX*{ronsqr}
  compute/table {phottable} :{s2n} = :{flux}*{conexp} / sqrt(:{s2n})
  !
  !
  !        COMPUTATION OF LIMITING MAGNITUDE :
  !        *********************************
  !
  silent = 0
  if nogood .ne. 1 then
    !
        mess/out 4 {PROID} " "
        mess/out 3 {PROID} "  limiting magnitude :"
        longlog/out 1 "     limiting magnitude :" >Null
    !
    !- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    !
    select/table {phottable} :FLAGS.eq."  -------".and.:{s2n}.lt.{MAXS2N} >Null
                                     ! "  -------": 2 blanks, 7 dashes !
    !
    if outputi(1) .ge. MINOBJ then
      !
      !********************************************************************
      @p fors_photo_fitlim {phottable},FLAGS,{flux},{s2n} {LIMS2N},{MAXS2N}
      !********************************************************************
      if {q1} .eq. -1 then
        warn = 1
      elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Computation of limiting magnitude failed !"
        mess/out 1 {PROID} "({phottable})"
        goto error
      endif
      limflux = outputr(1)
      limmag = outputr(2)
      !
    else
      select/table {phottable} all >Null
      limflux = FAILNO
      limmag = FAILNO
        set/format F2.0
        mess/out 2 {PROID} "    Found less than {MINOBJ} proper objects"
        mess/out 2 {PROID} "    with S/N lower than {MAXS2N} -"
        mess/out 2 {PROID} "    no computation of limiting mag. possible !"
        set/format {FMTDEF}
      warn = 1
    endif
    !
    !- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    !
    if limflux .gt. 0 then
        inputd(1) = LIMS2N
        set/format T7.3,T7.3
        mess/out 3 {PROID} "    detection limit (S/N)  = {inputd(1)}"
        longlog/out 1 "       detection limit (S/N)  = {inputd(1)}" >Null
        mess/out 3 {PROID} "    not flux calib. value  = {limmag}"
        longlog/out 1 "       not flux calib. value  = {limmag}" >Null
        set/format {FMTDEF}
    else
        longlog/out 1 "       calibration failed !" >Null
    endif
    ! outputi(2) = totrowno+2
    ! write/table {phottable} :{mag} @{outputi(2)} {LIMS2N}
    outputi(2) = totrowno+3
    write/table {phottable} :{mag} @{outputi(2)} {limmag}
    !
  else
    !
    silent = 1
    goto no_s2n
    !
  endif
  !
  goto s2n_done
    no_s2n:
    compute/table {phottable} :{s2n} = {FAILNO}
    outputi(2) = totrowno+3
    write/table {phottable} :{mag} @{outputi(2)} {FAILNO}
    !
    goto no_seeing
    !
  s2n_done:
  !
  !
  !        CALCULATION OF MEDIAN SEEING VALUE (FWHM) :
  !        *****************************************
  !
  if m$existd(topicima,pixscale) .eq. 0 then
        write/out " "
        mess/out 2 {PROID} "Entered desc. for pixel scale does not exist -"
        mess/out 2 {PROID} "({topicima},{pixscale})"
        mess/out 2 {PROID} "no calculation of seeing possible !"
        write/out " "
    warn = 1
    goto no_seeing
  endif
  if {{topicima},{pixscale}} .le. 0 then
        set/format F15.14
        write/out " "
        mess/out 2 {PROID} "Impossible value for descriptor"
        mess/out 2 {PROID} "`{pixscale}': {{topicima},{pixscale}} -"
        mess/out 2 {PROID} "no calculation of seeing possible !"
        write/out " "
        set/format {FMTDEF}
    warn = 1
    goto no_seeing
  endif
  compute/table {phottable} :FWHM = :FWHM*{{topicima},{pixscale}} !pixel->arcsec
  !
  if nogood .ne. 1 then  ! we have good objects for median seeing calculation
    !
        mess/out 4 {PROID} " "
        mess/out 3 {PROID} "  seeing :"
        longlog/out 1 "     seeing :" >Null
    !
    select/table {phottable} :ELLIP.le.{MAXELLIP} -
                        .and.:STELLAR.ge.{MINSTELL} -
                        .and.:FWHM.ge.{MINFWHM} -
                        .and.:FLAGS.eq."  -------" >Null
    !                                ! "  -------": 2 blanks, 7 dashes !
    !
    if outputi(1) .ge. MINOBJ then
      !
      ! Create basic qc1log information
      qc1log/out 2
      if m$existd(topicima,h_dprtype) .eq. 1 then
        copy/dk {topicima} {h_dprtype} dpr_type
        qc1log/out 1 {topicima} {h_dprtype} {dpr_type} + DPR
      endif
 
      if m$existd(topicima,h_tpid) .eq. 1 then
        copy/dk {topicima} {h_tpid} tpl_id
        qc1log/out 1 {topicima} {h_tpid} "{tpl_id}" + TPL
      endif

      if m$existd(topicima,h_arcfile) .eq. 1 then
        copy/dk {topicima} {h_arcfile} arcfile
        qc1log/out 1 {topicima} {h_arcfile} {arcfile} + PRIMARY-FITS
      endif

      ! Get instrument name
      copy/dk {topicima} {h_instrume} dictinst
      if dictinst(1:4) .eq. "FORS" then
        dictinst = "FORS"
      endif

      if m$existd(topicima,h_filti(1)) .eq. 1 then
        copy/dk {topicima} {h_filti(1)} filter
        qc1log/out 1 {topicima} {h_filti(1)} "{filter}" + {dictinst}
      endif

      copy/ti {phottable} tmp.bdf
      set/refc {phottable} :FWHM
      show/table {phottable} >Null
      extract/image fwhm.bdf = tmp.bdf [<,@{outputi(4)}:>,@{outputi(4)}]
      set/refc {phottable} seq
      statistics/image fwhm.bdf + {FWHMBINSZ}  ? gn >Null
      fmode = outputr(12)
      ! Average of all points in the central three bins
      outputr(1) = fmode - 1.5 * FWHMBINSZ
      outputr(2) = fmode + 1.5 * FWHMBINSZ
      statistics/image fwhm.bdf + {FWHMBINSZ} {outputr(1)},{outputr(2)} gn >Null
      fmode = outputr(3)

      ! Robust sigma yields the uncertainty of the image quality
      @p fors_robust_sigma fwhm.bdf
      fsigma = outputr(1)
      -delete fwhm.bdf fwhm.tbl
      outputd = fsigma
      set/format T7.3,T5.3
      mess/out 3 {PROID} -
           "    measured FWHM (arcs)   = {fmode}  +- {outputd}"
      longlog/out 1 -
        "       measured FWHM (arcs)   = {fmode}  +- {outputd}" >Null
      qc1log/out 1 {topicima} {h_proqcfwhm} {fmode}
      qc1log/out 1 {topicima} {h_proqcfwsg} {fsigma}

      !
      ! SEEING FROM AMBIENT SERVER:
      ! --------------------------
      if m$existk("h_seestrt") .eq. 1 .and. m$existk("h_seeend") .eq. 1 then
        if m$existd(topicima,h_seestrt) .eq. 1 -
          .and. m$existd(topicima,h_seeend) .eq. 1 then
          outputr(1) = {{topicima},{h_seestrt}}
          outputr(2) = {{topicima},{h_seeend}}
          !
          mess/out 3 {PROID} "    ambi.serv.(expo.start) = {outputr(1)}"
          mess/out 3 {PROID} "    ambi.serv.(expo.end)   = {outputr(1)}"
          longlog/out 1 "       ambi.serv.(expo.start) = {outputr(1)}" >Null
          longlog/out 1 "       ambi.serv.(expo.end)   = {outputr(1)}" >Null
          !
        endif
      endif
      !
      set/format {FMTDEF}
      outputi(2) = totrowno+2
      write/table {phottable} :FWHM @{outputi(2)} {fmode}
      outputi(2) = totrowno+3
      write/table {phottable} :FWHM @{outputi(2)} {fsigma}


      ! Mean stellarity
      set/format T7.3,T5.3
               statistics/table {phottable} :STELLAR >Null
      mess/out 3 {PROID} "    mean stellarity        = {outputr(3)}"
      longlog/out 1 "       mean stellarity        = {outputr(3)}"
      qc1log/out 1 {topicima} {h_proqcstel} {outputr(3)}
      set/format {FMTDEF}
 
      qc1log/out 3
      !
      !- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      !
      if nice_crea .eq. "y" then
        !
        !*********************************************************************
        @p fors_photo_plotsee {phottable},{s2n},ELLIP,NPIX,FLAGS,{flux},FWHM -
          {MINS2N} {MAXELLIP} {MINPIX} {fmode},{fsigma}
        !*********************************************************************
        if {q1} .eq. -1 then
          warn = 1
        elseif {q1} .eq. 1 then
          write/out " "
          mess/out 1 {PROID} "Plotting of FWHM versus flux failed !"
          mess/out 1 {PROID} "({phottable})"
          goto error
        endif
        !
      endif
      !
      !- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      !
    else
      outputi(2) = totrowno+2
      write/table {phottable} :FWHM @{outputi(2)} {FAILNO}
      outputi(2) = totrowno+3
      write/table {phottable} :FWHM @{outputi(2)} {FAILNO}
        set/format F3.2,
        mess/out 2 {PROID} "    Too few objects !"
        mess/out 2 {PROID} "    no calculation of seeing possible !"
        set/format {FMTDEF}
      warn = 1
        longlog/out 1 "       calibration failed !" >Null
    endif
    select/table {phottable} all >Null
    !
  else
    !
    outputi(2) = totrowno+2
    write/table {phottable} :FWHM @{outputi(2)} {FAILNO}
    outputi(2) = totrowno+3
    write/table {phottable} :FWHM @{outputi(2)} {FAILNO}
    !
  endif
  !
  goto seeing_done
    no_seeing:
    compute/table {phottable} :FWHM = {FAILNO}
    outputi(2) = totrowno+2
    write/table {phottable} :FWHM @{outputi(2)} {FAILNO}
    outputi(2) = totrowno+3
    write/table {phottable} :FWHM @{outputi(2)} {FAILNO}
  seeing_done:
  !
  !
  !        CALCULATION OF MEDIAN TRAILING VALUE (ELLIPTICITY) :
  !        **************************************************
  !
  if nogood .ne. 1 then  ! we have good objects for median ellipt. calculation
    !
        mess/out 4 {PROID} " "
        mess/out 3 {PROID} "  object shapes :"
        longlog/out 1 "     object shapes :" >Null
    !
    select/table {phottable} -
      :NPIX.ge.{MINPIX}.and.:FLAGS.eq."  -------" >Null
    !                               ! "  -------": 2 blanks, 7 dashes !
    !
    if outputi(1) .ge. 3 then
      !
      copy/ti {phottable} tmp.bdf
      set/refc {phottable} :ELLIP
      show/table {phottable} >Null
      extract/image ellip.bdf = tmp.bdf [<,@{outputi(4)}:>,@{outputi(4)}]
      set/refc {phottable} seq
      statistics/image ellip.bdf + ? ? xn >Null
      emedian = outputr(8)
      esigma = outputr(4)
      -delete ellip.bdf
      !
        outputd = esigma
        set/format T7.3,T5.3
        mess/out 3 {PROID} -
               "    median ellipticity     = {emedian}  +- {outputd}"
        longlog/out 1 -
          "       median ellipticity     = {emedian}  +- {outputd}" >Null
        set/format {FMTDEF}
      outputi(2) = totrowno+2
      write/table {phottable} :ELLIP @{outputi(2)} {emedian}
      outputi(2) = totrowno+3
      write/table {phottable} :ELLIP @{outputi(2)} {esigma}
      !
    else
      outputi(2) = totrowno+2
      write/table {phottable} :ELLIP @{outputi(2)} {FAILNO}
      outputi(2) = totrowno+3
      write/table {phottable} :ELLIP @{outputi(2)} {FAILNO}
        mess/out 2 {PROID} "    Found less than 3 proper objects"
        mess/out 2 {PROID} "    with number of pixels >= {MINPIX} -"
        mess/out 2 {PROID} "    no calculation of median ellip. possible !"
      warn = 1
        longlog/out 1 "       calibration failed !" >Null
    endif
    select/table {phottable} all >Null
    !
  endif
  -delete tmp.bdf
  !
        shortlog/out 1 -
          "S/N, limit.mag., seeing, ellipticity computed." >Null
  !
  !
  !        CALCULATION OF AIRMASS AND MJD :
  !        ******************************
  !
        mess/out 4 {PROID} " "
        mess/out 3 {PROID} "  airmass :"
  !
  if m$existd(topicima,geolon) .eq. 1 -
    .and. m$existd(topicima,geolat) .eq. 1 then
    !
    if {{topicima},O_TIME(1)} .lt. 1990. .or. -
       {{topicima},O_TIME(1)} .ge. 2100. then
        set/format F15.14
        write/out " "
        mess/out 1 {PROID} "Impossible value for descriptor"
        mess/out 1 {PROID} "`O_TIME(1)': {{topicima},O_TIME(1)}"
        set/format {FMTDEF}
      goto error
    endif
    if {{topicima},O_TIME(5)} .lt. 0. .or. -
       {{topicima},O_TIME(5)} .ge. 24. then
        set/format F15.14
        write/out " "
        mess/out 1 {PROID} "Impossible value for descriptor"
        mess/out 1 {PROID} "`O_TIME(5)': {{topicima},O_TIME(5)}"
        set/format {FMTDEF}
      goto error
    endif
    if {{topicima},O_TIME(7)} .lt. 0. .or. -
       {{topicima},O_TIME(7)} .ge. 43200. then
        set/format F15.14
        write/out " "
        mess/out 1 {PROID} "Impossible value for descriptor"
        mess/out 1 {PROID} "`O_TIME(7)': {{topicima},O_TIME(7)}"
        set/format {FMTDEF}
      goto error
    endif
    if {{topicima},O_POS(1)} .lt. 0. .or. -
       {{topicima},O_POS(1)} .gt. 360. then
        set/format F15.14
        write/out " "
        mess/out 1 {PROID} "Impossible value for descriptor"
        mess/out 1 {PROID} "`O_POS(1)': {{topicima},O_POS(1)}"
        set/format {FMTDEF}
      goto error
    endif
    if {{topicima},O_POS(2)} .lt. -90. .or. -
       {{topicima},O_POS(2)} .gt. 90. then
        set/format F15.14
        write/out " "
        mess/out 1 {PROID} "Impossible value for descriptor"
        mess/out 1 {PROID} "`O_POS(2)': {{topicima},O_POS(2)}"
        set/format {FMTDEF}
      goto error
    endif
    if {{topicima},{geolon}} .lt. -180. .or. -
       {{topicima},{geolon}} .gt. 180. then
        set/format F15.14
        write/out " "
        mess/out 1 {PROID} "Impossible value for descriptor"
        mess/out 1 {PROID} "`{geolon}': {{topicima},{geolon}}"
        set/format {FMTDEF}
      goto error
    endif
    if {{topicima},{geolat}} .lt. -90. .or. -
       {{topicima},{geolat}} .gt. 90. then
        set/format F15.14
        write/out " "
        mess/out 1 {PROID} "Impossible value for descriptor"
        mess/out 1 {PROID} "`{geolat}': {{topicima},{geolat}}"
        set/format {FMTDEF}
      goto error
    endif
    !
    create/column {phottable} :{airmass} " " F7.4 R*4  ! col.21
    create/column {phottable} :MJD "day" F13.6 R*8   ! col.22
      -rename {topicima} currima.bdf  ! 'com/air' doesn't work with
                                      ! filenames longer than 20 or 30 chars. !
    compute/airmass currima.bdf -
      {currima.bdf,{geolon}} {currima.bdf,{geolat}} >Null
      -rename currima.bdf {topicima}
    !
    if outputd(20) .lt. 47892. .or. -
       outputd(20) .ge. 88069. then    ! [ 1990/01/01 - 2100/01/01 [
        write/out " "
        set/format T16.7
        mess/out 1 {PROID} "No calculation of airmass possible - impossible"
        mess/out 1 {PROID} "MJD : {outputd(20)} (1990/01/01=47892.)"
        mess/out 1 {PROID} "({topicima})"
        mess/out 1 {PROID} "one of these descriptors must be wrong:"
        mess/out 1 {PROID} "O_TIME(1) [yr.day] = {{topicima},O_TIME(1)}"
        mess/out 1 {PROID} "O_TIME(5) [hr.m+s] = {{topicima},O_TIME(5)}"
        set/format {FMTDEF}
      goto error
    endif
    !
    if outputr(1) .lt. 1. .or. outputr(1) .gt. 10. then
        write/out " "
        set/format T16.7
        mess/out 1 {PROID} "Impossible airmass : {outputr(1)}"
        mess/out 1 {PROID} "({topicima})"
        mess/out 1 {PROID} "one of these descriptors must be wrong:"
        mess/out 1 {PROID} "O_TIME(1) [yr.day] = {{topicima},O_TIME(1)}"
        mess/out 1 {PROID} "O_TIME(5) [hr.m+s] = {{topicima},O_TIME(5)}"
        mess/out 1 {PROID} "O_TIME(7) [expo_s] = {{topicima},O_TIME(7)}"
        mess/out 1 {PROID} "O_POS(1)  [RA]     = {{topicima},O_POS(1)}"
        mess/out 1 {PROID} "O_POS(2)  [DEC]    = {{topicima},O_POS(2)}"
        mess/out 1 {PROID} "{geolon}     =      {{topicima},{geolon}}"
        mess/out 1 {PROID} "{geolat}     =      {{topicima},{geolat}}"
        set/format {FMTDEF}
      goto error
    endif
    !
        set/format T7.3,
        mess/out 3 {PROID} "    time-weighted value    = {outputr(1)}"
        longlog/out 1 "     time-weighted value      = {outputr(1)}" >Null
        set/format {FMTDEF}
    !
    write/table {phottable} :{airmass} @1..{totrowno} {outputr(1)}
    write/table {phottable} :MJD @1..{totrowno} {outputd(20)}
    outputi(2) = totrowno+3  ! necessary for later 'fors_calflux' !
    write/table {phottable} :{airmass} @{outputi(2)} {outputr(1)} !
    !
  else
    !
        write/out " "
    if m$existd(topicima,geolon) .eq. 0 then
        mess/out 1 {PROID} "Header keyword missing:"
        mess/out 1 {PROID} "({topicima},{geolon})"
    endif
    if m$existd(topicima,geolat) .eq. 0 then
        mess/out 1 {PROID} "Header keyword missing:"
        mess/out 1 {PROID} "({topicima},{geolat})"
    endif
        mess/out 1 {PROID} "no calculation of airmass possible !"
    goto error
    !
  endif
  !
        shortlog/out 1 "Weighted airmass calculated." >Null
  !
  !
  !        FINISHING OUTPUT TABLE :
  !        **********************
  !
  if type .eq. "c" then  ! i.e., we have NO std.star frames
    !
    !if nogood .ne. 1 then  ! if only bad objects: "{FAILNO}"
      outputi(2) = totrowno+2
      ! {phottable},:FLAGS,{outputi(2)} = "<sn/mdgd>"
      {phottable},:FLAGS,{outputi(2)} = "<medgood>"
      outputi(2) = totrowno+3
      {phottable},:FLAGS,{outputi(2)} = "<lm/sggd>"
    !endif
    !
  else
    !
    outputi(2) = totrowno+2
    ! {phottable},:FLAGS,{outputi(2)} = "<s2n/med>"
    {phottable},:FLAGS,{outputi(2)} = "<median >"
    outputi(2) = totrowno+3
    {phottable},:FLAGS,{outputi(2)} = "<lim/sig>"
    !
  endif
  !
  !
  if delworld .eq. 1 then
    delete/column {phottable} :{ra}
    delete/column {phottable} :{dec}
    delworld = 0
  endif
  !
  !
  goto catloop
end_catloop:
!
!
!          E N D   O F   L O O P   W I T H   F R A M E S
!          * * * * * * * * * * * * * * * * * * * * * * *
!
!
!
!          OUTPUT :
!          ******
!
        mess/out 4 {PROID} " "
        mess/out 4 {PROID} ">>> All output files are in <{OUTCAT}> : <<<"
        mess/out 4 {PROID} ">>> Sky background frames are <*_skymap.bdf>. <<<"
        mess/out 4 {PROID} ">>> Sky-subtracted frames are <*_skysub.bdf>. <<<"
        mess/out 4 {PROID} ">>> Pos. and brightness. are in <*_phot.tbl>. <<<"
if mess_level .eq. 4 then
  read/icat {OUTCAT}
endif
if goodcount .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No proper objects found at all !"
  goto error
endif
!
!
set/format
if nice_crea .eq. "y" set/graphic default
!
!
if warn .eq. 0 then
  return 0 {skip} {OUTCAT}
else
  return -1 {skip} {OUTCAT}
endif
!
error:
  if m$existk("phottable") .eq. 1 then
    if m$exist(phottable) .eq. 1 select/table {phottable} all >Null
  endif
  set/format
  if m$existk("nice_crea") .eq. 1 then
    if nice_crea .eq. "y" set/graphic default
  endif
  return 1 -99 -99
!
!
!-------------------------------------------------------------------------------
