! $Id: fors_prepareext.prg,v 1.11 1999/09/16 21:07:06 nrainer Exp $   
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1998  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_prepareext.prg,v $
!
!.KEYWORDS    Prepare standard star spectra for extinction coefficient
!             computation, spectroscopy
!
!.USAGE       prepare/extcoeff stdcat extcat standards response area
!                              fittype fitpar rbpar
!
!.INPUT       stdcat      (P1)    Catalog with the standard star tables
!                                 to be merged.
!             standards   (P3)    3d table with positions and model spectra.
!                                 of standard stars in erg/s/cm^2/A.
!             response    (P4)    Table with instrumental response
!             area        (P5)    Telescope effective area.
!             ftype       (P6)    Type of response fit.
!             fpar        (P7)    Parameters for response fit.
!             rbpar       (P8)    Parameters for spectrum rebinning (wstart,
!                                 wend, wstep) 
!
!.OUTPUT      extcat      (P2)    Catalog with spectrum scale factor tables.
!
!.RETURNS     The recipe returns 0 on success, otherwise 1. Keyword Q1 is
!             updated with the exit status.
!
!.PURPOSE     The procedure takes all 3d standard star tables listed in the 
!             input catalog and corrects all standard star spectra for the
!             efficiency of the combination telescope and instrument. The
!             corrected observations are then divided by the reference
!             spectrum to obtain the spectrum scale factor that represents
!             the influence of atmospheric extinction effects for each
!             wavelength bin. Both, the input spectra and the reference flux
!             are rebinned to constant wavelength steps to make sure that
!             all extinction correction factors are defined over the same
!             wavelength range and that they have the same wavelength binning.
!
!.ENVIRON     FORS context, XMOS context
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1998/12/02:  Creation
!
!------------------------------------------------------------------------------

crossref  stdcat extcat standards response area ftype fpar rbpar

define/param p1 ? c "Standard star spectrum catalog:"
define/param p2 ? c "Catalog of scale factor tables:"
define/param p3 ? c "Standard flux database:"
define/param p4 ? c "System response table:"
define/param p5 ? c "Telescope effective area:"
define/param p6 + c "Type of response fit:"
define/param p7 + c "Parameter for reponse fit:"
define/param p8 ? c "Rebinning parameters:"
define/maxpar 8


! Procedure identification
define/local procID/c/1/20      "PrepareExt"


! Local constants
define/local sqm2sqcm/d/1/1   1.0e+4           ! conversion factor m^2 to cm^2
define/local A2cm/d/1/1       1.0e-8           ! conversion factor A to cm
define/local hplanck/d/1/1    6.6261e-27       ! Planck constant (cgs)
define/local clight/d/1/1     2.9979e+10       ! lightspeed in cm/s
define/local refstbl/c/1/16   "fors_refspec.tbl"
define/local refsbin/c/1/16   "fors_refsbin.tbl"
define/local refsfrm/c/1/16   "fors_refspec.bdf"
define/local extcor/c/1/16    "fors_extcor.bdf"
define/local extbase/c/1/16   "fors_ext"
define/local binspec/c/1/16   "fors_binspec.bdf"
define/local stdspec/c/1/16   "fors_stdspec.bdf"
define/local stdbase/c/1/8    "fors_std"


! Local variables
define/local i/i/1/1           0
define/local j/i/1/1           0
define/local catal/i/1/1       0
define/local ntbl/i/1/1        0
define/local nspc/i/1/1        0
define/local npar/i/1/1        0
define/local refno/i/1/1       0
define/local wnpix/i/1/1       0 
define/local area/r/1/1        0.
define/local exptime/r/1/1     0.
define/local airmass/r/1/1     0.
define/local conad/r/1/1       0.
define/local wstart/r/1/1      0.
define/local wend/r/1/1        0.
define/local wstep/r/1/1       0.
define/local cat_entry/c/1/80  " "  all
define/local tmpspec/c/1/100   " "  all
define/local exttbl/c/1/100    " "  all
define/local stdcat/c/1/100    {p1}
define/local extcat/c/1/100    {p2}
define/local flxlst/c/1/100    {p3}
define/local instrsp/c/1/100   {p4}
define/local fittyp/c/1/100    {p6}
define/local fitpar/c/1/100    {p7}



! Process parameters
if m$tstno(p5) .eq. 0 then
  mess/out 1 {procID} "Missing telescope effective area!"
  return 1
else
  area = {p5}
  if area .le. 0. then
    mess/out 1 {procID} "Invalid telescope effective area!"
    return 1
  else
    area = area * sqm2sqcm
  endif
endif

if m$tstno(p8(1:1)) .eq. 0 then
  mess/out 1 {procID} "No rebinning parameters provided!"
  return 1
else
  npar = m$parse(p8, "rbpar")
  if npar .ne. 3 then
    mess/out 1 {procID} "Invalid rebinning parameters encountered!"
    return 1
  else
    wstart = {rbpar01}
    wend = {rbpar02}
    wstep = {rbpar03}
    wnpix = (wend - wstart) / wstep
  endif
endif


! Get number of tables to be processed
show/icat {stdcat} N
ntbl = outputi(1)
if ntbl .le. 0 then
  mess/out 1 {procID} "Empty standard star catalog encountered!"
  return 1
endif


! Create output catalog
create/icat {extcat} NULL {FORS_IDSTRNG} >Null


! Convert spectra from 3d table format to one dimensional images
read_cat_entry:

store/frame cat_entry {stdcat} 1 exit_loop
j = j + 1

set/format i1
mess/out 3 {procID} " "
mess/out 3 {procID} "preparing standard star table {j} / {ntbl}"

nspc = m$value({cat_entry},tblcontr(4))


! Compute airmass from header information
get/airmass {cat_entry}
if {q1} .eq. 0 then
  airmass = outputr(1)
else
  mess/out 1 {procID} "Airmass computation failed!"
  return 1
endif

! Get exposure time and conversion factor (e- to adu) from spectrum header
exptime = m$value({cat_entry},o_time(7))

conad = m$value({cat_entry},{h_confac({fors_outi(1)})})
if fors_nout .gt. 1 then
  do i = 2 fors_nout
    conad = conad + m$value({cat_entry},{h_confac({fors_outi({i})})})
  enddo
  conad = conad / fors_nout

  set/format i1 f6.3
  mess/out 2 {procID} -
    "Multi port readout encountered! Using mean conversion factor {conad}!"
  set/format
endif


! Check existence of columns START, STEP and NDATA in object list.
if m$existc(cat_entry,":{c_sostart}") .lt. 0 then
  mess/out 1 {procID} "Column :{c_sostart} not found in {cat_entry}!"
  return 1
else if m$existc(cat_entry,":{c_sostep}") .lt. 0 then
  mess/out 1 {procID} "Column :{c_sostep} not found in {cat_entry}!"
  return 1
else if m$existc(cat_entry,":{c_sondata}") .lt. 0 then
  mess/out 1 {procID} "Column :{c_sondata} not found in {cat_entry}!"
  return 1
else
  copy/teima {cat_entry} :{c_soflux} 1,{nspc} :{c_sondata} -
         :{c_sostart},:{c_sostep} {stdbase} >Null
endif


! Create output 3d table for extinction correction factors
create/exttbl {extbase} {j} {wnpix}
if {q1} .eq. 0 then
  exttbl = outputc
  copy/dd {cat_entry} *,3 {exttbl}
else
  mess/out 1 {procID} "Cannot create extinction correction factor table!"
  return 1
endif


! Compute extinction correction factor for each standard star spectrum
do i = 1 nspc

  set/format i1
  mess/out 3 {procID} "  computing extinct.coeff. for spectrum {i} / {nspc}"

  set/format i4
    tmpspec = stdbase // "{i}" // ".bdf"
  set/format
  copy/dd {cat_entry} *,3 {tmpspec}


! Convert observed spectrum from ADU to electrons/cm^2/s/A
  compute/image {tmpspec} = {tmpspec} * {conad} / ({area} * {exptime} * -
                            {wstep}) >Null
  rebin/image {tmpspec} {binspec} {wstart},{wstep},{wnpix} pixel
  copy/dd {tmpspec} *,3 {binspec}


! Correct observed standard star spectrum for the instrument response
  calibrate/flux {binspec} {stdspec} {instrsp} no ratio {fittyp} -
                 {fitpar} spectrum
  copy/dd {binspec} *,3 {stdspec}


! Extract reference spectrum for current standard star from flux database
  refno = m$value({cat_entry},:{c_sorefid},{i})
  get/refflux {flxlst} {refno} {refstbl}


! Convert the reference spectrum from 1e+16 erg/cm^2/s/A to 
! electrons/cm^2/s/A (each registered photon produces 1 electron)
! and convert it to an image. 
  compute/table {refstbl} :{c_fxflux} = 1.0e-16 * :{c_fxflux} * -
         :{c_fxwave} * {A2cm} / ({hplanck} * {clight})


  rebin/table {refstbl} :{c_fxwave},{c_fxflux},{c_fxbwidth} {refsbin} -
       {wstart},{wstep},{wnpix} pixel

  convert/table {refsfrm} = {refsbin} :{c_fxwave} :{c_fxflux} {stdspec} plot


! Calculate extinction correction factor
  compute/ima {extcor} = -2.5 * log10({stdspec} / {refsfrm}) >Null


! Save correction factor to output table
  add/exttbl {extcor} {i} {exttbl} >Null


! Intermediate cleanup
  -delete {tmpspec}

enddo


! Add the preprocessed standard star table to the scale factor catalog
add/icat {extcat} {exttbl} >Null

goto read_cat_entry

exit_loop:


! Cleanup
-delete {stdspec}
-delete {binspec}
-delete {refstbl}
-delete {refsfrm}
-delete {refsbin}
-delete {extcor}

return 0
