! $Id: fors_preparestd.prg,v 1.6 1999/09/10 12:20:40 rpalsa Exp $   
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1998  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_preparestd.prg,v $
!
!.KEYWORDS    Prepare standards for response calculation, spectroscopy
!
!.USAGE       prepare/standard objects objno standards area stdspec refspec
!
!.INPUT       objects     (P1)    3d table with extracted standard star
!                                 spectra and reference to flux database
!                                 entries.
!             objno       (P2)    Entry number in object list to prepare.
!             standards   (P3)    3d table with positions and model spectra.
!                                 of standard stars in erg/s/cm^2/A.
!             area        (P4)    Telescope effective area.
!
!.OUTPUT      stdspec     (P5)    Single standard star spectrum stored as
!                                 1d image in terms of photons/s/cm^2/A.
!             refflux     (P6)    Reference flux spectrum table for the
!                                 selected standard star with flux in terms
!                                 of photons/s/cm^2/A.
!
!.RETURNS     The recipe returns 0 on success, otherwise 1. Keyword Q1 is
!             updated with the exit status.
!
!.PURPOSE     According to the given object number a standard star spectrum 
!             is selected from the input list and converted to a 1-dimensional
!             image.  The observed flux (ADU) is converted to photons/s/cm^2/A
!             using the exposure time and the conversion factor stored in the
!             spectrum header and the telescope effective area. The associated
!             reference flux table, given by the flux database entry column,
!             is extracted from the database and the flux values are converted
!             to photons/s/cm^2/A. 
!
!.ENVIRON     FORS context, XMOS context
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1999/01/04:  Creation
!
!------------------------------------------------------------------------------

crossref  objects objno standards area stdspec refspec

define/param p1 ? c "Standard star spectrum table:"
define/param p2 ? c "Standard star entry number:"
define/param p3 ? c "Standard flux database:"
define/param p4 ? c "Telescope effective area:"
define/param p5 + c "Standard star spectrum frame:"
define/param p6 + c "Reference flux table:"
define/maxpar 6


! Procedure identification
define/local procID/c/1/80	"PrepareStd"


! Local constants
define/local sqm2sqcm/d/1/1   1.0e+4           ! conversion factor m^2 to cm^2
define/local A2cm/d/1/1       1.0e-8           ! conversion factor A to cm
define/local hplanck/d/1/1    6.6261e-27       ! Planck constant (cgs)
define/local clight/d/1/1     2.9979e+10       ! lightspeed in cm/s
define/local stdbase/c/1/8    "fors_std"


! Local variables
define/local i/i/1/1          0
define/local objno/i/1/1      0
define/local flxno/i/1/1      0
define/local ndata/i/1/1      0
define/local area/r/1/1       0.
define/local exptime/r/1/1    0.
define/local conad/r/1/1      0.
define/local wstep/r/1/1      0.
define/local tmpfile/c/1/100  " "   all
define/local stdlst/c/1/100   {p1}
define/local flxlst/c/1/100   {p3}
define/local stdspec/c/1/100  {p5}
define/local refspec/c/1/100  {p6}



! Process parameters
if m$tstno(p2) .eq. 0 then
  mess/out 1 {procID} "Missing object entry number!"
  return 1
else
  objno = {p2}
  if objno .lt. 1 .or. objno .gt. m$value({stdlst},tblcontr(4)) then
    mess/out 1 {procID} "Invalid object entry number!"
    return 1
  endif
endif

if m$tstno(p4) .eq. 0 then
  mess/out 1 {procID} "Missing telescope effective area!"
  return 1
else
  area = {p4}
  if area .le. 0. then
    mess/out 1 {procID} "Invalid telescope effective area!"
    return 1
  else
    area = area * sqm2sqcm
  endif
endif


! Check existence of flux database reference column
if m$existc(stdlst,":{c_sorefid}") .lt. 0 then
  mess/out 1 {procID} -
      "Reference column to flux database not found in {stdlst}!"
  return 1
else
  flxno = m$value({stdlst},:{c_sorefid},{objno})
endif


! Check existence of columns START, STEP and NDATA in object list.
if m$existc(stdlst,":{c_sostart}") .lt. 0 then
  mess/out 1 {procID} "Column :{c_sostart} not found in {stdlst}!"
  return 1
else if m$existc(stdlst,":{c_sostep}") .lt. 0 then
  mess/out 1 {procID} "Column :{c_sostep} not found in {stdlst}!"
  return 1
else if m$existc(stdlst,":{c_sondata}") .lt. 0 then
  mess/out 1 {procID} "Column :{c_sondata} not found in {stdlst}!"
  return 1
else
  copy/teima {stdlst} :{c_soflux} {objno} :{c_sondata} -
     :{c_sostart},:{c_sostep} {stdbase} >Null
  set/format i4
  tmpfile = stdbase // "{objno}" // ".bdf"
  set/format
  copy/dd {stdlst} *,3 {tmpfile}


! Get exposure time and conversion factor (e- to adu) from spectrum header
  exptime = m$value({tmpfile},o_time(7))

  conad = m$value({tmpfile},{h_confac({fors_outi(1)})})
  if fors_nout .gt. 1 then
    do i = 2 fors_nout
      conad = conad + m$value({tmpfile},{h_confac({fors_outi({i})})})
    enddo
    conad = conad / fors_nout

    set/format i1 f6.3
    mess/out 2 {procID} -
      "Multi port readout encountered! Using mean conversion factor {conad}!"
    set/format
  endif

  wstep = m$value({tmpfile},step(1))
  compute/dimage {stdspec} = {tmpfile} * {conad} / ({area} * {exptime} * -
                             {wstep})
endif


! Get reference spectrum from flux database
get/refflux {flxlst} {flxno} {refspec}


! Convert fluxes from 1e+16 erg/s/cm^2/A to photons/s/cm^2/A
compute/table {refspec} :{c_fxflux} = 1.0e-16 * :{c_fxflux} * -
       :{c_fxwave} * {A2cm} / ({hplanck} * {clight})


! Cleanup
-delete {tmpfile}

return 0
