! $Id: regress.prg,v 1.16 2000/12/14 17:03:14 nrainer Exp $
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
! .COPYRIGHT  (C) 1999  European Southern Observatory (ESO)
!
! .IDENT      regress/general
!
! .KEYWORDS   General regression
!
! .PURPOSE    Computation of regression coefficients and their errors for one
!             variable with different methods and calculation of resulting
!             fitted values and residuals.
!             The original values and the fitted ones and their errors can be
!             plotted.
!             The plot can be saved as postscript file.
! __________
!
! .USAGE      regress/general tbl cols [meth] [mode] [graph] [show] [ps] [flags]
!
! .INPUT      tbl   (P1): table name;
!                           default: none
!
!             cols  (P2): independent (x) and dependent (y) column,
!                         separated by a comma;
!                           default: none,none
!
!             meth  (P3): one of the following regression methods:
!                         <poly>,deg : polynomial with degree "deg" (<=6),
!                         <lin>      : linear,
!                         <slope>,slp: linear w. const. slope "slp",
!                         <zero>,zero: linear w. const. zeropoint "zero";
!                           default: <lin>
!                           examples:  poly,5  OR  slope,-0.25  OR  zero,0.
!
!             mode  (P4): mode of calculating the methods:
!                         <avg>: normal regression,
!                         <rob>: for meth=lin       : Hough transform. (only
!                                                     reasonable f.#val.>>5),
!                                for meth=slope/zero: median instead of mean;
!                           default: <avg>
!                           hint: ignored, if meth=poly
!
!             graph (P5): <n> for no graphic OR number of graph. window (0-9)
!                         -if the window exists it is used, else it is created-
!                         AND flag for background colour: <w>hite or <b>lack,
!                         separated by a comma;
!                           default: 1,b
!
!             show  (P6): flag for result output on terminal: <y>:show, <n>o;
!                           default: <y>
!
!             ps    (P7): flag for creation of a postscript file of the plot:
!                         <n>:do not create, <y>es,create;
!                           default: <n>
!                           hint: ignored, if graph=n
!
!             flags (P8): flag for setting plot characteristics to defaults:
!                         <y>es, <n>:accept previous setttings (e.g. scales
!                         and offsets, but not stype/ltype/bcolour/colour/
!                         yaxis/lwidth) and do not reset to defaults at the end
!                         AND flag for plotting mean(mode(P4)=avg) or median
!                         (mode(P4)=rob) "centre of gravity":
!                         <y>es, <n>o,
!                         AND flag for plotting name of table above plot:
!                         <y>es, <n>o,
!                         separated by commas;
!                           default: <y>,<y>,<y>
!                           hint: ignored, if graph=n
!
! .OUTPUT     OUTPUTD  (1): zero point
!                      (2): slope
!                 meth=poly:
!                   (3..7): higher coefficients
!
!                     (11): standard deviation of zero point
!                     (12): standard deviation of slope
!                 meth=poly:
!                 (13..17): higher deviations
!
!             Q2: name of table column with fitted values
!             Q3: name of table column with residuals
!                 "{P2[2](1:max12)}_[FIT|RES]" type:R*8 format:{OUT_FMT}
!                 (see below) - existing columns with type R*8 are used, else
!                 they are created - the formats are corrected, if necessary
!
! .RETURN     Q1:  0: successful return
!                  1: an error occured
! __________
!
! .ALGORITHM  1. Parameter tests
!             2. Creating of table columns for results
!             3. Checking number of table rows
!             -  (Plotting of original values)
!             4. Computation of regression coefficients and fitted values
!             -  (Plotting fitted values / saving plot as postscript file)
!             5. Saving fitted values in table columns
!                and regression coefficients in global key
!
! .COMMENT    -
!
! .ENVIRON    -
!
! .LANGUAGE   MIDAS 99NOV
!
! .AUTHOR     N.Rainer  DPG/DMD/ESO
!
! .VERSION    1.5  99/11/25  Creation
!                  00/       Release
!
!------------------------------------------------------------------------------
!
!
crossref tbl cols meth mode graph show ps flags
!
!
define/local MAXDEGREE/i/1/1 6          ! limit of 'regress/poly'
define/local MINGRA/i/1/1    0          ! low limit of graph.win.no.
define/local MAXGRA/i/1/1    9          ! high limit of graph.win.no.
! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
define/local R8_TCOL/i/1/1   452984833
!                            four-five-two nine-eight-four eight-three-three
define/local I4_TCOL/i/1/1   167772161
!                            one-six-seven seven-seven-two one-six-one
!  $MIDASHOME/$MIDVERS/incl/tblsys.h:
!  bits: TBL_D_R8 033; TBL_ElementType(t) (t>>24); TBL_Items(t) (t&((1<<24)-1))
!  ==> decimal:    27                      *2^24                    +1 (depth)
!  bits: TBL_D_I4 012;
!  ==> decimal:    10
! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
define/local OUT_FMT/c/1/6   "E23.15"   ! format for cols. w. fitted vals./res.
define/local REGDESC/c/1/6   "REGRESS"  ! name for new desc. w. regress.coeffs.
!
!
!  For `set/format':
define/local FMTDEF/c/1/17 "I1 E14.7,E23.15"
!
!  For `mess/out':
define/local PROID/c/1/40 "regress/general"
if m$existk("mess_level") .eq. 0 write/keyword mess_level/i/1/1 3
!
!  For return values:
write/keyword q1 0
write/keyword q2 0
write/keyword q3 0
!
!
if p1 .eq. "??" then
  w/o "      *****************************************************************"
  w/o "      tbl   (P1): table name;"
  w/o "                    default: none"
  w/o " "
  w/o "      cols  (P2): independent (x) and dependent (y) column,"
  w/o "                  separated by a comma;"
  w/o "                    default: none,none"
  w/o " "
  w/o "      meth  (P3): one of the following regression methods:"
  w/o "                  <poly>,deg : polynomial with degree "deg" (<=6),"
  w/o "                  <lin>      : linear,"
  w/o "                  <slope>,slp: linear w. const. slope "slp","
  w/o "                  <zero>,zero: linear w. const. zeropoint "zero";"
  w/o "                    default: <lin>"
  w/o "                    examples:  poly,5  OR  slope,-0.25  OR  zero,0."
  w/o " "
  w/o "      mode  (P4): mode of calculating the methods:"
  w/o "                  <avg>: normal regression,"
  w/o "                  <rob>: for meth=lin       : Hough transform. (only"
  w/o "                                              reasonable f.#val.>>5),"
  w/o "                         for meth=slope/zero: median instead of mean;"
  w/o "                    default: <avg>"
  w/o "                    hint: ignored, if meth=poly"
  w/o " "
  w/o "      graph (P5): <n> for no graphic OR number of graph. window (0-9)"
  w/o "                  -if the window exists it is used, else it is created-"
  w/o "                  AND flag for background colour: <b>lack or <w>hite,"
  w/o "                  separated by a comma;"
  w/o "                    default: 1,b"
  w/o " "
  w/o "      show  (P6): flag for result output on terminal: <y>:show, <n>o;"
  w/o "                    default: <y>"
  w/o " "
  w/o "      ps    (P7): flag for creation of a postscript file of the plot:"
  w/o "                  <n>:do not create, <y>es,create;"
  w/o "                    default: <n>"
  w/o "                    hint: ignored, if graph=n"
  w/o " "
  w/o "      flags (P8): flag for setting plot characteristics to defaults:"
  w/o "                  <y>es, <n>:accept previous setttings (e.g. scales"
  w/o "                  and offsets, but not stype/ltype/bcolour/colour/"
  w/o "                  yaxis/lwidth) and do not reset to defaults at the end"
  w/o "                  AND flag for plotting mean(mode(P4)=avg) or median"
  w/o "                  (mode(P4)=rob) "centre of gravity":"
  w/o "                  <y>es, <n>o,"
  w/o "                  AND flag for plotting name of table above plot:"
  w/o "                  <y>es, <n>o,"
  w/o "                  separated by commas;"
  w/o "                    default: <y>,<y>,<y>"
  w/o "                    hint: ignored, if graph=n"
  w/o "      _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _"
  w/o "      OUTPUTD  (1): zero point"
  w/o "               (2): slope"
  w/o "          meth=poly:"
  w/o "            (3..7): higher coefficients"
  w/o " "
  w/o "              (11): standard deviation of zero point"
  w/o "              (12): standard deviation of slope"
  w/o "          meth=poly:"
  w/o "          (13..17): higher deviations"
  w/o " "
  w/o "      Q2: name of table column with fitted values"
  w/o "      Q3: name of table column with residuals"
  w/o "          P2[2](1:max12)_[FIT,RES] type:R*8 format:{OUT_FMT}"
  w/o "          - existing columns with type R*8 are used, else created"
  w/o "          - the formats are corrected, if necessary"
  w/o "      *****************************************************************"
  return/exit
endif
!
!
set/format {FMTDEF}
!
!
!          INPUT / PARAMETER TESTS :
!          ***********************
!
!P1(tbl):
!-------
if p1 .eq. "?" mess/out 2 {PROID} " "
define/parameter p1 ? ? -
  "        Enter name of table [def.:none]: "
define/local intab/c/1/64 "{p1}"
if m$index(intab,".tbl") .eq. 0 intab = intab//".tbl"
!
define/local updaflag/i/1/1 1
if m$exist(intab) .eq. 0 then
        write/out "      "
        mess/out 1 {PROID} "Entered table is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({intab})"
  goto error
endif
if m$existd(intab,"HISTORY_UPDA") .eq. 1 then
  updaflag = {{intab},HISTORY_UPDA}
endif
write/desc {intab} HISTORY_UPDA/i/1/1 0  ! disable update of HISTORY desc.
!
!
!P2(cols):
!--------
if p2 .eq. "?" mess/out 2 {PROID} " "
define/parameter p2 ? ? -
  "        Enter indep.(x),dep.(y) col. [def.:none,none]: "
outputi = m$parse(p2,"cols")
if outputi .ne. 2 then
        write/out "      "
        mess/out 1 {PROID} "Wrong number of elements in P2(cols) !"
  goto error
endif
define/local xcol/c/1/16 "{cols01}"
if xcol(1:1) .eq. ":" xcol = xcol(2:)
define/local ycol/c/1/16 "{cols02}"
if ycol(1:1) .eq. ":" ycol = ycol(2:)
!
if m$existc(intab,":{xcol}") .eq. -1 then
        write/out "      "
        mess/out 1 {PROID} "Entered independent column is not present !"
        mess/out 1 {PROID} "({intab},:{xcol})"
  goto error
elseif m$existc(intab,":{ycol}") .eq. -1 then
        write/out "      "
        mess/out 1 {PROID} "Entered dependent column is not present !"
        mess/out 1 {PROID} "({intab},:{ycol})"
  goto error
endif
!
!
!P3(meth):
!--------
define/parameter p3 l ? -
  "        Regress. method [<poly>,deg|def.:<lin>|<slope>,slp|<zero>,zero]: "
define/local meth/c/1/1 "{p3(1:1)}"
if meth .eq. "+" meth = "l"
inputi = 0
if meth .eq. "l" inputi = -1
if meth .eq. "p" inputi = 1
if meth .eq. "s" inputi = 1
if meth .eq. "z" inputi = 1
if inputi .eq. 0 then
        write/out "      "
        mess/out 1 {PROID} "Entered regression method does not exist !"
        mess/out 1 {PROID} "({p3})"
  goto error
endif
if inputi .eq. 1 then
  outputi = m$parse(p3,"meths")
  if outputi .ne. 2 then
        write/out "      "
        mess/out 1 {PROID} "Wrong number of elements in P3(meth) !"
    goto error
  endif
endif
define/local degree/i/1/1 1  ! "1" as initial value needed for below
define/local slope/d/1/1 0
define/local zero/d/1/1 0
if meth .eq. "p" then
  degree = {meths02}
  if degree .lt. 1 .or. degree .gt. {MAXDEGREE} then
        write/out "      "
        mess/out 1 {PROID} "Degree for polynomial out of range !"
        mess/out 1 {PROID} "({degree} ; 1<=degree<={MAXDEGREE})"
    goto error
  endif
elseif meth .eq. "s" then
  slope = {meths02}
elseif meth .eq. "z" then
  zero = {meths02}
endif
!
!
!P4(mode):
!--------
define/parameter p4 a ? -
  "        Calculation mode [def.:<avg>|<rob>]: "
define/local mode/c/1/1 "{p4(1:1)}"
if mode .eq. "+" mode = "a"
if mode .ne. "a" .and. mode .ne. "r" then
        write/out "      "
        mess/out 1 {PROID} "Entered calculation mode does not exist !"
        mess/out 1 {PROID} "({p4})"
  goto error
endif
!
!
!P5(graph):
!---------
define/parameter p5 1,b ? -
  "        Graph.wnd.[<n>|def.:1|{MINGRA}-{MAXGRA}],bck.col.[<w>|def.<b>]: "
outputi = m$parse(p5,"graph")
if m$tstno(graph01) .eq. 1 .and. outputi .ne. 2 then
        write/out "      "
        mess/out 1 {PROID} "Wrong number of elements in P5(graph) !"
  goto error
endif
define/local graphcrea/c/1/1 " "
define/local graphwin/i/1/1 0
if m$tstno(graph01) .eq. 1 then
  graphcrea = "y"
  graphwin = {graph01}
  if graphwin .lt. MINGRA .or. graphwin .gt. MAXGRA then
        write/out "      "
        mess/out 1 {PROID} "Number of graphic window out of range !"
        mess/out 1 {PROID} "({graphwin} ; {MINGRA}<=window#<={MAXGRA})"
    goto error
  endif
else
  graphcrea = "n"
endif
if m$tstno(graph01) .eq. 1 define/local backcol/c/1/1 "{graph02}"
!
!
!P6(show):
!--------
define/parameter p6 y ? -
  "        Flag for result output on terminal [def.:<y>|<n>]: "
define/local show/c/1/1 "{p6}"
!
!
!P7(ps):
!------
define/parameter p7 + ? -
  "        Flag for PS file of plot [def.:<n>|<y>es,create]: "
define/local ps/c/1/1 "{p7}"
!
!
!P8(flags):
!---------
define/parameter p8 + ? -
  "        Clear plot set.,plot centre of grav.,tbl.name [def.:<y,y,y>|<n..>]: "
outputi = m$parse(p8,"flags")
define/local pclear/c/1/1 " "
define/local pcentre/c/1/1 " "
define/local pname/c/1/1 " "
if outputi .eq. 3 then
  pclear = flags01
  pcentre = flags02
  pname = flags03
elseif outputi .eq. 1 then  ! defaults
  pclear = "y"
  pcentre = "y"
  pname = "y"
else
        write/out "      "
        mess/out 1 {PROID} "Wrong number of elements in P8(cc) !"
  goto error
endif
!
!
!
!          DEFINITION OF FURTHER VARIABLES :
!          *******************************
!
  define/local refcol/i/1/1      0
  define/local fitcol/c/1/16     " " all
  define/local rescol/c/1/16     " " all
  define/local selcol/c/1/16     " " all
  define/local rowloop/i/1/1     0
  define/local idxcol/c/1/16     " " all
  !
  define/local wrktab/c/1/64     " " all
  define/local nsel/i/1/1        0
  define/local toofew/i/1/1      0
  !
  define/local degloop/i/1/1     0
  define/local plusone/i/1/1     0
  inputi = degree+1
  define/local coef/d/1/{inputi} 0. all
  define/local cerr/d/1/{inputi} 0. all
  define/local xval/d/1/1 0.
  define/local xerr/d/1/1 0.
  define/local yval/d/1/1 0.
  define/local yerr/d/1/1 0.
  define/local zval/d/1/1 0.
  define/local zerr/d/1/1 0.
  define/local sval/d/1/1 0.
  define/local serr/d/1/1 0.
  define/local ymin/d/1/1 0.
  define/local ymax/d/1/1 0.
  define/local fmin/d/1/1 0.
  define/local fmax/d/1/1 0.
  define/local yws/r/1/1 0.
  define/local ywe/r/1/1 0.
  define/local ycs/r/1/1 0.
  define/local yce/r/1/1 0.
  define/local yw1/r/1/1 0.
  define/local ywl/r/1/1 0.
!
!
!
!          CREATING ADDITIONAL TABLE COLUMNS :
!          *********************************
!
        mess/out 4 {PROID} "    preparing output"
!
refcol = {{intab},TBLCONTR(5)}
!
fitcol = m$upper(ycol(1:12)) // "_FIT"  ! max.length of tbl.cols. = 16 chars.
if m$existc(intab,":{fitcol}") .eq. -1 then            ! missing column
  create/column {intab} :{fitcol} " " {OUT_FMT} R*8
elseif m$existc(intab,":{fitcol}") .gt. 0 then
  set/refc {intab} :{fitcol}
  inputi = {{intab},TBLCONTR(5)}
  if {{intab},TBLENGTH({inputi})} .ne. R8_TCOL then    ! wrong column type
    delete/column {intab} :{fitcol}
    create/column {intab} :{fitcol} " " {OUT_FMT} R*8
  else
    set/format I3
    inputc = "{{intab},TLABL{inputi}(33:38)}"          ! removes trail. blanks
    if inputc .ne. OUT_FMT then                        ! wrong column format
      name/column {intab} :{fitcol} ? " " {OUT_FMT}
    endif
    set/format {FMTDEF}
  endif
endif
!
rescol = m$upper(ycol(1:12)) // "_RES"  ! max.length of tbl.cols. = 16 chars.
if m$existc(intab,":{rescol}") .eq. -1 then            ! missing column
  create/column {intab} :{rescol} " " {OUT_FMT} R*8
elseif m$existc(intab,":{rescol}") .gt. 0 then
  set/refc {intab} :{rescol}
  inputi = {{intab},TBLCONTR(5)}
  if {{intab},TBLENGTH({inputi})} .ne. R8_TCOL then    ! wrong column type
    delete/column {intab} :{rescol}
    create/column {intab} :{rescol} " " {OUT_FMT} R*8
  else
    set/format I3
    inputc = "{{intab},TLABL{inputi}(33:38)}"          ! removes trail. blanks
    if inputc .ne. OUT_FMT then                        ! wrong column format
      name/column {intab} :{rescol} ? " " {OUT_FMT}
    endif
    set/format {FMTDEF}
  endif
endif
!
selcol = "PreSel"
if m$existc(intab,":{selcol}") .eq. -1 then            ! missing column
  create/column {intab} :{selcol} " " I6 I*4
elseif m$existc(intab,":{selcol}") .gt. 0 then
  set/refc {intab} :{selcol}
  inputi = {{intab},TBLCONTR(5)}
  if {{intab},TBLENGTH({inputi})} .ne. I4_TCOL then    ! wrong column type
    delete/column {intab} :{selcol}
    create/column {intab} :{selcol} " " I6 I*4
  else
    set/format I3
    inputc = "{{intab},TLABL{inputi}(33:38)}"          ! removes trail. blanks
    if inputc .ne. "I6" then                           ! wrong column format
      name/column {intab} :{selcol} ? " " I6
    endif
    set/format {FMTDEF}
  endif
endif
!
do rowloop = 1 {{intab},TBLCONTR(4)}
  {intab},:{selcol},@{rowloop} = {{intab},sel,@{rowloop}}  ! for restor. below
enddo
!
idxcol = "IdxSel"
if m$existc(intab,":{idxcol}") .eq. -1 then            ! missing column
  create/column {intab} :{idxcol} " " I6 I*4
elseif m$existc(intab,":{idxcol}") .gt. 0 then
  set/refc {intab} :{idxcol}
  inputi = {{intab},TBLCONTR(5)}
  if {{intab},TBLENGTH({inputi})} .ne. I4_TCOL then    ! wrong column type
    delete/column {intab} :{idxcol}
    create/column {intab} :{idxcol} " " I6 I*4
  else
    set/format I3
    inputc = "{{intab},TLABL{inputi}(33:38)}"          ! removes trail. blanks
    if inputc .ne. "I6" then                           ! wrong column format
      name/column {intab} :{idxcol} ? " " I6
    endif
    set/format {FMTDEF}
  endif
endif
!
compute/table {intab} :{idxcol} = seq                      ! for copy/tt below
!
set/refc {intab} #{refcol}
!
!
!          CHECKING NUMBER OF SELECTED AND NON-EMPTY TABLE ROWS :
!          ****************************************************
!
        mess/out 4 {PROID} "    checking number of values"
!
select/table {intab} :{selcol}.eq.1 >Null
select/table {intab} sel.and.:{xcol}.ne.NULL.and.:{ycol}.ne.NULL >Null
nsel = outputi
wrktab = "midreg.tbl"
copy/table {intab} {wrktab}
select/table {intab} :{selcol}.eq.1 >Null
!
!
!       ^^^^^^^^^^^^^^^^^
!       ^   I N T A B   ^
! - - - + - - - - - - - + - - - - - - - - - - - - - - - - - - - - - - - - - - -
!       v  W R K T A B  v
!       vvvvvvvvvvvvvvvvv
!
!
if meth .eq. "p" then                ! <POLY>,DEG
    if nsel .le. degree toofew = 1
elseif meth .eq. "l" then            ! <LIN>
  if mode .eq. "a" then
    if nsel .lt. 3 toofew = 1        ! ### reg/lin not w. 2 vals. !
  elseif mode .eq. "r" then
    if nsel .lt. 5 toofew = 1
  endif
elseif meth .eq. "s" then            ! <SLOPE>,SLP
    if nsel .lt. 1 toofew = 1
elseif meth .eq. "z" then            ! <ZERO>,ZERO
    if nsel .lt. 1 toofew = 1
endif
!
if toofew .eq. 1 then
        write/out "      "
        mess/out 1 {PROID} "Too few (selected) values - at least necessary:"
        mess/out 1 {PROID} "  meth \ mode | AVG ROB"
        mess/out 1 {PROID} "  ------------+--------"
        mess/out 1 {PROID} "  POLY,d      | d+1  -"
        mess/out 1 {PROID} "  LIN         |  3   5"  ! ### reg/lin not w.2vals.
        mess/out 1 {PROID} "  SLOPE       |  1   1"
        mess/out 1 {PROID} "  ZERO        |  1   1"
  goto error
endif
!
!
!          CREATION OR ASSIGNMENT OF GRAPHIC WINDOW :
!          ****************************************
!
if graphcrea .eq. "y" then
  !
        mess/out 4 {PROID} "    preparing plot"
  !
  inputi = graphwin+11  ! winopen(11+i) reflects gra.wnd. 0+i
  if winopen({inputi}) .eq. 0 then
    inputi(7) = ididev(7)  ! create/graphic deletes TrueColour info
    create/graphic {graphwin} 710,506,0,692 450000 >Null
    ididev(7) = inputi(7)
  elseif winopen({inputi}) .eq. 1 then
    assign/graphic g,{graphwin} >Null
  elseif winopen({inputi}) .eq. 2 then
    assign/graphic g,{graphwin} >Null
    modify/graphic window
  endif
  if pclear .ne. "n" then
    set/graphic default
    set/graphic pmode=0
  endif
  set/graphic stype=1 ltype=0
  !
endif
!
!
!          COMPUTATION OF REGRESSION COEFFICIENTS
!          AND FITTED VALUES :
!          **************************************
!
        mess/out 4 {PROID} "    computing regression coeff. / fitted values"
!
if meth .eq. "p" then        ! <POLY>,DEG
  !
  regression/poly {wrktab} :{ycol} :{xcol} {degree} >Null
  plusone = degree+1
  do degloop = 1 plusone
    coef({degloop}) = outputd({degloop})  ! coef(1) is zero point
    cerr({degloop}) = 0.                  ! ### not supported yet
  enddo
  save/regression {wrktab} {REGDESC}
  compute/regression {wrktab} :{fitcol} = {REGDESC}
  compute/table {wrktab} :{rescol} = :{ycol}-:{fitcol}
  if graphcrea .eq. "y" then
    statistics/table {wrktab} :{xcol} >Null
    xval = outputr(3)
    statistics/table {wrktab} :{ycol} >Null
    ymin = outputr(1)
    yval = outputr(3)
    ymax = outputr(2)
    statistics/table {wrktab} :{fitcol} >Null
    fmin = outputr(1)
    fmax = outputr(2)
  endif
  !
elseif meth .eq. "l" then    ! <LIN>
  if mode .eq. "a" then
    !
    regression/linear {wrktab} :{ycol} :{xcol} >Null
    coef(1) = outputd(1)     ! zero point
    cerr(1) = outputr(2)
    coef(2) = outputd(2)     ! slope
    cerr(2) = outputr(4)
    compute/table {wrktab} :{fitcol} = {coef(1)}+{coef(2)}*:{xcol}
    compute/table {wrktab} :{rescol} = :{ycol}-:{fitcol}
    if graphcrea .eq. "y" then
      statistics/table {wrktab} :{xcol} >Null
      xval = outputr(3)
      statistics/table {wrktab} :{ycol} >Null
      ymin = outputr(1)
      yval = outputr(3)
      ymax = outputr(2)
      statistics/table {wrktab} :{fitcol} >Null
      fmin = outputr(1)
      fmax = outputr(2)
    endif
    !
  elseif mode .eq. "r" then
    !
        ! Only because of error in regres/robu up to version 99NOVpl2.0:
        define/local rename_fit/i/1/1 0
        if fitcol .ne. "TMP_FIT" then
          name/column {wrktab} :{fitcol} :TMP_FIT
          rename_fit = 1
        endif
        define/local rename_res/i/1/1 0
        if rescol .ne. "TMP_RES" then
          name/column {wrktab} :{rescol} :TMP_RES
          rename_res = 1
        endif
    !regression/robust {wrktab} :{ycol} :{xcol} -
    !                  midtmp.dat :{fitcol} :{rescol} y >Null
    regression/robust {wrktab} :{ycol} :{xcol} -
                      midtmp.dat :TMP_FIT :TMP_RES y >Null
        ! Only because of error in regres/robu up to version 99NOVpl2.0:
        if rename_fit .eq. 1 then
          name/column {wrktab} :TMP_FIT :{fitcol}
        endif
        if rename_res .eq. 1 then
          name/column {wrktab} :TMP_RES :{rescol}
        endif
    coef(1) = outputr(2)     ! zero point
    cerr(1) = outputr(4)
    coef(2) = outputr(1)     ! slope
    cerr(2) = outputr(3)
    -delete midtmp.dat
    if graphcrea .eq. "y" then
      copy/ti {wrktab} midtmp.bdf
      set/refc {wrktab} :{xcol}
      inputi = {{wrktab},TBLCONTR(5)}
      statistics/image midtmp.bdf [<,@{inputi}:>,@{inputi}] ? ? xn >Null
      xval = outputr(8)
      set/refc {wrktab} :{ycol}
      inputi = {{wrktab},TBLCONTR(5)}
      statistics/image midtmp.bdf [<,@{inputi}:>,@{inputi}] ? ? xn >Null
      yval = outputr(8)
      ymin = outputr(1)
      ymax = outputr(2)
      set/refc {wrktab} seq
      -delete midtmp.bdf
      statistics/table {wrktab} :{fitcol} >Null
      fmin = outputr(1)
      fmax = outputr(2)
    endif
    !
  endif
elseif meth .eq. "s" then    ! <SLOPE>,SLP
  if mode .eq. "a" then
    !
    statistics/table {wrktab} :{xcol} >Null
    xval = outputr(3)
    xerr = outputr(4)
    statistics/table {wrktab} :{ycol} >Null
    yval = outputr(3)
    yerr = outputr(4)
    if graphcrea .eq. "y" then
      ymin = outputr(1)
      ymax = outputr(2)
    endif
    zval = yval-slope*xval
    zerr = m$sqrt(yerr**2+(slope*xerr)**2)
    coef(1) = zval           ! zero point (mean)
    cerr(1) = zerr
    coef(2) = slope          ! slope
    cerr(2) = 0.
    compute/table {wrktab} :{fitcol} = {coef(1)}+{coef(2)}*:{xcol}
    compute/table {wrktab} :{rescol} = :{ycol}-:{fitcol}
    if graphcrea .eq. "y" then
      statistics/table {wrktab} :{fitcol} >Null
      fmin = outputr(1)
      fmax = outputr(2)
    endif
    !
  elseif mode .eq. "r" then
    !
    copy/ti {wrktab} midtmp.bdf
    set/refc {wrktab} :{xcol}
    inputi = {{wrktab},TBLCONTR(5)}
    statistics/image midtmp.bdf [<,@{inputi}:>,@{inputi}] ? ? xn >Null
    xval = outputr(8)
    xerr = outputr(4)
    set/refc {wrktab} :{ycol}
    inputi = {{wrktab},TBLCONTR(5)}
    statistics/image midtmp.bdf [<,@{inputi}:>,@{inputi}] ? ? xn >Null
    yval = outputr(8)
    yerr = outputr(4)
    if graphcrea .eq. "y" then
      ymin = outputr(1)
      ymax = outputr(2)
    endif
    set/refc {wrktab} seq
    -delete midtmp.bdf
    zval = yval-slope*xval
    zerr = m$sqrt(yerr**2+(slope*xerr)**2)
    coef(1) = zval           ! zero point (median)
    cerr(1) = zerr
    coef(2) = slope          ! slope
    cerr(2) = 0.
    compute/table {wrktab} :{fitcol} = {coef(1)}+{coef(2)}*:{xcol}
    compute/table {wrktab} :{rescol} = :{ycol}-:{fitcol}
    if graphcrea .eq. "y" then
      statistics/table {wrktab} :{fitcol} >Null
      fmin = outputr(1)
      fmax = outputr(2)
    endif
    !
  endif
elseif meth .eq. "z" then    ! <ZERO>,ZERO
  if mode .eq. "a" then
    !
    statistics/table {wrktab} :{xcol} >Null
    xval = outputr(3)
    if xval .eq. 0. then
        write/out "      "
        mess/out 1 {PROID} "Infinite slope - try meth=[lin|slope] instead !"
      goto error
    endif
    xerr = outputr(4)
    statistics/table {wrktab} :{ycol} >Null
    yval = outputr(3)
    yerr = outputr(4)
    if graphcrea .eq. "y" then
      ymin = outputr(1)
      ymax = outputr(2)
    endif
    sval = (yval-zero)/xval
    serr = m$sqrt((yerr**2+(sval*xerr)**2)/xval**2)
    coef(1) = zero           ! zero point
    cerr(1) = 0.
    coef(2) = sval           ! slope (mean)
    cerr(2) = serr
    compute/table {wrktab} :{fitcol} = {coef(1)}+{coef(2)}*:{xcol}
    compute/table {wrktab} :{rescol} = :{ycol}-:{fitcol}
    if graphcrea .eq. "y" then
      statistics/table {wrktab} :{fitcol} >Null
      fmin = outputr(1)
      fmax = outputr(2)
    endif
    !
  elseif mode .eq. "r" then
    !
    copy/ti {wrktab} midtmp.bdf
    set/refc {wrktab} :{xcol}
    inputi = {{wrktab},TBLCONTR(5)}
    statistics/image midtmp.bdf [<,@{inputi}:>,@{inputi}] ? ? xn >Null
    xval = outputr(8)
    if xval .eq. 0. then
        write/out "      "
        mess/out 1 {PROID} "Infinite slope - try meth=[lin|slope] instead !"
      goto error
    endif
    xerr = outputr(4)
    set/refc {wrktab} :{ycol}
    inputi = {{wrktab},TBLCONTR(5)}
    statistics/image midtmp.bdf [<,@{inputi}:>,@{inputi}] ? ? xn >Null
    yval = outputr(8)
    yerr = outputr(4)
    if graphcrea .eq. "y" then
      ymin = outputr(1)
      ymax = outputr(2)
    endif
    set/refc {wrktab} seq
    -delete midtmp.bdf
    sval = (yval-zero)/xval
    serr = m$sqrt((yerr**2+(sval*xerr)**2)/xval**2)
    coef(1) = zero           ! zero point
    cerr(1) = 0.
    coef(2) = sval           ! slope (median)
    cerr(2) = serr
    compute/table {wrktab} :{fitcol} = {coef(1)}+{coef(2)}*:{xcol}
    compute/table {wrktab} :{rescol} = :{ycol}-:{fitcol}
    if graphcrea .eq. "y" then
      statistics/table {wrktab} :{fitcol} >Null
      fmin = outputr(1)
      fmax = outputr(2)
    endif
    !
  endif
endif
!
!
!          PLOTTING ORIGINAL AND FITTED VALUES
!          (/ CREATING PS-FILE) :
!          ***********************************
!
if graphcrea .eq. "y" then
  !
        mess/out 4 {PROID} "    plotting values"
  !
  if backcol .eq. "w" then
    set/graphic bcolour=8 colour=8
  else
    set/graphic bcolour=1 colour=1
  endif
  plot/table {wrktab} :{xcol} :{ycol} >Null
  if fmin .lt. ymin then
    inputd(1) = fmin
  else
    inputd(1) = ymin
  endif
  inputd(1) = inputd(1)-plrgrap(7)
  if fmax .gt. ymax then
    inputd(2) = fmax
  else
    inputd(2) = ymax
  endif
  inputd(2) = inputd(2)+plrgrap(7)
  set/graphic yaxis={inputd(1)},{inputd(2)}
  !
  if backcol .eq. "w" then
    set/graphic bcolour=8 colour=1
  else
    set/graphic bcolour=1 colour=3
  endif
  plot/table {wrktab} :{xcol} :{ycol} >Null
    yws = plrgrap(5)
    ywe = plrgrap(6)
    ycs = plrgrap(23)
    yce = plrgrap(24)
    yw1 = ywe + (1-yce)/(yce-ycs)*(ywe-yws)
    ywl = yw1 - 0.6*(yw1-ywe)
  if pname .ne. "n" then
    label/graphic "{intab}" {plrgrap(1)},{ywl} ? 0.8 1
  endif
  !
  set/graphic stype=0 ltype=1
  set/graphic colour=2
  if pcentre .ne. "n" then
    overplot/symbol 5 {xval},{yval} 1.
  endif
  set/graphic lwidth=2
  overplot/table {wrktab} :{xcol} :{fitcol} >Null
  if coef(1) .ge. plrgrap(5) .and. coef(1) .le. plrgrap(6) then
    set/graphic lwidth=3
    if plrgrap(1) .lt. 0. then
      inputr(1) = 0.
      inputr(2) = plrgrap(5)
      inputr(3) = 0.
      inputr(4) = coef(1)
      overplot/line 3 {inputr(1)},{inputr(2)} {inputr(3)},{inputr(4)}
    endif
    if plrgrap(1) .lt. 0. then
      inputr(1) = 0.
      inputr(2) = coef(1)
      inputr(3) = plrgrap(1)
      inputr(4) = coef(1)
    else
      inputr(1) = plrgrap(1)+plrgrap(4)
      inputr(2) = coef(1)
      inputr(3) = plrgrap(1)-plrgrap(4)
      inputr(4) = coef(1)
    endif
    overplot/line 3 {inputr(1)},{inputr(2)} {inputr(3)},{inputr(4)}
    set/graphic lwidth=2
  endif
  !
  if ps .eq. "y" then
        mess/out 4 {PROID} "    creating postscript file of plot"
    copy/graphic postscript
    inputi = m$index(intab,".tbl")-1
    inputc = intab(1:{inputi})//".ps"
    -rename postscript.ps {inputc}
        mess/out 3 {PROID} "      PS file of the plot is <{inputc}>"
  endif
  !
endif
!
!
!          WRITING RESULTS INTO INPUT TABLE :
!          ********************************
!
set/refc {wrktab} :{idxcol}
!
!
!       ^^^^^^^^^^^^^^^^^
!       ^  W R K T A B  ^
! - - - + - - - - - - - + - - - - - - - - - - - - - - - - - - - - - - - - - - -
!       v   I N T A B   v
!       vvvvvvvvvvvvvvvvv
!
!
set/refc {intab} :{idxcol}
copy/tt {wrktab} :{fitcol} {intab} :{fitcol}
copy/tt {wrktab} :{rescol} {intab} :{rescol}
set/refc {intab} #{refcol}
-delete {wrktab}
!
!
!          OUTPUT :
!          ******
!
if show .ne. "n" then
        set/format ,T14.7
        mess/out 3 {PROID} "    Resulting regression coefficients:"
        mess/out 3 {PROID} "      zero point: {coef(1)}  +- {cerr(1)}"
  plusone = 1+degree
  do degloop = 2 plusone
    inputi = degloop-1
        mess/out 3 {PROID} -
          "      coef. ^{inputi}  : {coef({degloop})}  +- {cerr({degloop})}"
  enddo
        set/format {FMTDEF}
endif
!
write/keyword outputi 0 all
write/keyword outputr 0. all
write/keyword outputd 0. all
plusone = 1+degree
do degloop = 1 plusone
  outputd({degloop}) = coef({degloop})  ! outputd(1) is zero point
enddo
plusone = 11+degree
do degloop = 11 plusone
  inputd = degloop-10
  outputd({degloop}) = cerr({inputd})  ! outputd(11) is std.dev. of zero point
enddo
!
!
write/desc {intab} HISTORY_UPDA/i/1/1 {updaflag}
set/format
if graphcrea .eq. "y" then
  if pclear .ne. "n" set/graphic default
endif
!
return 0 {fitcol} {rescol}
!
error:
  if m$exist(intab) .eq. 1 -
    write/desc {intab} HISTORY_UPDA/i/1/1 {updaflag}
  return 1 -99 -99
!
!
!------------------------------------------------------------------------------
