/* $Id: fors_badpix.c,v 1.16 1999/04/19 11:01:35 nrainer Exp $    */
/*+++ FORS_BADPIX +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 
 .COPYRIGHT	(C) 1998  European Southern Observatory
 
 .IDENT		fors_badpix
 
 .KEYWORDS	Determine bad pixels on an image frame (flatfield of a CCD)
 
 .USAGE		run PIPE_EXE:(IDENT)
 
 .INPUT		IFRM	   (char)  area definition table
                                   (output from fors_badpix.exe)
				  
		MASK       (char)  name of bad pixel mask (already exist)
		TABL       (char)  name of bad pixel table (already exist)
		LOW        (float) low level factor (LOW*mean)
		HGH        (float) high level factor (HGH*mean)
		MNUM       (int)   minimum number of bad pixels defining a
		                   bad column.
		ONUM       (int)   number of ports on the CCD
		START      (float) start position of world coordinates (x,y)
		STEP       (float) step sizes of world coordinates (x,y)
		IM_SEC     (int)   image section (pixel coordinates)
		MESS_LEVEL (int)   message level
		MEAN       (float) mean of IFRM on image section (IM_SEC)
		EXPFACT    (float) expand factor to get 'best' areas.

 .OUTPUT	OUTPUTI(1) (int)   number of bad pixels found.
                                   -1 in case an error had occured.
 
 .RETURN	

 .PURPOSE	Check if intensities of the input image lay outside the inter-
                val [lo,hi] (lo=LOW*mean, hi=HGH*mean), which defines a pixel
		to be bad. Finally bad pixels will be combined to bad pixel
		areas.
 
 .ALGORITHM	Find bad pixels on the selected image section (IM_SEC) of the
                input frame (IFRM). In order to be more accurate the bad pixel
		will be determined with respect to a 'local mean', i.e. means
		of a 4x4 binned input image. This mean array will finally be
		stored as a descriptor BADPIXREF.
		All detected bad pixel will be stored in the bad pixel mask
		(MASK).
		The bad pixel table should be used for bad pixel correction,
		thus the allocated bad pixels will be combined to bad pixel
		areas. Furthermore, these areas will be expanded in the way
		that the already allocated bad pixels represent the core of
		an area while the boundaries end at a 3% limit of the local
		mean. This leads to a better bad pixel correction -- not only
		the cores will be replaced by a surface interpolation but
		also the wings. This 3% limit may be changed by the parameter
		EXPFACT.
		If no local means should be used then MEAN has to be set to
		the desired MEAN of the frame, otherwise set it to a negative
		value.
		Bad columns are treated seperatly. Bad columns are defined
		as columns of minimum length (MNUM) containing bad pixels.
 
 .ENVIRON	MIDAS
    
 .LANGUAGE	C
 
 .AUTHOR	Sebastian Wolf, ESO-DMO
 
 .VERSION	1.0 1998/04/02	SW Creation
                    1998/09/21  SW low, hgh should be in the range of 0..65535
 
 .COMMENT
  OUTPUT table:	TABL -- 1. column: >0: bad column
                                    0: bad area (area defined by next 4 cols)
				   -2: bad pixel area at the edge of the CCD
			2.-5.      :XSTART,:XEND,:YSTART,:YEND
			6.-10.     :XSTARTPIX,:XENDPIX,:YSTARTPIX,:YENDPIX
-----------------------------------------------------------------------------*/

#include <string.h>
#include <stdio.h>
#include <tblsys.h>
#include <tblerr.h>
#include <tbldef.h>
#include <macrogen.h>
#include <atype.h>
#include <midas_def.h>
#include <math.h>

#define TRUE 1
#define FALSE 0
#define BELL  0x7
#define MLEN  255
#define NLEN  160
#define NCOL  9
#define NBINX 4
#define NBINY 4
#define MXAREAWIDTH 160
#define COORDSOK ((x>Imsc[0]) && (x<Imsc[2]) && (y>Imsc[1]) && (y<Imsc[3]) && (k<kmax))
#define EXPAND 1

int   NMx, NMy;
int   Cnt;                              /* counts checked edges */
int   Ifrm;                             /* file id of image frame */
int   Imsk;                             /* file id of mask frame */
int   Ptid;                             /* table id of bad pixel table */
int   Acol[4], Ajrow;                   /* columns indices, row counter */
float Sta[2], Ste[2];                   /* start and step of worl coord. */
float Low, Hgh;                         /* low and high limit (*mean) filter */
float Expandfact;                       /* expand factor to get 'best' area */
float *Mean;                            /* mean of each bin of Frm */
float *Frm, **MFrm;                     /* image frame, matrix represent. */
char *Msk, **MMsk;                      /* bad pixel mask, matrix repres. */
/* float *Msk, **MMsk; */                     /* bad pixel mask, matrix repres. */
int   *Dir, **MDir;                     /* search direction check, matrix */
int   Num;                              /* total number of pixels in img */
int   NMea;                             /* total number of pixels in Mean */
int   Cnum;                             /* number of rows in check table */
int   Mnum;                             /* max.num. of pixels def. bad col. */
int   Onum;                             /* number of outputs per CCD (ports) */
int   Mlev;                             /* message level MESS_LEVEL */
int   Iar, Nar, Nco;                    /* area counter, num. of areas, col. */
int   Bin = TRUE;                       /* bin flag */
int   Binx, Biny;                       /* bin size in x and y */
int   SI, SF;                           /* size of int, size of float */
int   **Ctab, **Ptab;                   /* check table, bad pixel table */
int   Imsc[5];                          /* image section */
int   Bit[5] = {0,1,2,4,8};             /* bit pattern */
int   **i_matrix();                     /* int matrix + elem. initialization */
int   **i_rmatrix();                    /* same as i_matrix but use realloc */
int   **imap2mat();                     /* map a matrix on a integer vector */
float **map2mat();                      /* map a matrix on a float vector */
char  **cmap2mat();                     /* map a matrix on a char vector */
char  Mes[255];                         /* message buffer */
char  Imgname[NLEN] = "combff.bdf";     /* filename of image frame */
char  Mskname[NLEN] = "badmsk.bdf";     /* filename of bad pixel mask */
char  Pixname[NLEN] = "badall.tbl";     /* filename of pixel table */
char Hdr[4][20] = {
  "[badpix]: ","INFO    [badpix]:","WARNING [badpix]:","ERROR   [badpix]:"};

char  Prgnam[]  = "badpix";
char  Version[] = "1998.04.15-SW";

char *malloc();

main()
{
  int    i, j, k;
  int    nbad, uni;
  float  row[NCOL];
  char   *ptr;

  SCSPRO("badpix");
  
  SF = sizeof(float);
  SI = sizeof(int);

  nbad = 0;
  SCKWRI("OUTPUTI",&nbad,1,1,&uni);                  /* return nbad */

  get_params();                                      /* get MIDAS parameters */
  if (Bin>0)
    binning(MFrm,Imsc[0],Imsc[2],Imsc[1],Imsc[3],NBINX,NBINY,Mean);
                                                     /* do the binning to    */
                                                     /* get float array Mean */
  nbad = findbadpixel(Low,Hgh);                      /* find bad pixel */
  searchareas(nbad);                                 /* define badpixelareas */

  SCKWRI("OUTPUTI",&nbad,1,1,&uni);                  /* return nbad */

  TCTINI(Pixname,F_TRANS,F_O_MODE,NCOL,Nar,&Ptid);   /* create MIDAS table */
  TCCINI(Ptid,D_R4_FORMAT,1,"E12.6","PIXEL",":X",&k);
  TCCINI(Ptid,D_R4_FORMAT,1,"E12.6","PIXEL",":XSTART",&k);
  TCCINI(Ptid,D_R4_FORMAT,1,"E12.6","PIXEL",":XEND",&k);
  TCCINI(Ptid,D_R4_FORMAT,1,"E12.6","PIXEL",":YSTART",&k);
  TCCINI(Ptid,D_R4_FORMAT,1,"E12.6","PIXEL",":YEND",&k);
  TCCINI(Ptid,D_R4_FORMAT,1,"E12.6","PIXEL",":XSTARTPIX",&k);
  TCCINI(Ptid,D_R4_FORMAT,1,"E12.6","PIXEL",":XENDPIX",&k);
  TCCINI(Ptid,D_R4_FORMAT,1,"E12.6","PIXEL",":YSTARTPIX",&k);
  TCCINI(Ptid,D_R4_FORMAT,1,"E12.6","PIXEL",":YENDPIX",&k);
  for (k=1,i=nbad;i>Iar;i--,k++){                    /* save bad pixel table */
    row[0] = Ptab[i][0];
    if (Ptab[i][0] > 0)                              /* bad column */
      row[0] = Sta[0] + (Ptab[i][0] - 1.0)*Ste[0];   /* :X */
    row[1] = Sta[0] + (Ptab[i][1] - 1.0)*Ste[0];     /* :XSTART */
    row[2] = Sta[0] + (Ptab[i][2] - 1.0)*Ste[0];     /* :XEND */
    row[3] = Sta[1] + (Ptab[i][3] - 1.0)*Ste[1];     /* :YSTART */
    row[4] = Sta[1] + (Ptab[i][4] - 1.0)*Ste[1];     /* :YEND */
    row[5] = Ptab[i][1] - Imsc[0];                   /* :XSTARTPIX */
    row[6] = Ptab[i][2] - Imsc[0];                   /* :XENDPIX */
    row[7] = Ptab[i][3] - Imsc[1];                   /* :YSTARTPIX */
    row[8] = Ptab[i][4] - Imsc[1];                   /* :YENDPIX */
    for (j=0;j<NCOL;j++) TCEWRR(Ptid,k,j+1,&row[j]); /* write to table */
  }
  SCDWRR(Ptid,"BADPIXREF",Mean,1,NMea,&uni);         /* write mean-ref. to */
  TCTCLO(Ptid);                                      /* descriptor BADPIXREF */

  SCSEPI();
}


int get_params()
{
  char   *ptr;
  int    i, len;
  int    uni, null;
  int    npix[2];
  float  mean;

  SCKGETC("IFRM",1,NLEN,&len,Imgname);      /* Image to determine bad pix on */
  if (Imgname[0] == '?'){
    i=-1;
    sprintf(Mes,"PROGRAM: badpix -- %s",Version);
    SCTPUT(Mes);
    SCKWRI("OUTPUTI",&i,1,1,&uni);          /* return nbad */
    SCSEPI();
    exit(0);
  }
  SCKGETC("MASK",1,NLEN,&len,Mskname);      /* mask image */
  SCKGETC("TABL",1,NLEN,&len,Pixname);      /* bad pixel table */
  SCKRDR("MEAN",1,1,&len,&mean,&uni,&null); /* lower filter limit */
  SCKRDR("EXPFACT",1,1,&len,&Expandfact,&uni,&null);   /* expand factor */
  SCKRDR("LOW",1,1,&len,&Low,&uni,&null);   /* lower filter limit */
  SCKRDR("HGH",1,1,&len,&Hgh,&uni,&null);   /* upper filter limit */
  SCKRDR("START",1,2,&len,Sta,&uni,&null);  /* start of world coordinates */
  SCKRDR("STEP",1,2,&len,Ste,&uni,&null);   /* stepsize of world coordinates */
  SCKRDI("MNUM",1,1,&len,&Mnum,&uni,&null); /* max.number of pixel def. col. */
  SCKRDI("ONUM",1,1,&len,&Onum,&uni,&null); /* number of outputs per CCD */
  SCKRDI("IMSEC",1,4,&len,Imsc,&uni,&null); /* image section */
  Imsc[0]--; Imsc[1]--; Imsc[2]--; Imsc[3]--; /* work with indices */
  SCKRDI("MESS_LEVEL",1,1,&len,&Mlev,&uni,&null);

  SCFOPN(Imgname,D_R4_FORMAT,0,F_IMA_TYPE,&Ifrm);
  SCDRDI(Ifrm,"NPIX",1,2,&len,npix,&uni,&null);
  NMx = npix[0];
  NMy = npix[1];
  Num = npix[0]*npix[1];
  Frm = (float*) malloc((size_t) Num*SF);
  if (Frm == NULL) no_memory("Frm",Num*SF);
  ptr = (char*) Frm;
  SCFGET(Ifrm,1,Num,&len,ptr);
  MFrm = map2mat(Frm,NMx,NMy,0,0);          /* matrix start at index [0,0] */


  SCFCRE(Mskname,D_I1_FORMAT,F_O_MODE,F_IMA_TYPE,Num,&Imsk);
  Msk = (char*) malloc((size_t) Num);
  if (Msk == NULL) no_memory("Msk",Num);
  ptr = (char*) Msk;
  MMsk = cmap2mat(Msk,NMx,NMy,0,0);         /* matrix start at index [0,0] */

  Dir = (int*) malloc((size_t) Num*SI);
  if (Dir == NULL) no_memory("Dir",Num*sizeof(int));
  for (i=0;i<Num;i++) Dir[i] = 0;
  MDir = imap2mat(Dir,NMx,NMy,0,0);         /* matrix start at index [0,0] */

  NMea = NBINX * NBINY;
  Mean = (float*) malloc((size_t) NMea*SF);
  if (Mean == NULL) no_memory("Mean",NMea*SF);
  if (mean>0.0){
    Bin = FALSE;
    for (i=0;i<NMea;i++) Mean[i] = mean;

  }

  Binx = Imsc[2]-Imsc[0]+1;
  Biny = Imsc[3]-Imsc[1]+1;
  Binx = Binx/NBINX;                              /* x-step size in mat */
  Biny = Biny/NBINY;                              /* y-step size in mat */

}


int binning(mat,ilo,ihi,jlo,jhi,nbinx,nbiny,bin)
/*
  Do a nbinx x nbiny binning in memory in order to get more sensitive reference
  values (means) as the flatfields are not as flat as expected.
*/
float **mat, *bin;
int   ilo, ihi, jlo, jhi, nbinx, nbiny;
{
  int i, j, i1, i2, j1, j2, k, l, m;
  int stat, cnt, dif;
  double sigma;
  double calcmean();
  char   *ptr;

  sprintf(Mes,"Determine means on a %dx%d raster as reference.",nbinx,nbiny);
  messout(Mlev,4,Prgnam,Mes);
  if (Binx % nbinx != 0) warn(2);
  if (Biny % nbiny != 0) warn(3);

  j1 = jlo;
  j2 = Biny-1 + j1;
  m  = 0;                                         /* index of bin[] */
  for (l=0;l<nbiny;l++){                          /* loop through bin[] */
    i1 = ilo;
    i2 = Binx-1 + i1;
    for (k=0;k<nbinx;k++){
      bin[m++] = (float) calcmean(mat,i1,i2,j1,j2,0.0,0.0,0,&sigma,&cnt);
      if (Mlev > 4){
	printf("REFERENCE %d,%d.: %d %d %d %d %f\n",k,l,i1,i2,j1,j2,bin[m-1]);
      }
      i1 += Binx;
      i2 += Binx;
      if (i2>ihi){        /* upper limit too high: last column */
	dif = i2-ihi;     /* shift back, so that bins have same size as */
	i1 -= dif;        /* all the others --> will lead to an overlap */
	i2 = ihi;         /* with previous column */
      }
    }
    j1 += Biny;
    j2 += Biny;
    if (j2>jhi){          /* upper limit too high: last row */
      dif = j2-jhi;       /* shift back, so that bins have same size as */
      j1 -= dif;          /* all the others --> will lead to an overlap */
      j2 = jhi;           /* with previous row */
    }
  }
}



int findbadpixel(low,hgh)
/* 
   Find bad pixel. A bad pixel is defined as a pixel having an intensity which
   lays outside the data interval: [low*mean,hgh*mean].
   In case a flatfield had been obtained using the 4-port readout this will be
   recognized and all comparisons will be made within each CCD-quadrant. That
   is necessary due to different biases of each quadrant. BUT usually the
   bad pixel determination should be done with 1-port flatfields.
*/

float low, hgh;
{
  int   nbad;
  int   ncs, nrs, uni;
  char  *ptr;

  nbad = 0;
  if (Onum == 1){
    nbad = findpix(MFrm,MMsk,Imsc[0],Imsc[2],Imsc[1],Imsc[3],low,hgh,Mean);
  }else if (Onum == 4){
    ncs = Imsc[0] + (Imsc[2] - Imsc[0] + 1) / 2;
    nrs = Imsc[1] + (Imsc[3] - Imsc[1] + 1) / 2;
    nbad += findpix(MFrm,MMsk,Imsc[0],ncs-1,Imsc[1],nrs-1,low,hgh,Mean);
    nbad += findpix(MFrm,MMsk,ncs,Imsc[2],Imsc[1],nrs-1,low,hgh,Mean);
    nbad += findpix(MFrm,MMsk,Imsc[0],ncs-1,nrs,Imsc[3],low,hgh,Mean);
    nbad += findpix(MFrm,MMsk,ncs,Imsc[2],nrs,Imsc[3],low,hgh,Mean);
  }

  if (nbad == 0) nobadpix();            /* no bad pixels found --> exit */

  sprintf(Mes,"%d bad pixels found.",nbad);
  messout(Mlev,4,Prgnam,Mes);

  ptr = (char*) Msk;
  SCFPUT(Imsk,1,Num,ptr);               /* store in MASK frame */
  SCDCOP(Ifrm,Imsk,1,ptr);                     /* copy descriptors to output */
  SCDWRR(Imsk,"BADPIXREF",Mean,1,NMea,&uni);

  SCFCLO(Imsk);
  SCFCLO(Ifrm);

  if (nbad > Num/2) warn(1);                        /* ABORT due too many */
                                                    /* bad pixels found ! */

  Cnum = nbad;
  Ctab = i_matrix(0,Cnum,0,1);                      /* 0:X,1:Y */
  if (Ctab == NULL) no_memory("Ctab",Cnum*2*SI);
  Ptab = i_matrix(0,nbad,0,NCOL-1);                 /* 0:X,1:XSTART,2:XEND, */
  if (Ptab == NULL) no_memory("Ptab",nbad*NCOL*SI); /* 3:YSTART,4:YEND */

  creatbl(MMsk,Ptab,Imsc[0],Imsc[2],Imsc[1],Imsc[3]);

  return(nbad);
}


int searchareas(nbad)
/*
  All bad pixels have been detected. Now clusters of bad pixels will be
  searched for. These clusters will defined through rectangular areas on the
  frame.
 */
int nbad;
{
  int num;
  int idir;
  int xsta, ysta;
  int area[4];

  strncpy(Mes,"Search areas ...",MLEN);
  messout(Mlev,4,Prgnam,Mes);
  sprintf(Mes,"... and check for bad columns (>%d).",Mnum);
  messout(Mlev,4,Prgnam,Mes);
  if (EXPAND){
    strncpy(Mes,"Expand areas to improve area-fixing.",MLEN);
    messout(Mlev,4,Prgnam,Mes);
  }
  Mnum = Mnum + 2;              /* as we are saving extended areas! */
  xsta = Ptab[0][1];            /* x-index(start-pixel) */
  ysta = Ptab[0][2];            /* y-index(start-pixel) */
  idir = 0;                     /* search direction index (0: all 4 dir.) */
  num = nbad;                   /* store number of bad pixels found */
  Iar = num;                    /* area table index (starts at bottom) */
  Nar = 0;                      /* area counter (= number of bad areas) */
  Nco = 0;                      /* column counter (= number of bad columns) */
  do{
    Cnt = 0;                    /* counts number of entries in check table */
    search_4(idir,xsta,ysta);   /* search starting at (xsta,ysta), fill Ctab */
    detarea(area,EXPAND);       /* determine smallest rectangular area */
    num = savearea(num,area,&xsta,&ysta); /* save area in Ptab */
  }while (num>0);

  if (Nco>0){
    strncpy(Mes,"Bad columns detected.",MLEN);
    messout(Mlev,4,Prgnam,Mes);
  }
  sprintf(Mes,"Number of bad pixel areas found: %d",Nar);
  messout(Mlev,4,Prgnam,Mes);
}



/* === subroutines doing the main job ====================================== */

int findpix(img,msk,ncl,nch,nrl,nrh,lwf,hgf,mean)
float **img, lwf, hgf, *mean;
char **msk;
int   ncl, nch, nrl, nrh;
{
  int   i, j, k, nbad;
  float low, hgh;

  nbad = 0;
  for (j=nrl;j<=nrh;j++)
    for (i=ncl;i<=nch;i++){
      k = (i-ncl)/Binx + ((j-nrl)/Biny)*NBINX;
      low = lwf * mean[k];
      hgh = hgf * mean[k];
      if (low < 0) low = 0.0;
      if (hgh > 65534.0) hgh = 65534.0;

      if ((img[j][i] < low) || (img[j][i] > hgh)){
	msk[j][i] = 1;
	nbad++;
      }else
	msk[j][i] = 0;
    }
  return(nbad);
}


int creatbl(msk,tbl,ncl,nch,nrl,nrh)
char **msk;
int   **tbl;
int   ncl, nch, nrl, nrh;
{
  int i, j, k, l;

  FILE *fp, *fopen();
  fp = (FILE*) fopen("badpixall.dat","w");

  l=0;                         /* table row counter */
  for (j=nrl;j<=nrh;j++)
    for (i=ncl;i<=nch;i++)
      if (msk[j][i] == 1){
	tbl[l][0] = 0;         /* store bad pixel coordinates */
	tbl[l][1] = i;         /* x-index (--> pixel: +1) */
	tbl[l++][2] = j;       /* y-index (--> pixel: +1) */
	fprintf(fp,"%d %d\n",i,j);
      }
  fclose(fp);
}


int search_4(idir,xsta,ysta)
/*
  The search algorithms walks in the four main directions through the frame
  starting at the first bad pixel in the list. As soon as a pixel is not
  declared as a bad pixel one boundary of a bad pixel areas is found.
 */
int idir, xsta, ysta;
{
  int   i, j, il, ih, off;
  int   xcor[4], ycor[4];
  float x, y;

  for (i=0;i<4;i++){ xcor[i]=-1; ycor[i]=-1; }

  if (idir > 0){
    il = idir;
    ih = il + 1;
  }else{
    il = 1;
    ih = 4;
  }
  
  for (i=il;i<=ih;i++) search_1(i,xsta,ysta,&xcor[i-1],&ycor[i-1]);
  
  for (i=0;i<2;i++){
    if ((xcor[i] != -1) && (ycor[i] != -1)){
      if (MMsk[ycor[i]][xcor[i]] < 5) search_4(3, xcor[i], ycor[i]);
    }
  }

  for (i=2;i<4;i++){
    if ((xcor[i] != -1) && (ycor[i] != -1)){
      if (MMsk[ycor[i]][xcor[i]] < 5) search_4(1, xcor[i], ycor[i]);
    }
  }

  if (Cnt+4 > Cnum){               /* check the size of the check-table */
    i = Cnum;
    Cnum += Cnum+4;
    Ctab = i_rmatrix(Ctab,0,Cnum,0,1,i);
  }
  for (i=0;i<4;i++){
    if (xcor[i]!=-1){
      Ctab[Cnt][0] = xcor[i];      /* save edge points of the */
      Ctab[Cnt++][1] = ycor[i];    /* detected areas (INDEX)  */
    }
  }
}


int search_1(idir,xsta,ysta,xcor,ycor)
/*
  Do the search in one of the main direction. The desired direction may be
  selected by idir (0: LEFT, 1: RIGHT, 2: DOWN, 3: UP).
  Finally the detected corner point will be returned by (xcor,ycor).
 */
int idir, xsta, ysta, *xcor, *ycor;
{
  int   x1, y1, x2, y2;
  int   nb;

  x1 = xsta;
  y1 = ysta;
  x2 = x1;
  y2 = y1;

  if (MDir[y1][x1] & Bit[idir]){
    *xcor = x1;
    *ycor = y1;
    return(0);
  }

  MMsk[y1][x1] = MMsk[y1][x1] + 1;
  MDir[y1][x1] = MDir[y1][x1] | Bit[idir];
  switch(idir){
  case 1:
    do{
      x2 = MIN(x1 + 1,Imsc[2]);
      nb = (int) MMsk[y2][x2];
      if ((nb >= 1) && ((MDir[y2][x2] & Bit[2])==0)){
	MMsk[y2][x2] = MMsk[y2][x2] + 1;
	MDir[y2][x2] = MDir[y2][x2] | Bit[1];
	x1 = x2;
      }
    }while ((nb >= 1) && ((MDir[y2][x2] & Bit[2])==0) && (x2 < Imsc[2]));
    break;
  case 2:
    do{
      x2 = MAX(x1 - 1,Imsc[0]);
      nb = (int) MMsk[y2][x2];
      if ((nb >= 1) && ((MDir[y2][x2] & Bit[1])==0)){
	MMsk[y2][x2] = MMsk[y2][x2] + 1;
	MDir[y2][x2] = MDir[y2][x2] | Bit[2];
	x1 = x2;
      }
    }while ((nb >= 1) && ((MDir[y2][x2] & Bit[1])==0) && (x2 > Imsc[0]));
    break;
  case 3:
    do{
      y2 = MIN(y1 + 1,Imsc[3]);
      nb = (int) MMsk[y2][x2];
      if ((nb >= 1) && ((MDir[y2][x2] & Bit[4])==0)){
	MMsk[y2][x2] = MMsk[y2][x2] + 1;
	MDir[y2][x2] = MDir[y2][x2] | Bit[3];
	y1 = y2;
      }
    }while ((nb >= 1) && ((MDir[y2][x2] & Bit[4])==0) && (y2 < Imsc[3]));
    break;
  case 4:
    do{
      y2 = MAX(y1 - 1,Imsc[1]);
      nb = (int) MMsk[y2][x2];
      if ((nb >= 1) && ((MDir[y2][x2] & Bit[3])==0)){
	MMsk[y2][x2] = MMsk[y2][x2] + 1;
	MDir[y2][x2] = MDir[y2][x2] | Bit[4];
	y1 = y2;
      }
    }while ((nb >= 1) && ((MDir[y2][x2] & Bit[3])==0) && (y2 > Imsc[1]));
    break;
  }                                        /* switch */
  *xcor = x1;
  *ycor = y1;
}


int detarea(area,expand)
/*
  Determine bad pixel area. Loop through the bad pixel cluster table (or check
  table) Ctab and determine RECTANGULAR area in which all pixels of the cluster
  are included.
  In case of no bad column the area will be expanded. Expansion means that
  starting from the center of the area the bounds will be shifted in all 4 main
  directions in the way that they lay on a intensity level which fullfills the
  following condition:

                  | mat[x_bound,y_bound] - MEAN |  <  Expandfac*MEAN
 */
int *area, expand;
{
  int   i, j, k, kmax, num, null;
  int   x, y, xs, ys, xcen, ycen;
  float limit;
  
  area[0] = 65535;                         /* xmin */
  area[1] = -area[0];                      /* xmax */
  area[2] = area[0];                       /* ymin */
  area[3] = -area[0];                      /* ymax */
  num = Cnt;
  for (i=0;i<num;i++){
    x = Ctab[i][0];
    y = Ctab[i][1];
    if ((x>-1) && (y>-1)){
      area[0] = MIN(area[0],x);            /* xmin */
      area[1] = MAX(area[1],x);            /* xmax */
      area[2] = MIN(area[2],y);            /* ymin */
      area[3] = MAX(area[3],y);            /* ymax */
    }
  }
                                           /* expand areas for better       */
                                           /* results when using MODIFY/PIX */ 
                                           /* afterwards in order to do a   */ 
                                           /* bad pixel correction.*/ 
  if ((expand == 1) && (area[1]>area[0])){ /* only if not a column */

    xcen = (area[0] + area[1]) / 2;        /* center coordinates of */
    ycen = (area[2] + area[3]) / 2;        /* the bad pixel area    */

    j = (xcen-Imsc[0])/Binx + ((ycen-Imsc[1])/Biny)*NBINX;
    limit = Mean[j] * Expandfact;          /* 0.03; */
    for (i=0;i<num;i++){
      switch (i)
	{
	case 0: x = area[0];  y = ycen;    xs = -1; ys = 0;  break;  /* L */
	case 1: x = area[1];  y = ycen;    xs = 1;  ys = 0;  break;  /* R */
	case 2: x = xcen;     y = area[2]; xs = 0;  ys = -1; break;  /* D */
	case 3: x = xcen;     y = area[3]; xs = 0;  ys = 1;  break;  /* U */
	}

      if (xs!=0) kmax = (MXAREAWIDTH-(area[1]-area[0]))/2; /* MXAREAWIDTH is */
      if (ys!=0) kmax = (MXAREAWIDTH-(area[3]-area[2]))/2; /* half the size  */
                                                           /* of the allowed */
                                                           /* area size used */
                                                           /* in MODIFY/PIX. */
      k=0;
      do{
	x = x + xs;
	y = y + ys;
	k++;
	if (!COORDSOK) break;
	j = (x-Imsc[0])/Binx + ((y-Imsc[1])/Biny)*NBINX;
      }while ((fabs(MFrm[y][x] - Mean[j]) >= limit) && COORDSOK);

      x = MIN( MAX(Imsc[0],x) , Imsc[2] );   /* check if coordinates still */
      y = MIN( MAX(Imsc[1],y) , Imsc[3] );   /* lay inside the image section */

      switch (i)
	{
	case 0: area[0] = x;  break;  /* L */
	case 1: area[1] = x;  break;  /* R */
	case 2: area[2] = y;  break;  /* D */
	case 3: area[3] = y;  break;  /* U */
	}      
    }
  }
}


int savearea(num,area,xsta,ysta)
int   *area, *xsta, *ysta, num;
{
  int    i, nnum, null, chg;
  int    x, y;
  
  nnum = num - 1;
  for (i=num-1;i>=0;i--){
    x = Ptab[i][1];
    y = Ptab[i][2];
    if ((x>=area[0]) && (x<=area[1]) && (y>=area[2]) && (y<=area[3])){
      if (i != nnum){
	Ptab[i][1] = Ptab[nnum][1];  /* shift the already detected bad */
	Ptab[i][2] = Ptab[nnum][2];  /* pixel to the bottom of the table: */
	Ptab[nnum][1] = x;           /* -->  exchange table rows */
	Ptab[nnum][2] = y;
      }
      nnum--;
    }
  }
  nnum++;
  if (nnum>0){              /* get next starting pixel to examine */
    *xsta = Ptab[0][1];     /* .. take the first one in the table */
    *ysta = Ptab[0][2];
  }

  /* In the following correction of the bad pixel table matrix we will
     consider both:
     (1) we need pixels starting at 1 instead of 0 for the index which
         are actually stored in area[].
     (2) we need "expanded areas"
         == smallest area surrounding a bad pixel area
   */
  Ptab[Iar][1] = area[0];     /* save area on bottom of the table */
  Ptab[Iar][2] = area[1] + 2; /* may be overwritten, as these pixels */
  Ptab[Iar][3] = area[2];     /* are already detected ! */
  Ptab[Iar][4] = area[3] + 2; /* add 1 as PIXELS have to be saved */
  if (((Ptab[Iar][4]-Ptab[Iar][3]) > Mnum)
   && ((Ptab[Iar][2]-Ptab[Iar][1]) == 2)){
    Ptab[Iar][0] = Ptab[Iar][1]+1; /* bad pixel column add 1 (s.above) */
    Nco++;
  }

  chg = 0;
  if (Ptab[Iar][1] == 0){     /* check coordinates at the edge of the CCD */
    Ptab[Iar][1] = 1;         /* --> correct if necessary */
    chg = 1;
  }
  if (Ptab[Iar][2] > Imsc[2]+1){
    Ptab[Iar][2] = Imsc[2]+1;
    chg = 1;
  }
  if (Ptab[Iar][3] == 0){
    Ptab[Iar][3] = 1;
    chg = 1;
  }
  if (Ptab[Iar][4] > Imsc[3]+1){
    Ptab[Iar][4] = Imsc[3]+1;
    chg = 1;
  }
  if ((chg) && (Ptab[Iar][0]==0))
    Ptab[Iar][0] = -2;        /* marked as an area at the edge of the CCD */
  Iar--;                      /* area table index */
  Nar++;                      /* area counter */

  return(nnum);
}


/* === utilities =========================================================== */
int no_memory(svar,num)
int   num;
char  *svar;
{
  int nbad, uni;

  nbad = -1;
  sprintf(Mes,"Cannot allocate memory for %s [%d]!\n",svar,num);
  messout(Mlev,1,Prgnam,Mes);
  SCKWRI("OUTPUTI",&nbad,1,1,&uni);                  /* return nbad */
  SCETER(-1,"*** memory-error");
  SCSEPI();
  exit(1);
}


int nobadpix()
{
  int nbad, uni;

  nbad = 0;
  strncpy(Mes,"No bad pixels found!",MLEN);
  messout(Mlev,1,Prgnam,Mes);
  SCKWRI("OUTPUTI",&nbad,1,1,&uni);                  /* return nbad */
  SCSEPI();
  exit(0);
}


int warn(i)
int i;
{
  int uni;

  switch(i){
  case 1:
    i = -1;
    strncpy(Mes,"More than half of all pixels are bad !",MLEN);
    SCKWRI("OUTPUTI",&i,1,1,&uni);                  /* return nbad */
    SCSEPI();
    exit(0);
    break;
  case 2:
    strncpy(Mes,"Number of bins in X causes overlap at last column!",MLEN);
    break;
  case 3:
    strncpy(Mes,"Number of bins in Y causes overlap at last row!",MLEN);
    break;
  default:
    printf("warning code not found!\n");
    return(0);
  }
  messout(Mlev,2,Prgnam,Mes);
}
