/* $Id: fors_bindif.c,v 1.12 1999/01/27 17:46:47 swolf Exp $ */
/*+++ FORS_BINDIF +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 
 .COPYRIGHT	(C) 1998  European Southern Observatory
 
 .IDENT		fors_bindif.exe
 
 .KEYWORDS	Do the binning on a single or a (scaled) difference frame
 
 .USAGE		run PIPE_EXE:(IDENT)
 
 .INPUT		IN_A	   (char)  1. Image
 		IN_B	   (char)  2. Image
 		OUT_A	   (char)  binned image
 		OUT_B	   (char)  sigma image
		INPUTI(1)  (int)   number of CCD-ports
		IMSEC      (int)   image section (pixel coordinates)
		NBINX      (int)   binning in x-direction
		NBINY      (int)   binning in y-direction
		KAPPA      (int)   kappa of kappa-sigma-clipping
		SCALA      (float) scale of Image 1
		SCALB      (float) scale of Image 2
 		MESS_LEVEL (int)   message level
 .OUTPUT	OUTPUTI(1) (int)   number of outliers found
                OUTPUTD    (int)   (1..4): mean of each quadrant

 .RETURN	Q1:  0: successful return
 		    -1: a successful return failed
 
 .PURPOSE	Do a binning of the single input frame IN_A or of the scaled
                difference frame (IN_A - IN_B/SCALB)/SCALA
 
 .ALGORITHM	Build the diffrence frame of (Image1 - Image2/SCALB) / SCALA
                and do the binning. The difference frame will only be computed
		as soon as both input frames are specified.
		Beside the binned frame (OUT_A) the sigmas of each bin will be
		stored in an extra frame OUT_B.

                Recognizes 4-port readout and does the binning without leaving
		the bounds of each port!

		In case of bin sizes which do not fit into the dimensions of
		the input frame the last row and/or column will have an overlap
		with its predecessor in order to get identical number of pixels
		within the averaging process. A warning will be put on screen.
 
 .ENVIRON	MIDAS
    
 .LANGUAGE	C
 
 .AUTHOR	Sebastian Wolf, ESO-DMO
 
 .VERSION	1.0 1998/09/04	Creation
 
 .COMMENT       Set KAPPA to negative value in order to switch the kappa-sigma-
		clipping OFF

-----------------------------------------------------------------------------*/


#include <string.h>
#include <stdio.h>
#include <tblsys.h>
#include <tblerr.h>
#include <tbldef.h>
#include <macrogen.h>
#include <atype.h>
#include <midas_def.h>
#include <math.h>

#define TRUE 1
#define FALSE 0
#define BELL  0x7
#define NLEN  160
#define MLEN  255

int   SI, SF;                           /* size of int, size of float */
int   Mlev;                             /* message level MESS_LEVEL */
int   Imsc[4];                          /* image section */

int   Stat=0;                           /* status of program 0=ok, -1 error */
int   Id1;                              /* file id of IMAGE frame */
int   Iid;                              /* file id of IMAGE frame */
int   Nid;                              /* file id of MEDIAN frame */
int   Mid;                              /* file id of MEAN frame */
int   Sid;                              /* file id of SIGMA frame */
int   NIx, NIy;                         /* dimensions of INPUT frame */
int   NNx, NNy;                         /* dimensions of OUTPUT frame */
int   NMx, NMy;                         /* dimensions of OUTPUT frame */
int   NSx, NSy;                         /* dimensions of OUTPUT frame */
int   NI;                               /* total number of pixels in Ima */
int   NN;                               /* total number of pixels in Med */
int   NM;                               /* total number of pixels in Mea */
int   NS;                               /* total number of pixels in Sig */
int   Num;                              /* total number of used pixels */
int   NOut;                             /* total number of outlayers */
int   Idx, Idy;                         /* bin-size in x and y */
int   Nbin[2];                          /* number of bins in x and y */
int   Nport;                            /* number of ports used on CCD */
int   Kappa;                            /* kappa-sigma clipping factor */
float Sta[2], Ste[2];                   /* start and step of worl coord. */
float Low, Hgh;                         /* low and high limit (*mean) filter */
float *Ima, **MIma;                     /* image frame, matrix represent. */
float *Im1, **MIm1;                     /* image frame, matrix represent. */
float *Mea, **MMea;                     /* MEAN frame, matrix repres. */
float *Med, **MMed;                     /* MEDIAN frame, matrix repres. */
float *Sig, **MSig;                     /* SIGMA frame, matrix repres. */

float **map2mat();                      /* map a matrix on a float vector */

double Sumq[4];                         /* sum over all pix. of each quadr. */
double Sum, Ceni, Cenj, Vari, Varj;     /* moments of intensity distribution */

char  Mes[MLEN];                        /* message buffer */
char  Ima1name[NLEN] = "image.bdf";     /* filename of image frame */
char  Imagname[NLEN] = "image.bdf";     /* filename of image frame */
char  Mediname[NLEN] = "medi.bdf";      /* filename of medi image */
char  Meanname[NLEN] = "mean.bdf";      /* filename of mean image */
char  Sigmname[NLEN] = "sigma.bdf";     /* filename of sigma image */

char  Prgnam[]  = "bindif.c";
char  Version[] = "1999.01.18-SW";

char *malloc();

main()

{
  int uni;
  int i, j, k;


  SCSPRO("bindif");
  
  SF = sizeof(float);
  SI = sizeof(int);

  get_params();                                      /* get MIDAS parameters */
  calcmean_and_sig(Imsc[0],Imsc[2],Imsc[1],Imsc[3],Nport);

  free_mem();                                /* free allocated memory */
  Stat = NOut;
  SCKWRI("OUTPUTI",&Stat,1,1,&uni);            /* return number of outlayers */
  SCKWRD("OUTPUTD",Sumq,1,4,&uni);             /* return */
  SCSEPI();
}


/* === interface to MIDAS ================================================== */

int get_params()
{
  char   *ptr;
  int    i, j, len;
  int    uni, null;
  int    nx, ny;
  int    npix[2];
  float  scala, scalb;

  SCKGETC("IN_A",1,NLEN,&len,Ima1name);           /* get name of image frame */
  if (Imagname[0] == '?'){
    i=0;
    sprintf(Mes,"PROGRAM: bindif -- %s",Version);
    SCTPUT(Mes);
    SCSEPI();
    exit(0);
  }
  SCKGETC("IN_B",1,NLEN,&len,Imagname);           /* get name of image frame */
  SCKGETC("OUT_A",1,NLEN,&len,Meanname);          /* mean image */
  SCKGETC("OUT_B",1,NLEN,&len,Sigmname);          /* sigma image */
  SCKRDR("SCALA",1,1,&len,&scala,&uni,&null);     /* mean of first image */
  SCKRDR("SCALB",1,1,&len,&scalb,&uni,&null);     /* scale of second image */
  SCKRDI("KAPPA",1,1,&len,&Kappa,&uni,&null);     /* kappa sigma clipping */
  SCKRDI("NBINX",1,1,&len,&Nbin[0],&uni,&null);   /* bin to nbin[0] in x-dir */
  SCKRDI("NBINY",1,1,&len,&Nbin[1],&uni,&null);   /* bin to nbin[0] in x-dir */
  SCKRDI("MESS_LEVEL",1,1,&len,&Mlev,&uni,&null); /* message level */
  SCKRDI("INPUTI",1,1,&len,&Nport,&uni,&null);    /* number of ports */
  SCKRDI("IMSEC",1,4,&len,Imsc,&uni,&null);       /* image section */
  nx = Imsc[2] - Imsc[0] + 1;
  ny = Imsc[3] - Imsc[1] + 1;

  Idx = nx/Nbin[0];                               /* x-steps in MIma */
  Idy = ny/Nbin[1];                               /* y-steps in MIma */
  if (nx % Nbin[0] != 0) warn(1);
  if (ny % Nbin[1] != 0) warn(2);
  Num = nx * ny;                                  /* total num. (used pixel) */

  SCFOPN(Ima1name,D_R4_FORMAT,0,F_IMA_TYPE,&Id1);
  SCDRDI(Id1,"NPIX",1,2,&len,npix,&uni,&null);
  NIx = npix[0];                               /* x-size of complete frame */
  NIy = npix[1];                               /* y-size of complete frame */
  NI  = npix[0]*npix[1];                       /* total number of pixels */
  Im1 = (float*) malloc((size_t) NI*SF);
  if (Im1 == NULL) no_memory("Im1",NI*SF);
  ptr = (char*) Im1;
  SCFGET(Id1,1,NI,&len,ptr);                   /* read the whole frame */
  SCFCLO(Id1);
  MIm1 = map2mat(Im1,NIx,NIy,1,1);             /* map data for matrix access */

  NIx = npix[0];                               /* x-size of complete frame */
  NIy = npix[1];                               /* y-size of complete frame */
  NI  = npix[0]*npix[1];                       /* total number of pixels */
  Ima = (float*) malloc((size_t) NI*SF);
  if (Ima == NULL) no_memory("Ima",NI*SF);
  MIma = map2mat(Ima,NIx,NIy,1,1);             /* map data for matrix access */
  if (Imagname[0] != '+'){
    SCFOPN(Imagname,D_R4_FORMAT,0,F_IMA_TYPE,&Iid);
    SCDRDI(Iid,"NPIX",1,2,&len,npix,&uni,&null);
    ptr = (char*) Ima;
    SCFGET(Iid,1,NI,&len,ptr);                 /* read the whole frame */
    SCFCLO(Iid);

    messout(Mlev,4,Prgnam,"    determine normalized difference image:");
    sprintf(Mes,"    (%s/m - %s/k)",Ima1name,Imagname);
    messout(Mlev,4,Prgnam,Mes);
    sprintf(Mes,"    m=%f\tk=%f",scala,scalb);
    messout(Mlev,4,Prgnam,Mes);

    for (j=1;j<=NIy;j++)
      for (i=1;i<=NIx;i++)
	MIm1[j][i] = (MIm1[j][i]/scala - MIma[j][i]/scalb);
  }

  SCFOPN(Mediname,D_R4_FORMAT,0,F_IMA_TYPE,&Nid);
  SCDRDI(Nid,"NPIX",1,2,&len,npix,&uni,&null);
  NNx = npix[0];                               /* x-size of complete frame */
  NNy = npix[1];                               /* y-size of complete frame */
  NN  = npix[0]*npix[1];                       /* total number of pixels */
  Med = (float*) malloc((size_t) NN*SF);
  if (Med == NULL) no_memory("Med",NN*SF);
  ptr = (char*) Med;
  SCFGET(Nid,1,NN,&len,ptr);
  MMed = map2mat(Med,NNx,NNy,1,1);             /* matrix representation */

  SCFOPN(Meanname,D_R4_FORMAT,0,F_IMA_TYPE,&Mid);
  SCDRDI(Mid,"NPIX",1,2,&len,npix,&uni,&null);
  NMx = npix[0];                               /* x-size of complete frame */
  NMy = npix[1];                               /* y-size of complete frame */
  NM  = npix[0]*npix[1];                       /* total number of pixels */
  Mea = (float*) malloc((size_t) NM*SF);
  if (Mea == NULL) no_memory("Mea",NM*SF);
  ptr = (char*) Mea;
  SCFGET(Mid,1,NM,&len,ptr);
  MMea = map2mat(Mea,NMx,NMy,1,1);             /* matrix representation */

  SCFOPN(Sigmname,D_R4_FORMAT,0,F_IMA_TYPE,&Sid);
  SCDRDI(Sid,"NPIX",1,2,&len,npix,&uni,&null);
  NSx = npix[0];                               /* x-size of complete frame */
  NSy = npix[1];                               /* y-size of complete frame */
  NS  = npix[0]*npix[1];                       /* total number of pixels */
  if ((NSx != NMx) || (NSy != NMy)) myerror(1);
  Sig = (float*) malloc((size_t) NS*SF);
  if (Sig == NULL) no_memory("Sig",NS*SF);
  ptr = (char*) Sig;
  SCFGET(Sid,1,NS,&len,ptr);
  MSig = map2mat(Sig,NSx,NSy,1,1);             /* matrix representation */




  messout(Mlev,4,Prgnam,"    calculate MEANS and SIGMAS of difference image");
  if (Kappa<0)
    sprintf(Mes,"    in %dx%d bins",Nbin[0],Nbin[1]);
  else
    sprintf(Mes,"    in %dx%d bins ignoring outliers",Nbin[0],Nbin[1]);

  messout(Mlev,4,Prgnam,Mes);
  sprintf(Mes,"      bin-size:             xdim=%d, ydim=%d",Idx,Idy);
  messout(Mlev,4,Prgnam,Mes);
  sprintf(Mes,"      use image area:       [%d,%d]:[%d,%d]"
	  ,Imsc[0],Imsc[1],Imsc[2],Imsc[3]);
  messout(Mlev,4,Prgnam,Mes);
  if (Kappa<0)
    strncpy(Mes,"      kappa-sigma clipping: OFF",MLEN);
  else
    sprintf(Mes,"      kappa-sigma clipping: kappa=%d",Kappa);
  messout(Mlev,4,Prgnam,Mes);
}


/* === subroutines doing the main job ====================================== */

int calcmeans(ilo,ihi,jlo,jhi,klo,khi,llo,lhi)
int ilo,ihi,jlo,jhi;
int klo,khi,llo,lhi;
{
  int i, j, i1, i2, j1, j2, k, l;
  int stat, cnt, dif;
  float median, sigmed, calcmedian();
  float *ima, *sig;
  double sigma;
  double calcmean();
  char   *ptr;

  ima = Im1;
  sig = Ima;
  median = 0.0;
  sigmed = 0.0;

  j1 = jlo;
  j2 = Idy-1 + j1;
  NOut = 0;
  for (l=llo;l<=lhi;l++){
    i1 = ilo;
    i2 = Idx-1 + i1;
    for (k=klo;k<=khi;k++){
      if (Kappa>=0) {
	MMed[l][k] = calcmedian(ima,sig,i1,i2,j1,j2,NIx,&sigmed);
	median = MMed[l][k];
      }
      MMea[l][k] =
	(float) calcmean(MIm1,i1,i2,j1,j2,median,sigmed,Kappa,&sigma,&cnt);
      MSig[l][k] = (float) sigma;
      if (Mlev > 4){
	printf("%d,%d.: %d %d %d %d\n",k,l,i1,i2,j1,j2);
	printf("       median=%f sigmed=%f\n",median,sigmed);
	printf("       nout=%d\n",cnt);
      }
      NOut += cnt;
      Sum  += MMea[l][k];
      Ceni += k * MMea[l][k];
      Cenj += l * MMea[l][k];
      i1 += Idx;
      i2 += Idx;
      if (i2>ihi){        /* upper limit too high: last column */
	dif = i2-ihi;     /* shift back, so that bins have same size as */
	i1 -= dif;        /* all the others --> will lead to an overlap */
	i2 = ihi;         /* with previous column */
      }
    }
    j1 += Idy;
    j2 += Idy;
    if (j2>jhi){          /* upper limit too high: last row */
      dif = j2-jhi;       /* shift back, so that bins have same size as */
      j1 -= dif;          /* all the others --> will lead to an overlap */
      j2 = jhi;           /* with previous row */
    }
  }
}


int calcmean_and_sig(ilo,ihi,jlo,jhi,nport)
int ilo,ihi,jlo,jhi,nport;
{
  int i1, i2, j1, j2;
  int k1, k2, l1, l2;
  int ni2, nj2, nq;
  int stat;
  char *ptr;


  Sum  = 0.0;
  Ceni = 0.0;
  Cenj = 0.0;
  Vari = 0.0;
  Varj = 0.0;

  if (nport == 4){
    strncpy(Mes,"      4-port readout",MLEN);
    messout(Mlev,4,Prgnam,Mes);
  }
  ni2 = (ihi-ilo+1)/2;
  nj2 = (jhi-jlo+1)/2;
  nq  = NMx * NMy / 4;                  /* number of pixels per quadrant */
  i1 = ilo; i2 = ilo + ni2 - 1;
  j1 = jlo; j2 = jlo + nj2 - 1;
  k1 = 1; k2 = NMx/2;
  l1 = 1; l2 = NMy/2;
  calcmeans(i1,i2,j1,j2,k1,k2,l1,l2);          /* lower left quadrant */
  Sumq[0] = Sum;

  i1 = i2+1; i2 = ihi;
  k1 = k2+1; k2 = NMx;
  calcmeans(i1,i2,j1,j2,k1,k2,l1,l2);          /* lower right quadrant */
  Sumq[1] = Sum - Sumq[0];

  i1 = ilo; i2 = ilo + ni2 - 1;
  j1 = j2+1; j2 = jhi;
  k1 = 1; k2 = NMx/2;
  l1 = l2+1; l2 = NMy;
  calcmeans(i1,i2,j1,j2,k1,k2,l1,l2);          /* upper left quadrant */
  Sumq[2] = Sum - (Sumq[0] + Sumq[1]);

  i1 = i2+1; i2 = ihi;
  k1 = k2+1; k2 = NMx;
  calcmeans(i1,i2,j1,j2,k1,k2,l1,l2);          /* upper right quadrant */
  Sumq[3] = Sum - (Sumq[0] + Sumq[1] + Sumq[2]);

  Sumq[0] = Sumq[0]/nq;                           /* mean of each quadrant */
  Sumq[1] = Sumq[1]/nq;                           /* will be stored in */
  Sumq[2] = Sumq[2]/nq;                           /* global MIDAS keyword */
  Sumq[3] = Sumq[3]/nq;                           /* OUTPUTR */

  if (Sum != 0.0){                                /* Determine moments of */
    Ceni = Ceni / Sum;                            /* the intensity distr. */
    Cenj = Cenj / Sum;
  }
  for (j1=1;j1<NMy+1;j1++){                       /* ..and now the variances */
    for (i1=1;i1<NMx+1;i1++){
      Sum  += MMea[j1][i1];
      Vari += (i1 - Ceni) * (i1 - Ceni) * MMea[j1][i1];
      Varj += (j1 - Cenj) * (j1 - Cenj) * MMea[j1][i1];
    }
  }
  Vari = Vari / Sum;
  Varj = Varj / Sum;

  ptr = (char*) Mea;                              /* save the data on disk */
  stat=SCFPUT(Mid,1,NM,ptr);
  SCFCLO(Mid);
  ptr = (char*) Med;
  stat=SCFPUT(Nid,1,NN,ptr);
  SCFCLO(Nid);
  ptr = (char*) Sig;
  stat=SCFPUT(Sid,1,NS,ptr);
  SCFCLO(Sid);


  messout(Mlev,4,Prgnam,"    MEANS and SIGMAS are calculated and stored in:");
  sprintf(Mes,"    %s, %s",Meanname,Sigmname);
  messout(Mlev,4,Prgnam,Mes);

  sprintf(Mes,"    Statistic of %s:",Meanname);
  messout(Mlev,4,Prgnam,Mes);
  sprintf(Mes,"    Centre:   (%f,%f) [pixel]",Ceni,Cenj);
  messout(Mlev,4,Prgnam,Mes);
  sprintf(Mes,"    Variance: (%f,%f) [pixel]",Vari,Varj);
  messout(Mlev,4,Prgnam,Mes);
  sprintf(Mes,"    Total Intensity: %f",Sum);
  messout(Mlev,4,Prgnam,Mes);
  sprintf(Mes,"    Total number of outliers: %d (%6.2f %%)"
          ,NOut,100.0*NOut/Num);
  messout(Mlev,4,Prgnam,Mes);
}


int calcmean_and_sigOLD(ilo,ihi,jlo,jhi,nport)
int ilo,ihi,jlo,jhi,nport;
{
  int i1, i2, j1, j2;
  int k1, k2, l1, l2;
  int ni2, nj2, nq;
  int stat;
  char *ptr;


  Sum  = 0.0;
  Ceni = 0.0;
  Cenj = 0.0;
  Vari = 0.0;
  Varj = 0.0;

  switch (nport)
    {
    case 1:
      calcmeans(ilo,ihi,jlo,jhi,1,NMx,1,NMy);
      Sumq[0] = Sum;
      Sumq[1] = Sumq[0];
      Sumq[2] = Sumq[0];
      Sumq[3] = Sumq[0];
      nq = NMx * NMy;                       /* number of pixels per quadrant */
      break;
    case 4:
      strncpy(Mes,"  4-port readout",MLEN);
      messout(Mlev,4,Prgnam,Mes);
      ni2 = (ihi-ilo+1)/2;
      nj2 = (jhi-jlo+1)/2;
      nq  = NMx * NMy / 4;                  /* number of pixels per quadrant */
      i1 = ilo; i2 = ilo + ni2 - 1;
      j1 = jlo; j2 = jlo + nj2 - 1;
      k1 = 1; k2 = NMx/2;
      l1 = 1; l2 = NMy/2;
      calcmeans(i1,i2,j1,j2,k1,k2,l1,l2);          /* lower left quadrant */
      Sumq[0] = Sum;

      i1 = i2+1; i2 = ihi;
      k1 = k2+1; k2 = NMx;
      calcmeans(i1,i2,j1,j2,k1,k2,l1,l2);          /* lower right quadrant */
      Sumq[1] = Sum - Sumq[0];

      i1 = ilo; i2 = ilo + ni2 - 1;
      j1 = j2+1; j2 = jhi;
      k1 = 1; k2 = NMx/2;
      l1 = l2+1; l2 = NMy;
      calcmeans(i1,i2,j1,j2,k1,k2,l1,l2);          /* upper left quadrant */
      Sumq[2] = Sum - (Sumq[0] + Sumq[1]);

      i1 = i2+1; i2 = ihi;
      k1 = k2+1; k2 = NMx;
      calcmeans(i1,i2,j1,j2,k1,k2,l1,l2);          /* upper right quadrant */
      Sumq[3] = Sum - (Sumq[0] + Sumq[1] + Sumq[2]);
      break;
    default:
      myerror(2);
    }
  Sumq[0] = Sumq[0]/nq;                           /* mean of each quadrant */
  Sumq[1] = Sumq[1]/nq;                           /* will be stored in */
  Sumq[2] = Sumq[2]/nq;                           /* global MIDAS keyword */
  Sumq[3] = Sumq[3]/nq;                           /* OUTPUTR */

  if (Sum != 0.0){                                /* Determine moments of */
    Ceni = Ceni / Sum;                            /* the intensity distr. */
    Cenj = Cenj / Sum;
  }
  for (j1=1;j1<NMy+1;j1++){                       /* ..and now the variances */
    for (i1=1;i1<NMx+1;i1++){
      Sum  += MMea[j1][i1];
      Vari += (i1 - Ceni) * (i1 - Ceni) * MMea[j1][i1];
      Varj += (j1 - Cenj) * (j1 - Cenj) * MMea[j1][i1];
    }
  }
  Vari = Vari / Sum;
  Varj = Varj / Sum;

  ptr = (char*) Mea;                              /* save the data on disk */
  stat=SCFPUT(Mid,1,NM,ptr);
  SCFCLO(Mid);
  ptr = (char*) Sig;
  stat=SCFPUT(Sid,1,NS,ptr);
  SCFCLO(Sid);


  sprintf(Mes,"MEANS and SIGMAS are calculated and stored in:\n%s, %s",
	  Meanname,Sigmname);
  messout(Mlev,4,Prgnam,Mes);
  sprintf(Mes,"Statistic of %s:\nCentre:   (%f,%f) [pixel]\nVariance: (%f,%f) [pixel]\nTotal Intensity: %f\nTotal number of outlayers: %d (%6.2f %%)",
	  Meanname,Ceni,Cenj,Vari,Varj,Sum,NOut,100.0*NOut/Num);
  messout(Mlev,4,Prgnam,Mes);
}


/* === utilities =========================================================== */
int free_mem()
{
  free((char*) (MIm1+1));
  free((char*) (MIma+1));
  free((char*) (MMea+1));
  free((char*) (MSig+1));
  free((char*) Im1);
  free((char*) Ima);
  free((char*) Mea);
  free((char*) Sig);

}


int no_memory(svar,num)
int   num;
char  *svar;
{
  int nbad, uni;

  nbad = 0;
  sprintf(Mes,"Cannot allocate memory for %s [%d]!\n",svar,num);
  messout(Mlev,1,Prgnam,Mes);

  Stat = -1;
  SCKWRI("OUTPUTI",&Stat,1,1,&uni);                  /* return status */
  SCETER(-1,"*** memory-error");
  SCSEPI();
  exit(1);
}


int warn(i)
int i;
{
  switch(i){
  case 1:
    strncpy(Mes,"    No. of bins in X causes overlap at last col. !",MLEN);
    break;
  case 2:
    strncpy(Mes,"    No. of bins in Y causes overlap at last row !",MLEN);
    break;
  default:
    printf("warning code not found!\n");
    return(0);
  }
  messout(Mlev,2,Prgnam,Mes);
}


int myerror(i)
int i;
{
  switch(i){
  case 1:
    strncpy(Mes,"Incompatible dimensions for output frames !",MLEN);
    messout(Mlev,1,Prgnam,Mes);
    break;
  case 2:
    sprintf(Mes,"Number of ports: %d is NOT supported!",Nport);
    messout(Mlev,1,Prgnam,Mes);
    break;
  default:
    printf("error code not found!\n");
  }
  i = -i - 1;
  Stat = -1;
  SCKWRI("OUTPUTI",&Stat,1,1,&i);                  /* return status */
  SCETER(i,"*** bindif-error");
  SCSEPI();
  exit(1);
}
