/* $Id: fors_check_overl.c,v 1.8 1998/12/10 14:51:16 swolf Exp $    */
/*+++ FORS_CHECK_OVERL ++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 
 .COPYRIGHT	(C) 1998  European Southern Observatory
 
 .IDENT		fors_check_overl
 
 .KEYWORDS	Check bad pixel table for bad-pixel-area overlaps
 
 .USAGE		run PIPE_EXE:(IDENT)
 
 .INPUT		IN_A	   (char) area definition table
                                  (output from fors_badpix.exe)

		STEPX      (int)  step size of image in x-direction
		NDIMX      (int)  x-dimension of image (STEPX * NPIX(1))
		MESS_LEVEL (int)  message level

 .OUTPUT	OUT_A	(char) area definition table without any
                               area-overlaps
 
 .RETURN	

 .PURPOSE	Check if areas defined through input table show overlaps.
                In case of overlaps the smallest union of the overlapping
		areas will be stored in the output table.
 
 .ALGORITHM	For more details see procedure headers, below.
 
 .ENVIRON	MIDAS
    
 .LANGUAGE	C
 
 .AUTHOR	Sebastian Wolf, ESO-DMO
 
 .VERSION	1.0 1998/04/02	SW Creation
                    1998/09/28  SW bug fixed: bad column detection.
 
 .COMMENT
  INPUT table:  Atid -- readonly except last column (:VALUE1) will be updated:
                        set to 0.0 if an area has an overlap to another.
  OUTPUT table:	Btid -- 1. column: >0: bad column
                                    0: bad area (area defined by next 4 cols)
				   -1: single bad pixel
				   -2: bad pixel area at the edge of the CCD
				   -3: single bad pixel at the edge of the CCD

-----------------------------------------------------------------------------*/
#include <string.h>
#include <stdio.h>
#include <midas_def.h>
#include <math.h>

#define TRUE 1
#define FALSE 0
#define BELL  0x7
#define NLEN  80
#define VALCOL 10                       /* value table column */
#define MIN(A,B) ((A)<(B)?(A):(B))      /* minimum function */
#define MAX(A,B) ((A)>(B)?(A):(B))      /* maximum function */
#define SNGLPOINT ( (*xend-*xsta==2.0) && (*yend-*ysta==2.0) )

static int   Mlev;
static int   Ncol;                      /* number of bad columns found */
static int   Atid, Btid;
static int   Anrow, Ancol;
static int   Bnrow, Bncol, Bjrow;
static int   **Imap;
static int   **Omap;
static int   Stepx, Npixx;
static float Colwd;
static char  Mes[255];
static char  atnam[NLEN] = "wrk.tbl";
static char  btnam[NLEN] = "badpix.tbl";
static float *Bcol;
static int   *BcolSaved;

char  Prgnam[]  = "check_overl";
static char  Version[]="1998.09.28-SW";

main()
{
  int    err;
  int    len;
  int    nsort, nallc, nallr;
  int    uni, null;
  float  *vector();
   

  SCSPRO("check_overl");

  SCKGETC("IN_A",1,NLEN,&len,atnam);
  SCKGETC("OUT_A",1,NLEN,&len,btnam);
  SCKRDI("STEPX",1,1,&len,&Stepx,&uni,&null);
  SCKRDI("NDIMX",1,1,&len,&Npixx,&uni,&null);
  SCKRDI("MESS_LEVEL",1,1,&len,&Mlev,&uni,&null);
  Colwd = 2.0*Stepx - 1.0;                /* column width in an area def. */
  Bcol = vector(0,Npixx-1);               /* bad column vector */

  TCTOPN(atnam,F_IO_MODE,&Atid);          /* area table */
  TCTOPN(btnam,F_IO_MODE,&Btid);          /* bad pixel table [updated] */

  TCIGET(Atid,&Ancol,&Anrow,&nsort,&nallc,&nallr);
  TCIGET(Btid,&Bncol,&Bnrow,&nsort,&nallc,&nallr);
  Bjrow = 1;

  write_overl();
  get_overl();
  save_nooverl();

  TCTCLO(Atid);
  TCTCLO(Btid);
  SCSEPI();
}


int write_overl()
/*
  METHOD:
  . Produce an Input-Map (Imap) in which the bad pixel areas are set to their
    rownumber of the area-definition table while the rest is set to 0. Bad
    pixel columns will be ignored, as they nevertheless will be handled in
    an extra way during badpixel correction.
    For example, Imap will hold 5 for each pixel in case of area 5 of the
    input table.
  . Create an Overlap-Map (Omap) of dimension (Anrow x Anrow) while Anrow
    defines the number of areas stored in the area-definition table. Each
    overlap of two areas are indicated as 1 in this quadratic matrix.
    For example, area 5 and area 9 show an overlap then Omap[5][9] = 1.
 */
{
  int   jrow, ix, iy, null;
  int   *i_vector();
  int   **i_matrix();
  int   iMxsta, iMxend, iMysta, iMyend;
  float Mxsta, Mxend, Mysta, Myend;
  float *xsta, *xend, *ysta, *yend;
  float *colu;
  float *vector();

  xsta = vector(1,Anrow);
  xend = vector(1,Anrow);
  ysta = vector(1,Anrow);
  yend = vector(1,Anrow);
  colu = vector(1,Anrow);

  Ncol = 0;                                 /* number of bad columns found */
  Mxsta = 1e33;  Mysta = 1e33;
  Mxend = -1e33; Myend = -1e33;
  for (jrow=1;jrow<=Anrow;jrow++){
    TCERDR(Atid,jrow,1,&colu[jrow],&null);  /* bad columns? (>0) */
    TCERDR(Atid,jrow,2,&xsta[jrow],&null);  /* get areas: */
    TCERDR(Atid,jrow,3,&xend[jrow],&null);  /* [xsta,ysta:xend,yend] */
    TCERDR(Atid,jrow,4,&ysta[jrow],&null);
    TCERDR(Atid,jrow,5,&yend[jrow],&null);

    if (xsta[jrow] < Mxsta) Mxsta = xsta[jrow];  /* get biggest intervals */
    if (xend[jrow] > Mxend) Mxend = xend[jrow];  /* in x and y in order to */
    if (ysta[jrow] < Mysta) Mysta = ysta[jrow];  /* a sufficiently big */
    if (yend[jrow] > Myend) Myend = yend[jrow];  /* area index map */

    Ncol = collect_col(colu[jrow],Ncol);         /* register bad columns */
  }                                              /* save in Bcol[] */
  iMxsta = (int)Mxsta;
  iMxend = (int)Mxend;
  iMysta = (int)Mysta;
  iMyend = (int)Myend;

  if (Ncol>0){
    BcolSaved = i_vector(0,Ncol-1);
    if (BcolSaved == NULL) nomemory("BcolSaved",Ncol);
  }
  Omap = i_matrix(1,Anrow,1,Anrow);
  if (Omap == NULL) nomemory("Omap",Anrow*Anrow*4);
  Imap = i_matrix(iMxsta,iMxend,iMysta,iMyend);
  if (Imap == NULL) nomemory("Imap",(iMxend-iMxsta+1)*(iMyend-iMysta+1)*4);

  jrow = 1;
  for (iy=(int)ysta[jrow];iy<=(int)yend[jrow];iy++)
    for (ix=(int)xsta[jrow];ix<=(int)xend[jrow];ix++){
      if (colu[jrow] < 1) Imap[ix][iy] = jrow; /* only work with non columns */
    }
  Omap[jrow][jrow] = 1;
  for (jrow=2;jrow<=Anrow;Omap[jrow][jrow] = 1,jrow++)
    for (iy=(int)ysta[jrow];iy<=(int)yend[jrow];iy++)
      for (ix=(int)xsta[jrow];ix<=(int)xend[jrow];ix++){
	if (colu[jrow] < 1){                   /* only work with non columns */
	  if (Imap[ix][iy] != 0){
	    Omap[jrow][Imap[ix][iy]] = 1;
	  }
	  Imap[ix][iy] = jrow;
	}
      }

  /*  show_mat(Imap,iMxsta,iMxend,iMysta,iMyend);
      show_mat(Omap,1,Anrow,1,Anrow);
  */
}


int get_overl()
/* METHOD:
   Each to be studied area is represented by a power of 2. Thus, you are able
   to do bit-operations onto these areas.
   The overlap table consists of all to be studied area pixels. Each pixel has
   a value between 2^0 until 2^32, their bits indicate the overlapps of the
   corresponding areas (e.g. 5 == 101  overlap of area 0 and 2 (or 1 and 3)).
   Only overlapping pixels will be analysed during this procedure.
   This is in principle what is used in this procedure. But in order to be more
   flexible the overlaps are not defined by INTEGERs but by STRINGs having the
   same meaning as an integer. Here, you are not restricted to an upper limit
   of number of areas (32) -- it may be as big as needed. So, each character
   in the string may be interpreted as a bit and typical AND/OR operations
   may be applied to these strings using the user-defined functions AND() and
   OR().

   (1) At the beginning the overlap of the first three areas are analysed by
   a simple bit-AND-operation. If they do not overlap the result would be 0.
   As only overlapping areas are analysed at least 2 bits are set of each pixel
   value. Thus, as soon as only one bit is the result of the AND operation, all
   three regions are overlapping (e.g.: binary: 101 AND 110 = 100).
   (2) In the next steps for each following region a overlap-check to the
   previous regions will be done in the following way:
        The pixels of the next region (i-th region) may take values from 1
	to 2^i. The combination of all these values will show all links
	of all previous regions to the i-th region. This combination will
	be done by the bit-OR-operation (e.g.: 100101 OR 101101 = 111101).
	Finally, a bit-AND-operation of this result with the previous one
	will be checked. If it is non zero an overlap of the i-th region
	with the previous ones has occured, otherwise the OR-result itself
	is an overlapping region. The overlap of the nonzero AND-result may
	be determined by a final bit-OR-operation.
	Examples:
	a)  old result:   10101
	    1.OR-result: 111101                     543210
	    AND:         010101  = nonzero -->  OR: 111101
	    ==> regions 0,2,3,4, and 5 are overlapping regions
	b)  old result:     101
	    1.OR-result:  11000
	    AND:          00000  = zero    -->  101 and 11000 are single
	                                        overlapping regions
            ==>	regions 0, and 2, are overlapping as well as 3, and 4.

   After all regions have been analysed the coordinates of the combination
   of all overlapping regions will be saved in the area definition table Btid.
   Moreover, the :VALUE column of all overlapping regions will be set to
   1.0 in order to find all non overlapping regions easily in a later step.
*/
{
  /*  int     res, res1, res2; */
  int     *i_vector();
  int     *res, *res1, *one, *sng;
  int     jrow, result, sz, nchk;  
  int i;

  sz = Anrow*sizeof(int);
  res1 = i_vector(1,Anrow);
  if (res1 == NULL) nomemory("res1",sz);
  res = i_vector(1,Anrow);
  if (res == NULL) nomemory("res",sz);
  one = i_vector(1,Anrow);
  if (one == NULL) nomemory("one",sz);
  fillmem(&one[1],1,Anrow);

  jrow = 1;                        /* skip 1st line: not necessary to check */
  nchk = 2;                        /* check first three areas */
  result = AND(&Omap[jrow][1],&Omap[jrow+1][1],nchk,&res[1]);
  if (result != 0){ res[1]=1; res[2]=1; }
  nchk++;
  jrow++;
  sng = one;
  memcpy(&res1[1],&res[1],sz);
  for (jrow++;jrow<=Anrow;jrow++){
    result = AND(&res1[1],&Omap[jrow][1],nchk,&res[1]);
    if (result != 0) OR(&res1[1],&Omap[jrow][1],nchk++,&res1[1]);
    else{
      sng[jrow] = 0;                        /* single area == not overlapped */
      if (AND(&Omap[jrow][1],&sng[1],nchk,&res[1]) != 0)          /* overlap */
	savearea(&Omap[jrow][1], nchk);
      else
	savearea(&res1[1], nchk-1);
      sng[jrow] = 1;
      nchk++;
    }
  }  

  if (result != 0) savearea(&res1[1], nchk-1);  /* save last occured overlap */

  sprintf(Mes,"%5d overlapping area(s) found.",Bjrow-1);
  messout(Mlev,4,Prgnam,Mes);
}


int savearea(x, wd)
int  *x, wd;
{
  int   i, num;
  int   null;
  float val, xcol, xsta, xend, ysta, yend;
  float xcolp, xstap, xendp, ystap, yendp;

  num = 0;
  xsta = 1e33; xend = -1e33;
  ysta = 1e33; yend = -1e33;
  xstap = 1e33; xendp = -1e33;
  ystap = 1e33; yendp = -1e33;
  xcol = 0.0;
  for (i=0;i<wd;i++){
    if (x[i] == 1){
      TCERDR(Atid,i+1,1,&val,&null);  /* negative xcol indicates areas at */
      xcol = MIN(xcol,val);           /* the edge; will be recognized */
      TCERDR(Atid,i+1,2,&val,&null);
      xsta = MIN(xsta,val);
      TCERDR(Atid,i+1,3,&val,&null);
      xend = MAX(xend,val);
      TCERDR(Atid,i+1,4,&val,&null);
      ysta = MIN(ysta,val);
      TCERDR(Atid,i+1,5,&val,&null);
      yend = MAX(yend,val);
      TCERDR(Atid,i+1,6,&val,&null);  /* pixel values (index-values) */
      xstap = MIN(xstap,val);
      TCERDR(Atid,i+1,7,&val,&null);
      xendp = MAX(xendp,val);
      TCERDR(Atid,i+1,8,&val,&null);
      ystap = MIN(ystap,val);
      TCERDR(Atid,i+1,9,&val,&null);
      yendp = MAX(yendp,val);
      val = 0.0;                       /* register as overlapped area: */
      TCEWRR(Atid,i+1,VALCOL,&val);    /* :VALUE!=0.0 indicates non-over- */
      num++;                           /* lapping area will be checked in */
                                       /* save_nooverl() */
    }
  }
  if (num>0) write_to_table(Bjrow++,&xcol,&xsta,&xend,&ysta,&yend,
			    &xstap,&xendp,&ystap,&yendp);
}


int save_nooverl()
{
  int   i, noo, null;
  float xcol, xsta, xend, ysta, yend, xstap, xendp, ystap, yendp;


  noo=0;
  for (i=1;i<Anrow+1;i++){
    TCERDR(Atid,i,VALCOL,&xsta,&null);   /* read :VALUE column */
    if (xsta != 0.0) {                   /* no overlapping area */
      TCERDR(Atid,i,1,&xcol,&null);      /* get entries of INPUT table */
      TCERDR(Atid,i,2,&xsta,&null);
      TCERDR(Atid,i,3,&xend,&null);
      TCERDR(Atid,i,4,&ysta,&null);
      TCERDR(Atid,i,5,&yend,&null);
      TCERDR(Atid,i,6,&xstap,&null);
      TCERDR(Atid,i,7,&xendp,&null);
      TCERDR(Atid,i,8,&ystap,&null);
      TCERDR(Atid,i,9,&yendp,&null);     /* append entries to OUTPUT table */
      write_to_table(Bjrow++,&xcol,&xsta,&xend,&ysta,&yend,
			    &xstap,&xendp,&ystap,&yendp);
      noo++;
    }
  }
  sprintf(Mes,"%5d non overlapping area(s) found.",noo);
  messout(Mlev,4,Prgnam,Mes);
}


int write_to_table(j,xcol,xsta,xend,ysta,yend,xstap,xendp,ystap,yendp)
float *xcol, *xsta, *xend, *ysta, *yend, *xstap, *xendp, *ystap, *yendp;
int   j;
{
  float val;

  val = 1;
  if (SNGLPOINT)
    *xcol = *xcol - 1.0;  /* single points are marked as -1 in :X */
  else
    val = MAX(*xend-*xsta+1,*yend-*ysta+1);

  if (check_column(xcol,*xsta,*xend) == 0){
    val = 0.0;            /* mark as not to be used area */
  }
  TCEWRR(Btid,j,1,xcol);
  TCEWRR(Btid,j,2,xsta);
  TCEWRR(Btid,j,3,xend);
  TCEWRR(Btid,j,4,ysta);
  TCEWRR(Btid,j,5,yend);
  TCEWRR(Btid,j,6,xstap);
  TCEWRR(Btid,j,7,xendp);
  TCEWRR(Btid,j,8,ystap);
  TCEWRR(Btid,j,9,yendp);
  TCEWRR(Btid,j,VALCOL,&val);

}


int check_column(xcol,xsta,xend,colwd)
float *xcol, xsta, xend, colwd;
{
  int i;
  float bcol;

  if (Ncol<1) return(1);           /* there are NO bad columns, SAVE */

  bcol = xend-xsta-1;              /* check if x-dimensions specify a column */
  if (bcol > Colwd) return(1);     /* NO column --> area may be saved */

  if (*xcol>0.0) bcol = *xcol;     /* already identified as a column */
  else bcol = (xsta+xend)/2;                  /* get column number */
  i = 0;
  while ((bcol != Bcol[i]) && (i<Ncol)) i++;  /* search bcol in Bcol[] */
  if (i<Ncol){                                /* bcol found in Bcol[] */
    if (BcolSaved[i]>0) return(0);            /* already saved */
    *xcol = bcol;
    BcolSaved[i] = 1;                         /* mark as saved */
    return(1);                                /* otherwise: SAVE area */
  }
                                              /* bcol not found */
  return(1);                                  /* --> no bad column */
                                              /* SAVE area */
}


int collect_col(bcol,ncol)
float bcol;
int   ncol;
{
  int i;
  i=0;
  if (bcol<=0) return(ncol);                     /* no real bad column */
  while ((bcol != Bcol[i]) && (i<ncol)) i++;     /* check if bcol is regist. */
  if (i<ncol)                                    /* bcol already registered */
    return(ncol);
  else                                           /* bcol not found */
    Bcol[i] = bcol;                              /* register bcol */
                                                 /* increase number of */
  return(ncol+1);                                /* registered bad columns */
}


int AND(vec1,vec2,num,res)
int *vec1, *vec2, *res, num;
{
  int i, cnt;

  for (i=0,cnt=0;i<num;i++){
    if ((vec1[i] != 0) && (vec1[i] == vec2[i])){
      res[i] = 1;
      cnt++;
    }else res[i] = 0;
  }
  /*  printf("AND:\n");
      printf("vec1: ");  show_vec(vec1,num);
      printf("vec2: ");  show_vec(vec2,num);
      printf("res:  ");  show_vec(res,num);
      */
  return(cnt);
}


int OR(vec1,vec2,num,res)
int *vec1, *vec2, *res, num;
{
  int i, cnt;

  for (i=0,cnt=0;i<num;i++){
    if ((vec1[i] == 0) && (vec1[i] == vec2[i]))
      res[i] = 0;
    else{
      res[i] = 1;
      cnt++;
    }
  }
  /*  printf("OR:\n");
      printf("vec1: ");  show_vec(vec1,num);
      printf("vec2: ");  show_vec(vec2,num);
      printf("res:  ");  show_vec(res,num);
      */
  return(cnt);
}


int show_vec(vec,num)
int *vec, num;
{
  int i;

  for (i=0;i<num;i++) printf("%d ",vec[i]);
  printf("\n");
}


int show_mat(map,ml,mh,nl,nh)
int **map, ml, mh, nl, nh;
{
  int i,j;

  for (j=nl;j<=nh;j++,printf("\n"))
    for (i=ml;i<=mh;i++) printf("%d ",map[i][j]);
  printf("\n");
}


int fillmem(vec,val,num)
int *vec, val, num;
{
  int i;

  for (i=0;i<num;i++) vec[i] = val;
}


int *i_vector(nl,nh)
int nl,nh;
{
        int *v;
	char *malloc();

        v=(int *)malloc((unsigned) (nh-nl+1)*sizeof(int));
        if (!v) nrerror("allocation failure in i_vector()");
	fillmem(v,0,nh-nl+1);
        return v-nl;
}


int **i_matrix(nrl,nrh,ncl,nch)
int nrl,nrh,ncl,nch;
{
        int i,j,**m,nsize;
	char *malloc();

        m=(int **)malloc((unsigned) (nrh-nrl+1)*sizeof(int*));
        if (!m) nrerror("allocation failure 1 in i_matrix()");
        m -= nrl;

	nsize = (nch-ncl+1);
        for(i=nrl;i<=nrh;i++) {
                m[i]=(int *)malloc((unsigned) (nch-ncl+1)*sizeof(int));
                if (!m[i]) nrerror("allocation failure 2 in i_matrix()");
                m[i] -= ncl;
		for (j=ncl;j<=nch;j++) m[i][j] = 0;
        }
        return m;
}



int nomemory(svar,num)
int   num;
char  *svar;
{
  sprintf(Mes,"Cannot allocate memory for %s [%d]!\n",svar,num);
  messout(Mlev,1,Prgnam,Mes);
  TCTCLO(Atid);
  TCTCLO(Btid);
  SCETER(-1,"*** memory-error");
  SCSEPI();
  exit(1);
}
