/* $Id: fors_convextract.c,v 1.3 1999/03/04 12:42:07 rpalsa Exp $   
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 *.COPYRIGHT   (C) 1998 European Southern Observatory
 *
 *.IDENT       $RCSfile: fors_convextract.c,v $
 *
 *.KEYWORDS    3d tables, multi object spectra
 *
 *.USAGE       run convextract.exe
 *
 *.INPUT       SPCFRM      2d frame with extracted spectra (cf. MIDAS MOS
 *                         documentation)
 *
 *.OUTPUT      SPCTBL      3d output table
 *
 *.RETURNS     The program returns 0 on success or 1 on error. The exit
 *             status is stored in keyword Q1.
 *
 *.PURPOSE     The spectra extracted from a MOS image by EXTRACT/MOS is
 *             written to a 3d table that contains the object positions in
 *             world coordinates and the wavelength, flux and flux error
 *             columns.
 *
 *.ENVIRON     MIDAS 98NOV, FORS pipeline
 *
 *.LANGUAGE    ANSI C/KR C
 *
 *.AUTHOR      R. Palsa, ESO-DMO
 *
 *.VERSION     0.1 1998/12/18: Creation, RP
 *
 *----------------------------------------------------------------------------
 */

#include <stdio.h>

#ifndef __STDC__
#include <string.h>
#endif

#include <midas_def.h>
#include <tblsys.h>
#include <tblerr.h>
#include <messages.h>



#define PROID_STRNG      "ConvExtract"
#ifndef __STDC__
#define PROID_STRNG_LEN  11
#endif

#define  MSG_BUFFER_SIZE  256



#ifdef __STDC__
void error(const char *name, int msglev, const char *msg)
#else
void error(name, msglev, msg)
  char *name, *msg;
  int msglev;
#endif
{

  int unit;


  messout(msglev, 1, name, msg);
  SCKWRC("Q1", 1, "1", 1, 60, &unit);
  SCSEPI();
}



void main() {

#ifdef __STDC__
  const char proId[] = PROID_STRNG;
#endif


  int    i, nval, unit, null, offset, idum;
  int    msglev;
  int    frm, naxis, npix[2];
  int    t3id, t3rno, t3nrow;
  int    t3start, t3step, t3ndata, t3wave, t3flux, t3ferr;

  size_t ndata, nspec;

  double start[2], step[2];
  float  *fptr, *eptr, *wptr;

  char   msg[MSG_BUFFER_SIZE];
  char   fname[FNAME_LEN + 1];
  char   t3name[TBL_NAMLEN + 1];
  char   lstart[TBL_LABLEN + 1], lstep[TBL_LABLEN + 1];
  char   lndata[TBL_LABLEN + 1], lwave[TBL_LABLEN + 1];
  char   lflux[TBL_LABLEN + 1], lferr[TBL_LABLEN + 1];



#ifndef __STDC__
  char proId[PROID_STRNG_LEN + 1];
  strcpy(proId, PROID_STRNG);
#endif


  /*
   * Connect to MIDAS
   */

  SCSPRO((char *)proId);


  /*
   * Get keywords from environment
   */

  SCKRDI("MESS_LEVEL", 1, 1, &nval, &msglev, &unit, &null);

  SCKGETC("SPCFRM", 1, FNAME_LEN, &nval, fname);
  SCKGETC("SPCTBL", 1, TBL_NAMLEN, &nval, t3name);


  /*
   * Get column labels for spectrum table from environment
   */

  if (SCKGETC("C_SOSTART", 1, TBL_LABLEN, &nval, lstart))
    error(proId, msglev, "Column label keyword C_SOSTART not found!");

  if (SCKGETC("C_SOSTEP", 1, TBL_LABLEN, &nval, lstep))
    error(proId, msglev, "Column label keyword C_SOSTEP not found!");

  if (SCKGETC("C_SONDATA", 1, TBL_LABLEN, &nval, lndata))
    error(proId, msglev, "Column label keyword C_SONDATA not found!");


  if (SCKGETC("C_SOWAVE", 1, TBL_LABLEN, &nval, lwave))
    error(proId, msglev, "Column label keyword C_SOWAVE not found!");

  if (SCKGETC("C_SOFLUX", 1, TBL_LABLEN, &nval, lflux))
    error(proId, msglev, "Column label keyword C_SOFLUX not found!");

  if (SCKGETC("C_SOFERR", 1, TBL_LABLEN, &nval, lferr))
    error(proId, msglev, "Column label keyword C_SOFERR not found!");


  /*
   * Read spectrum frame and image structure descriptors
   */

  if (SCFOPN(fname, D_R4_FORMAT, 0, F_IMA_TYPE, &frm))
    error(proId, msglev, "Cannot open input frame!");

  SCDRDI(frm, "NAXIS", 1, 1, &nval, &naxis, &unit, &null);
  if (naxis != 2)
    error(proId, msglev, "Input frame is not a 2-dimensional image");

  SCDRDI(frm, "NPIX", 1, 2, &nval, npix, &unit, &null);
  SCDRDD(frm, "START", 1, 2, &nval, start, &unit, &null);
  SCDRDD(frm, "STEP", 1, 2, &nval, step, &unit, &null);

  ndata = npix[0] * npix[1];
  if ((fptr = (float *)osmmget(ndata * sizeof(float))) == NULL)
    error(proId, msglev, "Not enough memory!");

  SCFGET(frm, 1, ndata, &nval, (char *)fptr);
  SCFCLO(frm);


  /*
   * Open output table
   */

  if (TCTOPN(t3name, F_IO_MODE, &t3id)) {
    osmmfree(fptr);
    error(proId, msglev, "Cannot open output table!");
  }


  /*
   * Check if number of spectra matches number of objects
   */

  TCIGET(t3id, &idum, &t3nrow, &idum, &idum, &idum);
  nspec = npix[1] / 2;

  if (t3nrow != nspec) {
    osmmfree(fptr);
    error(proId, msglev, "Number of objects do not match number of spectra!");
  }


  /*
   * Pointer to flux errors
   */

  eptr = fptr + nspec * npix[0];


  /*
   * Create output columns
   */

  if (TCCINI(t3id, D_R8_FORMAT, 1, "E18.8", "Angstrom", lstart, &t3start)) {
    osmmfree(fptr);
    sprintf(msg, "Cannot column :%s!", lstart);
    error(proId, msglev, msg);
  }

  if (TCCINI(t3id, D_R8_FORMAT, 1, "E18.8", "Angstrom", lstep, &t3step)) {
    osmmfree(fptr);
    sprintf(msg, "Cannot column :%s!", lstep);
    error(proId, msglev, msg);
  }

  if (TCCINI(t3id, D_I4_FORMAT, 1, "I6", "None", lndata, &t3ndata)) {
    osmmfree(fptr);
    sprintf(msg, "Cannot column :%s!", lndata);
    error(proId, msglev, msg);
  }

  if (TCCINI(t3id, D_R8_FORMAT, npix[0], "E18.8", "Angstrom", lwave,
	     &t3wave)) {
    osmmfree(fptr);
    sprintf(msg, "Cannot column :%s!", lwave);
    error(proId, msglev, msg);
  }

  if (TCCINI(t3id, D_R8_FORMAT, npix[0], "E18.8", "Adu", lflux, &t3flux)) {
    osmmfree(fptr);
    sprintf(msg, "Cannot column :%s!", lflux);
    error(proId, msglev, msg);
  }

  if (TCCINI(t3id, D_R8_FORMAT, npix[0], "E18.8", "Adu", lferr, &t3ferr)) {
    osmmfree(fptr);
    sprintf(msg, "Cannot column :%s!", lferr);
    error(proId, msglev, msg);
  }


  /*
   * Compute wavelength array for all spectra from frame descriptors
   */

  if ((wptr = (float *)osmmget(npix[0] * sizeof(float))) == NULL) {
    osmmfree(fptr);
    error(proId, msglev, "Not enough memory!");
  }
  else
    for (i = 0; i < npix[0]; i++)
      *(wptr + i) = start[0] + i * step[0];


  /*
   * Copy frame rows to table element arrays
   */

  for (i = 0; i < npix[1] / 2; i++) {
    offset = i * npix[0];
    t3rno = i + 1;

    TCEWRD(t3id, t3rno, t3start, start);
    TCEWRD(t3id, t3rno, t3step, step);
    TCEWRI(t3id, t3rno, t3ndata, npix);

    TCAWRR(t3id, t3rno, t3wave, 1, npix[0], wptr);
    TCAWRR(t3id, t3rno, t3flux, 1, npix[0], fptr + offset);
    TCAWRR(t3id, t3rno, t3ferr, 1, npix[0], eptr + offset);
  }


  /*
   * Cleanup
   */

  TCTCLO(t3id);
  osmmfree(fptr);
  osmmfree(wptr);


  /*
   * Disconnect from MIDAS
   */

  SCKWRC("Q1", 1, "0", 1, 60, &unit);
  SCSEPI();

}
