/* $Id: fors_extcoeff.c,v 1.6 1999/09/17 09:20:46 rpalsa Exp $   
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 *.COPYRIGHT   (C) 1998 European Southern Observatory
 *
 *.IDENT       $RCSfile: fors_extcoeff.c,v $
 *
 *.KEYWORDS    Extinction coefficients, spectroscopy
 *
 *.USAGE       run fors_extcoeff.exe
 *
 *.INPUT       EXTCAT      Catalog of 3d tables with atmospheric extinction
 *                         scale factors for several airmasses.
 *
 *.OUTPUT      EXTTBL      Output table of wavlength dependent atmospheric
 *                         extinction coefficients.
 *
 *.RETURNS     The program returns 0 on success or 1 on error. The exit
 *             status is stored in keyword Q1.
 *
 *.PURPOSE     Perform a robust, linear regression of atmospheric
 *             extinction scale factors as a function of wavelength, i.e.
 *             for all wavelength bins defined in the input tables. No
 *             rebinning of the input tables is done here, i.e. all input
 *             tables must be defined over the same wavelength range with
 *             the same wavelength bin size.
 *             
 *.ENVIRON     MIDAS 98NOV, FORS pipeline
 *
 *.LANGUAGE    ANSI C/KR C
 *
 *.AUTHOR      R. Palsa, ESO-DMO
 *
 *.VERSION     0.1 1998/12/18: Creation, RP
 *
 *----------------------------------------------------------------------------
 */

#include <stdio.h>

#ifndef __STDC__
#include <string.h>
#endif

#include <math.h>
#include <midas_def.h>
#include <tblsys.h>
#include <tblerr.h>
#include <nrutil.h>
#include <messages.h>



#define PROID_STRNG      "ExtCoeff"
#ifndef __STDC__
#define PROID_STRNG_LEN  8
#endif

#define MSG_BUFFER_SIZE  256
#define MIN_NUM_AIRMASS  3
#define EPS              1.e-7

#define MAX(A,B) ((A)  > (B) ? (A) : (B))
#define MIN(A,B) ((A)  < (B) ? (A) : (B))


typedef struct extinfo {
  int nspec;
  float airmass;
  float *spectra;
} XtInfo;



#ifdef __STDC__
void error(const char *name, int msglev, const char *msg)
#else
void error(name, msglev, msg)
  char *name, *msg;
  int msglev;
#endif
{

  int unit;


  messout(msglev, 1, name, msg);
  SCKWRC("Q1", 1, "1", 1, 60, &unit);
  SCSEPI();
}


#ifdef __STDC__
void xtfree(XtInfo *xt, size_t size)
#else
void xtfree(xt, size)
  XtInfo *xt;
  size_t size;
#endif
{

  register int i;


  for (i = 0; i < size; i++)
    if (xt->spectra)
      osmmfree(xt->spectra);

  osmmfree(xt);

  return;
}


#ifdef __STDC__
int transpose(float *a, int nx, int ny)
#else
int transpose(a, nx, ny)
     float *a;
     int   nx, ny;
#endif
{
  int i, j, k, l;
  float *b, tmp;

  if (nx < 0 || ny < 0)
    return 1;

  if (ny == 0 || ny == 1) 
    return 0;

  if (nx != ny)
    if ((b = (float *)osmmget(nx * ny * sizeof(float))) == 0)
      return 1;
    else {
      for (i = 0; i < ny; i++)
	for (j = 0; j < nx; j++) {
	  k = i * nx + j;
	  l = j * ny + i;

	  *(b + l) = *(a + k);
	}

      memcpy(a, b, nx * ny * sizeof(float));
      osmmfree(b);
    }
  else 
    for (i = 0; i < nx - 1; i++)
      for (j = i + 1 ; j < nx; j++) {
	k = i * nx + j;
	l = j * nx + i;

	tmp = *(a + k);
	*(a + k) = *(a + l);
        *(a + l) = tmp;
      }

  return 0;

}



void main() {

#ifdef __STDC__
  const char proId[] = PROID_STRNG;
#endif


  int    i, j, nval, unit, null, offset, idum;
  int    msglev;
  int    ntbl, nspec;
  int    t2id;
  int    t2wave, t2extc, t2zpnt, t2adev;
  int    t3id, t3nrow;
  int    t3start, t3step, t3ndata, t3flux;
  int    ndata, ndata1, ndata2;
  int    nairm;

  float  fdum;
  float  start, step, start1, step1, start2, step2;
  float  afit, bfit, absdev, wave;
  float  *airm, *xdata;

  char   msg[MSG_BUFFER_SIZE];
  char   catfile[FNAME_LEN + 1];
  char   t3name[TBL_NAMLEN + 1], t2name[TBL_NAMLEN + 1];
  char   lstart[TBL_LABLEN + 1], lstep[TBL_LABLEN + 1];
  char   lndata[TBL_LABLEN + 1], lflux[TBL_LABLEN + 1];
  char   lwave[TBL_LABLEN + 1], lextc[TBL_LABLEN + 1];
  char   lzpnt[TBL_LABLEN + 1], ladev[TBL_LABLEN + 1];

  XtInfo *xinfo, *xp;



#ifndef __STDC__
  char proId[PROID_STRNG_LEN + 1];
  strcpy(proId, PROID_STRNG);
#endif


  /*
   * Connect to MIDAS
   */

  SCSPRO((char *)proId);


  /*
   * Get keywords from environment
   */

  SCKRDI("MESS_LEVEL", 1, 1, &nval, &msglev, &unit, &null);

  SCKGETC("EXTCAT", 1, FNAME_LEN, &nval, catfile);
  SCKGETC("EXTTBL", 1, TBL_NAMLEN, &nval, t2name);


  /*
   * Get column labels for spectrum table from environment
   */

  if (SCKGETC("C_XTSTART", 1, TBL_LABLEN, &nval, lstart))
    error(proId, msglev, "Column label keyword C_XTSTART not found!");

  if (SCKGETC("C_XTSTEP", 1, TBL_LABLEN, &nval, lstep))
    error(proId, msglev, "Column label keyword C_XTSTEP not found!");

  if (SCKGETC("C_XTNDATA", 1, TBL_LABLEN, &nval, lndata))
    error(proId, msglev, "Column label keyword C_XTNDATA not found!");

  if (SCKGETC("C_XTRATIO", 1, TBL_LABLEN, &nval, lflux))
    error(proId, msglev, "Column label keyword C_XTRATIO not found!");

  if (SCKGETC("C_EXWAV", 1, TBL_LABLEN, &nval, lwave))
    error(proId, msglev, "Column label keyword C_EXWAV not found!");

  if (SCKGETC("C_EXTCO", 1, TBL_LABLEN, &nval, lextc))
    error(proId, msglev, "Column label keyword C_EXTCO not found!");

  if (SCKGETC("C_ZEROP", 1, TBL_LABLEN, &nval, lzpnt))
    error(proId, msglev, "Column label keyword C_ZEROP not found!");

  if (SCKGETC("C_XTADEV", 1, TBL_LABLEN, &nval, ladev))
    error(proId, msglev, "Column label keyword C_XTADEV not found!");


  /*
   * Get number of scale factor tables in input catalog
   */

  SCCSHO(catfile, &ntbl, &idum);


  /*
   *  Create housekeeping arrays to store all 'spectra' of each table
   *  and the corresponding airmass.
   */

  if ((xinfo = (XtInfo *)osmmget(ntbl * sizeof(XtInfo))) == NULL)
    error(proId, msglev, "Not enough memory!");

  for (i = 0; i < ntbl; i++)
    xinfo->spectra = NULL;

  
  /*
   * Go through catalog and check input tables and read extinction
   * data into memory.
   */

  for (i = 0; i < ntbl; i++) {
    SCCFND(catfile, i + 1, t3name);

    if (SCFINF(t3name, 99, NULL)) {
      xtfree(xinfo, ntbl);
      sprintf(msg, "Input table >%s< does not exist!", t3name);
      error(proId, msglev, msg);
    }
    else {
      if (TCTOPN(t3name, F_I_MODE, &t3id)) {
	xtfree(xinfo, ntbl);
	sprintf(msg, "Cannot open input table >%s<!", t3name);
	error(proId, msglev, msg);
      }

      TCCSER(t3id, lstart, &t3start);
      if (t3start < 0) {
	xtfree(xinfo, ntbl);
	sprintf(msg, "Column >%s< does not exist!", lstart);
	error(proId, msglev, msg);
      }

      TCCSER(t3id, lstep, &t3step);
      if (t3step < 0) {
	xtfree(xinfo, ntbl);
	sprintf(msg, "Column >%s< does not exist!", lstep);
	error(proId, msglev, msg);
      }

      TCCSER(t3id, lndata, &t3ndata);
      if (t3ndata < 0) {
	xtfree(xinfo, ntbl);
	sprintf(msg, "Column >%s< does not exist!", lndata);
	error(proId, msglev, msg);
      }

      TCCSER(t3id, lflux, &t3flux);
      if (t3flux < 0) {
	xtfree(xinfo, ntbl);
	sprintf(msg, "Column >%s< does not exist!", lflux);
	error(proId, msglev, msg);
      }

      TCIGET(t3id, &idum, &t3nrow, &idum, &idum, &idum);


      TCERDR(t3id, 1, t3start, &start1, &null);
      TCERDR(t3id, 1, t3step, &step1, &null);
      TCERDI(t3id, 1, t3ndata, &ndata1, &null);

      for (j = 2; j <= t3nrow; j++) {
	TCERDR(t3id, j, t3start, &start2, &null);
	TCERDR(t3id, j, t3step, &step2, &null);
	TCERDI(t3id, j, t3ndata, &ndata2, &null);
      
	if (start1 != start2 || step1 != step2 || ndata1 != ndata2) {
	  xtfree(xinfo, ntbl);
	  error(proId, msglev, "Invalid input table. Data grids differ!");
	}
      }

      if (i == 0) {
	start = start1;
	step = step1;
	ndata = ndata1;
      }
      else
	if (start1 != start || step1 != step || ndata1 != ndata) {
	  xtfree(xinfo, ntbl);
	  error(proId, msglev,
		"Invalid input tables. Data grids do not match!");
	}
    }


    /*
     * Read extinction data from table
     */

    xp = xinfo + i;
    xp->nspec = t3nrow;
    
    SCDRDR(t3id, "O_AIRM", 1, 1, &nval, &xp->airmass, &unit, &null);

    if ((xp->spectra = (float *)osmmget(t3nrow * ndata * sizeof(float))) ==
	NULL) {
      xtfree(xinfo, ntbl);
      error(proId, msglev, "Not enough memory!");
    }
    else 
      for (j = 0; j < t3nrow; j++)
	if (TCARDR(t3id, j + 1, t3flux, 1, ndata, xp->spectra + j * ndata)) {
	  xtfree(xinfo, ntbl);
	  sprintf(msg, "Cannot read extinction scale factors from >%s<!",
		  t3name);
	  error(proId, msglev, msg);
	}


    TCTCLO(t3id);
  }


  /*
   * Restructure extinction scale factor in memory for easy fitting
   */

  nspec = 0;
  for (i = 0; i < ntbl; i++) {
    xp = xinfo + i;
    nspec += xp->nspec;
  }
    
  if ((airm = (float *)osmmget(nspec * sizeof(float))) == NULL) {
    xtfree(xinfo, ntbl);
    error(proId, msglev, "Not enough memory!");
  }

  if ((xdata = (float *)osmmget(nspec * ndata * sizeof(float))) == NULL) {
    xtfree(xinfo, ntbl);
    osmmfree(airm);
    error(proId, msglev, "Not enough memory!");
  }

  offset = 0;
  for (i = 0; i < ntbl; i++) {
    xp = xinfo + i;

    for (j = 0; j < xp->nspec; j++) {
      airm[j + offset] = xp->airmass;
      memcpy(xdata + (offset + j) * ndata, xp->spectra + j * ndata, ndata * 
	     sizeof(float));
    }

    offset += xp->nspec;
  }

  xtfree(xinfo, ntbl);


  /*
   * Check number of available airmasses
   */

  nairm = 1;
  for (i = 1; i < nspec; i++)
    if (fabs(airm[i] - airm[i - 1]) > EPS)
      nairm++;

  if (nairm < MIN_NUM_AIRMASS) {
    osmmfree(airm);
    osmmfree(xdata);
    sprintf(msg, "Number of different airmass values is %d. Required minimum "
	    "number is %d!", nairm, MIN_NUM_AIRMASS);
    error(proId, msglev, msg);
  }
  else 
    transpose(xdata, ndata, nspec);


  /*
   * Create output table of fitted extinction coefficients
   */

  if (TCTINI(t2name, F_TRANS, F_O_MODE, 4, ndata, &t2id)) {
    osmmfree(airm);
    osmmfree(xdata);
    sprintf(msg, "Cannot create output table >%s<!", t2name);
    error(proId, msglev, msg);
  }

  TCCINI(t2id, D_R4_FORMAT, 1, "F10.4", "Angstrom", lwave, &t2wave);
  TCCINI(t2id, D_R4_FORMAT, 1, "F10.4", "mag", lzpnt, &t2zpnt);
  TCCINI(t2id, D_R4_FORMAT, 1, "F10.4", "mag/airmass", lextc, &t2extc);
  TCCINI(t2id, D_R4_FORMAT, 1, "F10.4", "mag", ladev, &t2adev);

    
  /*
   * Robust linear regression of atmospheric extinction data.
   * Since medfitc is a modified version of the Numerical Recipes'
   * medfit the pointers for airmass and extinction scale factor have
   * to be offset by -1!
   */

  for (i = 0; i < ndata; i++) {
    medfitc(airm - 1, xdata + i * nspec - 1, nspec, &afit, &bfit, &absdev,
	    &fdum);

    wave = start + i * step;
    TCEWRR(t2id, i + 1, t2wave, &wave);
    TCEWRR(t2id, i + 1, t2zpnt, &afit);
    TCEWRR(t2id, i + 1, t2extc, &bfit);
    TCEWRR(t2id, i + 1, t2adev, &absdev);
  }


  /*
   * Cleanup
   */

  osmmfree(airm);
  osmmfree(xdata);


  /*
   * Disconnect from MIDAS
   */

  SCKWRC("Q1", 1, "0", 1, 60, &unit);
  SCSEPI();

}
