/* $Id: fors_findorder_lin.c,v 1.2 2000/05/15 19:31:49 nrainer Exp $ */
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

 .COPYRIGHT  (C) 1999  European Southern Observatory (ESO)

 .IDENT      fors_findorder_lin.c

 .KEYWORDS   Order detection

 .PURPOSE    Hough transformation of the x- and y-positions of the orders
 __________

 .USAGE      run PIPE_EXE:(IDENT) (invoked from fors_findorder_lin.prg)

 .INPUT      IN_A     : input table with x- and y-positions
             IN_B     : name for output image
             INPUTR(1): minimal slope for Hough transformation
             INPUTR(2): maximal slope for Hough transformation
             INPUTR(3): slope step for Hough transformation
             INPUTR(4): zero point step for Hough transformation

 .OUTPUT     IN_B: Hough transformation image with slope as x-axis and
                   zero points as y-axis

 .RETURN     Q1:  0: successful return
                 -1: a warning message occured
                  1: an error occured
 __________

 .ALGORITHM  Hough transformation - see P. Ballester, Proceedings of the
             3rd Data Analysis Workshop, 1991.

 .COMMENT    Special programme for order detection - not to be used generally.

 .ENVIRON    MIDAS 99NOV

 .LANGUAGE   C (ANSI)

 .AUTHOR     N.Rainer  DPG/DMD/ESO

 .VERSION    1.4  99/07/12  Creation
                  99/10/21  Release

----------------------------------------------------------------------------*/


/*  PROTOTYPES  :
    **********/

#ifdef __STDC__  /* ANSI */
  static void  error(void);
#else  /* K&R */
  static void  error();
#endif


/*  INCLUDED FILES  :
    **************/

#include  <midas_def.h>
#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  <math.h>


/*  DEFINITIONS OF CONSTANTS AND MACROS  :
    ***********************************/

#define  PROID    "findorder_lin.c"


/*  GLOBAL VARIABLES  :
    ****************/

static int  messlevel;


/****************************************************************************/


#ifdef __STDC__  /* ANSI */
  void main(int argc, char *argv[])
#else  /* K&R */
  void main(argc, argv)
  int  argc;
  char *argv[];
#endif
{

/*  LOCAL VARIABLES  :
    ***************/

char          postbl[65], tmps[160], houima[65], ident[72], cunit[48];

double        start[2], step[2];

float         slopmin, slopmax, slopstep, zepostep, *xpos, *ypos;
float         xmin, xmax, ymin, ymax, zepomin, zepomax, zepowld;
float         *houpnt, *houpnt0, slop;

int           r, nval, unit, null, info, row, loop, slopidx, warn = 0;
int           idpostbl, nrows, npix[2], idhouima, zepopix;


SCSPRO("findorder_lin");


if (argc > 1 && *argv[1] == '-' && tolower(*(argv[1]+1)) == 'h') {
  SCTPUT("      ******************************************************************");
  SCTPUT("      IN_A     : input table with x- and y-positions");
  SCTPUT("      IN_B     : name for output image");
  SCTPUT(" ");
  SCTPUT("      INPUTR(1): minimal slope for Hough transformation");
  SCTPUT("      INPUTR(2): maximal slope for Hough transformation");
  SCTPUT("      INPUTR(3): slope step for Hough transformation");
  SCTPUT("      INPUTR(4): zero point step for Hough transformation");
  SCTPUT("      ******************************************************************");
  SCSEPI();
}


/*  INPUT  :
    *****/

r = SCKRDI("MESS_LEVEL",1,1,&nval,&messlevel,&unit,&null);

r = SCKGETC("IN_A",1,64,&nval,postbl);
r = SCFINF(postbl,99,&info);
if (r != 0) {
  SCTPUT(" ");
  messout(messlevel,1,PROID,"Table not found or can not be opened !");
  sprintf(&tmps[0],"(%s)",postbl);
  messout(messlevel,1,PROID,tmps);
  error();
}

r = SCKGETC("IN_B",1,16,&nval,houima);
r = SCKRDR("INPUTR",1,1,&nval,&slopmin,&unit,&null);
r = SCKRDR("INPUTR",2,1,&nval,&slopmax,&unit,&null);
r = SCKRDR("INPUTR",3,1,&nval,&slopstep,&unit,&null);
r = SCKRDR("INPUTR",4,1,&nval,&zepostep,&unit,&null);


/*  READ INPUT TABLE  :
    ****************/

r = TCTOPN(postbl,F_IO_MODE,&idpostbl);
r = TCIGET(idpostbl,&null,&nrows,&null,&null,&null);

xpos = (float *) malloc ((unsigned int)(nrows*sizeof(float)));
if (xpos == 0) {
  SCTPUT(" ");
  messout(messlevel,1,PROID,"Allocation not possible !");
  sprintf(&tmps[0],"(xpos)");
  messout(messlevel,1,PROID,tmps);
  error();
}
ypos = (float *) malloc ((unsigned int)(nrows*sizeof(float)));
if (ypos == 0) {
  SCTPUT(" ");
  messout(messlevel,1,PROID,"Allocation not possible !");
  sprintf(&tmps[0],"(ypos)");
  messout(messlevel,1,PROID,tmps);
  error();
}

for (row=1; row<=nrows; row++) {
  r = TCERDR(idpostbl,row,1,&xpos[row-1],&null);
  if (row == 1) {
    xmin = xpos[0];
    xmax = 0.;
  }
  if (xpos[row-1] < xmin) xmin = xpos[row-1];
  if (xpos[row-1] > xmax) xmax = xpos[row-1];
  r = TCERDR(idpostbl,row,2,&ypos[row-1],&null);
  if (row == 1) {
    ymin = ypos[0];
    ymax = 0.;
  }
  if (ypos[row-1] < ymin) ymin = ypos[row-1];
  if (ypos[row-1] > ymax) ymax = ypos[row-1];
}


/*  PREPARE OUTPUT IMAGE  :
    ********************/

npix[0] = (int) ((slopmax-slopmin)/slopstep);
zepomin = ymin - slopmax*(xmax-xmin);
zepomax = ymax + slopmin*(xmax-xmin);
npix[1] = (int) ((zepomax-zepomin)/zepostep);
start[0] = slopmin;
start[1] = zepomin;
step[0] = slopstep;
step[1] = zepostep;
strcpy(ident,"Hough transformation image");
strcpy(cunit,"HOUGH_INTENSITY SLOPES          ZERO_POINTS");
        /* 1--^           17--^           33--^ */

r = SCIPUT(houima,D_R4_FORMAT,F_O_MODE,F_IMA_TYPE,
           2,npix,start,step,ident,cunit,(char **)&houpnt,&idhouima);

houpnt0 = houpnt;
for (loop=0; loop<npix[0]*npix[1]; loop++) *houpnt++ = 0.;
houpnt = houpnt0;


/**/


/****************************
 *                          *
 *  HOUGH TRANSFORMATION :  *
 *                          *
 ****************************/


for (row=1; row<=nrows; row++) {

  for (slopidx=1; slopidx<=npix[0]; slopidx++) {
    slop = slopmin + (slopidx-1)*slopstep;
    zepowld = ypos[row-1] - slop*xpos[row-1];
    zepopix = (int) ((zepowld - zepomin)/zepostep + 0.5);
    if (zepopix >= 1 && zepopix < npix[1])  /* Not"<="(array bounds)! */
      houpnt[(int)(zepopix*npix[0]+(slopidx-1))] += 1.;
  }

}

r = TCTCLO(idpostbl);
free(xpos);
free(ypos);

r = SCFCLO(idhouima);


if (warn == 0)
  r = SCKWRC("Q1",1,"0",1,60,&unit);
else
  r = SCKWRC("Q1",1,"-1",1,60,&unit);
r = SCSEPI();


} /* void main(int argc, char *argv[]) */


/****************************************************************************/

#ifdef __STDC__  /* ANSI */
  static void error(void)
#else  /* K&R */
  static void error()
#endif
{
  int  r, unit;

  r = SCKWRC("Q1",1,"1",1,60,&unit);
  r = SCSEPI();
}

/******************************************************************************/
