/* $Id: fors_medsigmed.c,v 1.25 2000/05/15 19:31:50 nrainer Exp $    */
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

 .COPYRIGHT  (C) 1998  European Southern Observatory (ESO)

 .IDENT      fors_medsigmed.c

 .KEYWORDS   Statistics

 .PURPOSE    Calculation of a more robust value for deviations from median
             values.
 __________

 .USAGE      run PIPE_EXE:(IDENT) (invoked from fors_medsigmed.prg)

 .INPUT      IN_A     : input frame

             INPUTI(1): number of prescan columns,
             INPUTI(2): number of overscan columns

             INPUTC(1): calculation flag: <+>:calculate median,
                        <u>se median from INPUTR(1)

             INPUTC(2): additional calc. flag: <y>:calculate MAD_MED,
                        <n>o;

 .OUTPUT     OUTPUTR(8) : median
             OUTPUTR(15): median deviation ("sigma") of the median
             if INPUTC(2)!=n => OUTPUTR(14): mean abs.dev. of the median

             DESCRIPTOR STATISTIC(8) : the same
             DESCRIPTOR STATISTIC(15): the same
             if INPUTC(2)!=n => DESCRIPTOR STATISTIC(14): the same

 .RETURN     Q1:  0: successful return
                 -1: a warning message occured
                  1: an error occured
 __________

 .ALGORITHM  The output is the result of the following formula:
               SQRT ( MEDIAN [ (x[i]-<MEDIAN>)^2 ] ),
               with MEDIAN rank, if number of pixels is ODD: (N+1)/2  EVEN: N/2

 .COMMENT    Compare with the formula for standard deviation:
               SQRT ( MEAN [ (x[i]-<AVERAGE>)^2 ] ),
               with MEAN=(1/(N-1))

             The selection of the 50%-quantile ("median") is done by the
             "select" function from Numerical Recipes v2.08 (renamed to
             "selection" because of HP problems).

 .ENVIRON    MIDAS 98NOV

 .LANGUAGE   C (ANSI)

 .AUTHOR     N.Rainer  DPG/DMD/ESO

 .VERSION    1.0  98/08/21  Creation
                  98/12/15  Release
             1.1  99/01/14  Release
             1.2  99/03/15  Release
             1.3  99/09/23  Release

----------------------------------------------------------------------------*/


/*  PROTOTYPES  :
    **********/

#ifdef __STDC__  /* ANSI */

  static void    error( void);
  static float   selection(unsigned long k, unsigned long n, float *arr);

#else  /* K&R */

  unsigned long  k, n;
  float          *arr;

  static void    error();
  static float   selection();

#endif


/*  INCLUDED FILES  :
    **************/

#include  <midas_def.h>
/* ONLY FOR DEBUGGING WITH `cflow': */
/* #include  "/home/dfs1rp1/dmd_df/nrainer/LAN/MIDAS/C/CFLOW/midas_def.h" */

#include  <ctype.h>
#include  <stdio.h>
#include  <stdlib.h>
#include  <math.h>


/*  DEFINITIONS OF CONSTANTS AND MACROS  :
    ***********************************/

#define        PROID      "medsigmed.c"
#define        NPIXMIN    3  /* FEWER PIXELS NOT REASONABLE FOR MEDIAN */
#define        SWAP(a,b)  temp=(a);(a)=(b);(b)=temp;


/*  GLOBAL VARIABLES  :
    ****************/

static int  messlevel;


/****************************************************************************/


#ifdef __STDC__  /* ANSI */
  void main(int argc, char *argv[])
#else  /* K&R */
  void main(argc, argv)
  int  argc;
  char *argv[];
#endif
{

/*  LOCAL VARIABLES  :
    ***************/

char           fileimaorig[65];
char           newmed[2], addmad[2], tmps[160];

float          *imaorig, *imaorig0;
float          median, medsigmed, madmed;

unsigned long  i, j;
unsigned long  pixorig;
unsigned long  kthsmallest;

int            r, nval, unit, null, info;
int            warn = 0;
int            scan[2], naxisimaorig, npiximaorig[2];
int            idimaorig;


SCSPRO("medsigmed");


if (argc > 1 && *argv[1] == '-' && tolower(*(argv[1]+1)) == 'h') {
  SCTPUT("      ******************************************************************");
  SCTPUT("      IN_A     : input frame");
  SCTPUT(" ");
  SCTPUT("      INPUTI(1): number of prescan columns,");
  SCTPUT("      INPUTI(2): number of overscan columns");
  SCTPUT(" ");
  SCTPUT("      INPUTC(1): calculation flag: <+>:calculate median,");
  SCTPUT("                 <u>se median from INPUTR(1)");
  SCTPUT(" ");
  SCTPUT("      INPUTC(2): additional calc. flag: <y>:calculate MAD_MED,");
  SCTPUT("                 <n>o;");
  SCTPUT("      ******************************************************************");
  SCSEPI();
}


/*  INPUT  :
    *****/

r = SCKRDI("MESS_LEVEL",1,1,&nval,&messlevel,&unit,&null);

r = SCKGETC("IN_A",1,64,&nval,fileimaorig);
r = SCFINF(fileimaorig,99,&info);
if (r != 0) {
  SCTPUT(" ");
  messout(messlevel,1,PROID,"Frame not found or can not be opened !");
  sprintf(&tmps[0],"(%s)",fileimaorig);
  messout(messlevel,1,PROID,tmps);
  error();
}

r = SCKRDI("INPUTI",1,2,&nval,scan,&unit,&null);
if (scan[0] < 0 || scan[1] < 0) {
  SCTPUT(" ");
  messout(messlevel,1,PROID,"Impossible pre- or overscan value !");
  error();
}

r = SCKGETC("INPUTC",1,1,&nval,newmed);
if (tolower(newmed[0]) == 'u')
  r = SCKRDR("INPUTR",1,1,&nval,&median,&unit,&null);

r = SCKGETC("INPUTC",2,1,&nval,addmad);


/*  READING PIXEL VALUES  :
    ********************/

messout(messlevel,4,PROID,"      reading image data");

r = SCFOPN(fileimaorig,D_R4_FORMAT,0,F_IMA_TYPE,&idimaorig);

r = SCDRDI(idimaorig,"NAXIS",1,1,&nval,&naxisimaorig,&unit,&null);
if (naxisimaorig == 2) {
  r = SCDRDI(idimaorig,"NPIX",1,2,&nval,npiximaorig,&unit,&null);
  pixorig = npiximaorig[0]*npiximaorig[1];
} else if (naxisimaorig == 1) {
  r = SCDRDI(idimaorig,"NPIX",1,1,&nval,npiximaorig,&unit,&null);
  pixorig = npiximaorig[0];
  npiximaorig[1] = 1;  /* for removing of pre-/overscan below */
} else {
  SCTPUT(" ");
  messout(messlevel,1,PROID,"Frame dimension not supported !");
  sprintf(&tmps[0],"(%s has %d axes)!",fileimaorig,naxisimaorig);
  messout(messlevel,1,PROID,tmps);
  error();
}
if (pixorig < NPIXMIN) {
  SCTPUT(" ");
  sprintf(&tmps[0],"Too few pixels (at least %d necessary) !",NPIXMIN);
  messout(messlevel,1,PROID,tmps);
  sprintf(&tmps[0],"(%s)",fileimaorig);
  messout(messlevel,1,PROID,tmps);
  error();
}

imaorig = (float *) malloc ((unsigned int)(pixorig*sizeof(float)));
if (imaorig == 0) {
  SCTPUT(" ");
  messout(messlevel,1,PROID,"Allocation not possible !");
  sprintf(&tmps[0],"(%s)",fileimaorig);
  messout(messlevel,1,PROID,tmps);
  error();
}

r = SCFGET(idimaorig,1,pixorig,&nval,(char *)imaorig);


if (scan[0] != 0 || scan[1] != 0) {

  /*  REMOVING PRE- AND OVERSCAN REGIONS IN X-DIRECTION :
      ************************************************/

  messout(messlevel,4,PROID,"      removing pre- and overscan regions");

  if (npiximaorig[0] <= scan[0]+scan[1]) {
    SCTPUT(" ");
    messout(messlevel,1,PROID,"(pre- + overscan) >= (#pixels in x-axis) !");
    sprintf(&tmps[0],"(%s)",fileimaorig);
    messout(messlevel,1,PROID,tmps);
    error();
  }

  imaorig0 = imaorig;
  for (j=0; j<npiximaorig[1]; j++) {
    for (i=0; scan[0]+i<npiximaorig[0]-scan[1]; i++) {
      /* imaorig[j*(npiximaorig[0]-scan[0]-scan[1])+i] = */
      /*   imaorig[j*npiximaorig[0]+scan[0]+i];          */
         *imaorig++ =
           *(imaorig0+j*npiximaorig[0]+scan[0]+i);
    }
  }
  imaorig = imaorig0;
  pixorig = npiximaorig[1]*(npiximaorig[0]-scan[0]-scan[1]);

  if (pixorig < NPIXMIN) {
    SCTPUT(" ");
    messout(messlevel,1,PROID,
            "Too few pixels left after removing pre-/overscan !");
    sprintf(&tmps[0],"(< %d in %s)",NPIXMIN,fileimaorig);
    messout(messlevel,1,PROID,tmps);
    error();
  }

} /* if (scan[0] != 0 || scan[1] != 0) */


/**/


/***************************
 *                         *
 *  ACTUAL COMPUTATIONS  : *
 *                         *
 ***************************/

imaorig--;  /* The following "selection" function uses unit-offset */
            /* pointer addresses instead of zero-offset ones !     */

/* Median rank, if number of pixels is  ODD: (N+1)/2  EVEN: N/2 */
kthsmallest = (div(pixorig,2).rem == 1) ? (pixorig+1)/2 : pixorig/2 ;

if (tolower(newmed[0]) != 'u') {  /* no median value in INPUTR(1) */
  messout(messlevel,4,PROID,"      calculating median");
  median = selection(kthsmallest,pixorig,imaorig);
} else {
  messout(messlevel,4,PROID,"      using entered median");
}

messout(messlevel,4,PROID,"      calculating median deviation of the median");
imaorig++;
imaorig0 = imaorig;
for (i=0; i<pixorig; i++)
  *imaorig++ = (float)pow((double)(*imaorig-median),(double)2); 
imaorig = imaorig0;
imaorig--;

medsigmed = selection(kthsmallest,pixorig,imaorig);
medsigmed = sqrt(medsigmed);

if (tolower(addmad[0]) != 'n') {
  messout(messlevel,4,PROID,"      calculating mean abs.dev. of the median");
  imaorig++;
  imaorig0 = imaorig;
  for (i=0, madmed=0; i<pixorig; i++) {
    *imaorig = (float)sqrt((double)(*imaorig)); 
    madmed += *imaorig++;
  }
  imaorig = imaorig0;
  imaorig--;

  madmed /= pixorig;
}


free(++imaorig);  /* "+1" because of unit-offset */
                  /*   before median computation */


r = SCKWRR("OUTPUTR",&median,8,1,&unit);
r = SCKWRR("OUTPUTR",&madmed,14,1,&unit);
r = SCKWRR("OUTPUTR",&medsigmed,15,1,&unit);

r = SCDWRR(idimaorig,"STATISTIC",&median,8,1,&unit);
r = SCDWRR(idimaorig,"STATISTIC",&madmed,14,1,&unit);
r = SCDWRR(idimaorig,"STATISTIC",&medsigmed,15,1,&unit);
r = SCDWRH(idimaorig,"STATISTIC","14: mad_med, 15: medsigmed",1,26);

r = SCFCLO(idimaorig);


if (warn == 0)
  r = SCKWRC("Q1",1,"0",1,60,&unit);
else
  r = SCKWRC("Q1",1,"-1",1,60,&unit);
r = SCSEPI();


} /* void main(int argc, char *argv[]) */


/****************************************************************************/

#ifdef __STDC__  /* ANSI */
  static void error(void)
#else  /* K&R */
  static void error()
#endif
{
  int  r, unit;

  r = SCKWRC("Q1",1,"1",1,60,&unit);
  r = SCSEPI();
}

/*----------------------------------------------------------------------------*/

/* #define   SWAP(a,b)  temp=(a);(a)=(b);(b)=temp; */

#ifdef __STDC__  /* ANSI */
  static float selection(unsigned long k, unsigned long n, float *arr)
#else  /* K&R */
  static float selection(k, n, arr)
  unsigned long  k, n;
  float          *arr;
#endif
{
    unsigned long  i, ir, j, l, mid;
    float          a,temp;

    l=1;
    ir=n;
    for (;;) {

        if (ir <= l+1) {        
            if (ir == l+1 && arr[ir] < arr[l]) {
                SWAP(arr[l],arr[ir])
            }
            return arr[k];
        } else {
            mid=(l+ir) >> 1;
            SWAP(arr[mid],arr[l+1])
            if (arr[l] > arr[ir]) {
                SWAP(arr[l],arr[ir])
            }
            if (arr[l+1] > arr[ir]) {
                SWAP(arr[l+1],arr[ir])
            }
            if (arr[l] > arr[l+1]) {
                SWAP(arr[l],arr[l+1])
            }
            i=l+1;
            j=ir;
            a=arr[l+1];

            for (;;) {
                do i++; while (arr[i] < a);
                do j--; while (arr[j] > a);
                if (j < i) break;
                SWAP(arr[i],arr[j])
            }

            arr[l+1]=arr[j];
            arr[j]=a;
            if (j >= k) ir=j-1;
            if (j <= k) l=i;
        }

    }
}

/******************************************************************************/
