/* $Id: fors_ovexpomap.c,v 1.7 2001/07/11 12:00:39 sbogun Exp $ */
/*+++ FORS_OVEXPOMAP ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 
 .COPYRIGHT     (C) 1998  European Southern Observatory
 
 .IDENT         fors_ovexpomap
 
 .KEYWORDS      Determine bad pixels on an image frame (flatfield of a CCD)
 
 .USAGE         run PIPE_EXE:(IDENT)
 
 .INPUT         IFRM       (char)  input frame (flats, science, std.star)
                MASK       (char)  name of overexposed pixel mask
                                   (already exist)
                HGH        (int)   high level factor (HGH*mean)
                HGHFLG     (int)   high level flag (overexposed pixel will be
                                   set to this value in the mask image).
                NPIXMAX     (int)  maximum number of allowed bad pixel
                                   in the image section.
                IM_SEC     (int)   image section (pixel coordinates)
                MESS_LEVEL (int)   message level

 .OUTPUT        OUTPUTI(1) (int)   number of overexposed pixels found.
                                   -1 in case an error had occured.
 
 .RETURN        Number of detected overexposed pixel.
                If NPIXMAX is set to an positive value the return number will
                be at maximum the maximum allowed number of overexposed pixels.

 .PURPOSE       Detect overeposed pixels in specified image section.
 
 .ALGORITHM     Loops through the specified image section of IFRM and compares
                the value at each pixel position with the limit HGH. In the
                MASK frame all values greater than HGH will be indicated by
                HGHFLG, the rest remains 0. 
                In case NPIXMAX is a positive number (the maxim. allowed number
                of overexposed pixels) the loop stops as soon as this limit is
                reached and the program returns NPIXMAX.

 .ENVIRON       MIDAS
    
 .LANGUAGE      C
 
 .AUTHOR        Sebastian Wolf, ESO-DMO
 
 .VERSION       1.0 1998/12/04  SW Creation.
                1.2 1999/02/16  SW Stop the findpix()-loop as soon as the
                                   maximum allowed number of overexposed pixels
                                   has exceeded.
                1.7 2001/07/11  SB Cleaned code completely. Segmentation fault
                                   does not occur anymore.
 .COMMENT

-----------------------------------------------------------------------------*/

#include <string.h>
#include <stdio.h>
#include <tblsys.h>
#include <tblerr.h>
#include <tbldef.h>
#include <macrogen.h>
#include <atype.h>
#include <midas_def.h>
#include <math.h>

#define TRUE 1
#define FALSE 0
#define BELL  0x7
#define MLEN  255
#define NLEN  160


int   Mlev;                             /* message level MESS_LEVEL */
int   NMx, NMy;                         /* dimensions of mask and inputframe */
int   Cnt;                              /* counts checked edges */
int   Ifrm;                             /* file id of image frame */
int   Imsk;                             /* file id of mask frame */
int   Hgh, Hghflg;                      /* high limit and replace value */
int   Mpix;                             /* max. allowed detections */
int   Num;                              /* total number of pixels in img */
int   NMea;                             /* total number of pixels in Mean */
int   DatformI, DatformM;               /* dataformat of input files */
int   SI, SU, SF;                       /* size of int, size of float */
int   Imsc[5];                          /* image section */

int   **i_matrix();                     /* int matrix + elem. initialization */
int   **imap2mat();                     /* map a matrix on a integer vector */
unsigned short **usmap2mat();           /* map a matrix on a float vector */
float **map2mat();                      /* map a matrix on a float vector */

float *Frmf, **MFrmf;                   /* image frame, matrix represent. */
float *Mskf, **MMskf;                   /* bad pixel mask, matrix repres. */
unsigned short *Frmu, **MFrmu;          /* image frame, matrix represent. */
unsigned short *Msku, **MMsku;          /* bad pixel mask, matrix repres. */


char  Mes[MLEN];                        /* message buffer */
char  Imgname[NLEN] = "combff.bdf";     /* filename of image frame */
char  Mskname[NLEN] = "badmsk.bdf";     /* filename of bad pixel mask */
char  Pixname[NLEN] = "badall.tbl";     /* filename of pixel table */

char  Prgnam[]  = "ovexpomap";
char  Version[] = "1998.12.04-SW";

int get_params();
int findbadpixel(int low, int hgh, int mpix, int *npix);
int findpix(float **img, float **msk, int ncl, int nch, int nrl, int nrh,
                        int low, int hgh, int mpix);
int ufindpix(unsigned short **img, unsigned short **msk, int ncl, int nch,
                         int nrl, int nrh, int low, int hgh, int mpix);
void no_memory(char *svar, int num);
void nobadpix();
void warn(int i);


int main()
{
  int    nbad, uni, returncode = -1;

  SCSPRO(Prgnam);
  
  SU = sizeof(unsigned short);
  SF = sizeof(float);

  nbad = 0;
  SCKWRI("OUTPUTI",&nbad,1,1,&uni);                  /* return nbad */

  if (returncode = get_params()) {                   /* get MIDAS parameters */
    SCSEPI();
    exit(returncode);
  }

  if (returncode = findbadpixel(0,Hgh,Mpix,&nbad)) { /* find bad pixel and */
    SCSEPI();
    exit(returncode);
  }
                                                     
  SCKWRI("OUTPUTI",&nbad,1,1,&uni);                  /* return nbad */
  SCSEPI();

  return(0);
}





int get_params()
{
  char   *ptr;
  int    i, len;
  int    uni, null;
  int    npix[2];
  int    finfo[4];

  SCKGETC("IFRM",1,NLEN,&len,Imgname);      /* Image to determine bad pix on */
  if (Imgname[0] == '?'){
    i=-1;
    sprintf(Mes,"PROGRAM: %s -- %s",Prgnam,Version);
    SCTPUT(Mes);
    SCKWRI("OUTPUTI",&i,1,1,&uni);          /* return nbad */
    return(-1);
  }
  SCKGETC("MASK",1,NLEN,&len,Mskname);      /* mask image */
  SCKRDI("HGH",1,1,&len,&Hgh,&uni,&null);   /* upper filter limit */
  SCKRDI("HGHFLG",1,1,&len,&Hghflg,&uni,&null);
  SCKRDI("NPIXMAX",1,1,&len,&Mpix,&uni,&null);
  Mpix--;
  SCKRDI("IMSEC",1,4,&len,Imsc,&uni,&null); /* image section */
  Imsc[0]--; Imsc[1]--; Imsc[2]--; Imsc[3]--; /* work with indices */
  SCKRDI("MESS_LEVEL",1,1,&len,&Mlev,&uni,&null); /* message level */


  SCFINF(Mskname,4,&finfo[0]);              /* get file information(MASK) */
  DatformM = finfo[1];                      /* data format */
  SCFINF(Imgname,4,&finfo[0]);              /* get file information(IFRM) */
  DatformI = finfo[1];                      /* data format */

  if (DatformI != DatformM) {               /* have to have same datatype ! */
    warn(-2);
    return(1);
  }


  SCFOPN(Imgname,DatformI,0,F_IMA_TYPE,&Ifrm); /* open input frame */
  SCDRDI(Ifrm,"NPIX",1,2,&len,npix,&uni,&null);
  NMx = npix[0];
  NMy = npix[1];
  Num = npix[0]*npix[1];

  SCFOPN(Mskname,DatformM,0,F_IMA_TYPE,&Imsk); /* open mask frame */
  SCDRDI(Imsk,"NPIX",1,2,&len,npix,&uni,&null);

  if ((npix[0]!=NMx) || (npix[1]!=NMy)) {
    warn(-3);
    return(1);
  }


  switch(DatformI)
    {
    case D_R4_FORMAT:
      Frmf = (float*) malloc((size_t) Num*SF);
      if (Frmf == NULL) {
        no_memory("Frmf",Num*SF);
        return(1);
      }       
      ptr = (char*) Frmf;
      SCFGET(Ifrm,1,Num,&len,ptr);             /* read the whole frame */
      MFrmf = map2mat(Frmf,NMx,NMy,0,0);       /* map data for matrix access */

      Mskf = (float*) malloc((size_t) Num*SF);
      if (Mskf == NULL) {
        no_memory("Mskf",Num*SF);
        return(1);
      }       
      ptr = (char*) Mskf;
      SCFGET(Imsk,1,Num,&len,ptr);
      MMskf = map2mat(Mskf,NMx,NMy,0,0);     /* mat. start at index [0,0] */
      break;

    case D_UI2_FORMAT:
      Frmu = (unsigned short*) malloc((size_t) Num*SU);
      if (Frmu == NULL) {
        no_memory("Frmu",Num*SU);
        return(1);
      }       
      ptr = (char*) Frmu;
      SCFGET(Ifrm,1,Num,&len,ptr);             /* read the whole frame */
      MFrmu = usmap2mat(Frmu,NMx,NMy,0,0);     /* map data for matrix access */

      Msku = (unsigned short *) malloc((size_t) Num*SU);
      if (Msku == NULL) {
        no_memory("Msku",Num*SU);
        return(1);
      }       
      ptr = (char*) Msku;
      SCFGET(Imsk,1,Num,&len,ptr);
      MMsku = usmap2mat(Msku,NMx,NMy,0,0);     /* mat. start at index [0,0] */
      break;

    default:
      warn(-4);                                /* un-supported format, exit */
      return(1);
    }
  SCFCLO(Ifrm);

  return(0);
}


int findbadpixel(int low, int hgh, int mpix, int *nbad)
/* 
   Find bad pixel. A bad pixel is defined as a pixel having an intensity which
   lays outside the data interval: [low*mean,hgh*mean].
   In case a flatfield had been obtained using the 4-port readout this will be
   recognized and all comparisons will be made within each CCD-quadrant. That
   is necessary due to different biases of each quadrant. BUT usually the
   bad pixel determination should be done with 1-port flatfields.
*/

{
  char  *ptr;

  *nbad = 0;
  switch(DatformI) {
  case D_R4_FORMAT:
    *nbad = findpix(MFrmf,MMskf,Imsc[0],Imsc[2],Imsc[1],Imsc[3],low,hgh,mpix);
    if (*nbad>0) {
      ptr = (char*) Mskf;
      SCFPUT(Imsk,1,Num,ptr);                   /* store in MASK frame */
      SCFCLO(Imsk);
    } else {
      nobadpix();
      return(-1);
    }
    break;

  case D_UI2_FORMAT:
    *nbad = ufindpix(MFrmu,MMsku,Imsc[0],Imsc[2],Imsc[1],Imsc[3],low,hgh,mpix);
    if (*nbad>0) {
      ptr = (char*) Msku;
      SCFPUT(Imsk,1,Num,ptr);                    /* store in MASK frame */
      SCFCLO(Imsk);
    } else {
      nobadpix();
      return(-1);
    }
    break;
  }

  if (*nbad == 0) {                              /* no bad pixels found --> exit */
    nobadpix();
    return(-1);
  }
  return(0);
}


/* === subroutines doing the main job ====================================== */

int findpix(float **img, float **msk, int ncl, int nch, int nrl, int nrh,
                        int low, int hgh, int mpix)
{
  int   i, j, nbad, num;

  num = (nrh-nrl+1)*(nch-ncl+1);
  if (mpix >= 0) num = mpix;                   /* max.allow. bad pixel */

  nbad = 0;
  for (j=nrl;j<=nrh;j++)
    for (i=ncl;i<=nch;i++){
      if (img[j][i] >= hgh){
        msk[j][i] = Hghflg;
        nbad++;
        if (nbad > num) return(nbad);
      }else
        msk[j][i] = 0;
    }
  return(nbad);
}


int ufindpix(unsigned short **img, unsigned short **msk, int ncl, int nch,
                         int nrl, int nrh, int low, int hgh, int mpix)
{
  int   i, j, nbad, num;

  if (hgh > 65534) hgh = 65535;
  num = (nrh-nrl+1)*(nch-ncl+1);
  if (mpix >= 0) num = mpix;                 /* max. allow. bad pixel */

  nbad = 0;
  for (j=nrl;j<=nrh;j++)
    for (i=ncl;i<=nch;i++){
      if (img[j][i] >= hgh){
        msk[j][i] = Hghflg;
        nbad++;
        if (nbad > num) return(nbad);
      }else
        msk[j][i] = 0;
    }
  return(nbad);
}


/* === utilities =========================================================== */
void no_memory(char *svar, int num)
{
  int nbad, uni;

  nbad = -1;
  SCKWRI("OUTPUTI",&nbad,1,1,&uni);                  /* return nbad */
  SCETER(-1,"*** memory-error");
  return;
}


void nobadpix()
{
  int nbad, uni;

  nbad = 0;
  SCKWRI("OUTPUTI",&nbad,1,1,&uni);                  /* return nbad */
  return;
}



void warn(int i)
{
  if (i>0) {
    strncpy(Mes,"text",MLEN);
    messout(Mlev,2,Prgnam,Mes);
  } else {
    switch(-i) {
    case 2:
      strncpy(Mes,"Input frames must have same data type !",MLEN);
      messout(Mlev,1,Prgnam,Mes);
      SCETER(-2,"*** input-datatype-error");
      break;

    case 3:
      strncpy(Mes,"Input frames must have same dimensions !",MLEN);
      messout(Mlev,1,Prgnam,Mes);
      SCETER(-3,"*** dimension-error");
      break;

    case 4: sprintf(Mes,"Dataformat %d not supported !",DatformI);
      messout(Mlev,1,Prgnam,Mes);
      SCETER(-4,"*** dataformat-error");
      break;
    }
  }
  return;
}

