/* $Id: fors_rebima.c,v 1.1 1999/03/04 12:44:46 rpalsa Exp $   
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 *.COPYRIGHT   (C) 1999 European Southern Observatory
 *
 *.IDENT       $RCSfile: fors_rebima.c,v $
 *
 *.KEYWORDS    Linear rebinning, flux conservation, images
 *
 *.USAGE       run fors_rebima.exe
 *
 *.INPUT       INFRM      Input table to be rebinned. The input grid
 *                        specification is read from the descriptors
 *                        START, STEP and NPIX.
 *             REFGRD     Reference for the new data grid. This may be a list
 *                        of values interpreted as start, step and number of
 *                        bins or a reference frame with the descriptors
 *                        START, STEP and NPIX which are used to create the
 *                        new grid.
 *             METHOD     Rebinning method.
 *
 *.OUTPUT      OUTFRM     Rebinned output frame.
 *
 *.RETURNS     The program returns 0 on success or 1 on error. The exit
 *             status is stored in keyword Q1.
 *
 *.PURPOSE     Rebin the binned 1d data set given by INFRM to the new grid
 *             provided by REFGRD. The rebinned data set is written to OUTFRM
 *             which is overwritten if it exists. The rebinning conserves the
 *             total flux of the input data set.  
 *             
 *.ENVIRON     MIDAS 98NOV, FORS pipeline
 *
 *.LANGUAGE    ANSI C/KR C
 *
 *.AUTHOR      R. Palsa, ESO-DMO
 *
 *.VERSION     0.1 1999/02/28: Creation, RP
 *
 *----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <midas_def.h>
#include <tblsys.h>
#include <tblerr.h>
#include <rebin.h>
#include <messages.h>



#define PROID_STRNG      "RebinIma"
#ifndef __STDC__
#define PROID_STRNG_LEN  8
#endif

#define MSG_BUFFER_SIZE     256
#define MID_PARAMETER_SIZE  100
#define FRM_MAX_AXIS  1



typedef struct frame {
  int naxis;
  int ndata;
  int npix[FRM_MAX_AXIS];
  double start[FRM_MAX_AXIS];
  double step[FRM_MAX_AXIS];
  float *data;
} Frame;



#ifdef __STDC__
void error(const char *name, int msglev, const char *msg)
#else
void error(name, msglev, msg)
  char *name, *msg;
  int msglev;
#endif
{

  int unit;


  if (msg)
    messout(msglev, 1, name, msg);

  SCKWRC("Q1", 1, "1", 1, 60, &unit);
  SCSEPI();
}


#ifdef __STDC__
void bdinit(BinData *bd)
#else
void bdinit(bd)
  BinData *bd;
#endif
{

  bd->bval = NULL;
  bd->bwidth = NULL;
  bd->data = NULL;

  return;
}


#ifdef __STDC__
void bdfree(BinData *bd)
#else
void bdfree(bd)
  BinData *bd;
#endif
{

  if (bd->bval)
    osmmfree(bd->bval);

  if (bd->bwidth)
    osmmfree(bd->bwidth);

  if (bd->data)
    osmmfree(bd->data);

  return;
}



void main() {

#ifdef __STDC__
  const char proId[] = PROID_STRNG;
#endif


  int    i, nval, unit, null, idum;
  int    msglev, state;
  int    f1id, f1info[3];
  int    f2id, f3id;
  int    naxis;

  double start, step;

  char   msg[MSG_BUFFER_SIZE];
  char   f1name[FNAME_LEN + 1], f2name[FNAME_LEN + 1];
  char   refgrd[MID_PARAMETER_SIZE + 1];
  char   method[MID_PARAMETER_SIZE + 1];

  BinData iset, oset;

  Method  mtd;



#ifndef __STDC__
  char proId[PROID_STRNG_LEN + 1];
  strcpy(proId, PROID_STRNG);
#endif


  /*
   * Connect to MIDAS
   */

  SCSPRO((char *)proId);


  /*
   * Get keywords from environment
   */

  SCKRDI("MESS_LEVEL", 1, 1, &nval, &msglev, &unit, &null);

  SCKGETC("INFRM", 1, FNAME_LEN, &nval, f1name);
  SCKGETC("OUTFRM", 1, FNAME_LEN, &nval, f2name);
  SCKGETC("REFGRD", 1, MID_PARAMETER_SIZE, &nval, refgrd);
  SCKGETC("METHOD", 1, MID_PARAMETER_SIZE, &nval, method);

  if (strncmp(method, "LIN", 3) == 0)
    mtd = reblin;
  else
    mtd = rebpix;


  /*
   * Read input frame
   */

  if (SCFINF(f1name, 99, NULL)) {
    sprintf(msg, "Input frame >%s< does not exist!", f1name);
    error(proId, msglev, msg);
  }
  else {
    if (SCFOPN(f1name, D_R8_FORMAT, 0, F_IMA_TYPE, &f1id)) {
      sprintf(msg, "Cannot open input frame >%s<!", f1name);
      error(proId, msglev, msg);
    }

    SCFINF(f1name, 4, f1info);

    SCDRDI(f1id, "NAXIS", 1, 1, &nval, &naxis, &unit, &null);
    if (naxis <= 0 || naxis > FRM_MAX_AXIS)
      error(proId, msglev, "Input frame is not a 1d image!");
  }


  /*
   * Get input data grid from frame header.
   */

  bdinit(&iset);

  SCDRDI(f1id, "NPIX", 1, 1, &nval, &iset.nbin, &unit, &null);
  SCDRDD(f1id, "START", 1, 1, &nval, &start, &unit, &null);
  SCDRDD(f1id, "STEP", 1, 1, &nval, &step, &unit, &null);

  if (step < 0.)
      error(proId, msglev, "Invalid input grid specification!");

  if ((iset.bval = (double *)osmmget(iset.nbin * sizeof(double))) == NULL)
    error(proId, msglev, "Not enough memory!");

  if ((iset.bwidth = (double *)osmmget(iset.nbin * sizeof(double))) == NULL) {
    bdfree(&iset);
    error(proId, msglev, "Not enough memory!");
  }

  if ((iset.data = (double *)osmmget(iset.nbin * sizeof(double))) == NULL) {
    bdfree(&iset);
    error(proId, msglev, "Not enough memory!");
  }
  else 
    if (SCFGET(f1id, 1, iset.nbin, &nval, (char *)iset.data))
      error(proId, msglev, "Cannot read input frame into memory!");

  for (i = 0; i < iset.nbin; i++) {
    iset.bval[i] = start + i * step;
    iset.bwidth[i] = step;
  }


  /*
   * Decode output grid specification
   */

  if (SCFINF(refgrd, 99, NULL)) {
    if (!isdigit(*refgrd)) {
      sprintf(msg, "Reference frame >%s< does not exist!", refgrd);
      error(proId, msglev, msg);
    }
    else 
      if (sscanf(refgrd,"%lf,%lf,%d", &start, &step, &oset.nbin) < 3) {
	bdfree(&iset);
	error(proId, msglev, "Invalid grid specification!");
      }
  }
  else {
    if (SCFOPN(refgrd, D_R8_FORMAT, 0, F_IMA_TYPE, &f3id)) {
      sprintf(msg, "Cannot open input frame >%s<!", refgrd);
      error(proId, msglev, msg);
    }

    SCDRDI(f3id, "NPIX", 1, 1, &nval, &oset.nbin, &unit, &null);
    SCDRDD(f3id, "START", 1, 1, &nval, &start, &unit, &null);
    SCDRDD(f3id, "STEP", 1, 1, &nval, &step, &unit, &null);

    SCFCLO(f3id);
  }

  bdinit(&oset);

  if ((oset.bval = (double *)osmmget(oset.nbin * sizeof(double))) == NULL) {
    bdfree(&iset);
    error(proId, msglev, "Not enough memory!");
  }

  if ((oset.bwidth = (double *)osmmget(oset.nbin * sizeof(double))) == NULL) {
    bdfree(&iset);
    bdfree(&oset);
    error(proId, msglev, "Not enough memory!");
  }

  for (i = 0; i < oset.nbin; i++) {
    oset.bval[i] = start + i * step; 
    oset.bwidth[i] = step; 
  }


  /*
   * Rebin input data set
   */

  if ((oset.data = (double *)osmmget(oset.nbin * sizeof(double))) == NULL) {
    bdfree(&iset);
    bdfree(&oset);
    error(proId, msglev, "Not enough memory!");
  }

  if ((state = rebin(&iset, &oset, mtd))) {
    bdfree(&iset);
    bdfree(&oset);

    switch (state) {
    case -1:
      error(proId, msglev, "Data grids do not overlap!");
      break;

    case -2:
      error(proId, msglev, "Bin values are not in ascending order!");
      break;

    case -3:
      error(proId, msglev, "Non-continuous data set encountered!");
      break;
    default:
      error(proId, msglev, NULL);
      break;
    }
  }
  else
    bdfree(&iset);


  /*
   * Write rebinned data to output frame. The repeated opening and 
   * closing of the output image below is necessary so that MIDAS
   * is able to perform the appropriate type conversions (It's not
   * a bug, it's a feature!).
   */

  if (SCFCRE(f2name, f1info[1], F_O_MODE, F_IMA_TYPE, oset.nbin, &f2id)) {
    bdfree(&oset);
    sprintf(msg, "Cannot create output frame >%s<!", f2name);
    error(proId, msglev, msg);
  }

  SCFCLO(f2id);
  
  if (SCFOPN(f2name, D_R8_FORMAT, 0, F_IMA_TYPE, &f2id)) {
    bdfree(&oset);
    sprintf(msg, "Cannot open output frame >%s<!", f2name);
    error(proId, msglev, msg);
  }

  SCDWRI(f2id, "NAXIS", &naxis, 1, 1, &unit);
  SCDWRD(f2id, "START", &start, 1, 1, &unit);
  SCDWRD(f2id, "STEP", &step, 1, 1, &unit);
  SCDWRI(f2id, "NPIX", &oset.nbin, 1, 1, &unit);
  SCDWRC(f2id, "IDENT", 1, " ", 1, 72, &unit);

  SCFPUT(f2id, 1, oset.nbin, (char *)oset.data);
  
  SCDCOP(f1id, f2id, 4, "LHCUTS");
  SCDCOP(f1id, f2id, 4, "CUNIT");


  /*
   * Cleanup
   */
      
  SCFCLO(f1id);
  SCFCLO(f2id);
  bdfree(&oset);


  /*
   * Disconnect from MIDAS
   */

  SCKWRC("Q1", 1, "0", 1, 60, &unit);
  SCSEPI();

}
