/* $Id: fors_rebtbl.c,v 1.1 1999/03/04 12:43:21 rpalsa Exp $   
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 *.COPYRIGHT   (C) 1999 European Southern Observatory
 *
 *.IDENT       $RCSfile: fors_rebtbl.c,v $
 *
 *.KEYWORDS    Linear rebinning, flux conservation, tables
 *
 *.USAGE       run fors_rebtbl.exe
 *
 *.INPUT       INTBL      Input table to be rebinned.
 *             INGRD      Input grid specification, i.e column references
 *                        for independent variable, dependent variable and
 *                        bin size.
 *             REFGRD     Reference for the new data grid. This may be a list
 *                        of values interpreted as start, step and number of
 *                        bins or a list of column references for the 
 *                        independent and dependent variable and the new bin
 *                        size.
 *             METHOD     Rebinning method.
 *             REFTBL     Reference table name to look up the column 
 *                        references given in REFGRD.
 *
 *.OUTPUT      OUTTBL     Rebinned output table.
 *
 *.RETURNS     The program returns 0 on success or 1 on error. The exit
 *             status is stored in keyword Q1.
 *
 *.PURPOSE     Rebin the binned 1d data set specified by INTBL and INGRD to
 *             the new grid referenced by REFGRD. The rebinned data set is
 *             written to OUTTBL which is overwritten if it exists. The
 *             rebinning conserves the total flux of the input data set. 
 *             
 *.ENVIRON     MIDAS 98NOV, FORS pipeline
 *
 *.LANGUAGE    ANSI C/KR C
 *
 *.AUTHOR      R. Palsa, ESO-DMO
 *
 *.VERSION     0.1 1999/02/28: Creation, RP
 *
 *----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <midas_def.h>
#include <tblsys.h>
#include <tblerr.h>
#include <rebin.h>
#include <messages.h>



#define PROID_STRNG      "RebinTbl"
#ifndef __STDC__
#define PROID_STRNG_LEN  8
#endif

#define MSG_BUFFER_SIZE     256
#define MID_PARAMETER_SIZE  100



#ifdef __STDC__
void error(const char *name, int msglev, const char *msg)
#else
void error(name, msglev, msg)
  char *name, *msg;
  int msglev;
#endif
{

  int unit;


  if (msg)
    messout(msglev, 1, name, msg);

  SCKWRC("Q1", 1, "1", 1, 60, &unit);
  SCSEPI();
}


#ifdef __STDC__
void bdinit(BinData *bd)
#else
void bdinit(bd)
  BinData *bd;
#endif
{

  bd->bval = NULL;
  bd->bwidth = NULL;
  bd->data = NULL;

  return;
}


#ifdef __STDC__
void bdfree(BinData *bd)
#else
void bdfree(bd)
  BinData *bd;
#endif
{

  if (bd->bval)
    osmmfree(bd->bval);

  if (bd->bwidth)
    osmmfree(bd->bwidth);

  if (bd->data)
    osmmfree(bd->data);

  return;
}



void main() {

#ifdef __STDC__
  const char proId[] = PROID_STRNG;
#endif


  int    i, nval, unit, null, idum;
  int    msglev, state;
  int    t1id, t1col[3];
  int    t2id, t2col[3], t2type;
  int    t3id, t3col[3], t3flag;

  double start, step;

  char   *s;
  char   msg[MSG_BUFFER_SIZE];
  char   t1name[TBL_NAMLEN + 1];
  char   t2name[TBL_NAMLEN + 1], t2label[MID_PARAMETER_SIZE + 1];
  char   t2form[TBL_FORLEN + 1], t2unit[TBL_UNILEN + 1];
  char   t3name[TBL_NAMLEN + 1];
  char   ingrd[MID_PARAMETER_SIZE + 1], refgrd[MID_PARAMETER_SIZE + 1];
  char   method[MID_PARAMETER_SIZE + 1];

  BinData iset, oset;

  Method  mtd;



#ifndef __STDC__
  char proId[PROID_STRNG_LEN + 1];
  strcpy(proId, PROID_STRNG);
#endif


  /*
   * Connect to MIDAS
   */

  SCSPRO((char *)proId);


  /*
   * Get keywords from environment
   */

  SCKRDI("MESS_LEVEL", 1, 1, &nval, &msglev, &unit, &null);

  SCKGETC("INTBL", 1, TBL_NAMLEN, &nval, t1name);
  SCKGETC("OUTTBL", 1, TBL_NAMLEN, &nval, t2name);
  SCKGETC("INGRD", 1, MID_PARAMETER_SIZE, &nval, ingrd);
  SCKGETC("REFGRD", 1, MID_PARAMETER_SIZE, &nval, refgrd);

  if (*refgrd == ':') {
    SCKGETC("REFTBL", 1, TBL_NAMLEN, &nval, t3name);
    t3flag = 1;
  }
  else 
    t3flag = 0;


  SCKGETC("METHOD", 1, MID_PARAMETER_SIZE, &nval, method);
  if (strncmp(method, "LIN", 3) == 0)
    mtd = reblin;
  else
    mtd = rebpix;


  /*
   * Read input table data
   */

  if (SCFINF(t1name, 99, NULL)) {
    sprintf(msg, "Input table >%s< does not exist!", t1name);
    error(proId, msglev, msg);
  }
  else 
    if (TCTOPN(t1name, F_I_MODE, &t1id)) {
      sprintf(msg, "Cannot open input table >%s<!", t1name);
      error(proId, msglev, msg);
    }


  /*
   * Decode input grid specification and save column labels for
   * output.
   */

  strcpy(t2label, ingrd);

  if (!(s = strtok(ingrd, ",")))
    error(proId, msglev, "Invalid input grid specification!");
  else {
    i = 0;

    while (s && i < 3) {
      TCCSER(t1id, s, t1col + i);
      if (t1col[i] < 0) {
	sprintf(msg, "Column >%s< not found in %s!", s, t1name);
	error(proId, msglev, msg);
      }

      s = strtok(NULL, ",");
      i++;
    }

    if (i != 3)
      error(proId, msglev, "Invalid input grid specification!");
  }

  TCIGET(t1id, &idum, &iset.nbin, &idum, &idum, &idum);

  bdinit(&iset);

  if ((iset.bval = (double *)osmmget(iset.nbin * sizeof(double))) == NULL)
    error(proId, msglev, "Not enough memory!");

  if ((iset.bwidth = (double *)osmmget(iset.nbin * sizeof(double))) == NULL) {
    bdfree(&iset);
    error(proId, msglev, "Not enough memory!");
  }

  if ((iset.data = (double *)osmmget(iset.nbin * sizeof(double))) == NULL) {
    bdfree(&iset);
    error(proId, msglev, "Not enough memory!");
  }

  for (i = 0; i < iset.nbin; i++) {
    TCERDD(t1id, i + 1, t1col[0], iset.bval + i, &null);
    TCERDD(t1id, i + 1, t1col[1], iset.data + i, &null);
    TCERDD(t1id, i + 1, t1col[2], iset.bwidth + i, &null);
  }


  /*
   * Decode output grid specification
   */

  if (t3flag) 
    if (SCFINF(t3name, 99, NULL)) {
      sprintf(msg, "Reference table >%s< does not exist!", t3name);
      error(proId, msglev, msg);
    }
    else {
      if (TCTOPN(t3name, F_I_MODE, &t3id)) {
	sprintf(msg, "Cannot open reference table >%s<!", t3name);
	error(proId, msglev, msg);
      }
      else 
	if (!(s = strtok(refgrd, ",")))
	  error(proId, msglev, "Invalid reference grid specification!");
	else {
	  i = 0;

	  while (s && i < 3) {
	    TCCSER(t3id, s, t3col + i);
	    if (t3col[i] < 0) {
	      sprintf(msg, "Column >%s< not found in %s!", s, t3name);
	      error(proId, msglev, msg);
	    }

	    s = strtok(NULL, ",");
	    i++;
	  }

	  if (i != 3)
	    error(proId, msglev, "Invalid input grid specification!");

	  TCIGET(t3id, &idum, &oset.nbin, &idum, &idum, &idum);
	}
    }
  else {
    if (sscanf(refgrd,"%lf,%lf,%d", &start, &step, &oset.nbin) < 3) {
      bdfree(&iset);
      error(proId, msglev, "Invalid grid specification!");
    }
  }

  bdinit(&oset);

  if ((oset.bval = (double *)osmmget(oset.nbin * sizeof(double))) == NULL) {
    bdfree(&iset);
    error(proId, msglev, "Not enough memory!");
  }

  if ((oset.bwidth = (double *)osmmget(oset.nbin * sizeof(double))) == NULL) {
    bdfree(&iset);
    bdfree(&oset);
    error(proId, msglev, "Not enough memory!");
  }

  if (t3flag) {
    for (i = 0; i < oset.nbin; i++) {
      TCERDD(t3id, i + 1, t3col[0], oset.bval + i, &null);
      TCERDD(t3id, i + 1, t3col[2], oset.bwidth + i, &null);
    }

    TCTCLO(t3id);
  }
  else 
    for (i = 0; i < oset.nbin; i++) {
      oset.bval[i] = start + i * step; 
      oset.bwidth[i] = step; 
    }


  /*
   * Rebin input data set
   */

  if ((oset.data = (double *)osmmget(oset.nbin * sizeof(double))) == NULL) {
    bdfree(&iset);
    bdfree(&oset);
    error(proId, msglev, "Not enough memory!");
  }

  if ((state = rebin(&iset, &oset, mtd))) {
    bdfree(&iset);
    bdfree(&oset);

    switch (state) {
    case -1:
      error(proId, msglev, "Data grids do not overlap!");
      break;

    case -2:
      error(proId, msglev, "Bin values are not in ascending order!");
      break;

    case -3:
      error(proId, msglev, "Non-continuous data set encountered!");
      break;
    default:
      error(proId, msglev, NULL);
      break;
    }
  }
  else
    bdfree(&iset);


  /*
   * Write rebinned data to output table
   */

  if (TCTINI(t2name, F_TRANS, F_O_MODE, 3, oset.nbin, &t2id)) {
    bdfree(&oset);
    sprintf(msg, "Cannot create output table >%s<!", t2name);
    error(proId, msglev, msg);
  }

  i = 0;
  s = strtok(t2label, ",");

  while (s) {
    TCFGET(t1id, t1col[i], t2form, &idum, &t2type);
    TCUGET(t1id, t1col[i], t2unit);

    if (TCCINI(t2id, t2type, 1, t2form, t2unit, s, t2col + i)) {
      bdfree(&oset);
      sprintf(msg, "Cannot create output column >%s<!", s);
      error(proId, msglev, msg);
    }

    s = strtok(NULL, ",");
    i++;
  }

  for (i = 0; i < oset.nbin; i++) {
    TCEWRD(t2id, i + 1, t2col[0], oset.bval + i);
    TCEWRD(t2id, i + 1, t2col[1], oset.data + i);
    TCEWRD(t2id, i + 1, t2col[2], oset.bwidth + i);
  }


  /*
   * Cleanup
   */
      
  TCTCLO(t1id);
  TCTCLO(t2id);
  bdfree(&oset);


  /*
   * Disconnect from MIDAS
   */

  SCKWRC("Q1", 1, "0", 1, 60, &unit);
  SCSEPI();

}
