/* $Id: fors_remout.c,v 1.13 2000/05/15 19:31:50 nrainer Exp $    */
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

 .COPYRIGHT  (C) 1998  European Southern Observatory (ESO)

 .IDENT      fors_remout.c

 .KEYWORDS   Statistics

 .PURPOSE    Removing of pre- and overscan regions and filtering of out-
             liers in MIDAS BDF frames; 1dim storing of filtered frames.
 __________

 .USAGE      run PIPE_EXE:(IDENT) (invoked from fors_remout.prg)

 .INPUT      IN_A     : input frame

             INPUTI(1): number of prescan columns,
             INPUTI(2): number of overscan columns

             INPUTR(1): lower threshold (ADU) for outliers,
             INPUTR(2): upper threshold (ADU) for outliers -
                          pixels outside of the interval are ignored

 .OUTPUT     OUT_A     : Name of filtered 1D frame

             OUTPUTI(1): Number of all pixels without pre-/overscan
             OUTPUTI(2): Number of outliers
             OUTPUTI(3): Number of left pixels

             OUTPUTR(1): Percentage of outliers

 .RETURN     Q1:  0: successful return
                 -1: a warning message occured
                  1: an error occured
 __________

 .ALGORITHM  -

 .COMMENT    -

 .ENVIRON    MIDAS 98NOV

 .LANGUAGE   C (ANSI)

 .AUTHOR     N.Rainer  DPG/DMD/ESO

 .VERSION    1.0  98/09/21  Creation
                  98/12/15  Release
             1.1  99/01/14  Release
             1.2  99/03/15  Release
             1.3  99/09/23  Release

----------------------------------------------------------------------------*/


/*  PROTOTYPES  :
    **********/

#ifdef __STDC__  /* ANSI */
  static void  error(void);
#else  /* K&R */
  static void  error();
#endif


/*  INCLUDED FILES  :
    **************/

#include  <midas_def.h>
/* ONLY FOR DEBUGGING WITH `cflow': */
/* #include  "/home/dfs1rp1/dmd_df/nrainer/LAN/MIDAS/C/CFLOW/midas_def.h" */

#include  <ctype.h>
#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>


/*  DEFINITIONS OF CONSTANTS AND MACROS  :
    ***********************************/

/* main  :
   ---*/
#define       PROID      "remout.c"
#define       NPIXMIN    20  /* FEWER PIXELS NOT REASONABLE FOR STATISTICS */


/*  GLOBAL VARIABLES  :
    ****************/

static int  messlevel;


/******************************************************************************/
/****************************************************************************/


#ifdef __STDC__  /* ANSI */
  void main(int argc, char *argv[])
#else  /* K&R */
  void main(argc, argv)
  int  argc;
  char *argv[];
#endif
{

/*  LOCAL VARIABLES  :
    ***************/

char              fileimaorig[65];
char              *char0, *char1, fileimafilt[72];
char              tmps[160];

double            start, step;

float             *imaorig, *imafilt;
float             *imaorig0, *imafilt0;
float             thresh[2], pixrem;
register float    imaorigi;

unsigned long     i, j; 
unsigned long     pixorig, pixfilt, pixout;

int               r, nval, unit, null, info, slen;
int               warn = 0;
int               scan[2];
int               naxisimaorig, npiximaorig[2];
int               idimaorig, idimafilt;
int               naxis, npix;
int               ipixorig, ipixfilt, ipixout;


SCSPRO("remout");


if (argc > 1 && *argv[1] == '-' && tolower(*(argv[1]+1)) == 'h') {
  SCTPUT("      ******************************************************************");
  SCTPUT("      IN_A     : input frame");
  SCTPUT(" ");
  SCTPUT("      INPUTI(1): number of prescan columns,");
  SCTPUT("      INPUTI(2): number of overscan columns");
  SCTPUT(" ");
  SCTPUT("      INPUTR(1): lower threshold (ADU) for outliers,");
  SCTPUT("      INPUTR(2): upper threshold (ADU) for outliers -");
  SCTPUT("                   pixels outside of the interval are ignored");
  SCTPUT("      ******************************************************************");
  SCSEPI();
}


/*  INPUT  :
    *****/

r = SCKRDI("MESS_LEVEL",1,1,&nval,&messlevel,&unit,&null);

r = SCKGETC("IN_A",1,64,&nval,fileimaorig);
r = SCFINF(fileimaorig,99,&info);
if (r != 0) {
  SCTPUT(" ");
  messout(messlevel,1,PROID,"Frame not found or can not be opened !");
  sprintf(&tmps[0],"(%s)",fileimaorig);
  messout(messlevel,1,PROID,tmps);
  error();
}

r = SCKRDI("INPUTI",1,2,&nval,scan,&unit,&null);
if (scan[0] < 0 || scan[1] < 0) {
  SCTPUT(" ");
  messout(messlevel,1,PROID,"Impossible pre- or overscan value !");
  error();
}

r = SCKRDR("INPUTR",1,2,&nval,thresh,&unit,&null);
if (thresh[0] >= thresh[1]) {
  SCTPUT(" ");
  messout(messlevel,1,PROID,"Impossible thresh. f. filt. outl. !");
  error();
}


/*  READING PIXEL VALUES  :
    ********************/

messout(messlevel,4,PROID,"      reading image data");

r = SCFOPN(fileimaorig,D_R4_FORMAT,0,F_IMA_TYPE,&idimaorig);

r = SCDRDI(idimaorig,"NAXIS",1,1,&nval,&naxisimaorig,&unit,&null);

if (naxisimaorig == 2) {
  r = SCDRDI(idimaorig,"NPIX",1,2,&nval,npiximaorig,&unit,&null);
  pixorig = npiximaorig[0]*npiximaorig[1];
} else if (naxisimaorig == 1) {
  r = SCDRDI(idimaorig,"NPIX",1,1,&nval,npiximaorig,&unit,&null);
  pixorig = npiximaorig[0];
  npiximaorig[1] = 1;  /* for removing of pre-/overscan below */
} else {
  SCTPUT(" ");
  sprintf(&tmps[0],
   "Frame dimension is not supported !");
  messout(messlevel,1,PROID,tmps);
  sprintf(&tmps[0],"(%s has %d axes)",fileimaorig,naxisimaorig);
  messout(messlevel,1,PROID,tmps);
  error();
}

if (pixorig < NPIXMIN) {
  SCTPUT(" ");
  sprintf(&tmps[0],"Too few pixels (at least %d necessary) !",NPIXMIN);
  messout(messlevel,1,PROID,tmps);
  sprintf(&tmps[0],"(%s)",fileimaorig);
  messout(messlevel,1,PROID,tmps);
  error();
}

imaorig = (float *) malloc ((unsigned int)(pixorig*sizeof(float)));
if (imaorig == 0) {
  SCTPUT(" ");
  messout(messlevel,1,PROID,"Allocation for frame not possible !");
  sprintf(&tmps[0],"(%s)",fileimaorig);
  messout(messlevel,1,PROID,tmps);
  error();
}

r = SCFGET(idimaorig,1,pixorig,&nval,(char *)imaorig);


if (scan[0] != 0 || scan[1] != 0) {

  /*  REMOVING PRE- AND OVERSCAN REGIONS IN X-DIRECTION :
      ************************************************/

  messout(messlevel,4,PROID,"      removing pre- and overscan regions");

  if (npiximaorig[0] <= scan[0]+scan[1]) {
    SCTPUT(" ");
    messout(messlevel,1,PROID,"(pre- + overscan) >= (#pixels in x-axis) !");
    sprintf(&tmps[0],"(%s)",fileimaorig);
    messout(messlevel,1,PROID,tmps);
    error();
  }

  imaorig0 = imaorig;
  for (j=0; j<npiximaorig[1]; j++) {
    for (i=0; scan[0]+i<npiximaorig[0]-scan[1]; i++)
      *imaorig++ = *(imaorig0+j*npiximaorig[0]+scan[0]+i);
  }
  imaorig = imaorig0;
  pixorig = npiximaorig[1]*(npiximaorig[0]-scan[0]-scan[1]);

  if (pixorig < NPIXMIN) {
    SCTPUT(" ");
    messout(messlevel,1,PROID,
            "Too few pixels left after removing pre-/overscan !");
    sprintf(&tmps[0],"(< %d in %s)",NPIXMIN,fileimaorig);
    messout(messlevel,1,PROID,tmps);
    error();
  }

} /* if (scan[0] != 0 || scan[1] != 0) */


/*  REMOVING OUTLIERS  :
    *****************/

messout(messlevel,4,PROID,"      filtering outliers");

imafilt = imaorig;

imaorig0 = imaorig;
imafilt0 = imafilt;
for (i=0, pixfilt=0; i<pixorig; i++) {
  imaorigi = *imaorig++;
  if (imaorigi >= thresh[0] && imaorigi <= thresh[1]) {
    *imafilt++ = imaorigi;
    pixfilt++;
  }
}
imaorig = imaorig0;
imafilt = imafilt0;

pixout = pixorig-pixfilt;
pixrem = (float)pixout/pixorig*100.;
sprintf(&tmps[0],"        removed :  %ld / %ld = %.4f%%"
        ,pixout,pixorig,pixrem);
messout(messlevel,4,PROID,tmps);
if (pixfilt < NPIXMIN) {
  SCTPUT(" ");
  messout(messlevel,1,PROID,"Too few pixels left after filtering outliers !");
  sprintf(&tmps[0],"(< %d IN %s)",NPIXMIN,fileimaorig);
  messout(messlevel,1,PROID,tmps);
  error();
}

ipixorig = (int)pixorig;
ipixout = (int)pixout;
ipixfilt = (int)pixfilt;
r = SCKWRI("OUTPUTI",&ipixorig,1,1,&unit);
r = SCKWRI("OUTPUTI",&ipixout,2,1,&unit);
r = SCKWRI("OUTPUTI",&ipixfilt,3,1,&unit);
r = SCKWRR("OUTPUTR",&pixrem,1,1,&unit);


  /*  STORING FILTERED DATA  :
      *********************/

messout(messlevel,4,PROID,"      storing (extracted/filtered) data");
messout(messlevel,4,PROID,"      in new 1D-frame :");

  char1 = strrchr(fileimaorig,'.');
  if (char1) {
    char0 = &fileimaorig[0];
    slen = char1-char0;
    strncpy(fileimafilt,fileimaorig,slen);
    fileimafilt[slen] = '\0';
  }
  else {
    strcpy(fileimafilt,fileimaorig);
  }
  strcat(fileimafilt,"_f.bdf");

sprintf(&tmps[0],"        %s",fileimafilt);
messout(messlevel,4,PROID,tmps);

r = SCFCRE(fileimafilt,D_R4_FORMAT,F_O_MODE,F_IMA_TYPE,pixfilt,&idimafilt);

naxis = 1;
npix  = pixfilt;
start = 1.0;
step  = 1.0;
r = SCDWRI(idimafilt,"NAXIS",&naxis,1,1,&unit);
r = SCDWRI(idimafilt,"NPIX",&npix,1,1,&unit);
r = SCDWRD(idimafilt,"START",&start,1,1,&unit);
r = SCDWRD(idimafilt,"STEP",&step,1,1,&unit);
r = SCDWRC(idimafilt,"IDENT",1,fileimafilt,1,72,&unit);
r = SCDWRC(idimafilt,"CUNIT",1,"PIXEL",1,48,&unit);
r = SCDCOP(idimaorig,idimafilt,4,"LHCUTS");
r = SCDCOP(idimaorig,idimafilt,5,"");

r = SCFPUT(idimafilt,1,pixfilt,(char *)imafilt);

r = SCKWRC("OUT_A",1,fileimafilt,1,80,&unit);

r = SCFCLO(idimafilt);
r = SCFCLO(idimaorig);


free(imafilt);


if (warn == 0)
  r = SCKWRC("Q1",1,"0",1,60,&unit);
else
  r = SCKWRC("Q1",1,"-1",1,60,&unit);
r = SCSEPI();


} /* void main(int argc, char *argv[]) */


/******************************************************************************/
/****************************************************************************/

#ifdef __STDC__  /* ANSI */
  static void error(void)
#else  /* K&R */
  static void error()
#endif
{
  int  r, unit;

  r = SCKWRC("Q1",1,"1",1,60,&unit);
  r = SCSEPI();
}

/****************************************************************************/
