/* $Id: fors_robust_sigma.c,v 1.3 2001/06/15 14:21:53 sbogun Exp $
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * COPYRIGHT (c) 2001 European Southern Observatory
 * LICENSE: GNU General Public License version 2 or later
 *
 * PROJECT:   VLT Data Flow System
 * AUTHOR:    Stefan Bogun -- ESO/DMD/DPG
 * SUBSYSTEM: Instrument pipelines
 *
 * PURPOSE:
 *   Calculate a resistant estimate of the dispersion of a MIDAS image.
 *   For an uncontaminated distribution, this is identical to the standard
 *   deviation.
 *
 * USAGE:
 *   RUN PIPE_EXE:fors_robust_sigma.exe 
 *
 * INPUT:
 *   image:    input frame
 *   zero:     if set to 1 the dispersion is calculated w.r.t. 0.0 rather than the
 *             central value of the vector. If Y is a vector of residuals, this
 *             should be set.
 *
 * OUTPUT:
 *   q1:        0: successful return
 *             -1: a warning message occured
 *              1: an error occured
 *
 *   sigma:    contains the dispersion. In case of failure, returns a
 *             value of -1.0
 *
 * DESCRIPTION:
 *   Use the median absolute deviation as the initial estimate, then weight
 *   points using Tukey's Biweight. See, for example, "Understanding Robust
 *   and Exploratory Data Analysis," by Hoaglin, Mosteller and Tukey, John
 *   Wiley & Sons, 1983.
 *
 * $Name: rel1_8_2 $
 * $Revision: 1.3 $
 * ----------------------------------------------------------------------------
 */

#include <midas_def.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>


#define PROID "robust_sigma.c"

static void error(void);

double robust_sigma(double*, int, int);

static int messlevel;

void main(int argc, char *argv[])
{
    int i, r, nval, unit, null, info;
    int fimageid;
    float *fimage;
    double *dimage;
    double sigma;
    int naxis, npix[2], npixel, zero;
    char file[81], tmps[160];
	
    SCSPRO("robust_sigma");
	
    r = SCKRDI("mess_level",1,1,&nval,&messlevel,&unit,&null);
    r = SCKGETC("image",1,80,&nval,file);
    r = SCKRDI("zero",1,1,&nval,&zero,&unit,&null);
	
    r = SCFINF(file,99,&info);
    if (r != 0) {
		messout(messlevel,1,PROID,"Frame not found or can not be opened !");
		sprintf(&tmps[0],"(%s)",file);
		messout(messlevel,1,PROID,tmps);
		error();
    }
	
    r = SCFOPN(file,D_R4_FORMAT,0,F_IMA_TYPE,&fimageid);
	
    r = SCDRDI(fimageid,"NAXIS",1,1,&nval,&naxis,&unit,&null);
    if (naxis == 2) {
		r = SCDRDI(fimageid,"NPIX",1,2,&nval,npix,&unit,&null);
		npixel = npix[0]*npix[1];
    } else if (naxis == 1) {
		r = SCDRDI(fimageid,"NPIX",1,1,&nval,npix,&unit,&null);
		npixel = npix[0];
		npix[1] = 1;
    } else {
		SCTPUT(" ");
		messout(messlevel,1,PROID,"Frame dimension not supported !");
		sprintf(&tmps[0],"(%s has %d axes)!",file,naxis);
		messout(messlevel,1,PROID,tmps);
		error();
    }
    
    fimage = (float *) malloc ((unsigned int)(npixel*sizeof(float)));
    if (fimage == 0) {
		SCTPUT(" ");
		messout(messlevel,1,PROID,"Allocation not possible !");
		sprintf(&tmps[0],"(%s)",file);
		messout(messlevel,1,PROID,tmps);
		error();
    }
    dimage = (double *) malloc ((unsigned int)(npixel*sizeof(double)));
    if (dimage == 0) {
		SCTPUT(" ");
		messout(messlevel,1,PROID,"Allocation not possible !");
		sprintf(&tmps[0],"(%s)",file);
		messout(messlevel,1,PROID,tmps);
		free(fimage);
		error();
    }
 
    r = SCFGET(fimageid,1,npixel,&nval,(char *)fimage);
	
    for (i=0; i<npixel; i++)
		dimage[i] = (double)fimage[i];
	
    /* this is the call of the function robust_sigma */
	
    sigma = robust_sigma(dimage, npixel, zero);
	
    if (sigma < 0) {
		SCTPUT(" ");
		messout(messlevel,2,PROID,"This distribution is TOO WEIRD! Returning -1!");
		sprintf(&tmps[0],"(%s)",file);
		messout(messlevel,2,PROID,tmps);
		r = SCKWRC("Q1",1,"1",1,60,&unit);
    }
    else
		r = SCKWRC("Q1",1,"0",1,60,&unit);
	
    r = SCKWRD("sigma",&sigma,1,1,&unit);
    free(fimage);
    free(dimage);
    r = SCFCLO(fimageid);
    r = SCSEPI();
	
}



static void error(void)
{
	int r, unit;

	r = SCKWRC("Q1",1,"1",1,60,&unit);
	r = SCSEPI();
}


