/* $Id: fors_t3dcpyima.c,v 1.2 1999/03/04 12:41:55 rpalsa Exp $   
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 *.COPYRIGHT   (C) 1998 European Southern Observatory
 *
 *.IDENT       $RCSfile: fors_t3dcpyima.c,v $
 *
 *.KEYWORDS    3d tables, insert a image row as table element array
 *
 *.USAGE       run t3dcpyima.exe
 *
 *.INPUT       SOURCE      Source frame
 *             SROW        Frame row to be copied
 *             CRANGE      Range of frame columns to be copied (i.e. 
 *                         first,last pixel of frame row)
 *             TCOLUMN     Label of target column
 *             TROW        Row number of target table element
 *
 *.OUTPUT      TARGET      3d output table
 *
 *.RETURNS     The program returns 0 on success or 1 on error. The exit
 *             status is stored in keyword Q1.
 *
 *.PURPOSE     Insert a row of the input image into a table element of the
 *             target table, whose elements are one dimensional arrays.
 *
 *.ENVIRON     MIDAS 98NOV, FORS pipeline
 *
 *.LANGUAGE    ANSI C/KR C
 *
 *.AUTHOR      R. Palsa, ESO-DMO
 *
 *.VERSION     0.1 1998/12/18: Creation, RP
 *
 *----------------------------------------------------------------------------
 */

#include <stdio.h>

#ifndef __STDC__
#include <string.h>
#endif

#include <midas_def.h>
#include <tblsys.h>
#include <tblerr.h>
#include <messages.h>



#define PROID_STRNG      "T3dCpyima"
#ifndef __STDC__
#define PROID_STRNG_LEN  9
#endif

#define MSG_BUFFER_SIZE  256
#define RSTR_BUFFER_SIZE 100
#define FRM_MAX_AXIS     2



typedef struct frame {
  int naxis;
  int ndata;
  int npix[FRM_MAX_AXIS];
  double start[FRM_MAX_AXIS];
  double step[FRM_MAX_AXIS];
  float *data;
} Frame;



#ifdef __STDC__
void error(const char *name, int msglev, const char *msg)
#else
void error(name, msglev, msg)
  char *name, *msg;
  int msglev;
#endif
{

  int unit;


  messout(msglev, 1, name, msg);
  SCKWRC("Q1", 1, "1", 1, 60, &unit);
  SCSEPI();
}



void main() {

#ifdef __STDC__
  const char proId[] = PROID_STRNG;
#endif


  int    i, nval, unit, null, idum;
  int    nchar, ncol;
  int    fid, frno, fcrng[2];
  int    t3id, t3cno, t3rno, t3nrow, t3ctype, t3csize, t3cbytes;
  int    msglev;
  int    state = 0;

  double rval;

  char   *t3eptr;
  char   msg[MSG_BUFFER_SIZE];
  char   rstr[RSTR_BUFFER_SIZE + 1];
  char   fname[FNAME_LEN + 1];
  char   t3name[TBL_NAMLEN + 1], t3column[TBL_LABLEN + 1];

  Frame  ifrm;



#ifndef __STDC__
  char proId[PROID_STRNG_LEN + 1];
  strcpy(proId, PROID_STRNG);
#endif


  /*
   * Connect to MIDAS
   */

  SCSPRO((char *)proId);


  /*
   * Get keywords from environment
   */

  SCKRDI("MESS_LEVEL", 1, 1, &nval, &msglev, &unit, &null);

  SCKGETC("SOURCE", 1, FNAME_LEN, &nval, fname);
  SCKGETC("TARGET", 1, TBL_NAMLEN, &nval, t3name);
  SCKGETC("TCOLUMN", 1, TBL_LABLEN, &nval, t3column);
  SCKGETC("SROW", 1, RSTR_BUFFER_SIZE, &nval, rstr);

  SCKRDI("CRANGE", 1, 2, &nval, fcrng, &unit, &null);
  SCKRDI("TROW", 1, 1, &nval, &t3rno, &unit, &null);


  /*
   * Get source frame information
   */

  if (SCFOPN(fname, D_R4_FORMAT, 0, F_IMA_TYPE, &fid))
    error(proId, msglev, "Cannot open input frame!");

  SCDRDI(fid, "NAXIS", 1, 1, &nval, &ifrm.naxis, &unit, &null);
  if (ifrm.naxis < 1 || ifrm.naxis > FRM_MAX_AXIS)
    error(proId, msglev, "Number of frame axes is out of range!");

  SCDRDI(fid, "NPIX", 1, 2, &nval, ifrm.npix, &unit, &null);
  SCDRDD(fid, "START", 1, 2, &nval, ifrm.start, &unit, &null);
  SCDRDD(fid, "STEP", 1, 2, &nval, ifrm.step, &unit, &null);
  if (ifrm.naxis == 1) {
    ifrm.npix[1] = 1;
    ifrm.start[1] = 1.;
    ifrm.step[1] = 1.;
  }

  ifrm.ndata = ifrm.npix[0] * ifrm.npix[1];


  /*
   * Convert world coordinates to frame coordinates if necessary
   */

  if (*rstr != '@')
    if (sscanf(rstr, "%lf", &rval) != 1)
      error(proId, msglev, 
	    "Conversion failed! Invalid format in frame row specification!");
    else 
      frno = (int)((rval - ifrm.start[1]) / ifrm.step[1]) + 1;
  else 
    if (sscanf(rstr + 1, "%lf", &rval) != 1)
      error(proId, msglev, 
	    "Conversion failed! Invalid format in frame row specification!");
    else 
      frno = (int)rval;
    
  if (frno < 1 || frno > ifrm.npix[1]) {
    sprintf(msg, "Selected frame row #%d is out of range!", frno);
    error(proId, msglev, msg);
  }


  /*
   * Check frame column selection
   */

  if (fcrng[0] < 1 || fcrng[1] > ifrm.npix[0] || fcrng[0] > fcrng[1])
    error(proId, msglev, "Invalid frame column selection!");
  else 
    ncol = fcrng[1] - fcrng[0] + 1;


  /*
   * Get frame data
   */

  if ((ifrm.data = (float *)osmmget(ncol * sizeof(float))) == NULL)
    error(proId, msglev, "Not enough memory to store frame row!");

  if (SCFGET(fid, (frno - 1) * ifrm.npix[0] + fcrng[0], ncol, &nval, 
	     (char *)ifrm.data)) {
      SCFCLO(fid);
      osmmfree(ifrm.data);
      error(proId, msglev, "Reading selected frame row failed!");    
    }
  else 
    SCFCLO(fid);

  
  /*
   * Get target table information
   */

  if (TCTOPN(t3name, F_IO_MODE, &t3id)) {
    osmmfree(ifrm.data);
    error(proId, msglev, "Cannot open output table!");
  }

  TCCSER(t3id, t3column, &t3cno);
  if (t3cno < 0) {
    TCTCLO(t3id);
    osmmfree(ifrm.data);

    sprintf(msg, "Column %s not found in output table %s.", t3column, t3name);
    error(proId, msglev, msg);    
  }
  else {
    TCIGET(t3id, &idum, &t3nrow, &idum, &idum, &idum);
    TCBGET(t3id, t3cno, &t3ctype, &t3csize, &t3cbytes);
  }


  /*
   * Check if target column is a character column. Data conversion from
   * numerical formats to character strings is not yet supported
   * due to incapabilities of the MIDAS TCA interface.
   */

  if (t3ctype == D_C_FORMAT) {
    TCTCLO(t3id);
    osmmfree(ifrm.data);

    error(proId, msglev, "Conversions number to string are not supported!");
  }


  /*
   * Check if array table element is large enough to hold
   * frame data
   */

  if (ncol > t3csize) {
    TCTCLO(t3id);
    osmmfree(ifrm.data);

    error(proId, msglev,
	  "Number of frame pixels is too large for output table!");
  }


  /*
   * Reset the whole table element array to null value
   */

  TCADEL(t3id, t3rno, t3cno, 1, 0);
    

  /*
   * Copy frame row to output table element
   */

  if (TCAWRR(t3id, t3rno, t3cno, 1, ncol, ifrm.data)) {
    TCTCLO(t3id);
    osmmfree(ifrm.data);

    error(proId, msglev, "Cannot write element array!");
  }


  /*
   * Cleanup 
   */

  TCTCLO(t3id);
  osmmfree(ifrm.data);


  /*
   * Disconnect from MIDAS
   */

  SCKWRC("Q1", 1, "0", 1, 60, &unit);
  SCSEPI();

}
