/* $Id: fors_t3dinsert.c,v 1.2 1999/03/04 12:41:49 rpalsa Exp $   
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 *.COPYRIGHT   (C) 1998 European Southern Observatory
 *
 *.IDENT       $RCSfile: fors_t3dinsert.c,v $
 *
 *.KEYWORDS    3d tables, insert table column as table element array
 *
 *.USAGE       run t3dinsert.exe
 *
 *.INPUT       SOURCE      Source table
 *             SCOLUMN     Label of source column
 *             RRANGE      Range of source rows to be copied (first,last)
 *             TCOLUMN     Label of target column
 *             TROW        Row number of target table element
 *
 *.OUTPUT      TARGET      3d output table
 *
 *.RETURNS     The program returns 0 on success or 1 on error. The exit
 *             status is stored in keyword Q1.
 *
 *.PURPOSE     Insert a column of the source table into a table element of
 *             the target table, whose elements are one dimensional arrays.
 *
 *.ENVIRON     MIDAS 98NOV, FORS pipeline
 *
 *.LANGUAGE    ANSI C/KR C
 *
 *.AUTHOR      R. Palsa, ESO-DMO
 *
 *.VERSION     0.1 1998/12/18: Creation, RP
 *
 *----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <midas_def.h>
#include <tblsys.h>
#include <tblerr.h>
#include <messages.h>



#define PROID_STRNG      "T3dInsert"
#ifndef __STDC__
#define PROID_STRNG_LEN  9
#endif

#define  MSG_BUFFER_SIZE  256



#ifdef __STDC__
void error(const char *name, int msglev, const char *msg)
#else
void error(name, msglev, msg)
  char *name, *msg;
  int msglev;
#endif
{

  int unit;


  messout(msglev, 1, name, msg);
  SCKWRC("Q1", 1, "1", 1, 60, &unit);
  SCSEPI();
}


void main() {

#ifdef __STDC__
  const char proId[] = PROID_STRNG;
#endif


  int    i, nval, unit, null, idum;
  int    nchar, nrow, offset;
  int    t2id, t2cno, t2nrow, t2ctype, t2csize, t2cbytes, t2store, t2rrng[2];
  int    t3id, t3cno, t3rno, t3nrow, t3ctype, t3csize, t3cbytes;
  int    msglev;
  int    state = 0;

  double sval;

  char   *t2cptr, *t2cbuf, *t3eptr;
  char   msg[MSG_BUFFER_SIZE];
  char   t2name[TBL_NAMLEN + 1], t2column[TBL_LABLEN + 1];
  char   t3name[TBL_NAMLEN + 1], t3column[TBL_LABLEN + 1];



#ifndef __STDC__
  char proId[PROID_STRNG_LEN + 1];
  strcpy(proId, PROID_STRNG);
#endif


  /*
   * Connect to MIDAS
   */

  SCSPRO((char *)proId);


  /*
   * Get keywords from environment
   */

  SCKRDI("MESS_LEVEL", 1, 1, &nval, &msglev, &unit, &null);

  SCKGETC("SOURCE", 1, TBL_NAMLEN, &nval, t2name);
  SCKGETC("TARGET", 1, TBL_NAMLEN, &nval, t3name);
  SCKGETC("SCOLUMN", 1, TBL_LABLEN, &nval, t2column);
  SCKGETC("TCOLUMN", 1, TBL_LABLEN, &nval, t3column);
  SCKRDI("RRANGE", 1, 2, &nval, t2rrng, &unit, &null);
  SCKRDI("TROW", 1, 1, &nval, &t3rno, &unit, &null);


  /*
   * Check range of source row selection
   */

  if (t2rrng[0] <= 0 || t2rrng[1] <= 0 || t2rrng[1] < t2rrng[0])
    error(proId, msglev, "Invalid source row selection!");


  /*
   * Get source table information
   */

  if (TCTOPN(t2name, F_I_MODE, &t2id))
    error(proId, msglev, "Cannot open input table");

  TCCSER(t2id, t2column, &t2cno);
  if (t2cno < 0) {
    TCTCLO(t2id);

    sprintf(msg, "Column %s not found in input table %s.", t2column, t2name);
    error(proId, msglev, msg);    
  }

  TCIGET(t2id, &idum, &t2nrow, &idum, &idum, &idum);
  TCBGET(t2id, t2cno, &t2ctype, &t2csize, &t2cbytes);


  /*
   * Adjust user selected range of rows if necessary.
   */

  if (t2rrng[1] > t2nrow)
    t2rrng[1] = t2nrow;


  TCDGET(t2id, &t2store);
  if (t2store != F_TRANS) {
    TCTCLO(t2id);

    sprintf(msg, "%s uses unsupported storage format!", t2name);
    error(proId, msglev, msg);    
  }
  else
    if (TCCMAP(t2id, t2cno, &t2cptr)) {
      TCTCLO(t2id);

      sprintf(msg, "Cannot map column %s into memory!", t2column);
      error(proId, msglev, msg);    
    }

  
  /*
   * Get target table information
   */

  if (TCTOPN(t3name, F_IO_MODE, &t3id)) {
    TCCUNM(t2id, t2cptr);
    TCTCLO(t2id);

    error(proId, msglev, "Cannot open output table!");
  }

  TCCSER(t3id, t3column, &t3cno);
  if (t3cno < 0) {
    TCCUNM(t2id, t2cptr);
    TCTCLO(t2id);
    TCTCLO(t3id);

    sprintf(msg, "Column %s not found in output table %s.", t3column, t3name);
    error(proId, msglev, msg);    
  }
  else {
    TCIGET(t3id, &idum, &t3nrow, &idum, &idum, &idum);
    TCBGET(t3id, t3cno, &t3ctype, &t3csize, &t3cbytes);
  }


  /*
   * Check if array table element is large enough to hold
   * the source column's row selection
   */

  nrow = t2rrng[1] - t2rrng[0] + 1;
  offset = t2rrng[0] - 1;

  if (nrow > t3csize) {
    TCCUNM(t2id, t2cptr);
    TCTCLO(t2id);
    TCTCLO(t3id);

    error(proId, msglev, "Range of input rows is too large for output table!");
  }


  /*
   * Check for column data type mismatch. Data conversion from 
   * numerical formats to character strings is not yet supported
   * due to incapabilities of the MIDAS TCA interface.
   */

  if (t2ctype != D_C_FORMAT && t3ctype == D_C_FORMAT) {
    TCCUNM(t2id, t2cptr);
    TCTCLO(t2id);
    TCTCLO(t3id);

    error(proId, msglev, "Conversions number to string are not supported!");
  }

    
  /*
   * Reset the whole table element array to null value
   */

  TCADEL(t3id, t3rno, t3cno, 1, 0);
    

  /*
   * Copy source column to output table element. For conversions
   * from character columns to numerical formats a workaround is
   * implemented to compensate for bugs in the MIDAS TCA interface.
   */

  switch (t2ctype) {
  case D_C_FORMAT:
    if (!(state = TCAMAP(t3id, t3rno, t3cno, &t3eptr))) {
      if (t3ctype == D_C_FORMAT) {
	nchar = t3cbytes / t3csize;
	nval = nchar < t2cbytes ? nchar : t2cbytes;

	for (i = 0; i < nrow && !state; i++)
	  state = TCATRC(t3id, t3eptr, t3cno, i * nchar + 1, nval,
			 t2cptr + (i + offset) * t2cbytes);
      }
      else 
	if ((t2cbuf = osmmget(nrow * t2cbytes + 1)) != NULL) {

	  /*
	   * Use internal zero-terminated string to avoid reading past
	   * the end of the table column.
	   */

	  memcpy(t2cbuf, t2cptr, nrow * t2cbytes);
          *(t2cbuf + nrow * t2cbytes) = '\0';

	  for (i = 0; i < nrow && !state; i++)
	    if (sscanf(t2cbuf + i * t2cbytes,"%lf", &sval) == 1)
	      state = TCAWRD(t3id, t3rno, t3cno, i + 1, 1, &sval);
	    else {
	      state = 1;
	      messout(msglev, 1, proId, 
		      "Conversion failed. Input string is not a number!");
	    }
	     
	  osmmfree(t2cbuf);

	}
	else 
	  state = 1;

      TCAUNM(t3id, t3eptr);
    }
    break;

  case D_I4_FORMAT:
    state = TCAWRI(t3id, t3rno, t3cno, 1, nrow, (int *)t2cptr + offset);
    break;

  case D_R4_FORMAT:
    state = TCAWRR(t3id, t3rno, t3cno, 1, nrow, (float *)t2cptr + offset);
    break;

  case D_R8_FORMAT:
    state = TCAWRD(t3id, t3rno, t3cno, 1, nrow, (double *)t2cptr + offset);
    break;

  default:
    TCCUNM(t2id, t2cptr);
    TCTCLO(t2id);
    TCTCLO(t3id);

    sprintf(msg, "Data type of column %s is not supported!", t2column);
    error(proId, msglev, msg);
    break;
  }

  if (state) {
    TCCUNM(t2id, t2cptr);
    TCTCLO(t2id);
    TCTCLO(t3id);

    error(proId, msglev, "Cannot write element array!");
  }


  /*
   * Cleanup 
   */

  TCCUNM(t2id, t2cptr);
  TCTCLO(t2id);
  TCTCLO(t3id);


  /*
   * Disconnect from MIDAS
   */

  SCKWRC("Q1", 1, "0", 1, 60, &unit);
  SCSEPI();

}
