/* $Id: robust_sigma.c,v 1.2 2001/04/11 14:35:56 sbogun Exp $
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * COPYRIGHT (c) 2001 European Southern Observatory
 * LICENSE: GNU General Public License version 2 or later
 *
 * PROJECT:   VLT Data Flow System
 * AUTHOR:    Stefan Bogun -- ESO/DMD/DPG
 * SUBSYSTEM: Instrument pipelines
 *
 * PURPOSE:
 *   Calculate a resistant estimate of the dispersion of a double array.
 *   For an uncontaminated distribution, this is identical to the standard
 *   deviation.
 *
 * USAGE:
 *   result = robust_sigma(y, npixel, zero) 
 *
 * INPUT:
 *   y:        double*: pointer to pixel array
 *   npixel:   int:     number of pixels
 *   zero:     int:     if set to 1 the dispersion is calculated w.r.t. 0.0 rather
 *                      than the central value of the vector. If Y is a vector
 *                      of residuals, this should be set.
 *
 * RETURNS:
 *   result:   double:  dispersion. In case of failure, returns a
 *                      value of -1.0
 *
 * DESCRIPTION:
 *   Use the median absolute deviation as the initial estimate, then weight
 *   points using Tukey's Biweight. See, for example, "Understanding Robust
 *   and Exploratory Data Analysis," by Hoaglin, Mosteller and Tukey, John
 *   Wiley & Sons, 1983.
 *
 * $Name: rel1_8_2 $
 * $Revision: 1.2 $
 * ----------------------------------------------------------------------------
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>


#define EPS 1.0E-20
#define SWAP(a,b) { double t=(a);(a)=(b);(b)=t; }
#define ABS(x) ((x) > 0 ? (x) : (-(x)))

int darray_sadd(double*, int, double, double*);
int darray_smult(double*, int, double, double*);
int darray_amult(double*, double*, int, double*);
int darray_abs(double*, int, double*);
double darray_total(double*, int);
double darray_median(double*, int);
int sort(double*, int);
double power(double, int);


double robust_sigma(double *y, int npixel, int zero)
{
    double y0, mad, numerator, den1, sigmaa;
    double ymy0[npixel], absymy0[npixel], u[npixel], uu[npixel];
    int i,count;

    if (zero) 
		y0 = 0.;
    else 
		y0 = darray_median(y, npixel);

    if (darray_sadd(y, npixel, -y0, ymy0) != 0)
		return -1.0;

    if (darray_abs(ymy0, npixel, absymy0) != 0)
		return -1.0;
 
/* First, the median absolute deviation about the median: */
    mad = darray_median(absymy0, npixel) / 0.6745;

/* If mad=0, try the MEAN absolute deviation: */
    if (mad < EPS) mad = darray_total(absymy0, npixel) / npixel / 0.80;

    if (mad < EPS) 
		return 0.0;

/* Now the biweighted value: */
    if (darray_smult(ymy0, npixel, 1./6./mad, u) != 0)
		return -1.0;

    if (darray_amult(u, u, npixel, uu) != 0)
		return -1.0;

    count = 0;
    numerator = 0.0;
    den1  = 0.0;
    for (i=0; i<npixel; i++) {
		if (uu[i] <= 1.0) {
			count++;
			numerator += power(y[i] - y0, 2) * power(1. - uu[i], 4);
			den1 += (1. - uu[i]) * (1. - 5 * uu[i]);
		}
    }

    if (count < 3)
		return -1.0;

    sigmaa = npixel * numerator / (den1 * (den1 - 1.));
    sigmaa = ((sigmaa > 0) ? sqrt(sigmaa) : 0.);

    return sigmaa;
}



/* Add a double scalar number to each of the elements on an array */
int darray_sadd(double *array1, int npixel, double scalar, double *array2)
{
    int i;

    for (i=0; i<npixel; i++) 
		array2[i] = array1[i] + scalar;
    
    return 0;
}


/* Multiply a scalar with each of the elements on an array */
int darray_smult(double *array1, int npixel, double scalar, double *array2)
{
    int i;

    for (i=0; i<npixel; i++) 
		array2[i] = array1[i] * scalar;
    
    return 0;
}


/* Multiply two arrays elementwise */
int darray_amult(double *array1, double *array2, int npixel, double *array3)
{
    int i;

    for (i=0; i<npixel; i++) 
	array3[i] = array1[i] * array2[i];
    
    return 0;
}


/* Compute the absolute value of each of the elements on an array */
int darray_abs(double *array1, int npixel, double *array2)
{
    int i;

    for (i=0; i<npixel; i++) 
		array2[i] = ABS(array1[i]);
    
    return 0;
}


/* Compute the sum of all elements of an array */
double darray_total(double *array, int npixel)
{
    int i;
    double sum = 0;

    for (i=0; i<npixel; i++) 
		sum += array[i];
    
    return sum;
}


/* Compute the median of all elements of an array 
   If the number of elements is even, use the average of the
   two central values                                         */
double darray_median(double *array, int npixel)
{
    double median;
    double tmparray[npixel];
    int i;

    for (i=0; i<npixel; i++) 
		tmparray[i] = array[i];

    sort(tmparray, npixel);
    
    median = ((npixel % 2) ? tmparray[(npixel-1)/2] : 
			  0.5 * (tmparray[npixel/2-1] + tmparray[npixel/2]));

    return median;
}


/* Sort an array in ascending order */
int sort(double *array, int npixel)
{
    int i,j;

    for (i=1; i<npixel; i++) 
		for (j=0; j<i; j++)
			if (array[j] > array[i]) SWAP(array[j],array[i]);
    
    return 0;
}


/* Compute the power of a double scalar to an integer exponent */
double power(double scalar, int exponent)
{
    int i;
    double power = 1.0;

    for (i=0; i<exponent; i++)
		power *= scalar;
   
    return power;
}
