! $Id: fors_ech_cal_findorder.prg,v 1.6 1999/11/25 11:24:24 sbogun Exp $
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
! .COPYRIGHT  (C) 1999  European Southern Observatory (ESO)
!
! .IDENT       $RCSfile: fors_ech_cal_findorder.prg,v $
!
! .KEYWORDS   Order detection
!
! .PURPOSE    Pipeline wrapper and frame classification for
!             fors_findorder.prg:
!             Determine positions of orders on a reference frame.
!             The found orders can be displayed.
! __________
!
! .USAGE      @p fors_ech_cal_findorder in ref [out] [geom] [hough]
!
! .INPUT      in    (P1): single input frame (ascii list with single frame);
!                         possible frame type:
!                           order reference image (SCREEN_FLAT_ECH);
!                             default: none
!
!             ref   (P2): super recipe mode: "+" (reference frame is included
!                         in P1),
!                         single recipe mode: set of reference frame (ascii
!                         list);
!                         needed frame type:
!                           table with purging information (DESC_PURGE_TABLE);
!                             default: none
!
!             out   (P3): single recipe mode: basename for ascii list with
!                         output frames,
!                         super recipe mode: again input frame and
!                         existing image catalogue (".cat"), separated by a
!                         comma;
!                           order table (ORDER_POS_ECH);
!                           background positions (BACKGROUND_POS_ECH);
!                             default: `order.tfits' (single mode),
!                                      `back.tfits' (single mode)
!
!             geom  (P4): >>> CHANGE ONLY IF YOU HAVE A CLEAR <<<
!                         >>> UNDERSTANDING OF THE METHOD !!! <<<
!                         distance of chosen columns AND
!                         no. of neighbour pixels for median filtering AND
!                         minimal value for first derivative AND
!                         tolerance for y-width of one order,
!                         separated by commas;
!                           default: 50,43,5,10
!
!             hough (P5): >>> CHANGE ONLY IF YOU HAVE A CLEAR <<<
!                         >>> UNDERSTANDING OF THE METHOD !!! <<<
!                         slope_start AND slope_end AND slope_step AND
!                         zero_point_step for the Hough transformation,
!                         separated by commas;
!                           default: 0.35,0.75,0.008,1.
!
! .OUTPUT     "order.tfits": order detection table with relative order numbers
!                            and fitted positions
!             "back.tfits":  table with background positions
!
! .RETURN     Q1:  0: successful return
!                 -1: a warning message occured
!                  1: an error occured
! __________
!
! .ALGORITHM  1. Detection of order positions in chosen columns by calculating
!                the column derivatives.
!             2. Computing straight coefficients for all orders by performing
!                a Hough transformation and cluster detection and thereby
!                finding the number of orders.
!             3. Assigning relative order numbers to order positions with the
!                help of the straight coefficients.
!             4. Improving the order positions with a 2nd order regression.
!             5. Computing positions of background areas with the help of the
!                regression coefficients.
!
! .COMMENT    The orientation of the input frame MUST be as follows, as
!             "fors_findorder_lin" (Hough trafo.) only works with POSITIVE
!             slopes !!!
!
!                           -------------
!                    low   |  / /    / / | higher
!                   (red)  |/ /    / /   |
!               Abs.       |/    / /    /|     Rel.
!               ord.       |   / /    / /|     ord.
!                    high  | / /    / /  | 1
!                   (blue) |      / /    |
!                          |             |
!                           -------------
!                           low      high
!                              Lambda
!
! .ENVIRON    FORS context
!
! .LANGUAGE   MIDAS 99NOV
!
! .AUTHOR     N.Rainer  DPG/DMD/ESO
!
! .VERSION    1.4  99/07/19  Creation
!                  99/10/21  Release
!
!------------------------------------------------------------------------------
!
!
crossref in ref out geom hough
!
!
define/local MAXSLITNO/i/1/1 2
!
!
define/local FMTDEF/c/1/17 "I1 E14.7,E23.15"
!
define/local PROID/c/1/40 "ech_cal_findorder"
!
define/local warn/i/1/1 0
write/keyword q1 0
!
!
if p1 .eq. "??" then
  w/o "      ******************************************************************"
  w/o "      in    (P1): single input frame (ascii list with single frame);"
  w/o "                  possible frame type:"
  w/o "                    order reference image (SCREEN_FLAT_ECH);"
  w/o "                      default: none"
  w/o " "
  w/o "      ref   (P2): super recipe mode: "+" (reference frame is included"
  w/o "                  in P1),"
  w/o "                  single recipe mode: set of reference frame (ascii"
  w/o "                  list);"
  w/o "                  needed frame type:"
  w/o "                    table with purging information (DESC_PURGE_TABLE);"
  w/o "                      default: none"
  w/o " "
  w/o "      out   (P2): single recipe mode: basename for ascii list with"
  w/o "                  output frame,"
  w/o "                  super recipe mode: again input frame and"
  w/o "                  existing image catalogue (".cat"), separated by a"
  w/o "                  comma;"
  w/o "                    order table (ORDER_POS_ECH);"
  w/o "                    background positions (BACKGROUND_POS_ECH);"
  w/o "                      default: `order.tfits' (single mode)"
  w/o "                               `back.tfits' (single mode)"
  w/o " "
  w/o "      geom  (P3): >>> CHANGE ONLY IF YOU HAVE A CLEAR <<<"
  w/o "                  >>> UNDERSTANDING OF THE METHOD !!! <<<"
  w/o "                  distance of chosen columns AND"
  w/o "                  no. of neighbour pixels for median filtering AND"
  w/o "                  minimal value for first derivative AND"
  w/o "                  tolerance for y-width of one order,"
  w/o "                  separated by commas;"
  w/o "                    default: 50,43,5,10"
  w/o " "
  w/o "      hough (P4): >>> CHANGE ONLY IF YOU HAVE A CLEAR <<<"
  w/o "                  >>> UNDERSTANDING OF THE METHOD !!! <<<"
  w/o "                  slope_start AND slope_end AND slope_step AND"
  w/o "                  zero_point_step for the Hough transformation,"
  w/o "                  separated by commas;"
  w/o "                    default: 0.35,0.75,0.008,1."
  w/o "      ******************************************************************"
  return/exit
endif
!
!
!          PREPARATIONS FOR RECIPE :
!          ***********************
!
        mess/out 3 {PROID} "Compute order positions ..."
!
@p fors_paragraph 1 1  ! -------- INITIALIZATION --------
!
log/recipe {PROID} in
!
set/format {FMTDEF}
!
!
!          INPUT :
!          *****
!
!P1(in):
!------
if p1 .eq. "?" mess/out 2 {PROID} " "
define/parameter p1 ? ? -
  "        Enter name of input frame: "
define/local inpipe_frm/c/1/76 "{p1}"
!
if m$exist(inpipe_frm) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered list with single input frame"
        mess/out 1 {PROID} "is not present or can not be opened !"
        mess/out 1 {PROID} "({inpipe_frm})"
  goto error
endif
!
define/local filecode/i/1/2 0 all
open/file {inpipe_frm} read filecode
if filecode(1) .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Unable to open entered list !"
        mess/out 1 {PROID} "({inpipe_frm})"
  goto error
endif
read/file {filecode(1)} outputc
if filecode(2) .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Entered list is empty !"
        mess/out 1 {PROID} "({inpipe_frm})"
  close/file {filecode(1)}
  goto error
endif
close/file {filecode(1)}
!
!
!P2(ref):
!-------
if p2 .eq. "?" mess/out 2 {PROID} " "
define/parameter p2 ? ? -
  "        Enter name of reference frame set: "
define/local inpipe_ref/c/1/76 "{p2}"
!
! No tests like for P1 - in super recipes reference frame is included in P1 .
!
!
!P3(out):
!-------
define/parameter p3 {PROID} ? -
  "        Enter basename of out.list(single)/exist.cat.".cat"(super): "
define/local outpipe/c/1/76 "{p3}"
!
!
!P4(geom):
!--------
define/parameter p4 50,43,5,10 ? -
  "        Enter geom.params. (see proc.header) [def:50,43,5,10]: "
outputi = m$parse(p4,"geom")
if outputi(1) .ne. 4 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P4 !"
  goto error
endif
define/local COLDIST/i/1/1 {geom01}
define/local YMED/i/1/1    {geom02}
define/local YDERTHR/i/1/1 {geom03}
define/local WIDTOL/i/1/1  {geom04}
!
! All checks are done in "fors_findorder.prg" !
!
!
!P5(hough):
!---------
define/parameter p5 0.35,0.75,0.008,1. ? -
  "        Enter Hough params. (see rec.header) [def:0.35,0.75,0.008,1.]: "
outputi = m$parse(p5,"hough")
if outputi(1) .ne. 4 then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in P5 !"
  goto error
endif
define/local SLOPMIN/r/1/1 {hough01}
define/local SLOPMAX/r/1/1 {hough02}
define/local SLOPSTEP/r/1/1 {hough03}
define/local ZEPOSTEP/r/1/1 {hough04}
!
define/local houghbad/i/1/1 0
set/format F5.3,
if SLOPMIN .ge. SLOPMAX then
        write/out " "
        mess/out 1 {PROID} "P3[1:2]: slope minimum >= slope maximum !"
        mess/out 1 {PROID} "({SLOPMIN} >= {SLOPMAX})"
  houghbad = 1
endif
inputr(1) = (SLOPMAX-SLOPMIN)/2.
if SLOPSTEP .ge. inputr(1) .and. houghbad .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "P3[3]: slope step too high !"
        mess/out 1 {PROID} "({SLOPSTEP} ; slope_step<{inputr(1)})"
  houghbad = 1
endif
! ZEPOSTEP can first be checked in "fors_findorder_lin.prg" !
set/format {FMTDEF}
if houghbad .eq. 1 goto error
!
!
!
!          DEFINITION OF VARIABLES :
!          ***********************
!
!     "Identifying input frames":
!
  define/local catal/i/1/1 0
  define/local topicima/c/1/76 " " all
  define/local id/c/1/25 " " all
  !
  define/local flatID/c/1/25 " " all
  !
  define/local ordima/c/1/76 " " all
  !
  define/local inord/i/1/1 0
  !
  define/local insdesc/c/1/60 " " all
!
!     "Calculating order width":
!
  define/local oldslitno/i/1/1 0
  define/local totslitno/i/1/1 0
  define/local descloop/i/1/1 0
  define/local ORDWID/r/1/1 0.
!
!     "Calling of procedure":
!
  define/local ordtbl/c/1/76 " " all
  define/local bactbl/c/1/76 " " all
!
!
!
!          LOADING OF GLOBAL KEYS (SINGLE RECIPE MODE) ,
!          CONVERSION FITS->BDF/TBL ,
!          CLASSIFICATION OF INPUT FRAME :
!          *********************************************
!
        mess/out 4 {PROID} "Getting input frame"
!
initialize/pipeline {inpipe_frm} {inpipe_ref} -
  inpipe_good.cat ? ? ECH inpipe_bad.cat ?
if {q1} .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Pipeline initialization failed !"
  goto error
endif
-delete inpipe_bad.cat
!
initialize/longlog {inpipe_frm} {inpipe_ref} 2 -
                   "geometrical parameters: {p3}" -
                   "Hough parameters:       {p4}"
!
!
!          IDENTIFICATION VIA DESCRIPTORS ,
!          MAPPING OF FRAME TO PARAMETER  :
!          ********************************
!
        mess/out 4 {PROID} "Identifying input frame"
!
flatID = "{FLAT_ID}"
!
create/icat {PROID}.cat null {FORS_IDSTRNG} >Null  ! only for finish/longlog
inord = 0
!
ident_loop:
  store/frame topicima inpipe_good.cat 1 end_ident_loop
  id = "{{topicima},{fors_idstrng}}"
  !
  if id .eq. "{flatID}" then
    ordima = topicima
    add/icat {PROID}.cat {topicima} >Null
        mess/out 4 {PROID} "{topicima} : {id} => input"
    inord = inord+1
  endif
  !
  goto ident_loop
end_ident_loop:
-delete inpipe_good.cat
!
if inord .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No matching order reference frame present !"
  goto error
elseif inord .gt. 1 then
        write/out " "
        mess/out 1 {PROID} "More than one order reference frame present !"
  goto error
endif
!
insdesc = "{{ordima},{h_instrume}}"
!
!
!          CHECKING NUMBER OF PORTS FOR READOUT :
!          ************************************
!
if fors_nout .gt. 1 then
        write/out " "
        mess/out 1 {PROID} "No 4-port readout possible for Echelle !"
        mess/out 1 {PROID} "({ordima})"
  goto error
endif
!
!
!          REMOVING OVERSCAN SECTIONS :
!          **************************
!
@p fors_paragraph 2 1  ! -------- PROCESSING --------
!
        mess/out 4 {PROID} "Cutting overscan regions from order ref. frame"
        mess/out 4 {PROID} "  {ordima}"
!
!************************
@p fors_trim {ordima} + +
!************************
if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Cutting of overscan areas failed !"
  goto error
endif
!
!
!O            ORIENTATING LOW (RED) ABS. ORDERS
!L            AT TOP OF ORDER REFERENCE FRAME :
!D            *********************************
!
!F         mess/out 4 {PROID} "Flipping order ref. frame in y-direction"
!R
!A inputd(1) = {{ordima},START(2)}
!M inputd(2) = {{ordima},STEP(2)}
!E flip/image {ordima} y
!S {ordima},START(2) = {inputd(1)}
!  {ordima},STEP(2) = {inputd(2)}
!
!
!          CALCULATING ORDER WIDTH :
!          ***********************
!
ORDWID = 0.
oldslitno = 0
totslitno = 0
do descloop = 1 FORS_MAXSLITLET
  if m$existd(ordima,"{h_moswid({descloop})}") .eq. 1 then
    if {{ordima},{h_moswid({descloop})}} .gt. 0. then
      inputi = oldslitno+1
      if inputi .gt. 1 .and. descloop .gt. inputi then
        write/out " "
        mess/out 1 {PROID} "Non-contiguous MOS slits seem to be open !"
        mess/out 1 {PROID} "( {ordima},{h_moswid({oldslitno})} and"
        mess/out 1 {PROID} "  {ordima},{h_moswid({descloop})} )"
        goto error
      endif
      !
      if m$existd(ordima,"{h_moslen({descloop})}") .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Descriptor with MOS slit length is missing !"
        mess/out 1 {PROID} "({ordima},{h_moslen({descloop})})"
        goto error
      elseif {{ordima},{h_moslen({descloop})}} .eq. 0. then
        write/out " "
        mess/out 1 {PROID} "Descriptor with MOS slit length = 0. !"
        mess/out 1 {PROID} "({ordima},{h_moslen({descloop})})"
        goto error
      else
        ORDWID = ORDWID + {{ordima},{h_moslen({descloop})}}
      endif
      !
      oldslitno = descloop
      totslitno = totslitno+1
    endif
  endif
enddo
if totslitno .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No descriptor w. non-zero MOS slit width present !"
        mess/out 1 {PROID} "({ordima},{kbase_mosi}i{ksuff_moswid})"
  goto error
elseif totslitno .gt. MAXSLITNO then
        mess/out 2 {PROID} -
          "More than {MAXSLITNO} descs.w.non-zero MOS slt.wid. present !"
        mess/out 2 {PROID} "({ordima},{kbase_mosi}i{ksuff_moswid})"
  goto error
endif
if m$existd(ordima,h_pixscale) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Descriptor w. pixel scale [arcsec/pix] is missing !"
        mess/out 1 {PROID} "({ordima},{h_pixscale})"
  goto error
endif
!
ORDWID = ORDWID / {{ordima},{h_pixscale}}  ! arcsec -> pixel
!
!
!          *********************************
!          *                               *
!          *  CALLING OF MAIN PROCEDURE :  *
!          *                               *
!          *********************************
!
        mess/out 4 {PROID} "Applying procedure "fors_findorder" :"
!
!************************************************************************
@p fors_findorder {ordima} {COLDIST},{YMED},{YDERTHR},{ORDWID},{WIDTOL} -
                           {SLOPMIN},{SLOPMAX},{SLOPSTEP},{ZEPOSTEP} n p
!************************************************************************
!
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Order detection failed !"
  goto error
endif
!
outputi = m$parse(q2,"tbl")
ordtbl = tbl01
bactbl = tbl02
!
create/icat outprog.cat null >Null
add/icat outprog.cat {ordtbl} >Null
add/icat outprog.cat {bactbl} >Null
-delete dirfile.ascii
!
!
!          UPDATING DESCRIPTORS OF OUT- AND INPUT FRAMES :
!          *********************************************
!
@p fors_paragraph 3 1  ! -------- TERMINATION --------
!
        mess/out 4 {PROID} "Classifying output tables"
        mess/out 4 {PROID} "({ordtbl}, {bactbl})"
!
classify/product {ordtbl} {ORDER_ID} {d_pre} {d_det} OK
write/descriptor {ordtbl} {h_instrume}/c/1/60 {insdesc}
write/descriptor {ordtbl} {FORS_IDSTRNG}/c/1/60 {ORDER_ID} !f.su.rec.
write/descriptor {ordtbl} {h_proqcsts}/c/1/9 "CHECKED"
write/dhelp {ordtbl} {h_proqcsts} "{i_proqcsts}"
longlog/out 1 "     Order table:      {ordtbl}"
!
classify/product {bactbl} {BACK_ID} {d_pre} {d_det} OK
write/descriptor {bactbl} {h_instrume}/c/1/60 {insdesc}
write/descriptor {bactbl} {FORS_IDSTRNG}/c/1/60 {BACK_ID} !f.su.rec.
write/descriptor {bactbl} {h_proqcsts}/c/1/9 "CHECKED"
write/dhelp {bactbl} {h_proqcsts} "{i_proqcsts}"
longlog/out 1 "     Background table: {bactbl}"
!
        mess/out 4 {PROID} "Endorsing QC status in input image"
!
write/descriptor {ordima} {h_proqcsts}/c/1/9 "CHECKED"
write/dhelp {ordima} {h_proqcsts} "{i_proqcsts}"
!
shortlog/out 1 "Order and background positions determined"
!
!
!          UPDATING OF OUTPUT CATALOGUE (SUPER RECIPE MODE) ,
!          CONVERSION BDF/TBL->FITS/TFITS (SINGLE MODE) :
!          **************************************************
!
        mess/out 4 {PROID} "Releasing output tables"
!
finish/pipeline outprog.cat {outpipe} + {ordtbl}
!
finish/longlog {PROID}.cat outprog.cat {ordtbl} 2 {outpipe}
!
-delete {PROID}.cat
-delete outprog.cat
!
!
!          RESETTINGS AFTER RECIPE :
!          ***********************
!
set/format
!
log/recipe {PROID} out
!
@p fors_paragraph 4 1  ! ================
!
!
if warn .eq. 0 then
  return 0
else
  return -1
endif
!
error:
  set/format
  finish/longlog ? -1 ? 2
  write/out " "
  return 1
!
!
!------------------------------------------------------------------------------
