! $Id: fors_gen_cal_badpix.prg,v 1.52 2000/04/12 10:27:50 sbogun Exp $   
!+++ FORS_GEN_CAL_BADPIX +++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT     (C) 1998  European Southern Observatory
!
!.IDENT         $RCSfile: fors_gen_cal_badpix.prg,v $
!
!.KEYWORDS      Bad pixel table/map creation
!
!.USAGE         @p fors_gen_cal_badpix sof_in sof_ref out [low] [high] [minnum] [noise] [expfact]
!
!.INPUT         sof_in  (P1): set of input frames (ascii list)
!               sof_ref (P2): set of reference frame (ascii list)
!               low     (P4): low level to identify bad pixels
!               high    (P5): high level to identify bad pixels
!               minnum  (P6): minimum number of bad pixels def. a bad column
!               noise   (P7): maximum allowed noise level
!               expfact (P8): expand factor to get best bad pixel areas
!
!.OUTPUT        out     (P3): path + basename of output files:
!                             1) bad pixel table [FITS]
!                             2) bad pixel map   [FITS]
!                             A BDF-copy of the output files are stored
!                             in the local directory:
!                             1) badpix.tbl      [BDF]
!                             2) badmsk.bdf      [BDF]
!
!.RETURN        Q1:  0: successful return
!                   -1: a successful return failed
!
!.DEFAULT       P3:     badpix  (output basename)
!               P4:     0.53    (low level)
!               P5:     1.89    (high level)
!               P6:     10      (minimum number)
!               P7:     0.02    (noise)
!               P8:     0.05    (expandfactor)
!
!.PURPOSE       Create a bad pixel mask from a set of flatfield measurements
!
!.ALGORITHM     A series of at least 10 low level flatfields  allows to check
!               for defects on the CCD.
!               The definition of a defect is given by (VLT-PLA-ESO-13110-1390)
!               if a pixel is
!                  a) lower than 50%
!                  b) higher than 200%
!               of the mean level. A series of 10 pixels defines a bad column
!               on the CCD.
!               The noise of the combined low level flatfields should be less
!               than 2% for at least 10 individual exposures.
!
!               (1) NO REDUCTION of the flat frames are necessary, because
!               possible cosmics would be eliminated as all flatfields will be
!               averaged to one frame which will be used for the bad pixel
!               detection (COMBINE/CCD).
!               (2) Using this averaged image the bad pixels may easily be
!               detected by the FIND/PIXEL command. All detected pixels will
!               be stored in a MIDAS table.
!               (3) This allows to create a bad pixel mask, i.e. an image
!               filled with 0 except those pixels listed in the MIDAS table,
!               which are set to 1.
!               (4) The next steps will analyse the bad pixel table. At first
!               bad columns will be searched for and will be stored in an
!               updated MIDAS table  BADPIX.
!               (5) After that the resting bad pixels will be grouped to
!               rectangular areas. These areas are necessary in order to
!               later fix these regions by the MODIFY/PIXEL command -- it does
!               a twodimensional fit to the data.
!               These areas will be optimized not only to contain the 'core'
!               of an bad pixel area but its surrounding too. This surrounding
!               ends until a certain intensity level has reached. This level
!               is determined through: |Int[i,j] - Mean|<limit
!               (6) At the end, the smallest non overlapping area of the whole
!               sample will be determined, in order to asure a correct fixing
!               process (s.(5)).
!               
!               ALL bad pixel columns and areas are finally stored in the table
!               BADPIX.TBL.
!               BADPIX-format: :X :XSTART :XEND :YSTART :YEND
!               Each row of the table defines a bad column or a bad pixel area.
!               In case of bad pixel areas the value of column :X is set to
!               0.0. In case of bad columns both, the column is specified by
!               the :X-value and an appropriate pixel area has been calculated.
!               Thus, bad columns may later be fixed by both methods
!               MODIFY/COLUMN and MODIFY/PIXEL.
!
!
!.ENVIRON       FORS, CCDRED context
!   
!.LANGUAGE      MIDAS 98NOV
!
!.AUTHOR        Sebastian Wolf, ESO-DMO
!
!.VERSION       0.5 1997/12/18  Creation
!               0.9 1998/08     TC release
!               1.0 1998/10/29  classify/product included
!                               main product specified: bad pixel table
!                   1998/12/21  Set IDENT to DO-classification
!                   1998/12/28  shortlog implemented
!               1.2 1999/02/03  update write/error
!                   1999/03/10  shortlog close shifted to FINISH/PIPELINE
!                   1999/03/15  obsolete descriptors deleted
!                               dhelps written
!                   1999/04/28  longlog implemented (SB).
!               1.3 1999/09/03  Product header generation implemented (SB).
!
!.COMMENT
!
!---------------------------------------------------------------------------
!*) at the moment (May'98) the Reduction Block Scheduler [RBS] sets this
!   parameter to the BASEname of all outputframes. Each outputframe will
!   be a construct of this BASEname and a number: {BASEname}+0001.fits
!---------------------------------------------------------------------------

define/local line/c/1/79 "-" ALL
write/out {line}
LOG/REC fors_gen_cal_badpix IN
!
! Check parameters
!
define/param P1 ? c "Enter name of ASCII filelist with input names:"
define/param P2 ? c "Enter name of ASCII filelist with reference names:"
define/param P3 badpix c "Enter name of set of result-frames (TBL and MAP):"
define/param P4 0.53 number     !low cut
define/param P5 1.89 number     !high cut
define/param P6 10   number     !min. number of pixels defining a bad col.
define/param P7 0.02 number     !max. noise
define/param P8 0.03 number     !expand factor
!
! local variable declaration
!
define/local IFRM/c/1/100 " " ALL
define/local MASK/c/1/100 " " ALL
define/local TABL/c/1/100 " " ALL

define/local IMSEC/i/1/4 0 ALL
define/local NDIMX/i/1/1 2080
define/local ONUM/i/1/1 1
define/local MNUM/i/1/1 0
define/local MEAN/r/1/1 -1.0
define/local EXPFACT/r/1/1 -1.0
define/local LOW/r/1/1 0
define/local HGH/r/1/1 0
define/local NPIX/i/1/2 0 ALL
define/local START/r/1/2 1.0 ALL
define/local STEP/r/1/2 1.0 ALL

define/local i1/i/1/1 0
define/local nare/i/1/1 0
define/local num/i/1/1 0
define/local ncol/i/1/1 0
define/local cnt/i/1/1 0
define/local catal/i/1/1 0
define/local procid/c/1/20 gen_cal_badpix
define/local frm/c/1/100 " " ALL
define/local warn0/c/1/80 " "
define/local warn1a/c/1/80 " "
define/local warn1b/c/1/80 " "
define/local warn2/c/1/80 " "
warn0 = "Too few INPUT frames! --> NO bad pixel detection."
warn1b = "--> NO bad pixel detection."
warn2 = "NO bad pixels found! -->END."

mess/out 3 {procid} "START to create a badpixel mask and table ..."
@p fors_paragraph 1

INITIALIZE/PIPELINE {P1} {P2} badpix.cat CCDRED + IMG reject.cat result.cat
if {Q1} .gt. 0 goto ABORT

init/prohead {p1} {p2} 2,"Badpixel mask and table generation" -
   "low_cut,{P4}" "high_cut,{P5}" "min_num,{P6}" "max_nois,{P7}" -
   "expfact,{P8}"
if {Q1} .gt. 0 goto ABORT

init/longlog {p1} {p2} 2 "low level: {P4}" "high level: {P5}" "minnum: {P6}" -
                      "noise: {P7}" "expfact: {P8}"
if {Q1} .gt. 0 goto ABORT

init/ccd
@p fors_gen_cal_badpix,BADPX_CAT badpix.cat
frm = outputc
if outputc(1:5) .eq. "empty" goto ABORT
show/icat badpix N
cnt = outputi(1)
@p fors_paragraph 2
!------------------------------------------------------------------------------
!
! combine set of low level flatfields
!

!IM_SEC = "[<,<:>,>]"           !reset image section (complete frame)
image/section {frm} X
IM_SEC = out_b
if cnt .gt. 1 then
  mess/out 3 {procid} "Create averaged flat field ..."
  @p fors_comb_setup FF
  if MESS_LEVEL .le. 3 then
    combine/ccd FF badpix.cat combff >Null
  else
    combine/ccd FF badpix.cat combff
  endif
  if m$existd("combff.bdf",h_origname) .gt. 0 -
    write/descr combff.bdf {h_origname} " " ALL
  if m$existd("combff.bdf",h_arcfile) .gt. 0 -
    write/descr combff.bdf {h_arcfile} " " ALL
  write/descr combff.bdf IDENT " " ALL
  combff.bdf,IDENT = "averaged flatfield"
else
  mess/out 2 {procid} "Only ONE frame specified!\nBad pixel table may be bad!"
  store/frame frm badpix.cat 1
  -copy {frm} combff.bdf
  mess/out 3 {procid} "INPUT frame: {frm} read."
endif

set/format I1
shortlog/out 1 "{cnt} screen flatfields averaged."      !reduction recipe log
set/format

!------------------------------------------------------------------------------
!
! check MAX. NOISE
!
mess/out 4 {procid} "Determine readout port(s) and image section"
CONFIGURE/PORTS combff
IMAGE/SECTION combff X
IM_SEC = out_b
copy/key OUTPUTI/i/1/4 IMSEC/i/1/4
mess/out 4 {procid} "Use image section: {IM_SEC}"

statistic/image combff {IM_SEC} ? ? SN
hgh = m$value(combff,statistic(4)) / m$value(combff,statistic(3))
if hgh .gt. {P7} then                   !check maximum NOISE
  warn1a = "Max. allowed noise exceeded: {hgh} > {P7}"
  mess/out 2 {procid} "{warn1a}"        !should be LESS ...
  mess/out 2 {procid} "{warn1b}"        !... than {P7} [0.02]
  goto ABORT
endif
!------------------------------------------------------------------------------
!
! find BAD PIXELS, save in table: badall
!
IFRM = "combff"         !input frame (averaged [median] screen flatfields)
MASK = "badmsk"         !name of bad pixel mask
TABL = "badpix"         !name of bad pixel table
MEAN = -1.0             !--> use subraster (4x4) means
EXPFACT = {P8}          !expand factor to get best areas (limit=EXPFACT*MEAN)
LOW = {P4}              !lower level factor to define bad pixel
HGH = {P5}              !upper level factor to define bad pixel
MNUM = {P6}             !minimum number of bad pixels defining a bad column
ONUM = fors_nout        !number of CCD-ports
if ONUM .eq. 4 then
  mess/out 3 {procid} "4-port readout: treat each quadrant separately"
endif

copy/dk combff START START/r/1/2
copy/dk combff STEP  STEP/r/1/2
copy/dk combff NPIX  NPIX/i/1/2
mess/out 3 gen_cal_badpix "Find bad pixel areas in the image frame ..."

run PIPE_EXE:fors_badpix.exe
if OUTPUTI(1) .le. 0 goto ABORT

show/table {TABL} >Null
nare = outputi(2)
if nare .gt. 1 then                     !check if any bad pixel area found
  create/column {TABL} :VALUE1 "PIXEL"
  write/table {TABL} :VALUE1 @1..{nare} 0.0
  -copy {TABL}.tbl wrk.tbl
  write/table wrk :VALUE1 @1..{nare} 1.0

  NDIMX = NPIX(1) * STEP(1)
  @p fors_check_overl wrk {TABL} {STEP(1)} {NDIMX}

  sort/table {TABL} :VALUE1,:XSTART
  select/tab {TABL} :VALUE1.eq.0.0 >Null !number of invalid rows in badpix
  num = outputi(1)                       !is less equal than maximum number
  if num .gt. 0 dele/row {TABL} @1..{num}               !of lines
  define/local mid$selidx/i/1/{nare} 0 ALL +lower_levels
  select/table {TABL} :X.gt.0 >Null

  ncol = outputi(1)                     !number of bad columns
  if ncol .gt. 0 then
    @a getcoord combff.bdf @{NPIX(1)},@{NPIX(2)} >Null
    HGH = outputd(2)
    do i1 = 1 {ncol}                    !define area as complete column
      {TABL},:YSTART,{mid$selidx({i1})} = START(2)
      {TABL},:YEND,{mid$selidx({i1})} = HGH
      {TABL},:YSTARTPIX,{mid$selidx({i1})} = 1.0
      {TABL},:YENDPIX,{mid$selidx({i1})} = {NPIX(2)} * 1.0
    enddo
  endif
  select/table {TABL} ALL
  show/table {TABL} >Null
  num = outputi(2)                      !total number of bad pixels

  create/column {TABL} :XW "PIXEL" >Null        !x-width
  create/column {TABL} :YW "PIXEL" >Null        !y-width
  create/column {TABL} :SZ "PIXEL" >Null        !size: num of pixel within area
  compute/table {TABL} :XW = :XENDPIX - :XSTARTPIX + 1
  compute/table {TABL} :YW = :YENDPIX - :YSTARTPIX + 1
  compute/table {TABL} :SZ = :XW * :YW
  name/column {TABL} :VALUE1 :MAXW

  mess/out 3 {procid} "Bad pixel area determination finished."
  set/format i1
  mess/out 3 {procid} "{num} bad pixel areas found ({ncol} bad columns)."
  longlog/out 1 "     {num} bad pixel areas found ({ncol} bad columns)."
  longlog/out 1 "     Bad pixel table is: {TABL}.tbl"
  longlog/out 1 "     Bad pixel mask is:  {MASK}.bdf"

  if num .eq. 0  mess/out 2 {procid} "NO bad pixels found, table is EMPTY!"
else
  mess/out 2 {procid} "NO bad pixels found, table is EMPTY!"
endif

@p fors_paragraph 3
mess/out 3 gen_cal_badpix -
  "OUTPUT frames:\nBad pixel table: {TABL}.tbl\nBad pixel mask:  {MASK}.bdf"

shortlog/out 1 "Bad pixel areas determined."


FIN:

TABL = "{TABL}.tbl"
MASK = "{MASK}.bdf"

copy/dd {MASK} *,3 {TABL}
if m$existd(MASK,h_origname) .gt. 0 write/descr {MASK} {h_origname} " " ALL
if m$existd(MASK,h_arcfile) .gt. 0 write/descr {MASK} {h_arcfile} " " ALL
write/descr {MASK} IDENT " " ALL
{MASK},IDENT = "{BPIXM_ID}"
!extract/frame {MASK} + +
trim/frame {MASK} + {IM_SEC} X
{MASK},LHCUTS(3) = 0.0
{MASK},LHCUTS(4) = 1.0

if m$existd(TABL,h_origname) .gt. 0 write/descr {TABL} {h_origname} " " ALL
if m$existd(TABL,h_arcfile) .gt. 0 write/descr {TABL} {h_arcfile} " " ALL

add/icat result.cat {TABL} >Null
add/icat result.cat {MASK} >Null
classify/product {TABL} {BPIXT_ID} {d_pre} {d_empty} OK
classify/product {MASK} {BPIXM_ID} {d_tmp} {d_empty} OK
copy/dd {TABL} {h_procatg} {TABL} {fors_idstrng}
copy/dd {MASK} {h_procatg} {MASK} {fors_idstrng}
copy/dd {MASK} NCOMBINE/i/1/1  {MASK} {h_proncom}/i/1/1 
write/dhelp {MASK} {h_proncom} "{i_proncom}"
delete/descr {MASK} NCOMBINE
delete/descr {MASK} MEAN
delete/descr {MASK} MEDIAN
delete/descr {MASK} MODE


if nice_crea .eq. "y" then     !display possible
  !
  @p fors_niceclear
  !
  !***********************************************
  @p fors_nicedisp {IFRM}.bdf ? 7. 0,0 ? ? 3.,3. n
  !***********************************************
  if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {procid} "Displaying of frame failed !"
        mess/out 1 {procid} "({IFRM}.bdf)"
    return -1
  endif
  !
        mess/out 4 {procid} "  loading image into display 1"
  assign/display d,1 >Null
  load/image {IFRM}.bdf ? {outputi(1)},avg c,c
  !
  mess/out 3 {procid} "DISPLAY: Mark rectangular bad pixel areas found."
  draw/rect {TABL} F N 5 >Null
  !
endif


! Create product header
finish/prohead result.cat 2


FINISH/PIPELINE result.cat {P3} CCDRED {TABL}
finish/longlog badpix.cat result.cat {TABL} 2 {P3}


LOG/REC fors_gen_cal_badpix OUT
@p fors_paragraph 4
return 0



ABORT:
finish/longlog badpix.cat -1 ? 1 {P4}
write/error 1,APP,NO "*** ERROR in gen_cal_badpix occured !"



!=== SUBROUTINES ==============================================================


ENTRY BADPX_CAT
define/param p1 ? c "Enter name of input catalog to be corrected:"

define/local catal/i/1/1 0
define/local isub/i/1/1 0
define/local frm/c/1/100 " " all
define/local expt/c/1/25 " " all

write/key outputc/c/1/80 " " ALL
outputc = "empty"
ima_loop:
  store/frame frm {p1} 1 finish
  expt = m$value({frm},{FORS_IDSTRNG})
  isub = 1
  if expt .eq. "{BPIXF_ID}" isub = 0
  if expt .eq. "{FLAT_ID}"  isub = 0
  if expt .eq. "{SKY_ID}"   isub = 0

  if isub .eq. 1 then
    subtr/icat {P1} {frm} >Null
  else
    outputc = frm
  endif
  goto ima_loop
finish:
