! $Id: fors_gen_cal_mkmaster.prg,v 1.92 2000/09/12 08:41:38 sbogun Exp $   
!+++ FORS_GEN_CAL_MKMASTER ++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT     (C) 1998  European Southern Observatory
!
!.IDENT         $RCSfile: fors_gen_cal_mkmaster.prg,v $
!
!.KEYWORDS      Create a master calibration frame
!
!.USAGE         @p (IDENT) sof_in sof_ref out [mode] [minnum] [thresh] [intfac]
!
!.INPUT         sof_in  (P1): set of input frames
!               sof_ref (P2): set of reference frames   (calibration frames
!                             used for the reduction of the input frames)
!               mode    (P4): controls the process and type of result frames
!                             (see .ALGORITHM)
!               minnum  (P5): minimum number of input frames
!               thresh  (P6): thresholds
!               intfac  (P7): interval factors for mmedian mode
!                             (BIAS,DARK,SCREEN-FLAT,SKY-FLAT)
!
!.OUTPUT        out     (P3): path + basename of output file [FITS]
!                             A copy of the output files are stored
!                             in the local directory:
!                             mkmaster.bdf      [BDF]
!               OUT_A   DO-classification of product (master frame)
!
!.RETURN        Q1:   0: successful return
!                     otherwise: successful return failed
!
!.DEFAULT       P3:     mkmaster  (basename of output)
!               P4:     22        (mode)
!               P5:     4         (minimum number of input frames)
!
!.PURPOSE       Create a master calibration file (reduction included) from
!               a set of calibration frames
!
!.ALGORITHM
!               Using a set of calibration frames (biases, darks, or,
!               flatfields) a master calibration frame will be created, i.e.
!               a median or mean averaged frame. The averaging will be made
!               iteratively by means of the MIDAS command COMBINE/CCD
!               (0) The averaging of the set of frames will be done as
!                   following:
!                   a) The median of all frames at each pixel will be
!                      determined and the pixelvalue at (x,y) of the frame
!                      nearest to the median found will be stored as the
!                      claimed median: med(x,y)
!                   b) According to a specified interval centered on the
!                      determined median med(x,y) the mean will be calculated
!                      and declared as the new median: med(x,y) (restored).
!                      This procedure has the advantage that the median may be
!                      corrected recognizing the surrondings of the median.
!                      In case this should be ignored the interval limits have
!                      to be set to fact=0. Finally, the interval will be
!                      determined by:
!                               delta = fact*MEAN
!                               [median-delta,median+delta]
!                      fact depends on the type of the input frames:
!                      bias and dark use 0.03, flat fields 0.01 but may be
!                      set by P7 (=bias,dark,screenflat,skyflat-factor), too.
!               (1) An averaged image of the whole set of frames will be
!                   created. This frame will be the first reference frame in
!                   the next step. No extra mean-averaging will be done -- 0b)
!                   will be omitted.
!               (2) Now, in an extra procedure each single frame will be
!                   compared with the reference frame. If the single frame
!                   differs too much it will be rejected. Finally, a further
!                   averaging will be performed only using the NOT rejected
!                   frames. This median will serve as a new reference frame.
!               (3) Repeat (2) as long as all used frames do not differ too 
!                   much with respect to the reference frame, OR if the 
!                   iteration ended in an empty set of frames. In the latter
!                   case, the quality criterium (P6) may be to strong and 
!                   should be adjusted accordingly.
!               Depending on the value set for P4=mode different types of
!               result frames may be produced:
!               mode<0: A constant frame will be produced as a master. The
!                       constant will be the absolut value |P4| passed to the
!                       program.
!               mode=0: Same as before, but the constant will be the mean of
!                       all frames specified in the list P1.
!               mode=1: A mean frame of all frames will be created.
!               mode=2: A median frame of all frames will be created.
!               mode=10,11,12: 
!                       The same as before, but the bias correction will
!                       be done using the scaled master bias. The scaling
!                       is done as follows:
!                       Each of the sections of the master bias is
!                       divided by its median and multiplied by the median
!                       of the corresponding pre-/overscan region of the
!                       input frame.
!               mode=20,21,22: 
!                       The same as before, but the bias correction will
!                       be done using the scaled master bias. The scaling
!                       is done as follows:
!                       Each of the sections of the master bias is
!                       divided by the median of its corresponding 
!                       pre-/overscan region and multiplied by the median
!                       of the pre-/overscan region of the input frame.
!
!.ENVIRON       FORS, CCDRED context
!
!.AUTHOR        Sebastian Wolf, ESO-DMO
!
!.VERSION       0.5 1997/10/28  Creation
!               0.9 1998/08     TC release
!               1.0 1998/10/27  Output improved
!                   1998/10/28  classify/frame obsolete,
!                               classify/product implemented
!                   1998/11/25  swap/units implemented (work on PIXEL system)
!                   1999/01/28  shortlog implemented
!                   1999/02/01  pre-selection test implemented.
!               1.2 1999/02/03  NO defaults for thresholds P6 and int.fact. P7,
!                               call to reduce/info adapted.
!                   1999/02/04  Create (night-)sky-flat from science and/or
!                               standard star exposures.
!                   1999/02/09  MASTER_NIGHT_FLAT classification added.
!                   1999/03/10  shortlog close shifted to FINISH/PIPELINE
!                   1999/03/15  obsolete descriptors deleted
!                               dhelps written
!               1.3 1999/04/29  Longlog implemented (SB)
!               1.4 1999/11/19  Scaled master bias used (SB)
!
!---------------------------------------------------------------------------
!*) at the moment (May'98) the Reduction Block Scheduler [RBS] sets this
!   parameter to the ROOTname of all outputframes. Each outputframe will
!   be a construct of this ROOTname and a number: {ROOTname}+0001.fits
!---------------------------------------------------------------------------
!
!
!          PREPARATIONS FOR RECIPE :
!          ***********************
!
define/local line/c/1/79 "-" ALL
write/out {line}
LOG/REC fors_gen_cal_mkmaster IN
!
!
!          INPUT :
!          *****
!
define/param p1 ? c "Enter name of ASCII filelist with input frames:"
define/param p2 + c "Enter name of ASCII filelist with calibration frames:"
define/param p3 mkmaster c "Name of result frame (master frame)"
define/param p4 22 n "Enter averaging mode and scaling method:"
define/param p5 4 n "Minimum number of input frames"

@p fors_gen_cal_mkmaster,CHCK_PARAM "{P6}" P6 fors_gen_cal_mkmaster
if {Q1} .ne. 0 return -1
@p fors_gen_cal_mkmaster,CHCK_PARAM "{P7}" P7 fors_gen_cal_mkmaster
if {Q1} .ne. 0 return -1
define/param p6 ? ? "Enter thresholds as defined in the rrd-files:"
define/param p7 ? ? "Enter mmedian-mode interval factors: "
!
!
!          DEFINITION OF VARIABLES :
!          ***********************
!
! constant declarations
!
define/local MAXNUM/i/1/1 20            !max.num of simultaneously opened files
define/local THRESH/r/1/1 {fors_maxlin}
define/local pid/c/1/20 "gen_cal_mkmaster"
!
! local variable declaration
!
define/local DO_RED/i/1/1 0             !flag: 1 if reduction should be done
define/local type/i/1/1 0
define/local i1/i/1/1 0
define/local j1/i/1/1 0
define/local nfrm/i/1/1 0               !number of input frames
define/local nfrm1/i/1/1 0
define/local ifnum/i/1/1 0              !number of interva factors read
define/local cnt/i/1/1 0
define/local chk/i/1/1 0
define/local catal/i/1/1 0              !used for reading catalogs by STORE
define/local avmode/i/1/1 0
define/local fact/r/1/1 0.03
define/local intfact/r/1/1 0.0 A +lower_levels          !interval factors for mmedian
define/local mnois/r/1/1 0.2
define/local qcrit/r/1/1 1.0
define/local sum/r/1/1 0.0
define/local summea/r/1/1 0.0
define/local summed/r/1/1 0.0
define/local limits/r/1/2 1.0,1.0
define/local binxy/i/1/1 0              !binning in x and y
define/local dots/c/1/79 "." ALL
define/local warn0/c/1/80 " " ALL
define/local expt/c/1/20 " " ALL
define/local etyp/c/1/2 " " ALL
define/local bs_scl/c/1/3 "NO"
define/local ofrm/c/1/100 mkmaster.bdf  !OUTPUT frame
define/local frm/c/1/80 " " ALL
define/local fidstr/c/1/25 " " ALL
define/local prefix/c/1/2 "./"
define/local csys/c/1/1 "W"
define/local switch/i/1/1 0
define/local minnum/i/1/1 0
define/local chktbl/c/1/64 " " all
define/local SC_BSTPL/c/1/100 " " all

fidstr = "MASTER_unknown"
warn0 = "Too few INPUT frames! --> NO MASTER-frame creation."
if {P5} .lt. 0 then 
  mess/out 2 {pid} "*** CHKCAL disabled ***"
  switch = 0
  minnum = -{P5}                        !minimum number of frames to be averag.
else
  switch = 1
  minnum = {P5}                         !minimum number of frames to be averag.
endif
define/local MINFRAMES/i/1/1 {minnum} ? +lower_levels
                                        !used by fors_chkcal

mess/out 3 {pid} -
  "Recipe GEN_CAL_MKMASTER will generate a\nmaster calibration frame."
@p fors_paragraph 1
!
!
!          LOADING OF GLOBAL KEYS (SINGLE RECIPE MODE) ,
!          CONVERSION FITS->BDF/TBL ,
!          CLASSIFICATION OF INPUT FRAMES :
!          *********************************************
!
INITIALIZE/PIPELINE {P1} {P2} mkmasterI.cat CCDRED + + reject.cat result.cat
if {Q1} .ne. 0 return {Q2}

init/prohead {p1} {p2} 2,"Master frame generation" "avmode,{p4}" -
             "minnum,{p5}" "thresholds,{p6}" "intfac,{p7}"
if {Q1} .ne. 0 goto ABO

init/longlog {p1} {p2} 2 "avmode: {p4}" "minnum: {p5}" "thresholds: {p6}" -
             "intfac: {p7}"
if {Q1} .gt. 0 goto ABO

@p fors_paragraph 2
mess/out 4 {pid} "Use local version of REDUCE/CCD and COMBINE/CCD"
!check at end: reset --> delete/comm
set/midas user=expert
delete/comm combine/ccd
create/comm combine/ccd @p ccdcombine >Null
delete/comm reduce/ccd
create/comm reduce/ccd @p ccdreduce >Null
set/midas user=user

avmode = m$nint({p4})

@p fors_initfors                                !initialize global keywords
                                                !used for IMG-mode of FORS.
                                                !exposure type descriptor
                                                !EXP_DESC used in CCDRED-
                                                !context for datatype
                                                !recognition is set.


show/icat mkmasterI N                           !get number of INPUT frames
nfrm = outputi(1)
nfrm1 = nfrm + 1
if nfrm .eq. 0 then
  mess/out 1 {pid} "NO frames specified --> EXIT!"
  return -1
endif
define/local useidx/r/1/{nfrm} 0.0 ALL +lower_levels !index of used INPUT frm.
define/local meanraw/r/1/{nfrm1} 0.0 ALL +lower_levels
define/local meanred/r/1/{nfrm1} 0.0 ALL +lower_levels
define/local medired/r/1/{nfrm1} 0.0 ALL +lower_levels
                                                !means of all raw INPUT frames
                                                !means of all reduced frames
                                                !median of all reduced frames
                                                !nfrm > number of raw INPUTs



!
!
!          CHECK IF SCALED BIAS CAN BE APPLIED AT ALL:
!          *******************************************
!
if avmode .gt. 22 then                          !check process-mode
  set/format i1
  mess/out 1 {pid} "Wrong MODE (P4={avmode}) selected !"
  set/format
  return -2

else if avmode .lt. 0 then                      !process-mode:
  store/frame frm mkmasterI.cat 1               !produce constant master frame
  compute/image {ofrm} = {frm} * 0.0 - {avmode}
  goto RDY                                      !OUTPUT created --> ready

else if {avmode} .ge. 10 then  ! use scaled master bias for reduction
  if {avmode} .ge. 20 then
    bs_scl = "OVE"
    avmode = avmode - 20
  else
    bs_scl = "MED"
    avmode = avmode - 10
  endif
c_loop:
  store/frame frm mkmasterI.cat 1 c_fini
  expt = "{{frm},{FORS_IDSTRNG}}"

  BRANCH expt {DARK_ID},{FLAT_ID},{SKY_ID},{SCIENCE_ID},{STD_ID} -
              chck,chck,chck,chck,chck
  goto c_loop
chck:
  compute/ovbias {frm} MEDIAN
  if {q1} .eq. -1  then
    mess/out 2 {pid} -
      "No overscan in input frame(s). Appylying standard bias correction."
    bs_scl = "NO"
  else if {q1} .ne. 0 then
    mess/out 1 {pid} -
      "Bias determination from pre-/overscan failed."
      return 1
  endif
  goto c_loop
    
c_fini:

endif

!
!
! Correct bad pixel, reduce, average frames
!
DO_RED   = 0                                    !reduction-flag: NO reduction
SC_BSCOR = "no"                                 !bias correction
SC_DKCOR = "no"                                 !dark correction
SC_FFCOR = "no"                                 !flat correction
SC_FXPIX = "no"                                 !bad pixel correction
!
!
!          IDENTIFICATION VIA DESCRIPTORS ,
!          MAPPING OF FRAMES TO PARAMETERS  :
!          ********************************
!
mess/out 4 {pid} "Identify calibration frames used for reduction ..."

search/icat mkmasterI {BPIXT_ID} >Null          !identify BAD PIXEL TABLE
if m$len(out_a) .ne. 0 then
  FX_TABLE = out_a
  mess/out 4 {pid} "BPIX: {FX_TABLE}"
else
  FX_TABLE = "not_used"
endif 

search/icat mkmasterI {MBIAS_ID} >Null          !identify BIAS frame
BS_TYP = "{BIAS_ID}"
if m$len(out_a) .ne. 0 then
  SC_BSCOR = "yes"
  if bs_scl .eq. "NO" then
    SC_BSFRM = out_a
    mess/out 4 {pid} "BIAS: {SC_BSFRM}"
  else
    SC_BSTPL = "TPL" // out_a
    SC_BSFRM = "SCL" // SC_BSTPL(4:)
    normalize/mbias {out_a} {SC_BSTPL} {bs_scl}
    if {q1} .ne. 0 then
      mess/out 1 {pid} "Normalization of master bias failed !"
      return 1
    endif
    mess/out 4 {pid} "BIAS TEMPLATE: {SC_BSTPL}"
    mess/out 4 {pid} "SCALED MASTER BIAS: {SC_BSFRM}"
  endif
endif 

search/icat mkmasterI {MDARK_ID} >Null          !identify DARK frame
DK_TYP = "{DARK_ID}"
if m$len(out_a) .ne. 0 then
  SC_DKCOR = "yes"
  SC_DKFRM = out_a
  mess/out 4 {pid} "DARK: {SC_DKFRM}"
else
  SC_DKFRM = "not_used"
endif
SC_FFFRM = "not_used"
SC_ILFRM = "not_used"
SC_FRFRM = "not_used"


if SC_BSCOR .eq. "yes" .OR. SC_DKCOR .eq. "yes" then
  DO_RED = 1                                    !REDUCTION will be performed
else
  mess/out 4 {pid} "NO calibration frames for reduction specified!"
endif
!
! Get interval factors
!
intfact = {P7}
!
!
!          REMOVING BAD PIXELS :
!          *******************
!
if FX_TABLE .ne. "not_used" then
  !
        mess/out 4 {pid} "Removing bad pixels from input frames"
  !
  create/icat mkmaster_noref.cat null >Null
  catal = 0
  bpx_loop:
    store/frame frm mkmasterI.cat 1 end_bpx_loop
    !
    expt = "{{frm},{FORS_IDSTRNG}}"
    branch expt {BIAS_ID},{DARK_ID},{FLAT_ID},{SKY_ID},{SCIENCE_ID},{STD_ID} -
      add_cat,add_cat,add_cat,add_cat,add_cat,add_cat
    goto bpx_loop
    add_cat:
      add/icat mkmaster_noref.cat {frm} >Null
    !
    goto bpx_loop
  end_bpx_loop:
  !
  correct/badpix {FX_TABLE} mkmaster_noref.cat
  if {q1} .ne. 0 then
    mess/out 1 {pid} "Correction of bad pixels failed !"
    return 1
  endif
  -delete mkmaster_noref.cat
  !
endif
!
!
!          REDUCTION OF INPUT FRAMES :
!          *************************
!
creat/icat mkmasterO.cat NULL {FORS_IDSTRNG} >Null

catal = 0
cnt = 1
write/out " "
mess/out 3 {pid} "Do the REDUCTION of ALL input frames ..."
r_loop:

  store/frame frm mkmasterI.cat 1 r_fini

  expt = "{{frm},{FORS_IDSTRNG}}"

  BRANCH expt {BIAS_ID},{DARK_ID},{FLAT_ID},{SKY_ID},{SCIENCE_ID},{STD_ID} -
              b_red,d_red,f_red,s_red,o_red,o_red
  goto r_loop

b_red:
  @p fors_gen_cal_mkmaster,UPD_IDENT {frm} {BIAS_ID}
  fidstr = "{MBIAS_ID}"
  etyp = "BS"
  type = 1
  fact = intfact
  csys = "W"
  DO_RED = 0                                    !BIAS: only bad pixel corr.
  goto red
d_red:
  @p fors_gen_cal_mkmaster,UPD_IDENT {frm} {DARK_ID}
  fidstr = "{MDARK_ID}"
  etyp = "DK"
  type = 2
  fact = intfact
  if fact .lt. 0.0 fact = 0.03                  !take the default
  csys = "W"
  goto red
f_red:
  @p fors_gen_cal_mkmaster,UPD_IDENT {frm} {FLAT_ID}
  fidstr = "{MFLAT_ID}"
  etyp = "FF"
  type = 4
  fact = intfact
  csys = "W"
  goto red
s_red:
  @p fors_gen_cal_mkmaster,UPD_IDENT {frm} {SKY_ID}
  fidstr = "{MSKY_ID}"
  etyp = "FF"                                   !OLD:  etyp = "SK"
  type = 3
  fact = intfact
  csys = "W"
  goto red
o_red:
  fidstr = "{MNIGHT_ID}"
  etyp = "FF"                                   !take the science and std-*
  type = 3                                      !as if they were sky-flats!
  fact = intfact
  csys = "W"

red:
  mess/out 4 {pid} "Process {expt}-frame: {frm}"

  {etyp}_TYP = "{expt}"                         !DO_CLASSIFICATION descriptor
                                                !.. used in CCDRED-context.
  swap/units {frm} PIXEL P5=1                   !force PIXEL coordinates!
  image/section {frm} X                         !get image section on CCD,
  binxy = outputi(9)                            !binning (1,2,4,...).
  IM_SEC = outputc                              !store interval in WCS
                                                !because of stat/ima call in
                                                !REDUC/CCD on reduced image
                                                !where pre-/postscan are cut
                                                !off as MASTER-frames have no
                                                !overscan regions anymore.


  SC_FXPIX = "no "
  statistic/image {frm} {IM_SEC} ? ? SN         !needed for first averaging
  meanraw({cnt}) = outputr(3)                   !mean of RAW-frame
  sum = sum + meanraw({cnt})                    !sum over all means



  if DO_RED .eq. 0 then                         !NO reduction necessary
    goto r_stat                                 !get next frame (prefix is still "./" !)
  endif
                                                !reduction is necessary
                                                !only for SKY, FLAT and DARK
  mess/out 3 {pid} "Do the reduction of frame: `{frm}`"
  prefix = "r_"
  parse/path {frm}
  CCD_IN = Q2
  frm = CCD_IN

! Scale master bias template 
  if SC_BSCOR .eq. "yes" .and. bs_scl .ne. "NO" then
    image/section {frm}
    if outputi(5) .gt. 0 .and. outputi(6) .gt. 0 then
      scale/mbias {frm} {SC_BSTPL} {SC_BSFRM}
      if {Q1} .ne. 0 return 1
    else
      mess/out 2 {pid} -
        "No overscan in input frame(s). Appylying standard bias correction."
    endif
  endif

  if MESS_LEVEL .lt. 4 then
    reduce/ccd {frm} {prefix}{frm} "Y" >Null
  else
    reduce/ccd {frm} {prefix}{frm} "Y"
    write/out {dots}
  endif
  @p fors_gen_cal_mkmaster,RM_OSCAN {prefix}{frm}.bdf


r_stat:
  statistic/image {prefix}{frm} {IM_SEC} ? ? GN >Null !for first averag. needed
  meanred({cnt}) = outputr(3)                   !mean of REDUCED-frame
  medired({cnt}) = outputr(8)                   !median of REDUCED-frame

  !*** Added (NR) ***
  if meanred({cnt}) .le. 0. .or. medired({cnt}) .le. 0. then
    !
    mess/out 2 {pid} "Count level of reduced frame <= 0. - skipped !"
    goto r_loop
    !
  endif
  !******************

  summea = summea + meanred({cnt})              !sum over means of positive frames
  summed = summed + medired({cnt})              !sum over medians of positive frames

  add/icat mkmasterO.cat {prefix}{frm} >Null    !DARK, FLAT, SKY

  if cnt .eq. 1 reduce/info {prefix}{frm} {type} {bs_scl} !reduction recipe log.
  !
  cnt = cnt + 1                                 !count processed frames
  !
  goto r_loop
r_fini:
!
!
!
!          CREATION OF PRELIMINARY MASTER FRAME :
!          ************************************
!
!                                               !Prepare the AVERAGING
show/icat mkmasterO N
nfrm = outputi(1)                               !number of to be averaged
nfrm1 = nfrm + 1                                !input frames.
if nfrm .lt. minnum then
  mess/out 1 {pid} "{warn0}"                    !Check if there are
  out_a = fidstr                                !enough frames
  return -3
endif
if nfrm .gt. MAXNUM then                        !Check if there are
  mess/out 1 {pid} "{warn0}"                    !too many frames, system allows
  out_a = fidstr                                !only a max. number of simult.
  return -3                                     !opened files
endif
sum = sum / nfrm
meanraw({nfrm1}) = sum
meanred({nfrm1}) = summea / nfrm                !mean of all means
medired({nfrm1}) = summed / nfrm                !mean of all median


                                                !PLOT the reduced  INPUT  data
@p fors_gen_cal_mkmaster,PLOT mkmasterO.cat {nfrm}

                                                !Do the setup for combine/ccd
                                                !AND   do the pre-selection
@p fors_comb_setup {etyp} mkmasterO.cat {nfrm} meanred medired
                                                !medired has now been changed
                                                !to     meanred(j)/scale(j)
                                                !scale(j) are the scal. factors
                                                !used in combine/ccd (s. below)

@p fors_gen_cal_mkmaster,PRESEL mkmasterO.cat {nfrm} medired meanraw intfact
nfrm  = {Q1}                                    !new number of input frames
nfrm1 = {Q2}
if nfrm .lt. minnum then
  mess/out 1 {pid} "{warn0}"                    !Check if there are
  out_a = fidstr                                !still enough frames
  return -3
endif

write/out
mess/out 3 {pid} "Do the first AVERAGING of ALL frames ..."
if MESS_LEVEL .gt. 3 then
  verbose = "yes"
else
  verbose = "no"
endif

BRANCH {avmode} 0,1,2 CST,MEA,MED

CST:
MEA:
  {etyp}_MET = "average"
  if MESS_LEVEL .lt. 4 then
    combine/ccd {etyp} mkmasterO.cat {ofrm} >Null
  else
    combine/ccd {etyp} mkmasterO.cat {ofrm}
  endif

  if avmode .eq. 0 then                                 !constant frame
    compute/image {ofrm} =  0.0 * {ofrm} + m$value({ofrm},STATISTIC(3))
  endif                                                 !mean of all frames
  goto a_fini

MED:
  limits(1) = fact * meanraw({nfrm1}) !BIAS, DARK: fact=0.03, FLATS: fact=0.01
  limits(2) = limits(1)
  {etyp}_CLP(1) = limits(1)     ![median-limits(1),median+limits(2)] defines
  {etyp}_CLP(2) = limits(2)     !interval within the MEAN-MEDIAN method oerates
  {etyp}_MET    = "mmedian"     !MEAN-MEDIAN method (same as in aver/ima)
  if MESS_LEVEL .lt. 4 then
    combine/ccd {etyp} mkmasterO.cat {ofrm} >Null
  else
    combine/ccd {etyp} mkmasterO.cat {ofrm}
  endif

a_fini:

if m$existd(ofrm,h_origname) .gt. 0 write/descr {ofrm} {h_origname} " " ALL
if m$existd(ofrm,h_arcfile) .gt. 0 write/descr {ofrm} {h_arcfile} " " ALL
if m$existd(ofrm,h_proqcobia) .gt. 0 delete/descr {ofrm} {h_proqcobia}
if m$existd(ofrm,h_proqconoi) .gt. 0 delete/descr {ofrm} {h_proqconoi}
if m$existd(ofrm,h_prqcobi(1)) .gt. 0 delete/descr {ofrm} {h_prqcobi(1)}
if m$existd(ofrm,h_prqcobi(2)) .gt. 0 delete/descr {ofrm} {h_prqcobi(2)}
if m$existd(ofrm,h_prqcobi(3)) .gt. 0 delete/descr {ofrm} {h_prqcobi(3)}
if m$existd(ofrm,h_prqcobi(4)) .gt. 0 delete/descr {ofrm} {h_prqcobi(4)}
if m$existd(ofrm,h_prqcono(1)) .gt. 0 delete/descr {ofrm} {h_prqcono(1)}
if m$existd(ofrm,h_prqcono(2)) .gt. 0 delete/descr {ofrm} {h_prqcono(2)}
if m$existd(ofrm,h_prqcono(3)) .gt. 0 delete/descr {ofrm} {h_prqcono(3)}
if m$existd(ofrm,h_prqcono(4)) .gt. 0 delete/descr {ofrm} {h_prqcono(4)}
write/descr {ofrm} IDENT " " ALL
{ofrm},IDENT = "{fidstr}"
{ofrm},LHCUTS(1) = m$value({ofrm},STATISTIC(1))
{ofrm},LHCUTS(2) = m$value({ofrm},STATISTIC(2))

if type .gt. 1 @p fors_gen_cal_mkmaster,RM_OSCAN {ofrm}
                                                       !remove SCAN descript.
!
!
!          ITERATIVE REJECTIONS OF SINGLE FRAMES
!          AND ITERATIVE CREATION OF NEW MASTER FRAME :
!          ******************************************
!
if switch .ge. 1 then                                   !check-frames-switch
  mess/out 3 {pid} "Start doing the AVERAGING ITERATIVELY ..."
else
  mess/out 2 {pid} "ITERATION switched OFF!"
  goto RDY
endif


i_loop:
  mess/out 3 {pid} "COMPARE input frames with master frame ..."
  @p fors_chkcal -
    mkmasterO.cat,meanraw,{binxy} {ofrm} {type} p u 0,0 + {p6}
  if {Q1} .gt. 0 then
    mess/out 3 {pid} "ABORT due to error in in fors_chkcal !"
    out_a = fidstr
    return -4
  endif
  chktbl = "mkmasterO_chk.tbl"


  BRANCH {Q2} -1,0,1 ABO,RDY,AVG
  goto i_fini
  AVG:
    sum = 0.0
    do i1 = 1 nfrm 1                                    !loop through all input
      j1 = useidx({i1})
      if j1 .gt. 0 then                                 !frames and check if
        meanraw({i1}) = meanraw( {useidx({i1})} )       !they are still not
        sum = sum + meanraw({i1})                       !rejected then push
      else                                              !them into the array of
        nfrm = i1 - 1                                   !means of raw frames
      endif                                             !and det. new MEAN and
    enddo                                               !update nfrm.
    sum = sum / nfrm

    limits(1) = fact * sum      !BIAS, DARK: fact=0.03, FLATS: fact=0.01
    limits(2) = limits(1)

    {etyp}_CLP(1) = limits(1)   ![median-limits(1),median+limits(2)] defines
    {etyp}_CLP(2) = limits(2)   !interval within the MEAN-MEDIAN method oerates
    {etyp}_MET    = "mmedian"   !MEAN-MEDIAN method (same as in aver/ima)

    mess/out 3 {pid} "AVERAGE revised image catalog ..."
    if MESS_LEVEL .lt. 4 then
      combine/ccd {etyp} mkmasterO.cat {ofrm} >Null
    else
      combine/ccd {etyp} mkmasterO.cat {ofrm}
    endif
    if type .gt. 1 @p fors_gen_cal_mkmaster,RM_OSCAN {ofrm}
    mess/out 3 {pid} "AVERAGING finished !"
    if m$existd(ofrm,h_origname) .gt. 0 write/descr {ofrm} {h_origname} " " ALL
    if m$existd(ofrm,h_arcfile) .gt. 0 write/descr {ofrm} {h_arcfile} " " ALL
    write/descr {ofrm} IDENT " " ALL
    {ofrm},IDENT = "{fidstr}"
    {ofrm},LHCUTS(1) = m$value({ofrm},STATISTIC(1))
    {ofrm},LHCUTS(2) = m$value({ofrm},STATISTIC(2))
    goto i_loop

  RDY:
    if m$existd(ofrm,h_origname) .gt. 0 write/descr {ofrm} {h_origname} " " ALL
    if m$existd(ofrm,h_arcfile) .gt. 0 write/descr {ofrm} {h_arcfile} " " ALL
    write/descr {ofrm} IDENT " " ALL
    {ofrm},IDENT = "{fidstr}"
    {ofrm},LHCUTS(1) = m$value({ofrm},STATISTIC(1))
    {ofrm},LHCUTS(2) = m$value({ofrm},STATISTIC(2))
    add/icat result.cat {ofrm} >Null
    add/icat result.cat {chktbl} >Null
    compute/ovbias {ofrm} MED
    trim/frame {ofrm} + +                               !cut off overscan reg.
    @p fors_gen_cal_mkmaster,RM_OSCAN {ofrm} 0          !set CCDSTAT(2) for log
    @p fors_paragraph 3
    @p fors_gen_cal_mkmaster,STATISTIC {ofrm} GN Y {pid} {fidstr} {nfrm}
    shortlog/out 1 "{outputc}"

    @p fors_qc1report_master {ofrm} mkmasterO.cat {fidstr}
    if {q1} .gt. 0 then
      goto ABO
    endif

    !
    ! only if CHKCAL is disabled - i.e., not for official pipeline :
    if switch .eq. 0 .and. nice_crea .eq. "y" then
      !
      !*******************************************
      @p fors_nicedisp {ofrm} ? 7. 0,0 ? ? 3.,3. n
      !*******************************************
      if {q1} .eq. 1 then
            write/out " "
            mess/out 1 {pid} "Displaying of frame failed !"
            mess/out 1 {pid} "({ofrm})"
        return 1
      endif
      !
        mess/out 4 {pid} "  loading image into display 2"
      assign/display d,2 >Null
      load/image {ofrm} ? {outputi(1)},avg c,c
      !
    endif
    !
    !
    !      UPDATING DESCRIPTORS OF OUTPUT FRAMES :
    !      *************************************
    !
    classify/product {ofrm} {fidstr} {d_pre} {d_empty} OK
    copy/dd {ofrm} {h_procatg} {ofrm} {fors_idstrng}
    delete/descr {ofrm} {h_tpexpno} NO
    delete/descr {ofrm} BPXLCORR NO
    delete/descr {ofrm} MEAN NO
    delete/descr {ofrm} MEDIAN NO
    delete/descr {ofrm} MODE NO
    delete/descr {ofrm} CCDMMM NO
    delete/descr {ofrm} CCDSCALE NO 
    !
    classify/product {chktbl} {CHKCALT_ID} {d_qcp} {d_empty} OK y
    copy/dd {ofrm} {h_instrume} {chktbl} {h_instrume}
    write/descriptor {chktbl} {FORS_IDSTRNG}/c/1/60 {CHKCALT_ID} !f.su.rec.
    write/descriptor {chktbl} {h_proqcsts}/c/1/9 "CHECKED"
    write/dhelp {chktbl} {h_proqcsts} "{i_proqcsts}"
    !
    !
    !      UPDATING OF OUTPUT CATALOGUE (SUPER RECIPE MODE) ,
    !      CONVERSION BDF/TBL->FITS/TFITS (SINGLE MODE) :
    !      **************************************************
    !
    finish/prohead result.cat 2
    FINISH/PIPELINE result.cat {P3} CCDRED
    finish/longlog mkmasterI.cat result.cat {ofrm} 2 {p3}
    goto i_fini

  ABO:
    @p fors_paragraph 3
    mess/out 1 {pid} "NO master frame could be created!"
    outputi = 1
    out_a = fidstr
    finish/longlog mkmasterI.cat -1 ? 1 {p3}
    return -5

i_fini:
!
!
!          RESETTINGS AFTER RECIPE :
!          ***********************
!
LOG/REC fors_gen_cal_mkmaster OUT
@p fors_paragraph 4
set/format
set/midas user=expert
delete/comm reduce/ccd
delete/comm combine/ccd
set/midas user=user
out_a = fidstr
return 0


!=== SUBROUTINES ==============================================================

ENTRY CHCK_PARAM
define/param p1 ? c "Enter contents of MIDAS parameter (P1, P2, ...):"
define/param p2 ? c "Enter Name of MIDAS parameter (P1, P2, ...):"
define/param p3 ? c "Enter Name of calling procedure:"

if m$index(p1,"?") .gt. 0 then
  mess/out 1 {p3} "Parameter {p2} has NO defaults!"
  mess/out 3 {p3} "Parameter {p2} has to be set as defined in rrd-files!"
  return -1
endif
return 0


!..............................................................................
ENTRY PLOT
define/parameter P1 ? c "image catalog which should be displayed"
define/parameter P2 ? n "number of images"

outputi(1) = m$sqrt({P2})               !Determine num. of columns and rows
outputr(1) = 1.0*{P2} / outputi(1)      !takes only integer part  --> columns
outputi(2) = m$nint(outputr(1))         !nearest integer (3.5->4) --> rows
outputi(3) = outputi(1) * outputi(2)
if outputi(3) .lt. {P2} then
  if outputi(1) .lt. outputi(2) then
    outputi(1) = outputi(1) + 1
  else
    outputi(2) = outputi(2) + 1
  endif
endif
@p fors_niceclear                       !Clear displays (image and graphic)
outputc = "REDUCED IMAGES will build MASTER FRAME"
@p fors_multiplt {P1} {outputi(1)} {outputi(2)} -1 P7="{outputc}"


!..............................................................................
ENTRY RM_OSCAN
define/parameter P1 ? c "image:"
define/parameter P2 1 c "mode:"

define/local i1/i/1/1 0
define/local j1/i/1/1 0

if {P2} .eq. 0 goto SETSTAT
mess/out 4 gen_cal_mkmaster "Delete overscan descriptors."
do i1 = 1 fors_nout
  j1 = fors_outi({i1})                          !remove pre-/postscan
  if m$existd(P1,h_pscan({j1})) .eq. 1  -
    delete/descr {P1} {h_pscan({j1})}           !descriptors as the overscan
  if m$existd(P1,h_oscan({j1})) .eq. 1  -
    delete/descr {P1} {h_oscan({j1})}           !regions have been cut off.
end do
SETSTAT:
if m$existd(P1,"CCDSTAT") .le. 0 -
  write/desc {P1} CCDSTAT/i/1/10 0 ALL
{P1},CCDSTAT(2) = 1                             !indicates overscan cut off.


!..............................................................................
ENTRY STATISTIC
define/parameter P1 ? c "image:"
define/parameter P2 ? c "stat/ima options (SN,FN,GN,...):"
define/parameter P3 ? c "info on screen Y/N"
define/parameter P4 ? c "name of calling procedure"
define/parameter P5 ? c "output category (FORS_IDSTRING)"
define/parameter P6 ? n "number of frames which produced master frame"
image/section {P1} X
statistic/image {P1} {out_b} ? ? {P2} >Null
{p1},LHCUTS(1) = m$value({p1},STATISTIC(1))
{p1},LHCUTS(2) = m$value({p1},STATISTIC(2))

copy/dd {P1} STATISTIC/r/1/1 {P1} {h_datamin}/d/1/1 
copy/dd {P1} STATISTIC/r/2/1 {P1} {h_datamax}/d/1/1 
copy/dd {P1} STATISTIC/r/3/1 {P1} {h_promean}/d/1/1 
copy/dd {P1} STATISTIC/r/4/1 {P1} {h_prosigm}/d/1/1 
copy/dd {P1} STATISTIC/r/8/1 {P1} {h_promedi}/d/1/1 
copy/dd {P1} NCOMBINE/i/1/1  {P1} {h_proncom}/i/1/1 
write/dhelp {P1} {h_datamin} "{i_datamin}"
write/dhelp {P1} {h_datamax} "{i_datamax}"
write/dhelp {P1} {h_promean} "{i_promean}"
write/dhelp {P1} {h_prosigm} "{i_prosigm}"
write/dhelp {P1} {h_promedi} "{i_promedi}"
write/dhelp {P1} {h_proncom} "{i_proncom}"
delete/desc {P1} NCOMBINE

outputi(20) = {P6}

if P3(1:1) .EQ. "Y" then
  set/format I1 F8.2
  write/out
  outputc = "{P5} created out of {outputi(20)} frames"
  mess/out 3 {P4} -
    "{outputc}:\n{P1} (MEAN: {outputr(3)})"
  set/format
endif


!..............................................................................
ENTRY UPD_IDENT
define/param p1 ? i "Enter image name:"
define/param p2 ? c "Enter image exposure type:"

!define/local ident/c/1/72

if m$strlen("{{p1},IDENT}") .eq. 0 then
  {p1},IDENT  = {P2}
endif



!..............................................................................
ENTRY PRESEL
!Do a preselection of the input frames:
!Check: abs(n_j - median_of_all(n_j)) > limit
!Update image catalog (p1)
!Update array P4 (meanraw)
!
define/param p1 ? c "Enter name of image catalog:"
define/param p2 ? n "Enter number of frames in the catalog:"
define/param p3 ? c "Enter name of check-array:"
define/param p4 ? c "Enter name of array of means of raw frames:"
define/param p5 ? c "Enter name of interval factors:"

define/local i1/i/1/1 0
define/local j1/i/1/1 0
define/local changed/i/1/1 0
define/local catal/i/1/1 0 +lower_levels
define/local x0/i/1/1 0.0
define/local y0/i/1/1 0.0
define/local sum/r/1/1 0.0
define/local test/r/1/1 0.0
define/local limit/r/1/1 0.0
define/local median/r/1/1 0.0
define/local pid/c/1/20 gen_cal_mkmaster

i1 = {p2} + 1
limit = {p5} * {p4}({i1})                               !check criterium
-rename {p1} tmp{p1}                                    !save 'old' catalog
crea/icat {p1} null >Null                               !create empty catalog

copy/ki {p3}/r/1/{p2} mediarr                           !determine the median
statistic/image mediarr ? ? ? GN >Null                  !.. of an array
median = outputr(8)                                     !median of median-array


mess/out 3 {pid} "Do the pre-selection test for input-frames:"
                                                        !Check input frames on:
                                                        !|n_j - median| > limit
j1 = 1
sum = 0.0
catal = 0
do i1 = 1 {p2}
  store/frame in_a tmp{p1} 1                            !get filename
  test = m$abs({p3}({i1}) - median)
  if test .gt. limit then
    changed = 1                                         !flags changed catalog
    if nice_crea .eq. "y" then
      y0 = m$nint({{in_a},DISPLAY_DATA(4)} / {{in_a},DISPLAY_DATA(2)})
      y0 = {{in_a},DISPLAY_DATA(9)} + y0
      x0 = m$nint({{in_a},DISPLAY_DATA(3)} / {{in_a},DISPLAY_DATA(1)})
      x0 = {{in_a},DISPLAY_DATA(8)} + x0 - 24
      {in_a},lhcuts(1) = {{in_a},lhcuts(2)} - 0.001     !set cuts in order to
      @% loadima {in_a} dirs=up,O                       !make image invisib.
      label/disp REMOVED {y0},{x0}
    endif
    append/out ". Frame: `{in_a}` REMOVED from catalog\n"
    append/out "  as |scaled_mean - median(all_scaled_means)|\n"
    append/out "  exceeds limit:   {test} > {limit}"
    mess/out 3 {pid}
    mess/out 4 {pid} "   (|{{p3}({i1})} - {median}| > {limit})"
  else
    append/out ". Frame: `{in_a}` is OK\n"
    append/out "  as |scaled_mean - median(all_scaled_means)|\n"
    append/out "  within limit:   {test} < {limit}"
    mess/out 3 {pid}
    mess/out 4 {pid} "  (|{{p3}({i1})} - {median}| < {limit})"
    add/icat {p1} {in_a} >Null
    {p4}({j1}) = {p4}({i1})                             !p4 = meanraw
    sum = sum + {p4}({j1})
    j1 = j1 + 1
  endif
enddo
sum = sum / ( j1 - 1 )
{p4}({j1}) = sum                                        !MEAN of all means
i1 = j1 - 1                                             !number of remaining
                                                        !frames.
if changed .eq. 0 mess/out 3 {pid} "  All frames are ok!"

-delete mediarr.bdf
return {i1} {j1} {changed}
