! $Id: fors_gen_qc0_overscan.prg,v 1.70 2000/05/23 16:06:57 nrainer Exp $
!+++ FORS_GEN_QC0_OVERSCAN ++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT     (C) 1998  European Southern Observatory
!
!.IDENT       $RCSfile: fors_gen_qc0_overscan.prg,v $
!
!.KEYWORDS      Determine bias from pre- and postscan region
!
!.USAGE         @p fors_gen_qc0_overscan sof_in out [accuracy] [port]
!
!.INPUT         sof_in  (P1): set of input frames (ascii list)
!               sof_ref (P2): set of reference frame (ascii list)
!               accur.  (P4): accuracy for the fit
!
!.OUTPUT        out     (P3): path + basename of output files:
!
!.RETURN        Q1:  0: successful return
!                    otherwise: successful return failed
!                    (>0 error, <0 warning)
!
!.DEFAULT
!
!.PURPOSE       Determine bias from pre- and postscan region and do
!               trend check by fitting a cubic polynomial to the overscans.
!
!.ALGORITHM     (1) determine pre-/overscan region using the FITS header
!                   keyword (DICB: ESO.DET.OUT{port}.aaSCb
!                   [aa:PR or OV, b:X or Y])
!               (2) call MIDAS command OVERSCAN/CCD in order to obtain a
!                   polynomial fit (degree=3) of the region perpendicular
!                   to the readout direction.
!               (3) determine a mean profile in the readout direction of the
!                   region and carry out an exponential non-linear least
!                   squares fit fit-function:  A*exp(-B*x) + C
!                   The constant  C  will finally be the claimed BIAS of the
!                   actual frame.
!                   If P4<0.0 then the median of the overscans will be
!                   determined and defined as the desired BIAS.
!               (4) Finally, the mean bias and its standard deviation will be
!                   saved in the hierachical FITS keywords;
!                       ESO.PRO.QC.OVBIAS and ESO.PRO.QC.OVNOISE
!                   within the output table.
!
!               CCD:
!                     p  sci   o         p: prescan region
!                    +-+------+-+        o: overscan (=postscan) region
!                 ^  |#|  *   |#|  ^     sci: scientific region
!                 |  |#|      |#|  |
!                (2) |#|    * |#| (2)
!                    +-+------+-+
!                        ----> readout direction
!                    (3)->    (3)->
!
!
!
!.ENVIRON       FORS, CCDRED context
!   
!.LANGUAGE      MIDAS 98NOV
!
!.AUTHOR        Sebastian Wolf, ESO-DMO
!
!.VERSION       0.5 1998/02/04  Created.
!               1.0 1998/10/27  PRO-Dictionary descriptors added.
!                   1998/11/25  swap/units implemented (work on PIXEL system).
!                   1998/12/09  fors_paragraph implemented.
!               1.1 1999/01/07  shortlog implemented.
!                   1999/02/16  comment string for QC-descriptors added.
!                   1999/03/10  shortlog close shifted to FINISH/PIPELINE
!               1.2 1999/04/28  longlog implemented (SB)
!               1.3 1999/09/03  Product header generation implemented (SB).
!
!.COMMENT       Output tables for each area (pre-/overscan and x-,y-direction):
!                  a) fit perpendicular to the scan area            (case (2))
!                  b) fit in readout-direction of the scan area     (case (3))
!                  c) parameters of the exponential fit, the third parameter
!                     -- the constant term -- reflects the desired bias.
!               In case of P4 is set to a negative value the MEDIAN of
!               the scan-regions will be determined instead of the
!               exponential fit.
!               At the end a combination of all tables will be stored as the
!               specified table (P3). The sequence of saving the fits are the
!               same as described above
!               a) and b), whereas the three parameters of the exponential fit
!               are stored in a set of descriptors: PRE_P# and/or POST_P#
!               (#:1,2,3). Additionally the MEAN BIAS and its "STANDARD
!               DEVIATION" are kept in the descritors BIAS_MEAN and BIAS_STDV.
!
!               Structure of the output table:
!               5 columns: ROW  COUNTRATE  FIT  RESIDUAL   COLUMN   
!               ROW and COLUMN:   the x-axis of the fits (negative values
!                                       indicate the actually not used column)
!               COUNTRATE:        the y-axis: the measurment
!               FIT:              the y-axis: the fit to the measurement
!               RESIDUAL:         the y-axis: the residuals of the fit
!
!
!------------------------------------------------------------------------------
!
!*) at the moment (May'98) the Reduction Block Scheduler [RBS] sets this
!   parameter to the ROOTname of all outputframes. Each outputframe will
!   be a construct of this ROOTname and a number: {ROOTname}+0001.fits
!
! TODO:
!   (1) quality control criteria!
!   (2) add the exposure type to the parameters in order to be sure that
!       the right type will be processed. At the moment it will be detected
!       automatically, is possibly not secure.
!-----------------------------------------------------------------------------

define/local line/c/1/79 "-" ALL
write/out {line}
LOG/REC fors_gen_qc0_overscan IN

!
! Check parameters
!
define/param P1 ? c "Enter name of INPUT frame:"
define/param P2 ? c "Enter name of REFERENCE frame:"
define/param P3 ? c "Enter name of OUTPUT table:"
define/param P4 -1 n "Enter accuracy (difference of succesive chi^2):"
!
! local variable declaration
!
define/local SCOFF/i/1/1 3
define/local port/i/1/1 1
define/local catal/i/1/1 0
define/local fid/i/1/2 0 ALL
define/local cnt/i/1/1 0
define/local iquad/i/1/1 -4
define/local ifnm/i/1/1 0
define/local i1/i/1/1 0
define/local num/i/1/1 0
define/local dumy/i/1/1 0
define/local xdim/i/1/1 0
define/local ydim/i/1/1 0
define/local ydir/i/1/1 0
define/local prdone/i/1/1 0
define/local ovdone/i/1/1 0
define/local prdony/i/1/1 0
define/local ovdony/i/1/1 0
define/local ovx/i/1/1 0
define/local prx/i/1/1 0
define/local ovy/i/1/1 0
define/local pry/i/1/1 0
define/local x1/i/1/1 0
define/local x2/i/1/1 0
define/local y1/i/1/1 0
define/local y2/i/1/1 0
define/local xd/i/1/1 0
define/local yd/i/1/1 0
define/local binx/i/1/1 1
define/local biny/i/1/1 1
define/local bx1/r/1/1 -1.0
define/local bx2/r/1/1 -1.0
define/local by1/r/1/1 -1.0
define/local by2/r/1/1 -1.0
define/local bias/r/1/1 0.0
define/local sdev/r/1/1 0.0
define/local sigAC/r/1/1 0.0
define/local sigBD/r/1/1 0.0
define/local sig/r/1/1 0.0
define/local BIAS_MEAN/r/1/4 -1.0 ALL
define/local BIAS_SIGM/r/1/4 -1.0 ALL
define/local procid/c/1/20 gen_qc0_overscan
define/local CA/c/1/2 CA
define/local DB/c/1/2 DB
define/local suffix/c/1/1 "x"
define/local ifrm/c/1/100 " " ALL
define/local rtbl/c*25/1/12 " " ALL
define/local errm/c/1/200 " " ALL
define/local warn0/c/1/200 " " ALL
warn0 = "NO INPUT frame specified !"
errm = "NO pre- and postscan region found!\nMay be regions are not properly defined."

mess/out 3 {procid} "Recipe GEN_QC0_OVERSCAN will determine the bias\nfrom pre- and postscan region."
@p fors_paragraph 1


INITIALIZE/PIPELINE {P1} {P2} overscI.cat CCDRED + + overscR.cat overscO.cat
if {Q1} .gt. 0 return {Q1}

init/prohead {p1} {p2} 2,"Bias determination from pre-/overscan"  "accuracy,{P4}"
if {Q1} .gt. 0 return 1

init/longlog {p1} {p2} 2 "accuracy: {P4}"
if {Q1} .gt. 0 return 1

@p fors_paragraph 2
port = fors_outi(1)
@p fors_initfors                        !set correct DESCRIPTORS

set/midas user=expert
mess/out 4 {procid} "LOCAL version of OVERSCAN/CCD in use!"
delete/comm overscan/ccd
create/comm overscan/ccd @p ccdscan >Null
set/midas user=user
set/format I1 F7.2

rtbl(1) = "{BIAS_ID}"
rtbl(2) = "{DARK_ID}"
rtbl(3) = "{FLAT_ID}"
rtbl(4) = "{SKY_ID}"
rtbl(5) = "{STD_ID}"
rtbl(6) = "{SCIENCE_ID}"
rtbl(7) = "{WLC_ID}"


do i1 = 1 7
  search/icat overscI.cat {rtbl({i1})} >Null
  if m$len(out_a) .ne. 0 then
    ifrm = out_a
    if "{{ifrm},{FORS_IDSTRNG}(1:6)}" .eq. "MASTER" .OR. -
       "{{ifrm},{FORS_IDSTRNG}(1:5)}" .eq. "SUPER" then
!   SUPER and MASTER frames in the catalog have to be skipped!
!   They are only for calibration purposes in reduction recipes.
     else
       i1 = 7
    endif
  endif
enddo
if ifrm(1:1) .eq. " " then
  mess/out 1 {procid} "NO input frame found!"
  return 1
endif

@p fors_getwin {ifrm}                           !check for window readout
if {Q1} .eq. 1 then
  append/out "Input frame `{ifrm}` is made\nin window readout mode.\n"
  append/out "There are NO pre- and postscan regions!"
  mess/out 2 {procid}
  return -1
endif





mess/out 3 {procid} "Determine OVERSCAN from: {ifrm}"
ifnm = m$len(ifrm)
!
!initialize keywords
!
xdim = m$value({ifrm},NPIX(1))
ydim = m$value({ifrm},NPIX(2))

OV_IMODE = "no"         !no interactive fitting of overscan region
OV_FUNCT = "poly"
OV_ORDER = 3
OV_AVER  = 1
OV_ITER  = 1
OV_REJEC = 3
cnt = 0

if m$existd(ifrm,h_nout) .eq. 1 then
  num = m$value({ifrm},{h_nout})                !number of outputs
  if num .gt. 1 then                            !FAST (4-port) readout
    iquad = 0
    num = num / 2
  else
    CA = "AA"
    DB = "AA"
  endif
  cnt = num
endif

swap/units {ifrm} PIXEL
image/section {ifrm} X
IM_SEC = outputc
prx  = outputi(5)
ovx  = outputi(6)
pry  = outputi(7)
ovy  = outputi(8)
binx = outputi(9)
biny = outputi(10)


if num .eq. 1 then
  longlog/out 1 "     +---+---+"
  longlog/out 1 "     |       |"
  longlog/out 1 "     |       |"
  longlog/out 1 "     |       |"
  longlog/out 1 "     +   A   +"
  longlog/out 1 "     |       |"
  longlog/out 1 "     |       |"
  longlog/out 1 "     |       |"
  longlog/out 1 "     +---+---+"
  longlog/out 1 " "
else if num .eq. 2 then
  longlog/out 1 "     +---+---+"
  longlog/out 1 "     |   |   |"
  longlog/out 1 "     | C | D |"
  longlog/out 1 "     |   |   |"
  longlog/out 1 "     +---+---+"
  longlog/out 1 "     |   |   |"
  longlog/out 1 "     | A | B |"
  longlog/out 1 "     |   |   |"
  longlog/out 1 "     +---+---+"
  longlog/out 1 " "
else
  mess/out 2 {procid} "Neither one nor four port output. FUNNY!"
endif


!
! Clear displays (image and graphic)
!
@p fors_niceclear
!
! Do the CALCULATION
!
x2 = 0
y2 = 0
xd = xdim / num
yd = ydim / num
loop:
  rtbl(1) = "overscfipx1{CA({cnt}:{cnt})}.tbl"
  rtbl(2) = "overscfipx2{CA({cnt}:{cnt})}.tbl"
  rtbl(3) = "overscparpx{CA({cnt}:{cnt})}.tbl"
  rtbl(4) = "overscfiox1{DB({cnt}:{cnt})}.tbl"
  rtbl(5) = "overscfiox2{DB({cnt}:{cnt})}.tbl"
  rtbl(6) = "overscparox{DB({cnt}:{cnt})}.tbl"
  rtbl(7) = "overscfipy1{CA({cnt}:{cnt})}.tbl"
  rtbl(8) = "overscfipy2{CA({cnt}:{cnt})}.tbl"
  rtbl(9) = "overscparpy{CA({cnt}:{cnt})}.tbl"
  rtbl(10) = "overscfioy1{DB({cnt}:{cnt})}.tbl"
  rtbl(11) = "overscfioy2{DB({cnt}:{cnt})}.tbl"
  rtbl(12) = "overscparoy{DB({cnt}:{cnt})}.tbl"
  x1  = 1 + x2
  x2  = x2 + xd
  y1  = 1 + y2
  y2  = y2 + yd

  if prx .gt. 1 then
    iquad = iquad + 1
    OV_SEC = "[@1,@{y1}:@{prx},@{y2}]"
    overscan/ccd {ifrm} {rtbl(1)},TABLE {OV_SEC} x >Null
    average/row  scan = {ifrm} [@{y1},@{y2}]
    extract/imag scan = scan [@1:@{prx}]
    plot/oscan {rtbl(1)} :ROW "PRESCAN (X)" 1 {NICE_HCOPY} {iquad}
    fit/exponential scan {rtbl(2)} {rtbl(3)} {P4} >Null
    if {Q1} .ne. 0 then
      if {P4} .gt. 0 then
        write/out
        mess/out 2 {procid} "Exponential fit FAILED !\nUse MEDIAN instead !"
      endif

      statistic/image {ifrm} {OV_SEC} ? ? SN >Null
      sigAC = outputr(4)

      statistic/image scan [@1:@{prx}] ? ? GN >Null
      compute/table {rtbl(2)} :FIT = {outputr(8)}               !MEDIAN
      compute/table {rtbl(2)} :RESIDUAL = :Y - :FIT
      {rtbl(3)},:RESULT,3 = outputr(8)
      append/out "PRESCAN region   (X-dir.): {OV_SEC}\n"
      append/out "Median of prescan region = {outputr(8)}"
      mess/out 3 {procid}
      longlog/out 1 "     PRESCAN region   (X-dir.): {OV_SEC}"
      longlog/out 1 "     Median of prescan region = {outputr(8)}"
    else
      mess/out 3 {procid} "CHI^2 of exponential fit: {outputr(1)}"
      longlog/out "     CHI^2 of exponential fit: {outputr(1)}"
      plot/oscan {rtbl(2)} :COLUMN "PRESCAN (X)" 2 {NICE_HCOPY} {iquad}
    endif
    create/column {rtbl(1)} :COLUMN
    write/table   {rtbl(1)} :COLUMN @1..{yd} -1
    create/column {rtbl(2)} :ROW
    write/table   {rtbl(2)} :ROW @1..{prx} -1
    delete/image scan N

    bx1 = m$value({rtbl(3)},:RESULT,3)
    bias  = bx1
    merge/tab {rtbl(1)} {rtbl(2)} prescanx
    prdone = 1
  endif

  dumy = ovx
  ovx = xdim - ovx + 1
  if ovx .lt. xdim then
    iquad = iquad + 1
    OV_SEC = "[@{ovx},@{y1}:@{xdim},@{y2}]"
    overscan/ccd {ifrm} {rtbl(4)},TABLE {OV_SEC} x >Null
    average/row  scan = {ifrm} [@{y1},@{y2}]
    extract/imag scan = scan [@{ovx}:@{xdim}]

    plot/oscan {rtbl(4)} :ROW "POSTSCAN (X)" 1 {NICE_HCOPY} {iquad}
    fit/exponential scan {rtbl(5)} {rtbl(6)} {P4} >Null
    if {Q1} .ne. 0 then
      if {P4} .gt. 0 then
        write/out
        mess/out 2 {procid} "Exponential fit FAILED !\nUse MEDIAN instead !"
      endif

      statistic/image {ifrm} {OV_SEC} ? ? SN >Null
      sigBD = outputr(4)

      statistic/image scan [@1:@{dumy}] ? ? GN >Null
      compute/table {rtbl(5)} :FIT = {outputr(8)}
      compute/table {rtbl(5)} :RESIDUAL = :Y - :FIT
      {rtbl(6)},:RESULT,3 = outputr(8)
      append/out "POSTSCAN region  (X-dir.): {OV_SEC}\n"
      append/out "Median of postscan region = {outputr(8)}"
      mess/out 3 {procid}
      longlog/out 1 "     POSTSCAN region  (X-dir.): {OV_SEC}"
      longlog/out 1 "     Median of postscan region = {outputr(8)}"
    else
      mess/out 3 {procid} "CHI^2 of exponential fit: {outputr(1)}"
      longlog/out 1 "     CHI^2 of exponential fit: {outputr(1)}"
      plot/oscan {rtbl(5)} :COLUMN "POSTSCAN (X)" 2 {NICE_HCOPY} {iquad}
    endif
    create/column {rtbl(4)} :COLUMN
    write/table   {rtbl(4)} :COLUMN @1..{yd} -1
    create/column {rtbl(5)} :ROW
    write/table   {rtbl(5)} :ROW @1..{dumy} -1
    delete/image scan N

    bx2 = m$value({rtbl(6)},:RESULT,3)
    bias  = bias + bx2
    MERGE/TAB {rtbl(4)} {rtbl(5)} postscanx
    ovdone = 1
  endif
  ovx = dumy


  @p fors_combtab {prdone} {ovdone} x {rtbl(3)} {rtbl(6)}

  if pry .gt. 1 then
    ydir = 1
    suffix = "y"
    iquad = iquad + 1
    OV_SEC = "[@{x1},@1:@{x2},@{pry}]"
    overscan/ccd {ifrm} {rtbl(7)},TABLE {OV_SEC} y >Null
    average/col  scan = {ifrm} [@{x1},@{x2}]
    extract/imag scan = scan [@1:@{pry}]

    plot/oscan {rtbl(7)} :ROW "PRESCAN (Y)" 1 {NICE_HCOPY} {iquad}
    fit/exponential scan {rtbl(8)} {rtbl(9)} {P4} >Null
    if {Q1} .ne. 0 then
      if {P4} .gt. 0 then
        write/out
        mess/out 2 {procid} "Exponential fit FAILED !\nUse MEDIAN instead !"
      endif

      statistic/image {ifrm} {OV_SEC} ? ? SN >Null
      sigAC = outputr(4)

      statistic/image scan [@1:@{pry}] ? ? GN >Null
      compute/table {rtbl(8)} :FIT = {outputr(8)}               !MEDIAN
      compute/table {rtbl(8)} :RESIDUAL = :Y - :FIT
      {rtbl(9)},:RESULT,3 = outputr(8)
      append/out "PRESCAN region   (Y-dir.): {OV_SEC}\n"
      append/out "Median of prescan region = {outputr(8)}"
      mess/out 3 {procid}
    else
      mess/out 3 {procid} "CHI^2 of exponential fit: {outputr(1)}"
      plot/oscan {rtbl(8)} :COLUMN "PRESCAN (Y)" 2 {NICE_HCOPY} {iquad}
    endif
    create/column {rtbl(7)} :COLUMN
    write/table   {rtbl(7)} :COLUMN @1..{xd} -1
    create/column {rtbl(8)} :ROW
    write/table   {rtbl(8)} :ROW @1..{pry} -1
    delete/image scan N

    by1 = m$value({rtbl(9)},:RESULT,3)
    bias  = bias + by1
    MERGE/TAB {rtbl(8)} {rtbl(8)} prescany
    prdone = 1 + prdone
    prdony = 1
  endif

  dumy = ovy
  ovy = ydim - dumy + 1
  if ovy .lt. ydim then
    ydir = ydir + 1
    suffix = "y"
    iquad = iquad + 1
    OV_SEC = "[@{x1},@{ovy}:@{x2},@{ydim}]"
    overscan/ccd {ifrm} {rtbl(10)},TABLE {OV_SEC} y >Null
    average/col  scan = {ifrm} [@{x1},@{x2}]
    extract/imag scan = scan [@{ovy}:@{ydim}]

    plot/oscan {rtbl(10)} :ROW "POSTSCAN (Y)" 1 {NICE_HCOPY} {iquad}
    fit/exponential scan {rtbl(11)} {rtbl(12)} {P4} >Null
    if {Q1} .ne. 0 then
      if {P4} .gt. 0 then
        write/out
        mess/out 2 {procid} "Exponential fit FAILED !\nUse MEDIAN instead !"
      endif

      statistic/image {ifrm} {OV_SEC} ? ? SN >Null
      sigBD = outputr(4)

      statistic/image scan [@1:@{dumy}] ? ? GN >Null
      compute/table {rtbl(11)} :FIT = {outputr(8)}
      compute/table {rtbl(11)} :RESIDUAL = :Y - :FIT
      {rtbl(12)},:RESULT,3 = outputr(8)
      append/out "POSTSCAN region  (Y-dir.): {OV_SEC}\n"
      append/out "Median of postscan region = {outputr(8)}"
      mess/out 3 {procid}
    else
      mess/out 3 {procid} "CHI^2 of exponential fit: {outputr(1)}"
      plot/oscan {rtbl(11)} :COLUMN "POSTSCAN (Y)" 2 {NICE_HCOPY} {iquad}
    endif
    create/column {rtbl(10)} :COLUMN
    write/table   {rtbl(10)} :COLUMN @1..{xd} -1
    create/column {rtbl(11)} :ROW
    write/table   {rtbl(11)} :ROW @1..{dumy} -1
    delete/image scan N

    by2 = m$value({rtbl(12)},:RESULT,3)
    bias  = bias + by2
    MERGE/TAB {rtbl(10)} {rtbl(11)} postscany
    ovdone = 1 + ovdone
    ovdony = 1
  endif
  ovy = dumy

  if ydir .ge. 1 then
    @p fors_combtab {prdony} {ovdony} y {rtbl(9)} {rtbl(12)}
  endif
!
! NO scan regions found --> ERROR
!
  if ovdone .eq. 0 .AND. prdone .eq. 0 then
    mess/out 1 {procid} "{errm}"
    return 2
  endif
  cnt = cnt - 1
  if cnt .gt. 0 then
    write/keyw BIAS_MEAN/r/1/4 {b{suffix}1},{b{suffix}2}
    write/keyw BIAS_SIGM/r/1/2 {sigAC},{sigBD}
    mess/out 3 {procid} "MEAN BIAS = {b{suffix}1}, {b{suffix}2} [A,B-port]"
    longlog/out 1 "     MEAN BIAS = {b{suffix}1}, {b{suffix}2} [A,B-port]"
    rename/table overscan{suffix} overscan{suffix}_AB
    @p fors_max {sigAC},{sigBD}
    sig = {Q1}
    goto loop
  else if num .gt. 1 then                       ! FAST readout (4-ports)
    sdev = -1.0                                 ! no std. deviation calculated
    write/keyw BIAS_MEAN/r/3/4 {b{suffix}1},{b{suffix}2}
    write/keyw BIAS_SIGM/r/3/2 {sigAC},{sigBD}
    mess/out 3 {procid} "MEAN BIAS = {b{suffix}1}, {b{suffix}2} [C,D-port]"
    longlog/out 1 "     MEAN BIAS = {b{suffix}1}, {b{suffix}2} [C,D-port]"
    rename/table overscan{suffix} overscan{suffix}_CD
    merge/table  overscan{suffix}_AB overscan{suffix}_CD overscan{suffix}
    write/descr  overscan{suffix}.tbl PORTS/i/1/1 4
    @p fors_max {sigAC},{sigBD}
    bias = {Q1}
    @p fors_max {sig},{bias}
    sig = {Q1}                                  !max of all 4 sigmas in ov.scan
    bias = (BIAS_MEAN(1) + BIAS_MEAN(2) + BIAS_MEAN(3) + BIAS_MEAN(4)) / 4.0

    copy/kd bias overscan{suffix}.tbl {h_proqcobia}/r/1/1
    copy/kd sig  overscan{suffix}.tbl {h_proqconoi}/r/1/1
    copy/kd bias_mean/r/1/1 overscan{suffix}.tbl {h_prqcobi(1)}/r/1/1
    copy/kd bias_mean/r/2/1 overscan{suffix}.tbl {h_prqcobi(2)}/r/1/1
    copy/kd bias_mean/r/3/1 overscan{suffix}.tbl {h_prqcobi(3)}/r/1/1
    copy/kd bias_mean/r/4/1 overscan{suffix}.tbl {h_prqcobi(4)}/r/1/1
    copy/kd bias_sigm/r/1/1 overscan{suffix}.tbl {h_prqcono(1)}/r/1/1
    copy/kd bias_sigm/r/2/1 overscan{suffix}.tbl {h_prqcono(2)}/r/1/1
    copy/kd bias_sigm/r/3/1 overscan{suffix}.tbl {h_prqcono(3)}/r/1/1
    copy/kd bias_sigm/r/4/1 overscan{suffix}.tbl {h_prqcono(4)}/r/1/1

    write/dhelp overscan{suffix}.tbl {h_proqcobia} "{i_proqcobia}"
    write/dhelp overscan{suffix}.tbl {h_proqconoi} "{i_proqconoi}"
    write/dhelp overscan{suffix}.tbl {h_prqcobi(1)} "{i_prqcobi}"
    write/dhelp overscan{suffix}.tbl {h_prqcobi(2)} "{i_prqcobi}"
    write/dhelp overscan{suffix}.tbl {h_prqcobi(3)} "{i_prqcobi}"
    write/dhelp overscan{suffix}.tbl {h_prqcobi(4)} "{i_prqcobi}"
    write/dhelp overscan{suffix}.tbl {h_prqcono(1)} "{i_prqcono}"
    write/dhelp overscan{suffix}.tbl {h_prqcono(2)} "{i_prqcono}"
    write/dhelp overscan{suffix}.tbl {h_prqcono(3)} "{i_prqcono}"
    write/dhelp overscan{suffix}.tbl {h_prqcono(4)} "{i_prqcono}"

    copy/kd bias {ifrm} {h_proqcobia}/r/1/1
    copy/kd sig  {ifrm} {h_proqconoi}/r/1/1
    copy/kd bias_mean/r/1/1 {ifrm} {h_prqcobi(1)}/r/1/1
    copy/kd bias_mean/r/2/1 {ifrm} {h_prqcobi(2)}/r/1/1
    copy/kd bias_mean/r/3/1 {ifrm} {h_prqcobi(3)}/r/1/1
    copy/kd bias_mean/r/4/1 {ifrm} {h_prqcobi(4)}/r/1/1
    copy/kd bias_sigm/r/1/1 {ifrm} {h_prqcono(1)}/r/1/1
    copy/kd bias_sigm/r/2/1 {ifrm} {h_prqcono(2)}/r/1/1
    copy/kd bias_sigm/r/3/1 {ifrm} {h_prqcono(3)}/r/1/1
    copy/kd bias_sigm/r/4/1 {ifrm} {h_prqcono(4)}/r/1/1

    write/dhelp {ifrm} {h_proqcobia} "{i_proqcobia}"
    write/dhelp {ifrm} {h_proqconoi} "{i_proqconoi}"
    write/dhelp {ifrm} {h_prqcobi(1)} "{i_prqcobi}"
    write/dhelp {ifrm} {h_prqcobi(2)} "{i_prqcobi}"
    write/dhelp {ifrm} {h_prqcobi(3)} "{i_prqcobi}"
    write/dhelp {ifrm} {h_prqcobi(4)} "{i_prqcobi}"
    write/dhelp {ifrm} {h_prqcono(1)} "{i_prqcono}"
    write/dhelp {ifrm} {h_prqcono(2)} "{i_prqcono}"
    write/dhelp {ifrm} {h_prqcono(3)} "{i_prqcono}"
    write/dhelp {ifrm} {h_prqcono(4)} "{i_prqcono}"

  else                                          ! SLOW readout (1 port)

    bias = bias / (prdone + ovdone)             ! calculate MEAN bias and rms
    if bx1 .ge. 0.0 sdev = sdev + (bx1 - bias) * (bx1 - bias)
    if bx2 .ge. 0.0 sdev = sdev + (bx2 - bias) * (bx2 - bias)
    if by1 .ge. 0.0 sdev = sdev + (by1 - bias) * (by1 - bias)
    if by2 .ge. 0.0 sdev = sdev + (by2 - bias) * (by2 - bias)
    sdev = sdev / (prdone + ovdone - 1)
    sdev = m$sqrt( sdev ) 
    mess/out 4 {procid} "MEAN BIAS = {bias} +/- {sdev}\n" 
    @p fors_max {sigAC},{sigBD}
    sig = {Q1}                                  !max of sigmas in pre-/postscan
    write/keyw BIAS_MEAN/r/1/4 {bias},{b{suffix}1}, {b{suffix}2},-1
    write/descr overscan{suffix}.tbl PORTS/i/1/1 1

    copy/kd bias overscan{suffix}.tbl {h_proqcobia}/r/1/1
    copy/kd sig  overscan{suffix}.tbl {h_proqconoi}/r/1/1
    copy/kd bias overscan{suffix}.tbl {h_prqcobi(1)}/r/1/1
    copy/kd sig  overscan{suffix}.tbl {h_prqcono(1)}/r/1/1

    write/dhelp overscan{suffix}.tbl {h_proqcobia} "{i_proqcobia}"
    write/dhelp overscan{suffix}.tbl {h_proqconoi} "{i_proqconoi}"
    write/dhelp overscan{suffix}.tbl {h_prqcobi(1)} "{i_prqcobi}"
    write/dhelp overscan{suffix}.tbl {h_prqcono(1)} "{i_prqcono}"

    copy/kd bias {ifrm} {h_proqcobia}/r/1/1
    copy/kd sig  {ifrm} {h_proqconoi}/r/1/1
    copy/kd bias {ifrm} {h_prqcobi(1)}/r/1/1
    copy/kd sig  {ifrm} {h_prqcono(1)}/r/1/1

    write/dhelp {ifrm} {h_proqcobia} "{i_proqcobia}"
    write/dhelp {ifrm} {h_proqconoi} "{i_proqconoi}"
    write/dhelp {ifrm} {h_prqcobi(1)} "{i_prqcobi}"
    write/dhelp {ifrm} {h_prqcono(1)} "{i_prqcono}"
  endif
end_loop:

copy/kd BIAS_MEAN/r/1/4 overscan{suffix}.tbl BIAS_MEAN/r/1/4
write/descr overscan{suffix}.tbl BIAS_STDV/r/1/1 {sdev}
write/descr overscan{suffix}.tbl {FORS_IDSTRNG}/c/1/20 {OVSCAN_ID}
write/descr overscan{suffix}.tbl {h_proqcsts}/c/1/9 "CHECKED"
write/dhelp overscan{suffix}.tbl {h_proqcsts} "{i_proqcsts}"

@p fors_paragraph 3
swap/units {ifrm} WORLD                         !restore world coordinates
add/icat overscO.cat overscan{suffix}.tbl >Null
classify/product overscan{suffix}.tbl {OVSCAN_ID} {d_qcp} {d_empty}
write/descriptor overscan{suffix}.tbl {h_instrume}/c/1/60 {{ifrm},{h_instrume}}
copy/dd overscan{suffix}.tbl {h_procatg} overscan{suffix}.tbl {fors_idstrng}
longlog/out 1 "     Overscan results in overscan{suffix}.tbl"
shortlog/out  1 "Bias from pre-/overscan regions determined."
finish/prohead overscO.cat 2
FINISH/PIPELINE overscO.cat {P3} CCDRED overscan{suffix}.tbl
if {Q1} .gt. 0 goto ABORT
finish/longlog overscI.cat overscO.cat overscan{suffix}.tbl 2 {P3}

delete/image scan N
delete/image {ifrm(1:{ifnm})}_scan N

finish:
set/format


LOG/REC fors_gen_qc0_overscan OUT
@p fors_paragraph 4

set/midas user=expert
delete/comm overscan/ccd
set/midas user=user
return 0


ABORT:
finish/longlog overscI.cat -1 ? 1 {P3}
mess/out 1 {procid} "Error in FINISH/PIPELINE !"
write/error 9,APP,NO "Error in FINISH/PIPELINE !"
