! $Id: fors_img_cal_calcoeff.prg,v 1.29 2001/06/28 09:37:42 sbogun Exp $
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
! .COPYRIGHT  (C) 2000  European Southern Observatory (ESO)
!
! .IDENT      fors_img_cal_calcoeff.prg
!
! .KEYWORDS   Calibration coefficients
!
! .PURPOSE    Pipeline wrapper and frame classification for
!             fors_calcoeff.prg:
!             Before the invocation of fors_calcoeff.prg the object bright-
!             nesses are converted from ADU to electrons.
!             Computation of magnitude zero point, extinction coefficient
!             and colour coefficient for each filter.
!             The different regression steps can be plotted.
! __________
!
! .USAGE      @p fors_img_cal_calcoeff in ref [out] ncoef colours thresh
!
! .INPUT      in     (P1): set of input frames (ascii list);
!                          needed frame type:
!                            aligned photometry tables
!                            (ALIGNED_PHOTOMETRY_TABLE);
!                              default: none
!
!             ref    (P2): super recipe mode: "+" (reference frames are
!                          included in P1),
!                          single recipe mode: set of reference frames (ascii
!                          list);
!                          needed frames, if ncoef=1 or ncoef=2 (else ignored):
!                            tables with colour coefficients (one per UBVRI
!                            Bessel filter)
!                            (COLCOEFF_TABLE_IMG),
!                          additionally needed frames, if ncoef=2 (else
!                          ignored):
!                            tables with extinction coefficients (one per UBVRI
!                            Bessel filter)
!                            (EXTCOEFF_TABLE_IMG),
!                          also needed frame type:
!                            table with purging information (DESC_PURGE_TABLE);
!                              default: none
!
!             out    (P3): single recipe mode: basename for ascii list with
!                          output frames,
!                          super recipe mode: again set of input frames and
!                          existing image catalogue (".cat"), separated by a
!                          comma;
!                          frames for input set:
!                            tables with zero points (one per UBVRI
!                            Bessel filter)
!                            (ZEROPOINT_TABLE_IMG),
!                          additional frames for input set, if ncoef=0 or
!                          ncoef=1:
!                            tables with extinction coefficients (one per UBVRI
!                            Bessel filter)
!                            (EXTCOEFF_TABLE_IMG),
!                          additional frames for input set, if ncoef=0:
!                            tables with colour coefficients (one per UBVRI
!                            Bessel filter)
!                            (COLCOEFF_TABLE_IMG);
!                              default: `PROID' defined in proc. (single mode)
!
!             ncoef  (P4): number of different types of input tables with
!                          calibration coefficients:
!                            |      => per filter:
!                            V      zero_point  extinction.coef.  colour.coef.
!                                + - - - - - - - - - - - - - - - - - - - - - -
!                           <0>  |  calculated  calculated        calculated
!                           <1>  |  calculated  calculated        input_tables
!                           <2>  |  calculated  input_tables      input_tables
!                            default: none
!
!             colours(P5): colours to be calculated from the std.mags. for the
!                          filters U/B/V/R/I (Bessel), separated by commas;
!                          the colours must be entered for all filters, even if
!                          not all filters should be present;
!                            default: none
!                            example: U-B,B-V,B-V,V-R,V-R
!
!             thresh (P6): If P4(ncoef)=[0|1] (else ignored):
!                          ---------------------------------
!                            1. minimum number of measurements which should
!                               have been used to derive the std.mags.
!                            2. binsize for determination of number of diffe-
!                               rent airmasses, >= 10^(-AIRACC) (see below)
!                            3. minimum number of different airmasses, >= 2
!                            4. minimum range of airmasses
!                            5. minimum number of stars within the thresholds
!                          and if P4(ncoef)=[0] (else ignored):
!                          -----------------------------------
!                            6. binsize for determination of number of diffe-
!                               rent colours, >= 10^(-COLACC) (see below)
!                            7. minimum number of different colours, >= 5
!                            8. minimum range of colours
!                          all separated by commas;
!                            default: none
!                            example: 5,0.1,2,0.5,3,0.2,5,1.0
!
! .OUTPUT     Tables with calibration coefficients; the number of different
!             types of tables depends on P4(ncoef) according to the following
!             scheme:
!                          |  "<PROID>_<filter>_
!                          |   //zero.tfits  //kext.tfits  //kcol.tfits"
!             - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!             P4(ncoef)=0  |        x             x             x
!                       1  |        x             x
!                       2  |        x
!
! .RETURN     Q1:  0: successful return
!                 -1: a warning message occured
!                  1: an error occured
! __________
!
! .ALGORITHM  LOOP WITH ALIGNED PHOTOMETRY TABLES:
!             a  Determination of used filter
!             b  Merging tables per filter
!        
!             LOOP WITH MERGED TABLES:
!             a  Calculation of colours
!             b  Calculation of standard minus instrument magnitudes
!             c  Selection of high quality objects (column FLAGS .ne. "Y")
!             d  Extinction coefficient:
!                NCOEF=[0|1]:
!                -----------
!                - Checking extinction thresholds / indexing of good obj.
!                - Calculation of extinction coef.s / zero points without
!                  colour correction per good object
!                - Calc. of mean ext.coef.
!                NCOEF=[2]:
!                ---------
!                - Getting ext.coeff. from input
!                NCOEF=[0|1]:
!                -----------
!                - Re-calc. of zero points without colour correction with
!                  mean ext.coef.
!                NCOEF=[2]:
!                ---------
!                - Calc. of zero points without colour correction with
!                  input ext.coef.
!             e  Colour coefficient / zero point:
!                NCOEF=[0]:
!                ---------
!                - Selection of good objects / checking colour thresholds
!                - Calc. of colour coef. / zero point with colour correc.
!                  with all good objects
!                NCOEF=[1]:
!                ---------
!                - Calc. of zero point with colour correction with input
!                  colour coef. with all good objects
!                NCOEF=[2]:
!                ---------
!                - Calc. of zero point with colour correction with input
!                  colour coef. with all objects
!                NCOEF=[0|1|2]:
!                -------------
!                - Calc. of all single zero points with extinction
!                  and colour correction
!
!             USED FORMULA:    /U\                       /col=P5(colours)[1]\
!                             | B |                     | col=P5(colours)[2] |
!             mag_lit -mag_mea| V | = z +k_ext*X +k_col*| col=P5(colours)[3] |
!                             | R |                     | col=P5(colours)[4] |
!                              \I/                       \col=P5(colours)[5]/
!
! .COMMENT    Filter identifications:
!             ----------------------
!             * Aligned photometry tables:
!               - Brightness conversion from ADU to electrons:
!                 The FITS keys ESO.INS.FILT1.NAME are checked and read !
!               - Loop with aligned tables in fors_calcoeff.prg:
!                 It is checked whether exactly one column with measured magni-
!                 tudes is present per table. Then the last character of the
!                 name of the found column determines the filter of this table
!                 (e.g. "MAG_V" is recognized as filter "V") !
!             * Tables w. colour coeffs. and tables w. extinction coeffs.:
!                 The FITS keys ESO.INS.FILT1.NAME are checked and read !
!
! .ENVIRON    FORS context
!
! .LANGUAGE   MIDAS 99NOV
!
! .AUTHOR     N.Rainer  DPG/DMD/ESO
!
! .VERSION    1.5  00/01/20  Creation
!                  00/       Release
!
!------------------------------------------------------------------------------
!
!
crossref in ref out ncoef colours thresh
!
!
define/local TOTFILT/i/1/1 5                        ! currently fixed to the 5
define/local ALLFILT/c/1/{TOTFILT} "UBVRI"          ! "UBVRI"
do inputi = 1 TOTFILT                               ! ..
  inputc = ALLFILT({inputi}:{inputi})               ! ..
  define/local FILT_{inputc}/c/1/9 "{inputc}_BESS"  ! Bessel filters - f. FORS2
  !                                                 ! R_BESS->R_SPECIAL below !
  !                                                 ! U_BESS->U_SPECIAL below !
enddo
define/local AIRACC/i/1/1 3 ? +lower_levels  ! #deci. f. airmass column
define/local COLACC/i/1/1 3 ? +lower_levels  ! #deci. f. col.column / col.coef.
!
!
define/local FMTDEF/c/1/17 "I1 E14.7,E23.15"
!
define/local PROID/c/1/40 "img_cal_calcoeff"
!
define/local CAT_PLOT/c/1/40 "plot.cat"
define/local warn/i/1/1 0
write/keyword q1 0
!
!
if p1 .eq. "??" then
  w/o "      ******************************************************************"
  w/o "      in     (P1): set of input frames (ascii list);"
  w/o "                   possible frame type:"
  w/o "                     aligned photometry tables"
  w/o "                     (ALIGNED_PHOTOMETRY_TABLE);"
  w/o "                       default: none"
  w/o " "
  w/o "      ref    (P2): super recipe mode: "+" (reference frames are"
  w/o "                   included in P1),"
  w/o "                   single recipe mode: set of reference frames (ascii"
  w/o "                   list);"
  w/o "                   needed frames, if ncoef=1 or ncoef=2 (else ignored):"
  w/o "                     tables with colour coefficients (one per UBVRI"
  w/o "                     Bessel filter)"
  w/o "                     (COLCOEFF_TABLE_IMG),"
  w/o "                   additionally needed frames, if ncoef=2 (else"
  w/o "                   ignored):"
  w/o "                     tables with extinction coefficients (one per UBVRI"
  w/o "                     Bessel filter)"
  w/o "                     (EXTCOEFF_TABLE_IMG),"
  w/o "                   also needed frame type:"
  w/o "                     table with purging information (DESC_PURGE_TABLE);"
  w/o "                       default: none"
  w/o " "
  w/o "      out    (P3): single recipe mode: basename for ascii list with"
  w/o "                   output frames,"
  w/o "                   super recipe mode: again set of input frames and"
  w/o "                   existing image catalogue (".cat"), separated by a"
  w/o "                   comma;"
  w/o "                   frames for input set:"
  w/o "                     tables with zero points (one per UBVRI"
  w/o "                     Bessel filter)"
  w/o "                     (ZEROPOINT_TABLE_IMG),"
  w/o "                   additional frames for input set, if ncoef=0 or"
  w/o "                   ncoef=1:"
  w/o "                     tables with extinction coefficients (one per UBVRI"
  w/o "                     Bessel filter)"
  w/o "                     (EXTCOEFF_TABLE_IMG),"
  w/o "                   additional frames for input set, if ncoef=0:"
  w/o "                     tables with colour coefficients (one per UBVRI"
  w/o "                     Bessel filter)"
  w/o "                     (COLCOEFF_TABLE_IMG);"
  w/o "                       default: `PROID' defined in proc. (single mode)"
  w/o " "
  w/o "      ncoef  (P4): number of different types of input tables with"
  w/o "                   calibration coefficients:"
  w/o "                     |      => per filter:"
  w/o "                     V      zero_point  extinction.coef.  colour.coef."
  w/o "                         + - - - - - - - - - - - - - - - - - - - - - -"
  w/o "                    <0>  |  calculated  calculated        calculated"
  w/o "                    <1>  |  calculated  calculated        input_tables"
  w/o "                    <2>  |  calculated  input_tables      input_tables"
  w/o "                     default: none"
  w/o " "
  w/o "      colours(P5): colours to be calculated from the std.mags. for the"
  w/o "                   filters U/B/V/R/I, separated by commas;"
  w/o "                     default: none"
  w/o "                     example: U-B,B-V,B-V,V-R,V-R"
  w/o " "
  w/o "      thresh (P6): If P4(ncoef)=[0|1] (else ignored):"
  w/o "                   ---------------------------------"
  w/o "                     1. minimum number of measurements which should"
  w/o "                        have been used to derive the std.mags."
  w/o "                     2. binsize for determination of number of diffe-"
  w/o "                        rent airmasses, >= 10^(-AIRACC) (see below)"
  w/o "                     3. minimum number of different airmasses, >= 2"
  w/o "                     4. minimum range of airmasses"
  w/o "                     5. minimum number of stars within the thresholds"
  w/o "                   and if P4(ncoef)=[0] (else ignored):"
  w/o "                   -----------------------------------"
  w/o "                     6. binsize for determination of number of diffe-"
  w/o "                        rent colours, >= 10^(-COLACC) (see below)"
  w/o "                     7. minimum number of different colours, >= 5"
  w/o "                     8. minimum range of colours"
  w/o "                   all separated by commas;"
  w/o "                     default: none"
  w/o "                     example: 5,0.1,2,0.5,3,0.2,2,1.0"
  w/o "      ******************************************************************"
  return/exit
endif
!
!          PREPARATIONS FOR RECIPE :
!          ***********************
!
        mess/out 3 {PROID} "Determine calibration coefficients ..."
!
@p fors_paragraph 1 1  ! -------- INITIALIZATION --------
!
log/recipe {PROID} in
!
set/format {FMTDEF}
!
!
!          INPUT :
!          *****
!
!P1(in):
!------
if p1 .eq. "?" mess/out 2 {PROID} " "
define/parameter p1 ? ? -
  "        Enter name of input frame set: "
define/local inpipe_frm/c/1/76 "{p1}"
!
if m$exist(inpipe_frm) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered list is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({inpipe_frm})"
  goto error
endif
!
define/local filecode/i/1/2 0 all
open/file {inpipe_frm} read filecode
if {filecode(1)} .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Unable to open entered list !"
        mess/out 1 {PROID} "({inpipe_frm})"
  goto error
endif
read/file {filecode(1)} outputc
if {filecode(2)} .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Entered list is empty !"
        mess/out 1 {PROID} "({inpipe_frm})"
  close/file {inpipe_frm}
  goto error
endif
close/file {filecode(1)}
!
!
!P2(ref):
!-------
if p2 .eq. "?" mess/out 2 {PROID} " "
define/parameter p2 ? ? -
  "        Enter name of reference frame set: "
define/local inpipe_ref/c/1/76 "{p2}"
!
! No tests like for P1 - in super recipes reference frames are included in P1 .
!
!
!P3(out):
!-------
define/parameter p3 {PROID} ? -
  "        Enter basename of out.list(single)/exist.cat.".cat"(super): "
define/local outpipe/c/1/76 "{p3}"
!
!
!P4(ncoef):
!---------
if p4 .eq. "?" .or. p4 .eq. "+" then
        write/out " "
        mess/out 1 {PROID} "No mode has been entered (P4) !"
  goto error
endif
define/parameter p4 ? n -
  "Mode = no. of input coeffs. [def.:<0>|col=<1>|col+ext=<2>]: "
define/local ncoef/i/1/1 {p4}
if ncoef .lt. 0 .or. ncoef .gt. 2 then
        write/out " "
        mess/out 1 {PROID} "Entered mode out of range (P4) !"
  goto error
endif
!
!
!P5(colours):
!-----------
outputi = m$parse(p5,"colour")
if p5 .eq. "?" .or. p5 .eq. "+" then
        write/out " "
        mess/out 1 {PROID} -
          "No colours have been entered (P5) - 5 needed !"
  goto error
elseif outputi .ne. 5 then
        write/out " "
        mess/out 1 {PROID} -
          "Wrong number of colours(P5): {outputi} entered - 5 needed !"
  goto error
endif
define/parameter p5 ? ? -
  "5 colours (see procedure header): "
define/local colU/c/1/16 "{colour01}"; if colU(1:1) .eq. ":" colU = colU(2:)
define/local colB/c/1/16 "{colour02}"; if colB(1:1) .eq. ":" colB = colB(2:)
define/local colV/c/1/16 "{colour03}"; if colV(1:1) .eq. ":" colV = colV(2:)
define/local colR/c/1/16 "{colour04}"; if colR(1:1) .eq. ":" colR = colR(2:)
define/local colI/c/1/16 "{colour05}"; if colI(1:1) .eq. ":" colI = colI(2:)
!
define/local badcol/i/1/1 0
if m$len(colU) .lt. 3 badcol = 1
if m$len(colB) .lt. 3 badcol = 1
if m$len(colV) .lt. 3 badcol = 1
if m$len(colR) .lt. 3 badcol = 1
if m$len(colI) .lt. 3 badcol = 1
if badcol .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "At least one colour string (P5) is too short !"
        mess/out 1 {PROID} "(Characters.: 1.<filterX> 2.<-> 3.<filterY>)"
  goto error
endif
if m$index(ALLFILT,colU(1:1)) .eq. 0 .or. -
   m$index(ALLFILT,colU(3:3)) .eq. 0 badcol = 1
if m$index(ALLFILT,colB(1:1)) .eq. 0 .or. -
   m$index(ALLFILT,colB(3:3)) .eq. 0 badcol = 1
if m$index(ALLFILT,colV(1:1)) .eq. 0 .or. -
   m$index(ALLFILT,colV(3:3)) .eq. 0 badcol = 1
if m$index(ALLFILT,colR(1:1)) .eq. 0 .or. -
   m$index(ALLFILT,colR(3:3)) .eq. 0 badcol = 1
if m$index(ALLFILT,colI(1:1)) .eq. 0 .or. -
   m$index(ALLFILT,colI(3:3)) .eq. 0 badcol = 1
if badcol .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "At least one colour string (P5) has a bad filter !"
        mess/out 1 {PROID} "(Only <{ALLFILT}> are possible)"
  goto error
endif
!
!
!P6(thresh):
!----------
if ncoef .le. 1 then
  !
  outputi = m$parse(p6,"limit")
  if ncoef .eq. 0 then
    inputi(1) = 8
    inputi(2) = 8
  elseif ncoef .eq. 1 then
    inputi(1) = 5
    inputi(2) = 8  ! allowed, but #6-8 unused
  endif
  if p6 .eq. "?" .or. p6 .eq. "+" then
        write/out " "
        mess/out 1 {PROID} -
          "No thresholds have been entered (P6) - {inputi(1)} needed !"
    goto error
  elseif outputi .ne. inputi(1) .and. outputi .ne. inputi(2) then
        write/out " "
        mess/out 1 {PROID} -
        "Wrong number of thresh.(P6): {outputi} entered - {inputi(1)} needed !"
    goto error
  endif
  !
  define/parameter p6 ? ? -
    "[8|5] thresholds (see procedure header): "
  ! only for initialize/prohead and initialize/longlog:
  define/local thresh/c/1/80 "{p6}"
  ! actual input:                                            ! ALLOWED :
  define/local min_stdmea/i/1/1 {limit01} ? +lower_levels    ! all
  define/local airbinsize/r/1/1 {limit02} ? +lower_levels    ! >=10.^(-AIRACC)
  define/local min_noairs/i/1/1 {limit03} ? +lower_levels    ! >=2
  define/local min_airrng/r/1/1 {limit04} ? +lower_levels    ! all
  define/local min_nstars/i/1/1 {limit05} ? +lower_levels    ! all
  if ncoef .eq. 0 then
    define/local colbinsize/r/1/1 {limit06} ? +lower_levels  ! >=10.^(-COLACC)
    define/local min_nocols/i/1/1 {limit07} ? +lower_levels  ! >=5
    define/local min_colrng/r/1/1 {limit08} ? +lower_levels  ! all
  endif
  !
  define/local badthr/i/1/1 0
  inputr = 10.**(-AIRACC)
  if airbinsize .lt. inputr badthr = 1
  if min_noairs .lt. 2 badthr = 1
  if ncoef .eq. 0 then
    inputr = 10.**(-COLACC)
    if colbinsize .lt. inputr badthr = 1
    if min_nocols .lt. 5 badthr = 1
  endif
  if badthr .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "At least one threshold (P6) is out of range !"
    goto error
  endif
  !
endif
!
!
!          DEFINITION OF VARIABLES :
!          ***********************
!
!     "Identifying input frames":
!
  define/local aptID/c/1/25 " " all
  define/local colID/c/1/25 " " all
  define/local extID/c/1/25 " " all
  define/local in/i/1/1 0
  define/local filtloop/i/1/1 0
  define/local filter/c/1/1 " "
  do filtloop = 1 TOTFILT
    filter = ALLFILT({filtloop}:{filtloop})
    define/local refcol{filter}/i/1/1 0
    define/local refext{filter}/i/1/1 0
    define/local coltab{filter}/c/1/76 " "
    define/local exttab{filter}/c/1/76 " "
  enddo
  define/local catal/i/1/1 0
  define/local id/c/1/25 " " all
  define/local topicima/c/1/76 " " all
  define/local once/i/1/1 0
  define/local filtername/c/1/25 " " all
  define/local refcol/i/1/1 0
  define/local refext/i/1/1 0
  define/local insdesc/c/1/60 " " all
!
!     "Converting brightnesses":
!
  define/local topictab/c/1/76 " " all
  define/local portloop/i/1/1 0
  define/local cftopic/d/1/1 0.
  define/local cfsum/d/1/1 0.
  define/local cf/d/1/1 0.
  define/local filt/c/1/1 " "
  define/local async/i/1/1 0
!
!     "Calling of procedure":
!
  define/local label_a/c/1/97 " " all
  define/local label_b/c/1/97 " " all
  define/local label_c/c/1/97 " " all
  define/local label_d/c/1/97 " " all
  define/local label_e/c/1/97 " " all
  define/local label_f/c/1/97 " " all
  define/local thresh_a/c/1/97 " " all
  define/local thresh_b/c/1/97 " " all
  define/local coltabs/c/1/97 " " all
  define/local exttabs/c/1/97 " " all
  define/local zerocat/c/1/64 " " all
  define/local kextcat/c/1/64 " " all
  define/local kcolcat/c/1/64 " " all
!
!
!          LOADING OF GLOBAL KEYS (SINGLE RECIPE MODE) ,
!          CONVERSION TFITS->TBL ,
!          CLASSIFICATION OF INPUT FRAMES :
!          *********************************************
!
        mess/out 4 {PROID} "Getting input frames"
!
initialize/pipeline {inpipe_frm} {inpipe_ref} -
  inpipe_good.cat ? ? + inpipe_bad.cat ?
if {q1} .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Pipeline initialization failed !"
  goto error
endif
-delete inpipe_bad.cat
!
! Initialize plot file catalog
initialize/plot {cat_plot} 2

! Initialize product header:
if ncoef .le. 1 then
  initialize/prohead {inpipe_frm} {inpipe_ref} -
    2,"Compute calibration coefficients" ncoef,{ncoef} thresh,{thresh}
elseif ncoef .eq. 2 then
  initialize/prohead {inpipe_frm} {inpipe_ref} -
    2,"Compute calibration coefficients" ncoef,{ncoef}
endif
if {q1} .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Product header initialization failed !"
  goto error
endif
!
if ncoef .le. 1 then
  initialize/longlog {inpipe_frm} {inpipe_ref} 2 -
           "calc. mode: {ncoef}" -
           "colours: {colU},{colB},{colV},{colR},{colI} (for U,B,V,R,I resp.)" -
           "thresholds: {thresh}"
elseif ncoef .eq. 2 then
  initialize/longlog {inpipe_frm} {inpipe_ref} 2 -
                   "calc. mode: {ncoef}" -
              "filters>colours: U>{colU},B>{colB},V>{colV},R>{colR},I>{colI}"
endif
!
!
!          IDENTIFICATION VIA DESCRIPTORS ,
!          MAPPING OF FRAMES TO PARAMETERS  :
!          ********************************
!
        mess/out 4 {PROID} "Identifying input and reference frames"
!
aptID = "{PHOTTALI_ID}"
colID = "{COLCOEFF_ID}"
extID = "{EXTCOEFF_ID}"
!
create/icat {PROID}.cat null {FORS_IDSTRNG} >Null
in = 0
do filtloop = 1 TOTFILT
  filter = ALLFILT({filtloop}:{filtloop})
  refcol{filter} = 0
  refext{filter} = 0
  coltab{filter} = "?"
  exttab{filter} = "?"
enddo
!
once = 0
ident_loop:
  store/frame topicima inpipe_good.cat 1 end_ident_loop
  id = "{{topicima},{fors_idstrng}}"
  !
  !--- FOR FORS2 ONLY - BEGIN ---
  if once .eq. 0 then
    if id .eq. "{aptID}" once = 1
    if id .eq. "{colID}" once = 1
    if id .eq. "{extID}" once = 1
    if once .eq. 1 then
      insdesc = "{{topicima},{h_instrume}}"
      if insdesc .eq. "FORS2" then
        FILT_R = "R_SPECIAL"
        FILT_U = "U_SPECIAL"
        mess/out 2 {PROID} -
          "For FORS2 the R and U filters are called <R_SPECIAL> and <U_SPECIAL> !"
        warn = 1
      endif
    endif
  endif
  !--- FOR FORS2 ONLY - END ---
  !
  if id .eq. "{aptID}" then
    if m$existd(topicima,h_coll) .eq. 1 then
      if "{{topicima},{h_coll}}" .ne. "COLL_SR" then
          mess/out 2 {PROID} "Wrong collimator - only "COLL_SR" is allowed -"
          mess/out 2 {PROID} "skipping {topicima} !"
          warn = 1
          goto ident_loop
      endif
    else
          mess/out 2 {PROID} "Descriptor with collimator name is missing"
          mess/out 2 {PROID} "({h_coll}) -"
          mess/out 2 {PROID} "skipping {topicima} !"
          warn = 1
          goto ident_loop
    endif
    add/icat {PROID}.cat {topicima} >Null
        mess/out 4 {PROID} "{topicima} : {id} => input"
    in = in+1
  else
    !
    if ncoef .eq. 1 .or. ncoef .eq. 2 then
      if id .eq. "{colID}" then
        if m$existd(topicima,h_filti(1)) .eq. 1 then
          filtername = "{{topicima},{h_filti(1)}}"
        else
          mess/out 2 {PROID} "Descriptor with filter name is missing"
          mess/out 2 {PROID} "({h_filti(1)}) -"
          mess/out 2 {PROID} "skipping {topicima} !"
          warn = 1
          goto ident_loop
        endif
        do filtloop = 1 TOTFILT
          filter = ALLFILT({filtloop}:{filtloop})
          if filtername .eq. FILT_{filter} then
            coltab{filter} = topicima
            mess/out 4 {PROID} "{topicima} : {id} - <{filter}> => reference"
            refcol{filter} = refcol{filter}+1
          endif
        enddo
      endif
    endif
    !
    if ncoef .eq. 2 then
      if id .eq. "{extID}" then
        if m$existd(topicima,h_filti(1)) .eq. 1 then
          filtername = "{{topicima},{h_filti(1)}}"
        else
          mess/out 2 {PROID} "Descriptor with filter name is missing"
          mess/out 2 {PROID} "({h_filti(1)}) -"
          mess/out 2 {PROID} "skipping {topicima} !"
          warn = 1
          goto ident_loop
        endif
        do filtloop = 1 TOTFILT
          filter = ALLFILT({filtloop}:{filtloop})
          if filtername .eq. FILT_{filter} then
            exttab{filter} = topicima
            mess/out 4 {PROID} "{topicima} : {id} - <{filter}> => reference"
            refext{filter} = refext{filter}+1
          endif
        enddo
      endif
    endif
    !
  endif
  !
  goto ident_loop
end_ident_loop:
-delete inpipe_good.cat
!
if in .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No aligned photometry table present !"
  goto error
endif
if ncoef .eq. 1 .or. ncoef .eq. 2 then
  do filtloop = 1 TOTFILT
    filter = ALLFILT({filtloop}:{filtloop})
    refcol = refcol + refcol{filter}
  enddo
  if refcol .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No table with a colour coefficient present !"
    goto error
  endif
  inputc = " "
  do filtloop = 1 TOTFILT
    filter = ALLFILT({filtloop}:{filtloop})
    if refcol{filter} .gt. 1 inputc = inputc // "{filter}"
  enddo
  if m$len(inputc) .gt. 0 then
        write/out " "
        mess/out 1 {PROID} -
          "Found >1 tbl. w. colour coef. f. filter <{inputc}> !"
    goto error
  endif
endif
if ncoef .eq. 2 then
  do filtloop = 1 TOTFILT
    filter = ALLFILT({filtloop}:{filtloop})
    refext = refext + refext{filter}
  enddo
  if refext .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No table with an extinction coefficient present !"
    goto error
  endif
  inputc = " "
  do filtloop = 1 TOTFILT
    filter = ALLFILT({filtloop}:{filtloop})
    if refext{filter} .gt. 1 inputc = inputc // filter
  enddo
  if m$len(inputc) .gt. 0 then
        write/out " "
        mess/out 1 {PROID} -
          "Found >1 tbl. w. extin. coef. f. filter <{inputc}> !"
    goto error
  endif
endif
!
insdesc = "{#1_{PROID}.cat,{h_instrume}}"
!
!
!          CONVERTING BRIGHTNESSES FROM `ADU' TO `ELECTRONS'
!          (MEAN CONVERSION FACTOR FOR MULTIPLE PORTS) :
!          *************************************************
!
@p fors_paragraph 2 1  ! -------- PROCESSING --------
!
        mess/out 3 {PROID} -
          "Converting photometry tables from ADU to electrons"
!
catal = 0
conv_loop:
  store/frame topictab {PROID}.cat 1 end_conv_loop
        mess/out 3 {PROID} "  photometry table: {topictab}"
  !
  !**************************
  @p fors_getports {topictab}
  !**************************
  if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Determination of ports failed !"
    goto error
  endif
  cfsum = 0.
  do portloop = 1 fors_nout
    cftopic = {{topictab},{h_confac({fors_outi({portloop})})}}
    if cftopic .le. 0. then
        set/format F15.14
        write/out " "
        mess/out 1 {PROID} "Impossible value for descriptor"
        mess/out 1 {PROID} "`{h_confac({fors_outi({portloop})})}': {cftopic} !"
        set/format {FMTDEF}
      goto error
    endif
    cfsum = cfsum + cftopic
  enddo
  cf = cfsum / fors_nout
        set/format F15.14
      if fors_nout .gt. 1 then
        mess/out 3 {PROID} "    mean conversion factor: {cf}"
        mess/out 3 {PROID} "    readout ports: {fors_nout}"
      else
        mess/out 3 {PROID} "    conversion factor: {cf}"
      endif
        set/format {FMTDEF}
  !
  !***************************
  @p fors_getfilter {topictab}
  !***************************
  outputc = fors_filti(1)
  filt = outputc(1:1)
        mess/out 3 {PROID} "    filter: {filt}"
  if m$existc(topictab,":{c_flx}_{filt}") .eq. 0 then
        mess/out 2 {PROID} "    Column :{c_flx}_{filt} is missing -"
        mess/out 2 {PROID} "    skipping topical ali.phot.tbl. !"
    subtract/icat {PROID}.cat {topictab} >Null
    async = 1
    warn = 1
    goto conv_loop
  elseif m$existc(topictab,":{c_mag}_{filt}") .eq. 0 then
        mess/out 2 {PROID} "    Column :{c_mag}_{filt} is missing -"
        mess/out 2 {PROID} "    skipping topical ali.phot.tbl. !"
    subtract/icat {PROID}.cat {topictab} >Null
    async = 1
    warn = 1
    goto conv_loop
  elseif m$existc(topictab,":SKY") .eq. 0 then
        mess/out 2 {PROID} "    Column :SKY is missing -"
        mess/out 2 {PROID} "    skipping topical ali.phot.tbl. !"
    subtract/icat {PROID}.cat {topictab} >Null
    async = 1
    warn = 1
    goto conv_loop
  endif
  !
  compute/table {topictab} :{c_flx}_{filt} = :{c_flx}_{filt}*{cf}
  compute/table {topictab} :{c_mag}_{filt} = -
                                   :{c_mag}_{filt}-2.5*log10({cf})
  compute/table {topictab} :SKY = :SKY*{cf}
  name/column {topictab} :{c_flx}_{filt} ? "e-/sec"
  name/column {topictab} :{c_mag}_{filt} ? "e-/sec"
  name/column {topictab} :SKY ? "e-/sec"
  !
  goto conv_loop
end_conv_loop:
!
show/icat {PROID}.cat n
if outputi .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No aligned photometry table left !"
  goto error
elseif async .eq. 1 then
  @a catalsync {PROID}.cat
endif
!
!
!          *********************************
!          *                               *
!          *  CALLING OF MAIN PROCEDURE :  *
!          *                               *
!          *********************************
!
label_a = "{c_aptobj}"
label_b = "{c_aptmeaU},{c_aptmeaB},{c_aptmeaV},{c_aptmeaR},{c_aptmeaI}"
label_c = "{c_aptair},{c_aptmjd},{c_aptflg}"
label_d = "{c_aptstdU},{c_aptstdB},{c_aptstdV},{c_aptstdR},{c_aptstdI}"
label_e = "{colU},{colB},{colV},{colR},{colI}"
if ncoef .eq. 0 .or. ncoef .eq. 1 then
  label_f = "{c_aptNstd}"
  thresh_a = "{min_stdmea},{airbinsize},{min_noairs},{min_airrng},{min_nstars}"
endif
if ncoef .eq. 0 -
  thresh_b = "{colbinsize},{min_nocols},{min_colrng}"
if ncoef .eq. 1 .or. ncoef .eq. 2 -
  coltabs = "{coltabU},{coltabB},{coltabV},{coltabR},{coltabI}"
if ncoef .eq. 2 -
  exttabs = "{exttabU},{exttabB},{exttabV},{exttabR},{exttabI}"
!
        mess/out 4 {PROID} "Applying procedure "fors_calcoeff" :"
!
!*****************************************************************
if ncoef .eq. 0 then
  !
  @p fors_calcoeff 0 {PROID}.cat -
     {label_a},{label_b},{label_c},{label_d},{label_e},{label_f} -
     {thresh_a},{thresh_b} ? ? p n
  !
elseif ncoef .eq. 1 then
  !
  @p fors_calcoeff 1 {PROID}.cat -
     {label_a},{label_b},{label_c},{label_d},{label_e},{label_f} -
     {thresh_a} {coltabs},1 ? p n
  !
elseif ncoef .eq. 2 then
  !
  @p fors_calcoeff 2 {PROID}.cat -
     {label_a},{label_b},{label_c},{label_d},{label_e} -
     ? {coltabs},1 {exttabs},1 p n
  !
endif
!*****************************************************************
!
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Determination of calibration coefficients failed !"
  goto error
endif
!
outputi = m$parse(out_a,"cats")
if ncoef .eq. 0 inputi = 3
if ncoef .eq. 1 inputi = 2
if ncoef .eq. 2 inputi = 1
if outputi .ne. inputi then
        write/out " "
        mess/out 1 {PROID} "Wrong number of elements in OUT_A !"
        mess/out 1 {PROID} "(= output from fors_calcoeff.prg)"
  goto error
endif
zerocat = cats01
if ncoef .eq. 0 .or. ncoef .eq. 1 then
  kextcat = cats02
endif
if ncoef .eq. 0 then
  kcolcat = cats03
endif
!
! Add plot file to plot catalog
if out_b(1:1) .ne. " " then     !we have a plot file
  if m$exist(out_b) .ne. 0 then
    add/acat {CAT_PLOT} {out_b} >Null
  endif
endif

shortlog/out 1 "Calibration coefficients computed"
!
!
!          UPDATING DESCRIPTORS OF OUT- AND INPUT FRAMES :
!          *********************************************
!
@p fors_paragraph 3 1  ! -------- TERMINATION --------
!
        mess/out 4 {PROID} "Classifying output tables"
        mess/out 4 {PROID} "(*_zero.tbl[, *_kext.tbl[, *_kcol.tbl]])"
!
catal = 0
zero_loop:
  store/frame topicima {zerocat} 1 end_zero_loop
  classify/product {topicima} {ZEROPOINT_ID} {d_pre} {d_pho},{d_det} "OK"
  write/descriptor {topicima} {h_instrume}/c/1/60 {insdesc}
  write/descriptor {topicima} {FORS_IDSTRNG}/c/1/60 {ZEROPOINT_ID} !surec.
  write/descriptor {topicima} {h_proqcsts}/c/1/9 "CHECKED"
  write/dhelp {topicima} {h_proqcsts} "{i_proqcsts}"
    inputi = m$indexb(topicima,"_zero.tbl")-1  ! = filter
    inputc = topicima({inputi}:{inputi})
  write/descriptor {topicima} {h_filti(1)}/c/1/8 "{FILT_{inputc}}"
  write/dhelp {topicima} {h_filti(1)} "{i_filti}"
  goto zero_loop
end_zero_loop:
!
if ncoef .eq. 0 .or. ncoef .eq. 1 then
  catal = 0
  kext_loop:
    store/frame topicima {kextcat} 1 end_kext_loop
    classify/product {topicima} {EXTCOEFF_ID} {d_pre} {d_pho},{d_det} "OK"
    write/descriptor {topicima} {h_instrume}/c/1/60 {insdesc}
    write/descriptor {topicima} {FORS_IDSTRNG}/c/1/60 {EXTCOEFF_ID} !surec.
    write/descriptor {topicima} {h_proqcsts}/c/1/9 "CHECKED"
    write/dhelp {topicima} {h_proqcsts} "{i_proqcsts}"
      inputi = m$indexb(topicima,"_kext.tbl")-1  ! = filter
      inputc = topicima({inputi}:{inputi})
    write/descriptor {topicima} {h_filti(1)}/c/1/8 "{FILT_{inputc}}"
    write/dhelp {topicima} {h_filti(1)} "{i_filti}"
    goto kext_loop
  end_kext_loop:
endif
!
if ncoef .eq. 0 then
  catal = 0
  kcol_loop:
    store/frame topicima {kcolcat} 1 end_kcol_loop
    classify/product {topicima} {COLCOEFF_ID} {d_pre} {d_pho},{d_det} "OK"
    write/descriptor {topicima} {h_instrume}/c/1/60 {insdesc}
    write/descriptor {topicima} {FORS_IDSTRNG}/c/1/60 {COLCOEFF_ID} !surec.
    write/descriptor {topicima} {h_proqcsts}/c/1/9 "CHECKED"
    write/dhelp {topicima} {h_proqcsts} "{i_proqcsts}"
      inputi = m$indexb(topicima,"_kcol.tbl")-1  ! = filter
      inputc = topicima({inputi}:{inputi})
    write/descriptor {topicima} {h_filti(1)}/c/1/8 "{FILT_{inputc}}"
    write/dhelp {topicima} {h_filti(1)} "{i_filti}"
    goto kcol_loop
  end_kcol_loop:
endif
!
        mess/out 4 {PROID} "Endorsing QC status in input tables"
!
catal = 0
desc_loop_in:
  store/frame topicima {PROID}.cat 1 end_desc_loop_in
  !
  write/descriptor {topicima} {h_proqcsts}/c/1/9 "CHECKED"
  write/dhelp {topicima} {h_proqcsts} "{i_proqcsts}"
  !
  goto desc_loop_in
end_desc_loop_in:
!
!
!          UPDATING OF OUTPUT CATALOGUE (SUPER RECIPE MODE) ,
!          CONVERSION TBL->TFITS (SINGLE MODE) :
!          **************************************************
!
        mess/out 4 {PROID} "Releasing output tables"
!
create/icat outprog.cat null {FORS_IDSTRNG} >Null
add/icat outprog.cat {zerocat} >Null
if ncoef .eq. 0 .or. ncoef .eq. 1 then
  add/icat outprog.cat {kextcat} >Null
endif
if ncoef .eq. 0 then
  add/icat outprog.cat {kcolcat} >Null
endif
!
finish/prohead outprog.cat 2
!
catal = 0  ! => store/frame gets "catal+1 = 1st_entry"
store/frame outputc outprog.cat
finish/pipeline outprog.cat {outpipe} + {outputc} {CAT_PLOT}
!
catal = 0  ! => store/frame gets "catal+1 = 1st_entry"
store/frame outputc outprog.cat
finish/longlog {PROID}.cat outprog.cat {outputc} 2 {outpipe}
!
-delete {PROID}.cat
-delete outprog.cat
!
!
!          RESETTINGS AFTER RECIPE :
!          ***********************
!
set/format
!
log/recipe {PROID} out
!
@p fors_paragraph 4 1  ! ================
!
!
if warn .eq. 0 then
  return 0
else
  return -1
endif
!
error:
  set/format
  finish/longlog ? -1 ? 2
  write/out " "
  return 1
!
!
!-------------------------------------------------------------------------------
