! $Id: fors_img_obs_photometry.prg,v 1.194 2001/06/15 14:15:38 sbogun Exp $   
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
! .COPYRIGHT  (C) 1998  European Southern Observatory (ESO)
!
! .IDENT       $RCSfile: fors_img_obs_photometry.prg,v $
!
! .KEYWORDS   Photometry
!
! .PURPOSE    Pipeline wrapper and frame classification for
!             fors_photometry.prg:
!
!             A) Before the invocation of fors_photometry.prg the input frames
!                are normalized by their exposure times. If an overexposure
!                mask is present for an input frame, it is applied. Thus pixels
!                which were overexposed (before the basic reductions) are reset
!                to their original values and objects with such pixels are
!                flagged.
!
!             B) Now, by means of the SExtractor object extraction software,
!                object positions (pixel and WCS), sky background frame, sky-
!                subtracted frame and brightnesses of each object are computed.
!                SExtractor works as explained in .ALGORITHM under B).
!
!             C) Afterwards generally objects with properties which are ex-
!                plained in .ALGORITHM under C) are REMOVED.
!
!             D) Furthermore signal quality values, and that signal-to-noise,
!                limiting magnitude, seeing (FWHM) and trailing (ellipticity),
!                are calculated.
!                The values are SELECTED and calculated as explained in
!                .ALGORITHM under D).
!
!             E) Additionally the weighted airmass is computed.
! __________
!
! .USAGE      @p fors_img_obs_photometry in ref [out]
!
! .INPUT      in   (P1): set of input frames (ascii list);
!                        possible frame types:
!                          reduced standard star images (STANDARD_REDUCED_IMG),
!                          reduced science images (SCIENCE_REDUCED_IMG);
!                            default: none
!
!             ref  (P2): super recipe mode: "+" (reference frames are included
!                        in P1),
!                        single recipe mode: set of reference frames (ascii
!                        list);
!                        needed frames:
!                          table with SExtractor input parameters
!                          (PHOTOMETRY_INPUT_PAR),
!                          table with SExtractor convolution mask
!                          (PHOTOMETRY_CONVOL_MASK),
!                          table with SExtractor output parameters
!                          (PHOTOMETRY_OUTPUT_PAR);
!                          table with neuronal network weights
!                          (PHOTOMETRY_NEURON_PAR);
!                        optional:
!                          mask frame with overexposed pixels
!                          (OVEREXPO_MASK);
!                        also needed frame type:
!                          table with purging information (DESC_PURGE_TABLE);
!                            default: none
!
!             out  (P3): single recipe mode: basename for ascii list with
!                        output frames,
!                        super recipe mode: again set of input frames and
!                        existing image catalogue (".cat"), separated by a
!                        comma;
!                        frames for each input image:
!                          sky background images (PHOT_BACKGROUND_IMG),
!                          sky-subtracted images (PHOT_BACKSUBTR_IMG);
!                        tables for each input image with found objects:
!                          positions/brightnesses/object_quality/airmass
!                          (PHOTOMETRY_TABLE);
!                            default: `PROID' defined in procedure (single mode)
!
! .OUTPUT     "<P1-entries>_skymap.fits":  sky background frames
!
!             "<P1-entries>_skysub.fits":  sky-subtracted frames
!
!             If objects were found:
!
!             "<P1-entries>_phot.tfits": tables with positions and brightnesses
!                                        of objects (for standard star input
!                                        frames only high quality objects are
!                                        used) :
!
!             - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
!                               Columns:  1 XPIX,
!                                         2 YPIX,
!                                         3 RA,
!                                         4 DEC     : POSITIONS
!                                        ----------
!                                         5 FLUX...,  (sky-subtracted)
!                                         6 MAG...,   (sky-subtracted)
!                                         7 SKY     : BRIGHTNESSES
!                                        ----------
!                                         8 FWHM    : SEEING
!                                         9 ELLIP   : TRAILING
!                                        10 NPIX    : # pixels per object
!                                        11 ELL_A   : maj.ell.half-axis (1sigma)
!                                        12 ELL_B   : min.ell.half-axis (1sigma)
!                                        13 ROTANG  : angle(maj.axis,x-axis)
!                                        14 FLAGS   : OBJ.QUAL. (not f. std.)
!                                        ==========
!                                        15 COLOR   : colours f. ellipses
!                                        16 XSTARTPIX,
!                                        17 YSTARTPIX,
!                                        18 XENDPIX,
!                                        19 YENDPIX : rectangle surround. ellip.
!                                        20 S2N     : SIGNAL-TO-NOISE
!                                        21 AIRMASS : WEIGHTED AIRMASS
!                                        22 MJD     : mod.Jul.date
!
!               Hint for columns 13, 15 and 16-19 (object marks):
!                 With `draw/ellipse <P1-entries>_phot.tbl p'
!                 the found objects can be marked onto the displayed current
!                 <P1-entry> frame.
!
!               Legend for column 14 (obj.qual.flags):
!
!                 <C>lose    : area of neighbour overlaps more than 10%
!                 <M>ultiple : multiple object detection after deblending
!                 <S>aturated: at least one pixel of object is saturated
!                 <T>runcated: object is too close to image boundary
!                 <A>perture data corrupt : incomplete data
!                 memory overflow during <D>eblending: memory not enough
!                 memory <O>verflow during extraction: memory not enough
!
!               Formula for column 20 (signal-to-noise):
!
!                 All fluxes and readout noise in e- , w. original expo. time ,
!                 the object flux is integrated :
!
!                                     flux_obj.
!                 -------------------------------------------------------------
!                 sqrt( flux_obj. + #pix./obj.*flux_sky + #pix./obj.*(RON**2) )
!
!               The airmass (X) in column 21 is weighted as follows:
!                 <X> = (X_expostart + 4*X_expomiddle + X_expoend) / 6.
!
!               In the last but one row the medians of SEEING and TRAILING
!               are given (negative number: calculation failed),
!               in the last row the median deviations from the median ("more
!               robust sigma").
!
!               In the last but one row of the MAG column the relative (i.e.,
!               not flux calibrated) limiting magnitude is given (for S/N=5 ,
!               defined in fors_photometry.prg; negative number: calculation
!               failed).
!
!             - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
!             Into all headers descriptors for the sky background (mean, sigma,
!             median), the FWHM and the limiting magnitude (for S/N=5("LIMS2N"))
!             are written - negative values for the latter two signal quality
!             values reflect a failure in their computation.
!
! .RETURN     Q1:  0: successful return
!                 -1: a warning message occured
!                  1: an error occured
! __________
!
! .ALGORITHM  See .PURPOSE under B):
!               SExtractor works as follows:
!               1. Estimation of the sky background (kappa-sigma and mode)
!               2. Detection of objects by thresholding instead of peak finding
!                  (improved by a convolution mask PHOTOMETRY_CONVOL_MASK)
!               3. Deblending (splitting of close detections, reallocation of
!                  low-flux pixels)
!               4. Filtering of the detections (eliminating noise peaks in
!                  wings)
!               5. Photometry (fixed aperture in the pipeline; this and the
!                  other output values are determined by PHOTOMETRY_OUTPUT_PAR)
!               Important input parameters (passed by PHOTOMETRY_INPUT_PAR):
!               for 1. mesh size of grid, size of median filter;
!                   2. convolution mask, threshold, min. number of connected
!                      pixels, hardware dependent buffer sizes;
!                   3. number of additional thresholds, fraction of total
!                      intensity above which a component should be considered
!                      to be separate;
!                   4. factor for strength of cleaning, hardware dependent
!                      buffer size;
!                   5. aperture diameter in pixels, saturation level, no zero-
!                      point correction;
!               For more detailed descriptions see corresponding documents in
!               $PIPE_HOME/fors/sext/doc/, especially sex1_doc.ps.gz and
!               sex2_doc.ps.gz (the latter one for the valid current version
!               2.0.15 has not been completed by the author E.Bertin yet).
!
!             See .PURPOSE under C):
!               Generally objects with the following properties are REMOVED
!               (the constants `("...")' are defined in fors_photometry.prg):
!               - negative flux (because of sky subtraction),
!               - negative FWHM (bug in SExtractor),
!               - too close to the image boundary,
!               - incomplete data,
!               - ellipticity > 0.9 ("BADELLIP"),
!               - only for standard stars: #pixels/object < 40 ("MINSTDPIX").
!
!             See .PURPOSE under D):
!               The values are SELECTED and calculated in the following way:
!               * S/N:
!                   (no selection and no median value)
!               * LIMITING MAGNITUDE:
!                   - S/N <= 10. ("MAXS2N"),
!                   - without error flags;
!                   at least 9 ("MINOBJ") objects must be left, then S/N is
!                   fitted against the flux (with mode of all quotients (S/N /
!                   flux) as slope and 0.,0. as zero point). The limiting mag-
!                   nitude is taken at S/N=5 ("LIMS2N") .
!               * FWHM:
!                   - ellipticity < 0.5 ("MAXELLIP")
!                   - stellarity index > 0.5 ("MINSTELL")
!                   - without error flags (created by SExtractor);
!                   at least 9 ("MINOBJ") objects must be left, then the median
!                   of the FWHMs of the remaining objects is calculated.
!               * ELLIPTICITY:
!                   - #pixels/object >= 25 ("MINPIX"),
!                   - without error flags;
!                   at least 3 objects must be left, then the median
!                   of the ellipticities of the remaining objects is calculated.
!
! .COMMENT    The ascii files with input parameters and output parameters (P2)
!             must NOT be changed; otherwise "fors_photometry.prg" must be al-
!             tered partially (see comment in header of fors_photometry.prg) !!!
!
! .ENVIRON    FORS context
!
! .LANGUAGE   MIDAS 98NOV
!
! .AUTHOR     N.Rainer  DPG/DMD/ESO
!
! .VERSION    0.9  98/07/08  Creation
!                  98/07/15  Release
!             1.0  98/12/15  Release
!             1.1  99/01/14  Release
!             1.2  99/03/15  Release
!             1.3  99/09/23  Release
!
!-------------------------------------------------------------------------------
!
!
crossref in ref out
!
!
define/local CFMDESC/c/1/7 "CF_MEAN"
define/local RONMDESC/c/1/8 "RON_MEAN"
!
define/local FMTDEF/c/1/17 "I1 E14.7,E23.15"
!
define/local PROID/c/1/40 "img_obs_photometry"
!
if fors_maxlin .le. 65534. then
  define/local OVEXVAL/r/1/1 {fors_maxlin}
else
        write/out " "
        mess/out 1 {PROID} "Global key fors_maxlin={fors_maxlin} >65534. !"
  goto error
endif
!
define/local warn/i/1/1 0
write/keyword q1 0
!
!
if p1 .eq. "??" then
  w/o "      ******************************************************************"
  w/o "      in   (P1): set of input frames (ascii list);"
  w/o "                 possible frame types:"
  w/o "                   reduced standard star images (STANDARD_REDUCED_IMG),"
  w/o "                   reduced science images (SCIENCE_REDUCED_IMG);"
  w/o "                     default: none"
  w/o " "
  w/o "      ref  (P2): super recipe mode: "+" (reference frames are included"
  w/o "                 in P1),"
  w/o "                 single recipe mode: set of reference frames (ascii"
  w/o "                 list);"
  w/o "                 needed frames:"
  w/o "                   table with SExtractor input parameters"
  w/o "                   (PHOTOMETRY_INPUT_PAR),"
  w/o "                   table with SExtractor convolution mask"
  w/o "                   (PHOTOMETRY_CONVOL_MASK),"
  w/o "                   table with SExtractor output parameters"
  w/o "                   (PHOTOMETRY_OUTPUT_PAR);"
  w/o "                   table with SExtractor output parameters"
  w/o "                   (PHOTOMETRY_NEURON_PAR);"
  w/o "                 optional:"
  w/o "                   mask frame with overexposed pixels"
  w/o "                   (OVEREXPO_MASK);"
  w/o "                 also needed frame type:"
  w/o "                   table with purging information (DESC_PURGE_TABLE);"
  w/o "                     default: none"
  w/o " "
  w/o "      out  (P3): single recipe mode: basename for ascii list with"
  w/o "                 output frame,"
  w/o "                 super recipe mode: again set of input frames and"
  w/o "                 existing image catalogue (".cat"), separated by a"
  w/o "                 comma;"
  w/o "                 frames for each input image:"
  w/o "                   sky background images (PHOT_BACKGROUND_IMG),"
  w/o "                   sky-subtracted images (PHOT_BACKSUBTR_IMG);"
  w/o "                 tables for each input image with found objects:"
  w/o "                   positions/brightnesses/object_quality/airmass"
  w/o "                   (PHOTOMETRY_TABLE);"
  w/o "                   default: `PROID' defined in procedure (single mode)"
  w/o "      ******************************************************************"
  return/exit
endif
!
!
!          PREPARATIONS FOR RECIPE :
!          ***********************
!
log/recipe {PROID} in
!
        mess/out 3 {PROID} "Do photometry ..."
!
@p fors_paragraph 1 1  ! -------- INITIALIZATION --------
!
set/format {FMTDEF}
!
!
!          INPUT :
!          *****
!
!P1(in):
!------
if p1 .eq. "?" mess/out 2 {PROID} " "
define/parameter p1 ? ? -
  "        Enter name of input frame set: "
define/local inpipe_frm/c/1/76 "{p1}"
!
if m$exist(inpipe_frm) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered list is not present"
        mess/out 1 {PROID} "or can not be opened !"
        mess/out 1 {PROID} "({inpipe_frm})"
  goto error
endif
!
define/local filecode/i/1/2 0 all
open/file {inpipe_frm} read filecode
if filecode(1) .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Unable to open entered list !"
        mess/out 1 {PROID} "({inpipe_frm})"
  goto error
endif
read/file {filecode(1)} outputc
if filecode(2) .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Entered list is empty !"
        mess/out 1 {PROID} "({inpipe_frm})"
  close/file {filecode(1)}
  goto error
endif
close/file {filecode(1)}
!
!
!P2(ref):
!-------
if p2 .eq. "?" mess/out 2 {PROID} " "
define/parameter p2 ? ? -
  "        Enter name of reference frame set: "
define/local inpipe_ref/c/1/76 "{p2}"
!
! No tests like for P1 - in super recipes reference frames are included in P1 .
!
!
!P3(out):
!-------
define/parameter p3 {PROID} ? -
  "        Enter basename of out.list(single)/exist.cat.".cat"(super): "
define/local outpipe/c/1/76 "{p3}"
!
!
!          DEFINITION OF VARIABLES :
!          ***********************
!
!     "Identifying input frames":
!
  define/local catal/i/1/1 0
  define/local topicima/c/1/76 " " all
  define/local id/c/1/25 " " all
  !
  define/local stdID/c/1/25 " " all
  define/local sciID/c/1/25 " " all
  define/local sexiID/c/1/25 " " all
  define/local sexcID/c/1/25 " " all
  define/local sexoID/c/1/25 " " all
  define/local sexnID/c/1/25 " " all
  define/local oexmskID/c/1/25 " " all
  !
  define/local inpartbl/c/1/76 " " all
  define/local convoltbl/c/1/76 " " all
  define/local outpartbl/c/1/76 " " all
  define/local neupartbl/c/1/76 " " all
  !
  define/local instd/i/1/1 0
  define/local insci/i/1/1 0
  define/local refsex/i/1/1 0
  !
  define/local insdesc/c/1/60 " " all
!
!     "Conversion of ... tables":
!
  define/local totrowsin/i/1/1 0
  define/local row/i/1/1 0
  define/local totrowscon/i/1/1 0
  define/local totrowsout/i/1/1 0
!
!     "O_TIME(1)":
!
  define/local date/r/1/1 0.
!
!     "Normalization":
!
  define/local exptime/r/1/1 0.
  define/local char/i/1/1 0
  define/local topinorm/c/1/76 " " all
  define/local normalized/i/1/1 0 ? +lower_levels
!
!     "Calculating mean":
!
  define/local portloop/i/1/1 0
  define/local cftopic/d/1/1 0.
  define/local cfsum/d/1/1 0.
  define/local rontopic/d/1/1 0.
  define/local ronsum/d/1/1 0.
  define/local cf/d/1/1 0.
  define/local ron/d/1/1 0.
!
!     "Overexposure masks":
!
  define/local loopcount/i/1/2 0 all
  define/local topicmsk/c/1/76 " " all
  define/local badsize/i/1/1 0
!
!     "Calling of procedure":
!
  define/local typ/c/1/1 " "
  define/local scl/c/1/60 " " all
  define/local exp/c/1/60 " " all
  define/local lon/c/1/60 " " all
  define/local lat/c/1/60 " " all
  define/local psc/i/1/1 0
  define/local osc/i/1/1 0
  define/local cola/c/1/100 " " all
  define/local filt/c/1/1 " "
  define/local colb/c/1/100 " " all
  define/local colc/c/1/100 " " all
  define/local tcol/c/1/100 " " all
  define/local outprog_cat/c/1/76 " " all
!
!     "Updating descriptors":
!
  define/local procount/i/1/1 0
  define/local topicmap/c/1/76 " " all
  define/local topicsub/c/1/76 " " all
  define/local topictab/c/1/76 " " all
  define/local topicin/c/1/76 " " all
  define/local subxs/i/1/1 0
  define/local subxe/i/1/1 0
  define/local areaunit/r/1/1 0.
  define/local skymean/r/1/1 0.
  define/local skysig/r/1/1 0.
  define/local skymedi/r/1/1 0.
  define/local rowfwhm/i/1/1 0
  define/local rowfwsg/i/1/1 0
  define/local rowlimg/i/1/1 0
  define/local fwhm/r/1/1 0.
  define/local fwsg/r/1/1 0.
  define/local limg/r/1/1 0.
  define/local phottable/c/1/76 " " all
!
!
!          LOADING OF GLOBAL KEYS (SINGLE RECIPE MODE) ,
!          CONVERSION FITS->BDF/TBL ,
!          CLASSIFICATION OF INPUT FRAMES :
!          *********************************************
!
        mess/out 4 {PROID} "Getting input frames"
!
initialize/pipeline {inpipe_frm} {inpipe_ref} -
  inpipe_good.cat ? ? + inpipe_bad.cat ?
if {q1} .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Pipeline initialization failed !"
  goto error
endif
-delete inpipe_bad.cat
!
! Initialize product header:
init/prohead {inpipe_frm} {inpipe_ref} 2,"Photometry"
if {q1} .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Product header initialization failed !"
  goto error
endif
!
initialize/longlog {inpipe_frm} {inpipe_ref} 2
!
!
!          IDENTIFICATION VIA DESCRIPTORS ,
!          MAPPING OF FRAMES TO PARAMETERS  :
!          ********************************
!
        mess/out 4 {PROID} "Identifying input and reference frames"
!
stdID = "{RSTD_ID}"
sciID = "{RSCIENCE_ID}" 
sexiID = "{SEXINPAR_ID}"
sexcID = "{SEXCONVOL_ID}"
sexoID = "{SEXOUTPAR_ID}"
sexnID = "{SEXNEUPAR_ID}"
!oexmskID = "{OVEXPM_ID}"
!
create/icat {PROID}.cat null {FORS_IDSTRNG} >Null
instd = 0
insci = 0
refsex = 0
!
ident_loop:
  store/frame topicima inpipe_good.cat 1 end_ident_loop
  id = "{{topicima},{fors_idstrng}}"
  !
  if id .eq. "{stdID}" then
    add/icat {PROID}.cat {topicima} >Null
        mess/out 4 {PROID} "{topicima} : {id} => input"
    insdesc = "{{topicima},{h_instrume}}"
    instd = 1
  elseif id .eq. "{sciID}" then
    add/icat {PROID}.cat {topicima} >Null
        mess/out 4 {PROID} "{topicima} : {id} => input"
    insdesc = "{{topicima},{h_instrume}}"
    insci = 1
  elseif id .eq. "{sexiID}" then
    inpartbl = topicima
        mess/out 4 {PROID} "{topicima} : {id} => reference"
    refsex = refsex+1
  elseif id .eq. "{sexcID}" then
    convoltbl = topicima
        mess/out 4 {PROID} "{topicima} : {id} => reference"
    refsex = refsex+1
  elseif id .eq. "{sexoID}" then
    outpartbl = topicima
        mess/out 4 {PROID} "{topicima} : {id} => reference"
    refsex = refsex+1
  elseif id .eq. "{sexnID}" then
    neupartbl = topicima
        mess/out 4 {PROID} "{topicima} : {id} => reference"
    refsex = refsex+1
  endif
  !
  goto ident_loop
end_ident_loop:
-delete inpipe_good.cat
!
if instd .eq. 0 .and. insci .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No matching input science or"
        mess/out 1 {PROID} "standard star frame present !"
  goto error
endif
!
if instd .ne. 0 .and. insci .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Both standard and science frames present -"
        mess/out 1 {PROID} "only one kind possible per invocation !"
  goto error
endif
!
if refsex .lt. 4 then
        write/out " "
        mess/out 1 {PROID} "Not enough reference frames present :"
  !
  if m$len(inpartbl) .eq. 0 then
        mess/out 1 {PROID} "No reference file with sextractor"
        mess/out 1 {PROID} "input parameters present !"
  endif
  !
  if m$len(convoltbl) .eq. 0 then
        mess/out 1 {PROID} "No reference file with sextractor"
        mess/out 1 {PROID} "convolution mask present !"
  endif
  !
  if m$len(outpartbl) .eq. 0 then
        mess/out 1 {PROID} "No reference file with sextractor"
        mess/out 1 {PROID} "output parameters present !"
  endif
  !
  if m$len(neupartbl) .eq. 0 then
        mess/out 1 {PROID} "No reference file with sextractor"
        mess/out 1 {PROID} "neuronal network weights present !"
  endif
  !
  goto error
endif
!
!
!          CONVERSION OF REFERENCE TABLES
!          TO ASCII  :
!          ******************************
!
        mess/out 4 {PROID} "Converting reference tables to ascii"
!
!***************************
@p fors_tbl2ascii {inpartbl}
!***************************
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Table conversion failed !"
        mess/out 1 {PROID} "({inpartbl})"
  goto error
endif
if q2 .ne. "sex__inpar.dat" -rename {q2} sex__inpar.dat  ! def.f. `fors_photom.'
!
!****************************
@p fors_tbl2ascii {convoltbl}
!****************************
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Table conversion failed !"
        mess/out 1 {PROID} "({convoltbl})"
  goto error
endif
if q2 .ne. "sex_convol.dat" -rename {q2} sex_convol.dat
!
!****************************
@p fors_tbl2ascii {outpartbl}
!****************************
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Table conversion failed !"
        mess/out 1 {PROID} "({outpartbl})"
  goto error
endif
if q2 .ne. "sex_outpar.dat" -rename {q2} sex_outpar.dat
!
!****************************
@p fors_tbl2ascii {neupartbl}
!****************************
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Table conversion failed !"
        mess/out 1 {PROID} "({neupartbl})"
  goto error
endif
if q2 .ne. "sex_neupar.dat" -rename {q2} sex_neupar.dat
!
!
!          CORRECTION OF "O_TIME(1) = 0." BUG :
!          **********************************
!
catal = 0
date_loop:
  store/frame topicima {PROID}.cat 1 end_date_loop
  date = {{topicima},O_TIME(1)}
  if date .lt. 1990. .or. date .ge. 2100. then
        mess/out 2 {PROID} "Date in O_TIME(1) <1990. or >2100. -"
        mess/out 2 {PROID} "({topicima})"
        mess/out 2 {PROID} "calculating date from MJD !"
    !*************************
    @p fors_jd2date {topicima}
    !*************************
    {topicima},O_TIME(1) = outputi(1)
    {topicima},O_TIME(2) = outputi(2)
    {topicima},O_TIME(3) = outputi(3)
    warn = 1
  endif
  goto date_loop
end_date_loop:
!
!
!          CREATION OF NORMALIZED INPUT
!          IMAGES (EXPOSURE TIME = 1sec) :
!          *****************************
!
@p fors_paragraph 2 1  ! -------- PROCESSING --------
!
        mess/out 3 {PROID} "Creating exposure normalized input frame(s)"
!
create/icat {PROID}_norm.cat null {FORS_IDSTRNG} >Null
!
catal = 0
norm_loop:
  store/frame topicima {PROID}.cat 1 end_norm_loop
  exptime = {{topicima},O_TIME(7)}
  if exptime .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Exposure time is zero -"
        mess/out 1 {PROID} "({topicima})"
        mess/out 1 {PROID} "no normalization possible !"
    goto error
  endif
  !
    char = m$indexb(topicima,".bdf")-1
    topinorm = topicima(1:{char})//"_norm.bdf"
  compute/pixel {topinorm} = {topicima} / {exptime}
  {topinorm},O_TIME(7) = 1.
  add/icat {PROID}_norm.cat {topinorm} >Null
  !
        set/format F7.2,
        longlog/out 1 "     {topicima} was divided by {exptime} sec" >Null
        set/format {FMTDEF}
  !
  goto norm_loop
end_norm_loop:
!
normalized = 1  ! "+lower_levels" for main procedure below
!
        shortlog/out 1 "Normalized by exposure time." >Null
!
!
!          CALCULATING MEAN CONVERSION FACTOR
!          AND MEAN READOUT NOISE FOR MULTIPLE
!          PORT READOUT :
!          ***********************************
!
if fors_nout .gt. 1 then
  !
        mess/out 4 {PROID} "Calculating mean conv.fact. and RON of all ports"
  !
endif
!
catal = 0
cfsum = 0.
ronsum = 0.
conv_loop:
  store/frame topinorm {PROID}_norm.cat 1 end_conv_loop
  do portloop = 1 fors_nout
    !
    cftopic = {{topinorm},{h_confac({fors_outi({portloop})})}}
    if cftopic .le. 0. then
        set/format F15.14
        write/out " "
        mess/out 2 {PROID} "Impossible value for descriptor"
        mess/out 2 {PROID} "`{h_confac({fors_outi({portloop})})}': {cftopic} -"
        mess/out 2 {PROID} "not all calculations will be possible !"
        write/out " "
        set/format {FMTDEF}
      warn = 1
    endif
    cfsum = cfsum + cftopic
    !
    rontopic = {{topinorm},{h_ron({fors_outi({portloop})})}}
    if rontopic .le. 0. then
        set/format F15.14
        write/out " "
        mess/out 2 {PROID} "Impossible value for descriptor"
        mess/out 2 {PROID} "`{h_ron({fors_outi({portloop})})}': {rontopic} -"
        mess/out 2 {PROID} "not all calculations will be possible !"
        write/out " "
        set/format {FMTDEF}
      warn = 1
    endif
    ronsum = ronsum + rontopic
    !
  enddo
  cf = cfsum / fors_nout
  ron = ronsum / fors_nout
  write/descriptor {topinorm} {CFMDESC}/d/1/1 {cf}
  write/descriptor {topinorm} {RONMDESC}/d/1/1 {ron}
  goto conv_loop
end_conv_loop:
!
!
!          APPLYING OF OVEREXPOSURE MASKS
!          (IF PRESENT):
!          ******************************
!
catal = 0
loopcount(1) = 0
loopcount(2) = 0
ovex_loop:
  !
  catal = loopcount(1)
  store/frame topicima {PROID}.cat 1 end_ovex_loop
  loopcount(1) = catal
  !
  catal = loopcount(2)
  store/frame topinorm {PROID}_norm.cat 1 end_ovex_loop
  loopcount(2) = catal
  !
  if m$existd(topicima,h_ovexmask) .eq. 1 then
    !
    topicmsk = "{{topicima},{h_ovexmask}}"
    badsize = 0
    if m$exist(topicmsk) .eq. 1 then
      if {{topicima},NPIX(1)} .ne. {{topicmsk},NPIX(1)} badsize = 1
      if {{topicima},NPIX(2)} .ne. {{topicmsk},NPIX(2)} badsize = 1
      if badsize .eq. 0 then
        !
        mess/out 3 {PROID} "Overexposure mask {topicmsk}"
        mess/out 3 {PROID} "  is applied to   {topicima}"
        !
        replace/image {topinorm} + {topicmsk}/{OVEXVAL},>={topicmsk} >Null
        !
      else
        mess/out 2 {PROID} "Sizes of input frame and overexpo. mask differ -"
        mess/out 2 {PROID} "overexp.mask {topicmsk}"
        mess/out 2 {PROID} "not used for {topicima} !"
        warn = 1
      endif
    else
        mess/out 2 {PROID} "Descriptor w. overexpo. mask present -"
        mess/out 2 {PROID} "({topicima},{h_ovexmask})"
        mess/out 2 {PROID} "but mask frame not !"
        mess/out 2 {PROID} "({topicmsk})"
      warn = 1
    endif
    !
    delete/descriptor {topicima} {h_ovexmask} n
    !
  else  ! no overexposed pixels
    !
        mess/out 3 {PROID} "No overexpo. mask is assigned to"
        mess/out 3 {PROID} "  {topicima}"
    !
  endif
  goto ovex_loop
end_ovex_loop:
!
-delete {PROID}.cat
!
!
!          DETERMINING OVERSCAN SECTIONS :
!          *****************************
!
!************************************
@p fors_get_imsec #1_{PROID}_norm.cat
!************************************
if {q1} .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Determination of overscan sections failed !"
  goto error
endif
psc = outputi(5)
osc = outputi(6)
!
!
!          *********************************
!          *                               *
!          *  CALLING OF MAIN PROCEDURE :  *
!          *                               *
!          *********************************
!
if instd .ne. 0 typ = "t"
if insci .ne. 0 typ = "c"
!
scl = h_pixscale
exp = h_dit1  ! -> EXPOSURE TIME OF 1ST WINDOW !!!
!
lon = h_geolon
lat = h_geolat
!
  !************************************
  @p fors_getfilter #1_{PROID}_norm.cat
  !************************************
  outputc = fors_filti(1)
  filt = outputc(1:1)
cola = "{c_rect},{c_decl}"
colb = "{c_flx}_{filt},{c_mag}_{filt}"
colc = "{c_s2n},{c_airm}"
tcol = "{cola},{colb},{colc}"
!
        mess/out 4 {PROID} "Applying procedure  "fors_photometry" :"
!
!*************************************************************************
@p fors_photometry {PROID}_norm,{typ} ? {RONMDESC} {scl},{CFMDESC},{exp} -
   {lon},{lat} p {psc},{osc},{OVEXVAL} {tcol}
!*************************************************************************
!
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Photometry failed !"
  goto error
endif
!
outprog_cat = q3
!
!
!          UPDATING DESCRIPTORS OF OUT- AND INPUT FRAMES :
!          *********************************************
!
@p fors_paragraph 3 1  ! -------- TERMINATION --------
!
        mess/out 4 {PROID} "Classifying output table(s) and image(s)"
        mess/out 4 {PROID} "(*_phot.tbl, *_skymap.bdf, *_skysub.bdf) and"
        mess/out 4 {PROID} "endorsing QC status and values in input image(s)"
!
catal = 0
procount = 0
desc_loop:
  store/frame topicima {outprog_cat} 1 end_desc_loop
  !
  if m$index(topicima,"_skymap.bdf") .gt. 1 then
    procount = procount+1
    topicmap = topicima
      char = m$indexb(topicima,"_skymap.bdf")-1
    topicsub = topicima(1:{char})//"_skysub.bdf"
    topictab = topicima(1:{char})//"_phot.tbl"
      char = m$indexb(topicima,"_norm_skymap.bdf")-1
    topicin = topicima(1:{char})//".bdf"
  else
    goto desc_loop
  endif
  !
    subxs = psc+1
    subxe = {{topicmap},NPIX(1)}-osc
    statistics/image {topicmap} [@{subxs},<:@{subxe},>] ? ? xn >Null
    areaunit = {{topicmap},{h_pixscale}}**2.
    skymean = -2.5*m$log10(outputr(3)/areaunit)
    skymedi = -2.5*m$log10(outputr(8)/areaunit)
    !
    ! -------------------------------------------
    ! mag = -2.5*log(flux)
    ! mag_sigma != -2.5*log(flux_sigma)
    ! mag_sigma = |dmag/dflux*flux_sigma|
    !           = |-2.5*log(e)*flux_sigma/<flux>|
    ! -------------------------------------------
    !
    skysig  = m$abs(-2.5*m$log(m$exp(1))*outputr(4)/outputr(3))
  !
    show/table {topictab} >Null
    rowfwhm = outputi(2)-1
    rowfwsg = outputi(2)
    rowlimg = outputi(2)
    write/keyword fwhm {{topictab},:FWHM,{rowfwhm}} >Null
    write/keyword fwsg {{topictab},:FWHM,{rowfwsg}} >Null
    write/keyword limg {{topictab},:{c_mag}_{filt},{rowlimg}} >Null
  !
  classify/product {topicmap} {SKYBACK_ID} {d_tmp} {d_det} OK ? y
  write/descriptor {topicmap} {h_instrume}/c/1/60 {insdesc}
  write/descriptor {topicmap} {FORS_IDSTRNG}/c/1/60 {SKYBACK_ID} !f.su.rec.
  write/descriptor {topicmap} {h_proqcsts}/c/1/9 "CHECKED"
  write/dhelp {topicmap} {h_proqcsts} "{i_proqcsts}"
  write/descriptor {topicmap} {h_proqcsmea}/r/1/1 {skymean}
  write/dhelp {topicmap} {h_proqcsmea} "{i_proqcsmea}"
  write/descriptor {topicmap} {h_proqcssig}/r/1/1 {skysig}
  write/dhelp {topicmap} {h_proqcssig} "{i_proqcssig}"
  write/descriptor {topicmap} {h_proqcsmed}/r/1/1 {skymedi}
  write/dhelp {topicmap} {h_proqcsmed} "{i_proqcsmed}"
  write/descriptor {topicmap} {h_proqcfwhm}/r/1/1 {fwhm}
  write/dhelp {topicmap} {h_proqcfwhm} "{i_proqcfwhm}"
  write/descriptor {topicmap} {h_proqcfwsg}/r/1/1 {fwsg}
  write/dhelp {topicmap} {h_proqcfwsg} "{i_proqcfwsg}"
  write/descriptor {topicmap} {h_proqclimg}/r/1/1 {limg}
  write/dhelp {topicmap} {h_proqclimg} "{i_proqclimg}"
  !
  classify/product {topicsub} {SKYSUB_ID} {d_tmp} {d_det} OK ? y
  write/descriptor {topicsub} {h_instrume}/c/1/60 {insdesc}
  write/descriptor {topicsub} {FORS_IDSTRNG}/c/1/60 {SKYSUB_ID} !f.su.rec.
  write/descriptor {topicsub} {h_proqcsts}/c/1/9 "CHECKED"
  write/dhelp {topicsub} {h_proqcsts} "{i_proqcsts}"
  write/descriptor {topicsub} {h_proqcsmea}/r/1/1 {skymean}
  write/dhelp {topicsub} {h_proqcsmea} "{i_proqcsmea}"
  write/descriptor {topicsub} {h_proqcssig}/r/1/1 {skysig}
  write/dhelp {topicsub} {h_proqcssig} "{i_proqcssig}"
  write/descriptor {topicsub} {h_proqcsmed}/r/1/1 {skymedi}
  write/dhelp {topicsub} {h_proqcsmed} "{i_proqcsmed}"
  write/descriptor {topicsub} {h_proqcfwhm}/r/1/1 {fwhm}
  write/dhelp {topicsub} {h_proqcfwhm} "{i_proqcfwhm}"
  write/descriptor {topicsub} {h_proqcfwsg}/r/1/1 {fwsg}
  write/dhelp {topicsub} {h_proqcfwsg} "{i_proqcfwsg}"
  write/descriptor {topicsub} {h_proqclimg}/r/1/1 {limg}
  write/dhelp {topicsub} {h_proqclimg} "{i_proqclimg}"
  !
  classify/product {topictab} {PHOTT_ID} {d_tmp} {d_det} OK
  write/descriptor {topictab} {h_instrume}/c/1/60 {insdesc}
  write/descriptor {topictab} {FORS_IDSTRNG}/c/1/60 {PHOTT_ID} !f.su.rec.
  write/descriptor {topictab} {h_proqcsts}/c/1/9 "CHECKED"
  write/dhelp {topictab} {h_proqcsts} "{i_proqcsts}"
  write/descriptor {topictab} {h_proqcsmea}/r/1/1 {skymean}
  write/dhelp {topictab} {h_proqcsmea} "{i_proqcsmea}"
  write/descriptor {topictab} {h_proqcssig}/r/1/1 {skysig}
  write/dhelp {topictab} {h_proqcssig} "{i_proqcssig}"
  write/descriptor {topictab} {h_proqcsmed}/r/1/1 {skymedi}
  write/dhelp {topictab} {h_proqcsmed} "{i_proqcsmed}"
  write/descriptor {topictab} {h_proqcfwhm}/r/1/1 {fwhm}
  write/dhelp {topictab} {h_proqcfwhm} "{i_proqcfwhm}"
  write/descriptor {topictab} {h_proqcfwsg}/r/1/1 {fwsg}
  write/dhelp {topictab} {h_proqcfwsg} "{i_proqcfwsg}"
  write/descriptor {topictab} {h_proqclimg}/r/1/1 {limg}
  write/dhelp {topictab} {h_proqclimg} "{i_proqclimg}"
  phottable = topictab
  !
  write/descriptor {topicin} {h_proqcsts}/c/1/9 "CHECKED"
  write/dhelp {topicin} {h_proqcsts} "{i_proqcsts}"
  write/descriptor {topicin} {h_proqcsmea}/r/1/1 {skymean}
  write/dhelp {topicin} {h_proqcsmea} "{i_proqcsmea}"
  write/descriptor {topicin} {h_proqcssig}/r/1/1 {skysig}
  write/dhelp {topicin} {h_proqcssig} "{i_proqcssig}"
  write/descriptor {topicin} {h_proqcsmed}/r/1/1 {skymedi}
  write/dhelp {topicin} {h_proqcsmed} "{i_proqcsmed}"
  write/descriptor {topicin} {h_proqclimg}/r/1/1 {limg}
  write/dhelp {topicin} {h_proqclimg} "{i_proqclimg}"
  !
  delete/descriptor {topicmap} {CFMDESC} n
  delete/descriptor {topicmap} {RONMDESC} n
  delete/descriptor {topicsub} {CFMDESC} n
  delete/descriptor {topicsub} {RONMDESC} n
  delete/descriptor {topictab} {CFMDESC} n
  delete/descriptor {topictab} {RONMDESC} n
  delete/descriptor {topicin} {CFMDESC} n
  delete/descriptor {topicin} {RONMDESC} n
  !
  goto desc_loop
end_desc_loop:
!
!
!          UPDATING OF OUTPUT CATALOGUE (SUPER RECIPE MODE) ,
!          CONVERSION BDF/TBL->FITS/TFITS (SINGLE MODE) :
!          **************************************************
!
        mess/out 4 {PROID} "Releasing output table(s) and image(s)"
!
! Create product header:
finish/prohead {outprog_cat} 2
!
finish/pipeline {outprog_cat} {outpipe} + {phottable}
!
finish/longlog {PROID}_norm.cat {outprog_cat} {phottable} 2 {outpipe}
!
-delete {PROID}_norm.cat
-delete {outprog_cat}
!
!
!          RESETTINGS AFTER RECIPE :
!          ***********************
!
set/format
!
log/recipe {PROID} out
!
@p fors_paragraph 4 1  ! ================
!
!
if warn .eq. 0 then
  return 0
else
  return -1
endif
!
error:
  set/format
  finish/longlog ? -1 ? 2
  write/out " "
  return 1
!
!
!-------------------------------------------------------------------------------
