! $Id: fors_img_obs_reduce.prg,v 1.73 2000/09/12 08:41:05 sbogun Exp $   
!+++ FORS_IMG_OBS_REDUCE +++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT     (C) 1998  European Southern Observatory
!
!.IDENT         $RCSfile: fors_img_obs_reduce.prg,v $
!
!.KEYWORDS      Reduction of images and creation of auxiliary frames
!
!.USAGE         @p (IDENT) sof_in sof_ref out [BIAS] [DARK] [switch] [scale]
!
!.INPUT         sof_in  (P1): set of input frames (ascii list)
!               sof_ref (P2): set of reference frames (ascii list)
!               BIAS    (P4): BIAS constant (only used if there is no
!                             bias frame in the sof_ref)
!               DARK    (P5): DARK constant (only used if there is no
!                             dark frame in the sof_ref)
!               switch  (P6) -1: only cosmics correction
!                             0: standard reduction
!                             1: create illumination corr. of screen flatfields
!                             2: create corr.flatfield of screen flatfields
!                             3: create illumination corr. of blank sky expo.
!                             4: create super flat out of sky- and screen-flats
!                             5: fringe correction
!                             6: bias will be processed instead of sci. frame
!                             7: dark will be processed instead of sci. frame
!                             8: flat (sky) will be processed instead
!                             9: flat (screen) will be processed instead
!                            10: standard star will be processed instead
!               scale   (P7) "MED": the bias correction will be done using the 
!                                   scaled master bias. The scaling is done as
!                                   follows: 
!                                   Each of the sections of the master bias is 
!                                   divided by its median and multiplied by the
!                                   median of the corresponding pre-/overscan 
!                                   region of the input frame.
!                            "OVE": like "MED", but each of the sections of the
!                                   master bias is divided by the median of its
!                                   corresponding pre-/overscan region and
!                                   multiplied by the median of the pre-/overscan
!                                   region of the input frame.
!                            "NO":  The master bias will not be scaled.
!
!.OUTPUT        out     (P3): path + basename of output files:
!                             reduced (processed) image         [FITS]
!                             A copy in the local directory:
!                             r_<name of input frame>.bdf       [BDF]
!
!.RETURN        Q1:  0: successful return
!                    otherwise: successful return failed
!               OUT_A: name of reduced frame
!
!.DEFAULT       P4:     0.0     (bias)
!               P5:     0.0     (dark)
!               P6:     0       (standard reduction)
!               p7:     OVE     (bias method)
!
!.PURPOSE       reduce original image frame or create correction frames which
!               may be needed for the reduction process.
!
!.ALGORITHM     
!               Create an image catalog of set of input frames (P1). Copy all
!               files to the local directory. Check exposure type of each
!               frame. Only if a complete set of frames are found the reduction
!               may be started. If a bad pixel table has been passed to the
!               program the correction will be done depending on the labels
!               inside the MIDAS-table. After this optional pixel correction
!               the program  REDUCE/CCD will be called to do the rest of the
!               reduction.
!                    Depending on which value the switch parameter (P6) is set,
!               the program operates in a different way. Above the default
!               operation (P6=0), the reduction, has been described. In case
!               of P6 is set to 1,..,4 then the illumination correction frames
!               or the illumination corr. flatfields will be produced,
!               respectively.
!               All possible modes are based on MIDAS commands of the CCDRED
!               context:
!               REDUCE/CCD,
!               ILLCOR/CCD, ILLFLAT/CCD, SKYCOR/CCD, SKYFLAT/CCD, FRCOR/CCD.
!                    Besides the bias and dark correction the illumination and
!               fringe correction may be considered as soon as the correction
!               frames are specified! If no bias and/or dark frame is defined
!               then a constant value may be passed instead (BIAS_CONS,
!               DARK_CONS). In case the MEAN BIAS of the overscan regions
!               should be used (gen_qc0_overscan) then pass the name of the
!               overscan table [FITS] to parameter P4. The bad pixel
!               correction will be done using the MIDAS commands MODIFY/COLUMN
!               (in case the value in column :X is a positive number)
!               and MODIFY/PIXEL (in case :X has a zero or negative number).
!
!               Finally, a scaling factor (CCDSCALE) used in fors_chkcal in
!               order to be able to compare a calibration frame with a refe-
!               rence frame (MASTER) will be computed. For bias CCDSCALE is
!               simply 1.0. In case of flatfields this factor will be the
!               the MEDIAN of the frame. In case of darks it will be set to
!               the exposure time.
!               At the end CHKCAL may examine the following difference:
!                       master/master,CCDSCALE -  cal/cal,CCDSCALE
!
!               Then this difference is automatically defined in a correct way.
!               But this will be performed in a separate step.
!
!.ENVIRON       FORS, CCDRED context
!   
!.LANGUAGE      MIDAS 98NOV
!
!.AUTHOR        Sebastian Wolf, ESO-DMO
!
!.VERSION       0.5 1997/10/28  Creation.
!               1.0 1998/12/09  fors_paragraph implemented.
!                   1999/01/07  shortlog implemented.
!                   1999/03/10  shortlog close shifted to FINISH/PIPELINE
!               1.1 1999/04/28  longlog implemented (SB)
!               1.3 1999/09/03  Product header generation implemented (SB).
!
!.COMMENT       FORMAT of the bad pixel MIDAS-table:
!               5 columns:  :X, :XSTART, :XEND, :YSTART, :YEND
!                            |  |____________________________|
!                            |                |defines the corners of an area
!                            |defines a detected bad column
!
!               Be sure that the bad pixel table defines an AREA, i.e. both
!               dimensions in X and Y have to be greater than 1! So the table
!               defines an area which surrounds the known bad pixels. Each
!               line defines another area or column.
!
!               In case P6 > 6 && P6 <10 then NO output frame will be stored
!               in the result-catalog.
!------------------------------------------------------------------------------

define/local line/c/1/79 "-" ALL
write/out {line}
LOG/REC fors_img_obs_reduce IN
!
! Check parameters
!
define/param p1 ? c "Enter name of ASCII filelist with science frames:"
define/param p2 ? c "Enter name of ASCII filelist with calibr. frames:"
define/param p4 0 c "Enter BIAS correction constant or overscan table:"
define/param p5 0 c "Enter DARK correction constant:"
define/param p6 0 number        !see comments in the header!!
define/param p7 OVE c "Use scaled master bias (MED/OVE/NO):"
!
! local variable declaration
!
define/local CALIBFRM/i/1/1 0
define/local ityp/i/1/1 0
define/local etyp/c/1/25 " "  ALL
define/local rtyp/c/1/25 " "  ALL
define/local rfrm/c/1/25 " "  ALL
define/local ptyp/c/1/80 " "  ALL
define/local plev/c/1/80 " "  ALL
define/local procid/c/1/25 " "  ALL
define/local warn0a/c/1/80 " " ALL
define/local warn0b/c/1/80 " " ALL
define/local out_c/c/1/100 " " ALL +lower_levels
define/local scale/c/1/3   " "
define/local trimflag/i/1/1 0

warn0a = "Too few INPUT frames! --> NO reduction."
warn0b = "Too few CALIBRATION frames! --> NO reduction."

mess/out 3 img_obs_reduce -
  "Recipe IMG_OBS_REDUCE performs standard reduction"
@p fors_paragraph 1


! Process parameters
scale = m$upper(p7(1:3))
if scale .ne. "MED" .and. scale .ne. "OVE" then
  if scale .ne. "NO" then
    mess/out 1 img_obs_reduce -
       "Value entered for SCALE (P7) is not supported: >{scale}<!"
    return 1
  endif
endif


!
! init pipeline
!
INITIALIZE/PIPELINE {P1} {P2} reduceI.cat CCDRED + + reject.cat result.cat
if {Q1} .gt. 0 return {Q1}
!
init/prohead {p1} {p2} 2,"Standard reduction of images" -
  "bias,{p4}" "dark,{p5}" "switch,{p6}" "scale,{p7}"
if {Q1} .gt. 0 return {Q1}
!
init/longlog {p1} {p2} 2 "bias: {p4}" "dark: {p5}" "switch: {p6}" "scale: {p7}"
if {Q1} .gt. 0 return 1

@p fors_paragraph 2


@p fors_img_obs_reduce,EXPO_TYPE {P6} reduceI.cat
if {Q1} .lt. 0 return -1
CALIBFRM = {Q2}                         !indicates if INPUT is a calibr.frame
procid = Q3(1:25)                       !procedure ID
etyp   = out_a                          !exposure type (INPUT)
rtyp   = out_b                          !exposure type (OUTPUT)
ptyp   = out_c                          !ESO.PRO.TYPE
plev   = outputc                        !ESO.PRO.REDLEVEL
ityp   = {P6}                           !data type
!
! create command: use local version of CCDRED-context commands due to bugs!
!
set/midas user=expert
mess/out 4 {procid} "Use local version of REDUCE/CCD,FLAT/CCD,SKYFLA/CCD"
delete/comm reduce/ccd
delete/comm flat/ccd
delete/comm skyflat/ccd
create/comm reduce/ccd @p ccdreduce
create/comm flat/ccd @p ccdflat
create/comm skyfla/ccd @p ccdskycor
!create/comm illflat/ccd @p ccdflatcor
set/midas user=user,nopromp


@p fors_initfors        !initialize global keywords used for IMG-mode of FORS


search/icat reduceI {BPIXT_ID} >Null
if m$len(out_a) .ne. 0 then
  FX_TABLE = out_a                              !bad pixel table
else
  FX_TABLE = " "
endif 




SC_DKFRM = P5                                   !dark constant
@p fors_preset_bias {P4}                        !preset bias and dark
                                                !by passed parameters

if {p6} .lt. 1 .or. {p6} .gt. 5 then
! Compute pre-/overscan from input frames
  write/out " "
  mess/out 3 {procid} "Computing bias values from pre-/overscan regions..."
  overscan/frames reduceI.cat {etyp} median
  if scale .ne. "NO" then
    if {Q1} .eq. -1 then
      mess/out 2 {procid} -
        "No overscan in input frame(s). Appylying standard bias correction."
      scale = "NO"
    else if {Q1} .ne. 0 then
      mess/out 1 {procid} "Bias computation from overscan areas failed!"
      goto ABORT
    endif
  endif 
endif


! check input frames & preset the CCD-keyword for
! science, bias, dark, and flat

mess/out 3 {procid} "Process frame of exposure type:\n{etyp} ..."
@p fors_chck_iset reduceI.cat {etyp} reduceC.cat {scale}  !check set of frames:
                                                !complete calibration set
                                                !and desired input frame
                                                !specified ?
                                                !CCDRED global keywords will
                                                !be set (CCD_IN, SC_BSFRM, ...)
                                                !It will be guaranteed that all
                                                !all frames have a PIXEL coord.
                                                !system (CUNIT: PIXEL).
if {q1} .ne. 0 then
  mess/out 1 {procid} "NO reduction can be performed!"
  return -1
endif
!
! check/define OUTPUT frame
!
if {p6} .eq. 0 then
  define/param p3 {CCD_IN}_red.bdf c "Enter name of reduced frame:"
else if {p6} .eq. 1 .OR. {p6} .eq. 3 then
  define/param p3 {CCD_IN}_ill c "Enter name of illum.corr. frame:"
else if {p6} .eq. 2 .OR. {p6} .eq. 4 then
  define/param p3 {CCD_IN}_sup c "Enter name of super flatfield:"
endif

rfrm = "{CCD_IN}_red.bdf"                       !BDF name of res.frame


if nice_crea .eq. "y" then     !plot INPUT frame (DISP: 1)
  !
  @p fors_niceclear
  !
  !*************************************************
  @p fors_nicedisp {CCD_IN}.bdf ? 7. 0,0 ? ? 1.,3. n
  !*************************************************
  if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {procid} "Displaying of frame failed !"
        mess/out 1 {procid} "({CCD_IN}.bdf)"
    return 1
  endif
  !
        mess/out 4 {procid} "  loading image into display 1"
  assign/display d,1 >Null
  load/image {CCD_IN}.bdf ? {outputi(1)},avg c,c
  !
  if NICE_HCOPY(:1) .ne. "N" then
    mess/out 3 {procid} "  creating hardcopy <{CCD_IN}.ps>"
    copy/display p5=noprint >Null
    -delete screen{mid$sess(11:12)}.ima
    -rename screen{mid$sess(11:12)}.ps {CCD_IN}.ps   
  endif
  !
endif
!
@p fors_img_obs_reduce,PLOT reduceC.cat {ityp}  !plot CALIB.frames


                                                !Do the bad pixel correction
                                                !ONLY FOR CALIBRATION frames
!if FX_TABLE(1:1) .ne. " " .AND. CALIBFRM .eq. 1 then
if FX_TABLE(1:1) .ne. " " then                  !Will be controled by DO-rules
  swap/units {CCD_IN} PIXEL P5=1                !force PIXEL coordinates
  correct/badpix {FX_TABLE}                     !correct BAD pixels
  FX_TABLE = " "                                !.. only for calibration frames
endif
if {P6} .eq. -1 @p fors_corr_cosm.prg           !correct cosmics




BRANCH {p6} 0,1,2,3,4,5,6,7,8,9,10 RED,FFC,FFF,SKC,SKF,FRI,BIA,RED,RED,RED,RED
goto fin

BIA:
      -copy {CCD_IN}.bdf {rfrm}
      {rfrm},{FORS_IDSTRNG} = "{RBIAS_ID}"
      reduce/info {rfrm} 1 {scale}
      goto fini
RED:                                            !coordinates have to be the
      mess/out 3 {procid} "Do the REDUCTION ..."

      if MESS_LEVEL .gt. 3 then
        reduc/ccd P2={rfrm} P3="Y"              !standard reduction
      else
        reduc/ccd P2={rfrm} P3="Y" >Null        !standard reduction
      endif
      if m$exist(rfrm) .gt. 0 then
        mess/out 3 {procid} "REDUCTION finished."
      else
        if m$existd("{CCD_IN}.bdf","BPXLCORR") .gt. 0 then
          -copy {CCD_IN}.bdf {rfrm}
          write/desc {rfrm} CCDSTAT/i/3/1 1
          mess/out 3 {procid} "REDUCTION finished."
        else
          mess/out 1 {procid} "REDUCTION failed !"
          return -1
        endif
      endif
      if {P6} .ge. 8 .AND. {P6} .lt. 10 then    !CCDSCALE is used by chkcal
        statistic/image {CCD_IN} {IM_SEC} ? ? GN >Null
        write/desc {CCD_IN} CCDSCALE/r/1/1 {OUTPUTR(8)} !flat: median
      else if {P6} .eq. 6 then                          !bias: 1.0
        write/desc {CCD_IN} CCDSCALE/r/1/1 1.0
      else                                              !dark, science: exptime
        write/desc {CCD_IN} CCDSCALE/r/1/1 {{CCD_IN},O_TIME(7)}
      endif
      ityp = ityp - 5   !1:bias, 2:dark, 3:skyff, 4:scrff, 5:std, <0:science
      reduce/info {rfrm} {ityp} {scale}
      goto fini
FFC:  
      mess/out 3 {procid} "Determine the ILLUMINATION CORRECTION ..."
      illcor/ccd  {CCD_IN} {rfrm}               !illumination correction (FF)
      mess/out 3 {procid} "ILLUMINATION CORRECTION finished."
      goto fini
FFF:  
      mess/out 3 {procid} "Determine the ILLUMINATION CORRECTED FLATFIELD ..."
      illflat/ccd {CCD_IN} {rfrm}               !illum. corrected flatfield
      mess/out 3 {procid} "ILLUMINATION CORRECTED FLATFIELD finished."
      goto fini
SKC:  
      mess/out 3 {procid} "Determine the SKY-ILLUMINATION CORRECTION ..."
      skycor/ccd  {CCD_IN} {rfrm}               !illumination correction (SKY)
      if m$exist(rfrm) .gt. 0 then
        mess/out 3 {procid} "SKY-ILLUMINATION CORRECTION finished."
        shortlog/out 1 "ILLUMINATION correction determined."
      else
        mess/out 1 {procid} "SKY-ILLUMINATION CORRECTION failed."
        return -1
      endif
      goto fini
SKF:  
      rfrm = "{CCD_IN}_sup.bdf"
      mess/out 3 {procid} "Determine {rtyp} ..."
      @p fors_parsepath {SC_FFFRM}
      SC_FFFRM = Q1 // Q2
      mess/out 4 {procid} ". Do the flatfielding of {CCD_IN}."
      flat/ccd {CCD_IN} f{CCD_IN} >Null
      !
      view/mult f{CCD_IN}.bdf 1 1 -3 P7="Flatfielded {etyp}"
      !
      if m$exist("f{CCD_IN}.bdf") .gt. 0 then
        mess/out 4 {procid} ". Do the smoothing."
        skyflat/ccd f{CCD_IN} {rfrm} >Null      != scr * [sky/scr]_smoothed
        !
        view/mult f{CCD_IN}.bdf 1 1 -3 P7="Smoothed & Flatfielded {etyp}"
        !
        if m$exist(rfrm) .gt. 0 then
          mess/out 3 {procid} "{rtyp} created."
          shortlog/out 1 "{rtyp} created."
        else
          mess/out 1 {procid} "Creation of {rtyp} failed !"
          return -1
        endif
      else
          mess/out 1 {procid} "Creation of {rtyp} failed !"
          return -1
      endif
      write/desc {rfrm} {h_dprtype}/c/1/15 "FLAT           "
      write/desc {rfrm} IDENT " " ALL
      {rfrm},IDENT = {rtyp}
      plev = d_empty

      goto fini

FRI:
      mess/out 3 {procid} "Determine the FRINGE CORRECTION ..."
      frcor/ccd {CCD_IN} {rfrm}                 !fringe correction
      if m$exist(rfrm) .gt. 0 then
        mess/out 3 {procid} "FRINGE CORRECTION finished."
        shortlog/out 1 "FRINGE correction done."
      else
        mess/out 1 {procid} "FRINGE CORRECTION failed."
        return -1
      endif




fini:

if m$exist(rfrm) .eq. 1 then
  swap/units {CCD_IN} WORLD                     !restore WORLD coordinates.
  swap/units {rfrm} WORLD                       !overscan regions had been
  if m$existd(rfrm,"CCDSTAT") .gt. 0 trimflag = {{rfrm},CCDSTAT(2)}
  @p fors_img_obs_reduce,RM_OSCAN {CCD_IN} {P6} !cut off already for the input
  @p fors_img_obs_reduce,RM_OSCAN {rfrm} {P6}   !science & std.star frames
                                                !so BOTH have no scanregions
                                                !anymore.

  if m$existd(rfrm,"CCDSTAT") .gt. 0 .and. {{rfrm},CCDSTAT(2)} .eq. 1 then
    if trimflag .eq. 0 shortlog/out 1 "Pre- and overscan regions chopped off."
  endif
  @p fors_img_obs_reduce,CHECK_O_TIME {rfrm} {procid}

  add/icat result.cat {rfrm} >Null
  get/outname result.cat {P3} {rfrm}            !output name (without path)


  if nice_crea .eq. "y" then     !plot result frame (DISP: 2)
    !
    !*******************************************
    @p fors_nicedisp {rfrm} ? 7. 0,0 ? ? 1.,3. n
    !*******************************************
    if {q1} .eq. 1 then
        write/out " "
        mess/out 1 {procid} "Displaying of frame failed !"
        mess/out 1 {procid} "({rfrm})"
      return 1
    endif
    !
        mess/out 4 {procid} "  loading image into display 2"
    assign/display d,2 >Null
    load/image {rfrm} ? {outputi(1)},avg c,c
    !
    if NICE_HCOPY(:1) .ne. "N" then
      mess/out 3 {procid} "  creating hardcopy <{CCD_IN}_red.ps>"
      copy/display p5=noprint >Null
      -delete screen{mid$sess(11:12)}.ima
      -rename screen{mid$sess(11:12)}.ps {CCD_IN}_red.ps
    endif
    !
  endif


  @p fors_paragraph 3

  procid = m$upper(procid)
  classify/product {rfrm} {rtyp} {ptyp} {plev} OK ? y
  write/desc {rfrm} {fors_idstrng}/c/1/30 {rtyp}
  delete/desc {rfrm} BPXLCORR n
  longlog/out 3 "Reduced image is: {rfrm}"
  longlog/out 3 "  (not normalized by exposure time)"

! Create product header
  finish/prohead result.cat 2
  FINISH/PIPELINE result.cat {P3} CCDRED
  if {Q1} .gt. 0 goto ABORT
  finish/longlog reduceI.cat result.cat {rfrm} 2 {p3}

endif

fin:

set/midas user=expert
delete/comm reduce/ccd
delete/comm flat/ccd
delete/comm skyflat/ccd
! mess/out 4 {procid} "Reset commands"
set/midas user=user

LOG/REC fors_img_obs_reduce OUT
@p fors_paragraph 4

out_a = rfrm
return 0


ABORT:
finish/longlog reduceI.cat -1 ? 1 {p3}
mess/out 1 "img_obs_reduce" "Error in FINISH/PIPELINE !"
write/error 9,APP,NO "Error in FINISH/PIPELINE !"


!=== SUBROUTINES ==============================================================

ENTRY EXPO_TYPE

define/param P1 ? c "Enter process-mode:"
define/param P2 ? c "Enter input image catalog:"
!
! local keywords
!
define/local ityp/i/1/1 0
define/local noff/i/1/1 0
define/local procid/c/1/25 " " ALL
!
! Procedure NAME due to selected MODE (P1):
!
define/local pnam/c*25/1/11
!
pnam(1) = "img_obs_reduce"
pnam(2) = "img_cal_illcor"
pnam(3) = "img_cal_illflat"
pnam(4) = "img_cal_skycor"
pnam(5) = "img_cal_superflat"
pnam(6) = "img_cal_fringe"
pnam(7) = "img_cal_reduce"
pnam(8) = "img_obs_reduce"
pnam(9) = "img_obs_reduce"
pnam(10) = "img_obs_reduce"
pnam(11) = "img_obs_reduce"
!
! INPUT exposure types due to selected MODE
!
define/local etyp/c*25/1/11
etyp(1) = SCIENCE_ID            !SCIENCE_IMG
etyp(2) = MFLAT_ID              !MASTER_SCREEN_FLAT  (-> illum. corr.)
etyp(3) = MFLAT_ID              !MASTER_SCREEN_FLAT  (-> illum. corr. flatf.)
etyp(4) = MSKY_ID               !MASTER_SKY_FLAT     (-> illum. corr.)
etyp(5) = MSKY_ID               !MASTER_SKY_FLAT     (-> illum. corr. flatf.)
etyp(6) = MSKY_ID               !MASTER_SKY_FLAT     (-> fringe corr.)
etyp(7) = BIAS_ID               !BIAS
etyp(8) = DARK_ID               !DARK
etyp(9) = SKY_ID                !SKY_FLAT
etyp(10) = FLAT_ID              !SCREEN_FLAT
etyp(11) = STD_ID               !STANDARD
!
! OUTPUT exposure types due to selected MODE
!
define/local rtyp/c*25/1/11
rtyp(1) = RSCIENCE_ID           !SCIENCE_REDUCED
rtyp(2) = "undefined"           !
rtyp(3) = "undefined"           !
rtyp(4) = "undefined"           !
rtyp(5) = SSKY_ID               !SUPER_SKY_FLAT
rtyp(6) = "undefined"           !
rtyp(7) = RBIAS_ID              !REDUCED_BIAS (bad pixels removed)
rtyp(8) = RDARK_ID              !REDUCED_DARK
rtyp(9) = RSKY_ID               !REDUCED_SKY_FLAT
rtyp(10) = RFLAT_ID             !REDUCED_SCREEN_FLAT
rtyp(11) = RSTD_ID              !STANDARD_REDUCED

if {P1} .lt. -1 .OR. {P1} .gt. 10 then
  mess/out 1 img_obs_reduce "Wrong mode selected: P6={P1}! ABORT!"
  return -1
endif


ityp = {P1} + 1
if ityp .lt. 1 then     !cosmic removal
  ityp = 1
endif
procid = pnam({ityp})


if {P1} .eq. 0 .OR. {P1} .ge. 6 then
  mess/out 3 {procid} "START to reduce image ..."
else if {P1} .gt. 0 .AND. {P1} .lt. 6 then
  mess/out 3 {procid} "START to perform flatfield correction ..."
  search/icat {p2} {MSKY_ID} >Null              !identify MASTER_SKY_FLAT
  if m$len(out_a) .ne. 0 then
    etyp(5) = MSKY_ID
    rtyp(5) = SSKY_ID
    noff = 0
  else
    noff = 1
  endif
  if noff .eq. 1 then                           !NO master_sky, so look for:
    search/icat {p2} {MNIGHT_ID} >Null          !... MASTER_NIGHT_FLAT
    if m$len(out_a) .ne. 0 then
      etyp(5) = MNIGHT_ID
      rtyp(5) = SNIGHT_ID
      noff = 0
    else
      noff = 1
    endif
  endif

  if noff .eq. 1 then
    mess/out 1 {procid} "Neither {MSKY_ID}\n nor {MNIGHT_ID} frame found !"
    return -1
  endif 
  
endif
if {P1} .gt. 0 .AND. {P1} .lt. 10 then
  outputi(1) = 1
else
  outputi(1) = 0
endif

branch ityp 1,11,5 SCI,STD,SKY          !ESO.PRO-keywords depend on ityp
SCI:
  out_c   = d_red
  outputc = d_det
  goto hcnarb
STD:
  out_c   = d_tmp
  outputc = d_det
  goto hcnarb
SKY:
  out_c   = d_tmp
  outputc = d_empty
hcnarb:
out_a   = etyp({ityp})                  !exposure type of input
out_b   = rtyp({ityp})                  !exposure type of result
return {ityp} {outputi(1)} {procid}


ENTRY CHECK_O_TIME

define/param P1 ? c "Enter frame name: "
define/param P2 ? c "Enter name of calling procedure: "

if {{P1},O_TIME(1)} .le. 1997.0 then
  mess/out 2 {P2} "Bad observation date detected; recalculate !"
  @p fors_jd2date {P1}
  {P1},O_TIME(1) = outputi(1)
  {P1},O_TIME(2) = outputi(2)
  {P1},O_TIME(3) = outputi(3)
endif



ENTRY RM_OSCAN
define/parameter P1 ? c "image:"
define/parameter P2 ? n "exposure type:"

define/local i1/i/1/1 0
define/local j1/i/1/1 0

if {P2} .eq. 6 then                     !bias has still overscan regions
  trim/frame {P1}
  if {Q1} .eq. 0 {P1},CCDSTAT(2) = 1
else                                    !overscan regions already cut-off
  do i1 = 1 fors_nout
    j1 = fors_outi({i1})                        !remove pre-/postscan
    if m$existd(P1,h_pscan({j1})) .eq. 1 then
      delete/descr {P1} {h_pscan({j1})}         !descriptors as the overscan
      if m$existd(P1,"CCDSTAT") .gt. 0 {P1},CCDSTAT(2) = 1
    endif
    if m$existd(P1,h_oscan({j1})) .eq. 1 then
      delete/descr {P1} {h_oscan({j1})}         !regions have been cut off.
      if m$existd(P1,"CCDSTAT") .gt. 0 {P1},CCDSTAT(2) = 1
    endif
  end do
endif


ENTRY PLOT
define/parameter P1 ? c "Enter image catalog which should be displayed:"
define/parameter P2 ? c "Enter data type (1..10):"

if {p2} .eq. 6 return 0                 !bias: uses NO calibration frames.

define/local num/i/1/1 0

show/icat {P1} >Null
num = outputi(1)                        !number of frames in catalog

if num .gt. 0 then
  outputi(1) = m$sqrt({num})            !Determine num. of columns and rows
  outputr(1) = 1.0*{num} / outputi(1)   !takes only integer part  --> columns
  outputi(2) = m$nint(outputr(1))       !nearest integer (3.5->4) --> rows
  outputi(3) = outputi(1) * outputi(2)
  if outputi(3) .lt. {num} then
    if outputi(1) .lt. outputi(2) then
      outputi(1) = outputi(1) + 1
    else
      outputi(2) = outputi(2) + 1
    endif
  endif
  outputc = "CALIBRATION FRAMES used for REDUCTION"
  inputc = "N"
  if NICE_HCOPY(1:1) .ne. "N" inputc = "H,{P1}.ps"
  !
  view/mult {P1} {outputi(1)} {outputi(2)} -3 P7="{outputc}" P8={inputc}
  !
endif
