! $Id: fors_lss_cal_align.prg,v 1.8 1999/09/16 16:08:47 nrainer Exp $   
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1999  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_lss_cal_align.prg,v $
!
!.KEYWORDS    LSS/MOS standard star identification
!
!.USAGE       @p fors_lss_cal_align  frm_in frm_ref frm_out mode climit
!
!.INPUT       frm_in      (P1)    Set of raw input frames to be processed
!                                 (relative flux spectrum of star(s) with
!                                 object world coordinates).
!             frm_ref     (P2)    Set of reference frames (reference flux
!                                 table with standard star world coordinates).
!             mode        (P4)    Indicator of observing mode (LSS, MOS)
!             climit      (P5)    Confidence limit for standard star
!                                 identification (Not yet supported!).
!
!.OUTPUT      frm_out     (P3)    Set of output frames (relative flux spectrum
!                                 table of standard stars) 
!
!.RETURNS     The recipe returns 0 on success, otherwise 1. Keyword Q1 is
!             updated with the exit status. 
!
!.PURPOSE     The input set of frames consists of a table of extracted spectra
!             and the associated source positions. The objects are associated
!             with spectrophotometric standard stars by looking up these
!             positions in the standard star flux table, taking position
!             uncertainties into account. The extracted spectra of all
!             identified standard stars are then written to the output
!             table.
!
!.ENVIRON     FORS context, MOS context
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1999/02/06:  Creation
!             0.6  1999/04/28:  Longlog implemented (SB)
!             1.3  1999/09/03:  Product header generation implemented (SB).
!
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

define/param p1 frm_in.set   c  "Enter input frame set:"
define/param p2 frm_ref.set  c  "Enter reference frame set:"
define/param p3 frm_out      c  "Enter output frame set:"
define/param p4 LSS          c  "Enter observing mode:"
define/param p5 +            c  "Enter source identification confidence limit:"


! Procedure identification
define/local procID/c/1/80      "{p4}CalAlign"


! Recipe identification
define/local recID/c/1/9        "CAL_ALIGN"


! Local constants
define/local stdlst/c/1/20      "fors_standards.tbl"


! Local variables
define/local nstd/i/1/1         0
define/local ctrl/i/1/3         0             all
define/local catal/i/1/1        0
define/local climit/r/1/1       0.
define/local obsmod/c/1/3       " "           all
define/local setup/c/1/20       " "           all
define/local cat_accept/c/1/20  "accept.cat"
define/local cat_reject/c/1/20  "reject.cat"
define/local cat_result/c/1/20  "result.cat"
define/local cat_entry/c/1/512  " "           all



! Connect to pipeline
log/rec "{p4}_{recID}" IN
mess/out 3 {procID} -
    "Recipe {p4}_{recID}: Identify std. star spectra"


! Process parameters
obsmod = m$upper(p4)
setup = "{obsmod}pipe"

if m$tstno(p5) .eq. 1 then
  climit = {p5}
else
  mess/out 1 {procID} "Invalid confidence limit for spectrum identifications!"
  clean/pipeline xmos,spec
  return 1
endif


! Startup context, initialize environment and prepare input 
! frames for processing
@p fors_paragraph 1

initialize/pipeline {p1} {p2} {cat_accept} xmos {setup} {obsmod} -
                    {cat_reject} {cat_result}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Pipeline initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Initialize product header
init/prohead {p1} {p2} 2,"Identification of standard star spectra" -
   obsmode,{p4} climit,{p5}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Product header initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Initialize longlog
init/longlog {p1} {p2} 2 "obsmode: {p4}" "climit: {p5}"
if {q1} .ne. 0 then
  mess/out 1 {procID} "Longlog initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


@p fors_paragraph 2


! Update session keywords
read_cat_entry:

store/frame cat_entry {cat_accept} 1 exit_loop
fors_id = m$value({cat_entry},{FORS_IDSTRNG})
fors_id = m$upper(fors_id) 

if fors_id .eq. ESTD_ID then
  extobjec = cat_entry
  ctrl(1) = 1
elseif fors_id .eq. STDFLX_ID then
  refflux = cat_entry
  ctrl(2) = 1
endif

goto read_cat_entry


exit_loop:

if ctrl(1) .eq. 0 then
  mess/out 1 {procID} "No extracted spectrum table found!"
  clean/pipeline xmos,spec
  return 1
else if ctrl(2) .eq. 0 then
  mess/out 1 {procID} "No reference flux table present!"
  clean/pipeline xmos,spec
  return 1
endif


! Identify standard star spectra
identify/standard {extobjec} {refflux} {stdlst}
if {q1} .eq. 0 then
  if outputi(1) .gt. 0 then
    nstd = outputi(1)
  else
    mess/out 1 {procID} "No standard stars found in {extobjec}!"
    clean/pipeline xmos,spec
    return 1
  endif
else
  mess/out 1 {procID} "Standard star identification failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Copy all but the standard descriptors to the standard star table
copy/dd {extobjec} *,3 {stdlst}

@p fors_paragraph 3


! Prepare output catalog
classify/product {stdlst} {ASTD_ID} {d_pre} {d_geo},{d_det} OK
write/desc {stdlst} {FORS_IDSTRNG}/c/1/20 {ASTD_ID}

add/icat {cat_result} {stdlst} >Null


! Update internal log file
shortlog/out 1 "Standard star spectra identified."


! Create product header
finish/prohead {cat_result} 2


! Shutdown pipeline
finish/pipeline {cat_result} {p3} xmos,spec {stdlst}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Error leaving pipeline environment!"
  finish/longlog {cat_accept} -1 ? 1 {p3}
  clean/pipeline xmos,spec
  return 1
endif


! Close longlog
finish/longlog {cat_accept} {cat_result} {stdlst} 2 {p3}


! Disconnect from pipeline
log/rec "{p4}_{recID}" OUT

@p fors_paragraph 4

return 0
