! $Id: fors_lss_qc1_chkobs.prg,v 1.37 2000/04/12 11:06:44 nrainer Exp $   
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
! .COPYRIGHT  (C) 1999  European Southern Observatory (ESO)
!
! .IDENT       $RCSfile: fors_lss_qc1_chkobs.prg,v $
!
! .KEYWORDS   Spectroscopical quality
!
! .PURPOSE    Pipeline wrapper and frame classification for
!             fors_chkobs.prg:
!             Compute signal quality values of each object spectrum, and that
!             seeing (FWHM) per wavelength bin and signal-to-noise per
!             wavelength.
!
! __________
!
! .USAGE      @p fors_lss_qc1_chkobs in ref [out] [bin] [mod]
!
! .INPUT      in   (P1): single input frame (ascii list with single frame);
!                        possible frame type (m(o)d(e) = [LSS|MOS]):
!                          reduced standard star image (STANDARD_REDUCED_<md>);
!                            default: none
!
!             ref  (P2): super recipe mode: "+" (reference frames are included
!                        in P1),
!                        single recipe mode: set of reference frames (ascii
!                        list);
!                        needed frames:
!                          sky background image (STANDARD_SKY_<md>),
!                          tbl.w. aligned spectr[um|a] (STANDARD_ALIGNED_<md>) -
!                            if not found:
!                          table w. extracted spectra (STANDARD_EXTRACTED_<md>);
!                        also needed frame type:
!                          table with purging information (DESC_PURGE_TABLE);
!                            default: none
!
!             out  (P3): single recipe mode: basename for ascii list with
!                        output frames,
!                        super recipe mode: again input frame and
!                        existing image catalogue (".cat"), separated by a
!                        comma;
!                          sky-subtracted image (STANDARD_SOURCE_<md>),
!                          object_quality table (CHKOBS_TABLE);
!                            default: `PROID' defined in procedure (single mode)
! 
!             bin  (P4): number of x-pixels per wavelength bin for computation
!                        of FWHM, {MINFITXPIX} (see below) <= bin <= total no.
!                        of x-pixels of <P1>;
!                          default: <>:sqrt(total no. of x-pixels), at least
!                                   {MINFITXPIX}
!
!             mod  (P5): flag for observing mode <md>;
!                          default: LSS
!
! .OUTPUT     "<P1>_skysub.fits": sky-subtracted frame
!
!             "<P1>_qc3d.tfits" : This is a three-dimensional table !
!                                 It is built as follows:
!
!                                     _  array (values of columns;
!                                     /| depth = tot. no. of <P1>'s x-axis pix.)
!                                    /
!                                   +-------> columns (see below)
!                                   |
!                                   V rows (objects)
!
!                                 The object planes (i.e. rows) can be converted
!                                 to normal two-dimensional tables (after con-
!                                 version TFITS->TBL) with the help of
!
!                                   extract/t3d 3Dtbl 3Dcol 3Drow 2Dtbl 2Dcol +
!
!                                 for each column; `2Dtbl' will be created.
!
!               The columns are:  1 SLIT    : Slit number
!                                 2 RA,
!                                 3 DEC     : Positions
!                                 4 LAMBDA  : WAVELENGTHS
!                                 5 YSTART,
!                                 6 YEND    : Y-RANGES OF OBJECTS
!                                -----------
!                                 7 FLUX,     (sky-subtracted)
!                                 8 SKY     : BRIGHTNESSES
!                                -----------
!                                 9 LAM_CEN : Central wavelenghts of bins
!                                10 STATUS  : BIN QUALITY (mnd.f.FWHM,opt.f.S2N)
!                                11 FWHM    : SEEING (per bin)
!                                12 S2N     : SIGNAL-TO-NOISE (per wavelength)
!
!               Legend for column 10 (bin qual.flags):
!
!                 <0 : estimation failed (during Gauss fit)
!                  0 : O.K.
!                  1 : ambiguous sources
!                  2 : no source
!                  3 : iteration failed
!                 >3 : routine error
!
!               Formula for column 12 (signal-to-noise):
!
!                 All fluxes and readout noise in e- , w. original expo. time ,
!                 the object and sky fluxes are integrated (per wavelength,
!                 perpendicular to dispersion direction):
!
!                                     flux_obj.
!                 ----------------------------------------------------
!                 sqrt( flux_obj. + flux_sky + #ypix./wave.*(RON**2) )
!
! .RETURN     Q1:  0: successful return
!                 -1: a warning message occured
!                  1: an error occured
! __________
!
! .ALGORITHM  -
!
! .COMMENT    -
!
! .ENVIRON    FORS context
!
! .LANGUAGE   MIDAS 98NOV
!
! .AUTHOR     N.Rainer  DPG/DMD/ESO
!
! .VERSION    1.2  99/02/26  Creation
!                  99/03/15  Release
!             1.3  99/09/23  Release
!
!-------------------------------------------------------------------------------
!
!
crossref in ref out bin mod
!
!
define/local CFMDESC/c/1/7 "CF_MEAN"
define/local RONMDESC/c/1/8 "RON_MEAN"
!
define/local MINFITXPIX/i/1/1 27 ? +lower_levels
!                             ! cen/gau min.(x,y): 3 pixel ; better: 27
!
define/local FMTDEF/c/1/17 "I1 E14.7,E23.15"
!
define/local warn/i/1/1 0
write/keyword q1 0
!
!
!P5(mod):
!-------
define/parameter p5 lss ? -
  "        Enter flag for observing mode  [def:lss|mos]: "
define/local insmod/c/1/10 "{p5}"
!
inputc = m$lower(insmod)
define/local PROID/c/1/40 "{inputc}_qc1_chkobs"
!
if insmod .ne. "lss" .and. insmod .ne. "mos" then
        write/out " "
        mess/out 1 {PROID} "Entered observing mode unknown !"
        mess/out 1 {PROID} "({insmod})"
  goto error
endif
!
if fors_maxlin .le. 65534. then
  define/local FINDVAL/i/1/1 {fors_maxlin}
  define/local MASKVAL/i/1/1 65535
else
        write/out " "
        mess/out 1 {PROID} "Global key fors_maxlin={fors_maxlin} >65534. !"
  goto error
endif
!
!
if p1 .eq. "??" then
  set/format I1
  w/o "      ******************************************************************"
  w/o "      in   (P1): single input frame (ascii list with single frame);"
  w/o "                 possible frame type (m(o)d(e) = [LSS|MOS]):"
  w/o "                   reduced standard star image (STANDARD_REDUCED_<md>);"
  w/o "                     default: none"
  w/o " "
  w/o "      ref  (P2): super recipe mode: "+" (reference frames are included"
  w/o "                 in P1),"
  w/o "                 single recipe mode: set of reference frames (ascii"
  w/o "                 list);"
  w/o "                 needed frames:"
  w/o "                   sky background image (STANDARD_SKY_<md>),"
  w/o "                   tbl.w. aligned spectr[um|a] (STANDARD_ALIGNED_<md>) -"
  w/o "                     if not found:"
  w/o "                   table w. extracted spectra (STANDARD_EXTRACTED_<md>);"
  w/o "                 also needed frame type:"
  w/o "                   table with purging information (DESC_PURGE_TABLE);"
  w/o "                     default: none"
  w/o " "
  w/o "      out  (P3): single recipe mode: basename for ascii list with"
  w/o "                 output frames,"
  w/o "                 super recipe mode: again input frame and"
  w/o "                 existing image catalogue (".cat"), separated by a"
  w/o "                 comma;"
  w/o "                   sky-subtracted image (STANDARD_SOURCE_<md>),"
  w/o "                   object_quality table (CHKOBS_TABLE);"
  w/o "                     default: `PROID' defined in procedure (single mode)"
  w/o " "
  w/o "      bin  (P4): number of x-pixels per wavelength bin for computation"
  w/o "                 of FWHM, {MINFITXPIX} <= bin <= total no. of x-pixels"
  w/o "                 of <P1>;"
  w/o "                   default: <>:sqrt(total no. of x-pixels), at least"
  w/o "                            {MINFITXPIX}"
  w/o " "
  w/o "      mod  (P5): flag for observing mode <md>;"
  w/o "                   default: LSS"
  w/o "      ******************************************************************"
  set/format
  return/exit
endif
!
!
!          PREPARATIONS FOR RECIPE :
!          ***********************
!
log/recipe {PROID} in
!
        mess/out 3 {PROID} "Check quality of standard star spectra ..."
!
@p fors_paragraph 1 1  ! -------- INITIALIZATION --------
!
set/format {FMTDEF}
!
!
!          INPUT :
!          *****
!
!P1(in):
!------
if p1 .eq. "?" mess/out 2 {PROID} " "
define/parameter p1 ? ? -
  "        Enter name of input frame: "
define/local inpipe_frm/c/1/76 "{p1}"
!
if m$exist(inpipe_frm) .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Entered list with single input frame"
        mess/out 1 {PROID} "is not present or can not be opened !"
        mess/out 1 {PROID} "({inpipe_frm})"
  goto error
endif
!
define/local filecode/i/1/2 0 all
open/file {inpipe_frm} read filecode
if filecode(1) .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Unable to open entered list !"
        mess/out 1 {PROID} "({inpipe_frm})"
  goto error
endif
read/file {filecode(1)} outputc
if filecode(2) .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Entered list is empty !"
        mess/out 1 {PROID} "({inpipe_frm})"
  close/file {filecode(1)}
  goto error
endif
close/file {filecode(1)}
!
!
!P2(ref):
!-------
if p2 .eq. "?" mess/out 2 {PROID} " "
define/parameter p2 ? ? -
  "        Enter name of reference frame set: "
define/local inpipe_ref/c/1/76 "{p2}"
!
! No tests like for P1 - in super recipes reference frame is included in P1 .
!
!
!P3(out):
!-------
define/parameter p3 {PROID} ? -
  "        Enter basename of out.list(single)/exist.cat.".cat"(super): "
define/local outpipe/c/1/76 "{p3}"
!
!
!P4(bin):
!-------
define/parameter p4 c ? -
  "        Enter binsize [def.:<c>:sqrt(xpix),>={MINFITXPIX}]: "
if m$tstno(p4) .eq. 1 then  ! use entered binsize
  outputi(1) = {p4}         ! cut possible decimals
  define/local binflag/c/1/23 "{outputi(1)}"
  !
  define/local binsize/i/1/1 {p4}
  if binsize .lt. MINFITXPIX then
        write/out " "
        mess/out 1 {PROID} -
          "Binsize out of range ({binsize} ; binsize>={MINFITXPIX}) !"
    goto error
  endif
  !
else                        ! calculate binsize=sqrt(xpix)
  define/local binflag/c/1/1 "?"
endif
!
!
!
!          DEFINITION OF VARIABLES :
!          ***********************
!
!     "Getting input frames":
!
  define/local obsmod/c/1/10 " " all
  define/local par4/c/1/60 " " all
!
!     "Identifying input frames":
!
  define/local catal/i/1/1 0
  define/local topicima/c/1/76 " " all
  define/local id/c/1/25 " " all
  !
  define/local stdID/c/1/25 " " all
  define/local skyID/c/1/25 " " all
  define/local t3dID1/c/1/25 " " all
  define/local t3dID2/c/1/25 " " all
  !
  define/local redima/c/1/76 " " all
  define/local skyima/c/1/76 " " all
  define/local dim3tbl/c/1/76 " " all
  !
  define/local instd/i/1/1 0
  define/local refsky/i/1/1 0
  define/local reft3d/i/1/1 0
  !
  define/local insdesc/c/1/60 " " all
!
!     "Overexposure":
!
  define/local ystartpix/i/1/1 0
  define/local yendpix/i/1/1 0
!
!     "Mean CF/RON":
!
  define/local portloop/i/1/1 0
  define/local cfsum/d/1/1 0.
  define/local ronsum/d/1/1 0.
  define/local cf/d/1/1 0.
  define/local ron/d/1/1 0.
!
!     "Overscan sections":
!
  define/local psc/i/1/1 0
  define/local osc/i/1/1 0
!
!     "Calling of procedure":
!
  define/local cola/c/1/100 " " all
  define/local colb/c/1/100 " " all
  define/local colc/c/1/100 " " all
  define/local cold/c/1/100 " " all
  define/local tcol/c/1/100 " " all
  define/local char/i/1/1 0
  define/local qc3dtbl/c/1/76 " " all
  define/local subima/c/1/76 " " all
!
!
!
!          LOADING OF GLOBAL KEYS (SINGLE RECIPE MODE) ,
!          CONVERSION FITS->BDF/TBL ,
!          CLASSIFICATION OF INPUT FRAMES :
!          *********************************************
!
        mess/out 4 {PROID} "Getting input frames"
!
!
obsmod = m$upper(insmod)
!
initialize/pipeline {inpipe_frm} {inpipe_ref} -
  inpipe_good.cat ? + {obsmod} inpipe_bad.cat ?
if {q1} .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Pipeline initialization failed !"
  goto error
endif
-delete inpipe_bad.cat
!
if m$tstno(binflag) .eq. 1 then
  par4 = "{binflag}"
else
  par4 = "sqrt(total no. of x-pixels), at least {MINFITXPIX}"
endif
!
! Initialize product header:
init/prohead {inpipe_frm} {inpipe_ref} -
  2,"Check quality of standard star spectra" binsize,"{par4}" obsmode,{obsmod}
if {q1} .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Product header initialization failed !"
  goto error
endif
!
initialize/longlog {inpipe_frm} {inpipe_ref} 2 -
                   "no. of x-pixels per wavelength bin: {par4}" -
                   "obs.mode: {obsmod}"
!
!
!          IDENTIFICATION VIA DESCRIPTORS ,
!          MAPPING OF FRAMES TO PARAMETERS  :
!          ********************************
!
        mess/out 4 {PROID} "Identifying input and reference frames"
!
stdID = RSTD_ID
skyID = SBKG_ID
t3dID1 = ASTD_ID
t3dID2 = ESTD_ID
!
create/icat {PROID}.cat null {FORS_IDSTRNG} >Null  ! only for finish/longlog
instd = 0
refsky = 0
reft3d = 0
!
ident_loop1:
  store/frame topicima inpipe_good.cat 1 end_ident_loop1
  id = "{{topicima},{fors_idstrng}}"
  !
  if id .eq. "{stdID}" then
    redima = topicima
    add/icat {PROID}.cat {topicima} >Null
        mess/out 4 {PROID} "{topicima} : {id} => input"
    instd = instd+1
  elseif id .eq. "{skyID}" then
    skyima = topicima
        mess/out 4 {PROID} "{topicima} : {id} => reference"
    refsky = refsky+1
  elseif id .eq. "{t3dID1}" then  ! aligned spectr[um|a]
    dim3tbl = topicima
        mess/out 4 {PROID} "{topicima} : {id} => reference"
    reft3d = reft3d+1
  endif
  !
  goto ident_loop1
end_ident_loop1:
if reft3d .eq. 0 then
  catal = 0
  ident_loop2:
    store/frame topicima inpipe_good.cat 1 end_ident_loop2
    id = "{{topicima},{fors_idstrng}}"
    !
    if id .eq. "{t3dID2}" then  ! extracted spectra
      dim3tbl = topicima
        mess/out 4 {PROID} "{topicima} : {id} => reference"
      reft3d = reft3d+1
    endif
    !
    goto ident_loop2
  end_ident_loop2:
endif
-delete inpipe_good.cat
!
if instd .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No matching input standard star frame present !"
  goto error
elseif instd .gt. 1 then
        write/out " "
        mess/out 1 {PROID} "More than one input standard star frame present !"
  goto error
endif
!
if refsky .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No matching sky background frame present !"
  goto error
elseif refsky .gt. 1 then
        write/out " "
        mess/out 1 {PROID} "More than one sky background frame present !"
  goto error
endif
!
if reft3d .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "No matching spectroscopy table present !"
  goto error
elseif reft3d .gt. 1 then
        write/out " "
        mess/out 1 {PROID} "More than one spectroscopy table present !"
  goto error
endif
!
insdesc = "{{redima},{h_instrume}}"
!
!
!          EXTRACTING SPECTRUM AS SUBFRAME
!          AND CHECKING FOR OVEREXPOSED PIXELS :
!          ***********************************
!
if m$existc(dim3tbl,":{c_wostart}") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Column is not present !"
        mess/out 1 {PROID} "({dim3tbl},:{c_wostart})"
  goto error
elseif m$existc(dim3tbl,":{c_woend}") .eq. -1 then
        write/out " "
        mess/out 1 {PROID} "Column is not present !"
        mess/out 1 {PROID} "({dim3tbl},:{c_woend})"
  goto error
endif
select/table {dim3tbl} :{c_woslit}.ne.NULL >Null  ! removal of empty rows
copy/table {dim3tbl} tmp.tbl
-rename tmp.tbl {dim3tbl}
if {{dim3tbl},TBLCONTR(4)} .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Table is empty !"
        mess/out 1 {PROID} "({dim3tbl})"
  goto error
endif
!
ystartpix = {{dim3tbl},:{c_wostart},@1}
yendpix = {{dim3tbl},:{c_woend},@1}
if ystartpix .ge. yendpix then  ! possible error in input table !
        write/out " "
        mess/out 1 {PROID} "Spectrum y-start >= y-end in 3D input table !"
        mess/out 1 {PROID} "({dim3tbl})"
  goto error
endif
!
        mess/out 3 {PROID} -
          "Extracting spectrum (ystart=@{ystartpix} yend=@{yendpix})"
!
extract/image spectrum.bdf = {redima} [<,@{ystartpix}:>,@{yendpix}]
create/icat spectrum.cat null {FORS_IDSTRNG} >Null
add/icat spectrum.cat spectrum.bdf >Null
!
        mess/out 3 {PROID} "Checking spectrum for overexposed pixels"
!
!******************************************************************
@p fors_overexpo spectrum.cat {FINDVAL},{MASKVAL} 0. {h_ovexmask} p
!******************************************************************
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Overexposure check failed !"
  goto error
endif
if {q2} .eq. 0 then
        write/out " "
        mess/out 1 {PROID} "Spectrum is overexposed !"
  goto error
endif
-delete spectrum.cat
!
!
!          CALCULATING MEAN CONVERSION FACTOR
!          AND MEAN READOUT NOISE FOR MULTIPLE
!          PORT READOUT  :
!          ***********************************
!
@p fors_paragraph 2 1  ! -------- PROCESSING --------
!
if fors_nout .gt. 1 then
  !
        mess/out 4 {PROID} "Calculating mean conv.fact. and RON of all ports"
  !
endif
!
cfsum = 0.
ronsum = 0.
do portloop = 1 fors_nout
  cfsum = cfsum + {{redima},{h_confac({fors_outi({portloop})})}}
  ronsum = ronsum + {{redima},{h_ron({fors_outi({portloop})})}}
enddo
cf = cfsum / fors_nout
ron = ronsum / fors_nout
write/descriptor {redima} {CFMDESC}/d/1/1 {cf}
write/descriptor {redima} {RONMDESC}/d/1/1 {ron}
!
!
!          DETERMINING OVERSCAN SECTIONS :
!          *****************************
!
!*************************
@p fors_get_imsec {redima}
!*************************
if {q1} .ne. 0 then
        write/out " "
        mess/out 1 {PROID} "Determination of overscan sections failed !"
  goto error
endif
psc = outputi(5)
osc = outputi(6)
!
!
!          *********************************
!          *                               *
!          *  CALLING OF MAIN PROCEDURE :  *
!          *                               *
!          *********************************
!
cola = "{c_woslit}"
colb = "{c_radeg},{c_decdeg}"
colc = "{c_sowave},{c_soflux}"
cold = "{c_wostart},{c_woend}"
tcol = "{cola},{colb},{colc},{cold}"
!
        mess/out 4 {PROID} "Applying procedure  "fors_chkobs" :"
!
!************************************************************
@p fors_chkobs {redima} {skyima} {dim3tbl},{tcol} {binflag} -
   {CFMDESC},{RONMDESC} p {psc},{osc}
!************************************************************
!
if {q1} .eq. -1 then
  warn = 1
elseif {q1} .eq. 1 then
        write/out " "
        mess/out 1 {PROID} "Quality determination of objects failed !"
  goto error
endif
!
  char = m$indexb(redima,".bdf")-1
  qc3dtbl = redima(1:{char})//"_qc3d.tbl"
  subima = redima(1:{char})//"_skysub.bdf"
!
create/icat outprog.cat null >Null
add/icat outprog.cat {subima} >Null
add/icat outprog.cat {qc3dtbl} >Null
-delete dirfile.ascii
!
!
!          UPDATING DESCRIPTORS OF OUT- AND INPUT FRAMES :
!          *********************************************
!
@p fors_paragraph 3 1  ! -------- TERMINATION --------
!
        mess/out 4 {PROID} "Classifying output table and image"
        mess/out 4 {PROID} "({qc3dtbl}, {subima})"
!
classify/product {subima} {SSRC_ID} {d_tmp} {d_geo},{d_det} OK ? y
write/descriptor {subima} {h_instrume}/c/1/60 {insdesc}
write/descriptor {subima} {FORS_IDSTRNG}/c/1/60 {SSRC_ID} !f.su.rec.
write/descriptor {subima} {h_proqcsts}/c/1/9 "CHECKED"
write/dhelp {subima} {h_proqcsts} "{i_proqcsts}"
longlog/out 1 "     Sky subtracted frame: {subima}"
!
shortlog/out 1 "Sky background subtracted"
!
classify/product {qc3dtbl} {CHKOBST_ID} {d_qcp} {d_geo},{d_det} OK
write/descriptor {qc3dtbl} {h_instrume}/c/1/60 {insdesc}
write/descriptor {qc3dtbl} {FORS_IDSTRNG}/c/1/60 {CHKOBST_ID} !f.su.rec.
write/descriptor {qc3dtbl} {h_proqcsts}/c/1/9 "CHECKED"
write/dhelp {qc3dtbl} {h_proqcsts} "{i_proqcsts}"
longlog/out 1 "     Table with signal qualities: {qc3dtbl}"
!
shortlog/out 1 "Seeing and S/N computed"
!
        mess/out 4 {PROID} "Endorsing QC status in input image"
!
write/descriptor {redima} {h_proqcsts}/c/1/9 "CHECKED"
write/dhelp {redima} {h_proqcsts} "{i_proqcsts}"
!
delete/descriptor {subima} {CFMDESC} n
delete/descriptor {subima} {RONMDESC} n
delete/descriptor {redima} {CFMDESC} n
delete/descriptor {redima} {RONMDESC} n
!
!
!          UPDATING OF OUTPUT CATALOGUE (SUPER RECIPE MODE) ,
!          CONVERSION BDF/TBL->FITS/TFITS (SINGLE MODE) :
!          **************************************************
!
        mess/out 4 {PROID} "Releasing output table and image"
!
! Create product header:
finish/prohead outprog.cat 2
!
finish/pipeline outprog.cat {outpipe} + {qc3dtbl}
!
finish/longlog {PROID}.cat outprog.cat {qc3dtbl} 2 {outpipe}
!
-delete {PROID}.cat
-delete outprog.cat
!
!
!          RESETTINGS AFTER RECIPE :
!          ***********************
!
set/format
!
log/recipe {PROID} out
!
@p fors_paragraph 4 1  ! ================
!
!
if warn .eq. 0 then
  return 0
else
  return -1
endif
!
error:
  set/format
  finish/longlog ? -1 ? 2
  write/out " "
  return 1
!
!
!-------------------------------------------------------------------------------
