! $Id: fors_mos_cal_extcoeff.prg,v 1.3 2000/06/20 09:04:43 nrainer Exp $   
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1998  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_mos_cal_extcoeff.prg,v $
!
!.KEYWORDS    LSS/MOS extinction coefficients
!
!.USAGE       @p fors_mos_cal_extcoeff  frm_in frm_ref frm_out mode 
!                                       area binsize
!
!.INPUT       frm_in      (P1)    Set of raw input frames to be processed
!                                 (relative flux spectrum of standard star(s)).
!             frm_ref     (P2)    Set of reference frames (reference flux
!                                 table, instrument response table).
!             mode        (P4)    Indicator of observing mode (LSS, MOS)
!             area        (P5)    Telescope effective area (area taking
!                                 occultation effects into account)
!             nbins       (P6)    Number of wavelength bins of final
!                                 extinction coefficient table.
!
!.OUTPUT      frm_out     (P3)    Set of output frames (table of extinction
!                                 coefficients)
!
!.RETURNS     The recipe returns 0 on success, otherwise 1. Keyword Q1 is
!             updated with the exit status. 
!
!.PURPOSE     The input set of frames consists of a set of tables of extracted
!             standard star spectra obtained at diffferent airmasses. The
!             extracted spectra of each standard star is rebinned using the
!             given bin size. All standard star spectra must be defined over
!             the same wavelength range. Each extracted spectrum is then
!             corrected for the wavelength dependent instrument response and
!             divided by the reference flux. Finally the preprocessed flux
!             ratios are converted to magnitudes a linear regression of the
!             flux ratios at different airmasses is performed for each 
!             wavelength bin to yield the atmospheric extinction coefficients.
!
!.ENVIRON     FORS context, MOS context
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1998/12/02:  Creation
!             1.3  1999/09/03:  Product header generation implemented (SB).
!
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

define/param p1 frm_in.set   c  "Enter input frame set:"
define/param p2 frm_ref.set  c  "Enter reference frame set:"
define/param p3 frm_out      c  "Enter output frame set:"
define/param p4 MOS          c  "Enter observing mode:"
define/param p5 49.0         c  "Enter telescope effective area:"
define/param p6 +            c  "Enter number of wavelength bins:"


! Procedure identification
define/local procID/c/1/80      "{p4}CalExtcoeff"


! Recipe identification
define/local recID/c/1/12       "CAL_EXTCOEFF"


! Local constants
define/local dwin/i/1/2         1,2
define/local gwin/i/1/2         1,2
define/local min_nstd/i/1/1     3
define/local extcat/c/1/20      "fors_extcoef.cat"
define/local stdcat/c/1/20      "fors_stdspec.cat"
define/local stdspec/c/1/20     "fors_stdspec.tbl"
define/local exttbl/c/1/20      "fors_extcoef.tbl"


! Local variables
define/local i/i/1/1            0
define/local j/i/1/1            0
define/local pos/i/1/1          0
define/local ctrl/i/1/3         0             all
define/local catal/i/1/1        0
define/local nbins/i/1/1        0
define/local effarea/r/1/1      0.
define/local obsmod/c/1/3       " "           all
define/local setup/c/1/20       " "           all
define/local cat_accept/c/1/20  "accept.cat"
define/local cat_reject/c/1/20  "reject.cat"
define/local cat_result/c/1/20  "result.cat"
define/local cat_entry/c/1/512  " "           all



! Connect to pipeline
log/rec "{p4}_{recID}" IN
mess/out 3 {procID} -
    "Recipe {p4}_{recID}: Atmospheric extintion coefficient computation."


! Process parameters
obsmod = m$upper(p4)
setup = "{obsmod}pipe"

if m$tstno(p5) .eq. 1 then
  effarea = {p5}
else
  mess/out 1 {procID} "Invalid telescope effective area!"
  clean/pipeline xmos,spec
  return 1
endif

if m$tstno(p6) .eq. 1 then
  nbins = {p6}
else
  mess/out 1 {procID} "Invalid number of wavelength bins given!"
  clean/pipeline xmos,spec
  return 1
endif


! Startup context, initialize environment and prepare input 
! frames for processing
@p fors_paragraph 1

initialize/pipeline {p1} {p2} {cat_accept} xmos {setup} {obsmod} -
                    {cat_reject} {cat_result}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Pipeline initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Initialize product header
init/prohead {p1} {p2} 2,"Atmospheric extintion coefficient computation" -
   obsmode,{p4} area,{p5} nbins,{P6}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Product header initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Initialize longlog
init/longlog {p1} {p2} 2 "obsmode: {p4}" "area: {p5}" "nbins: {P6}"
if {q1} .ne. 0 then
  mess/out 1 {procID} "Longlog initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


@p fors_paragraph 2


! Update session keywords
read_cat_entry:

store/frame cat_entry {cat_accept} 1 exit_loop
fors_id = m$value({cat_entry},{FORS_IDSTRNG})
fors_id = m$upper(fors_id) 

if fors_id .eq. ASTD_ID then
  ctrl(1) = ctrl(1) + 1
  if ctrl(1) .eq. 1 then
    create/icat {stdcat} NULL {FORS_IDSTRNG} >Null
  endif
  add/icat {stdcat} {cat_entry} >Null
elseif fors_id .eq. RESP_ID then
  resptab = cat_entry
  ctrl(2) = 1
elseif fors_id .eq. STDFLX_ID then
  refflux = cat_entry
  ctrl(3) = 1
endif

goto read_cat_entry


exit_loop:

if ctrl(1) .eq. 0 then
  mess/out 1 {procID} "No standard star spectrum tables found!"
  clean/pipeline xmos,spec
  return 1
else if ctrl(1) .lt. min_nstd then
  mess/out 1 {procID} "Number of standard star tables found is {ctrl(1)}!"
  mess/out 1 {procID} -
     "Required minimum number of standard star tables is {min_nstd}!"
  clean/pipeline xmos,spec
  return 1
else if ctrl(2) .eq. 0 then
  mess/out 1 {procID} "No instrument response table present!"
  clean/pipeline xmos,spec
  return 1
else if ctrl(3) .eq. 0 then
  mess/out 1 {procID} "No reference flux table present!"
  clean/pipeline xmos,spec
  return 1
endif


! Check wavelength ranges of all spectra
verify/ranges {stdcat} 0. 0.
if {q1} .ne. 0 .or. outputi(1) .lt. min_nstd then
  mess/out 1 {procID} "Invalid set of standard star spectra encountered!"
  clean/pipeline xmos,spec
  return 1
else
  rebstp = (rebend - rebstrt) / nbins
endif


! Prepare standard star spectra for fitting the extinction coefficients
prepare/extcoeff {stdcat} {extcat} {refflux} {resptab} {effarea} {rspftyp} -
       {rspfpar} {rebstrt},{rebend},{rebstp}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Cannot prepare standard star spectra for fitting!"
  clean/pipeline xmos,spec
  return 1
endif


! Fit atmospheric extinction coefficients
fit/extinction {extcat} {exttbl}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Fitting of extinction coefficients failed!"
  clean/pipeline xmos,spec
  return 1
endif

@p fors_paragraph 3


! Prepare output catalog
classify/product {exttbl} {ATMEXT_ID} {d_pre} {d_geo},{d_det} OK
write/descriptor {exttbl} {h_instrume}/c/1/60 {#1_{stdcat},{h_instrume}}
write/desc {exttbl} {FORS_IDSTRNG}/c/1/20 {ATMEXT_ID}

! Remove template descriptor H_TPEXPNO
delete/desc {exttbl} {h_tpexpno} no

add/icat {cat_result} {exttbl} >Null


! Update internal log file
shortlog/out 1 "Atmospheric extinction coefficients determined."


! Create product header
finish/prohead {cat_result} 2


! Shutdown pipeline
finish/pipeline {cat_result} {p3} xmos,spec {exttbl}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Error leaving pipeline environment!"
  finish/longlog {cat_accept} -1 ? 1 {p3}
  clean/pipeline xmos,spec
  return 1
endif


! Close longlog
finish/longlog {cat_accept} {cat_result} {exttbl} 2 {p3}


! Disconnect from pipeline
log/rec "{p4}_{recID}" OUT

@p fors_paragraph 4

return 0
