! $Id: fors_mos_cal_normflat.prg,v 1.5 2000/09/12 07:16:14 sbogun Exp $   
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1998  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_mos_cal_normflat.prg,v $
!
!.KEYWORDS    MOS flat field normalization.
!
!.USAGE       @p fors_mos_cal_normflat  frm_in frm_ref frm_out mode method
!                                       mtdpar detpar 
!
!.INPUT      frm_in       (P1)    Set of input frames containing a single
!                                 spectroscopic master screen flat field.
!            frm_ref      (P2)    Set of reference frames (bias, dark ...)
!            mode         (P4)    Indicator of observing mode (LSS, MOS)
!            method       (P5)	  Flat field normalization method.
!            mtdpar       (P6)	  Parameter for normalization method. The
!                                 values, separated by a comma, correspond
!                                 to the smoothing along the x and y axis
!                                 respectively.
!            detpar       (P7)    Optional bias, dark constants and bias scaling
!                                 method ("MED", "OVE" or "NO").
!            fftype       (P8)    Current flat field type (UNPROCESSED,
!                                 REDUCED, MASTER)
!
!.OUTPUT     frm_out      (P3)	  Set of output frames (normalized flat field)
!
!.RETURNS    The recipe returns 0 on success, otherwise 1. Keyword Q1 is
!            updated with the exit status. 
!
!.PURPOSE    The recipe normalizes the incoming flat field frame by dividing
!            it by a smoothed version of itself.
!
!.ALGORITHM  The flat field is corrected for bias and dark if certified frames
!            of these types are found in the reference set of frames. If no
!            such frames are present, the bias and dark correction may be
!            enforced by non zero values for the bias and/or dark constants.
!            Generally it is assumed that the provided flat field frame
!            is already a certified version so that the above corrections are
!            unnecessary.
!            The flat field is then smoothed along the slit and the dispersion
!            axis using the specified method (cf. documentation of MOS command
!            NORMALIZE/FLAT). The flat field is then divided by the smoothed
!            version to obtain a normalized flat field frame.
!
!            If the observing mode is set to MOS the averaging, fitting and
!            the normalization are done for each individual slitlet. The
!            slitlet positions are detected on the fly.
!   
!.ENVIRON     FORS context
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1998/04/01:  Creation
!             1.3  1999/09/03:  Product header generation implemented (SB).
!
!------------------------------------------------------------------------------

define/param p1 frm_in.set   c    "Enter input frame set:"
define/param p2 frm_ref.set  c    "Enter reference frame set:"
define/param p3 frm_out      c    "Enter output frame set:"
define/param p4 MOS          c    "Enter observing mode:"
define/param p5 +            c    "Enter fit method:"
define/param p6 +            c    "Enter degree of fit:"
define/param p7 0.,0.,med    c    "Enter bias/dark const. and scaling method:"
define/param p8 MASTER       c    "Enter flat field frame type:"


! Procedure identification
define/local procID/c/1/80	"{p4}CalNormFlat"

! Recipe identification
define/local recID/c/1/12	"CAL_NORMFLAT"


! Local constants
define/local dwin/i/1/3         1,2


! Local variables
define/local nopt/i/1/1         0
define/local pos/i/1/1          0
define/local ctrl/i/1/4         0             all
define/local catal/i/1/1        0
define/local ffcode/i/1/1       0
define/local detpar/r/1/2       0.            all
define/local bscale/c/1/3       " "
define/local obsmod/c/1/3       " "           all
define/local setup/c/1/20       " "           all
define/local method/c/1/20      {p5}
define/local mtdpar/c/1/20      {p6}
define/local fftype/c/1/11      " "           all
define/local cat_accept/c/1/20  "accept.cat"
define/local cat_reject/c/1/20  "reject.cat"
define/local cat_result/c/1/20  "result.cat"
define/local cat_entry/c/1/512  " "           all

! Unfortunately needed for FINISH/LONGLOG
define/local normflat_name/c/1/32 " "



! Connect to pipeline
log/rec "{p4}_{recID}" IN
mess/out 3 {procID} "Recipe {p4}_{recID}: Flat field normalization"


! Process parameters
obsmod = m$upper(p4)
setup = "{obsmod}pipe"

nopt = m$parse(p7,"opt")
if nopt .gt. 3 .or. nopt .lt. 1 then
  mess/out 1 {procID} "Invalid detector parameter options given!"
  clean/pipeline xmos,spec
  return 1
else
  if m$tstno(opt01) .eq. 1 then
    detpar(1) = {opt01}
  else
    mess/out 1 {procID} "Invalid bias constant given!"
    clean/pipeline xmos,spec
    return 1
  endif

  if nopt .gt. 1 then
    if m$tstno(opt02) .eq. 1 then
      detpar(2) = {opt02}
    else
      mess/out 1 {procID} "Invalid dark constant given!"
      clean/pipeline xmos,spec
      return 1
    endif

    if nopt .gt. 2 then
      bscale = m$upper(opt03(1:3))
      if bscale .ne. "MED" .and. bscale .ne. "OVE" then
        if bscale .ne. "NO" then
          mess/out 1 {procID} "Invalid bias correction option given!"
          clean/pipeline xmos,spec
          return 1
        endif
      endif
    else
      bscale = "NO"
    endif
  else
    detpar(2) = 0.
    bscale = "NO"
  endif
endif

fftype = m$upper(p8)
if fftype(1:1) .eq. "U" then
  fftype = "unprocessed"
  ffcode = 0
elseif fftype(1:1) .eq. "R" then
  fftype = "reduced"
  ffcode = 1
else
  fftype = "master"
  ffcode = 2
endif


! Startup context, initialize environment and prepare input 
! frames for processing
@p fors_paragraph 1

initialize/pipeline {p1} {p2} {cat_accept} xmos {setup} {obsmod} -
    {cat_reject} {cat_result}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Pipeline initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif
normflat_name = "{normflat}"


! Initialize product header
init/prohead {p1} {p2} 2,"Flat field normalization" -
  "obsmode,{p4}" "method,{p5}" "metpar,{P6}" "detpar,{p7}" "fftype,{p8}"
if {q1} .ne. 0 then
  mess/out 1 {procID} "Product header initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Initialize longlog
init/longlog {p1} {p2} 2 "obsmode: {p4}" "method: {p5}" "metpar: {P6}" -
    "detpar: {p7}" "fftype: {p8}"
if {q1} .ne. 0 then
  mess/out 1 {procID} "Longlog initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Clear display and graphic windows, if present
@p fors_niceclear


! Update context specific parameters given on the command line
if method(1:1) .ne. "+" then
  method = m$upper(method)
  set/mos normmet={method}
endif

if m$tstno(p6) .eq. 1 then
  pos = m$index(mtdpar,",") - 1
  if pos .le. 0 then
    mess/out 1 {procID} "Invalid smoothing parameters!"
  else
    ffwin = {mtdpar(1:{pos})}
    pos = pos + 2
    fford = {mtdpar({pos}:)}
  endif
endif


@p fors_paragraph 2

! Update session keywords BIAS, DARK and FLAT
read_cat_entry:

store/frame cat_entry {cat_accept} 1 exit_loop
fors_id = m$value({cat_entry},{FORS_IDSTRNG})
fors_id = m$upper(fors_id) 

if fors_id .eq. MBIAS_ID then
  bias = cat_entry
  ctrl(1) = 1
elseif fors_id .eq. MDARK_ID then
  dark = cat_entry
  ctrl(2) = 1
elseif fors_id .eq. MFLAT_ID .and. ffcode .eq. 2 then
  flat = cat_entry
  ctrl(3) = 1
elseif fors_id .eq. RFLAT_ID .and. ffcode .eq. 1 then
  flat = cat_entry
  ctrl(3) = 1
elseif fors_id .eq. FLAT_ID .and. ffcode .eq. 0 then
  flat = cat_entry
  ctrl(3) = 1
elseif fors_id .eq. SLITS_ID then
  mos = cat_entry
  ctrl(4) = 1
endif

goto read_cat_entry


exit_loop:

if ctrl(3) .eq. 0 then
  mess/out 1 {procID} "No {fftype} flat field frame present!"
  clean/pipeline xmos,spec
  return 1
endif

if obsmod .eq. "LSS" .and. ctrl(4) .eq. 0 then
  mess/out 1 {procID} "No slit position table present!"
  clean/pipeline xmos,spec
  return 1
endif


! Check if a basic reduction of the flat field is required and check
! if constants for bias and dark should be used.
if ffcode .eq. 0 then
  if ctrl(1) .eq. 0 .and. detpar(1) .gt. 0. then
    bias = "{detpar(1)}"
    ctrl(1) = 1
  endif

  if ctrl(2) .eq. 0 .and. detpar(2) .gt. 0. then
    dark = "{detpar(2)}"
    ctrl(2) = 1
  endif
else
  if ctrl(1) .ne. 0 .or. ctrl(2) .ne. 0 then
    mess/out 4 {procID} -
        "Flat field frame already corrected for bias and/or dark."
    mess/out 4 {procID} "Bias and dark correction disabled!"
  endif
  ctrl(1) = 0
  ctrl(2) = 0
endif


! Get lamp setup from flat field header keywords
config/lamps {flat}


! Compute overscan bias for raw frame
if ctrl(1) .ne. 0 then
  compute/ovbias {flat} median
  if bscale .ne. "NO" then
    if {q1} .eq. -1 then
      mess/out 2 {procID} -
        "No overscan in input frame. Appylying standard bias correction."
      bscale = "NO"
    else if {q1} .ne. 0 then
      mess/out 1 {procID} "Bias computation from overscan areas failed!"
      clean/pipeline xmos,spec
      return 1
    endif
  endif
endif


! Remove pre- and overscan from master flat field
trim/frame {flat}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Cannot remove pre- and overscan from {flat}!"
  clean/pipeline xmos,spec
  return 1
endif


! Basic reduction steps
if ctrl(1) .ne. 0 then
  @p fors_corr_bias {flat} {bias} + + {bscale}
  if {q1} .ne. 0 then
    mess/out 1 {procID} "Bias correction failed!"
    clean/pipeline xmos,spec
    return 1
  endif
  mess/out 3 {procID} "Bias correction applied!"
endif

if ctrl(2) .ne. 0 then
  @p fors_corr_dark {flat} {dark}
  if {q1} .ne. 0 then
    mess/out 1 {procID} "Dark correction failed!"
    clean/pipeline xmos,spec
    return 1
  endif
  mess/out 3 {procID} "Dark correction applied!"
endif


! Display unprocessed master flat field
if nice_dids({dwin(1)}) .eq. 1 then
  view/spectrum {flat} {dwin(1)} f,min real avr "RAW FLAT FIELD"
endif


! Setup slit(let) position(s)
define/position {flat} {mos} {obsmod}
if {q1} .ne. 0 then
  mess/out 1 {procID} "{obsmod} slit setup failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Normalize flat field for each individual slit(let)
pos = m$index(normmet,",") - 1

set/format i1
if normmet(1:1) .eq. "A" then  
  mess/out 3 {procID} "Averaging {ffwin} image rows along the slit."
  longlog/out 1 "     Averaged {ffwin} image rows along the slit."
else
  mess/out 3 {procID} "Smoothing image along slit."
  mess/out 3 {procID} "Filter {normmet(1:{pos})}. Filter size: {ffwin} pixel."
  longlog/out 1 "     Smoothed image along slit."
  longlog/out 1 "     Filter {normmet(1:{pos})}. Filter size: {ffwin} pixel."
endif

longlog/out 1 " "

pos = pos + 2
if normmet({pos}:{pos}) .eq. "P" then  
  mess/out 3 {procID} "Fitting image along dispersion axis."
  mess/out 3 {procID} "Order of polynomial fit: {fford}."
  longlog/out 1 "     Fitted image along dispersion axis."
  longlog/out 1 "     Order of polynomial fit: {fford}."
else
  mess/out 3 {procID} "Smoothing image along dispersion axis."
  mess/out 3 {procID} "Filter {normmet({pos}:)}. Filter size: {fford} pixel."
  longlog/out 1 "     Smoothed image along dispersion axis."
  longlog/out 1 "     Filter {normmet({pos}:)}. Filter size: {fford} pixel."
endif
set/format

! Table has to be sorted by column :YSTART otherwise NORMALIZE/FLAT fails.
sort/table {mos} :{c_slystart}
normalize/flat {flat} {mos} {normflat} {normmet} {ffwin} {fford}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Flat field normalization failed!"
  clean/pipeline xmos,spec
  return 1
endif
sort/table {mos} :{c_slsno}

! Copy all but the standard descriptors to the new frame and prepare
! output catalog
copy/dd {flat} *,3 {normflat}


! Update statistic keywords
statistics/image {normflat} option=GN
write/desc {normflat} {h_promean}/d/1/1 {outputr(3)}
write/desc {normflat} {h_promedi}/d/1/1 {outputr(8)}
write/desc {normflat} {h_prosigm}/d/1/1 {outputr(4)}

@p fors_paragraph 3

if ffcode .le. 1 then
  classify/product {normflat} {NFLAT_ID} {d_tmp} {d_empty} OK ? y
  write/desc {normflat} {FORS_IDSTRNG}/c/1/20 {NFLAT_ID}
else
  classify/product {normflat} {MNFLAT_ID} {d_pre} {d_empty} OK ? y
  write/desc {normflat} {FORS_IDSTRNG}/c/1/20 {MNFLAT_ID}
endif
add/icat {cat_result} {normflat} >Null


! Display normalized master flat field
if nice_dids({dwin(2)}) .eq. 1 then
  if m$strlen("{{normflat},IDENT}") .eq. 0 then
    copy/dd {normflat} {h_procatg} {normflat} IDENT
  endif
  view/spectrum {normflat} {dwin(2)} f,2sigma real avr -
      "NORMALIZED FLAT FIELD"
endif


! Update internal log file
shortlog/out 1 "Flatfield normalized."


longlog/out 3 " "
longlog/out 3 "     Normalized flatfield is: {normflat}"


! Create product header
finish/prohead {cat_result} 2


! Shutdown pipeline
finish/pipeline {cat_result} {p3} xmos,spec {normflat}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Error leaving pipeline environment!"
  finish/longlog {cat_accept} -1 ? 1 {p3}
  clean/pipeline xmos,spec
  return 1
endif


! Close longlog
finish/longlog {cat_accept} {cat_result} {normflat_name} 2 {p3}


! Disconnect from pipeline
log/rec "{p4}_{recID}" OUT

@p fors_paragraph 4

return 0
