! $Id: fors_mos_cal_wave.prg,v 1.45 2001/03/13 09:42:09 sbogun Exp $
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1998  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_mos_cal_wave.prg,v $
!
!.KEYWORDS    MOS wavelength calibration
!
!.USAGE       @p fors_mos_cal_wave  frm_in frm_ref frm_out mode ccdpar
!
!.INPUT       frm_in      (P1)    Set of raw input frames to be processed.
!             frm_ref     (P2)    Set of reference frames
!             mode        (P4)    Indicator of observing mode (LSS, MOS)
!
!.OUTPUT      frm_out     (P3)	  Set of output frames (dispersion relation
!                                 coefficients)
!
!.RETURNS     The recipe returns 0 on success, otherwise 1. Keyword Q1 is
!             updated with the exit status. 
!
!.PURPOSE     The recipe computes the dispersion relation from a raw
!             wavelength calibration frame. The
!             background of the calibration frame is removed by fitting
!             a spline to local minima or by filtering the frame. The 
!             emission lines are searched and the dispersion coefficients
!             are determined from the emission line positions. Finally
!             the computed dispersion relation is consistent with the
!             linear approximation within adjustable limits.
!
!.ALGORITHM   The preprocesssed wavelength calibration frame is scanned for
!             emission lines. The pixel positions of the lines are stored in
!             an intermediate file. A first approximation of the dispersion
!             relation is determined using the optical properties of the
!             fors instrument. Using this approximation as initial guess
!             as many lines as possible are identified. Starting with these
!             identifications the final, two dimensional dispersion
!             relation is determined in an iterative process by fitting the
!             line positions and comparing the computed wavelengths to 
!             the laboratory wavelength of the lines under consideration.
!
!             If the observing mode is set to MOS the wavelength calibration
!             is done for each individual slitlet. The slitlet positions are
!             taken from the slit positions table.
!   
!.ENVIRON     FORS context
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1998/04/01:  Creation
!             0.9  1999/03/02:  Checks for computed dispersion coefficients
!                               implemented.
!             1.0  1999/03/09:  Automatic threshold selection implemented
!             1.3  1999/09/03:  Product header generation implemented (SB).
!             1.5  2000/08/01:  Checks of dispersion relation improved (NR).
!
!------------------------------------------------------------------------------

define/param p1 frm_in.set   c    "Enter input frame set:"
define/param p2 frm_ref.set  c    "Enter reference frame set:"
define/param p3 frm_out      c    "Enter output frame set:"
define/param p4 MOS          c    "Enter observing mode:"


! Procedure identification
define/local procID/c/1/80	"{p4}CalWave"


! Recipe identification
define/local recID/c/1/80	"CAL_WAVE"


! Local constants
define/local dwin/i/1/2         1,2
define/local gwin/i/1/2         1,2
define/local mu2mm/r/1/1        1.e-3
define/local nm2A/r/1/1         10.
define/local sfrm/c/1/20        "fors_linsrch.bdf"
define/local CAT_PLOT/c/1/40    "plot.cat"


! Local variables
define/local ctrl/i/1/8         0             all
define/local pos/i/1/1          0
define/local catal/i/1/1        0
define/local pxlsize/r/1/1      0.
define/local shift/r/1/1        0.
define/local sthresh/r/1/1      0.
define/local swindow/r/1/1      0.
define/local center/r/1/3       0.            all
define/local obsmod/c/1/3       " "           all
define/local setup/c/1/20       " "           all
define/local lmaster/c/1/100    " "           all
define/local fitord/c/1/100     " "           all
define/local centxy/c/1/100     " "           all
define/local cat_accept/c/1/20  "accept.cat"
define/local cat_reject/c/1/20  "reject.cat"
define/local cat_result/c/1/20  "result.cat"
define/local cat_entry/c/1/512  " "           all
define/local tol_pixs/r/1/1     0.
define/local tol_angs/r/1/1     0.
define/local loop/i/1/1         0
define/local chosenslit/i/1/1   0
define/local graphflag/c/1/3    " " all
define/local dico_qctbl/c/1/64  " " all
define/local slitchk/i/1/{FORS_MAXSLITLET} 0 all
define/local good/i/1/1 0
define/local bad/i/1/1 0
define/local closed/i/1/1 0
define/local longlog1/c/1/240 " " all
define/local longlog2/c/1/240 " " all
define/local longlog3/c/1/240 " " all
define/local sea_iter_cnt/i/1/1 0
define/local sea_iter_max/i/1/1 0
define/local seapar01/c/1/7 " " all +low  ! needed because of Midas error
define/local seapar02/c/1/7 " " all +low  ! needed because of Midas error

! Unfortunately needed for FINISH/LONGLOG
! (finish/pipeline before finish/longlog deletes contexts => "linfit" is gone)
define/local linfit_name/c/1/32 " "



! Connect to pipeline
log/rec "{p4}_{recID}" IN
mess/out 3 {procID} "Recipe {p4}_{recID}: Performs wavelength calibration"


! Process parameters
obsmod = m$upper(p4)
setup = "{obsmod}pipe"


! Startup context, initialize environment and prepare input 
! frames for processing
@p fors_paragraph 1

initialize/pipeline {p1} {p2} {cat_accept} xmos {setup} {obsmod} -
    {cat_reject} {cat_result}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Pipeline initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif
linfit_name = "{linfit}"


! Initialize plot file catalogue
initialize/plot {CAT_PLOT} 2


! Initialize product header
init/prohead {p1} {p2} 2,"Wavelength calibration" obsmode,{p4}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Product header initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Initialize longlog
init/longlog {p1} {p2} 2 "obsmode: {p4}"
if {q1} .ne. 0 then
  mess/out 1 {procID} "Longlog initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


!
@p fors_paragraph 2

! Update session keywords
read_cat_entry:

store/frame cat_entry {cat_accept} 1 exit_loop
fors_id = m$value({cat_entry},{FORS_IDSTRNG})
fors_id = m$upper(fors_id) 

if fors_id .eq. SLITS_ID then
  mos = cat_entry
  ctrl(4) = 1
elseif fors_id .eq. LINCAT_ID then
  lmaster = cat_entry
  ctrl(5) = 1
elseif fors_id .eq. GRISMS_ID then
  grispar = cat_entry
  ctrl(6) = 1
elseif fors_id .eq. FCAMERA_ID then
  fcamtbl = cat_entry
  ctrl(7) = 1
elseif fors_id .eq. WLC_ID then
  wlc = cat_entry
  ctrl(8) = 1
endif

goto read_cat_entry


exit_loop:

if ctrl(4) .eq. 0 then
  mess/out 1 {procID} "No slit position table present!"
  clean/pipeline xmos,spec
  return 1
else if ctrl(5) .eq. 0 then
  mess/out 1 {procID} "No line catalog present!"
  clean/pipeline xmos,spec
  return 1
else if ctrl(6) .eq. 0 then
  mess/out 1 {procID} "No grism data table present!"
  clean/pipeline xmos,spec
  return 1
else if ctrl(7) .eq. 0 then
  mess/out 1 {procID} "No camera focal length table present!"
  clean/pipeline xmos,spec
  return 1
else if ctrl(8) .eq. 0 then
  mess/out 1 {procID} "No wavelength calibration frame present!"
  clean/pipeline xmos,spec
  return 1
endif


! Setup slit(let) position(s)
mess/out 3 {procID} " "
define/position {wlc} {mos} {obsmod}
if {q1} .ne. 0 then
  mess/out 1 {procID} "{obsmod} slit setup failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Change frame coordinate system of incoming calibration frame to PIXEL.
swap/units {wlc} PIXEL
if {q1} .lt. 0 then
  mess/out 1 {procID} "Cannot establish PIXEL coordinate system in {wlc}!"
  clean/pipeline xmos,spec
  return 1
endif


! Remove pre- and overscan from wavelength calibration frame
trim/frame {wlc}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Cannot remove pre- and overscan from {wlc}!"
  clean/pipeline xmos,spec
  return 1
endif


! Select grism from parameter table
define/grism {wlc} {grispar}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Grism selection failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Convert detector pixel size from micro meter to milli meter
pxlsize = m$value({wlc},{h_xpixsize}) * mu2mm


! Remove background from calibration frame
mess/out 3 {procID} " "
mess/out 3 {procID} "Prepare emission line search."
prepare/search {wlc} {sfrm} {wlcsmmeth} {wlcsmpar(1)},{wlcsmpar(2)} -
               {wlcsmooth(1)} {wlcsmooth(3)},{wlcsmooth(4)} {mos}
if {q1} .eq. 0 then
  shift = outputr(1)
  mess/out 3 {procID} "Background emission removed from calibration frame."
  longlog/out 1 "     Background emission removed from calibration frame."
else
  mess/out 1 {procID} "Preparation of emission line search failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Adjust search threshold and search window to current exposure time
! and slit width
!
sea_iter_cnt = 0
!------------------------------------------------
search_iteration:
!
if sea_iter_cnt .gt. 0 then
  outputi = m$parse(SEAPAR,"seapar")
  set/format F5.2,
  SEAPAR = "{SEAITER({sea_iter_cnt})},{seapar02}"
  set/format
endif
!------------------------------------------------
!
adjust/search {sfrm} {seapar} {mos}
if {q1} .eq. 0 then
  sthresh = outputr(1)
  swindow = outputr(2)
else
 mess/out 1 {procID} -
     "Cannot adjust search parameters to exposure time and slit width!"
 clean/pipeline xmos,spec
 return 1
endif


! select appropriate emission line catalog using a representative slitwidth
mess/out 4 {procID} " "
define/linecat {wlc} {lmaster} {linecat} {h_moswid({slitidx})}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Preparation of line catalog failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Search for emission lines in wavelength calibration frame
pos = m$index(centmet," ") - 1

set/format i1 f1.2
mess/out 3 {procID} " "
mess/out 3 {procID} "Searching for emission lines in >{sfrm}<."
mess/out 3 {procID} "Using detection threshold:\t\t{sthresh} adu/s."
mess/out 3 {procID} "Search window size:\t\t\t{swindow} pxl."
mess/out 3 {procID} "Line centering method:\t\t{centmet(1:{pos})}"
mess/out 3 {procID} "Scan row step size and half width:\t{ybin} pxl"

longlog/out 1 " "
longlog/out 1 "     Searched for emission lines on calibration frame."
longlog/out 1 "     Used detection threshold:          {sthresh} adu/s."
longlog/out 1 "     Search window size:                {swindow} pxl."
longlog/out 1 "     Line centering method:             {centmet(1:{pos})}"
longlog/out 1 "     Scan row step size and half width: {ybin} pxl"
set/format

mess/out 4 {procID} " "
mess/out 4 {procID} "HINT1: IF YOU GET A RUN-TIME ERROR, YOU MOST"
mess/out 4 {procID} "       PROBABLY HAVE"
mess/out 4 {procID} "       - EITHER A WRONG BIAS VALUE SUBTRACTED"
mess/out 4 {procID} "       - OR A WRONG SLIT POSITION TABLE"
mess/out 4 {procID} "       - OR AN IMPROPER DETECTION THRESHOLD"
mess/out 4 {procID} "       - OR AN IMPROPER SEARCH WINDOW SIZE !"
if mess_level .le. 3 then
  search/mos {sfrm} {sthresh},{swindow} {ybin} {centmet} {mos} {linpos} >Null
else
  search/mos {sfrm} {sthresh},{swindow} {ybin} {centmet} {mos} {linpos}
endif
!--------------------------------------------------------------------------
if {q1} .eq. 1 then  ! DETECTION THRESHOLD TOO LOW - ITERATE
  sea_iter_cnt = sea_iter_cnt + 1
  info/key SEAITER
  sea_iter_max = mid$info(2)
  if sea_iter_cnt .le. sea_iter_max then
    set/format I1
    mess/out 2 {procID} " "
    mess/out 2 {procID} "Search for emission lines failed -"
    mess/out 2 {procID} "trying higher detection threshold"
    mess/out 2 {procID} "( Iteration {sea_iter_cnt} / max. {sea_iter_max} ):"
    mess/out 2 {procID} "- - - - - - - - - - - - - - - - -"
    goto search_iteration
  else
    mess/out 1 {procID} " "
    mess/out 1 {procID} "Search for emission lines finally failed!"
    return 1
  endif
endif
!--------------------------------------------------------------------------

! Check if there are enough lines for the choosen order of the
! fit polynomial.
mess/out 3 {procID} " "
check/nlines {sfrm} {wlcord(1)} {mos}
if {q1} .eq. 0 then
  if outputi(1) .ne. 0 then
    mess/out 1 {procID} -
      "Too few lines detected. Wavelength calibration failed!"
    clean/pipeline xmos,spec
    return 1
  endif
else
  mess/out 1 {procID} "Cannot check number of line detections!"
  clean/pipeline xmos,spec
  return 1
endif
mess/out 3 {procID} " "


if nice_dids({dwin(1)}) .eq. 1 then
  @p fors_niceclear
  view/search {wlc} {linpos},{mos} {dwin(1)} mean,2 real avr
  if {q1} .ne. 0 then
    mess/out 2 {procID} "Problems showing detected emission lines!"
  endif
endif


! Get pixel position of optical axis on the CCD in detector coordinates
! REFPIX is the reference pixel in frame coordinates as in extract/frame.
if m$existd(wlc,"REFPIX") .eq. 1 then
  center(1) = (m$value({wlc},refpix(1)) - 1) * m$value({wlc},step(1)) + -
               m$value({wlc},start(1))
  center(2) = (m$value({wlc},refpix(2)) - 1) * m$value({wlc},step(2)) + -
               m$value({wlc},start(2))
  center(3) = 0.5 * m$value({wlc},npix(1)) * m$value({wlc},step(1))

  if m$index("{{wlc},cunit(17:32)}","PIXEL") .eq. 1 .and. -
     m$index("{{wlc},{h_instrume}}","FORS") .gt. 0 then
    if m$value({wlc},refpix(1)) .lt. 2.0 .and. -
       m$value({wlc},refpix(2)) .lt. 2.0 then
      mess/out 2 {procID} "Reference pixel keywords seem to be incorrect."
      mess/out 2 {procID} "Trying temporary FORS workaround!"
      @p fors_getrefpix {wlc}
      if {q1} .ne. 0 then
        mess/out 1 {procID} "Keywords CRPIXi missing in header comment section!"
        mess/out 1 {procID} "FORS workaround failed!"
        mess/out 1 {procID} "Cannot perform wavelength calibration!"
        clean/pipeline xmos,spec
        return 1
      else
        center(1) = outputr(1)
        center(2) = outputr(2)
        mess/out 2 {procID} "FORS workaround was successful!"
      endif
    endif
  endif
else
  mess/out 1 {procID} "Reference pixel not found in {wlc}!"
  clean/pipeline xmos,spec
  return 1
endif


! Do the calibration
fitord = "{wlcord(1)},{wlcord(2)},{wlcord(3)}"
centxy = "{center(1)},{center(2)},{center(3)}"

pos = m$index(poltyp," ") - 1

! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
! TOL is now used as tolerance factor for the FWHM;
! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
! tolerance in pixel is:
!
tol_pixs = TOL *5.0 *{{wlc},{h_moswid({slitidx})}}/1. *0.024/pxlsize
!
! - The factor 5.0px is nearly a constant for the result of the following
!   equation for each grism (FWHM):
!   5.0px = mean_of_wavelength_range / spectral_resolution / linear_dispersion
! - This factor was calculated for a 1'' slit - for the other slit widths
!   a linear relation is applied with the term
!   "{{wlc},{h_moswid({slitidx})}}/1.''"
! - This factor was calculated for a CCD with a pixel size of 0.024mm -
!   for a future CCD with a different pixel size a linear relation is applied
!   with the term "0.024mm/pxlsize" ("pxlsize" is converted to mm above)
!
! tolerance in angstrom is:
!
tol_angs = {{wlc},{h_avdisp}}*nm2A*pxlsize *tol_pixs
!
! - "{{wlc},{h_avdisp}}" is given in nm/mm
! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

set/format i1 f7.2
mess/out 3 {procID} " "
mess/out 3 {procID} "Calibrating grism #{grism} ({{wlc},{h_grism1}})"
mess/out 3 {procID} "Using CCD reference pixel:\t\t{center(1)},{center(2)}"
mess/out 3 {procID} "Line identification parameter:\t{alpha}"
mess/out 3 {procID} "Type of fit polynom:\t\t{poltyp(1:{pos})}"
mess/out 3 {procID} "Selected order of fit (x,y):\t{wlcord(1)},{wlcord(2)}"
mess/out 3 {procID} "Fit tolerance value [px]:\t\t{tol_pixs}"
mess/out 3 {procID} "Fit tolerance value [A]:\t\t{tol_angs}"
longlog/out 1 " "
longlog/out 1 "     Calibrated grism #{grism} ({{wlc},{h_grism1}})"
longlog/out 1 "     Used CCD reference pixel:      {center(1)},{center(2)}"
longlog/out 1 "     Line identification parameter: {alpha}"
longlog/out 1 "     Type of fit polynom:           {poltyp(1:{pos})}"
longlog/out 1 "     Selected order of fit (x,y):   {wlcord(1)},{wlcord(2)}"
longlog/out 1 "     Fit tolerance value [px]:      {tol_pixs}"
longlog/out 1 "     Fit tolerance value [A]:       {tol_angs}"
set/format

mess/out 4 {procID} " "
mess/out 4 {procID} "HINT2: - IF YOU GET A RUN-TIME ERROR, YOU VERY PRO-"
mess/out 4 {procID} "         BABLY HAVE AN IMPROPER MASTER LINE CATAL.!"
mess/out 4 {procID} "       - IF TOO FEW LINES ARE DETECTED, THE DETECT."
mess/out 4 {procID} "         THRESHOLD FOR THE 1.ITER. CAN BE TOO LOW !"

if mess_level .le. 3 then
  calibrate/grism {tol_angs},{tolguess} {fitord} {grism} -
       -9999.,-9999.,{pxlsize},1 {centxy} {grispar} {fcamtbl} 0 >Null
else
  calibrate/grism {tol_angs},{tolguess} {fitord} {grism} -
       -9999.,-9999.,{pxlsize},1 {centxy} {grispar} {fcamtbl} 1
endif

if nice_dids({dwin(2)}) .eq. 1 then
  view/ident {wlc} {linpos},{mos} final {dwin(2)} mean,2 real,avr
  if {q1} .ne. 0 then
    mess/out 2 {procID} "Problems showing identified emission lines!"
  endif
endif

! Calculate median of the number of active slits:
create/image midslit.bdf 1,{FORS_MAXSLITLET} 1.,1. poly 0.,0.,0.
do loop = 1 FORS_MAXSLITLET
  select/table {mos} :Slit.eq.{loop} >Null  ! NOT {linpos} instead of {mos} -
  if outputi .gt. 0 then                    ! failed slits not deleted there !
    midslit.bdf[@{loop}] = {loop}
  endif
enddo
inputr = FORS_MAXSLITLET+1
statistics/image midslit.bdf + ? 1.,{inputr} xn >Null
chosenslit = outputr(8)  ! median slit
-delete midslit.bdf

set/format I1
if nice_gids({gwin(1)}) .eq. 1 then
  mess/out 4 {procID} -
    "  plotting ident. lines for median(#slits)={chosenslit}.slit"
  plot/lident {wlc} {linpos} {mos} {chosenslit} {ybin} {gwin(1)}
endif
if nice_gids({gwin(2)}) .eq. 1 then
  mess/out 4 {procID} -
    "  plotting  residuals   for median(#slits)={chosenslit}.slit"
  view/residuals {linpos} {chosenslit} {gwin(2)}
endif

! Check dispersion relation:
mess/out 3 {procID} " "
mess/out 3 {procID} "Checking computed dispersion relation"
!
check/dispersion {wlc} {linfit} {mos} {tolcoef(1)},{tolcoef(2)} {chosenslit}
copy/keyword outputi slitchk
!
if {q1} .eq. 0 then
  if {q2} .ne. 0 then
    write/out " "
    mess/out 1 {procID} "Wavelength calibra. failed (for all active slits) !"
    clean/pipeline xmos,spec
    return 1
  else
    !
    mess/out 3 {procID} " "
    mess/out 3 {procID} "<S>lits / <R>esults of dispersion checks"
    mess/out 3 {procID} "          ("0"=good "1"=bad "2"=unused):"
    mess/out 3 {procID} "---------------------------------------------------"
    set/format I1
    append/out "<S>"
    do inputi = 1 FORS_MAXSLITLET
      append/out " {inputi}"
    enddo
    copy/keyword cbuf longlog1
    mess/out 3 {procID}
    append/out "<R>"
    do inputi = 1 FORS_MAXSLITLET
      if inputi .ge. 10 set/format J2
      append/out " {slitchk({inputi})}"
    enddo
    copy/keyword cbuf longlog2
    mess/out 3 {procID}
    mess/out 3 {procID} "---------------------------------------------------"
    set/format I1
    do inputi = 1 FORS_MAXSLITLET
      if slitchk({inputi}) .eq. 0 good = good+1
      if slitchk({inputi}) .eq. 1 bad = bad+1
      if slitchk({inputi}) .eq. 2 closed = closed+1
    enddo
    append/out -
  "#Slits: {FORS_MAXSLITLET} / #good: {good} , #bad: {bad} , #unused: {closed}"
    copy/keyword cbuf longlog3
    mess/out 3 {procID}
    !
    longlog/out 1 " "
    longlog/out 1 "     Checked computed dispersion relation:"
    longlog/out 1 "     <S>lits / <R>esults of dispersion checks"
    longlog/out 1 "               ("0"=good "1"=bad "2"=unused):"
    longlog/out 1 "     ---------------------------------------------------"
    longlog/out 1 "     {longlog1}"
    longlog/out 1 "     {longlog2}"
    longlog/out 1 "     ---------------------------------------------------"
    longlog/out 1 "     {longlog3}"
    !
  endif
else
  write/out " "
  mess/out 1 {procID} "Check of dispersion relation failed !"
  clean/pipeline xmos,spec
  return 1
endif

! Compute high order differences (non-linearities):
set/format I2
mess/out 3 {procID} " "
mess/out 3 {procID} "Computing high order differences of slit #{chosenslit}"
set/format
!
if nice_crea .eq. "y" then
  graphflag = "2,w"
else
  graphflag = "n"
endif
view/delta {wlc} {chosenslit} {linpos} {mos} {linfit} {graphflag}
if {q1} .eq. 1 then
  mess/out 1 {procID} "Computation of high order differences failed !"
  clean/pipeline xmos,spec
  return 1
endif
dico_qctbl = q2
if q3(1:1) .ne. " " then
  if m$exist(q3) .ne. 0 then
    add/acat {CAT_PLOT} {q3} >Null
  endif
endif

! Remove bad slits (ONLY noticed by check/dispersion - NOT by view/delta)
! in tables with slit positions / line positions / dispersion coefficients:
mess/out 4 {procID} "  removing bad slits in tables with:"
set/format I1  ! important for 'delete/desriptor' below

mess/out 4 {procID} "    slit positions  : {mos}"
select/table {mos} seq.eq.0 >Null
copy/table {mos} tmpall.tbl
do loop = 1 FORS_MAXSLITLET
  if slitchk({loop}) .eq. 0 then  ! 0="good slit"
    select/table {mos} :SLIT.eq.{loop} >Null
    copy/table {mos} tmp{loop}.tbl
    merge/table tmpall.tbl tmp{loop}.tbl tmptmp.tbl
    -rename tmptmp.tbl tmpall.tbl
  endif
enddo
-rename tmpall.tbl {mos}
-delete tmp*.tbl

mess/out 4 {procID} "    identified lines: {linpos}"
select/table {linpos} seq.eq.0 >Null
copy/table {linpos} tmpall.tbl
do loop = 1 FORS_MAXSLITLET
  if slitchk({loop}) .eq. 0 then  ! 0="good slit"
    select/table {linpos} :Slit.eq.{loop} >Null
    copy/table {linpos} tmp{loop}.tbl
    merge/table tmpall.tbl tmp{loop}.tbl tmptmp.tbl
    -rename tmptmp.tbl tmpall.tbl
  endif
enddo
-rename tmpall.tbl {linpos}
-delete tmp*.tbl

mess/out 4 {procID} "    dispersion coef.: {linfit}"
select/table {linfit} seq.eq.0 >Null
copy/table {linfit} tmpall.tbl
do loop = 1 FORS_MAXSLITLET
  if slitchk({loop}) .eq. 0 then  ! 0="good slit"
    select/table {linfit} :SLIT.eq.{loop} >Null
    copy/table {linfit} tmp{loop}.tbl
    merge/table tmpall.tbl tmp{loop}.tbl tmptmp.tbl
    -rename tmptmp.tbl tmpall.tbl
  endif
enddo
-rename tmpall.tbl {linfit}
-delete tmp*.tbl
do loop = 1 FORS_MAXSLITLET
  if slitchk({loop}) .eq. 1 then  ! 1="bad slit"
    delete/descriptor {linfit} COEFSLIT{loop}
  endif
enddo


! Copy all but the standard descriptors to the new frame
copy/dd {wlc} *,5 {linfit}
copy/dd {wlc} *,5 {linpos}
copy/dd {wlc} *,5 {dico_qctbl}

@p fors_paragraph 3


! Prepare output catalog
classify/product {linfit} {DISPREL_ID} {d_pre} {d_det} OK
classify/product {linpos} {LINPOS_ID} {d_tmp} {d_det} OK
classify/product {dico_qctbl} {DISPQCT_ID} {d_qcp} {d_det} OK

write/desc {linfit} {FORS_IDSTRNG}/c/1/60 {DISPREL_ID}
write/desc {linpos} {FORS_IDSTRNG}/c/1/60 {LINPOS_ID}
write/desc {dico_qctbl} {FORS_IDSTRNG}/c/1/60 {DISPQCT_ID}

write/desc {dico_qctbl} {h_proqcsts}/c/1/9 "CHECKED"
write/dhelp {dico_qctbl} {h_proqcsts} "{i_proqcsts}"
longlog/out 1 " "
longlog/out 1 "     Table with quality of dispersion solution: {dico_qctbl}"

add/icat {cat_result} {linfit} >Null
add/icat {cat_result} {linpos} >Null
add/icat {cat_result} {dico_qctbl} >Null


! Update internal log file
shortlog/out 1 "Wavelength calibration performed."


longlog/out 3 " "
longlog/out 3 "     Table containing dispersion coeffients is: {linfit}"


! Create product header
finish/prohead {cat_result} 2


! Shutdown pipeline
finish/pipeline {cat_result} {p3} xmos,spec {linfit} {CAT_PLOT}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Error leaving pipeline environment!"
  finish/longlog {cat_accept} -1 ? 1 {p3}
  clean/pipeline xmos,spec
  return 1
endif


! Close longlog
finish/longlog {cat_accept} {cat_result} {linfit_name} 2 {p3}


! Disconnect from pipeline
log/rec "{p4}_{recID}" OUT

@p fors_paragraph 4

return 0
