! $Id: fors_mos_obs_reduce.prg,v 1.16 2001/03/21 17:38:39 sbogun Exp $   
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1998  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_mos_obs_reduce.prg,v $
!
!.KEYWORDS    MOS science frame reduction and rebinning
!
!.USAGE       @p fors_mos_obs_reduce  frm_in frm_ref frm_out mode
!                                     ccdpar calflg
!
!.INPUT       frm_in      (P1)    Science frame to be reduced. 
!             frm_ref     (P2)    Set of reference frames (bias, dark,
!                                 normalized flat field, dispersion coeff.,
!                                 atm. extinction coeff., instrument
!                                 response).  
!             mode        (P4)    Indicator of observing mode (LSS, MOS).
!             ccdpar      (P5)    Optional bias, dark constants and bias scaling
!                                 method ("MED", "OVE" or "NO").
!             calflg      (P6)    Controls if flux calibration is done.
!
!.OUTPUT      frm_out     (P3)	  Set of output frames (reduced and
!                                 rebinned and flux calibrated object
!                                 spectrum (2-dimensional)).
!
!.RETURNS     The recipe returns 0 on success, otherwise 1. Keyword Q1 is
!             updated with the exit status. 
!
!.PURPOSE     The reduction recipe performs a standard reduction of a raw
!             spectroscopic science frame. This includes the following
!             reduction steps:  
!
!                - Bias correction         (optional)
!                - Dark correction         (optional)
!
!                - Flat-field correction   (required)
!                - Resampling of the data
!
!                - Extinction correction   (optional)
!                - Flux calibration        (optional)
!
!                - Detection of (point source) spectra
!                - Sky background fitting
!                - Optimal extraction of the detected spectra
!
!             The procedure applies only the above corrections, assuming that
!             all calibration frames are certified master versions. No
!             combination and/or cleaning of individual images is done.
!             
!             After rebinning the science frame to constant wavelength steps
!             the frame units along the dispersion axis are Angstrom. Along
!             the slit, the pixel coordinates are converted to arcsec. As
!             origin the reference pixel is used, that corresponds in
!             principle the telescope pointing direction (given in the
!             header). This should be sufficient for the object spectrum
!             identification by the observer.
!
!             The extracted spectra are stored in a 3d table.
!
!.ENVIRON     FORS context, MOS context
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1998/04/01:  Creation
!             0.6  1999/04/28:  Longlog implemented (SB)
!             1.3  1999/09/03:  Product header generation implemented (SB).
!
!------------------------------------------------------------------------------

define/param p1 frm_in.set   c     "Enter input frame set:"
define/param p2 frm_ref.set  c     "Enter reference frame set:"
define/param p3 frm_out.set  c     "Enter output frame set:"
define/param p4 MOS          c     "Enter observing mode:"
define/param p5 0.,0.,med    c     "Enter bias/dark const. and scaling method:"
define/param p6 No           c     "Do flux calibration?"


! Procedure identification
define/local procID/c/1/80	"{p4}ObsReduce"


! Recipe identification
define/local recID/c/1/10       "OBS_REDUCE"


! Local constants
define/local dwin/i/1/3         1,2,3
define/local redobj/c/1/20      "fors_reduced.bdf"
define/local corobj/c/1/20      "fors_extcorr.bdf"
define/local calobj/c/1/20      "fors_calibrated.bdf"
define/local exttbl/c/1/20      "fors_extspectra.tbl"
define/local wcsref/c/1/20      "wcsref.bdf"


! Local variables
define/local nopt/i/1/1         0
define/local pos/i/1/1          0
define/local ctrl/i/1/8         0             all
define/local catal/i/1/1        0
define/local calflg/i/1/1       0
define/local airmass/r/1/1      0.
define/local ron/r/1/1          0.
define/local poserr/r/1/1       15.
define/local conad/r/1/1        0.
define/local detpar/r/1/2       0.            all
define/local bscale/c/1/3       " "
define/local key/c/1/1          " "
define/local obsmod/c/1/3       " "           all
define/local calib/c/1/3        " "           all
define/local setup/c/1/20       " "           all
define/local finobj/c/1/100     " "           all
define/local cat_accept/c/1/20  "accept.cat"
define/local cat_reject/c/1/20  "reject.cat"
define/local cat_result/c/1/20  "result.cat"
define/local cat_entry/c/1/512  " "           all
define/local slitno/i/1/1       0
define/local loop/i/1/1         0
define/local usedslits/i/1/{FORS_MAXSLITLET} 0 all
define/local longlog1/c/1/240   " "           all
define/local longlog2/c/1/240   " "           all


! Connect to pipeline
log/rec "{p4}_{recID}" IN
mess/out 3 {procID} "Recipe {p4}_{recID}: Science exposure reduction."


! Process parameters
obsmod = m$upper(p4)
setup = "{obsmod}pipe"

nopt = m$parse(p5,"opt")
if nopt .gt. 3 .or. nopt .lt. 1 then
  mess/out 1 {procID} "Invalid detector parameter options given!"
  clean/pipeline xmos,spec
  return 1
else
  if m$tstno(opt01) .eq. 1 then
    detpar(1) = {opt01}
  else
    mess/out 1 {procID} "Invalid bias constant given!"
    clean/pipeline xmos,spec
    return 1
  endif

  if nopt .gt. 1 then
    if m$tstno(opt02) .eq. 1 then
      detpar(2) = {opt02}
    else
      mess/out 1 {procID} "Invalid dark constant given!"
      clean/pipeline xmos,spec
      return 1
    endif

    if nopt .gt. 2 then
      bscale = m$upper(opt03(1:3))
      if bscale .ne. "MED" .and. bscale .ne. "OVE" then
        if bscale .ne. "NO" then
          mess/out 1 {procID} "Invalid bias correction option given!"
          clean/pipeline xmos,spec
          return 1
        endif
      endif
    else
      bscale = "NO"
    endif
  else
    detpar(2) = 0.
    bscale = "NO"
  endif
endif

calib = m$upper(p6)
if calib(1:1) .eq. "Y" then
  calflg = 1
else
  calflg = 0
endif



! Startup context, initialize environment and prepare input 
! frames for processing
@p fors_paragraph 1

initialize/pipeline {p1} {p2} {cat_accept} xmos {setup} {obsmod} -
                    {cat_reject} {cat_result}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Pipeline initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Initialize product header
init/prohead {p1} {p2} 2,"Standard reduction of science spectrum" -
   obsmode,{p4} ccdpar,{p5} optcal,{p6}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Product header initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Initialize longlog
init/longlog {p1} {p2} 2 "obsmode: {p4}" "bias,dark,scale: {p5}" "calflag: {P6}"
if {q1} .ne. 0 then
  mess/out 1 {procID} "Longlog initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


@p fors_paragraph 2


! Update session keywords
read_cat_entry:

store/frame cat_entry {cat_accept} 1 exit_loop
fors_id = m$value({cat_entry},{FORS_IDSTRNG})
fors_id = m$upper(fors_id) 

if fors_id .eq. SCIENCE_ID then
  obj = cat_entry
  ctrl(1) = 1
elseif fors_id .eq. MBIAS_ID then
  bias = cat_entry
  ctrl(2) = 1
elseif fors_id .eq. MDARK_ID then
  dark = cat_entry
  ctrl(3) = 1
elseif fors_id .eq. MNFLAT_ID then
  normflat = cat_entry
  ctrl(4) = 1
elseif fors_id .eq. DISPREL_ID then
  linfit = cat_entry
  ctrl(5) = 1
elseif fors_id .eq. SLITS_ID then
  mos = cat_entry
  ctrl(6) = 1
elseif fors_id .eq. ATMEXT_ID then
  atmext = cat_entry
  ctrl(7) = 1
elseif fors_id .eq. RESP_ID then
  response = cat_entry
  ctrl(8) = 1
endif

goto read_cat_entry


exit_loop:

if ctrl(1) .eq. 0 then
  mess/out 1 {procID} "No object frame present!"
  clean/pipeline xmos,spec
  return 1
else if ctrl(4) .eq. 0 then
  mess/out 1 {procID} "No flat field frame present!"
  clean/pipeline xmos,spec
  return 1
else if ctrl(5) .eq. 0 then
  mess/out 1 {procID} "No dispersion relation present!"
  clean/pipeline xmos,spec
  return 1
else if ctrl(6) .eq. 0 then
  mess/out 1 {procID} "No slit position table present!"
  clean/pipeline xmos,spec
  return 1
else if calflg .eq. 1 .and. ctrl(7) .eq. 0 then
  mess/out 1 {procID} "No extinction table present!"
  clean/pipeline xmos,spec
  return 1
else if calflg .eq. 1 .and. ctrl(8) .eq. 0 then
  mess/out 1 {procID} "No instrument response present!"
  clean/pipeline xmos,spec
  return 1
endif


! Check if constants for bias and dark should be used.
if ctrl(2) .eq. 0 .and. detpar(1) .gt. 0. then
  bias = "{detpar(1)}"
  ctrl(2) = 1
endif

if ctrl(3) .eq. 0 .and. detpar(2) .gt. 0. then
  dark = "{detpar(2)}"
  ctrl(3) = 1
endif


! Setup slit(let) position(s)
define/position {obj} {mos} {obsmod}
if {q1} .ne. 0 then
  mess/out 1 {procID} "{obsmod} slit setup failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Change frame coordinate system of incoming science frame to PIXEL.
swap/units {obj} PIXEL
if {q1} .lt. 0 then
  mess/out 1 {procID} "Cannot establish PIXEL coordinate system in {obj}!"
  clean/pipeline xmos,spec
  return 1
endif


! Compute overscan bias for raw frame
if ctrl(2) .ne. 0 then
  compute/ovbias {obj} median
  if bscale .ne. "NO" then
    if {q1} .eq. -1 then
      mess/out 2 {procID} -
        "No overscan in input frame. Appylying standard bias correction."
      bscale = "NO"
    else if {q1} .ne. 0 then
      mess/out 1 {procID} "Bias computation from overscan areas failed!"
      clean/pipeline xmos,spec
      return 1
    endif
  endif
endif


! Remove pre- and overscan from object frame
trim/frame {obj}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Cannot remove pre- and overscan from {obj}!"
  clean/pipeline xmos,spec
  return 1
endif
shortlog/out 1 "Pre- and overscan regions chopped off."
  

! Basic reduction steps
if ctrl(2) .ne. 0 then
  @p fors_corr_bias {obj} {bias} + + {bscale}
  if {q1} .ne. 0 then
    mess/out 1 {procID} "Bias correction failed!"
    clean/pipeline xmos,spec
    return 1
  endif
else
  mess/out 4 {procID} "Bias correction disabled!"
endif

if ctrl(3) .ne. 0 then
  @p fors_corr_dark {obj} {dark}
  if {q1} .ne. 0 then
    mess/out 1 {procID} "Dark correction failed!"
    clean/pipeline xmos,spec
    return 1
  endif
else
  mess/out 4 {procID} "Dark correction disabled!"
endif

! Flat field correction. This step is required!
@p fors_corr_nflat {obj} {normflat}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Flat field correction failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Display uncalibrated spectrum
if nice_dids({dwin(1)}) .eq. 1 then
  view/spectrum {obj} {dwin(1)} mean,3 real avr "RAW SPECTRA"
endif


! Remap spectrum from pixel to wavelength space
mess/out 3 {procID} " "
mess/out 3 {procID} "Remapping spectrum frame from pixel to wavelength:"

! Remove slits in slit positions table for which there are no
! entries in the dispersion coefficients table; this happens
! if the wavelength calibration had failed for single slits.
mess/out 4 {procID} "  removing bad slits in {mos}"
select/table {mos} seq.eq.0 >Null
copy/table {mos} tmpall.tbl
set/format I1
do loop = 1 FORS_MAXSLITLET
  if m$existd(linfit,"COEFSLIT{loop}") .eq. 1 then
    select/table {mos} :SLIT.eq.{loop} >Null
    copy/table {mos} tmp{loop}.tbl
    merge/table tmpall.tbl tmp{loop}.tbl tmptmp.tbl
    -rename tmptmp.tbl tmpall.tbl
    usedslits({loop}) = 0                           ! "good slit"
  else
    select/table {mos} :SLIT.eq.{loop} >Null
    if outputi(1) .eq. 0 then
      usedslits({loop}) = 2                         ! "unused"
    else
      usedslits({loop}) = 1                         ! "calibration failed"
    endif
  endif
enddo
-rename tmpall.tbl {mos}
-delete tmp*.tbl

! Calculation of start and end wavelength and wavelength step
mess/out 4 {procID} "  calculation of wavelength range/step for rebinning"
define/range {obj} {mos} {linfit} n
if {q1} .eq. 1 then
  mess/out 1 {procID} "Selection of wavelength range failed !"
  clean/pipeline xmos,spec
  return 1
endif
rebstrt = outputr(1)
rebend = outputr(2)
set/format I1
slitno = outputi(1)  ! slit nearest to frame centre
rebstp = {{linfit},COEFSLIT{slitno}(2)}/{{linfit},CENTERXY(3)}  ! ~= lin.disp.

set/format T7.2
mess/out 4 {procID} " "
mess/out 3 {procID} "  Using dispersion relation >{linfit}<."
mess/out 3 {procID} "  Rebinning method:\t{rebmet}."
mess/out 3 {procID} "  Start wavelength:\t{rebstrt} A"
mess/out 3 {procID} "  End wavelength:\t{rebend} A"
mess/out 3 {procID} "  Wavelength step:\t{rebstp} A"
set/format

if mess_level .le. 3 then
  rebin/mos {obj} {redobj} {rebstrt},{rebend},{rebstp} {rebmet} {linfit} -
            {mos} >Null
else
  rebin/mos {obj} {redobj} {rebstrt},{rebend},{rebstp} {rebmet} {linfit} -
            {mos}
endif
mess/out 3 {procID} " "

! Output to longlog
set/format T7.2
longlog/out 1 "     Dispersion relation used: {linfit}"
longlog/out 1 "     Rebinning method:         {rebmet}"
longlog/out 1 "     Start wavelength:         {rebstrt} Angstrom"
longlog/out 1 "     End wavelength:           {rebend} Angstrom"
longlog/out 1 "     Wavelength step:          {rebstp} Angstrom"
longlog/out 1 " "
shortlog/out 1 "Frame rebinned to wavelength space."
set/format

! Display rebinned spectrum
if nice_dids({dwin(2)}) .eq. 1 then
  view/spectrum {redobj} {dwin(2)} + real avr "REBINNED SPECTRA" 
endif

! Print overview of used slits
mess/out 3 {procID} " "
mess/out 3 {procID} "Reduced slits"
mess/out 3 {procID} "    ("0"=used "1"=not calibrated "2"=unused):"
mess/out 3 {procID} "-------------------------------------------------"
set/format I1
append/out "#"
do loop = 1 FORS_MAXSLITLET
  append/out " {loop}"
enddo
copy/keyword cbuf longlog1
mess/out 3 {procID}
append/out "."
do loop = 1 FORS_MAXSLITLET
  if loop .ge. 10 set/format J2
  append/out " {usedslits({loop})}"
enddo
copy/keyword cbuf longlog2
mess/out 3 {procID}

longlog/out 3 "     Reduced slits"
longlog/out 3 "         ("0"=used "1"=not calibrated "2"=unused):"
longlog/out 3 "     -------------------------------------------------"
longlog/out 3 "     {longlog1}"
longlog/out 3 "     {longlog2}"
longlog/out 3 "     -------------------------------------------------"

longlog/out 3 " "
longlog/out 3 "     Reduced 2-d spectrum is: {redobj}"


! Correct for atmospheric extinction effects and apply the flux
! calibration.
if calflg .eq. 1 then
  get/airmass {redobj}
  if {q1} .eq. 0 then
    airmass = outputr(1)
  else
    mess/out 1 {procID} "Airmass computation failed!"
    clean/pipeline xmos,spec
    return 1
  endif

  mess/out 3 {procID} "Correcting for atmospheric extinction."
  extinction/mos {redobj} {corobj} {atmext} :{c_sextco} {airmass}

  longlog/out 1 "     Airmass: {airmass}"
  longlog/out 1 "     Wavelength-dependent extinct.coeff.tbl. used: {atmext}"
  shortlog/out 1 "Corrected for atmospheric extinction."

  mess/out 3 {procID} "Correcting for instrumental response."
  calibrate/flux {corobj} {calobj} {response} Y RATIO {rspftyp} -
                 {rspfpar} spectrum
  
  longlog/out 1 "     Response table used:     {response}"
  longlog/out 1 "     Response fit type:       {rspftyp}"
  longlog/out 1 "     Response fit parameters: {rspfpar}"
  shortlog/out 1 "Corrected for instrumental response."

! Display calibrated spectrum
  if nice_dids({dwin(3)}) .eq. 1 then
    view/spectrum {calobj} {dwin(3)} mean,3 real avr "CALIBRATED SPECTRA"
  endif

  finobj = calobj

  longlog/out 3 " "
  longlog/out 3 "Calibrated 2-d spectrum is: {calobj}"

else
  finobj = redobj
endif


! Copy all but the standard descriptors to the reduced frame
copy/dd {obj} *,3,o_time {redobj}


! Define objects and sky regions for spectrum extraction
locate/spectrum {obj} {mos} {obsmod} {windows} {poserr} {wcsref}
if nice_dids({dwin(2)}) .eq. 1 then
  define/window load + + + {mos} {windows} >Null
  annotate/spectrum {finobj} {dwin(2)} "OBJECT/SKY DETERMINATION"
endif


! Setup CCD parameters conad (e-/ADU), readout noise and fit sky background
conad = m$value({finobj},{h_confac({fors_outi(1)})})
if fors_nout .gt. 1 then
  do i = 2 fors_nout
    conad = conad + m$value({finobj},{h_confac({fors_outi({i})})})
  enddo
  conad = conad / fors_nout

  set/format i1 f6.3
  mess/out 2 {procID} -
    "Multi port readout encountered! Using mean conversion factor {conad}!"
  set/format
endif

ron = m$value({finobj},{h_ron({fors_outi(1)})})
if fors_nout .gt. 1 then
  do i = 2 fors_nout
    ron = ron + m$value({finobj},{h_ron({fors_outi({i})})})
  enddo
  ron = ron / fors_nout
endif

key = m$upper(skymet(1:1))
set/format i1 f1.2
mess/out 3 {procID} "Fitting sky background."
mess/out 3 {procID} "Determining sky background of >{finobj}<"
mess/out 4 {procID} "Fitted Sky frame >{skyframe}<"
mess/out 3 {procID} " "
mess/out 3 {procID} "Using method:\t\t{skymet}"
if key .eq. "P" then
  mess/out 3 {procID} "Order of fitted polynomial:\t{skypar(1)}"
else
  mess/out 3 {procID} "Median window size:\t{skypar(1)}"
endif
mess/out 4 {procID} "Readout noise:\t\t{ron} e-."
mess/out 4 {procID} "Conversion factor e-/adu:\t{conad}."
mess/out 3 {procID} "Detection threshold:\t{rejthres}."
set/format

if mess_level .le. 3 then
  skyfit/mos {finobj} {mos} {windows} {skyframe} {skypar(1)} {skymet} -
             {ron},{conad},{rejthres} >Null
else
  skyfit/mos {finobj} {mos} {windows} {skyframe} {skypar(1)} {skymet} -
             {ron},{conad},{rejthres} >Null
endif
mess/out 3 {procID} " "


! Output to longlog
set/format i1 f1.2
longlog/out 1 "     Determining sky background of {finobj}"
longlog/out 1 "     Using method:                {skymet}"
if key .eq. "P" then
  longlog/out 1 "     Order of fitted polynomial:  {skypar(1)}"
else
  longlog/out 1 "     Median window size:          {skypar(1)}"
endif
longlog/out 1 "     Detection threshold:         {rejthres}"
longlog/out 1 "     Fitted Sky frame is:         {skyframe}"
shortlog/out 1 "Fitted sky bkgr. used f. extract. of spectra."
set/format


! Update sky frame header (mainly for display purposes)
copy/dd {finobj} *,3 {skyframe}


! Display sky frame
if nice_dids({dwin(3)}) .eq. 1 then
  view/spectrum {skyframe} {dwin(3)} mean,2 real avr "FITTED SKY"
endif

! Optimal extraction of object spectrum
set/format i1 f1.2
mess/out 3 {procID} "Extracting point source spectra."
mess/out 3 {procID} "Extracting object spectra from >{finobj}<"
mess/out 4 {procID} "Extracted object spectra >{extobjec}<"
mess/out 3 {procID} " "
mess/out 3 {procID} "Extraction parameter (order,iterations): {extpar}"
mess/out 4 {procID} "Readout noise:\t\t{ron} e-."
mess/out 4 {procID} "Conversion factor e-/adu:\t{conad}."
mess/out 3 {procID} "Pixel rejection threshold:\t{rejthres} sigma"
set/format

mess/out 3 {procID} "Extracting object ..."
mess/out -3 {procID} "..."
extract/mos {finobj} {skyframe} {windows} {extobjec} {extpar},{mess_level} -
            {ron},{conad} {rejthres}
mess/out 3 {procID} " "


convert/extract {extobjec} {windows} {exttbl}
copy/dd {finobj} *,3 {exttbl}


! Output to longlog
set/format i1 f1.2
longlog/out 1 "     Extracted object spectra from            {finobj}"
longlog/out 1 "     Extraction params (order,iterations):    {extpar}"
longlog/out 1 "     Pixel rejection threshold:               {rejthres} sigma"
longlog/out 1 "     Extracted 1-d object spectra table is:   {exttbl}"
shortlog/out 1 "Point source spectra extracted."
set/format


! For the rebinned standard star frames a new coordinate system along y-axis
! is established. The x-axis must mot be updated, since it has wavelength
! units after rebinning.
{redobj},cunit(33:48) = "ARCSEC"
{redobj},step(2) = m$value({redobj},{h_pixscale})
{redobj},start(2) = (1.0 - m$value({redobj},refpix(2))) * -
                    m$value({redobj},{h_pixscale})


! Prepare output catalog
classify/product {redobj} {RSCIENCE_ID} {d_tmp} {d_geo},{d_det} OK ? y
classify/product {exttbl} {ESCI_ID} {d_tmp} {d_geo},{d_det} OK
classify/product {skyframe} {SCIBKG_ID} {d_tmp} {d_geo},{d_det} OK ? y

write/desc {redobj} {FORS_IDSTRNG}/c/1/20 {RSCIENCE_ID}
write/desc {exttbl} {FORS_IDSTRNG}/c/1/20 {ESCI_ID}
write/desc {skyframe} {FORS_IDSTRNG}/c/1/20 {SCIBKG_ID}

add/icat {cat_result} {redobj} >Null
add/icat {cat_result} {exttbl} >Null
add/icat {cat_result} {skyframe} >Null


if calflg .eq. 1 then
! Copy all but the standard descriptors to the calibrated frame
  copy/dd {obj} *,3,o_time {calobj}


! A new coordinate system along y-axis
! is established. The x-axis must mot be updated, since it has wavelength
! units after rebinning.
  {calobj},cunit(33:48) = "ARCSEC"
  {calobj},step(2) = m$value({calobj},{h_pixscale})
  {calobj},start(2) = (1.0 - m$value({calobj},refpix(2))) * -
                      m$value({calobj},{h_pixscale})


! Classify calibrated frame
  classify/product {calobj} {FSCIENCE_ID} {d_red} {d_pho},{d_geo},{d_det} OK ? y
  write/desc {calobj} {FORS_IDSTRNG}/c/1/20 {FSCIENCE_ID}


  add/icat {cat_result} {calobj} >Null

endif

@p fors_paragraph 3


! Create product header
finish/prohead {cat_result} 2


! Shutdown pipeline
finish/pipeline {cat_result} {p3} xmos,spec {finobj}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Error leaving pipeline environment!"
  finish/longlog {cat_accept} -1 ? 1 {p3}
  clean/pipeline xmos,spec
  return 1
endif


! Close longlog
finish/longlog {cat_accept} {cat_result} {finobj} 2 {p3}


! Disconnect from pipeline
log/rec "{p4}_{recID}" OUT

@p fors_paragraph 4

return 0
