! $Id: fors_s_img_science.prg,v 1.46 2001/04/09 15:01:08 sbogun Exp $   
!+++ FORS_S_IMG_SCIENCE ++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT     (C) 1998  European Southern Observatory
!
!.IDENT         $RCSfile: fors_s_img_science.prg,v $
!
!.KEYWORDS      SUPER RECIPE: Complete science frame pipline processing
!
!.USAGE         @p (IDENT) sof_in sof_ref out bias dark
!
!.INPUT         sof_in  (P1): set of input frames (ascii list)
!                               name of science frame
!               sof_ref (P2): set of reference frames (ascii list)
!                               Badpixel-table, Bias, (Dark), Flat,
!                               Extinction Coefficients table
!                               Photometry Input parameter table (*)
!                               Photometry Output parameter table (*)
!                               Photometry Convolution mask (*)
!                               Photometry Neuronal network weights (*)
!                               (*): used by sextractor in photometry recipe
!               bias    (P4): Bias constant to be used instead of master.
!               dark    (P5): Dark constant to be used instead of master.
!               scale   (P6): "MED": the bias correction will be done using the 
!                               scaled master bias. The scaling is done as
!                               follows: 
!                               Each of the sections of the master bias is 
!                               divided by its median and multiplied by the
!                               median of the corresponding pre-/overscan 
!                               region of the input frame.
!                             "OVE": like "MED", but each of the sections of the
!                               master bias is divided by the median of its
!                               corresponding pre-/overscan region and
!                               multiplied by the median of the pre-/overscan
!                               region of the input frame.
!                             "NO":  The master bias will not be scaled.
!
!.OUTPUT        out     (P3): path + basename of output files:
!
!.RETURN        Q1:  0: successful return
!                   -1: a successful return failed
!
!.DEFAULT
!
!.PURPOSE       Performs the whole pipeline processing of a science frame.
!               It starts with two quality control procedures: check on
!               overexposed pixels and computation of basic statistical
!               parameters.
!               After having passed these quality control procedures, the 
!               science frame is reduced in a standard way (bad pixel
!               correction and
!               bias-, dark-subtraction as well as flatfield division). Now
!               the science frame is searched for objects. The image quality 
!               is determined from the extracted objects' parameters.
!               The derived object list and the frame itself are
!               flux-calibrated.
!
!.ALGORITHM     Calls single recipes in the following order:
!               - look for overexposed pixels             (QC)
!               - determination of statistical parameters (QC)
!               - reduce science frame
!               - extract sky and subtract from science frame
!               - find objects in the frame, determine image quality
!               - do the flux calibration of the frame
!
!.ENVIRON       FORS, CCDRED context
!   
!.LANGUAGE      MIDAS 98NOV
!
!.AUTHOR        Sebastian Wolf, ESO-DMO
!
!.VERSION       0.9 1998/07/09  Creation
!               1.0 1998/12/03  main product defined: FLUX-CALIBRATED science
!                               parameter handling implemented
!                   1998/12/08  paragraphs implemented
!                   1998/12/10  error handler implemented (write/error)
!                   1999/01/07  shortlog implemented.
!               1.2 1999/02/03  update write/error
!                   1999/03/10  shortlog close shifted to FINISH/PIPELINE
!               1.3 1999/09/03  Product header generation implemented (SB).
!
!.COMMENT       SUPER-RECIPES:
!               All output frames will have the following naming-convention:
!
!                    {P3}_{n}.fits              while {n} counts starting at 0
!
!               They will be finally saved as soon as FinishPipeline will be
!               called at the end of the SUPER-RECIPE. Until then only
!               temporary BDF-versions of the outputs will exist in the local
!               directory. All outputs have to be protocolled adding these
!               BDF-files to the result catalog created by the first
!               InitPipeline call.
!
!---------------------------------------------------------------------------

define/param p1 ? c "Enter name of ASCII filelist with science frames:"
define/param p2 ? c "Enter name of ASCII filelist with reference frames:"
define/param p3 ? c "Enter name of output-basename (+path):"
define/param p4 0 c "Enter bias constant to be used instead of master:"
define/param p5 0 c "Enter dark constant to be used instead of master:"
define/param p6 OVE c "Use scaled master bias (MED/OVE/NO):"

define/local part/i/1/1 0
define/local procname/c/1/25 "s_img_science"
define/local line/c/1/79 "-" ALL
define/local scienceflx/c/1/80 " " ALL
define/local errmsg/c*80/1/11 " " ALL

errmsg(1) = "Error in INIT/PIPELINE occured !"
errmsg(2) = "Error in fors_gen_qc0_overexpo occured !"
errmsg(3) = "Error in fors_gen_qc1_statistics occured !"
errmsg(4) = "Error in CHECK/OVERSCAN occured !"
errmsg(5) = "Error in OVERSCAN/FRAMES occured !"
errmsg(6) = "Error in REDUCE/IMAGE occured !"
errmsg(7) = "Error in fors_img_obs_photometry occured !"
errmsg(8) = "Error in CALCULATE/FLUX occured !"
errmsg(9) = "Error while identifying main product !"
errmsg(10) = "Error in FINISH/PIPELINE !"
errmsg(11) = "No good frames left after overexposure check !"


LOG/REC {procname} IN SUPER

mess/out 3 {procname} "START reduction of science frame ..."
write/out
@p fors_paragraph 1 0


part = 1
INIT/PIPELINE {P1} {P2} XSFinput.cat CCDRED + IMG XSFreject.cat XSFresult.cat
if {Q1} .gt. 0 goto ABORT
!
!
! now all input frames and reference frames are stored in the image catalog:
!    XSFinput.cat
! thus P2 may be ignored (+) in the following recipe-calls.
! (P2 usually refers to the set of reference frames)
!

init/prohead {p1} {p2} 1,"Science image reduction" "bias,{p4}" "dark,{p5}" -
  "scale,{p6}"
if {Q1} .gt. 0 goto ABORT

write/out
init/longlog {p1} {p2} 1 "bias: {p4}" "dark: {p5}" "scale: {p6}"
if {Q1} .gt. 0 goto ABORT

@p fors_paragraph 2 0
!------------------------------------------------------------------------------
!Check overexposed pixels
!
part = 2
mess/out 3 {procname} "Check for overexposed pixels (no rejection) ..."
write/out {line}
@p fors_gen_qc0_overexpo XSFinput.cat + XSFinput.cat,XSFresult.cat 100.
if {Q2} .eq. 0 mess/out 1 {procname} "{errmsg(11)}"
if {Q1} .gt. 0 goto ABORT

!------------------------------------------------------------------------------
!Perform statistic tests
!
part = 3
write/out
mess/out 3 {procname} "Perform statistic tests ..."
write/out {line}
@p fors_img_qc0_statistics XSFinput.cat + XSFinput.cat,XSFresult.cat {p6}
if {Q1} .gt. 0 goto ABORT

!------------------------------------------------------------------------------
!determine overscan
!
!! part = 4
!! write/out
!! mess/out 3 {procname} "Determine bias from overscan ..."
!! check/overscan XSFinput.cat + XSFinput.cat,XSFresult.cat
!! if {Q1} .gt. 0 goto ABORT
!! 
!------------------------------------------------------------------------------
!do the reduction
!
part = 6
write/out
mess/out 3 {procname} "Reduce science frame ..."
reduce/image XSFinput.cat + XSFinput.cat,XSFresult.cat {P4} {P5} 0 {P6}
if {Q1} .ne. 0 goto ABORT

!------------------------------------------------------------------------------
!do the photometry
!
part = 7
write/out
mess/out 3 {procname} "Determine photometric values ..."
write/out {line}
@p fors_img_obs_photometry XSFinput.cat + XSFinput.cat,XSFresult.cat
if {Q1} .gt. 0 goto ABORT

!------------------------------------------------------------------------------
!do the flux calibration
!
part = 8
write/out
mess/out 3 {procname} "Do the flux calibration ..."
calculate/flux XSFinput.cat + XSFinput.cat,XSFresult.cat
if {Q1} .ne. 0 goto ABORT


!------------------------------------------------------------------------------
!write FITS files to specified output directory
!
part = 9
write/out
@p fors_paragraph 3 0
search/icat XSFresult {FSCIENCE_ID} >Null
if m$len(out_a) .gt. 0 then
  scienceflx = out_a
else
  mess/out 1 {procname} "No result frame ({FSCIENCE_ID}) found !"
  goto ABORT
endif


part = 10
mess/out 3 {procname} "Finish Pipeline"
finish/prohead XSFresult.cat 1
FINISH/PIPELINE XSFresult.cat {P3} CCDRED {scienceflx}
if {Q1} .gt. 0 goto ABORT
finish/longlog XSFinput.cat XSFresult.cat {scienceflx} 1 {p3}

@p fors_s_img_science,CHK_STATUS

LOG/REC {procname} OUT SUPER
@p fors_paragraph 4 0
return 0


ABORT:
finish/longlog XSFinput.cat -1 ? 1 {p3}
mess/out 1 {procname} "{errmsg({part})}"
write/error {part},APP,NO "{errmsg({part})}"



ENTRY CHK_STATUS
if RECQUALI .eq. 1 then                                 !check recipe status
  RECQUALM = "Reduced frame of less quality (see log) !"
else
  write/key RECQUALM " " all
endif











