! $Id: fors_s_lss_science.prg,v 1.15 2001/03/21 14:49:04 sbogun Exp $   
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1999  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_s_lss_science.prg,v $
!
!.KEYWORDS    Complete science spectrum reduction and flux calibration,
!             super recipe
!
!.USAGE       @p fors_s_lss_science  sof_in sof_ref sof_out ccdpar
!
!.INPUT       sof_in      (P1)    Set of input frames to be processed.
!                                 Contains a single science observation
!             sof_ref     (P2)    Set of reference frames including:
!                                  - master bias
!                                  - [master dark]
!                                  - normalized master screen flat field
!                                  - dispersion relation table
!                                  - instrument response table
!                                  - atmospheric extinction coefficients
!                                  - slit positions table
!                                  - setup table for data reduction system
!             ccdpar      (P4)    Optional bias, dark constants and bias scaling
!                                 method ("MED", "OVE" or "NO").
!
!.OUTPUT      sof_out     (P3)    Set of output frames, i.e path and basename
!                                 for output files.
!
!.DEFAULTS    ccdpar = 0.,0.,med   (bias, dark constant, scaling method)
!
!.RETURNS     The recipe returns 0 on success. Keyword Q1 is updated with 
!             the exit status. 
!
!.PURPOSE     The recipe performs the complete science observation reduction,
!             including the flux calibration. 
!             The incoming science exposure is first searched for overexposed
!             pixels. If
!             the input frame passes this quality check the frame is
!             corrected for bias and dark if the appropriate frames are found
!             in the set of reference frames. If no such frames are present
!             the bias and the dark correction may be enforced by setting the
!             bias and dark constants (CCDPAR) to a value greater than
!             zero. The pixel to pixel variations are removed by dividing the
!             science exposure by a normalized flat field. Finally the
!             dispersion relation is applied to the science frame and is then
!             rebinned to constant wavelength steps.
!             A flux calibration is performed, i.e. the science
!             exposure is corrected for atmospheric extinction effects and
!             the wavelength dependent instrument efficiency.
!             In the end the rebinned frame is searched for point source
!             spectra. The sky background at the positions of the object
!             spectra are fitted and subtracted. Now the spectra are extracted
!             using the algorithm of Horne (Horne, 1986, PASP, 98, 609).
!
!.ALGORITHM   Executes individual tasks in the following order:
!              - search for overexposed pixels
!              - compute bias from overscan region
!              - apply the dispersion relation and rebin the 2d spectrum
!                to constant wavelength steps
!              - correction of atmospheric extinction effects
!              - correction of instrument efficiency
!              - search for point source spectra
!              - fit sky background
!              - extract the detected object spectra
!   
!.ENVIRON     FORS context, XMOS conext
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1999/03/10:  Creation
!             0.6  1999/04/28:  Longlog implemented (SB)
!             1.3  1999/09/03:  Product header generation implemented (SB).
!
!------------------------------------------------------------------------------

define/param p1 ?         c  "Enter input frame set:"
define/param p2 ?         c  "Enter reference frame set:"
define/param p3 ?         c  "Enter output frame set:"
define/param p4 0.,0.,med c  "Enter bias/dark const. and scaling method:"


! Procedure identification
define/local procID/c/1/80      "S_LSS_Science"


! Recipe identification
define/local recID/c/1/80       "S_LSS_SCIENCE"


! Local constants
define/local obsmod/c/1/3     "LSS"
define/local separator/c/1/79  "-"    all


! Local variables
define/local errno/i/1/1      0
define/local setup/c/1/20     " "  all
define/local ccdcnst/c/1/100  " "  all
define/local profrm/c/1/100   " "  all
define/local errmsg/c/1/100   " "  all



! Connect to pipeline
log/rec "{recID}" IN SUPER


! Process parameters
ccdcnst = p4
setup = "{obsmod}pipe"


! Startup pipeline
mess/out 3 {procID} "Starting science spectrum reduction ..."
errno = errno + 1

@p fors_paragraph 1 0

initialize/pipeline {p1} {p2} XSFinput.cat xmos {setup} {obsmod} -
          XSFreject.cat XSFresult.cat
if {q1} .ne. 0 then
  errmsg = "Pipeline initialization failed!"
  mess/out 1 {procID} "{errmsg}"
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Initialize product header
init/prohead {p1} {p2} 1,"Science spectrum reduction" ccdpar,{p4}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Product header initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Initialize longlog
init/longlog {p1} {p2} 1 "bias,dark,scale: {p4}"
if {q1} .ne. 0 then
  errmsg = "Longlog initialization failed!"
  mess/out 1 {procID} "{errmsg}"
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Search for overexposed pixels
@p fors_paragraph 2 0
write/out
mess/out 3 {procID} "Searching for overexposed pixels (no rejection) ..."
write/out {separator}
errno = errno + 1

@p fors_gen_qc0_overexpo XSFinput.cat + XSFinput.cat,XSFresult.cat 100.
if {q2} .eq. 0 then
  mess/out 1 {procID} "Science exposure rejected by overexposure check."
  errmsg = "No frames left in input. Execution of reduction block stopped!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 ? 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
else if {q1} .gt. 0 then
  errmsg = "FORS_GEN_QC0_OVEREXPO terminated with error status!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 ? 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


!! ! Bias determination using overscan regions
!! write/out
!! mess/out 3 {procID} "Bias determination using overscan regions ..."
errno = errno + 1
!!
!! check/overscan XSFinput.cat + XSFinput.cat,XSFresult.cat
!! if {q1} .gt. 0 then
!!   errmsg = "CHECK/OVERSCAN terminated with error status!"
!!   mess/out 1 {procID} "{errmsg}"
!!   finish/longlog XSFinput.cat -1 ? 1 {p3}
!!   clean/pipeline ccdred,xmos,spec
!!   write/error {errno},app,no "{errmsg}"
!! endif


! Do science spectrum reduction
write/out
mess/out 3 {procID} "Processing science exposure ..."
write/out {separator}
errno = errno + 1

reduce/obslss XSFinput.cat + XSFinput.cat,XSFresult.cat {obsmod} {ccdcnst} Yes
if {q1} .ne. 0 then
  errmsg = "REDUCE/OBSLSS terminated with error status!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 ? 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Lookup main product file in catalog
errno = errno + 1

@p fors_paragraph 3 0
search/icat XSFresult {FSCIENCE_ID} NB
if m$len(out_a) .ne. 0 then
  profrm = out_a
else
  errmsg = "Product frame ({FSCIENCE_ID}) not found in product catalog!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 ? 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Save product files
errno = errno + 1


! Create product header
finish/prohead XSFresult.cat 1


finish/pipeline XSFresult.cat {p3} xmos,spec {profrm}
if {q1} .ne. 0 then
  errmsg = "Error leaving pipeline environment!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 {profrm} 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Close longlog
finish/longlog XSFinput.cat XSFresult.cat {profrm} 1 {p3}


! Disconnect from pipeline
log/rec "{recID}" OUT SUPER

@p fors_paragraph 4

return 0

