! $Id: fors_s_lss_stdstar.prg,v 1.23 2000/09/12 07:16:16 sbogun Exp $   
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1999  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_s_lss_stdstar.prg,v $
!
!.KEYWORDS    Complete standard star spectrum reduction, super recipe
!
!.USAGE       @p fors_s_lss_stdstar sof_in sof_ref sof_out ccdpar
!                                   fwhm poserr
!
!.INPUT       sof_in      (P1)    Set of input frames to be processed.
!                                 Contains a single standard star 
!                                 observation
!             sof_ref     (P2)    Set of reference frames including:
!                                  - master bias
!                                  - [master dark]
!                                  - normalized master screen flat field
!                                  - dispersion relation table
!                                  - standard star flux table
!                                  - slit positions table
!                                  - extinction coefficients table
!                                  - setup table for data reduction system
!             ccdpar      (P4)    Optional bias, dark constants and bias scaling
!                                 method ("MED", "OVE" or "NO").
!             fwhm        (P5)    Initial, maximum width expected for point
!                                 sources in terms of arcsec
!             poserr      (P6)    Telescope pointing error in terms of arcsec
!             climit      (P7)    Confidence limit for source identification
!                                 (not yet supported!)
!
!.OUTPUT      sof_out     (P3)    Set of output frames, i.e path and basename
!                                 for output files.
!
!.DEFAULTS    ccdpar = 0.,0.,med   (bias, dark constant, scaling method)
!             fwhm   = 16.0 arcsec
!             poserr = 15.0 arcsec
!             climit = 0.
!
!.RETURNS     The recipe returns 0 on success. Keyword Q1 is updated with 
!             the exit status. 
!
!.PURPOSE     The recipe performs the complete standard star reduction and
!             spectrum extraction. Additionally the extracted spectra are
!             identified with standard stars and the instrument response is
!             computed. The incoming standard star frame is first checked for
!             overexposed pixels. If the input frame passes this quality check
!             the frame is corrected for bias and dark if the appropriate
!             frames are found in the set of reference frames. If no such
!             frames are present the bias and the dark correction may be
!             enforced by setting the bias and dark constants (CCDPAR) to a
!             value greater than zero. The standard star frame is divided by
!             the normalized flat field to remove pixel to pixel variations.
!             Now the dispersion relation is applied to the standard star
!             frame and it is rebinned to constant wavelength steps. The
!             rebinned frame is searched for point source spectra. The
!             spectrum position is determined by fitting a gaussian to the
!             profile of the spectrum along the slit. The initial width of the
!             gaussian is given by the parameter FWHM. The source position
!             corresponding with the spectrum is taken as the center of the
!             slit and the determined object region. Taking the pointing error
!             of the telescope given by POSERR into account, this position is
!             used to identify the source with a standard star from the list
!             of spectrophotometric standards. The sky background at the
!             position of the object spectrum is fitted and subtracted. Now
!             the spectra are extracted using the algorithm of Horne (Horne,
!             1986, PASP, 98, 609). In a final step the extracted spectra are
!             corrected for atmospheric extinction effects and divided by the
!             reference flux. This yields the instrumental response.
!
!.ALGORITHM   Executes individual tasks in the following order:
!              - search for overexposed pixels
!              - compute bias from overscan region
!              - apply the dispersion relation and rebin the 2d spectrum
!                to constant wavelength steps
!              - search for point source spectra
!              - fit sky background
!              - extract the detected object spectra
!              - identify the detected objects
!              - compute the instrument response
!   
!.ENVIRON     FORS context, XMOS conext
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1999/03/10:  Creation
!             0.6  1999/04/28:  Longlog implemented (SB)
!             1.3  1999/09/03:  Product header generation implemented (SB).
!
!------------------------------------------------------------------------------

define/param p1 ?         c  "Enter input frame set:"
define/param p2 ?         c  "Enter reference frame set:"
define/param p3 ?         c  "Enter output frame set:"
define/param p4 0.,0.,med c  "Enter bias/dark const. and scaling method:"
define/param p5 16.       c  "Enter maximum FWHM of spatial profile:"
define/param p6 15.       c  "Enter telescope pointing uncertainty:"
define/param p6 0.        c  "Enter confidence limit for soure identifications:"


! Procedure identification
define/local procID/c/1/80      "S_LSS_Stdstar"


! Recipe identification
define/local recID/c/1/80       "S_LSS_STDSTAR"


! Local constants
define/local obsmod/c/1/3     "LSS"
define/local separator/c/1/79  "-"    all


! Local variables
define/local errno/i/1/1      0
define/local setup/c/1/20     " "  all
define/local ccdcnst/c/1/100  " "  all
define/local profrm/c/1/100   " "  all
define/local errmsg/c/1/100   " "  all
define/local fwhm/c/1/100     {p5}
define/local poserr/c/1/100   {p6}
define/local climit/c/1/100   {p7}



! Connect to pipeline
log/rec "{recID}" IN SUPER


! Process parameters
ccdcnst = p4
setup = "{obsmod}pipe"


! Startup pipeline
mess/out 3 {procID} "Starting standard star spectrum reduction ..."
errno = errno + 1

@p fors_paragraph 1 0

initialize/pipeline {p1} {p2} XSFinput.cat xmos {setup} {obsmod} -
          XSFreject.cat XSFresult.cat
if {q1} .ne. 0 then
  errmsg = "Pipeline initialization failed!"
  mess/out 1 {procID} "{errmsg}"
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Initialize product header
init/prohead {p1} {p2} 1,"Standard star spectrum reduction" -
             ccdpar,{p4} fwhm,{p5} poserr,{p6} climit,{p7}
if {q1} .ne. 0 then
  mess/out 1 {procID} "Product header initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Initialize longlog
init/longlog {p1} {p2} 1 "bias,dark,scale: {p4}" "fwhm: {p5}" "poserr: {p6}" -
             "climit: {p7}"
if {q1} .ne. 0 then
  errmsg = "Longlog initialization failed!"
  mess/out 1 {procID} "{errmsg}"
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Search for overexposed pixels
@p fors_paragraph 2 0
write/out
mess/out 3 {procID} "Searching for overexposed pixels ..."
write/out {separator}
errno = errno + 1

@p fors_gen_qc0_overexpo XSFinput.cat + XSFinput.cat,XSFresult.cat
if {q2} .eq. 0 then
  mess/out 1 {procID} "Standard star frame rejected by overexposure check."
  errmsg = "No frames left in input. Execution of reduction block stopped!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 ? 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
else if {q1} .gt. 0 then
  errmsg = "FORS_GEN_QC0_OVEREXPO terminated with error status!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 ? 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Bias determination using overscan regions
!! write/out
!! mess/out 3 {procID} "Bias determination using overscan regions ..."
errno = errno + 1
!! 
!! check/overscan XSFinput.cat + XSFinput.cat,XSFresult.cat
!! if {q1} .gt. 0 then
!!   errmsg = "CHECK/OVERSCAN terminated with error status!"
!!   mess/out 1 {procID} "{errmsg}"
!!   finish/longlog XSFinput.cat -1 ? 1 {p3}
!!   clean/pipeline ccdred,xmos,spec
!!   write/error {errno},app,no "{errmsg}"
!! endif


! Do standard star spectrum reduction
write/out
mess/out 3 {procID} "Processing standard star frame ..."
write/out {separator}
errno = errno + 1

reduce/stdlss XSFinput.cat + XSFinput.cat,XSFresult.cat {obsmod} {ccdcnst} -
              {fwhm} {poserr}
if {q1} .ne. 0 then
  errmsg = "REDUCE/STDLSS terminated with error status!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 ? 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Identify standard star spectra
write/out
mess/out 3 {procID} "Identifing standard star spectra ..."
write/out {separator}
errno = errno + 1

align/stdlss XSFinput.cat + XSFinput.cat,XSFresult.cat {obsmod} {climit}
if {q1} .ne. 0 then
  errmsg = "ALIGN/STDLSS terminated with error status!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 ? 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Quality control for spectroscopic observations
write/out
mess/out 3 {procID} "Doing quality checks on identified spectra ..."
write/out {separator}
errno = errno + 1

@p fors_lss_qc1_chkobs XSFinput.cat + XSFinput.cat,XSFresult.cat c {obsmod}
if {q1} .eq. 1 then
  errmsg = "FORS_LSS_QC1_CHKOBS terminated with error status!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 ? 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Compute instrument response
write/out
mess/out 3 {procID} "Computing instrument response ..."
write/out {separator}
errno = errno + 1

@p fors_lss_cal_response XSFinput.cat + XSFinput.cat,XSFresult.cat {obsmod} -
   49.0
if {q1} .ne. 0 then
  errmsg = "FORS_LSS_CAL_RESPONSE terminated with error status!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 ? 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Lookup main product file in catalog
errno = errno + 1

@p fors_paragraph 3 0
search/icat XSFresult {RESP_ID} NB
if m$len(out_a) .ne. 0 then
  profrm = out_a
else
  errmsg = "Product frame ({RESP_ID}) not found in product catalog!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 ? 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Save product files
errno = errno + 1


! Create product header
finish/prohead XSFresult.cat 1


finish/pipeline XSFresult.cat {p3} xmos,spec {profrm}
if {q1} .ne. 0 then
  errmsg = "Error leaving pipeline environment!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 {profrm} 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Close longlog
finish/longlog XSFinput.cat XSFresult.cat {profrm} 1 {p3}


! Disconnect from pipeline
log/rec "{recID}" OUT SUPER

@p fors_paragraph 4

return 0
