! $Id: fors_s_mos_wave.prg,v 1.13 2001/03/08 12:58:02 sbogun Exp $   
!++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!.COPYRIGHT   (C) 1999  European Southern Observatory (ESO)
!
!.IDENT       $RCSfile: fors_s_mos_wave.prg,v $
!
!.KEYWORDS    Complete wavelength calibration, super recipe
!
!.USAGE       @p fors_s_mos_wave  sof_in sof_ref sof_out
!
!.INPUT       sof_in      (P1)    Set of input frames to be processed.
!                                 Contains a single calibration frame
!             sof_ref     (P2)    Set of reference frames including:
!                                  - slit positions table
!                                  - master line catalog
!                                  - grism parameter table
!                                  - table with color dependence of camera
!                                    focal length
!                                  - setup table for data reduction system
!
!.OUTPUT      sof_out     (P3)    Set of output frames, i.e path and basename
!                                 for output files.
!
!.DEFAULTS    
!
!.RETURNS     The recipe returns 0 on success. Keyword Q1 is updated with
!             the exit status. 
!
!.PURPOSE     The recipe performs the complete wavelength calibration for
!             a specific instrument setup in MOS mode. The incoming
!             calibration frame is first checked for overexposed pixels. 
!             The comparison spectrum is searched for emission lines which
!             are identified using the optical parameters of the FORS grisms.
!             The identified lines are used to perform a two-dimensional 
!             modelling of the dispersion relation. Finally the computed
!             dispersion relation is checked by comparing it to the 
!             corresponding linear approximation.
!
!.ALGORITHM   Executes individual tasks in the following order:
!              - search for overexposed pixels
!              - model two-dimensional dispersion relation (including simple
!                internal verification of the result)
!   
!.ENVIRON     FORS context, XMOS conext
!
!.LANGUAGE    MIDAS command language 98NOV.
!
!.AUTHOR      R. Palsa, ESO-DMO
!
!.VERSION     0.5  1999/03/10:  Creation
!             1.3  1999/09/03:  Product header generation implemented (SB).
!
!------------------------------------------------------------------------------

define/param p1 ?            c    "Enter input frame set:"
define/param p2 ?            c    "Enter reference frame set:"
define/param p3 ?            c    "Enter output frame set:"


! Procedure identification
define/local procID/c/1/80      "S_MOS_Wave"


! Recipe identification
define/local recID/c/1/80       "S_MOS_WAVE"


! Local constants
define/local obsmod/c/1/3      "MOS"
define/local separator/c/1/79  "-"    all
define/local CAT_PLOT/c/1/40   "plot.cat"


! Local variables
define/local errno/i/1/1      0
define/local setup/c/1/20     " "  all
define/local profrm/c/1/100   " "  all
define/local errmsg/c/1/100   " "  all



! Connect to pipeline
log/rec "{recID}" IN SUPER


! Process parameters
setup = "{obsmod}pipe"


! Startup pipeline
mess/out 3 {procID} "Starting wavelength calibration ..."
errno = errno + 1

@p fors_paragraph 1 0

initialize/pipeline {p1} {p2} XSFinput.cat xmos {setup} {obsmod} -
          XSFreject.cat XSFresult.cat
if {q1} .ne. 0 then
  errmsg = "Pipeline initialization failed!"
  mess/out 1 {procID} "{errmsg}"
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Initialize plot file catalogue
initialize/plot {CAT_PLOT} 1


! Initialize product header
init/prohead {p1} {p2} 1,"Wavelength calibration"
if {q1} .ne. 0 then
  mess/out 1 {procID} "Product header initialization failed!"
  clean/pipeline xmos,spec
  return 1
endif


! Initialize longlog
init/longlog {p1} {p2} 1
if {q1} .ne. 0 then
  errmsg = "Longlog initialization failed!"
  mess/out 1 {procID} "{errmsg}"
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Search for overexposed pixels
@p fors_paragraph 2 0
write/out
mess/out 3 {procID} "Searching for overexposed pixels ..."
write/out {separator}
errno = errno + 1

@p fors_gen_qc0_overexpo XSFinput.cat + XSFinput.cat,XSFresult.cat
if {q2} .eq. 0 then
  mess/out 1 {procID} "Calibration frame rejected by overexposure check."
  errmsg = "No frames left in input. Execution of reduction block stopped!"
  mess/out 1 {procID} "{errmsg}"
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
else if {q1} .gt. 0 then
  errmsg = "FORS_GEN_QC0_OVEREXPO terminated with error status!"
  mess/out 1 {procID} "{errmsg}"
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Perform wavelength calibration
write/out
mess/out 3 {procID} "Perform wavelength calibration ..."
write/out {separator}
errno = errno + 1

calibwave/mos XSFinput.cat + XSFinput.cat,XSFresult.cat {obsmod}
if {q1} .ne. 0 then
  errmsg = "CALIBWAVE/MOS terminated with error status!"
  mess/out 1 {procID} "{errmsg}"
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Determine effective spectral resolution
write/out
mess/out 3 {procID} "Determine effective spectral resolution ..."
write/out {separator}
errno = errno + 1

@p fors_mos_qc1_resol XSFinput.cat + XSFinput.cat,XSFresult.cat {obsmod}
if {q1} .eq. 1 then
  errmsg = "FORS_MOS_QC1_RESOL terminated with error status!"
  mess/out 1 {procID} "{errmsg}"
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif

  
! Lookup main product file in catalog
errno = errno + 1

@p fors_paragraph 3 0
search/icat XSFresult {DISPREL_ID} NB
if m$len(out_a) .ne. 0 then
  profrm = out_a
else
  errmsg = "Product frame ({DISPREL_ID}) not found in product catalog!"
  mess/out 1 {procID} "{errmsg}"
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Save product files
errno = errno + 1


! Create product header
finish/prohead XSFresult.cat 1


finish/pipeline XSFresult.cat {p3} xmos,spec {profrm} {CAT_PLOT}
if {q1} .ne. 0 then
  errmsg = "Error leaving pipeline environment!"
  mess/out 1 {procID} "{errmsg}"
  finish/longlog XSFinput.cat -1 {profrm} 1 {p3}
  clean/pipeline xmos,spec
  write/error {errno},app,no "{errmsg}"
endif


! Close longlog
finish/longlog XSFinput.cat XSFresult.cat {profrm} 1 {p3}


! Disconnect from pipeline
log/rec "{recID}" OUT SUPER

@p fors_paragraph 4

return 0
