/* $Id: mosnflat.c,v 1.10 2000/02/08 11:29:26 sbogun Exp $   
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 *.COPYRIGHT   (C) 1999 European Southern Observatory
 *
 *.IDENT       $RCSfile: mosnflat.c,v $
 *
 *.KEYWORDS    Flat field normalization
 *
 *.USAGE       run mosnflat.exe
 *
 *.INPUT       FFRM     Name of flat field frame.
 *             STBL     Name of slit position table.
 *             METH     List of smoothing methods.
 *             FFPAR    Parameter values for smoothing methods.
 *             CHAT     Control flag for amount of terminal output.
 *
 *.OUTPUT      NFRM     Name of normalized flat field frame.
 *
 *.RETURNS     The program returns 0 on success or 1 on error. The exit
 *             status is stored in keyword Q1.
 *
 *.PURPOSE     The program normalizes the input flat field by smoothing the
 *             frame along the slit direction using one of the supported
 *             methods (AVERAGE, SMOOTH and MEDIAN). Then this result is
 *             smoothed along the dispersion axis using one the following
 *             methods: POLYNOM, SMOOTH and MEDIAN. The method AVERAGE
 *             averages the frame columns and provides a one dimensional
 *             result and is therefore the only method for the y axis which
 *             may be combined with the option POLYNOM for the x axis. One
 *             should also be aware that the amount of required cpu time
 *             for the median filter may be large depending on the filter
 *             size.
 *             The input flat field is finally devided by the smoothed
 *             counterpart.
 *
 *.COMMENT     The slit position table STBL has to be sorted by ascending
 *             :YSTART/:YEND columns !!!
 *
 *.ENVIRON     MIDAS 98NOV, XMOS context
 *
 *.LANGUAGE    ANSI C/KR C
 *
 *.AUTHOR      R. Palsa, ESO-DMO
 *
 *.VERSION     0.1 1998/12/18: Creation, RP
 *             0.2 1999/09/27: little changes (transpose, smooth), JKR
 *                             MOS-mode successfully tested
 *----------------------------------------------------------------------------
*/


#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <ok.h>
#include <midas_def.h>
#include <tbldef.h>
#include <proto_fors.h>


#define FRM_MAX_AXIS  2
#define MAX_SLITS     100

#define FX_POLYNOM  01
#define FX_SMOOTH   02
#define FX_MEDIAN   04

#define FY_AVERAGE  16
#define FY_SMOOTH   32
#define FY_MEDIAN   64

#define MID_PAR_SIZE    100
#define MSG_BUFFER_SIZE 81

#define MAX(A,B) ((A)  > (B) ? (A) : (B))
#define MIN(A,B) ((A)  < (B) ? (A) : (B))


typedef struct frame {
  int naxis;
  int ndata;
  int npix[FRM_MAX_AXIS];
  double start[FRM_MAX_AXIS];
  double step[FRM_MAX_AXIS];
  float *data;
} Frame;


typedef struct slits {
  int nslits;
  int slitno[MAX_SLITS];
  int end[MAX_SLITS];
  int start[MAX_SLITS];
} Slits;


#ifdef __STDC__
typedef float (*Filter)(int, float *);
#else
typedef float (*Filter)();
#endif


#ifdef __STDC__
void error(char *name, char *msg)
#else
     void error(name, msg)
     char *name, *msg;
#endif
{

  int unit;
  char s[2 * MSG_BUFFER_SIZE - 1];

  sprintf(s, "%s: Error: %s", name, msg);
  SCTPUT(s);
  SCKWRC("Q1", 1, "1", 1, 60, &unit);
  SCSEPI();
}


/* 
 * Nearest integer
 */

#ifdef __STDC__
static int nint(float x)
#else
     static int nint(x)
     float x;
#endif
{
  int y;
 
  if (x >= 0)
    y = (int)(x+0.5);
  else
    y = (int)(x-0.5);

  return(y);
}


/*
 * Simple smoothing filter (running mean)
 */

#ifdef __STDC__
float mean(int n, float *v)
#else
     float mean(n, v)
     int n;
     float *v;
#endif
{
  int i;
  float sum;


  
  sum = 0.;

  for (i = 0; i < n; i++)
    sum += *(v + i);

  return sum / (float)n;

}


/*
 * Average frame rows
 */

#ifdef __STDC__
int average(float *rfrm, float *afrm, int nx, int iy, int width)
#else
     int average(rfrm, afrm, nx, iy, width)
     float *rfrm, *afrm;
     int nx, iy, width;
#endif
{

  int i, j;

  
  
  for (j = 0; j < nx; j++)
    *(afrm + j) = 0.;

  for (j = 0; j < nx; j++) {
    for (i = iy; i < iy + width; i++)
      *(afrm + j) += *(rfrm + i * nx + j); 

    *(afrm + j) /= width;
  }

  return 0;

}


/*
 * Median filtering of one or two dimensional framesalong the rows
 */

#ifdef __STDC__
int smooth(float *frm, int nx, int ny, int width, int fwidth, Filter filt)
#else
     int smooth(frm, nx, ny, width, fwidth, filt)
     float *frm;
     int   nx, ny, width, fwidth;
     Filter filt;
#endif
{

  int i, j, start, xoffset;
  float *frow;



  if (ny == 0)
    ny = 1;

  if (width == 0)
    width = nx;


  /*
   * Assure that the filter size is a odd number
   */

  start  = fwidth / 2;
  fwidth = 2 * start + 1;


  /*  printf("smooth: nx=%d, ny=%d, width=%d\n", nx,ny,width);
   */

  if ((frow = (float *)osmmget((2 * start + width) * sizeof(float))) == NULL)
    return 1;

  for (j = 0; j < ny; j++) {
    for (i = 0; i < start; i++) {
      xoffset = j * nx;
      *(frow + i) = *(frm + xoffset+ start - i);
      *(frow + start + width + i) = *(frm + xoffset + width - (i + 2));
    }

    memcpy(frow + start, frm + j * nx, width * sizeof(float));

    for (i = 0; i < width; i++)
      *(frm + j * nx + i) = filt(fwidth, frow + i);
  }

  osmmfree(frow);

  return 0;

}


/*
 * Transpose a two dimensional array. Original array is overwritten!
 */


#ifdef __STDC__
int transpose(float *a, int nx, int ny)
#else
     int transpose(a, nx, ny)
     float *a;
     int   nx, ny;
#endif
{
  int i, j, k, l;
  float *b, tmp;

  if (nx < 0 || ny < 0)
    return 1;

  if (ny == 0 || ny == 1) 
    return 0;

  if (nx != ny)
    if ((b = (float *)osmmget(nx * ny * sizeof(float))) == 0)
      return 1;
    else {
      for (i = 0; i < ny; i++)
	for (j = 0; j < nx; j++) {
	  k = i * nx + j;
	  l = j * ny + i;

	  *(b + l) = *(a + k);
	}

      memcpy(a, b, nx * ny * sizeof(float));
      osmmfree(b);
    }
  else 
    for (i = 0; i < nx - 1; i++)
      for (j = i + 1 ; j < nx; j++) {
	k = i * nx + j;
	l = j * nx + i;
	    
	tmp = *(a + k);
	*(a + k) = *(a + l);
	*(a + l) = tmp;
      }

  return 0;

}




/*
 * Normalize input frame. Input frame is overwritten!
 */

#ifdef __STDC__
int normalize(Frame *flat, Slits *ypos, int xwin, int ywin, int method)
#else
     int normalize(flat, ypos, xwin, ywin, method)
     Frame *flat;
     Slits *ypos;
     int   xwin, ywin, method;
#endif
{
  int i, j, k;
  int ystart, ywidth;
  int ywin0;
  int offset;

  float *xp;
  float *fbuf0, *fbuf1;

  double *ap;

  Filter filter;

  /* Store y filter size */
  ywin0 = ywin;

  /*
   * Create temporary buffers for polynomial fit
   */

  if (method & FY_AVERAGE) {
    if ((fbuf0 = (float *)osmmget(flat->npix[0] * sizeof(float))) == NULL)
      return 1;

    if (method & FX_POLYNOM) {
      xwin++;

      xp = vector(0, flat->npix[0]);
      ap = dvector(1, xwin);

      for (i = 1; i <= flat->npix[0]; i++)
	xp[i] = (float)i;

      if ((fbuf1 = (float *)osmmget(flat->npix[0] * sizeof(float))) == NULL) {
	osmmfree(fbuf0);
	return 1;
      }
    }

    for (k = 0; k < ypos->nslits; k++) {
      ywidth = ypos->end[k] - ypos->start[k] + 1;
      ywin = ywin0;

      if (ywin > 0 && ywin < ywidth) {
	if (ywidth % 2 == 1)
	  ywin = 2 * (ywin / 2) + 1;
	else
	  ywin = 2 * (ywin / 2);

	ystart = ypos->start[k] + (ywidth - ywin) / 2;
      }
      else {
	ystart = ypos->start[k];
	ywin = ywidth;
      }


      /*
       * Average slit rows and smooth the obtained 1-dimensional frame
       */

      if (average(flat->data, fbuf0, flat->npix[0], ystart - 1, ywin)) {
	if (method & FX_POLYNOM) {
	  osmmfree(fbuf1);
	  free_vector(xp, 0, flat->npix[0]);
	  free_dvector(ap, 1, xwin);
	}

	osmmfree(fbuf0);
	return 1;
      }

      if (method & FX_POLYNOM)
	fit_poly(fbuf0, xp, flat->npix[0], fbuf1, xp, flat->npix[0], xwin, ap);
      else {
	if (method & FX_SMOOTH)
	  filter = mean;
	else 
	  filter = heap_median;

	smooth(fbuf0, flat->npix[0], 0, 0, xwin, filter);
	fbuf1 = fbuf0;
      }

      for (i = ypos->start[k]; i < ypos->end[k]; i++)
	for (j = 0; j < flat->npix[0]; j++)
	  if (*(fbuf1 + j) != 0.)
	    *(flat->data + i * flat->npix[0] + j) /= *(fbuf1 + j);
	  else 
	    *(flat->data + i * flat->npix[0] + j) = 1.;
    }

    osmmfree(fbuf0);
    if (method & FX_POLYNOM) {
      osmmfree(fbuf1);
      free_vector(xp, 0, flat->npix[0]);
      free_dvector(ap, 1, xwin);
    }
  }
  else {
    if ((fbuf0 = (float *)osmmget(flat->ndata * sizeof(float))) == NULL)
      return 1;
    else 
      memcpy(fbuf0, flat->data, flat->ndata * sizeof(float));


    /*
     * Smooth slit columns with selected filter by smoothing the rows
     * of the rotated frame.
     */

    if (method & FY_SMOOTH)
      filter = mean;
    else 
      filter = heap_median;

    transpose(fbuf0, flat->npix[0], flat->npix[1]);

    for (k = 0; k < ypos->nslits; k++) {
      ywidth = ypos->end[k] - ypos->start[k] + 1;
      ywin = ywin0;

      if (ywin < 0 || ywin > ywidth)
    	ywin = ywidth;
      
      /*      printf("slit.no: %d\n", k);
       */

      smooth(fbuf0 + ypos->start[k],flat->npix[1], flat->npix[0], 
	     ywidth, ywin, filter);
    }

    transpose(fbuf0, flat->npix[1], flat->npix[0]);

    if (method & FX_SMOOTH)
      filter = mean;
    else 
      filter = heap_median;

    for (k = 0; k < ypos->nslits; k++) {
      smooth(fbuf0, flat->npix[0], ypos->end[k] - ypos->start[k] + 1, 0,
	     xwin, filter); 

      for (i = ypos->start[k]; i < ypos->end[k]; i++) 
	for (j = 0; j < flat->npix[0]; j++) {
	  offset = i * flat->npix[0] + j;
	  if (*(fbuf0 + offset) != 0.)
	    *(flat->data + offset) /= *(fbuf0 + offset);
	  else 
	    *(flat->data + offset) = 1.;
      
	}
    }

    osmmfree(fbuf0);
    
  }


  /*
   * Set pixel outside of slit limits to zero
   */

  for (i = 0; i < ypos->start[0]; i++)
    for (j = 0; j < flat->npix[0]; j++)
      *(flat->data + i * flat->npix[0] + j) = 0.;

  for (k = 1; k < ypos->nslits; k++)
    for (i = ypos->end[k - 1]; i < ypos->start[k]; i++)
      for (j = 0; j < flat->npix[0]; j++)
	*(flat->data + i * flat->npix[0] + j) = 0.;

  for (i = ypos->end[ypos->nslits - 1]; i < flat->npix[1]; i++)
    for (j = 0; j < flat->npix[0]; j++)
      *(flat->data + i * flat->npix[0] + j) = 0.;

  return 0;

}



void main() {

#ifdef __STDC__
  char proId[] = "mosnflat";
#else
  char proId[9];
#endif


  int nval, unit, null, idum;
  int i;
  int fid, nid, sid, snrow, slcno[3];
  int ffpar[2], sel;
  int method = 0, chat;

  char *s;
  char meth[20];
  char ffrm[MID_PAR_SIZE + 1], stbl[MID_PAR_SIZE + 1];
  char nfrm[MID_PAR_SIZE + 1];
  char msg[MSG_BUFFER_SIZE];

  Slits slits;

  Frame flat;

#ifndef __STDC__
  strcpy(proId, "mosnflat");
#endif

  /*
   * Connect to MIDAS
   */

  SCSPRO(proId);

  
  /*
   * Get keywords from environment
   */

  SCKGETC("FFRM", 1, MID_PAR_SIZE, &nval, ffrm);
  SCKGETC("NFRM", 1, MID_PAR_SIZE, &nval, nfrm);
  SCKGETC("STBL", 1, MID_PAR_SIZE, &nval, stbl);
  SCKGETC("METH", 1, 20, &nval, meth);
  SCKRDI("FFPAR", 1, 2, &nval, ffpar, &unit, &null);
  SCKRDI("CHAT", 1, 1, &nval, &chat, &unit, &null);


  switch (toupper(*meth)) {
  case 'A':
    method |= FY_AVERAGE;
    break;

  case 'S':
    method |= FY_SMOOTH;
    break;

  case 'M':
    method |= FY_MEDIAN;
    break;

  default:
    error(proId, "Invalid smoothing method!");
    break;
  }

  if ((s = strchr(meth, ',')) == NULL)
    method |= FX_MEDIAN;
  else 
    switch(toupper(*++s)) {
    case 'P':
      if (method & FY_SMOOTH || method & FY_MEDIAN)
	error(proId, "Method POLYNOM can only be used with option AVERAGE!");

      method |= FX_POLYNOM;
      break;

    case 'S':
      method |= FX_SMOOTH;
      break;

    case 'M':
      method |= FX_MEDIAN;
      break;

    default:
      error(proId, "Invalid smoothing method!");
      break;
    }
	
  if ((method & FY_SMOOTH || method & FY_MEDIAN) && ffpar[0] < 0)
    error(proId, "Invalid filter size!");

  if ((method & FX_SMOOTH || method & FX_MEDIAN) && ffpar[1] < 0)
    error(proId, "Invalid filter size!");


  /*
   * Read slit positions
   */

  if (TCTOPN(stbl, F_I_MODE, &sid)) {
    sprintf(msg, "Cannot read slit positions from %s!", stbl);
    error(proId, msg);
  }
  else 
    TCIGET(sid, &idum, &snrow, &idum, &idum, &idum);

  if (TCCSER(sid, ":SLIT", &slcno[0])) {
    sprintf(msg, "Column :SLIT not found in %s!", stbl);
    error(proId, msg);
  }
      
  if (TCCSER(sid, ":YSTART", &slcno[1])) {
    sprintf(msg, "Column :YSTART not found in %s!", stbl);
    error(proId, msg);
  }

  if (TCCSER(sid, ":YEND", &slcno[2])) {
    sprintf(msg, "Column :YEND not found in %s!", stbl);
    error(proId, msg);
  }

  slits.nslits = 0;
  for (i = 1; i <= snrow; i++) {
    TCSGET(sid, i, &sel);
    if (sel) {
      TCERDI(sid, i, slcno[0], &slits.slitno[slits.nslits], &null);
      TCERDI(sid, i, slcno[1], &slits.start[slits.nslits], &null);
      TCERDI(sid, i, slcno[2], &slits.end[slits.nslits], &null);
      slits.nslits++;
    }
  }

  TCTCLO(sid);


  /*
   * Read flat field frame
   */

  if (SCFOPN(ffrm, D_R4_FORMAT, 0, F_IMA_TYPE, &fid)) {
    sprintf(msg, "Cannot open flat field frame %s!", ffrm);
    error(proId, msg);
  }
  else {
    SCDRDI(fid, "NAXIS", 1, 1, &nval, &flat.naxis, &unit, &null);
    SCDRDI(fid, "NPIX", 1, 2, &nval, flat.npix, &unit, &null);
    SCDRDD(fid, "START", 1, 2, &nval, flat.start, &unit, &null);
    SCDRDD(fid, "STEP", 1, 2, &nval, flat.step, &unit, &null);
    if (flat.naxis == 1) {
      flat.npix[1] = 1;
      flat.start[1] = 1.;
      flat.step[1] = 1.;
    }

    flat.ndata = flat.npix[0] * flat.npix[1];

    if ((flat.data = (float *)osmmget(flat.ndata * sizeof(float))) == NULL)
      error(proId, "Not enough memory!");

    if (SCFGET(fid, 1, flat.ndata, &nval, (char *)flat.data)) {
      osmmfree(flat.data);
      sprintf(msg, "Cannot read flat field frame %s!", ffrm);
      error(proId, msg);
    }
  }
  

  if (chat) {
    sprintf(msg, "%s: Normalizing flat field %s for %3d out of %3d slitlets.",
	    proId, ffrm, slits.nslits, snrow);
    SCTPUT(msg);
    sprintf(msg, "%s: Normalization method: %s.", proId, meth);
    SCTPUT(msg);

    if (method & FY_AVERAGE) {
      sprintf(msg, "%s: Averaging window (0: full slit width): %d", 
	      proId, ffpar[0]);
      SCTPUT(msg);

      if (method & FX_POLYNOM) {
	sprintf(msg, "%s: Order of polynomial fit: %d", proId, ffpar[1]);
	SCTPUT(msg);
      }
      else {
	sprintf(msg, "%s: Filter size along x-axis: %d", proId, ffpar[1]);
	SCTPUT(msg);
      }
    }
    else {
      sprintf(msg, "%s: Filter size along y-axis: %d", proId, ffpar[0]);
      SCTPUT(msg);      
      sprintf(msg, "%s: Filter size along x-axis: %d", proId, ffpar[1]);
      SCTPUT(msg);
    }
  }


  /*
   * Convert slit positions from detector pixel to frame pixel
   */

  for (i = 0; i < slits.nslits; i++) {
    slits.start[i] = nint((slits.start[i] - flat.start[1]) / flat.step[1]);
    slits.end[i] = nint((slits.end[i] - flat.start[1]) / flat.step[1]);

    if (slits.start[i] < 0)
      slits.start[i] = 0;

    if (slits.end[i] >= flat.npix[1])
      slits.end[i] = flat.npix[1]-1;
  }


  if (normalize(&flat, &slits, ffpar[1], ffpar[0], method)) {
    osmmfree(flat.data);
    error(proId, "Flat field normalization failed!");
  }


  /*
   * Write normalized flat field. Restore standard descriptors.
   */

  if (SCFCRE(nfrm, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, flat.ndata, &nid)) {
    osmmfree(flat.data);
    sprintf(msg, "Cannot create normalized flat field %s!", nfrm);
    error(proId, msg);
  }
  else {
    SCDWRI(nid, "NAXIS", &flat.naxis, 1, 1, &unit);
    SCDWRI(nid, "NPIX", flat.npix, 1, 2, &unit);
    SCDWRD(nid, "START", flat.start, 1, 2, &unit);
    SCDWRD(nid, "STEP", flat.step, 1, 2, &unit);
    SCDWRC(nid, "IDENT", 1, " ", 1, 72, &unit);
    SCDCOP(fid, nid, 4, "CUNIT");
    SCDCOP(fid, nid, 4, "LHCUTS");
    SCDCOP(fid, nid, 3, NULL); 

    if (SCFPUT(nid, 1, flat.ndata, (char *)flat.data)) {
      osmmfree(flat.data);
      sprintf(msg, "Cannot write normalized flat field %s!", nfrm);
      error(proId, msg);
    }

  }

  SCFCLO(fid);
  SCFCLO(nid);

  osmmfree(flat.data);
  SCKWRC("Q1", 1, "0", 1, 60, &unit);


  /*
   * Disconnect from MIDAS
   */

  SCSEPI();

}
